package com.ibm.oti.appletviewer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.applet.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.*;
import java.util.Enumeration;
import java.util.Vector;

import com.ibm.oti.awt.AppletAudioClip;
import com.ibm.oti.awt.MessageStrings;

/**
 * An instance of AppletViewer represents the entire appletviewer
 * application, and implements AppletContext, providing the context
 * for the applets being viewed.  An AppletViewer instance deals
 * with only one HTML page which may contain multiple applets.  Each
 * applet is displayed in its own window.
 *
 * @author		IBM
 * @version		initial
 */
public class AppletViewer implements AppletContext {

	private static final boolean debug = true;

	/**
	 *  the default size for applet windows
	 */
	protected final int DEF_WIN_SIZE = 400;

	/**
	 *  the applet stubs of the applets existing in this context
	 */
	protected Vector appletStubs = new Vector();

	/**
	 *  the URL of the current document
	 */
	protected URL documentURL;

	/**
	 *  whether this instance should be treated as a full application
	 */
	protected boolean runAsApp = false;

	boolean windowsOpen = false;

	boolean appletLoadingProblem = false;

	static Image iconImage;

	public static class MsgDialog extends Frame implements ActionListener {

		public MsgDialog(String msg) {
			super(MessageStrings.getString("BBAWTIBM_AppletViewer") + " " + MessageStrings.getString("BBAWTMessage")); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			setIconImage(AppletViewer.getIconImage());
			add(new TextArea(msg), BorderLayout.NORTH);
			Button button = new Button(MessageStrings.getString("BBAWTOkButton")); //$NON-NLS-1$
			button.addActionListener(this);
			add(button, BorderLayout.SOUTH);
			pack();
			//setResizable(false);
			new WindowCloser(this, false);
			show();
		}

		public void actionPerformed(ActionEvent e) {
			dispose();
		}
	}

	/**
	 * Constructs an AppletViewer on a particular document.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param	documentURL		The URL of the document whose applets are
	 *					to be viewed.
	 */
	public AppletViewer(URL documentURL, boolean runAsApp) {
		this.documentURL = documentURL;
		this.runAsApp = runAsApp;
	}

	/**
	 * Adds an a stub for an applet to this applet context's list
	 * of applet stubs, and sets the stubs context to be this.
	 * This accomplishes the addition of an applet to the context.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param	stub		The stub to be added.
	 */
	public void addAppletStub(ViewerAppletStub stub) {
		stub.setAppletContext(this);
		appletStubs.add(stub);
	}

	/**
	 * Retrieves, by name, one of the other applets occurring in this
	 * applet context.  If more than one applet with the specified name exist, one
	 * of them is return.  If no applet with the specified name exists, this method
	 * returns <code>null</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param	name	the name of the applet to be retrieved (this is usually
	 *					specified as a parameter in the HTML <code>applet</code>
	 *					tag)
	 *
	 * @return	A reference to an applet with the specified name occurring in the
	 *			context, or <code>null</code> if no such applet exists.
	 *
	 */
	public Applet getApplet(String name) {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			if (name.equals(stub.getName()))
				return stub.getApplet();
		}
		return null;
	}

	/**
	 * Returns an Enumeration of all the applets occurring in this applet context.
	 * The references in the Enumeration are of type <code>Applet</code>.  The
	 * Enumeration may contain <code>null</code> references.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	A non-<code>null</code> Enumeration of all the applets occurring in
	 *			this applet context. Some elements in the Enumeration may be <code>null</code>.
	 *
	 *
	 */
	public Enumeration getApplets() {
		return new Enumeration() {
			int idx = 0;
			public boolean hasMoreElements() {
				return idx < appletStubs.size();
			}
			public Object nextElement() {
				ViewerAppletStub stub =
					(ViewerAppletStub) appletStubs.get(idx++);
				return stub.getApplet();
			}
		};

	}

	/**
	 * Retrieves an AudioClip object representing the audio clip located at the URL
	 * <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the non-<code>null</code> absolute URL from which the
	 *						audio clip should be loaded
	 *
	 * @return		an AudioClip object representing the audio clip to be retrieved
	 */
	public AudioClip getAudioClip(URL url) {
		return new AppletAudioClip(url);
	}

	/**
	 * Retrieves an Image object representing the image file located at the URL
	 * <code>url</code>.  This method always returns immediately, and it is only
	 * when the applet attempts to render the image that the image data is loaded.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the absolute URL from which the image should be loaded
	 *
	 * @return		an Image object representing the image to be retrieved
	 */
	public Image getImage(URL url) {
		return Toolkit.getDefaultToolkit().getImage(url);
	}

	/**
	 * Requests that a specified document be displayed in place of the document
	 * currently being displayed within the applet context. Such requests may be
	 * ignored by the applet context.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param	url		The non-<code>null</code> URL of the document to be displayed.
	 *
	 */
	public void showDocument(URL url) {
	}

	/**
	 * Requests that a specified document be displayed in a specified target window
	 * or HTML frame.  The <code>target</code> parameter specifies the name of the
	 * frame or window {} the frame or window usually acquires this name in the HTML or
	 * script code underlying the documents being displayed in the applet context.
	 * Such requests may be ignored by the applet context.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param	url		The non-<code>null</code> URL of the document to be displayed.
	 * @param   target	The name of the frame or window in which to display the document.
	 */
	public void showDocument(URL url, String target) {
	}

	/**
	 * Requests that the string <code>msg</code> be displayed in the 'status
	 * window'.  Many environments in which an applet can run provide a 'status
	 * window', which is an area always available for the display of text messages.
	 * The environment may ignore this request.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		status		the text to be displayed in the status window.
	 *						If <code>null</code> the previous message is cleared.
	 */
	public void showStatus(String status) {
		int size = appletStubs.size();
		for (int i = 0; i < size; i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			stub.getFrame().showStatus(status);
		}
	}

	/**
	 * Returns the document base URL using the URL of this document.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		the document base URL for the document being
	 *				examined by the AppletViewer
	 */
	protected URL determineDocumentBase(URLConnection connection)
		throws MalformedURLException {
		// Get the document base from the connection
		// since it could have changed due to a redirect

		String documentBaseString = connection.getURL().toString();
		return new URL(documentBaseString);
	}

	/**
	 * Uses the document base URL and a possibly relative
	 * codebase string parameter to determine the absolute code base URL.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		The absolute code base url corresponding to the
	 *				codebase string provided
	 */
	protected URL determineCodeBase(String codeBase, URLConnection connection)
		throws MalformedURLException {
		return determineCodeBase(documentURL, codeBase, connection);
	}

	/**
	 * Allows the use of an alternate document base to guess what the code base could be.
	 * bug 110639
	 */
	protected URL determineCodeBase(URL alternateDocumentURL, String codeBase, URLConnection connection)
		throws MalformedURLException {
		if (codeBase != null) {
			codeBase = codeBase.trim();
			// The codebase string should always refer to a directory.
			// We ensure that it ends in a slash so that future calls to
			// URL(codebaseURL, file) treat the codebase URL as a directory.
			if (codeBase.length() > 0) {
				if (codeBase.charAt(codeBase.length() - 1) != '/'
					&& codeBase.charAt(codeBase.length() - 1) != '\\') {
					codeBase += "/"; //$NON-NLS-1$
				}
			}
		} else {
			// no codebase string specified: the code base in the connection
			codeBase = connection.getURL().toString();
			int lastSlash = codeBase.lastIndexOf('\\');
			if (lastSlash == -1) lastSlash = codeBase.lastIndexOf('/');
			codeBase = codeBase.substring(0, lastSlash + 1);
		}

		try {
			// Attempt to treat the codeBase string as an absolute URL:
			return new URL(codeBase);
		} catch (MalformedURLException e) {
			// Treat the codeBase string as a path relative to the
			// document base:
			return new URL(alternateDocumentURL, codeBase);
		}
	}

	/**
	 * Retrieves the document, parses the applet tags, loads the
	 * applet classes and instantiates them, and provides windows for
	 * each applet, leaving them hidden.
	 *
	 * @author		IBM
	 * @version		initial
	 */
	public void loadApplets() {
		boolean tagFound = false;
		try {
			URLConnection connection = documentURL.openConnection();
			InputStream in = connection.getInputStream();
			Parser parser = new Parser(in);
			HTMLTag t = null;

			/*
			 * If we are currently constructing an applet and stub, or are
			 * parsing an applet's PARAMs, currentStub will reference this
			 * applet's stub.
			 */
			ViewerAppletStub currentStub = null;

			while ((t = parser.getNextTag()) != null) {
				if (t.getTagBase().equals("APPLET")) { //$NON-NLS-1$
					tagFound = true;
					if (t.isBeginTag() && currentStub == null) {
						try {
							currentStub = new ViewerAppletStub();
							addAppletStub(currentStub);

							currentStub.setDocumentBase(determineDocumentBase(connection));

							currentStub.setCode(t.getAttribValue("CODE")); //$NON-NLS-1$
							currentStub.setArchive(t.getAttribValue("ARCHIVE")); //$NON-NLS-1$
							currentStub.setObject(t.getAttribValue("OBJECT")); //$NON-NLS-1$

							//Determine and adjust the absolute codebase URL:
							String codeBase = t.getAttribValue("CODEBASE"); //$NON-NLS-1$
							currentStub.setCodeBase(determineCodeBase(codeBase, connection));

							currentStub.setName(t.getAttribValue("NAME")); //$NON-NLS-1$

							String width = t.getAttribValue("WIDTH"); //$NON-NLS-1$
							String height = t.getAttribValue("HEIGHT"); //$NON-NLS-1$
							int w = DEF_WIN_SIZE, h = DEF_WIN_SIZE;
							try {
								w = Integer.parseInt(width);
								currentStub.addParm("width", width); //JAG //$NON-NLS-1$
							} catch (NumberFormatException e) {
							}
							try {
								h = Integer.parseInt(height);
								currentStub.addParm("height", height); //JAG //$NON-NLS-1$
							} catch (NumberFormatException e) {
							}

							ViewerFrame window = new ViewerFrame(currentStub, runAsApp, w, h);
							currentStub.setFrame(window);
							showStatus(MessageStrings.getString("BBAWTLoadingAppletInURL") + " " + documentURL); //$NON-NLS-1$ //$NON-NLS-2$
							window.setVisible(true);
							windowsOpen = true;
							final ViewerAppletStub stub = currentStub;
							window.addWindowListener(new java.awt.event.WindowAdapter() {
								public void windowClosing(
									java.awt.event.WindowEvent ev) {
									_windowClosing(stub);
								}
							});

							//Ask the stub to load and instantiate the applet:
							try {
								currentStub.loadApplet();
							} catch (ClassNotFoundException classNotFoundException) {
								// check if the class may be found in the alternate document base - bug 110639
								if (isAmbiguousDocumentBase()) {
									// alternate document base taken into account for bug 110639
									String alternateDocumentBase = currentStub.getDocumentBase().toString()+"/";
									// alternate code base taken into account for bug 110639
									URL alternateCodeBase = determineCodeBase(new URL(alternateDocumentBase), codeBase, connection);
									currentStub.setDocumentBase(new URL(alternateDocumentBase));
									currentStub.setCodeBase(alternateCodeBase);
									currentStub.loadApplet();
								} else {
									throw classNotFoundException;
								}
							}

							//Add the applet to its container:
							//window.addApplet(); //JAG
							currentStub.getApplet().setVisible(false);

							//currentStub.appletResize(w, h);
						} catch (MalformedURLException e) {
							msg(MessageStrings.getString("BBAWTInvalidAppletTagMsg")); //$NON-NLS-1$
						} catch (Throwable e) { //loadApplet() failed
							msg(MessageStrings.getString("BBAWTAppletLoadFailureMsg")); //$NON-NLS-1$
							if (debug) {
								System.out.println("Debug info:  loadApplet() failed and threw throwable:"); //$NON-NLS-1$
								e.printStackTrace();
							}
						}
					} else //i.e. !t.isBeginTag()
						currentStub = null;
				} else if (t.getTagBase().equals("PARAM")) { //$NON-NLS-1$
					if (currentStub != null) {
						String name = t.getAttribValue("NAME"); //$NON-NLS-1$
						String value = t.getAttribValue("VALUE"); //$NON-NLS-1$
						if (name != null && value != null) {
							currentStub.addParm(name, value);
						}
					}
				}
			}
			if (!tagFound) {
				appletLoadingProblem = true;
				println(MessageStrings.getString("BBAWTNoAppletTagsFoundMsg")); //$NON-NLS-1$
			}
		} catch (IOException e) {
			appletLoadingProblem = true;
			println(MessageStrings.getString("BBAWTErrorReadingHTMLMsg")); //$NON-NLS-1$
		}
	}

	// for bug 110639
	private boolean isAmbiguousDocumentBase() {
		if (documentURL == null || !"http".equals(documentURL.getProtocol()))
			return false; // if it's not http, documentURL refers to a file
		int urlLength = documentURL.toString().length();
		if (urlLength == 0 || documentURL.toString().charAt(urlLength-1) == '/')
			return false; // if it is slash terminated, we know that this is a directory
		if (documentURL.toString().indexOf('?') != -1)
			return false; // there are GET parameters, so it's a cgi script and not a directory
		// we don't know if this is a directory or a file
		return true;
	}

	static void println(String msg) {
		System.out.println(msg);
	}

	void msg(String msg) {
		if (windowsOpen) {
			showStatus(msg);
		} else
			new MsgDialog(msg);
	}

	void _windowClosing(ViewerAppletStub appletStub) {

		Applet applet = appletStub.getApplet();
		if (applet == null)
			return;
		if (applet.isActive()) {
			appletStub.setActive(false);
			applet.stop();
		}
		applet.destroy();
	}

	/**
	 * Calls the init() method on all applets within the context, and
	 * shows their windows
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void initApplets() {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			if (!stub.wasDeserialized()) {
				Applet applet = (Applet) stub.getApplet();
				if (applet != null) {
					stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusInitializingApplet")); //$NON-NLS-1$
					applet.init();
					stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusAppletInitialized")); //$NON-NLS-1$
				}
			}
		}
	}

	/**
	 * Calls the start() method on all applets within the context
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void startApplets() {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			Applet applet = (Applet) stub.getApplet();
			if (applet != null) {
				stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusStartingApplet")); //$NON-NLS-1$
				stub.setActive(true);
				applet.start();
				stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusAppletStarted")); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Sets the applets visible.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void showApplets() {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			Applet applet = (Applet) stub.getApplet();
			if (applet != null) {
				stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusDisplayingApplet")); //$NON-NLS-1$
				applet.setVisible(true);
				stub.getFrame().showStatus(MessageStrings.getString("BBAWTStatusAppletDisplayed")); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Adds the applets to the frame.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void addApplets() {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			Applet applet = (Applet) stub.getApplet();
			if (applet != null) {
				ViewerFrame window;
				window = stub.getFrame();
				window.addApplet();
				applet.resize(applet.getSize());
			}

		}
	}

	/**
	 * Revalidates the applets between init() and start().
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void revalidateApplets() {
		for (int i = 0; i < appletStubs.size(); i++) {
			ViewerAppletStub stub = (ViewerAppletStub) appletStubs.get(i);
			Applet applet = (Applet) stub.getApplet();
			if (applet != null) {
				applet.resize(applet.getSize());
				applet.invalidate();
				applet.getParent().validate();
			}

		}
	}

	/**
	 * Creates an AppletViewer instance, then loads, initializes, and
	 * starts its applets
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		docURL	the URL of the document whose
	 *						applets are to be viewed
	 *
	 *				runAsApp	whether the instance should be treated as a
	 *							full application
	 */
	public static void createAndRunViewer(final URL docURL, final boolean runAsApp) {
		//System.setSecurityManager(new SecurityManager());
		System.setSecurityManager(new AppletSecurity());
		final AppletViewer viewer = new AppletViewer(docURL, runAsApp);
		class ViewerThreadGroup extends ThreadGroup {
			public ViewerThreadGroup(String name) {
				super(name);
			}
			public void uncaughtException(Thread t, Throwable e) {
				if (e instanceof ThreadDeath) {
					super.uncaughtException(t, e);
					return;
				}
				String message = "Uncaught exception in thread " + t.getName() + " :" + e; //$NON-NLS-1$ //$NON-NLS-2$
				if (debug) {
					System.out.println(message);
					e.printStackTrace();
				}
				viewer.msg(message);
			}
		};
		final ViewerThreadGroup catcher = new ViewerThreadGroup("appletviewer"); //$NON-NLS-1$
		Runnable runnable = new Runnable() {
			public void run() {
				viewer.loadApplets();
				if (viewer.appletLoadingProblem)
					System.exit(-1);
				viewer.addApplets();
				viewer.initApplets();
				viewer.revalidateApplets();
				viewer.startApplets();
				viewer.showApplets();
			}
		};

		new Thread(catcher, runnable, catcher.getName()).start();

		//	EventQueue.invokeLater(runnable);
	}

	/**
	 * Returns an URL for the applet to load based on the program
	 * argument.  If the argument is not already an URL, then it will be
	 * treated like a local file if a file exists at the specified location.
	 *
	 * @param arg the applet argument to AppletViewer
	 * @return the URL to load the applet from
	 * @throws MalformedURLException if the argument is not a valid applet location
	 */
	static URL getURL(String arg) throws MalformedURLException {
		try {
			return new URL(arg);
		} catch (MalformedURLException e) {
			// If the file exists locally then that location should
			// be used.  Otherwise, it must have been an invalid URL.
			File f = new File(arg);
			if (f.exists()) {
				return new URL("file", "", arg);
			} else {
				throw e;
			}
		}
	}

	/**
	 * The main method used in order to run AppletViewer as an application.
	 * Creates one AppletViewer on a particular document.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		args	an array which should contain one String
	 *						element:  The URL of the document whose
	 *						applets are to be viewed
	 *
	 */
	public static void main(String[] args) {
		Thread.currentThread().getThreadGroup().setMaxPriority(5);

		if (args.length == 1) {
			try {
				createAndRunViewer(getURL(args[0]), true);
			} catch (MalformedURLException e) {
				printUsage();
			}
		} else {
			printUsage();
		}
	}

	private static void printUsage() {
		println(MessageStrings.getString("BBAWTMessageAppletviewerUsage")); //$NON-NLS-1$
	}

	public static void main(String arg) {
		main(new String[] { arg });
	}

	public static Image getIconImage() {
		if (iconImage == null) {
			iconImage = Toolkit.getDefaultToolkit().getImage(Window.class.getResource("wsdd.gif")); //$NON-NLS-1$
		}
		return iconImage;
	}
}
