package com.ibm.oti.appletviewer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

/**
 * HTMLTag parses an HTML tag, including its attribute values and
 * answers questions about its values
 *
 * @author      IBM
 */

public class HTMLTag {

	/**
	 *  the tag's base type ie <IMG SRC="..."> would be IMG
	 */
	protected String tagBase;

	/**
	 *  list of the tag's attributes
	 */
	protected Vector tagAttribs;

	/**
	 *  is it a beginning tag ie <P> or ending tag ie </FONT>
	 */
	protected boolean tagBeginning;

/**
 * constructor - takes the tag string and breaks it down
 * @param tag	the tag text
 */
public HTMLTag(StringBuffer tag) {
	// we should never get an empty tag
	if ((tag==null) || (tag.length()==0))
		return;
	// it is assumed that we're getting a valid tag, so we blindly
	// clip the first and last characters which should be the <>

	tag.setLength(tag.length()-1);
	tag.deleteCharAt(0);

	// if the first char is a '/' then we know it's an end tag
	// so clip off that character and set the tagBeginning flag
	if (tag.charAt(0)=='/') {
		tagBeginning=false;
		tag.deleteCharAt(0);
	}
	else
		tagBeginning=true;
	// get the tag's base type
	parseTagBase(tag);

	// if the tag is now empty, then we're finished
	if (tag.length()==0) return;

	// make an enumeration for the attribute and values
	tagAttribs=new Vector();

	// parse all the attribute and values from the tag
	while (tag.length()>0) {
		parseTagAttribute(tag);
	}
}

/**
 * getAttribValue returns the value of an attribute of the tag if it exists
 * otherwise it returns null
 * @param	key		key to search for
 * @return			the attribute's value
 */
protected String getAttribValue(String key) {
	if (tagAttribs==null)
		return null;
	Enumeration e=tagAttribs.elements();
	String[] element;
	while (e.hasMoreElements()) {
		element=(String[])e.nextElement();
		if (element[0].equals(key))
			return element[1];
	}
	return null;
}

/**
 * parseTagBase - extracts, removes and stores the tag's base type
 * @param tag	the tag text
 */
protected void parseTagBase(StringBuffer tag) {
	// find the first space in the StringBuffer
	for (int i=0;i<tag.length();i++)
		if (tag.charAt(i) == ' ') {
			// create a string containing the text up to the first space
			tagBase=tag.substring(0,i).toUpperCase();
			tag.delete(0,i+1);
			return;
		}
	// there was no space, so the entire tag is the base type
	tagBase=tag.toString().toUpperCase();
	tag.setLength(0);
	return;
}

/**
 * getTagBase - returns tag's base type (e.g. "IMG")
 * @return 	the tags base type string
 */
protected String getTagBase() {
	return tagBase;
}

/**
 * parseTagAttribute - extracts and removes all tag attributes, adding them
 * to the attributes enumeration
 * @author IBM
 * @param	tag		the remaining tag text
 */
protected void parseTagAttribute(StringBuffer tag) {

   String name;
   String value;

   StreamTokenizer t = new StreamTokenizer(new StringReader(tag.toString()));
   // Reset the syntax so that numbers are not
   // paresed differently than words.  The StreamTokenizer
   // does not parse hexidecimal or octal numbers correctly
   // so instead they will be parsed as words and detected later
   t.resetSyntax();
   t.wordChars('a', 'z');
   t.wordChars('A','Z');
   t.wordChars('0','9');
   t.wordChars(160, 255);
   t.whitespaceChars(0, 32);
   t.commentChar('/');
   t.quoteChar('"');
   t.quoteChar('\'');
   t.ordinaryChar('=');
   t.wordChars('/', '/');
   t.wordChars('_', '_');   // per HTML spec
   t.wordChars('.', '.');   // per HTML spec
   t.wordChars(':', ':');   // per HTML spec
   t.wordChars('-', '-');   // per HTML spec
   t.eolIsSignificant(false);

   int token;

   /*
    * Parse through all the attributes passed to us.
    */

   try{
      token = t.nextToken();
      while(token !=  StreamTokenizer.TT_EOF){

         // Find the attribute name

         if(token == StreamTokenizer.TT_WORD){
            name=t.sval;

            token = t.nextToken();

            /* look for an '=' sign */

            if(token == '='){
               token = t.nextToken();
	           value = t.sval;
            } else {

               /* Wasn't an '=' sign so may an attribute with no value */

               t.pushBack();
               value = ""; //$NON-NLS-1$
            }

            tagAttribs.add(new String[] {name.toUpperCase(),value});
         }
         token = t.nextToken();
      }
   } catch (IOException ex){
      System.err.println("HTMLTag parsing caught exception"); //$NON-NLS-1$
      ex.printStackTrace(System.err);
      tag.setLength(0);
      return;
   }

   /* We now parse it all, so tell the caller that we are done. */

   tag.setLength(0);
   return;

}
/**
 * answers whether this tag is the beginning of an element or the
 * end. Tags beginning with '/' are always end elements
 * @return	whether this tag is the beginning of an element or the end
 */
public boolean isBeginTag() {
	return tagBeginning;
}
}
