package com.ibm.oti.appletviewer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005 All Rights Reserved
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

import com.ibm.oti.awt.MessageStrings;

/**
 * A window which is intended to contain an applet. It provides a menu which
 * allows users to manipulate and query the applet.
 *
 * @author IBM
 * @version initial
 */
public class ViewerFrame extends Frame {

	/**
	 * The stub for the applet contained in this window
	 */
	protected ViewerAppletStub appletStub;

	/**
	 * The panel containing the applet
	 */
	protected Panel appletContainer;

	/**
	 * The label providing the status bar
	 */
	protected Label statusLabel;

	/**
	 * Causes the receiver's status bar to display a specified string.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param stub
	 *            String the stub for the applet to be contained in this window
	 */
	public void showStatus(String msg) {
		statusLabel.setText(msg);
	}

	/**
	 * Change the size of the applet and also resize the parent container to
	 * hold the applet's new size
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param width
	 *            new width for the applet
	 * @param height
	 *            new height for the applet
	 */
	public void resizeApplet(int width, int height) {
		Applet applet = appletStub.getApplet();
		Rectangle appletBounds = applet.getBounds();
		applet.setBounds(appletBounds.x, appletBounds.y, width, height);
		appletContainer.setSize(width, height);
	}

	/**
	 * Constructs a ViewerFrame and adds to it the applet referenced by <code>stub</code>
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param stub
	 *            the stub for the applet to be contained in this window
	 *
	 * exit whether the closing of this window should halt the VM
	 */
	public ViewerFrame(
		ViewerAppletStub stub,
		boolean exit,
		final int initialWidth,
		final int initialHeight) {
		appletStub = stub;

		String title = MessageStrings.getString("BBAWTIBM_AppletViewer"); //$NON-NLS-1$
		String name = appletStub.getName();
		if (name != null) {
			title = title + " - " + name; //$NON-NLS-1$
		}
		setTitle(title);

		setIconImage(AppletViewer.getIconImage());

		//add an event listener to deal with the closing of the window
		new WindowCloser(this, exit);

		//create the menubar, and add items to it
		MenuBar mb = new MenuBar();
		Menu menu = new Menu(MessageStrings.getString("BBAWTAppletMenu")); //$NON-NLS-1$
		MenuItem item = new MenuItem(MessageStrings.getString("BBAWTStartMenuItem")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				appletStub.startApplet();
			}
		});
		menu.add(item);

		item = new MenuItem(MessageStrings.getString("BBAWTStopMenuItem")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				appletStub.stopApplet();
			}
		});
		menu.add(item);

		item = new MenuItem(MessageStrings.getString("BBAWTSerializeToFileMenuItem")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				FileDialog dialog = new FileDialog(ViewerFrame.this);
				dialog.setMode(FileDialog.SAVE);
				dialog.setFile("Applet.ser"); //$NON-NLS-1$
				dialog.show();
				File file = new File(dialog.getDirectory() + dialog.getFile());
				try {
					FileOutputStream fout = new FileOutputStream(file);
					ObjectOutputStream ostream = new ObjectOutputStream(fout);
					ostream.writeObject(appletStub.getApplet());
					ostream.close();
				} catch (IOException ex) {
				}
			}
		});
		menu.add(item);

		item = new MenuItem(MessageStrings.getString("BBAWTAppletAndParmInfoMenuItem")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				InfoFrame infoFrame = new InfoFrame(appletStub);
				infoFrame.pack();
				infoFrame.show();
			}
		});
		menu.add(item);

		item = new MenuItem(MessageStrings.getString("BBAWTExitMenuItem")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//post the window-closing event:
				dispatchEvent(new WindowEvent(ViewerFrame.this, WindowEvent.WINDOW_CLOSING));
				//If this WINDOW_CLOSING event sends dispose() to the
				//window, dispose() will post a WINDOW_CLOSED event.
			}
		});
		menu.add(item);
		mb.add(menu);
		setMenuBar(mb);

		//Create a panel to contain the applet. This panel's preferred size
		//is its current size:
		appletContainer = new Panel() {
			public Dimension getPreferredSize() {
				Dimension currentSize = getSize();
				return new Dimension(initialWidth, initialHeight);
			}
		};
		appletContainer.setLayout(new BorderLayout());
		appletContainer.setSize(initialWidth, initialHeight);
		//Add the panel to the window:
		add(appletContainer, BorderLayout.CENTER);
		add(statusLabel = new Label(), BorderLayout.SOUTH);
		pack();
	}

	public void addApplet() {
		Applet applet = appletStub.getApplet();
		appletContainer.add(applet, BorderLayout.CENTER);
		invalidate();
		validate();
	}

	public void setAppletName(String name) {
		setTitle(
			MessageStrings.getString("BBAWTIBM_AppletViewer") + " - " + name); //$NON-NLS-1$ //$NON-NLS-2$
	}

}
