package com.ibm.oti.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.metal.graphics.Device;
import com.ibm.oti.awt.metal.image.ImagePeer;

public class NativeImageConsumer extends AbstractImageConsumer {

	int fWidth= -1;
	int fHeight= -1;
	ColorModel fNativeColorModel;

	// Transient state.
	private byte[] byteData;
	private int[] rgbData;

	public NativeImageConsumer(IImageImpl imageImpl) {
		super(imageImpl);
		fNativeColorModel = ColorModel.getRGBdefault();
	}

	public void setDimensions(int width, int height) {
		fWidth= width;
		fHeight= height;
		super.setDimensions(width, height);
	}

	private int getBytesPerLine(int width, int depth) {
		int pixelsPerByte = 8 / depth;
		int bytesPerLine = width / pixelsPerByte;

		// If the incoming width is odd and the depth is less than 8 then
		// we might need to add an extra byte per line to account for the
		// integer rounding error.
		if (width % pixelsPerByte > 0) bytesPerLine++;
		return bytesPerLine;
	}

	private void copyPixelData(byte[] pixels, int destinationOffset, int depth) {
		int sourceLength = pixels.length;

		if (byteData == null) {
			int bytesPerLine = getBytesPerLine(fWidth, depth);
			byteData = new byte[bytesPerLine * fHeight];
		}
		System.arraycopy(pixels, 0, byteData, destinationOffset, sourceLength);
	}

	private void copyPixelData(int[] pixels, int destinationOffset) {
		int sourceLength = pixels.length;

		if (rgbData == null) {
			rgbData = new int[fWidth * fHeight];
		}
		System.arraycopy(pixels, 0, rgbData, destinationOffset, sourceLength);
	}

	/**
	 * Returns whether the specified depth is supported by UGL
	 *
	 * @param depth the color depth
	 * @return true if the depth is supported by UGL, false otherwise
	 */
	boolean isSupportedDepth(int depth) {
		return (depth == 1 || depth == 2 || depth == 4 || depth == 8 ||
				depth == 16 || depth == 24 || depth == 32);
	}

	/**
	 * Returns the number of bits per primary color available to the image.
	 * <p>
	 * This is normally the same as the pixel size of the color model.
	 * If the color model uses an unsupported color depth of less than 8 bpp
	 * (e.g., 7 bpp) it is possible that the color resolution (or color
	 * precision) is still higher and a compressed palette is used.
	 * See bug 108793.
	 * </p>
	 *
	 * @param colorModel color model to get the color resolution from
	 * @return the number of bits per primary color available to the image
	 */
	int getColorResolution(ColorModel colorModel) {
		int[] colorResolutions = colorModel.getComponentSize();
		int colorResolution = 8;

		if (colorResolutions != null) {
			colorResolution = colorResolutions[0];
			for (int i = 1; i < colorResolutions.length; i++) {
				if (colorResolutions[i] > colorResolution) {
					colorResolution = colorResolutions[i];
				}
			}
			if (isSupportedDepth(colorResolution) == false) {
				colorResolution = 8;
			}
		}
		return colorResolution;
	}

	public void setPixels(int x, int y, int width, int height, ColorModel cm, int[] pixels, int offset, int scanlineWidth) {
		// color conversion needed only when not using UGL compatible color model
		if (Device.convertToRGB(cm)) {
			setExpandedPixels(x, y, width, height, pixels, offset, scanlineWidth, cm);
			super.setPixels(x, y, width, height, cm, pixels, offset, scanlineWidth);
		} else {
			fNativeColorModel = cm;
			setExpandedPixels(x, y, width, height, pixels, offset, scanlineWidth, null);
			super.setPixels(x, y, width, height, cm, pixels, offset, scanlineWidth);
		}
	}

	/**
	 * Sets the pixel data supplied in <code>pixels</code>. Pixels are converted
	 * to the internal linear byte buffer. If a color model is supplied it is used
	 * to convert pixel data to the default RGB model.
	 * Pixel data is always expanded from int to a byte array. Data is expected
	 * in ARGB format.
	 * <code>pixels</code> may specify a scanlineWidth that is wider than the
	 * image width. In that case extra pixels are ignored on each scan line as
	 * suggested by the ImageConsumer.setPixels spec.
	 */
	void setExpandedPixels(int x, int y, int width, int height, int[] pixels, int offset, int scanlineWidth, ColorModel colorModel) {
		int[] rgbPixels = new int[width * height];
		int bufferIndex = 0;

		for (int n = 0; n < height; n++) {
			int pixelIndex = (n * scanlineWidth) + offset;
			for (int m = 0; m < width; m++) {
				int rgb;
				if (colorModel != null) {
					rgb = colorModel.getRGB(pixels[pixelIndex]);
				} else {
					rgb = pixels[pixelIndex];
				}
				rgbPixels[bufferIndex++] = rgb;
				pixelIndex++;
			}
		}
		copyPixelData(rgbPixels, y * width + x);
	}

	public void setPixels(int x, int y, int width, int height, ColorModel cm, byte[] pixels, int offset, int scanlineWidth) {
		int depth = cm.getPixelSize();

		// color conversion needed only when not using UGL compatible color model
		if (Device.convertToRGB(cm)) {
			setExpandedPixels(x, y, width, height, pixels, offset, scanlineWidth, cm);
		} else if (isSupportedDepth(depth) == false) {
			int colorResolution = getColorResolution(cm);
			fNativeColorModel = cm;
			setExpandedPixels(x, y, width, height, pixels, offset, scanlineWidth, colorResolution);
		} else {
			fNativeColorModel = cm;
			setExpandedPixels(x, y, width, height, pixels, offset, scanlineWidth, depth);
		}
		super.setPixels(x, y, width, height, cm, pixels, offset, scanlineWidth);
	}

	/**
	 * Sets the pixel data supplied in <code>pixels</code>. Pixels are converted
	 * to the internal linear byte buffer. If a color model is supplied it is used
	 * to convert pixel data to the default RGB model.
	 * Pixel data is always expanded from int to a byte array. Data is expected
	 * in ARGB format.
	 * <code>pixels</code> may specify a scanlineWidth that is wider than the
	 * image width. In that case extra pixels are ignored on each scan line as
	 * suggested by the ImageConsumer.setPixels spec.
	 */
	void setExpandedPixels(int x, int y, int width, int height, byte[] pixels, int offset, int scanlineWidth, ColorModel colorModel) {
		int bufferIndex = 0;
		int[] rgbPixels = new int[width * height];

		for (int n = 0; n < height; n++) {
			int pixelIndex = (n * scanlineWidth) + offset;
			for (int m = 0; m < width; m++, pixelIndex++) {
				int rgb = colorModel.getRGB(pixels[pixelIndex]);
				rgbPixels[bufferIndex++] = rgb;
			}
		}
		copyPixelData(rgbPixels, y * width + x);
	}

	/**
	 * Sets the pixel data supplied in <code>pixels</code>. Pixels are converted
	 * to the internal linear byte buffer.
	 * Pixel data is always expanded from int to a byte array. Data is expected
	 * in ARGB format.
	 * <code>pixels</code> may specify a scanlineWidth that is wider than the
	 * image width. In that case extra pixels are ignored on each scan line as
	 * suggested by the ImageConsumer.setPixels spec.
	 */
	void setExpandedPixels(int x, int y, int width, int height, byte[] pixels, int offset, int scanlineWidth, int depth) {
		int bufferIndex = 0;
		int pixelsPerByte = 8 / depth;
		int bytesPerLine = getBytesPerLine(width, depth);
		int byteIndex = x % pixelsPerByte;	// index of the first pixel in a partially written byte
		int destinationOffset = y * bytesPerLine + x / pixelsPerByte;
		int newByteCount = bytesPerLine * height;
		byte[] bytePixels;

		// do the new pixels span a partial byte?
		if ((byteIndex > 0) && (x / pixelsPerByte != (x + width - 1) / pixelsPerByte)) {
			newByteCount++;
		}
		bytePixels = new byte[newByteCount];

		// preserve existing pixel(s) of first byte
		if (byteData != null && byteIndex > 0) {
			int bitIndex = byteIndex * depth;
			byte firstPixel = byteData[destinationOffset];
			int newBitCount = 8 - bitIndex;

			// push all bits that will be new off the clif
			firstPixel >>= newBitCount;
			firstPixel <<= newBitCount;
			bytePixels[0] = firstPixel;
		}

		for (int n = 0; n < height; n++) {
			int pixelIndex = (n * scanlineWidth) + offset;
			for (int m = 0; m < width; ) {
				int i = 8 - depth;

				if (bufferIndex == 0 && byteIndex > 0) {
					int bitIndex = byteIndex * depth;
					i -= bitIndex;
					byteIndex = 0;
				}
				for (; i >= 0 && m < width; i -= depth, m++, pixelIndex++) {
					byte pixel = pixels[pixelIndex];
					bytePixels[bufferIndex] += pixel << i;
				}
				bufferIndex++;
			}
		}

		// preserve existing pixel(s) of last byte
		if (byteData != null) {
			int lastPixelIndex = width * height + x;
			byteIndex = lastPixelIndex % pixelsPerByte;
			// if last pixel does not fill up last byte,
			// preserve existing pixel(s) of last byte
			if (byteIndex > 0) {
				int bitIndex = byteIndex * depth;
				byte lastPixel = byteData[destinationOffset + bytePixels.length - 1];

				// mask out what will be new pixels in the last byte
				for (int i = 8 - bitIndex; i < 8; i++) {
					byte bit = (byte) ~(1 << i);
					lastPixel &= bit;
				}
				bytePixels[bytePixels.length - 1] += lastPixel;
			}
		}
		copyPixelData(bytePixels, destinationOffset, depth);
	}

	ImagePeer createImagePeer() {
		ImagePeer peer;

		if (rgbData != null) {
			int depth = fNativeColorModel.getPixelSize();
			peer = ImagePeer.createMutableCopy(Util.getDisplay(), fWidth, fHeight, rgbData, depth);
		} else {
			byte[] reds;
			byte[] greens;
			byte[] blues;
			int transparentPixel = -1;
			int depth = fNativeColorModel.getPixelSize();

			if (fNativeColorModel instanceof IndexColorModel) {
				IndexColorModel indexColorModel = (IndexColorModel) fNativeColorModel;
				int mapSize = indexColorModel.getMapSize();

				reds = new byte[mapSize];
				greens = new byte[mapSize];
				blues = new byte[mapSize];
				indexColorModel.getReds(reds);
				indexColorModel.getGreens(greens);
				indexColorModel.getBlues(blues);
				transparentPixel = indexColorModel.getTransparentPixel();
			} else {
				int numColors = (int) Math.pow(2, depth);

				reds = new byte[numColors];
				greens = new byte[numColors];
				blues = new byte[numColors];
				for (int i = 0; i < numColors; i++) {
					reds[i] = (byte) fNativeColorModel.getRed(i);
					greens[i] = (byte) fNativeColorModel.getGreen(i);
					blues[i] = (byte) fNativeColorModel.getBlue(i);
				}
			}
			if (isSupportedDepth(depth) == false) {
				depth = getColorResolution(fNativeColorModel);
			}
			peer = ImagePeer.createMutableCopy(Util.getDisplay(), fWidth, fHeight, byteData, depth, reds, greens, blues, transparentPixel);
		}
		return peer;
	}
}
