package com.ibm.oti.awt.metal.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2006  All Rights Reserved
 */

import java.awt.ActiveEvent;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.IndexColorModel;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;

import com.ibm.oti.awt.metal.Handle;
import com.ibm.oti.awt.metal.NativeObject;
import com.ibm.oti.awt.metal.graphics.Device;
import com.ibm.ugl.eswt.OS;

public class ImagePeer extends NativeObject {
	transient HashSet activeGraphics = new HashSet();

	public Rectangle getBounds() {
		int[] rect = OS.Image_GetBounds(fPeerHandle);
		java.awt.Rectangle returnValue = new java.awt.Rectangle(rect[OS.INDEX_X],rect[OS.INDEX_Y],rect[OS.INDEX_WIDTH],rect[OS.INDEX_HEIGHT]);
		return returnValue;
	}

	public int getDepth(){
		return OS.Image_GetDepth(fPeerHandle);
	}

	public static ImagePeer createMutableCopy(Device device, int width, int height, byte[] imageData, int depth, byte[] indexedReds, byte[] indexedGreens, byte[] indexedBlues, int transparentPixel) {
		int bytesPerLine = imageData.length / height;
		int returnValue = OS.Image_NewFromByteData(device.fPeerHandle, width, height, imageData, bytesPerLine, depth, indexedReds, indexedGreens, indexedBlues, indexedReds.length, transparentPixel);
		return new ImagePeer(returnValue);
	}

	public static ImagePeer createMutableCopy(Device device, int width, int height, int[] imageData, int depth) {
		int returnValue = OS.Image_NewFromIntData(device.fPeerHandle, width, height, imageData, depth, -1, -1, null);
		return new ImagePeer(returnValue);
	}

	public void newGraphicsHandle(Handle handle) {
		handle.value = OS.Image_NewGraphics(fPeerHandle);
		synchronized (activeGraphics) {
			activeGraphics.add(handle);
		}
	}

	public void disposeGraphicsHandle(Handle handle) {
		synchronized (handle) {
			if (handle.value != 0) {
				OS.Graphics_Dispose(handle.value);
				handle.value = 0;
			}
		}
		synchronized (activeGraphics) {
			activeGraphics.remove(handle);
		}
	}

	public static ImagePeer createMutableCopy(Device device, String fileName) {
		// Fix for bug 106469.  A File object is created and
		// the absolute path is used to load the image because
		// relative paths should always be resolved based on
		// 'user.dir' which the native API will not have access to
		File imageFile = new File(fileName);
		String path;
		try {
			// Get the canonical path. Fixes bug 119098
			path = imageFile.getCanonicalPath();
		} catch (IOException exception) {
			path = imageFile.getAbsolutePath();
		}
		int returnValue = OS.Image_NewFromFile(device.fPeerHandle, path);
		if (returnValue == 0) return null;
		return new ImagePeer(returnValue);
	}

	public ImagePeer(int handle) {
		super(handle);
	}

	public static ImagePeer createImmutableCopy(Device device, int width, int height, Color fillColor) {
		int rgb = (fillColor != null ? fillColor.getRGB() : Color.black.getRGB());
		int returnValue = OS.Image_NewFromSize(device.fPeerHandle, width, height, rgb);
		return new ImagePeer(returnValue);
	}

	public static ImagePeer createMutableCopy(Device device, ImagePeer peer) {
		int returnValue = OS.Image_NewFromHandle(device.fPeerHandle, (peer == null ? 0 : peer.fPeerHandle));
		return new ImagePeer(returnValue);
	}

	public ColorModel getColorModel() {
		ColorModel colorModel = null;
		int depth = OS.Image_GetDepth(fPeerHandle);

		if (OS.Image_IsDirect(fPeerHandle)) {
			int[] masks = OS.Image_GetDirectPaletteMasks(fPeerHandle);
			if (masks != null) {
				colorModel = new DirectColorModel(depth, masks[OS.PALETTE_MASK_RED], masks[OS.PALETTE_MASK_GREEN], masks[OS.PALETTE_MASK_BLUE], masks[OS.PALETTE_MASK_ALPHA]);
			} // TODO: should throw exception if 'masks' is null?
		} else {
			int numColors = (int) Math.pow(2, depth);
			byte[] red = new byte[numColors];
			byte[] green = new byte[numColors];
			byte[] blue = new byte[numColors];
			int transparentPixel = OS.Image_GetTransparentColor(fPeerHandle);

			OS.Image_GetIndexedPalette(fPeerHandle, red, green, blue);
			colorModel = new IndexColorModel(depth, numColors, red, green, blue, transparentPixel);
		}
		return colorModel;
	}

	public static ImagePeer createMutableCopy(Device device, byte[] imageData) {
		if (imageData.length == 0) return null;
		int returnValue = OS.Image_NewFromFileData(device.fPeerHandle, imageData, imageData.length);
		if (returnValue == 0) return null;
		return new ImagePeer(returnValue);
	}

	public void setRGB(int x, int y, int rgb) {
		OS.Image_SetRGB(fPeerHandle, x, y, rgb);
	}

	public final void dispose() {
		if (fPeerHandle == 0) return;

		synchronized (this) {
			synchronized (activeGraphics) {
				Object[] graphics = activeGraphics.toArray();
				for (int i = 0; i < graphics.length; i++) {
					Handle handle = (Handle) graphics[i];
					disposeGraphicsHandle(handle);
				}
			}
		}

		OS.Image_Dispose(fPeerHandle);
		fPeerHandle = 0;
	}

	/*
	 * Returns the flipped sub-image specified by sx, sy, sw, sh.
	 */
	public ImagePeer flip(boolean hFlip, boolean vFlip, int sx, int sy, int sw, int sh) {
		int returnValue = OS.Image_Flip(fPeerHandle, hFlip, vFlip, sx, sy, sw, sh);
		return new ImagePeer(returnValue);
	}

	public int[] getPixelsHighColor() {
		return OS.Image_GetPixelsHighColor(fPeerHandle);
	}

	public byte[] getPixelsLowColor() {
	    Rectangle bounds = getBounds();
		int depth = getDepth();
	    int pixelsPerByte = 8 / depth;
	    byte pixelMask;
	    byte[] pixels = new byte[bounds.width * bounds.height];
		byte[] packedPixels = OS.Image_GetPixelsLowColor(fPeerHandle);
	    int bytesPerLine = packedPixels.length / bounds.height;

	    if (depth == 1) {
	    	pixelMask = 0x1;
	    } else if (depth == 2) {
	    	pixelMask = 0x3;
	    } else if (depth == 4) {
	    	pixelMask = 0xF;
	    } else if (depth == 8) {
	    	pixelMask = (byte) 0xFF;
	    } else {
	    	return packedPixels;
	    }

	    for (int y = 0 ; y < bounds.height; y++){
	        for (int x = 0; x < bounds.width; ){
				byte aByte = packedPixels[y * bytesPerLine + x / pixelsPerByte];
				for (int i = 8 - depth; i >= 0 && x < bounds.width; i -= depth, x++) {
					byte pixel = (byte) ((aByte >> i) & pixelMask);
					pixels[y * bounds.width + x] = pixel;
				}
	        }
	    }
		return pixels;
	}

	public void getRGB(int startX, int startY, int width, int height, int[] rgbs, int offset, int scanSize) {
		OS.Image_GetRGB(fPeerHandle, startX, startY, width, height, offset, scanSize, rgbs);
	}

	/*
	 * This is implemented using other createMutable methods
	 * as there is no direct native support for creating images
	 * from URLs.
	 */
	public static ImagePeer createMutableCopy(Device device, URL url) {
		byte[] data;
		try {
			byte[] buf = new byte[1024];

			InputStream input = url.openStream();
			try {
				ByteArrayOutputStream bout = new ByteArrayOutputStream();

				while (true) {
					int read = input.read(buf);
					if (read == -1)
						break;
					bout.write(buf, 0, read);
				}
				data = bout.toByteArray();
			} finally {
				input.close();
			}
		} catch (IOException ex) {
			// data = new byte[0];
			return null;
		}
		return createMutableCopy(device, data);
	}

}
