package com.ibm.oti.awt.metal.sound;

import java.io.*;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Vector;

import com.ibm.oti.awt.metal.NativeObject;
import com.ibm.ugl.eswt.OS;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2005  All Rights Reserved
 */

public class SoundPeer extends NativeObject {

	private static final int DEFAULT_CHUNK_SIZE = 16384;
	private static final int PREFERRED_SIZE_ALL = -1;
	private static final int PLAY_ONCE = 0;

	private URL fUrl;
	private volatile boolean fRunning;
	private Thread fThread;
	private volatile Object fThreadLock;

	private static boolean DEBUG = false;
	private static PrintStream fLogStream = System.err;
	private static Vector gCleanupList;

	static {
		// TODO We need a better cleanup story than this.
		gCleanupList = new Vector();
		AccessController.doPrivileged(new PrivilegedAction() {
			public Object run() {
				Runtime.getRuntime().addShutdownHook(new Thread() {
					public void run() {
						int size = gCleanupList.size();
						for (int i=0; i<size; i++) {
							SoundPeer peer = (SoundPeer)gCleanupList.elementAt(i);
							peer.dispose();
						}
					}
				});
				return null;
			}
		});
	}

	private SoundPeer(URL url, int handle) {
		super(handle);
		fUrl = url;
		fRunning = false;
	}

	public static SoundPeer getInstance(URL url) {
		if (url == null) {
			return null;
		}
		int handle = OS.Sound_New();
		if (handle == 0) {
			return null;
		}
		SoundPeer peer = new SoundPeer(url,handle);
		gCleanupList.addElement(peer);
		return peer;
	}

	public void dispose() {
		if (fPeerHandle == 0) return;
		OS.Sound_Dispose(fPeerHandle);
		fPeerHandle = 0;
	}

	/*
	 * Play the clip in a loop.  This method cancels any existing
	 * play request, and executes the new request asynchronously.
	 */
	public void loop() {
		log(this+": loop()");
		playOrLoop(true);
	}

	/*
	 * Play the clip one time.  This method cancels any existing
	 * play request, and executes the new request asynchronously.
	 */
	public void play() {
		log(this+": play()");
		playOrLoop(false);
	}

	/*
	 * Stop any currently playing clip and wait
	 * for it to actually stop before returning.
	 */
	public synchronized void stop() {
		log(this+": stop()");
		if (!fRunning) {
			return;
		}
		OS.Sound_Stop(fPeerHandle);
		fRunning = false;

		// streaming audio spawns thread. wait for it to run to completion.
		if (fThread != null && fThread.isAlive()) {
			boolean stopped = false;
			while (!stopped) {
				try {
					fThread.join();
					stopped = true;
				} catch (InterruptedException e) {}
			}
		}
	}

	private void playAll(InputStream stream, boolean looped) {
		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		int bytesRead = 0;
		byte data[] = new byte[DEFAULT_CHUNK_SIZE];

		while (bytesRead != -1) {
			try {
				bytesRead = stream.read(data,0,DEFAULT_CHUNK_SIZE);
			} catch (IOException e1) {
				log(SoundPeer.this+": run2: "+e1.getMessage(),e1);
				fRunning = false;
				bytesRead = -1;
			}

			if (bytesRead != -1) {
				// Save the new data.
				outputStream.write(data,0,bytesRead);
			}
		}

		if (fRunning) {
			byte chunk[] = outputStream.toByteArray();
			try {
				OS.Sound_Play(fPeerHandle,chunk,chunk.length,looped ? OS.SOUND_REPEAT_FOREVER : PLAY_ONCE);
			} catch (OutOfMemoryError error) {
				log(SoundPeer.this+": run: writeWholeClip: nativeWriteChunk failed. Out of memory");
				fRunning = false;
			} catch (RuntimeException exception) {
				log(SoundPeer.this+": run: writeWholeClip: nativeWriteChunk failed:" + exception.getMessage());
				fRunning = false;
			}
		}
	}

	private void playAll(boolean looped) {
		InputStream inputStream = null;

		try {
			inputStream = fUrl.openStream();
		} catch (IOException e) {
			log(SoundPeer.this+": run1: "+e.getMessage(),e);
			fRunning = false;
		}

		if (inputStream != null) {
			playAll(inputStream, looped);

			try {
				log(SoundPeer.this+": closing stream");
				inputStream.close();
			} catch (IOException e1) {
				log(SoundPeer.this+": run3: "+e1.getMessage(),e1);
			}
		}
	}

	private synchronized void playOrLoop(boolean looped) {
		int bufferSize;

		stop();
		fRunning = true;

		bufferSize = OS.Sound_GetPreferredChunkSize(fPeerHandle);
		if (bufferSize == OS.SOUND_CHUNK_SIZE_ALL || looped) {
			playAll(looped);
		} else {
			fThreadLock = new Object();
			fThread = new Thread(getRunnable(looped, bufferSize));
			synchronized (fThreadLock) {
				fThread.start();
				try {
					fThreadLock.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/*
	 * The chunk reading should be done in Java to prevent
	 * each new native implementor from having to write URL
	 * reading code in the native.
	 */
	private Runnable getRunnable(final boolean looped, final int chunkSize) {

		Runnable r = new Runnable() {

			private void playStreaming(InputStream stream) {
				int bytesRead = 0;
				byte data[] = new byte[chunkSize];
				boolean playing = false;

				while (fRunning && bytesRead != -1) {
					try {
						bytesRead = stream.read(data,0,chunkSize);
					} catch (IOException e1) {
						log(SoundPeer.this+": run2: "+e1.getMessage(),e1);
						fRunning = false;
						bytesRead = -1;
						if (playing) OS.Sound_Stop(fPeerHandle);
					}

					if (bytesRead != -1) {
						// Write the next chunk to the peer.
						try {
							OS.Sound_Play(fPeerHandle,data,bytesRead,PLAY_ONCE);
							if (!playing) {
								playing = true;
								// notify thread spawner that playing has started
								synchronized (fThreadLock) {
									fThreadLock.notify();
								}
							}
						} catch (OutOfMemoryError error) {
							log(SoundPeer.this+": run: writeWholeClip: nativeWriteChunk failed. Out of memory");
							OS.Sound_Stop(fPeerHandle);
							fRunning = false;
						} catch (RuntimeException exception) {
							log(SoundPeer.this+": run: writeWholeClip: nativeWriteChunk failed:" + exception.getMessage());
							OS.Sound_Stop(fPeerHandle);
							fRunning = false;
						}
					}
				}
				if (!playing) {
					// let thread spawner continue
					synchronized (fThreadLock) {
						fThreadLock.notify();
					}
				}
			}
			public void run() {
				InputStream inputStream = null;

				try {
					log(SoundPeer.this+": opening stream");
					inputStream = fUrl.openStream();
				} catch (IOException e) {
					log(SoundPeer.this+": run1: "+e.getMessage(),e);
					fRunning = false;
					// let thread spawner continue
					synchronized (fThreadLock) {
						fThreadLock.notify();
					}
					return;
				}

				playStreaming(inputStream);
				try {
					if (inputStream != null) {
						log(SoundPeer.this+": closing stream");
						inputStream.close();
					}
				} catch (IOException e1) {
					log(SoundPeer.this+": run3: "+e1.getMessage(),e1);
				}
			}
		};
		return r;
	}

	public String toString() {
		return "SoundPeer [handle="+fPeerHandle+",url="+fUrl+"]";
	}

	private static synchronized void log(String string) {
		if (!DEBUG) return;
		fLogStream.println(string);
	}

	private static synchronized void log(String string, Exception exception) {
		if (!DEBUG) return;
		log(string);
		exception.printStackTrace(fLogStream);
	}
}
