package com.ibm.oti.awt.metal.widgets;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2006  All Rights Reserved
 */

import java.awt.*;

import com.ibm.oti.awt.metal.graphics.Device;
import com.ibm.ugl.eswt.OS;
public class Display extends Device {

	private static Thread eventThread;
	private boolean done=false;

	private static int systemColors[];

	private boolean shellCallbackRegistered = false;

	// Lightweight mouse event support
	private static Component lastComponentWithEvents;
	private static Component lastGrabbedComponentWithEvents;
	private static Component lastComponentEntered;

	public static Component getLastComponentWithEvents() {
		return lastComponentWithEvents;
	}

	public static void setLastComponentWithEvents(Component c) {
		lastComponentWithEvents = c;
	}

	public static Component getLastGrabbedComponentWithEvents() {
		return lastGrabbedComponentWithEvents;
	}

	public static void setLastGrabbedComponentWithEvents(Component c) {
		lastGrabbedComponentWithEvents = c;
	}

	public static Component getLastComponentEntered() {
		return lastComponentEntered;
	}

	public static void logComponentEntered(Component c) {
		lastComponentEntered = c;
	}

	public static void logComponentExited(Component c) {
		if (lastComponentEntered == c) {
			lastComponentEntered = null;
		}
	}

	public Rectangle getBounds() {
		int[] rect = OS.Device_GetBounds(fPeerHandle);
		java.awt.Rectangle returnValue = new java.awt.Rectangle(rect[OS.INDEX_X],rect[OS.INDEX_Y],rect[OS.INDEX_WIDTH],rect[OS.INDEX_HEIGHT]);
		return returnValue;
	}

	public boolean isValidPeerThread() {
		return isEventLoopThread(Thread.currentThread());
	}

	public boolean isEventLoopThread(Thread currentThread) {
		if (eventThread == null) return false;
		if (currentThread.equals(eventThread)) return true;
		return false;
	}

	public Display(int handle) {
		super(handle);

		registerEventCallbacks();
	}

	public static Display openDisplay() {
		final Object lock = new Object();
		final Display[] cell = new Display[1];

		eventThread = new Thread(new Runnable() {
			public void run() {
				synchronized(lock) {
					try {
						cell[0] = new Display(OS.Display_New());
					} finally {
						// This is just in case new Display throws an exception.   At least
						// now openDisplay() will return with null
						lock.notifyAll();
					}
				}

				if (cell[0] != null) {
					cell[0].runEventLoop();
				}
			}
		}, "Display thread");
		eventThread.setDaemon(true);

		synchronized(lock) {
			eventThread.start();
			try {
				lock.wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		return cell[0];
	}

	public int getSystemColor(int id) {
		int nativeColor = OS.COLOR_DESKTOP_BACKGROUND;

		if (systemColors == null) {
			systemColors = new int[SystemColor.NUM_COLORS];
		} else {
			int cachedColor = systemColors[id];
			if (cachedColor != 0) return cachedColor;
		}

		switch (id) {
			case SystemColor.DESKTOP:					nativeColor = OS.COLOR_DESKTOP_BACKGROUND; break;
			case SystemColor.ACTIVE_CAPTION:			nativeColor = OS.COLOR_ACTIVE_CAPTION_BACKGROUND; break;
			case SystemColor.ACTIVE_CAPTION_TEXT:		nativeColor = OS.COLOR_ACTIVE_CAPTION_FOREGROUND; break;
			case SystemColor.ACTIVE_CAPTION_BORDER:		nativeColor = OS.COLOR_ACTIVE_CAPTION_BORDER; break;
			case SystemColor.INACTIVE_CAPTION:			nativeColor = OS.COLOR_INACTIVE_CAPTION_BACKGROUND; break;
			case SystemColor.INACTIVE_CAPTION_TEXT:		nativeColor = OS.COLOR_INACTIVE_CAPTION_FOREGROUND; break;
			case SystemColor.INACTIVE_CAPTION_BORDER:	nativeColor = OS.COLOR_INACTIVE_CAPTION_BORDER; break;
			case SystemColor.WINDOW:					nativeColor = OS.COLOR_WINDOW_BACKGROUND; break;
			case SystemColor.WINDOW_BORDER:				nativeColor = OS.COLOR_WINDOW_BORDER; break;
			case SystemColor.WINDOW_TEXT:				nativeColor = OS.COLOR_WINDOW_FOREGROUND; break;
			case SystemColor.MENU:						nativeColor = OS.COLOR_MENU_BACKGROUND; break;
			case SystemColor.MENU_TEXT:					nativeColor = OS.COLOR_MENU_FOREGROUND; break;
			case SystemColor.TEXT:						nativeColor = OS.COLOR_TEXT_BACKGROUND; break;
			case SystemColor.TEXT_TEXT:					nativeColor = OS.COLOR_TEXT_FOREGROUND; break;
			case SystemColor.TEXT_HIGHLIGHT:			nativeColor = OS.COLOR_TEXT_HIGHLIGHT_BACKGROUND; break;
			case SystemColor.TEXT_HIGHLIGHT_TEXT:		nativeColor = OS.COLOR_TEXT_HIGHLIGHT_FOREGROUND; break;
			case SystemColor.TEXT_INACTIVE_TEXT:		nativeColor = OS.COLOR_TEXT_INACTIVE_FOREGROUND; break;
			case SystemColor.CONTROL:					nativeColor = OS.COLOR_CONTROL_BACKGROUND; break;
			case SystemColor.CONTROL_TEXT:				nativeColor = OS.COLOR_CONTROL_FOREGROUND; break;
			case SystemColor.CONTROL_HIGHLIGHT:			nativeColor = OS.COLOR_CONTROL_HIGHLIGHT_NORMAL; break;
			case SystemColor.CONTROL_LT_HIGHLIGHT:		nativeColor = OS.COLOR_CONTROL_HIGHLIGHT_LIGHT; break;
			case SystemColor.CONTROL_SHADOW:			nativeColor = OS.COLOR_CONTROL_SHADOW_NORMAL; break;
			case SystemColor.CONTROL_DK_SHADOW:			nativeColor = OS.COLOR_CONTROL_SHADOW_DARK; break;
			case SystemColor.SCROLLBAR:					nativeColor = OS.COLOR_SCROLLBAR_BACKGROUND; break;
			case SystemColor.INFO:						nativeColor = OS.COLOR_INFO_BACKGROUND; break;
			case SystemColor.INFO_TEXT:					nativeColor = OS.COLOR_INFO_FOREGROUND; break;

		}

		int nativeHandle = OS.Display_GetSystemColor(this.fPeerHandle, nativeColor);
		systemColors[id] = nativeHandle;
		return nativeHandle;
	}

	public void beep() {
		OS.Display_Beep(this.fPeerHandle);
	}

	private void runEventLoop() {
		while (!done) {
			try {
				while (OS.Display_ReadAndDispatch(fPeerHandle));
				OS.Display_Sleep(fPeerHandle);
			} catch (Throwable e) {
				e.printStackTrace();
			}
		}
	}

	public void dispose() {
		if (eventThread != null) {
			// signal the thread as done, and wait for it to stop
			done = true;
			OS.Display_Wake(fPeerHandle);
			try {
				eventThread.join();
			} catch (InterruptedException e) {
				// Do nothing
			}
			eventThread = null;
		}

		// Cleanup the system colors which have been initialized
		if (systemColors != null) {
			for (int i=0;i<systemColors.length;i++) {
				int handle = systemColors[i];
				if (handle != 0) OS.Color_Dispose(handle);
			}
		}

		super.dispose();
	}

	private void registerEventCallbacks() {
		OS.Display_InitCallbacks(fPeerHandle, this);
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_SHELL, "java/awt/Window", "shellCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_CONTROL_KEY, "java/awt/Component", "keyCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_CONTROL_POINTER, "java/awt/Component", "pointerCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_FOCUS, "java/awt/Component", "focusCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_PAINT, "java/awt/Component", "paintCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_SHELL_RESIZE, "java/awt/Window", "resizeCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_SHELL_MOVE, "java/awt/Window", "moveCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_GET_PROPERTY, "com/ibm/oti/awt/metal/widgets/Display", "getProperty");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_BUTTON, "java/awt/Component", "buttonCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_LIST, "java/awt/List", "listCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_SCROLLBAR, "java/awt/Scrollbar", "scrollbarCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_CHOICE, "java/awt/Choice", "choiceCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_TEXT_MODIFY, "java/awt/TextComponent", "textModifyCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_MENU_SELECTION, "java/awt/MenuItem", "selectionCallback");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_HAS_MENU, "java/awt/Component", "hasMenuInfo");
		OS.Display_RegisterCallback(fPeerHandle, OS.CALLBACK_CONTROL_WHEEL, "java/awt/Component", "mouseWheelCallback");
	}

	/*
	 * Return the value of the property.
	 */
	String getProperty(String property) {
		// In the future, System.getProperty() must be the default case.
		return System.getProperty(property);
	}

	public void clearFocus() {
		OS.Display_ClearFocus(fPeerHandle);
	}

	/**
	 * Returns true if the native windowing system supports
	 * a native full screen "exclusive" mode.  This means that
	 * other windows cannot overlap the window currently indicated
	 * as full screen, and appear below that window in z-order.
	 *
	 * @since PPro 1.1
	 */
	public boolean isFullScreenSupported() {
		return OS.Platform_GetCapability(OS.CAPABILITY_FULLSCREEN_EXCLUSIVE_MODE) != 0;
	}

	/**
	 * Sets the window indicated by the argument to be in
	 * exclusive full screen mode.  If the argument is null,
	 * then overlapping window mode is restored.
	 *
	 * @since PPro 1.1
	 */
	public void setFullScreenWindow(int windowHandle) {
		OS.Display_SetFullScreenShell(this.fPeerHandle, windowHandle);
	}

	public Insets getInsets() {
		int[] insetArray = OS.Display_GetInsets(this.fPeerHandle);
		Insets insets = new Insets(
			insetArray[OS.INDEX_TOP],
			insetArray[OS.INDEX_LEFT],
			insetArray[OS.INDEX_BOTTOM],
			insetArray[OS.INDEX_RIGHT]);
		return insets;
	}

}
