package com.ibm.oti.connection.datagram;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000, 2006  All Rights Reserved
 */

import java.io.*;

import javax.microedition.io.*;
import com.ibm.oti.connection.*;
import com.ibm.oti.connection.socket.Socket;
import java.net.*;

/**
 * Implements a datagram connection. Accepts the following parameters:
 * <UL>
 * <LI>so_timeout - set the receive timeout in milliseconds, 0 waits forever,
 * the default is zero. If the connection is opened with the timeout argument set
 * to true, and the so_timeout is not set, the default is 3000 milliseconds.</LI>
 * <LI>so_rcvbuf - set the socket receive buffer size in bytes.</LI>
 * <LI>so_sndbuf - set the socket send buffer size in bytes.</LI>
 * </UL>
 *
 * @author		OTI
 * @version		initial
 *
 * @see		javax.microedition.io.DatagramConnection
 */

public class Connection implements CreateConnection, UDPDatagramConnection {

	static final int DEFAULT_TIMEOUT = 8000;
	private String server;
	private boolean closed = false;
	private int access, port, timeout = 0;
	private java.net.DatagramSocket socket;

public Connection() {
}

public void close() throws IOException {
	closed = true;
	socket.close();
}

private static native int netMaxDatagramImpl(java.net.DatagramSocket socket);
private static native int netNominalDatagramImpl(java.net.DatagramSocket socket);

public int getMaximumLength() throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return netMaxDatagramImpl(socket);
}

public int getNominalLength() throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return netNominalDatagramImpl(socket);
}

public Datagram newDatagram(byte[] data, int length) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if ((length < 0) || isOverMaxLength(length)) throw new IllegalArgumentException();
	Datagram datagram = new DatagramPacket();
	datagram.setData(data, 0, length);
	StringBuffer address = new StringBuffer("datagram://");
	if (server != null) {
		address.append(server);
	}
	address.append(':');
	address.append(port);
	datagram.setAddress(address.toString());
	return datagram;
}

public Datagram newDatagram(byte[] data, int length, String address) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if ((length < 0) || isOverMaxLength(length)) throw new IllegalArgumentException();
	Datagram datagram = new DatagramPacket();
	datagram.setData(data, 0, length);
	datagram.setAddress(address);
	return datagram;
}

public Datagram newDatagram(int length) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if ((length < 0) || isOverMaxLength(length)) throw new IllegalArgumentException();
	return newDatagram(new byte[length], length);
}

public Datagram newDatagram(int length, String address) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if ((length < 0) || isOverMaxLength(length)) throw new IllegalArgumentException();
	Datagram datagram = new DatagramPacket();
	datagram.setData(new byte[length], 0, length);
	datagram.setAddress(address);
	return datagram;
}

private boolean isOverMaxLength(int length){
	try {
		if (getMaximumLength() < length)
			return true;
	} catch (IOException e) {}
	return false;
}

public void receive(javax.microedition.io.Datagram datagram) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.READ && access != Connector.READ_WRITE)
		// K00a9 = Not open for READ
		throw new IOException(com.ibm.oti.util.Msg.getString("K00a9"));

	DatagramPacket data;
	if (datagram instanceof DatagramPacket) {
		data = (DatagramPacket)datagram;
	} else {
		data = new DatagramPacket();
		data.setData(datagram.getData(), datagram.getOffset(), datagram.getLength());
		data.setAddress(datagram.getAddress());
	}
	try {
		socket.receive(data.getNetPacket());
		data.setData(data.getData(), data.getOffset(), data.getLength());
		if (data != datagram) {
			datagram.setAddress(data.getAddress());
		}
	} catch (java.net.SocketException e) {
		throw new IOException(e.getMessage());
	}
}

public void send(Datagram datagram) throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.WRITE && access != Connector.READ_WRITE)
		// K00aa = Not open for WRITE
		throw new IOException(com.ibm.oti.util.Msg.getString("K00aa"));
	DatagramPacket data;
	if (datagram instanceof DatagramPacket) {
		data = (DatagramPacket)datagram;
	} else {
		data = new DatagramPacket();
		data.setData(datagram.getData(), datagram.getOffset(), datagram.getLength());
		data.setAddress(datagram.getAddress());
	}

	String serverAddress = data.getAddress();
	if (serverAddress == null) {
		if (server == null) return;
		if (data == datagram) {
			data = new DatagramPacket();
			data.setData(datagram.getData(), datagram.getOffset(), datagram.getLength());
		}
		data.setAddress("datagram://" + server + ":" + this.port);
	}

	try {
		socket.send(data.getNetPacket());
	} catch (java.net.SocketException e) {
		throw new IOException(e.getMessage());
	}
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 2.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public javax.microedition.io.Connection setParameters2(String spec, int access, boolean timeout) throws IOException {

	setParameters(spec, access, timeout);
	return this;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 1.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, int access, boolean timeout) throws IOException {

	String[][] equates = ConnectionUtil.NO_PARAMETERS;
	int index = spec.indexOf(';');
	if (index != -1) {
		equates = ConnectionUtil.getParameters(spec.substring(index + 1));
		spec = spec.substring(0, index);
	}
	setParameters(spec, equates, access, timeout);
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		equates String[][]
 *					An 2-dimensional array of strings that
 *					contain the parameters and their keys
 *					passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		throwTimeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
private void setParameters(String spec, String[][] equates, int access, boolean throwTimeout) throws IOException {
	this.access = access;
	int rcvbuf = 0, sndbuf = 0;
	int[] result = new int[1];
	for (int i=0; i<equates.length; i++) {
		String key = equates[i][0];
		equates[i][0] = equates[i][0].toLowerCase();
		if (ConnectionUtil.intParam("so_timeout", equates[i], ConnectionUtil.NEGATIVE, result))
			timeout = result[0];
		else if (ConnectionUtil.intParam("so_rcvbuf", equates[i], ConnectionUtil.NEGATIVE_OR_ZERO, result))
			rcvbuf = result[0];
		else if (ConnectionUtil.intParam("so_sndbuf", equates[i], ConnectionUtil.NEGATIVE_OR_ZERO, result))
			sndbuf = result[0];
		// K00a5 = Invalid parameter - {0}
		else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a5", key));
	}
	if (throwTimeout && timeout == 0)
		timeout = DEFAULT_TIMEOUT;

	if (!spec.startsWith("//"))
		// K00a1 = Missing //
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a1"));
	String host = Socket.parseURL(spec, result, true, true);
	if (host.length() != 0) {
		server = host;
	}
	port = result[0];

	try {
		socket = new java.net.DatagramSocket(server == null ? port : 0);
		if (rcvbuf != 0) socket.setReceiveBufferSize(rcvbuf);
		if (sndbuf != 0) socket.setSendBufferSize(sndbuf);
		if (timeout != 0) socket.setSoTimeout(timeout);
	} catch (java.net.SocketException e) {
		if (socket != null) socket.close();
		throw new IOException(e.getMessage());
	}
}

private boolean addressZero(byte[] address) {
	boolean addressZero = true;
	for (int i=0; i<address.length; i++) {
		if(address[i] != 0 ) {
			addressZero = false;
			break;
		}
	}
	return addressZero;
}

/**
 * @see javax.microedition.io.UDPDatagramConnection#getLocalAddress()
 */
public String getLocalAddress() throws IOException {
	// K00ac = Connection is closed
	if(closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	InetAddress inet = socket.getLocalAddress();
	if (!addressZero(inet.getAddress()))
		return inet.getHostAddress();
	inet = InetAddress.getLocalHost();
	if (!addressZero(inet.getAddress()))
		return inet.getHostAddress();
	return "127.0.0.1";
}

/**
 * @see javax.microedition.io.UDPDatagramConnection#getLocalPort()
 */
public int getLocalPort() throws IOException {
	// K00ac = Connection is closed
	if(closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return socket.getLocalPort();
}

}
