package com.ibm.oti.connection.file.base;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000, 2006  All Rights Reserved
 */

import java.io.*;
import javax.microedition.io.*;
import com.ibm.oti.connection.*;

/**
 * Implements a file connection. Accepts the following parameters:
 * <UL>
 * <LI>append - true/false, append to an existing file, only valid when writing,
 * the default is false which will overwrite an existing file</LI>
 * </UL>
 *
 * @author		OTI
 * @version		initial
 *
 * @see		javax.microedition.io.InputConnection
 * @see		javax.microedition.io.OutputConnection
 */
public class Connection extends DataConnection implements CreateConnection, StreamConnection {
	private File file;
	private int access;
	private boolean append = false;

public Connection() {
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 2.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public javax.microedition.io.Connection setParameters2(String spec, int access, boolean timeout) throws IOException {
	setParameters(spec, access, timeout);
	return this;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 1.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, int access, boolean timeout) throws IOException {
	String[][] equates = ConnectionUtil.NO_PARAMETERS;
	int index = spec.indexOf(';');
	if (index != -1) {
		equates = ConnectionUtil.getParameters(spec.substring(index + 1));
		spec = spec.substring(0, index);
	}
	setParameters(spec, equates, access, timeout);
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		equates String[][]
 *					An 2-dimensional array of strings that
 *					contain the parameters and their keys
 *					passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, String[][] equates, int access, boolean timeout) throws IOException {
	int idx = 0;
	spec = com.ibm.oti.util.Util.decode(spec, false);
	if (spec.startsWith("//")) {
		boolean found = false;
		int length = spec.length();
		for (idx=2; idx<length; idx++) {
			char ch;
			if ((ch = spec.charAt(idx)) == '/' || ch == '\\') {
				found = true;
				break;
			} else if (ch == ':' || ch == ' ') break;
		}
		if (!found) throw new IllegalArgumentException(spec);
	}
	String name;
	if (idx == 2) name = spec.substring(2);
	else name = spec;
	file = new File(name);
	this.access = access;
	for (int i=0; i<equates.length; i++) {
		String key = equates[i][0].toLowerCase();
		if (key.equals("append") && equates[i][1] != null) {
			String value = equates[i][1].toLowerCase();
			if (value.equals("true")) append = true;
			else if (!value.equals("false"))
				// K00b5 = Invalid boolean value: {0}
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00b5"));
		// K00a5 = Invalid parameter - {0}
		} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a5", equates[i][0]));
	}
	if (access == Connector.READ && !file.exists()) {
		throw new ConnectionNotFoundException(name);
	}
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
		// READ is checked by file.exists() above
		if (access == Connector.READ_WRITE) {
			sm.checkRead(file.getPath());
		}
		if (access == Connector.WRITE || access == Connector.READ_WRITE) {
			sm.checkWrite(file.getPath());
		}
	}
}

public void close() throws IOException {
	file = null;
}

public InputStream openInputStream() throws IOException {
	// K00ac = Connection is closed
	if (file == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.READ && access != Connector.READ_WRITE)
		// K00a9 = Not open for READ
		throw new IOException(com.ibm.oti.util.Msg.getString("K00a9"));
	return new FileInputStream(file);
}

public OutputStream openOutputStream() throws IOException {
	// K00ac = Connection is closed
	if (file == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.WRITE && access != Connector.READ_WRITE)
		// K00aa = Not open for WRITE
		throw new IOException(com.ibm.oti.util.Msg.getString("K00aa"));
	return new FileOutputStream(file, append);
}

}
