package com.ibm.oti.connection.ftp;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000, 2006  All Rights Reserved
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import javax.microedition.io.*;
import com.ibm.oti.connection.*;

/**
 * Implements a FTP connection using the URL format - //login:password@host:port/file.
 * Accepts the following parameters:
 * <UL>
 * <LI>type - 'A' for an ASCII transfer, 'I' for a binary transfer, the default is binary</LI>
 * </UL>
 *
 * @author		OTI
 * @version		initial
 *
 * @see		javax.microedition.io.InputConnection
 * @see		javax.microedition.io.OutputConnection
 */
public class Connection extends DataConnection implements CreateConnection {
	private String host, file, user = "anonymous", password = "";
	private int access;
	private InputStream ctrlInput, inputStream;
	private OutputStream ctrlOutput, outputStream;
	private boolean streamOpen = false;

	//FTP Reply Constants
	private static final int FTP_DATAOPEN = 125;
	private static final int FTP_OPENDATA = 150;
	private static final int FTP_OK = 200;
	private static final int FTP_USERREADY = 220;
	private static final int FTP_TRANSFEROK = 226;
	private static final int FTP_LOGGEDIN = 230;
	private static final int FTP_FILEOK = 250;
	private static final int FTP_PASWD = 331;
	private static final int FTP_NOTFOUND = 550;

public Connection() {
}

public void close() throws IOException {
	host = null;
	if (!streamOpen) {
		if (inputStream != null) inputStream.close();
		else outputStream.close();
	}
}

public InputStream openInputStream() throws IOException {
	// K00ac = Connection is closed
	if (host == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.READ)
		// K00a9 = Not open for READ
		throw new IOException(com.ibm.oti.util.Msg.getString("K00a9"));
	streamOpen = true;
	return inputStream;
}

public OutputStream openOutputStream() throws IOException {
	// K00ac = Connection is closed
	if (host == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.WRITE)
		// K00aa = Not open for WRITE
		throw new IOException(com.ibm.oti.util.Msg.getString("K00aa"));
	streamOpen = true;
	return outputStream;
}

private void disconnect(StreamConnection c0, StreamConnectionNotifier c1) {
	try {if (c0 != null) c0.close();} catch (Exception e) {}
	try {if (ctrlInput != null) ctrlInput.close();} catch (Exception e) {}
	try {if (ctrlOutput != null) ctrlOutput.close();} catch (Exception e) {}
	try {if (c1 != null) c1.close();} catch (Exception e) {}
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 2.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public javax.microedition.io.Connection setParameters2(String spec, int access, boolean timeout) throws IOException {
	setParameters(spec, access, timeout);
	return this;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 1.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, int access, boolean timeout) throws IOException {
	String[][] equates = ConnectionUtil.NO_PARAMETERS;
	int index = spec.indexOf(';');
	if (index != -1) {
		equates = ConnectionUtil.getParameters(spec.substring(index + 1));
		spec = spec.substring(0, index);
	}
	setParameters(spec, equates, access, timeout);
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		equates String[][]
 *					An 2-dimensional array of strings that
 *					contain the parameters and their keys
 *					passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, String[][] equates, int access, boolean timeout) throws IOException {
	char type = 'I';
	for (int i=0; i<equates.length; i++) {
		String key = equates[i][0].toLowerCase();
		if (key.equals("type") && equates[i][1] != null) {
			char ch;
			if (equates[i][1].length() > 1 ||
				((ch = Character.toUpperCase(equates[i][1].charAt(0))) != 'A' &&
				ch != 'I'))
					// K0089 = Unknown type: {0}
					throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0089", equates[i][1]));
			type = ch;
		// K00a5 = Invalid parameter - {0}
		} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a5", equates[i][0]));
	}
	this.access = access;
	if (access == Connector.READ_WRITE)
		// K00b4 = READ_WRITE not supported
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00b4"));
	if (spec.startsWith("//")) {
		int index = spec.indexOf('/', 2);
		if (index == -1)
			// K00d6 = Missing file
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00d6"));
		String hostPart = spec.substring(2, index);
		file = spec.substring(index);
		if ((index = hostPart.indexOf('@')) != -1) {
			String login = hostPart.substring(0, index);
			hostPart = hostPart.substring(index + 1);
			if ((index = login.indexOf(':')) != -1) {
				user = login.substring(0, index);
				password = login.substring(index + 1);
			} else user = login;
		}
		if (hostPart.indexOf(":") == -1) host = hostPart + ":21";
		else host = hostPart;
	} else {
		host = "localhost:21";
		file = spec;
	}

	StreamConnection dataSocket = null;
	com.ibm.oti.connection.serversocket.Connection acceptSocket = null;
	com.ibm.oti.connection.socket.Connection controlSocket = new com.ibm.oti.connection.socket.Connection();
	controlSocket.setParameters("//" + host, Connector.READ_WRITE, false);
	ctrlOutput = controlSocket.openOutputStream();
	ctrlInput = new BufferedInputStream(controlSocket.openInputStream());
	try {
		login();
		setType(type);
		if (access == Connector.WRITE) cd();
		acceptSocket = new com.ibm.oti.connection.serversocket.Connection();
		acceptSocket.setParameters("//:0;so_timeout=3000", Connector.READ, true);
		/* Cannot set REUSEADDR so we need to send a PORT comannd */
		port(acceptSocket.getLocalPort(), controlSocket.getLocalAddress());
		if (access == Connector.READ) getFile();
		else sendFile();
		dataSocket = acceptSocket.acceptAndOpen();
	} catch (RuntimeException e) {
		disconnect(controlSocket, acceptSocket);
		throw e;
	} catch (IOException e) {
		disconnect(controlSocket, acceptSocket);
		throw e;
	}
	disconnect(controlSocket, acceptSocket);

	if (access == Connector.READ) {
		final InputStream dataIn = dataSocket.openInputStream();
		dataSocket.close();
		inputStream = new InputStream() {
			public int available() throws IOException {
				return dataIn.available();
			}
			public int read() throws IOException {
				return dataIn.read();
			}
			public int read(byte[] buf, int offset, int length) throws IOException {
				return dataIn.read(buf, offset, length);
			}
			public long skip(long amount) throws IOException {
				return dataIn.skip(amount);
			}
			public void close() throws IOException {
				dataIn.close();
			}
		};
	} else {
		final OutputStream dataOut = dataSocket.openOutputStream();
		dataSocket.close();
		outputStream = new OutputStream() {
			public void write(int value) throws IOException {
				dataOut.write(value);
			}
			public void write(byte[] buf, int offset, int length) throws IOException {
				dataOut.write(buf, offset, length);
			}
			public void close() throws IOException {
				dataOut.close();
			}
		};
	}
}

private int getReply() throws IOException {
	String line = readLine();
	if (line.length() >= 4) {
		String replyCode = line.substring(0, 3);
		if (line.charAt(3) == '-')
			while (readMultiLine(replyCode)); // Read all of a multiline reply
		try {
			return Integer.parseInt(replyCode);
		} catch (NumberFormatException e) {}
	}
	// K00dd = Invalid response: {0}
	throw new IOException(com.ibm.oti.util.Msg.getString("K00dd", line));
}

private void login() throws IOException {
	int reply;
	reply = getReply();
	if (reply == FTP_USERREADY) {
	} else {
		// K0097 = Unable to connect to server: {0}
		throw new IOException(com.ibm.oti.util.Msg.getString("K0097", host));
	}
	write("USER " + user + "\r\n");
	reply = getReply();
	if (reply != FTP_PASWD && reply != FTP_LOGGEDIN)
		// K0098 = Unable to log into server: {0}
		throw new IOException(com.ibm.oti.util.Msg.getString("K0098", host));
	if (reply == FTP_PASWD) {
		write("PASS " + password + "\r\n");
		reply = getReply();
		if (!(reply == FTP_OK || reply == FTP_USERREADY || reply == FTP_LOGGEDIN))
			// K0098 = Unable to log into server: {0}
			throw new IOException(com.ibm.oti.util.Msg.getString("K0098", host));
	}
}

/* Read a line of text and return it for posible parsing */
private String readLine() throws IOException {
	StringBuffer sb = new StringBuffer();
	int c;
	while ((c = ctrlInput.read()) != '\n') {
		sb.append((char)c);
	}
	return sb.toString();
}

private boolean readMultiLine(String replyCode) throws IOException {
	String line = readLine();
	if (line.length() < 4)
		return true;
	if (line.substring(0, 3).equals(replyCode) && (line.charAt(3) == (char)32))
		return false;
	return true;
}

/*
 Set the type of the file transfer. Only Image is suported
 */
private void setType(char type) throws IOException{
	write("TYPE " + type + "\r\n");
	if (getReply() != FTP_OK)
		// K009b = Unable to set transfer type
		throw new IOException(com.ibm.oti.util.Msg.getString("K009b"));
}

/* Change the server directory to that specfied in the file */
private void cd() throws IOException {
	int idx = file.lastIndexOf('/');
	if (idx > 0) {
		String dir = file.substring(0, idx);
		write("CWD " + dir + "\r\n");
		int reply = getReply();
		if (reply != FTP_FILEOK && dir.length() > 0 && dir.charAt(0) == '/') {
			write("CWD " + dir.substring(1) + "\r\n");
			reply = getReply();
		}
		if (reply != FTP_FILEOK)
			// K0094 = Unable to change directories
			throw new IOException(com.ibm.oti.util.Msg.getString("K0094"));
	}
}

/*
 Issue the STOR command to the server with the file as the parameter
 */
private void sendFile() throws IOException {
	int reply;
	write("STOR " + file.substring(file.lastIndexOf('/') + 1) + "\r\n");
	reply = getReply();
	if (!(reply == FTP_OPENDATA || reply == FTP_OK || reply == FTP_DATAOPEN))
		// K009a = Unable to store file
		throw new IOException(com.ibm.oti.util.Msg.getString("K009a"));
}

private void getFile() throws IOException {
	int reply;
	write("RETR " + file + "\r\n");
	reply = getReply();
	if (reply == FTP_NOTFOUND && file.length() > 0 && file.charAt(0) == '/') {
		write("RETR " + file.substring(1) + "\r\n");
		reply = getReply();
	}
	if (!(reply == FTP_OPENDATA || reply == FTP_TRANSFEROK))
		// K0096 = Unable to retrieve file: {0}
		throw new IOException(com.ibm.oti.util.Msg.getString("K0096", reply));
}

private void port(int dataPort, String ip) throws IOException{
	ip = ip.replace('.', ',');
	write("PORT " + ip + "," + (dataPort >> 8) + ',' + (dataPort & 255) + "\r\n");
	if (getReply() != FTP_OK)
		// K0099 = Unable to configure data port
		throw new IOException(com.ibm.oti.util.Msg.getString("K0099"));
}

private void write(String command) throws IOException {
	ctrlOutput.write(command.getBytes("ISO8859_1"));
}

}
