package com.ibm.oti.connection.serversocket;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000, 2006  All Rights Reserved
 */

import java.io.*;
import javax.microedition.io.*;
import com.ibm.oti.connection.*;
import com.ibm.oti.connection.socket.Socket;
import java.net.InetAddress;

/**
 * Implements a server socket connection. Accepts the following parameters:
 * <UL>
 * <LI>so_timeout - set the accept timeout in milliseconds, 0 waits forever,
 * the default is zero. If the connection is opened with the timeout argument set
 * to true, and the so_timeout is not set, the default is 3000 milliseconds.</LI>
 * <LI>backlog - the number of pending connection to accept, the default is 50.</LI>
 * </UL>
 *
 * @author		OTI
 * @version		initial
 *
 * @see		javax.microedition.io.StreamConnectionNotifier
 */
public class Connection implements CreateConnection, ServerSocketConnection {
	static final int DEFAULT_TIMEOUT = 8000;
	private boolean closed = false;
	private int timeout = 0, backlog = 50;

	private java.net.ServerSocket socket;

public Connection() {
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 2.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public javax.microedition.io.Connection setParameters2(String spec, int access, boolean timeout) throws IOException {

	setParameters(spec, access, timeout);
	return this;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 1.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, int access, boolean timeout) throws IOException {

	String[][] equates = ConnectionUtil.NO_PARAMETERS;
	int index = spec.indexOf(';');
	if (index != -1) {
		equates = ConnectionUtil.getParameters(spec.substring(index + 1));
		spec = spec.substring(0, index);
	}
	setParameters(spec, equates, access, timeout);
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		equates String[][]
 *					An 2-dimensional array of strings that
 *					contain the parameters and their keys
 *					passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		throwTimeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
private void setParameters(String spec, String[][] equates, int access, boolean throwTimeout) throws IOException {
	int[] result = new int[1];
	for (int i=0; i<equates.length; i++) {
		String key = equates[i][0];
		equates[i][0] = equates[i][0].toLowerCase();
		if (ConnectionUtil.intParam("so_timeout", equates[i], ConnectionUtil.NEGATIVE, result))
			timeout = result[0];
		else if (ConnectionUtil.intParam("backlog", equates[i], ConnectionUtil.NEGATIVE, result))
			backlog = result[0];
		// K00a5 = Invalid parameter - {0}
		else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a5", key));
	}
	if (throwTimeout && timeout == 0)
		timeout = DEFAULT_TIMEOUT;

	result[0] = 0;
	Socket.parseURL(spec, result, false, true);

	try {
		socket = new java.net.ServerSocket(result[0]);
	} catch (java.net.SocketException e) {
		throw new IOException(e.getMessage());
	}
}

public void close() throws IOException {
	closed = true;
	socket.close();
}

public StreamConnection acceptAndOpen() throws IOException {
	// K00ac = Connection is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	try {
		java.net.Socket acceptSocket = socket.accept();
		StreamConnection connection = new com.ibm.oti.connection.socket.Connection(acceptSocket);
		return connection;
	} catch (java.net.SocketException e) {
		throw new IOException(e.getMessage());
	}
}

/**
 * @see javax.microedition.io.ServerSocketConnection#getLocalAddress()
 */
public String getLocalAddress() throws IOException {
	// K00ac = Connection is closed
	if(closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	InetAddress inet = socket.getInetAddress();
	String result = inet.getHostAddress();
	if (inet.isAnyLocalAddress()) {
		return InetAddress.getLocalHost().getHostAddress();
	}
	return result;
}

/**
 * @see javax.microedition.io.ServerSocketConnection#getLocalPort()
 */
public int getLocalPort() throws IOException {
	// K00ac = Connection is closed
	if(closed) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return socket.getLocalPort();
}

}
