package com.ibm.oti.connection.socket;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000, 2006  All Rights Reserved
 */

import java.io.*;
import javax.microedition.io.*;
import com.ibm.oti.connection.*;

/**
 * Implements a socket connection. Accepts the following parameters:
 * <UL>
 * <LI>so_timeout - set the receive timeout in milliseconds, 0 waits forever,
 * the default is zero. If the connection is opened with the timeout argument set
 * to true, and the so_timeout is not set, the default is 3000 milliseconds.</LI>
 * <LI>so_linger - set the socket linger option in milliseconds.</LI>
 * <LI>tcp_nodelay - true/false, set the socket TCP nodelay option.</LI>
 * <LI>so_keepalive - true/false, set the socket keepalive option.</LI>
 * <LI>so_rcvbuf - set the socket receive buffer size in bytes.</LI>
 * <LI>so_sndbuf - set the socket send buffer size in bytes.</LI>
 * </UL>
 *
 * @author		OTI
 * @version		initial
 *
 * @see		javax.microedition.io.StreamConnection
 */
public class Connection implements CreateConnection, SocketConnection {
	static final int DEFAULT_TIMEOUT = 8000;
	private static final int UNOPENED = 0, OPEN = 1, CLOSED = 2;
	private String host;
	private int access, port = 0, timeout = 0;
	private int inputStatus = UNOPENED, outputStatus = UNOPENED;

	private java.net.Socket socket;

public Connection() {
}

public Connection(java.net.Socket socket) {
	host = "";
	access = Connector.READ_WRITE;
	this.socket = socket;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 2.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public javax.microedition.io.Connection setParameters2(String spec, int access, boolean timeout) throws IOException {

	// if the spec starts with "socket://:" then it is a server socket
	if (spec.startsWith("//") && (spec.length() == 2 || (spec.length() > 2 &&
		(spec.charAt(2) == ':' || spec.charAt(2) == ';'))))
	{
		Class connectionClass = null;
		try {
			connectionClass = Class.forName("com.ibm.oti.connection.serversocket.Connection");
		} catch (ClassNotFoundException e) {
			// K0002 = Scheme not found: {0}
			throw new ConnectionNotFoundException(com.ibm.oti.util.Msg.getString("K0002", "serversocket"));
		}
		boolean exception = false;
		CreateConnection connection = null;
		try {
			connection = (CreateConnection)connectionClass.newInstance();
		} catch (InstantiationException e) {
			exception = true;
		} catch (IllegalAccessException e) {
			exception = true;
		}
		if (exception)
			// K0003 = Cannot instantiate scheme: {0}
			throw new ConnectionNotFoundException(com.ibm.oti.util.Msg.getString("K0003", connectionClass));
		return connection.setParameters2(spec, access, timeout);
	}
	setParameters(spec, access, timeout);
	return this;
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object. Protocol used by MIDP 1.0
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		timeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
public void setParameters(String spec, int access, boolean timeout) throws IOException {

	String[][] equates = ConnectionUtil.NO_PARAMETERS;
	int index = spec.indexOf(';');
	if (index != -1) {
		equates = ConnectionUtil.getParameters(spec.substring(index + 1));
		spec = spec.substring(0, index);
	}
	setParameters(spec, equates, access, timeout);
}

/**
 * Passes the parameters from the Connector.open() method to this
 * object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		spec String
 *					The address passed to Connector.open()
 * @param		equates String[][]
 *					An 2-dimensional array of strings that
 *					contain the parameters and their keys
 *					passed to Connector.open()
 * @param		access int
 *					The type of access this Connection is
 *					granted (READ, WRITE, READ_WRITE)
 * @param		throwTimeout boolean
 *					A boolean indicating whether or not the
 *					caller to Connector.open() wants timeout
 *					exceptions
 * @exception	IOException
 *					If an error occured opening and configuring
 *					serial port.
 *
 * @see javax.microedition.io.Connector
 */
private void setParameters(String spec, String[][] equates, int access, boolean throwTimeout) throws IOException {
	int[] result = new int[1];
	boolean setNoDelay = false, setKeepAlive = false;
	int linger = -1, rcvbuf = 0, sndbuf = 0;
	for (int i=0; i<equates.length; i++) {
		String key = equates[i][0];
		equates[i][0] = equates[i][0].toLowerCase();
		if (ConnectionUtil.intParam("so_timeout", equates[i], ConnectionUtil.NEGATIVE, result))
			timeout = result[0];
		else if (ConnectionUtil.intParam("so_linger", equates[i], ConnectionUtil.NEGATIVE, result)) {
			linger = result[0];
			if (linger > 65535) linger = 65535;
		} else if (equates[i][0].equals("tcp_nodelay")) {
			String value = equates[i][1].toLowerCase();
			if (value.equals("true")) setNoDelay = true;
			else if (!value.equals("false"))
				// K00b5 = Invalid boolean value: {0}
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00b5", equates[i][1]));
		} else if (equates[i][0].equals("so_keepalive")) {
			String value = equates[i][1].toLowerCase();
			if (value.equals("true")) setKeepAlive = true;
			else if (!value.equals("false"))
				// K00b5 = Invalid boolean value: {0}
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00b5", equates[i][1]));
		} else if (ConnectionUtil.intParam("so_rcvbuf", equates[i], ConnectionUtil.NEGATIVE_OR_ZERO, result))
			rcvbuf = result[0];
		else if (ConnectionUtil.intParam("so_sndbuf", equates[i], ConnectionUtil.NEGATIVE_OR_ZERO, result))
			sndbuf = result[0];
		// K00a5 = Invalid parameter - {0}
		else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00a5", key));
	}
	if (throwTimeout && timeout == 0)
		timeout = DEFAULT_TIMEOUT;

	host = spec;
	this.access = access;
	if (!host.startsWith("//"))
		// K01cc = '//' must follow protocol
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K01cc"));
	host = Socket.parseURL(spec, result, true, false);

	try {
		socket = new java.net.Socket(host, result[0]);
	} catch (java.net.UnknownHostException e) {
		// K01ce = Host lookup failed: {0} - {1}
		throw new ConnectionNotFoundException(com.ibm.oti.util.Msg.getString("K01ce", host, e.getMessage()));
	} catch (IOException e) {
		// K01cf = Could not establish connection: {0} - {1}
		throw new ConnectionNotFoundException(com.ibm.oti.util.Msg.getString("K01cf", host + ":" + result[0], e.getMessage()));
	}
	try {
		if (rcvbuf != 0) socket.setReceiveBufferSize(rcvbuf);
		if (sndbuf != 0) socket.setSendBufferSize(sndbuf);
		if (linger != -1) socket.setSoLinger(true, linger);
		if (setNoDelay) socket.setTcpNoDelay(true);
		if (setKeepAlive) socket.setKeepAlive(true);
		if (timeout != 0) socket.setSoTimeout(timeout);
	} catch (java.net.SocketException e) {
		if (socket != null) socket.close();
		throw new IOException(e.getMessage());
	}
}

public void close() throws IOException {
	host = null;
	if (inputStatus != OPEN && outputStatus != OPEN) {
		socket.close();
	}
}

public InputStream openInputStream() throws IOException {
	// K00ac = Connection is closed
	if (host == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.READ && access != Connector.READ_WRITE)
		// K00a9 = Not open for READ
		throw new IOException(com.ibm.oti.util.Msg.getString("K00a9"));
	if (inputStatus == CLOSED) {
		// K0059 = Stream is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
	}
	inputStatus = OPEN;

	return new InputStream() {
		InputStream stream = socket.getInputStream();
		public int available() throws IOException {
			if (inputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			return stream.available();
		}
		public long skip(int amount) throws IOException {
			if (inputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			return stream.skip(amount);
		}
		public void close() throws IOException {
			if(inputStatus != CLOSED) {
				inputStatus = CLOSED;
				if(outputStatus != OPEN && host == null) {
					stream.close();
				}
			}
		}
		public int read(byte[] buffer, int offset, int count) throws IOException {
			if (inputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			return stream.read(buffer, offset, count);
		}
		public int read() throws IOException {
			if (inputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			return stream.read();
		}
	};

}

public OutputStream openOutputStream() throws IOException {
	// K00ac = Connection is closed
	if (host == null) throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	if (access != Connector.WRITE && access != Connector.READ_WRITE)
		// K00aa = Not open for WRITE
		throw new IOException(com.ibm.oti.util.Msg.getString("K00aa"));
	if (outputStatus == CLOSED) {
		// K0059 = Stream is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
	}
	outputStatus = OPEN;

	return new OutputStream() {
		OutputStream stream = socket.getOutputStream();
		public void close() throws IOException {
			if(outputStatus != CLOSED) {
				// don't shutdown output on Windows CE with the linger socket option set
				if ((Socket.getSocketFlags()&Socket.FLAG_BROKEN_SO_LINGER_SHUTDOWN)==0 || socket.getSoLinger()==-1) {
					socket.shutdownOutput();
				}
				outputStatus = CLOSED;
				if(inputStatus != OPEN && host == null) {
					stream.close();
				}
			}
		}
		public void write(int oneByte) throws IOException {
			if (outputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			stream.write(oneByte);
		}
		public void write(byte[] buffer, int offset, int count) throws IOException {
			if (outputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			stream.write(buffer, offset, count);
		}
		public void flush() throws IOException {
			if (outputStatus == CLOSED)
				// K0059 = Stream is closed
				throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
			stream.flush();
		}
	};

}

public DataInputStream openDataInputStream() throws IOException {
	return new DataInputStream(openInputStream());
}

public DataOutputStream openDataOutputStream() throws IOException {
	return new DataOutputStream(openOutputStream());
}

/**
 * @see javax.microedition.io.SocketConnection#getLocalAddress()
 */
public String getLocalAddress() throws IOException {
	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	return socket.getLocalAddress().getHostAddress();
}

/**
 * @see javax.microedition.io.SocketConnection#getLocalPort()
 */
public int getLocalPort() throws IOException {
	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return socket.getLocalPort();
}

/**
 * @see javax.microedition.io.SocketConnection#getAddress()
 */
public String getAddress() throws IOException {

	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	return socket.getInetAddress().getHostAddress();
}

/**
 * @see javax.microedition.io.SocketConnection#getPort()
 */
public int getPort() throws IOException {
	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));
	return socket.getPort();
}

/**
 * @see javax.microedition.io.SocketConnection#getSocketOption(byte)
 */
public int getSocketOption(byte optname)
	throws IllegalArgumentException, IOException {

	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	try {
		switch (optname) {
			case DELAY :
				return socket.getTcpNoDelay() ? 0 : 1;
			case KEEPALIVE:
				return socket.getKeepAlive() ? 1 : 0;
			case LINGER:
				int result = socket.getSoLinger();
				return result == -1 ? 0 : result;
			case RCVBUF:
				return socket.getReceiveBufferSize();
			case SNDBUF:
				return socket.getSendBufferSize();
			default:
				// K01c8 = Invalid socket option
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K01c8"));
		}
	} catch (IOException e) {
		return -1;
	}

}

/**
 * @see javax.microedition.io.SocketConnection#setSocketOption(byte, int)
 */
public void setSocketOption(byte optname, int optval)
	throws IllegalArgumentException, IOException {

	if(host == null)
		// K00ac = Connection is closed
		throw new IOException(com.ibm.oti.util.Msg.getString("K00ac"));

	if(optval < 0)
		// K01c9 = Socket option value cannot be negative
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K01c9"));

	try {
		switch (optname) {
			case DELAY :
				socket.setTcpNoDelay(optval == 0);
				break;
			case KEEPALIVE:
				socket.setKeepAlive(optval != 0);
				break;
			case LINGER:
				socket.setSoLinger(optval != 0, optval);
				break;
			case RCVBUF:
				socket.setReceiveBufferSize(optval);
				break;
			case SNDBUF:
				socket.setSendBufferSize(optval);
				break;
			default:
				// K01c8 = Invalid socket option
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K01c8"));
		}
	} catch (IOException e) {
	}

}

/**
 * Returns the timeout for the socket.  This is the value of so_timeout if it was set as
 * a connection parameter or the value of DEFAULT_TIMEOUT if none was set.
 */
public int getTimeout() {
	try {
		return socket.getSoTimeout();
	} catch (IOException e) {
		return 0;
	}
}

}
