
package com.ibm.oti.security.provider;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

import com.ibm.oti.util.ASN1Decoder;
import com.ibm.oti.util.ASN1Encoder;
import com.ibm.oti.util.ASN1Exception;
import com.ibm.oti.util.ASN1Decoder.Node;
import com.ibm.oti.util.ASN1Decoder.TypeMapper;
import com.ibm.oti.util.ASN1Decoder.UTCTime;
import java.util.Date;
import com.ibm.oti.util.BASE64Encoder;

import java.math.BigInteger;

import java.math.BigInteger;
import java.security.Principal;
import java.security.cert.CRLException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import com.ibm.oti.util.OIDDatabase;

import com.ibm.j9.bluez.crypto.CL3;

/**
 * This class implements a set of methods to facilitate reading
 * PKCS7 data from a stream of bytes. Instead of using an ASN.1
 * decoder instance to read the data and extract individual fields
 * manually, this class provides this service in the form of APIs.
 *
 * Not all fields were converted to API, only those that were needed
 * were converted.
 *
 * @author		OTI
 * @version		initial
 */
public class PKCS7 {

	private ASN1Decoder decoder;		// ASN.1 decoder used to parse the PKCS7 data from the input
	private ASN1Decoder.Node contents;	// The ASN.1 data decoded from the input stream
	private byte[] encoded;

/**
 * Returns a PKCS7 decoder for a given input stream.
 *
 * @param	inStream	InputStream		Where to read the PKCS7 data from
 * @return	ASN1Decoder	a ASN.1 decoder for PKCS7 data
 */
public static ASN1Decoder decoder (InputStream inStream) {
		ASN1Decoder decoder = new ASN1Decoder(inStream);
		TypeMapper mapper = new TypeMapper() {
			public int map (int originalType, int nesting, int sequenceItem) {
				// We are only called for nesting 1, for the 'content' field [0] EXPLICIT
				if( originalType == 0 ){
					return ASN1Decoder.SEQUENCE;
				} else {
					return originalType;
				}
			};
		};
		decoder.configureTypeRedirection (1, mapper); // when nesting == 1

		// For SignedData, field certificates
		mapper = new TypeMapper() {
			public int map (int originalType, int nesting, int sequenceItem) {
				// We are only called for nesting 3, for the field 'certificates' [0] IMPLICIT
				// or for the field 'crls' [1] IMPLICIT
				if( (originalType == 0) || (originalType == 1 )){
					return ASN1Decoder.SET;
				} else {
					return originalType;
				}
			};
		};
		decoder.configureTypeRedirection (3, mapper); // when nesting == 3

		// For SignedData, field eContent in EncapsulatedContentInfo
		mapper = new TypeMapper() {
			public int map (int originalType, int nesting, int sequenceItem) {
				// We are only called for nesting 4, for the field 'eContent' [0] EXPLICIT
				if( originalType == 0 ){
					return ASN1Decoder.SEQUENCE;
				} else {
					return originalType;
				}
			};
		};
		decoder.configureTypeRedirection (4, mapper); // when nesting == 4

		mapper = new TypeMapper() {
			public int map (int originalType, int nesting, int sequenceItem) {
				if(( originalType == 0 )&&(sequenceItem==4)){
					// For 'signedAttrs [0] IMPLICIT SignedAttributes OPTIONAL' in singerInfo in SignedData
					return ASN1Decoder.SET;
				} else if( originalType == 0 ){
					// For ExtendedCertificateOrCertificate, field certificates
					return ASN1Decoder.SEQUENCE;
				} else {
					return originalType;
				}
			};
		};
		decoder.configureTypeRedirection (5, mapper); // when nesting == 5

		// This is for the X509 certificate itself
		decoder.configureTypeRedirection (6, X509CertImpl.X509_MAPPER); // when nesting == 6
		return decoder;
}

/**
 * Constructs a PKCS7 structure for a SignedData object from a <code>List</code> of
 * <code>Certificates</code>.  Supports getting encoding for <code>CertPath</code>.
 * certificates is the only field to be populated.
 *
 * @param certificates List of <code>Certificate</code>s
 * @return SignedData with certificates field populated with <code>List</code>'s contents
 */
public static byte[] encodeCertList( List certificates ) throws ASN1Exception{

    // Build the set of certificates for the certificates field of the PKCS7 SignedData object
    Object[] certNodes = new Object[ certificates.size() ];

    int i = 0;
    for (Iterator iter = certificates.iterator(); iter.hasNext();) {
        try {
            java.security.cert.Certificate cert = (java.security.cert.Certificate) iter.next();
            byte[] certEncoding = cert.getEncoded();
            ASN1Decoder.Data newCertNode = new ASN1Decoder.Data(certEncoding );
            certNodes[ i++ ] = newCertNode;
        } catch (CertificateEncodingException e) {
            return null;
        }
    }

    ASN1Decoder.CertificateSet certSet = new ASN1Decoder.CertificateSet( certNodes );

    // Build the fields of the data
    Object[] fields = new Object[ 5 ];

    // version
    fields[ 0 ] = BigInteger.ONE;

    // digest algorithm identifiers
    fields[ 1 ] = new ASN1Decoder.Set2( new Object[0]);

    // content info
    fields[ 2 ] = DATA_OID;

    // certificates
    fields[ 3 ] = certSet;

    // SignerInfos
    fields[ 4 ] = new ASN1Decoder.Set2( new Object[0]);

    // Build the PKCS7 structure
    Object[] data = new Object[ 1 ];
    data[ 0 ] = fields;

    Object[] PKCS7Object = new Object[ 2 ];
    PKCS7Object[ 0 ] = SIGNED_DATA_OID;
    PKCS7Object[ 1 ] = data;

    ByteArrayOutputStream os = new ByteArrayOutputStream();
    ASN1Encoder encoder = new ASN1Encoder( os );
    encoder.writeObject( PKCS7Object );
    byte[] encodedBytes = os.toByteArray();

    return encodedBytes;

}

public static byte[] getASN1DEREncoded(
		            java.security.cert.Certificate[] signerCertChain,
		            String digestAlgorithm, String encryptionAlgorithm,
		            byte[] encryptedDigest) throws IOException {

	// just call other method indicating that we should include the certificates and are not including
	// the data that was signed
	return getASN1DEREncoded(signerCertChain, digestAlgorithm, encryptionAlgorithm, encryptedDigest, true, null,null);
}

    /**
     * Convenience method used to obtain the ASN1 DER encoded form of PKCS7
     * signed data. Introduced for the jarsigner tool which uses the encoded
     * form as the contents of the signature block file generated in the signing
     * of a jar file.
     *
     * @param signerCertChain
     *            the certificate chain for the signer
     * @param digestAlgorithm
     *            the name of the algorithm that was used for taking the message
     *            digest of the signature file generated in the course of the
     *            current signing attempt.
     * @param encryptionAlgorithm
     *            the name of the encryption algorithm that was used to sign the
     *            message digest value of the signature file generated in the
     *            course of the current signing attempt.
     * @param encryptedDigest
     *            the product of signing the generated signature file associated
     *            with the current signing attempt.
     * @return a <code>byte array</code> containing the ASN1 DER form of the
     *         PKCS7 signed data.
     * @throws IOException
     */
    public static byte[] getASN1DEREncoded(
            java.security.cert.Certificate[] signerCertChain,
            String digestAlgorithm, String encryptionAlgorithm,
            byte[] encryptedDigest,
            boolean includeCertificates,
            byte[] dataSigned,
            ASN1Decoder.Set2 signedAttributes) throws IOException {

        // Build the set of certificates for the certificates field of the PKCS7
        // SignedData object
        Object[] certNodes = new Object[signerCertChain.length];
        for (int i = 0; i < signerCertChain.length; i++) {
            try {
                java.security.cert.Certificate cert = signerCertChain[i];
                byte[] certEncoding = cert.getEncoded();
                ASN1Decoder.Data newCertNode = new ASN1Decoder.Data(
                        certEncoding);
 				certNodes[i] = newCertNode;
            } catch (CertificateEncodingException e) {
                return null;
            }
        }// end for

        ASN1Decoder.CertificateSet certSet = new ASN1Decoder.CertificateSet(
                certNodes);

        // Build the fields of the data
		Object[] fields = null;
		if (includeCertificates == true){
			fields = new Object[5];
		} else {
			fields = new Object[4];
		}

        // version
        fields[0] = BigInteger.ONE;

        // digest algorithm identifiers - a set of size 1 whose data is a
        // sequence of size 2.
        Object[] da = new Object[2];
        da[0] = algorithmNameToOID(digestAlgorithm);
        da[1] = null;
        fields[1] = new ASN1Decoder.Set2(new Object[] { da });

        // content info - as sequence of size 1.
		Object[] contentInfo = null;
		if (dataSigned == null){
			contentInfo = new Object[] { DATA_OID };
		} else {
			contentInfo = new Object[2];
			contentInfo[0] = DATA_OID;
			ASN1Decoder.Explicit explicitData = new ASN1Decoder.Explicit(dataSigned);
			contentInfo[1] = explicitData;
		}
		fields[2] = contentInfo;

        // certificates
		int signerInfosIndex = 3;
		if (includeCertificates == true){
			fields[3] = certSet;
			signerInfosIndex = 4;
		}

        // SignerInfos -- use the ASN1Decoder.Data type as we retrieve this
        // ready-encoded.
        java.security.cert.X509Certificate signerCert = null;
        if ((signerCertChain != null) && signerCertChain.length > 0) {
            signerCert = (java.security.cert.X509Certificate)signerCertChain[0];
        } else {
            return null;
        }

        byte[] encodedSignerInfo = SignerInfo.getASN1DEREncoded(signerCert,
                digestAlgorithm, encryptionAlgorithm, encryptedDigest,signedAttributes);

		// index into fields array depends on whether or not we include the certificates
	    fields[signerInfosIndex] = new ASN1Decoder.Set2(new Object[] { new ASN1Decoder.Data(
                encodedSignerInfo) });

        // Build the PKCS7 structure
        Object[] data = new Object[1];
        data[0] = fields;

        Object[] PKCS7Object = new Object[2];
        PKCS7Object[0] = SIGNED_DATA_OID;
        PKCS7Object[1] = data;

        ByteArrayOutputStream os = new ByteArrayOutputStream();
        ASN1Encoder encoder = new ASN1Encoder(os);
        encoder.writeObject(PKCS7Object);
        byte[] encodedBytes = os.toByteArray();

        return encodedBytes;
    }

    // Perhaps this method would be better off being
    // a public static method on X509Certificate (the owner of the OID_DATABASE).
    /**
     * Convenience method that returns the OID value corresponding to
     *
     * @param algorithmName
     *            the name of a security algorithm.
     * @return the object identifier value that corresponds to the named
     *         algorithm.
     * @throws IllegalArgumentException
     *             if the supplied algorithm name is not supported.
     */
    static int[] algorithmNameToOID(String algorithmName) {
        String oidString = OIDDatabase.getInstance().getFirstOIDForAlgorithm(
                algorithmName);
        return ASN1OID.stringToIntOID(oidString);
    }

public PKCS7(InputStream encodedStream) {
	decoder = decoder (encodedStream);
	try {
		this.decoder.collectBytes(true);
		contents = this.decoder.readContents();
		this.encoded = this.decoder.collectedBytes();
	} catch (ASN1Exception e) {
		throw new IllegalArgumentException();
	}
	if (!contentsIsPKCS7()) {
		throw new IllegalArgumentException();
	}
}

/**
 * Constructs a new instance of this class.
 *
 * @param	encoded		byte[]	the encoded PKCS7 data
 */
public PKCS7(byte[] encoded) throws IllegalArgumentException {
	ByteArrayInputStream in = new ByteArrayInputStream(encoded);
	decoder = decoder (in);
	try {
		contents = this.decoder.readContents();
		this.encoded = encoded;
	} catch (ASN1Exception e) {
		throw new IllegalArgumentException();
	}
	if (!contentsIsPKCS7())
		throw new IllegalArgumentException();
}

	static final int[] DATA_OID = {1,2,840,113549,1,7,1};	// OID that indicates this is a 'DATA' ASN.1 type
	static final int[] SIGNED_DATA_OID = {1,2,840,113549,1,7,2};	// OID that indicates this is a 'SIGNED DATA' ASN.1 type
	static final int[] ENVELOPED_DATA_OID = {1,2,840,113549,1,7,3};	// OID that indicates this is a 'ENVELOPED DATA' ASN.1 type
	static final int[] SIGNED_AND_ENVELOPED_DATA_OID = {1,2,840,113549,1,7,4};	// OID that indicates this is a 'SIGNED AND ENVELOPED DATA' ASN.1 type
	static final int[] DIGESTED_DATA_OID = {1,2,840,113549,1,7,5};	// OID that indicates this is a 'DIGESTED DATA' ASN.1 type
	static final int[] ENCRYPTED_DATA_OID = {1,2,840,113549,1,7,6};	// OID that indicates this is a 'ENCRYPTED DATA' ASN.1 type

	static final int[][] POSSIBLE_CONTENTS;	// These are the possible options for a PKCS7 ASN.1 structure
	static {
		POSSIBLE_CONTENTS = new int[6][];
		POSSIBLE_CONTENTS[0] = DATA_OID;
		POSSIBLE_CONTENTS[1] = SIGNED_DATA_OID;
		POSSIBLE_CONTENTS[2] = ENVELOPED_DATA_OID;
		POSSIBLE_CONTENTS[3] = SIGNED_AND_ENVELOPED_DATA_OID;
		POSSIBLE_CONTENTS[4] = DIGESTED_DATA_OID;
		POSSIBLE_CONTENTS[5] = ENCRYPTED_DATA_OID;
	}

/**
 * Answers true if the receiver contains a SignedData object, false otherwise.
 *
 * @return	true if the receiver contains a SignedData object, false otherwise.
 */

public boolean isSignedData () {
	try {
		Node subnode = (Node)contents.subnode (0);
		if (subnode==null) return false;
		int[] id = (int[]) subnode.data;
		return Arrays.equals (SIGNED_DATA_OID, id);
	} catch (ClassCastException e) {
		return false;
	}
}

/**
 * This class implements a SignedData object, usually contained in
 * a PKCS7 ASN.1 structure.
 */
public static class SignedData {
	private ASN1Decoder.Node contents;	// The actual ASN.1 Node with the representation for the SignedData
	private byte[] encoded;

	/**
	 * Constructs a new instance of this class.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @param	contents	Node	The ASN.1 representation for the node
	 */
	SignedData (ASN1Decoder.Node contents, byte[] encoded) {
		this.contents = contents;
		this.encoded = encoded;
	}

	/**
	 * Return a collection of SignerInfo information in the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	a collection (array) of SignerInfo information in the receiver
	 */
	public SignerInfo[] signerInfos() {
		try {
			Node[] signedDataContents = (Node[])contents.data;
			Node[] signerInfos = (Node[])signedDataContents[signedDataContents.length-1].data;
			SignerInfo[] result = new SignerInfo[signerInfos.length];
			for (int i = 0; i < signerInfos.length; i++) {
				result[i] = new SignerInfo (signerInfos[i]);
			}
			return result;
		} catch (ArrayIndexOutOfBoundsException e) {
			return null;
		} catch (ClassCastException e) {
			return null;
		}
	}

	public Vector certificates() throws CertificateException {
		Vector certificates = new Vector();
		try {
			Node certSequence = ((Node[])contents.data)[3];
			if (certSequence.originalType != 0) {
				// there are no certificates
				return certificates;
			}

			Node[] certNodes = ((Node[])certSequence.data);
			for(int i = 0; i < certNodes.length; i++) {
				certificates.add(X509Certificate.certificateFromASN1Object(certNodes[i], encoded));
			}
			return certificates;
		} catch (ArrayIndexOutOfBoundsException e) {
			return certificates;
		} catch (ClassCastException e) {
			return certificates;
		}
	}

	public Vector CRLs() throws CRLException {
		Vector crls = new Vector();
		try {
			Node crlSequence = ((Node[])contents.data)[3];
			if (crlSequence.originalType != 1) {
				crlSequence = ((Node[])contents.data)[4];
				if (crlSequence.originalType != 1) {
					// there are no CRLs
					return crls;
				}
			}

			Node[] certNodes = ((Node[])crlSequence.data);
			for(int i = 0; i < certNodes.length; i++) {
				crls.add(X509CRL.CRLFromASN1Object(certNodes[i], encoded));
			}
			return crls;
		} catch (ArrayIndexOutOfBoundsException e) {
			return crls;
		} catch (ClassCastException e) {
			return crls;
		}
	}
}

/**
 * This class implements a SignerInfo object, usually contained in
 * a SignedData PKCS7 ASN.1 structure.
 */
public static class SignerInfo {
	private ASN1Decoder.Node contents;	// The actual ASN.1 Node with the representation for the SignerInfo
	private Principal issuer = null;  // The entity that issued the certificate that signed the digest
	private BigInteger serialNumber = null;  // The issuer-specific serial number of the certificate that signed the certificate
	private static final int[] ATTRIBUTE_MESSAGE_DIGEST = new int[] { 1, 2,
                840, 113549, 1, 9, 4 };
	private ASN1Decoder.Set2 signedAttributes = null;

	/**
	 * Constructs a new instance of this class.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @param	contents	Node	The ASN.1 representation for the node
	 */
	SignerInfo (ASN1Decoder.Node contents) {
		this.contents = contents;
	}

	/**
	 * Return the encryptedDigest field for the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	byte[]	the encryptedDigest field for the receiver
	 */
	public byte[] encryptedDigest() {
		try {
			Node[] signerInfoContents = (Node[])contents.data;
			int index = 0;
			while (signerInfoContents[index].type != ASN1Decoder.OCTET_STRING) index++;
			return (byte[]) signerInfoContents[index].data;
		} catch (ArrayIndexOutOfBoundsException e) {
			return null;
		} catch (ClassCastException e) {
			return null;
		}
	}

	/**
	 * Return the digestEncryptionAlgorithm field for the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	byte[]	the digestEncryptionAlgorithm field for the receiver
	 */
	public String digestEncryptionAlgorithm() {
		int[] oid = digestEncryptionAlgorithmOID();
		if (oid==null) return null;
		String oidAsStr = ASN1OID.OIDToString (oid);
		return X509Certificate.getAlias("Alg.Alias.KeyFactory.", oidAsStr, null);
	}

	/**
	 * Return the digestAlgorithm field for the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	byte[]	the digestAlgorithm field for the receiver
	 */
	public String digestAlgorithm() {
		int[] oid = digestAlgorithmOID();
		if (oid==null) return null;
		String oidAsStr = ASN1OID.OIDToString (oid);
		return X509Certificate.getAlias("Alg.Alias.MessageDigest.", oidAsStr, null);
	}

	/**
	 * Return the digest encryption algorithm OID for the receiver, as an int[]
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	int[]	the digest encryption algorithm OID for the receiver
	 */
	public int[] digestEncryptionAlgorithmOID() {
		try {
			Node[] signerInfoContents = (Node[])contents.data;
			int index = 0;
			while (signerInfoContents[index].type != ASN1Decoder.OCTET_STRING) index++;
			index--; // alg. comes right before the octet string
			Node[] algData = (Node[])signerInfoContents[index].data;
			return (int[]) algData[0].data;
		} catch (ArrayIndexOutOfBoundsException e) {
			return null;
		} catch (ClassCastException e) {
			return null;
		}
	}

	/**
	 * Return the digest algorithm OID for the receiver, as an int[]
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	int[]	the digest algorithm OID for the receiver
	 */
	public int[] digestAlgorithmOID() {
		try {
			Node[] signerInfoContents = (Node[])contents.data;
			Node[] algData = (Node[])signerInfoContents[2].data;
			return (int[]) algData[0].data;
		} catch (ArrayIndexOutOfBoundsException e) {
			return null;
		} catch (ClassCastException e) {
			return null;
		}
	}

	/**
	 * Return the signature algorithm name for the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return	String	the signature algorithm name for the receiver
	 */
	public String signatureName() {
		String digestAlgorithm = digestAlgorithm();
		String digestEncryptionAlgorithm = digestEncryptionAlgorithm();
		if (digestAlgorithm==null || digestEncryptionAlgorithm==null) return null;
		return digestAlgorithm + "with" + digestEncryptionAlgorithm;

	}

	/**
	 * Return the authenticated attributes for the receiver
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return ASN1Decoder.Node[]  an array with each element representing
	 * 				an Attribute in the authenticatedAttributes SET.
	 */
	public ASN1Decoder.Node[] authenticatedAttributes() {
		// the authenticatedAttributes field is encoded with the [0] IMPLICIT tag
		ASN1Decoder.Node attributeSetNode = contents.subnodeWithOriginalType(0);
		if (attributeSetNode == null || attributeSetNode.data == null)
			return new ASN1Decoder.Node[0];
		return (ASN1Decoder.Node[])attributeSetNode.data;
	}

	/**
	 * Return the value of the message-digest attribute (in the authenticatedAttributes field)
	 * for the content that was signed.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @return byte[]	The message digest or null if the attribute wasn't specified.
	 */
	public byte[] contentMessageDigest() {
		ASN1Decoder.Node[] attributes = authenticatedAttributes();
		for(int i = 0; i < attributes.length; i++) {
			int[] attributeOid = (int[])attributes[i].subnode(0).data;
			if (Arrays.equals(attributeOid, ATTRIBUTE_MESSAGE_DIGEST)) {
				ASN1Decoder.Node value = attributes[i].subnode(1).subnode(0);
				return (byte[])value.data;
			}
		}
		return null;
	}

	/**
	 * Return the issuer of the certificate that signed the data
	 *
	 * @author OTI
	 *
	 * @return Principal issuer of the signing certificate
	 */
	public Principal getIssuer() {
		if (this.issuer == null) {
			X500Principal principal = new X500Principal(contents.subnode(1).subnode(0));
			this.issuer = principal;
		}
		return this.issuer;
	}

	/**
	 * Returns the serialNumber.
	 * @return BigInteger
	 */
	public BigInteger getSerialNumber() {

		if( this.serialNumber == null ){
			serialNumber = (BigInteger) contents.subnode( 1 ).subnode( 1 ).data;
		}

		return serialNumber;
	}

    private String digestAlgorithmName;

    private String encryptionAlgorithmName;

    private byte[] encryptedDigest;

    /**
     * Constructs a new instance of this class which can be used to provide an
     * ASN.1 encoded version through the {@link #getEncoded()} method.
     *
     * @param signerCert
     * @param digestAlgorithm
     * @param digest
     * @param encryptionAlgorithm
     * @param encryptedDigest
     * @param signedAttributes the signed attributes for the SingerInfo, if present
     */
    private SignerInfo (
            java.security.cert.X509Certificate signerCert,
            String digestAlgorithm,
            String encryptionAlgorithm,
            byte[] encryptedDigest,
            ASN1Decoder.Set2 signedAttributes) {
        this.digestAlgorithmName = digestAlgorithm;
        this.encryptionAlgorithmName = encryptionAlgorithm;
        this.encryptedDigest = encryptedDigest;
		this.signedAttributes = signedAttributes;
        this.issuer = signerCert.getIssuerDN();
		this.serialNumber = signerCert.getSerialNumber();
    }

    /**
     * Get the ASN.1 encoding of the SignedInfo
     *
     * @return byte array containing the encoded structure
     * @throws IOException
     *             an error occurred while attempting to encode the structure
     */
    private byte[] getEncoded() throws IOException {
        Object[] encSignedInfo;
		int signedAttributesOffset = 0;
		if(signedAttributes == null){
			encSignedInfo= new Object[5];
		} else {
			encSignedInfo= new Object[6];
			signedAttributesOffset = 1;
		}

        // version
        encSignedInfo[0] = BigInteger.ONE;

        // issuer and serial number...
        Object[] encIASN = new Object[2];

		 // issuer
		/* MHD - Design 344 for the SATSA tck there is a dependency on the ordering of the elements in the name
		 * Doing getName() ends up re-ordering the elements and causing failures.  While this should
		 * probably not be the case, to pass the tck we now only do the getName() when needed to support
		 * a class which is not of our type
		 */
		if (!(this.issuer instanceof com.ibm.oti.security.provider.X500Principal )){
			encIASN[0] = new ASN1Decoder.Data(new com.ibm.oti.security.provider.X500Principal(this.issuer.getName()).getEncoded());
		} else {
			encIASN[0] = new ASN1Decoder.Data(((X500Principal) this.issuer).getEncoded());
		}

        // serial number
        encIASN[1] = this.serialNumber;

        // Add issuer and serial number to the results
        encSignedInfo[1] = encIASN;

        // digest algorithm
        // -- need to convert the string identifier of the digest algorithm
        // into an OID that can be encoded.
        Object[] da = new Object[2];
        da[0] = algorithmNameToOID(this.digestAlgorithmName);
        da[1] = null;
        encSignedInfo[2] = da;

        // Add signed attributes if they are present, if so they are
		// implicitely tagged with 0
		if (signedAttributes != null){
			encSignedInfo[3] = new ASN1Decoder.ImplicitSet(signedAttributes,0);
		}

        // digest encryption algorithm
        // -- need to convert the string identifier of the digest algorithm
        // into an OID that can be encoded.
        Object[] dea = new Object[2];
        dea[0] = algorithmNameToOID(this.encryptionAlgorithmName);
        dea[1] = null;
        encSignedInfo[3+signedAttributesOffset] = dea;

        // encrypted digest
        encSignedInfo[4+signedAttributesOffset] = this.encryptedDigest;

        // NOTE : no unauthenticated attributes (which is optional anyway)

        // Return to caller the encoded version of encSignedInfo
        byte[] encoding = ASN1Encoder.getEncoding( encSignedInfo );
        return encoding;
    }

    /**
     * Convenience method
     *
     * @param signerCert
     * @param digestAlgorithm
     * @param encryptionAlgorithm
     * @param encryptedDigest
     * @return
     * @throws IOException
     */
    public static byte[] getASN1DEREncoded(
            java.security.cert.X509Certificate signerCert,
            String digestAlgorithm,
            String encryptionAlgorithm,
            byte[] encryptedDigest,
            ASN1Decoder.Set2 signedAttributes) throws IOException {
        return new SignerInfo(signerCert, digestAlgorithm,
                    encryptionAlgorithm, encryptedDigest, signedAttributes).getEncoded();
    }
}// end class SignerInfo

/**
 * Answers a SignedData object contained in this PKCS7 structure, or null
 * if the receiver does not represent a SignedData object
 *
 * @return	SignedData  The SignedData object contained in this PKCS7 structure, or null
 * 						if the receiver does not represent a SignedData object
 */
public SignedData signedData () {
	try {
		if (!isSignedData()) return null;
		int[] coords = {1,0};
		return new SignedData( (Node) contents.subnode (coords), encoded);
	} catch (ArrayIndexOutOfBoundsException e) {
		return null;
	} catch (ClassCastException e) {
		return null;
	}
}

/**
 * Answers an ASN.1 Node which was built as a result of parsing the input
 *
 * @return	an ASN.1 Node which was built as a result of parsing the input
 */
public Node contents () {
	return contents;
}

/**
 * Answers true is the input really contained PKCS7 data; false otherwise.
 *
 * @return	boolean	true is the input really contained PKCS7 data; false otherwise.
 */
boolean contentsIsPKCS7 () {
	int[] id = null;
	try {
		Node subnode = (Node)contents.subnode (0);
		if (subnode==null) return false;
		id = (int[]) subnode.data;
	} catch (ClassCastException e) {
		return false;
	}

	for (int i=0; i < POSSIBLE_CONTENTS.length; i++) {
		if ( Arrays.equals (POSSIBLE_CONTENTS[i], id)) return true;
	}
	return false;
}

/**
 * This method returns the signed attributes for a PKCS7 object.  Currently
 * the only attributes includes are the signingtime and the required
 * messagedigest and content type attributes
 *
 * @param data   	the data for the PKCS7 object
 * @param digestAlg	the digest algorithm for the pKCS7 object
 * @return the attribues as a Set2
 */
public static ASN1Decoder.Set2 getSignedAttributes(byte[] data, String digestAlg){

	//byte[] dataToDigest = ASN1Encoder.getEncoding(data);
	byte[] dataToDigest = data;

	// calculate the message digest of the data
	byte[] messageDigest = null;
	if (digestAlg.equals("SHA")){
		messageDigest = new byte[ 20 ];
		CL3.sha( null, dataToDigest, 0, dataToDigest.length, messageDigest, 0 );
	} else if (digestAlg.equals("MD5")) {
		messageDigest = new byte[ 16 ];
		CL3.md5( null, dataToDigest, 0, dataToDigest.length, messageDigest, 0 );
	}

	// create the contenttype attribute.  It is always id-data
	Object[] contenttype = new Object[2];
	Object[] contenttypeAttributeSet = new Object[1];
	contenttypeAttributeSet[0] =  DATA_OID ;
	contenttype[0] = ASN1OID.stringToIntOID("1.2.840.113549.1.9.3");
	contenttype[1] = new ASN1Decoder.Set2(contenttypeAttributeSet);

	// create messagedigest attribute with digest computed on the data passed in
	Object[] messagedigest = new Object[2];
	Object[] messagedigestAttributeSet = new Object[1];
	messagedigestAttributeSet[0] = messageDigest;
	messagedigest[0] = ASN1OID.stringToIntOID("1.2.840.113549.1.9.4");
	messagedigest[1] =  new ASN1Decoder.Set2(messagedigestAttributeSet);

	// create signingTime attribute with time as current time
	Object[] signingTime = new Object[2];
	Object[] signingTimeAttributeSet = new Object[1];
	signingTimeAttributeSet[0] = new ASN1Decoder.UTCTime(new Date());
	signingTime[0] = ASN1OID.stringToIntOID("1.2.840.113549.1.9.5");
	signingTime[1] = new ASN1Decoder.Set2(signingTimeAttributeSet);

	return new ASN1Decoder.Set2(new Object[] {contenttype, messagedigest,signingTime } );
};

/**
 * This method returns the signed attributes in a for which is suitable for digesting and
 * signing
 *
 * @param signedAttributes the signed attributes as a set
 * @return ANS1 DER encoded bytes for the signed attributes
 */
public static byte[] encodeSignedAttributesForDigest(ASN1Decoder.Set2 signedAttributes){
	byte result[] = ASN1Encoder.getEncoding(signedAttributes);
	return result;
}

}
