
package com.ibm.oti.security.provider;
/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 *	This class represents X.509 Certificate Revocation Lists (CRLs). They are used
 * to indicate that a given Certificate has expired already. This is the ASN.1
 * definition of a CRL
 *
 *
 *
 *   CertificateList  ::=  SEQUENCE  {
 *        tbsCertList          TBSCertList,
 *        signatureAlgorithm   AlgorithmIdentifier,
 *        signatureValue       BIT STRING  }
 *
 *   TBSCertList  ::=  SEQUENCE  {
 *        version                 Version OPTIONAL,
 *                                     -- if present, shall be v2
 *        signature               AlgorithmIdentifier,
 *        issuer                  Name,
 *        thisUpdate              Time,
 *        nextUpdate              Time OPTIONAL,
 *        revokedCertificates     SEQUENCE OF SEQUENCE  {
 *             userCertificate         CertificateSerialNumber,
 *             revocationDate          Time,
 *             crlEntryExtensions      Extensions OPTIONAL
 *                                           -- if present, shall be v2
 *                                  }  OPTIONAL,
 *        crlExtensions           [0]  EXPLICIT Extensions OPTIONAL
 *                                           -- if present, shall be v2
 *                                  }
 *
 * @author		OTI
 * @version		initial
 *
 * @see CRL
 */

import java.io.InputStream;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.Signature;
import java.security.SignatureException;
import java.security.cert.CRL;
import java.security.cert.CRLException;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import com.ibm.oti.util.ASN1Decoder;
import com.ibm.oti.util.ASN1Encoder;
import com.ibm.oti.util.ASN1Exception;
import com.ibm.oti.util.ASN1Decoder.BitString;
import com.ibm.oti.util.ASN1Decoder.Node;
import com.ibm.oti.util.ASN1Decoder.TypeMapper;

public class X509CRL extends java.security.cert.X509CRL {

	private byte[]	encoded; 		// ASN.1 DER encoded version of the CRL
	private int originalOffsetIntoRawBytes; // At which offset of the original byte[] the encoded version above really starts
	private Node tbsNode;			// Used to implement the API to fetch the encoded version of the tbsCertList
	private int version;			// Version of this CRL (0=v1, 1=v2, 2=v3)

	// The following ones represent the signature
	private String sigAlgOID;			// Algorithm ID for the signature (OID)
	private byte[] sigAlgParams;		// Algorithm parameters for the signature
	private byte[]	rawSignature;		// Actual raw bytes for the signature
	private String signatureName;			// JCA algorithm name for CRL signature
	private String signatureProviderName;	// provider name for CRL signature algorithm

	private com.ibm.oti.security.provider.X500Principal	issuer;		// Who issued this CRL

	// The following 2 are used to represent the update dates
	private Date	thisUpdate;			// When this update was made
	private Date	nextUpdate;			// When the next update is scheduled to be made

	private Hashtable revokedCertificates;	 // maps serialNumber(BigInteger)->X509CRLEntry

	// The optional extensions
	private Hashtable extensions;		// maps OID to X509Extension

/**
 * Constructs a new instance of this class
 *
 */
protected X509CRL() {
	super ();
}

/**
 * Return the CRL, DER-encoded.
 *
 * @return		byte[]		the parameters in their primary encoding format.
 * @throws	CRLException	if encoding errors are detected
 */
public  byte[] getEncoded() throws CRLException{
	return encoded;
}

/**
 * Return the issuer distinguished name for the receiver.
 *
 * @return		Principal		The entity that signed the receiver, the issuer
 */
public  java.security.Principal getIssuerDN(){
	return issuer;
}

/**
 * Return the CRL 'nextUpdate' field
 *
 * @return		Date	the CRL 'nextUpdate' field
 */
public  Date getNextUpdate(){
	return nextUpdate;
}

/**
 * Return the revoked CRL given its serial number.
 *
 * @param		serialNumber	BigInteger	The serial number of the desired Certificate
 * @return		X509CRLEntry	the revoked CRL given its serial number.
 */
public  java.security.cert.X509CRLEntry getRevokedCertificate(BigInteger serialNumber){
	if (revokedCertificates==null) return null;
	return (java.security.cert.X509CRLEntry) revokedCertificates.get (serialNumber);
}

/**
 * Return all the revoked Certificates in the receiver, as a Set of X509CRLEntry.
 *
 * @return		Set		all the revoked Certificates in the receiver, as a Set of X509CRLEntry.
 */
public  Set getRevokedCertificates(){
	if (this.revokedCertificates == null) return null;

	HashSet result = new HashSet();
	Enumeration elements = this.revokedCertificates.elements();
	while (elements.hasMoreElements()) {
		// we should collect this element
		result.add (elements.nextElement());
	}
	return result;
}

/**
 * Return the signature algorithm name for the receiver.
 *
 * @return		String	the signature algorithm name for the receiver.
 */
public  String getSigAlgName(){
	return signatureName;
}

/**
 * Return the signature algorithm ID (OID) for the receiver.
 *
 * @return		String	the signature algorithm ID (OID) for the receiver.
 */
public  String getSigAlgOID(){
	return sigAlgOID;
}

/**
 * Return the signature algorithm parameters for the receiver.
 *
 * @return		byte[]	the signature algorithm parameters for the receiver.
 */
public  byte[] getSigAlgParams(){
	return sigAlgParams;
}

/**
 * Return the actual signature for the receiver.
 *
 * @return		byte[]	the actual signature for the receiver.
 */
public  byte[] getSignature(){
	return rawSignature;
}

/**
 * Return the field 'tbsCertList' for the receiver.
 *
 * @throws	CRLException	if encoding errors are detected
 * @return		byte[]	the field 'tbsCertList' for the receiver.
 */
public  byte[] getTBSCertList() throws CRLException{
	byte[] result = new byte[tbsNode.endPosition - tbsNode.startPosition + 1];
	System.arraycopy (encoded, tbsNode.startPosition - originalOffsetIntoRawBytes, result, 0, result.length);
	return result;
}

/**
 * Return the field 'thisUpdate' for the receiver.
 *
 * @return		Date	the field 'thisUpdate' for the receiver.
 */
public  Date getThisUpdate(){
	return thisUpdate;
}

/**
 * Return the version number for the receiver.
 *
 * @return		Date	the version number for the receiver.
 */
public  int getVersion(){
	return version;
}

/**
 * Verifies if the CRL was signed with the PrivateKey corresponding to the
 * PublicKey passed.
 *
 * @param		key		PublicKey		PublicKey to be used for signature verification
 * @throws	CRLException				if encoding errors are detected
 * @throws	NoSuchAlgorithmException	if unsupported algorithms are encountered
 * @throws	InvalidKeyException			if the provided key is not valid
 * @throws	NoSuchProviderException		if a provider for the signature algorithm can not be found
 * @throws	SignatureException			if signature errors are found
 */
public  void verify(PublicKey key)
                     throws CRLException,
                            NoSuchAlgorithmException,
                            InvalidKeyException,
                            NoSuchProviderException,
                            SignatureException{

	if (signatureProviderName != null)
		verify (key, signatureProviderName);
	else {
		if (signatureName==null) // we had no map entry from ID to alg. name
			throw new NoSuchAlgorithmException(this.sigAlgOID);
		Signature signature = Signature.getInstance (signatureName);
		verify (key, signature);
	}
}

/**
 * Verifies if the CRL was signed with the PrivateKey corresponding to the
 * PublicKey passed, from the given signature provider.
 *
 * @param		key				PublicKey to be used for signature verification
 * @param		sigProvider		name of the signature provider to use
 *
 * @throws		CRLException				if encoding errors are detected
 * @throws		NoSuchAlgorithmException	if unsupported algorithms are encountered
 * @throws		InvalidKeyException			if the provided key is not valid
 * @throws		NoSuchProviderException		if a provider for the signature algorithm can not be found
 * @throws		SignatureException			if signature errors are found
 */
public  void verify(PublicKey key,
                            String sigProvider)
                     throws CRLException,
                            NoSuchAlgorithmException,
                            InvalidKeyException,
                            NoSuchProviderException,
                            SignatureException{

	if (signatureName==null) // we had no map entry from ID to alg. name
		throw new NoSuchAlgorithmException(this.sigAlgOID);
	Signature signature = Signature.getInstance (signatureName, sigProvider);
	verify(key, signature);
}

/**
 * Verifies if the CRL was signed with the PrivateKey corresponding to the
 * PublicKey passed. Uses the signature provided.
 *
 * @param		key			PublicKey	PublicKey to be used for signature verification
 * @param		signature	String		the signature provider to use
 *
 * @throws	CRLException				if encoding errors are detected
 * @throws	NoSuchAlgorithmException	if unsupported algorithms are encountered
 * @throws	InvalidKeyException			if the provided key is not valid
 * @throws	NoSuchProviderException		if a provider for the signature algorithm can not be found
 * @throws	SignatureException			if signature errors are found
 */
private void verify(PublicKey key,
                            Signature signature)
                     throws CRLException,
                            NoSuchAlgorithmException,
                            InvalidKeyException,
                            NoSuchProviderException,
                            SignatureException{

	signature.initVerify (key);
	signature.update (getTBSCertList());
	boolean ok = signature.verify (getSignature());
	if (!ok) throw new SignatureException();
}

/**
 * Answers if a given Certificate has been revoked or not.
 *
 * @param	cert	Certificate	The Certificate to test
 *
 * @return	true
 *	 			if the certificate has been revoked
 *	 		false
 *	 			if the certificate has not been revoked yet
 */
public boolean isRevoked(java.security.cert.Certificate cert) {
	try {
		java.security.cert.X509Certificate toVerify = (java.security.cert.X509Certificate) cert;
		java.security.cert.X509CRLEntry revoked = (java.security.cert.X509CRLEntry) getRevokedCertificate(toVerify.getSerialNumber());
		return revoked != null;
	} catch (ClassCastException e) { // has to be X.509 so that we can ask getSerialNumber
		return false;
	}
}

	private static final String[] SUPPORTED_CRITICAL_EXTENSIONS = {
	};

/**
 * Return true if the receiver has an unsupported, critical extension. Return
 * false otherwise.
 *
 * @return		boolean		true if the receiver has an unsupported, critical extension, false otherwise.
 */
public boolean hasUnsupportedCriticalExtension(){
	Set critical = getCriticalExtensionOIDs();
	if (critical==null) return false;
	if (critical.size() > SUPPORTED_CRITICAL_EXTENSIONS.length)
		return true;
	for (int i = 0; i < SUPPORTED_CRITICAL_EXTENSIONS.length; i++) {
		critical.remove (SUPPORTED_CRITICAL_EXTENSIONS[i]);
	}
	return critical.size() != 0;
}

/**
 * Return the Set of extensions in the receiver where isCritical matches the parameter
 *
 * @return		the Set of extensions in the receiver where isCritical matches the parameter
 */
private Set getExtensionOIDs(boolean isCritical){
	if (this.extensions == null) return null;

	HashSet result = new HashSet();
	Enumeration elements = this.extensions.elements();
	while (elements.hasMoreElements()) {
		X509Extension extension = (X509Extension) elements.nextElement();
		if (extension.isCritical() == isCritical) {
			// we should collect this element
			result.add (extension.name());
		}
	}
	return result;
}

/**
 * Return all the OID extensions marked as critical for the receiver.
 *
 * @return		Set		all the OID extensions marked as critical for the receiver.
 */
public Set getCriticalExtensionOIDs(){
	return getExtensionOIDs (true);
}

/**
 * Return all the OID extensions marked as non-critical for the receiver.
 *
 * @return		Set		all the OID extensions marked as non-critical for the receiver.
 */
public Set getNonCriticalExtensionOIDs(){
	return getExtensionOIDs (false);
}

/**
 * Return the DER-encoded OCTEC String for the extension value defined by oid in the receiver.
 *
 * @return		byte[]	the DER-encoded OCTEC String for the extension value defined by oid in the receiver.
 */
public byte[] getExtensionValue(String oid){
	if (this.extensions == null) return null;
	X509Extension extn = (X509Extension) this.extensions.get (oid);
	if (extn == null) return null;
	byte[] extnValue = extn.value();
	if (extnValue == null) extnValue = new byte[0];

	// encode the extension value with a DER OCTET STRING tag
	Node node = new ASN1Decoder.Node();
	node.type = ASN1Decoder.OCTET_STRING;
	node.data = extn.value();
	return ASN1Encoder.encodeNode(node);
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return		a printable representation for the receiver.
 */
public String toString () {
	StringBuffer result = new StringBuffer();
	result.append ("X.509 CRL v");
	result.append (getVersion());
	result.append ("\n\tThis update: ");
	result.append (getThisUpdate());
	result.append ("\n\tNext update: ");
	result.append (getNextUpdate());
	result.append ("\n\tExtensions: ");
	result.append (this.extensions==null ? 0 : this.extensions.size());
	result.append ("\n");
	return result.toString();
};

/**
 * Answers an ASN.1 decoder for X.509 CRLs
 *
 * @param		inStream	InputStream	where the ASN.1 raw bytes will be read from
 * @return		an ASN.1 decoder for X.509 CRLs
 */
public static ASN1Decoder CRLDecoder (InputStream inStream) {
		ASN1Decoder decoder = new ASN1Decoder(inStream);
		TypeMapper mapper = new TypeMapper() {
			public int map (int originalType, int nesting, int sequenceItem) {
				// We are only called for nesting 2, for the 'crlExtensions' field [0] EXPLICIT
				if( originalType == 0 ){
					return ASN1Decoder.SEQUENCE;
				} else {
					return originalType;
				}
			};
		};
		decoder.configureTypeRedirection (2, mapper); // when nesting == 2
		return decoder;
}

/**
 * Configure the revoked Certificates in the receiver according to
 * the ASN.1 Node read by a ASN.1 decoder
 *
 * @param		revokedCertificatesNode	Node	The ASN.1 node where to fetch the values from
 * @param		rawBytes	byte[]	bytes where the node was read from
 */
private void configureRevokedCertificates (Node revokedCertificatesNode, byte[] rawBytes) throws CRLException {
	Node[] allRevokedNodes = (Node[]) revokedCertificatesNode.data;
	this.revokedCertificates = new Hashtable();
	for (int i=0; i < allRevokedNodes.length; i++) {
		Node revokedCertificateNode = allRevokedNodes[i];
		X509CRLEntry entry = X509CRLEntry.X509CRLEntryFromASN1Object (revokedCertificateNode, rawBytes);
		this.revokedCertificates.put ( entry.getSerialNumber(), entry);
	}
}

/**
 * Configure the extensions in the receiver according to
 * the ASN.1 Node read by a ASN.1 decoder
 *
 * @param		extensions	Node	The ASN.1 node where to fetch the values from
 */
private void configureExtensions (Node extensions) {
	Node[] allExtensions = (Node[]) extensions.data;
	Node[] actualExtensions = (Node[]) allExtensions[0].data;
	this.extensions = new Hashtable();
	for (int i = 0; i < actualExtensions.length; i++) {
		Node extension = actualExtensions[i];
		Node[] extensionParts = (Node[]) extension.data;
		// Now we have  extnID, critical (optional), extnValue
		boolean isCritical = false;
		if (extensionParts.length > 2) {
			isCritical = ((Boolean)extensionParts[1].data).booleanValue();
		}
		ASN1OID extnID = new ASN1OID ((int[])extensionParts[0].data);
		byte[] extnValue = (byte[])extensionParts[extensionParts.length - 1].data;
		X509Extension extn = new X509Extension (extnID, extnValue, isCritical);
		this.extensions.put (extn.name(), extn);
	}

}

/**
 * Configure the signature in the receiver according to
 * the ASN.1 Node read by a ASN.1 decoder
 *
 * @param		asn1CRL	Node[]	The ASN.1 node array where to fetch the values from
 */
private void configureSignature (Node[] asn1CRL) {
	Node[] signatureAlgorithm = (Node[]) asn1CRL[1].data;
	BitString signature = (BitString) asn1CRL[2].data;
	// FIELD	signature
	this.rawSignature = signature.data;	// byte[] in BIT STRING
	this.sigAlgOID = ASN1OID.OIDToString ((int[])signatureAlgorithm[0].data);
	String[] found = new String[1];
	this.signatureName = X509Certificate.getAlias("Alg.Alias.Signature.", this.sigAlgOID, found);
	this.signatureProviderName = found[0];

	Node algParams;
	if (signatureAlgorithm.length > 1 && (algParams = signatureAlgorithm[1]).type != ASN1Decoder.NULL) {
		// Copy the bytes that form the ASN.1 object for the algorithm parameters. The public
		// API to fetch the parameters returns a byte[], so there's no need to parse. Besides, the bytes
		// depend on the OID.
		this.sigAlgParams = new byte[ algParams.endPosition - algParams.startPosition + 1];
		System.arraycopy (encoded, algParams.startPosition - originalOffsetIntoRawBytes, this.sigAlgParams, 0, this.sigAlgParams.length);
	}
}

/**
 * Return a CRL from an ASN.1 Node created with the ASN.1 decoder
 *
 * @param		node	Node	an ASN.1 Node created with the ASN.1 decoder
 * @param		rawBytes	byte[]	bytes where the node was read from
 * @return		CRL			The CRL for the generic node representation
 */

static CRL CRLFromASN1Object(Node node, byte[] rawBytes) throws CRLException {
	try {
		X509CRL result = new X509CRL();
		Node[] asn1CRL = (Node[]) node.data;	//CRL fields (tbsCertList,signatureAlgorithm,signatureValue)
		int start = node.startPosition;
		int end = node.endPosition;
		if (start==0 && end==(rawBytes.length-1)) {
			result.encoded = rawBytes;	// No need to copy bytes
		} else {
			result.encoded = new byte[end-start+1];
			System.arraycopy (rawBytes, start, result.encoded,0, result.encoded.length);
		}
		// This is needed in cases we need to get subparts of the byte[]
		// based on a ASN.1 Node. For instance, if the original rawBytes
		// have many CRLs, they start at positions other than 0. So, trying to
		// fetch subparts from the cert.encoded would be off by so many bytes.
		// That is why we need to remember by how many bytes we are off.
		result.originalOffsetIntoRawBytes = start;

		// Signature
		result.configureSignature (asn1CRL);

		// Now the contained structure, the tbsCertList
		Node tbsNode = asn1CRL[0];	// tbstbsCertList (SEQUENCE)
		Node[] tbsCertList = (Node[]) tbsNode.data; // tbscertList fields ( [version] , ...)
		result.tbsNode = tbsNode;

		// FIELD version
		int tbsFieldOffset;	// Where fields start
		if (tbsCertList[0].type == ASN1Decoder.INTEGER) {
			// Version is present.
			BigInteger version = (BigInteger) tbsCertList[0].data;
			result.version = version.intValue() + 1; // Encoded values: 0 = v1, 1 = v2, 2 = v3 , so we adjust
			tbsFieldOffset = 0;	// First field starts at 0
		} else {
			// Version is not present (optional) so it means v2
			result.version = 2;
			tbsFieldOffset = -1;	// Version is not present, so we must compensate for field offsets
		}

		// FIELD	issuer
		com.ibm.oti.security.provider.X500Principal issuer = new com.ibm.oti.security.provider.X500Principal((Node) tbsCertList[tbsFieldOffset+2]);
		result.issuer = issuer;

		// FIELDs	thisUpdate, nextUpdate
		Node updateNode = tbsCertList[tbsFieldOffset+3]; // thisUpdate
		result.thisUpdate = (Date) updateNode.data;

		if (tbsCertList.length <= tbsFieldOffset+4) return result;
		updateNode = tbsCertList[tbsFieldOffset+4]; // nextUpdate, OPTIONAL
		if (updateNode.type == ASN1Decoder.UTC_TIME || updateNode.type == ASN1Decoder.GENERALIZED_TIME ) {
			result.nextUpdate = (Date) updateNode.data;
		} else {
			tbsFieldOffset--;	// have to compensate for yet another missing (optional) field
		}

		// FIELD revokedCertificates
		if (tbsCertList.length <= tbsFieldOffset+5) return result;
		Node revokedCertificatesNode = tbsCertList[tbsFieldOffset+5];
		if (revokedCertificatesNode.originalType == ASN1Decoder.SEQUENCE) {
			result.configureRevokedCertificates (revokedCertificatesNode, rawBytes);
		} else	{
			tbsFieldOffset--;	// have to compensate for yet another missing (optional) field
		}

		// FIELD crlExtensions
		if (tbsCertList.length <= tbsFieldOffset+6) return result;
		Node crlExtensions = tbsCertList[tbsFieldOffset+6]; // crlExtensions [0] EXPLICIT ...
		if (crlExtensions.originalType==0) { // this CRL has extensions
			result.configureExtensions (crlExtensions);
		}
		return result;
	} catch (ClassCastException e) {
			throw new CRLException();
	}
}

/**
 * Return a CRL from an input stream with an encoded ASN.1 CRL
 *
 * @param		inStream	InputStream	Input where the ASN.1 data will be read from
 * @return		CRL			The CRL for the generic node representation
 */
static CRL CRLFromASN1Object(InputStream inStream) throws CRLException {
	try {
		ASN1Decoder decoder = CRLDecoder(inStream);
		decoder.collectBytes (true);
		Node node = decoder.readContents();
		X509CRL crl = (X509CRL) CRLFromASN1Object(node,decoder.collectedBytes());
		return crl;
	} catch (ASN1Exception e) {
		throw new CRLException();
	} catch (ClassCastException e) {
		throw new CRLException();
	}
}

public javax.security.auth.x500.X500Principal getIssuerX500Principal()
{
	return new javax.security.auth.x500.X500Principal( issuer.getName() );
}

}
