
package com.ibm.oti.security.provider;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

import java.util.*;
import java.math.*;
import java.security.cert.*;

import com.ibm.oti.util.ASN1Decoder;
import com.ibm.oti.util.ASN1Encoder;
import com.ibm.oti.util.ASN1Decoder.Node;

/**
 *	 Class for entries in a CRL, where the Certificates are of type X.509
 *
 * @author		OTI
 * @version		initial
 *
 * @see X509CRL
 */
public class X509CRLEntry extends java.security.cert.X509CRLEntry {

	private byte[]	encoded; 			// ASN.1 DER encoded version of the certificate
	private BigInteger	serialNumber;	// Serial number of this X509CRLEntry
	private Date		revocationDate;	// When this X509CRLEntry was revoked

	// The optional extensions
	private Hashtable extensions;			// maps OID to X509Extension

/**
 * Constructs a new instance of this class.
 *
 */
public X509CRLEntry() {
}

/**
 * Return the ASN.1 DER encoded representation for the receiver
 *
 * @throws	CRLException	if encoding problems are detected.
 * @return		byte[]		the ASN.1 DER encoded representation for the receiver
 *
 */
public  byte[] getEncoded() throws CRLException{
	return encoded;
}

/**
 * Return the serial number for the receiver
 *
 * @return		BigInteger		the serial number for the receiver
 *
 */
public  BigInteger getSerialNumber(){
	return serialNumber;
}

/**
 * Return the revocation date for the receiver.
 *
 * @return		boolean		the revocation date for the receiver.
 *
 */
public  Date getRevocationDate(){
	return revocationDate;
}

/**
 * Return true if the receiver has extensions, false otherwise.
 *
 * @return		boolean		true if the receiver has extensions, false otherwise.
 *
 */
public  boolean hasExtensions(){
	if (this.extensions == null) return false;
	return this.extensions.size() > 0;
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return		String
 *					a printable representation for the receiver.
 */
public  String toString () {
	StringBuffer result = new StringBuffer();
	result.append ("X.509 CRL Entry S/N=");
	result.append (getSerialNumber());
	result.append (" , Extensions: ");
	result.append (this.extensions==null ? 0 : this.extensions.size());
	result.append (" , Revocation: ");
	result.append (getRevocationDate());
	return result.toString();
};

	// No supported critical extensions yet
	private static final String[] SUPPORTED_CRITICAL_EXTENSIONS = {
	};

/**
 * Return true if the receiver has an unsupported, critical extension. Return
 * false otherwise.
 *
 * @return		boolean		true if the receiver has an unsupported, critical extension, false otherwise.
 */
public boolean hasUnsupportedCriticalExtension(){
	Set critical = getCriticalExtensionOIDs();
	if (critical==null) return false;
	if (critical.size() > SUPPORTED_CRITICAL_EXTENSIONS.length)
		return true;
	for (int i = 0; i < SUPPORTED_CRITICAL_EXTENSIONS.length; i++) {
		critical.remove (SUPPORTED_CRITICAL_EXTENSIONS[i]);
	}
	return critical.size() != 0;
}

/**
 * Return the Set of extensions in the receiver where isCritical matches the parameter
 *
 * @return		the Set of extensions in the receiver where isCritical matches the parameter
 */
private Set getExtensionOIDs(boolean isCritical){
	if (this.extensions == null) return null;

	HashSet result = new HashSet();
	Enumeration elements = this.extensions.elements();
	while (elements.hasMoreElements()) {
		X509Extension extension = (X509Extension) elements.nextElement();
		if (extension.isCritical() == isCritical) {
			// we should collect this element
			result.add (extension.name());
		}
	}
	return result;
}

/**
 * Return all the OID extensions marked as critical for the receiver.
 *
 * @return		Set		all the OID extensions marked as critical for the receiver.
 */
public Set getCriticalExtensionOIDs(){
	return getExtensionOIDs (true);
}

/**
 * Return all the OID extensions marked as non-critical for the receiver.
 *
 * @return		Set		all the OID extensions marked as non-critical for the receiver.
 */
public Set getNonCriticalExtensionOIDs(){
	return getExtensionOIDs (false);
}

/**
 * Return the DER-encoded OCTECT String for the extension value defined by oid in the receiver.
 *
 * @param		oid	String	The ASN.1 OID for which extension value will be fetched
 * @return		byte[]	the DER-encoded OCTECT String for the extension value defined by oid in the receiver.
 */
public byte[] getExtensionValue(String oid){
	if (this.extensions == null) return null;
	X509Extension extn = (X509Extension) this.extensions.get (oid);
	if (extn == null) return null;
	byte[] extnValue = extn.value();
	if (extnValue == null) extnValue = new byte[0];

	// encode the extension value with a DER OCTET STRING tag
	Node node = new ASN1Decoder.Node();
	node.type = ASN1Decoder.OCTET_STRING;
	node.data = extn.value();
	return ASN1Encoder.encodeNode(node);
}

/**
 * Return an X509CRLEntry from an ASN.1 Node created with the ASN.1 decoder
 *
 * @param		node	Node	an ASN.1 Node created with the ASN.1 decoder
 * @param		rawBytes	byte[]	bytes where the node was read from
 * @return		X509CRLEntry	The X509CRLEntry for teh generic node representation
 */
static X509CRLEntry X509CRLEntryFromASN1Object(Node node, byte[] rawBytes) throws CRLException {
	try {
		X509CRLEntry result = new X509CRLEntry();
		int start = node.startPosition;
		int end = node.endPosition;
		if (start==0 && end==(rawBytes.length-1)) {
			result.encoded = rawBytes;	// No need to copy bytes
		} else {
			result.encoded = new byte[end-start+1];
			System.arraycopy (rawBytes, start, result.encoded,0, result.encoded.length);
		}

		// FIELD serialVersion
		Node[] contents = (Node[]) node.data;
		result.serialNumber = (BigInteger) contents[0].data;

		// FIELD revocationDate
		result.revocationDate = (Date) contents[1].data;

		if (contents.length > 2) { // has extensions
			result.extensions = new Hashtable();
			Node[] extensions = (Node[]) contents[2].data;
			for (int i=0; i < extensions.length; i++) {
				Node extension = extensions[i];
				Node[] extensionFields = (Node[]) extension.data;
				result.extensions.put (extensionFields[0], extensionFields[1]); // OID, value (String)
			}
		}
		return result;
	} catch (ClassCastException e) {
			throw new CRLException();
	}
}

}
