
package com.ibm.oti.security.provider;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.InputStream;
import java.math.BigInteger;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Principal;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.util.List;
import java.util.Collection;
import java.security.cert.CertificateParsingException;

import java.util.Date;
import java.util.Set;

import com.ibm.oti.util.ASN1Decoder.Node;

/**
 *	This class represents X.509 Certificates. More details can be found at ftp://ftp.isi.edu/in-notes/rfc2459.txt
 *  The ASN.1 definition of a X.509 certificate is included below.
 *
 *   Certificate  ::=  SEQUENCE  {
 *        tbsCertificate       TBSCertificate,
 *        signatureAlgorithm   AlgorithmIdentifier,
 *        signatureValue       BIT STRING  }
 *
 *   TBSCertificate  ::=  SEQUENCE  {
 *        version         [0]  EXPLICIT Version DEFAULT v1,	-> SEQUENCE { Version }
 *        serialNumber         CertificateSerialNumber,
 *        signature            AlgorithmIdentifier,
 *        issuer               Name,
 *        validity             Validity,
 *        subject              Name,
 *        subjectPublicKeyInfo SubjectPublicKeyInfo,
 *        issuerUniqueID  [1]  IMPLICIT UniqueIdentifier OPTIONAL,
 *                             -- If present, version shall be v2 or v3
 *        subjectUniqueID [2]  IMPLICIT UniqueIdentifier OPTIONAL,
 *                             -- If present, version shall be v2 or v3
 *        extensions      [3]  EXPLICIT Extensions OPTIONAL
 *                             -- If present, version shall be v3
 *        }
 *
 *
 *   Version  ::=  INTEGER  {  v1(0), v2(1), v3(2)  }
 *
 *   CertificateSerialNumber  ::=  INTEGER
 *
 *   Validity ::= SEQUENCE {
 *        notBefore      Time,
 *        notAfter       Time }
 *
 *   Time ::= CHOICE {
 *        utcTime        UTCTime,
 *        generalTime    GeneralizedTime }
 *
 *   UniqueIdentifier  ::=  BIT STRING
 *
 *   SubjectPublicKeyInfo  ::=  SEQUENCE  {
 *        algorithm            AlgorithmIdentifier,
 *        subjectPublicKey     BIT STRING  }
 *
 *   Extensions  ::=  SEQUENCE SIZE (1..MAX) OF Extension
 *
 *   Extension  ::=  SEQUENCE  {
 *        extnID      OBJECT IDENTIFIER,
 *        critical    BOOLEAN DEFAULT FALSE,
 *        extnValue   OCTET STRING  }
 *
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @see java.security.cert.Certificate
 */
public class X509Certificate extends java.security.cert.X509Certificate {

	private final X509CertImpl impl;

	static final String extendedKeyUsageOID = "2.5.29.37";
	static final String issuerAltNamesOID = "2.5.29.18";
	static final String subjectAltNamesOID = "2.5.29.17";

	public static final java.util.Hashtable OID_DATABASE = new java.util.Hashtable(5);
	static {
		// encryption algorithms
		OID_DATABASE.put("1.2.840.113549.1.1.2", "MD2withRSA");
		OID_DATABASE.put("1.2.840.113549.1.1.4", "MD5withRSA");
		OID_DATABASE.put("1.2.840.113549.1.1.5", "SHA1withRSA");
		OID_DATABASE.put("1.2.840.10040.4.3", "SHA1withDSA");

		// message digest algorithms
		OID_DATABASE.put("1.3.14.3.2.26", "SHA");
		OID_DATABASE.put("1.2.840.113549.2.5", "MD5");
		OID_DATABASE.put("1.2.840.113549.2.2", "MD2");

		// key factories
		OID_DATABASE.put("1.2.840.113549.1.1.1", "RSA");
		OID_DATABASE.put("1.2.840.10040.4.1", "DSA");
		OID_DATABASE.put("1.3.14.3.2.12", "DSA");
		OID_DATABASE.put("1.2.840.10046.2.1", "DiffieHellman");
	}

	public X509Certificate( X509CertImpl impl ){
		this.impl = impl;
	}

	/**
	 * Checks if the certificate represented by the receiver is still valid.
	 * Throws an exception if the certificate is not valid.
	 *
	 * @throws	CertificateExpiredException		The certificate is no longer valid, it has expired
	 * @throws	CertificateNotYetValidException	The certificate is not valid yet
	 */
	public void checkValidity()
		throws CertificateExpiredException, CertificateNotYetValidException {
		impl.checkValidity (new Date());
	}
	/**
	 * Checks if the certificate represented by the receiver is valid for the
	 * date provided. Throws an exception if the certificate is not valid.
	 *
	 * @param		date	Date	date to use when testing the validity of the certificate
	 * @throws	CertificateExpiredException		The certificate is no longer valid, it has expired
	 * @throws	CertificateNotYetValidException	The certificate is not valid yet
	 */
	public void checkValidity(Date date)
		throws CertificateExpiredException, CertificateNotYetValidException {
		impl.checkValidity( date );
	}

	private static final String OID_BasicConstraints = "2.5.29.19";

	/**
	 * Return the path length constraint for the BasicConstraints for the receiver.
	 * The value comes from the BasicConstraints extension (OID = 2.5.29.19)
	 *
	 * @return		int		the path length constraint for the receiver if BasicConstraints is defined.
	 * 						Return -1 otherwise.
	 *
	 */
	public int getBasicConstraints() {
		return impl.getBasicConstraints();
	}

	/**
	 * Return the issuer distinguished name for the receiver
	 *
	 * @return		Principal	the issuer distinguished name for the receiver
	 */
	public java.security.Principal getIssuerDN() {
		return impl.getIssuerDN();
	}

	public javax.security.auth.x500.X500Principal getIssuerX500Principal(){
		return new javax.security.auth.x500.X500Principal( impl.getIssuerDN().getName() );
	}

	/**
	 * Return the issuer unique ID for the receiver
	 *
	 * @return		boolean[]	the issuer unique ID for the receiver. null if not present.
	 */
	public boolean[] getIssuerUniqueID() {
		return impl.getIssuerUniqueID();
	}

	private static final String OID_KeyUsage = "2.5.29.15";
	/**
	 * Return a bit array representation for the key usage for the receiver. The
	 * value comes from the KeyUsage extension (OID = 2.5.29.15)
	 *
	 * @return		boolean[]	 a bit array representation for the key usage for the receiver.
	 */
	public boolean[] getKeyUsage() {
		return impl.getKeyUsage();
	}

	/**
	 * Return the Date until when the certificate is valid
	 *
	 * @return		Date		the Date until when the certificate is valid
	 */
	public Date getNotAfter() {
		return impl.getNotAfter();
	}

	/**
	 * Return the Date when the certificate starts to be valid
	 *
	 * @return		Date		the Date when the certificate starts to be valid
	 */
	public Date getNotBefore() {
		return impl.getNotBefore();
	}

	/**
	 * Return the serial number for the receiver
	 *
	 * @return		BigInteger	the serial number for the receiver
	 */
	public BigInteger getSerialNumber() {
		return impl.getSerialNumber();
	}
	/**
	 * Return the certificate signature algorithm name for the receiver.
	 *
	 * @return		String		the certificate signature algorithm name for the receiver.
	 */
	public String getSigAlgName() {
		return impl.getSigAlgName();
	}
	/**
	 * Return the certificate signature OID for the receiver.
	 *
	 * @return		String		the certificate signature OID for the receiver.
	 */
	public String getSigAlgOID() {
		return impl.getSigAlgOID();
	}

	/**
	 * Return the signature algorithm parameters for the receiver, DER-encoded.
	 *
	 * @return		byte[]		the signature algorithm parameters for the receiver, DER-encoded. null if not present.
	 */
	public byte[] getSigAlgParams() {
		return impl.getSigAlgParams();
	}
	/**
	 * Return the actual signature value for the receiver
	 *
	 * @return		byte[]		the actual signature value for the receiver
	 */
	public byte[] getSignature() {
		return impl.getSignature();
	}

	/**
	 * Return the subject distinguished name for the receiver
	 *
	 * @return		Principal	the subject distinguished name for the receiver
	 */
	public java.security.Principal getSubjectDN() {
		return impl.getSubjectDN();
	}

	public javax.security.auth.x500.X500Principal getSubjectX500Principal(){
		return new javax.security.auth.x500.X500Principal( impl.getSubjectDN().getName() );
	}
	/**
	 * Return the subject unique ID for the receiver
	 *
	 * @return		boolean[]	the subject unique ID for the receiver
	 */
	public boolean[] getSubjectUniqueID() {
		return impl.getSubjectUniqueID();
	}
	/**
	 * Return the tbsCertificate for the receiver, DER-encoded
	 *
	 * @return		byte[]	the tbsCertificate for the receiver, DER-encoded
	 */
	public byte[] getTBSCertificate()
		throws CertificateEncodingException {
		return impl.getTBSCertificate();
	}
	/**
	 * Return the version number for this certificate (1 for v1, 2 for v2, 3 for v3)
	 *
	 * @return		int		the version number for this certificate (1 for v1, 2 for v2, 3 for v3)
	 */
	public int getVersion() {
		return impl.getVersion();
	}

	private static final String[] SUPPORTED_CRITICAL_EXTENSIONS = {
		OID_BasicConstraints,
		OID_KeyUsage,
	};

	/**
	 * Return true if the receiver has an unsupported, critical extension. Return
	 * false otherwise.
	 *
	 * @return		boolean		true if the receiver has an unsupported, critical extension, false otherwise.
	 */
	public boolean hasUnsupportedCriticalExtension(){
		return impl.hasUnsupportedCriticalExtension();
	}

	/**
	 * Return all the OID extensions marked as critical for the receiver.
	 *
	 * @return		Set		all the OID extensions marked as critical for the receiver.
	 */
	public Set getCriticalExtensionOIDs(){
		return impl.getCriticalExtensionOIDs();
	}

	/**
	 * Return all the OID extensions marked as non-critical for the receiver.
	 *
	 * @return		Set		all the OID extensions marked as non-critical for the receiver.
	 */
	public Set getNonCriticalExtensionOIDs(){
		return impl.getNonCriticalExtensionOIDs();
	}

	/**
	 * Return the DER-encoded OCTECT String for the extension value defined by oid in the receiver.
	 *
	 * @return		byte[]	the DER-encoded OCTECT String for the extension value defined by oid in the receiver.
	 */
	public byte[] getExtensionValue(String oid){
		return impl.getExtensionValue( oid );
	}

	public List getExtendedKeyUsage()
		throws CertificateParsingException
	{
		return impl.getExtendedKeyUsage();
	}

	/**
	 * Answers the encoded representation for this certificate.
	 *
	 * @return		the encoded representation for this certificate.
	 * @throws	CertificateEncodingException	if encoding errors are detected
	 */
	public byte[] getEncoded()
		throws CertificateEncodingException
	{
		return impl.getEncoded();
	}

	/**
	 * Answers the public key corresponding to this certificate.
	 *
	 * @return		the public key corresponding to this certificate.
	 */
	public PublicKey getPublicKey(){
		return impl.getPublicKey();
	}

	/**
	 * Verifies that this certificate was signed with the given public key.
	 *
	 * @param	key	PublicKey	public key for which verification should be performed.
	 *
	 * @throws CertificateException		if encoding errors are detected
	 * @throws NoSuchAlgorithmException	if an unsupported algorithm is detected
	 * @throws InvalidKeyException		if an invalid key is detected
	 * @throws NoSuchProviderException	if there is no default provider
	 * @throws SignatureException		if signature errors are detected
	 */

	 public void verify(PublicKey key)
				 throws CertificateException,
						NoSuchAlgorithmException,
						InvalidKeyException,
						NoSuchProviderException,
						SignatureException
	{
		 impl.verify( key );
	}

	/**
	 * Verifies that this certificate was signed with the given public key.
	 * Uses the signature algorithm given by the provider.
	 *
	 * @param	key	PublicKey		public key for which verification should be performed.
	 * @param	sigProvider	String	the name of teh signature provider.
	 *
	 * @throws CertificateException		if encoding errors are detected
	 * @throws NoSuchAlgorithmException	if an unsupported algorithm is detected
	 * @throws InvalidKeyException		if an invalid key is detected
	 * @throws NoSuchProviderException	if there is no default provider
	 * @throws SignatureException		if signature errors are detected
	 */
	public void verify(PublicKey key,
						String sigProvider)
				 throws CertificateException,
						NoSuchAlgorithmException,
						InvalidKeyException,
						NoSuchProviderException,
						SignatureException
	{

		impl.verify(key, sigProvider );
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver.
	 *
	 * @return		String
	 *					a printable representation for the receiver.
	 */
	public String toString () {
		StringBuffer result = new StringBuffer();
		result.append ("X.509 Certificate v");
		result.append (getVersion());
		result.append ("\nSubjectDN: ");
		result.append (getSubjectDN());
		result.append ("\n\tNot Before: ");
		result.append (getNotBefore());
		result.append ("\n\tNot After: ");
		result.append (getNotAfter());
		result.append ("\nIssuerDN: ");
		result.append (getIssuerDN());
		result.append ("\nSerialNumber: ");
		result.append (getSerialNumber().toString(16));
		result.append ("\nSigAlgName: ");
		result.append (getSigAlgName());
		result.append (", SigAlgOID: ");
		result.append (getSigAlgOID());
		return result.toString();
	}

	/**
	 * Lookup the OID alias for the specified key in the list of providers.
	 * Fill in the provider name, if found. Look in the OID database
	 * table if no alias is found in the providers.
	 */
	static String getAlias(String key, String oid, String[] providerName) {
		return X509CertImpl.getAlias( key, oid, providerName );
	}

	/**
	 * Generates and initializes a Certificate from a java byte[] object representing
	 * an ASN.1 DER certificate
	 *
	 * @param		certificateBytes	byte[]	The byte[] representing the ASN.1 certificate
	 *
	 * @return	Certificate 	an initialized Certificate
	 * @throws	CertificateException 	if parsing problems are detected
	 */
	public static X509Certificate certificateFromASN1Object(byte[] certificateBytes) throws CertificateException {
			X509CertImpl impl = X509CertImpl.certificateFromASN1Object( certificateBytes );
			return new X509Certificate( impl );
	}

	/**
	 * Generates and initializes a Certificate from an ASN.1 Node object
	 *
	 * @param		node	the ASN.1 Node structure
	 * @param		rawBytes	byte[]	bytes where the node was read from
	 * @return	Certificate 	an initialized Certificate
	 *
	 * @throws	CertificateException 	if parsing problems are detected
	 */
	public static X509Certificate certificateFromASN1Object(Node node, byte[] rawBytes) throws CertificateException {
		X509CertImpl impl = X509CertImpl.certificateFromASN1Object( node, rawBytes );
		return new X509Certificate( impl );
	}

	/**
	 * Generates and initializes a Certificate from a stream with bytes representing
	 * an ASN.1 DER certificate
	 *
	 * @param		in	InputStream stream with bytes representing the ASN.1 certificate
	 *
	 * @return	Certificate 	an initialized Certificate
	 * @throws	CertificateException 	if parsing problems are detected
	 */
	public static X509Certificate certificateFromASN1Object(InputStream in) throws CertificateException {
		X509CertImpl impl = X509CertImpl.certificateFromASN1Object( in );
		return new X509Certificate( impl );
	}

	/**
	 * Initializes and generates a Certificate from arguments
	 *
	 * @param	publicKey	PublicKey	the certificate's public key
	 * @param	principal	String	the name of issuer and subject
	 * @param	notBefore	Date	the date, when the certificates becomes valid
	 * @param	notAfter	Date	the date, when the certificates becomes invalid
	 *
	 * @return	Certifiace	an initialized Certificate
	 *
	 * @throws	CertificateException 	if parsing problems are detected
	 */
	public static java.security.cert.Certificate certificateFromData(PublicKey publicKey, String principal, Date notBefore, Date notAfter) throws CertificateException {
		X509CertImpl impl = X509CertImpl.certificateFromData( publicKey, principal, notBefore, notAfter );
		return new X509Certificate( impl );
	}

	/**
	 * Generates self-signed certificates, so that signs and encodes a certificate
	 * using a sign algorithm and private key.
	 *
	 * @param		signalg	the algorithm which is used to sign certificate
	 * @param		privateKey	the private key which is used to sign
	 *
	 * @return		byte[]	the encoded certificate
	 *
	 * @throws	CertificateException 	if parsing problems are detected
	 */
	public byte[] getSignedAndEncoded(String signalg, PrivateKey privateKey)  throws CertificateException {
		return impl.getSignedAndEncoded( signalg, privateKey );
	}

	/**
	 * Returns a Collection containing the SubjectAltName extension for the Certificate if present.
	 * null is returned if the extension is not present.
	 *
	 * The following is the definition of the ASN.1 SubjectAltName field of X509Certificates:
	 *     SubjectAltName ::= GeneralNames
	 *
	 *     GeneralNames ::= SEQUENCE SIZE (1..MAX) OF GeneralName
	 *
	 *     GeneralName ::= CHOICE {
	 *         otherName                       [0]     OtherName,
	 *         rfc822Name                      [1]     IA5String,
	 *         dNSName                         [2]     IA5String,
	 *         x400Address                     [3]     ORAddress,
	 *         directoryName                   [4]     Name,
	 *         ediPartyName                    [5]     EDIPartyName,
	 *         uniformResourceIdentifier       [6]     IA5String,
	 *         iPAddress                       [7]     OCTET STRING,
	 *         registeredID                    [8]     OBJECT IDENTIFIER}
	 *
	 * The elements in the Collection are Lists where the first element is an Integer defining the
	 * type of the GeneralName (i.e., 0=OtherName, 1=rfc822Name, ... ) and the second is the
	 * GeneralName.
	 *
	 * rfc822Name, dNSName, directoryName and uniformResourceIdentifiers are returned as Strings.
	 * iPAddress for IPv4 addresses is returned as a String in dotted-quad notation.  IPv6
	 * 	addresses are returned in the form "a1:a2:...:a8" where a1-a8 are the eight 16-bit pieces
	 * 	of the address.
	 * Elements of type OBJECT IDENTIFIER are returned in their String representation as a series
	 * 	of positive integers seperated by periods.
	 * Other types are returned as byte arrays.
	 *
	 * @return Collection Collection of Lists of General Names encoded as described above.
	 *
	 * @throws CertificateParsingException
	 */
	public Collection getSubjectAlternativeNames() throws CertificateParsingException {
		return impl.getSubjectAlternativeNames();
	}

	/**
	 * Returns a Collection containing the IssuerAltName extension for the Certificate if present.
	 * null is returned if the extension is not present.
	 *
	 * The following is the definition of the ASN.1 IssuerAltName field of X509Certificates:
	 *     IssuerAltName ::= GeneralNames
	 *
	 * The definition of GeneralNames and the manner in which its values are output
	 * is describe in getSubjectAlternativeNames.
	 *
	 * @return Collection Collection of Lists of General Names encoded as described above.
	 *
	 * @throws CertificateParsingException
	 */
	public Collection getIssuerAlternativeNames() throws CertificateParsingException {
		return impl.getIssuerAlternativeNames();
	}

	/**
	 * Set the issuer distinguished name
	 *
	 * @param		principal Principal	the subject distinguished
	*/
	public void setIssuerDN(Principal principal) {
		impl.setIssuerDN( principal );
	}

	/**
	 * Set the subject distinguished name
	 *
	 * @param		principal Principal	the subject distinguished
	*/
	public void setSubjectDN(Principal principal) {
		impl.setSubjectDN( principal );
	}

}
