package com.ibm.oti.security.provider;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2001, 2004  All Rights Reserved
 */

import java.security.cert.X509Certificate;
import java.security.Principal;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * @author 		OTI
 * @version		initial
 */
public class X509CertificateChain {

	LinkedList chain = null;

	/**
	 * Constructs the certificate chain for the subject given certificates that may
	 * form its certificate chain.  Every attempt should be made to keep the size of chainCandidates
	 * down, as the operation's time consumed will be dependent on its size.
	 *
	 * @param		subject				the certificate for which a certificate chain should be built
	 * @param		chainCandidates		certificates that may form the certificate chain of the subject
	 */
	public X509CertificateChain( X509Certificate subject, Collection chainCandidates ){

		// Null subject check
		if( ( subject == null ) ){
			throw new IllegalArgumentException();
		}

		// Create internal representation of cert chain and add subject to it
		chain = new LinkedList();
		chain.add( 0, subject );

		// Short-circuit if the subject is self-signed
		if( subject.getSubjectDN().equals( subject.getIssuerDN() ) ){
			return;
		}

		// No candidates provided?  Chain consists only of subject
		if( chainCandidates == null ){
			return;
		}

		// Build the certificate chain
		Principal searchDN = subject.getIssuerDN();
		while( true ){

			X509Certificate issuer = findCertForSubject( searchDN, chainCandidates );

			// Find issuer and add it to chain. No issuer found? Finished.
			if( issuer != null ){
				chain.add( issuer );
			} else {
				return;
			}

			// issuer self-signed?  Finished.
			if( issuer.getSubjectDN().equals( issuer.getIssuerDN() ) ){
				return;
			}

			// Now search for issuer of issuer
			searchDN = issuer.getIssuerDN();

		}

	}

	/**
	 * Returns an Iterator over the certificate chain where the first element is the subject of the chain and
	 * the subsequent elements are the certificate chain back to the root (last element)
	 *
	 * @return		Iterator over the certificate chain
	 */
	public Iterator iterator(){
		return chain.iterator();
	}

	/**
	 * Returns the first certificate found in the candidates array where the cert's subject DN is equal
	 * to the searchSubject argument.
	 *
	 * @param		searchSubject		subject of the certificate to be found
	 * @param		candidates			collection of certificates that may contain the certificate
	 * @return		certificate with searchSubject as subject DN if found, null if not found
	 */
	private X509Certificate findCertForSubject( Principal searchSubject, Collection candidates ){

		Iterator i = candidates.iterator();

		while( i.hasNext() ){
			X509Certificate testCert = (X509Certificate) i.next();
			if( testCert.getSubjectDN().equals( searchSubject ) ){
				return testCert;
			}
		}

		return null;

	}

}
