
package com.ibm.oti.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * IdentityHashtable associates keys with values, based on identity of keys (==).
 * Keys cannot be null, but values can.
 * The size of the IdentityHashtable is the number of key/value pairs it contains.
 * The capacity is the number of key/value pairs the IdentityHashtable can hold.
 * The load factor etermines how full the IdentityHashtable
 * gets before expanding the capacity. If the load factor of the IdentityHashtable
 * is exceeded, the capacity is doubled.
 *
 * @version		initial
 *
 * @see			java.lang.System#identityHashCode(Object)
 */

public class IdentityHashtable implements Cloneable {

	public static final Object ABSENT_FLAG = new Object();	// Used for checking if keys exist

	// So that users can walk the key/value pairs evaluating some code on them,
	// without having to know the table's internal representation
	public interface Iterator {
		public void iterate (Object key, Object value);
	}

	int elementCount;		// How many associations exist in the table
	Object [] elementKeys;	// Where the keys are stored
	Object [] elementData;	// Where the associated values are stored
	private int loadFactor;	// The ratio between elementCount and elementKeys.length before the table grows. 50 means 50% and so on
	private int maxSize;	// Maximum size of the table before it grows. Computed based on loadFactor.

	private static final int DEFAULT_SIZE = 101;	// Default initial capacity
	private static final int DEFAULT_LOAD_FACTOR = 50;	// Default load factor (when it hists 50% it grows)
/**
 * Constructs a new IdentityHashtable using the default capacity
 * and load factor.
 *
 * @version		initial
 */
public IdentityHashtable() {
	this (DEFAULT_SIZE);
}
/**
 * Constructs a new IdentityHashtable using the specified capacity
 * and the default load factor.
 *
 * @version		initial
 *
 * @param		capacity	the initial capacity
 */
public IdentityHashtable (int capacity) {
	if (capacity > 0) {
		elementCount = 0;
		elementKeys = new Object [capacity];
		elementData = new Object [capacity];
		loadFactor = DEFAULT_LOAD_FACTOR;
		computeMaxSize();
	} else throw new IllegalArgumentException();
}
/**
 * Constructs a new IdentityHashtable using the specified capacity
 * and load factor.
 *
 * @version		initial
 *
 * @param		capacity	the initial capacity
 * @param		loadFactor	the initial load factor
 */
public IdentityHashtable (int capacity, int loadFactor) {
	this(capacity);
	this.loadFactor = loadFactor;
	computeMaxSize();
}
/**
 * Removes all key/value pairs from this IdentityHashtable, leaving the size zero
 * and the capacity unchanged.
 *
 * @version		initial
 *
 * @see			#isEmpty
 * @see			#size
 */
public synchronized void clear() {
	elementCount = 0;
	for (int i = 0; i < elementKeys.length; i++) {
		elementKeys [i] = elementData [i] = null;
	}
}
/**
 * Answers a new IdentityHashtable with the same key/value pairs, capacity
 * and load factor.
 *
 * @version		initial
 *
 * @return		a shallow copy of this IdentityHashtable
 *
 * @see			java.lang.Cloneable
 */
public synchronized Object clone () {
	try {
		IdentityHashtable hashtable = (IdentityHashtable) super.clone ();
		hashtable.elementKeys = (Object[])elementKeys.clone();
		hashtable.elementData = (Object[])elementData.clone();
		return hashtable;
	} catch (CloneNotSupportedException e) {return null;}
}
/**
 * Computes the maximum number of elements this table can hold before it needs to grow.
 *
 * @version		initial
 *
 */
private void computeMaxSize() {
	maxSize = (int)((long)elementKeys.length * loadFactor / 100);
}
/**
 * Answers if this IdentityHashtable contains the specified object as the value
 * of at least one of the key/value pairs.
 *
 * @version		initial
 *
 * @param		object	the object to look for as a value in this IdentityHashtable
 * @return		true if object is a value in this IdentityHashtable, false otherwise
 *
 * @see			#containsKey
 * @see			java.lang.Object#equals
 */
public synchronized boolean contains (Object object) {
	if (object != null) {
		for (int i=0; i<elementData.length; i++)
			if (elementData[i] == object) return true;
		return false;
	}
	throw new NullPointerException ();
}
/**
 * Answers if this IdentityHashtable contains the specified object as a key
 * of one of the key/value pairs.
 *
 * @version		initial
 *
 * @param		object	the object to look for as a key in this IdentityHashtable
 * @return		true if object is a key in this IdentityHashtable, false otherwise
 *
 * @see			#contains
 * @see			java.lang.Object#equals
 */
public synchronized boolean containsKey (Object object) {
	return get (object, ABSENT_FLAG) != ABSENT_FLAG;
}
/**
 * Computes the index at which object <code>object</code> is in the array
 * <code>array</code>, or at which it should be if it were present.
 *
 * @version		initial
 *
 */
private int findIndex (Object object, Object [] array) {
	Object key;
	int length = array.length;
	int index = (System.identityHashCode(object) & 0x7FFFFFFF) % length;
	for (int i=index; i<length; i++)
		if ((key = array[i]) == null || key == object) return i;
	for (int i=0; i<index; i++)
		if ((key = array[i]) == null || key == object) return i;
	throw new RuntimeException();
}

/**
 * Answers the IdentityHashtable value that corresponds to input
 * <code>key</code>.
 *
 * @version initial
 *
 * @param key
 *            the key of the value to be returned
 * @return the value corresponding to <code>key</code>. May return
 *         <code>null</code> if <code>key</code> is not known or if the
 *         associated value is actually <code>null</code>.
 *
 * @see #put
 * @see #get(Object)
 * @see #get(Object, Object)
 */
public synchronized Object get (Object key) {
	return get (key, null);
}

/**
 * Answers the IdentityHashtable value that corresponds to input
 * <code>key</code>. If there is no such key, returns the parameter
 * <code>returnIfMissing</code>. This is useful because it allows the table
 * to have <code>null</code>s as associated values, and allows the caller to
 * determine if there's no such key or if the value is just <code>null</code>
 *
 * @version initial
 *
 * @param key
 *            the key of the value returned
 * @param returnIfMissing
 *            Object to return if there is no such key
 *
 * @return the value corresponding to <code>key</code> or else
 *         <code>returnIfMissing<code> if <code>key</code> is not known
 *
 * @see			#put
 */
public synchronized Object get (Object key, Object returnIfMissing) {
	int index = findIndex (key, elementKeys);
	if (elementKeys[index] != null) return elementData[index];
	else return returnIfMissing;
}
/**
 * Answers if this IdentityHashtable has no key/value pairs, a size of zero.
 *
 * @version		initial
 *
 * @return		true if this IdentityHashtable has no key/value pairs, false otherwise
 *
 * @see			#size
 */
public boolean isEmpty() {
	return elementCount == 0;
}
/**
 * Iterates over all keys&values using the provided iterator
 *
 * @version		initial
 *
 */
public synchronized void iterate (Iterator iterator) {
	for (int i=0; i<elementKeys.length; i++) {
		Object key;
		if ((key = elementKeys [i]) != null) {
			iterator.iterate (key, elementData [i]);
		}
	}
}

/**
 * Associates input object <code>value</code> with the input <code>key</code>
 * in this IdentityHashtable. If <code>key</code> already exists in the
 * IdentityHashtable, the old value is replaced.
 *
 * @version initial
 *
 * @param key
 *            the key to add. Cannot be <code>null</code>.
 * @param value
 *            the value to associate with <code>key</code>. May be
 *            <code>null</code>.
 * @return the old value associated with the <code>key</code>. A
 *         <code>null</code> return will result if the key did not exist in
 *         this IdentityHashtable or if the existing associated value really was
 *         <code>null</code>.
 *
 * @see #get(Object)
 * @see #get(Object, Object)
 */
public synchronized Object put (Object key, Object value) {
	if (key != null) {
		int index = findIndex (key, elementKeys);
		Object object = elementData [index];
		if (elementKeys[index] == null) {
			if (++elementCount > maxSize) {
				rehash ();
				index = findIndex (key, elementKeys);
			}
		}
		elementKeys [index] = key;
		elementData [index] = value;
		return object;
	} else throw new NullPointerException ();
}
/**
 * Re-arranges the elements in the receiver to be consistent with hash values of the keys.
 * This method is sent when the size of this IdentityHashtable exceeds the load factor.
 *
 * @version		initial
 */
protected void rehash () {
	Object key;
	int index, length = elementKeys.length<<1;
	Object [] newKeys = new Object [length];
	Object [] newData = new Object [length];
	for (int i=0; i<elementKeys.length; i++) {
		if ((key = elementKeys [i]) != null) {
			index = findIndex (key, newKeys);
			newKeys [index] = key;
			newData [index] = elementData [i];
		}
	}
	elementKeys = newKeys;
	elementData = newData;
	computeMaxSize();
}

/**
 * Remove the <code>key</code>/value pair from this IdentityHashtable.
 *
 * @version initial
 *
 * @param key
 *            the key in the IdentityHashtable identifying the key/value
 *            pair for removal
 * @return the value associated with
 *         <code>key<code>. This may be <code>null</code> if the key does not
 *          exist in this IdentityHashtable or if the associated value is
 *          really <code>null</code>.
 *
 * @see			#get(Object)
 * @see			#get(Object, Object)
 * @see			#put
 */
public synchronized Object remove (Object key) {
	boolean hashedOk;
	int index, next, hash;
	Object result, object;
	index = next = findIndex (key, elementKeys);
	if ((result = elementData [index]) == null) return null;
	int length = elementKeys.length;
	do {
		next = (next + 1) % length;
		if ((object = elementKeys [next]) == null) break;
		hash = (System.identityHashCode(object) & 0x7FFFFFFF) % length;
		hashedOk = hash > index;
		if (next < index) {
			hashedOk = hashedOk || (hash <= next);
		} else {
			hashedOk = hashedOk && (hash <= next);
		}
		if (!hashedOk) {
			elementKeys [index] = object;
			elementData [index] = elementData [next];
			index = next;
		}
	} while (true);
	elementCount--;
	elementKeys [index] = elementData [index] = null;
	return result;
}
/**
 * Answers the number of key/value pairs in this IdentityHashtable.
 *
 * @version		initial
 *
 * @return		the number of key/value pairs in this IdentityHashtable
 *
 * @see			#elementData
 * @see			#elementKeys
 */
public int size () {
	return elementCount;
}
/**
 * Answers the string representation of this IdentityHashtable.
 *
 * @version		initial
 *
 * @return		the string representation of this IdentityHashtable
 */
public synchronized String toString () {
	Object key;
	int count = 0;
	StringBuffer buffer = new StringBuffer ();
	buffer.append ('{');
	for (int i=0; i<elementKeys.length; i++) {
		if ((key = elementKeys [i]) != null) {
			buffer.append (key);
			buffer.append ('=');
			buffer.append (elementData [i]);
			if (++count < elementCount) buffer.append (',');
		}
	}
	buffer.append ('}');
	return buffer.toString ();
}
}
