/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2005, 2006  All Rights Reserved.
 */

package com.ibm.oti.util;

import java.io.*;
import java.math.BigInteger;
import java.util.*;
import java.security.GeneralSecurityException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.Principal;
import java.security.Signature;

import com.ibm.oti.security.provider.PKCS7;
import com.ibm.oti.security.provider.X509CertificateChain;
import com.ibm.oti.security.provider.PKCS7.SignerInfo;

/**
 * Helper class containing a number of static convenience methods for use
 * in the processing of JAR files.
 * <p>
 * This class cannot be instantiated.
 */
public class JarUtils {

    /**
     * Ensure that this utility class cannot be instantiated.
     */
    private JarUtils(){
        // NO OP
    }

    // ----------- Verification of a signed jar's signature file follows ------

    /**
     * Convenience method to used to check that a jar's signature file and its
     * corresponding signature block file (both supplied as {@link InputStream}
     * objects) are valid and that neither has been tampered with in respect to
     * the other.
     * <p>
     * In particular, this method will verify that:
     * <ul>
     * <li>the signature block file contains a signature that could only have
     * been created with the private key corresponding to the public key whose
     * certificate chain is also included in the signature block file.
     * <li>the message digest of the signature file is precisely equivalent to
     * the message digest value taken from the decrypted signature block file.
     * Confirming this demonstrates that there has been no tampering on the
     * contents of the signature file since the signing took place.
     *
     * @param signature
     *            the contents of a signed jar's signature file (.SF file) as an
     *            <code>InputStream</code>. This stream will be closed by the
     *            method immediately upon its contents being read.
     * @param signatureBlock
     *            the contents of a signed jar's binary signature block file as
     *            an <code>InputStream</code>. This stream will be closed by
     *            the method immediately upon its contents being read.
     * @return the certificate chains discovered in the supplied signature block
     *         file stream in an array of {@link Certificate}.
     * @throws IOException
     *             if there was an error encountered while reading from either
     *             of the supplied input streams.
     * @throws GeneralSecurityException
     *             if the signature verification process fails.
     */
    public static Certificate[] verifySignature(InputStream signature,
            InputStream signatureBlock) throws IOException,
            GeneralSecurityException {
        Certificate[] signerCertChain = null;

        byte[] sfBytes = new byte[signature.available()];
        signature.read(sfBytes, 0, sfBytes.length);
        byte[] digitalSigBytes = new byte[signatureBlock.available()];
        signatureBlock.read(digitalSigBytes, 0, digitalSigBytes.length);

        // TODO Confirm that the input InputStreams will be closed in this call.
        signature.close();
        signatureBlock.close();

        PKCS7 pkcs7;
        try {
            pkcs7 = new PKCS7(digitalSigBytes);
        } catch (IllegalArgumentException e) {
            return signerCertChain;
        }

        // Find the issuer and serial number of the certificate that signed the
        // JAR
        SignerInfo[] sInfos = pkcs7.signedData().signerInfos();
        SignerInfo sInfo = sInfos[0];
        Principal issuer = sInfo.getIssuer();
        BigInteger issuerSerialNumber = sInfo.getSerialNumber();

        // Locate the certificate that signed the JAR (it's issuer and serial
        // number were just found above...)
        X509Certificate signerCert = null;
        X509Certificate testCert = null;
        Collection certs = null;

        try {
            CertificateFactory cf = CertificateFactory.getInstance("X.509");
            certs = cf.generateCertificates(new ByteArrayInputStream(
                    digitalSigBytes));
            /* [IF] */
            // Probably should be verifying certificate chain
            /* [ENDIF] */

            /*
             * [PR97161] Check all certificates for a certificate whose issuer
             * and serial number match that used to sign data.
             */
            Iterator i = certs.iterator();
            while (signerCert == null && i.hasNext()) {
                testCert = (X509Certificate) i.next();
                if (testCert.getIssuerDN().equals(issuer)) {
                    if (testCert.getSerialNumber().equals(issuerSerialNumber)) {
                        signerCert = testCert;
                        break;
                    }// end if serial numbers match
                }// end if issuers match
            }// end while

            if (signerCert == null) {
                // no certificate for the signer so it can not be verified
                return signerCertChain;
            }
        } catch (CertificateException e) {
            // We can't read the certificate, algorithm may not be available
            return signerCertChain;
        }

        Signature sig = null;
        String algorithm = sInfo.signatureName();
        try {
            if (algorithm != null)
                sig = Signature.getInstance(algorithm);
        } catch (NoSuchAlgorithmException e) {
            // Intentionally not handling
        }
        if (sig == null)
            try {
                /* [PR 93954] */
                algorithm = sInfo.digestEncryptionAlgorithm();
                if (algorithm == null)
                    return signerCertChain;
                sig = Signature.getInstance(algorithm);
            } catch (NoSuchAlgorithmException e) {
                // We don't have the algorithm, don't verify this certificate
                return signerCertChain;
            }
        try {
            sig.initVerify(signerCert.getPublicKey());
        } catch (InvalidKeyException e) {
            return signerCertChain;
        }

        // If the authenticatedAttributes field of SignerInfo contains more than
        // zero attributes,
        // compute the message digest on the ASN.1 DER encoding of the
        // Attributes value.
        // Otherwise, compute the message digest on the data.
        ASN1Decoder.Node[] attributeNodes = sInfo.authenticatedAttributes();
        if (attributeNodes.length > 0) {
            ASN1Decoder.Node attributeSetNode = new ASN1Decoder.Node();
            attributeSetNode.type = ASN1Decoder.SET;
            attributeSetNode.data = attributeNodes;
            try {
                sig.update(ASN1Encoder.encodeNode(attributeSetNode));
            } catch (SignatureException e) {
                return signerCertChain;
            }
        } else {
            try {
                sig.update(sfBytes);
            } catch (SignatureException e) {
                return signerCertChain;
            }
        }

        // If the authenticatedAttributes field contains the message-digest
        // attribute, verify that it equals the computed digest of the
        // signature file
        byte[] existingDigest = sInfo.contentMessageDigest();
        if (existingDigest != null) {
            try {
                MessageDigest digestAlgorithm = MessageDigest.getInstance(sInfo
                        .digestAlgorithm());
                byte[] computedDigest = digestAlgorithm.digest(sfBytes);
                if (!Arrays.equals(existingDigest, computedDigest)) {
                    throw new SignatureException();
                }
            } catch (NoSuchAlgorithmException e) {
                return signerCertChain;
            }
        }

        if (!sig.verify(sInfo.encryptedDigest()))
            throw new SignatureException();

        if (((java.security.cert.X509Certificate) signerCert)
                .hasUnsupportedCriticalExtension()) {
            return signerCertChain;
        }

        // Signature verified. Return signer's certificate chain to caller.
        /*
         * [Bug 101102] Need to provide full certificate chain, not just the
         * signer certificate
         */
        X509CertificateChain certChain = new X509CertificateChain(signerCert,
                certs);

        // Take the X509CertificateChain object and retrieve an array of
        // java.security.cert.Certificate for the cert chain. The order of the
        // array entries is vitally important : first should be the subject,
        // the last should be the root.
        Vector allcerts = new Vector();
        Iterator it = certChain.iterator();
        while (it.hasNext()) {
            allcerts.add((Certificate) it.next());
        }// end while
        signerCertChain = (Certificate[]) allcerts.toArray(new Certificate[0]);
        return signerCertChain;
    }

    // ----------- Creating a signed jar file follows --------------

    /**
     * Creates the PKCS #7 version of a signed jar's digital signature which may
     * be directly used as the contents of the signature block file associated
     * with the signing. Receives the signed version of the digest (created
     * using the signing alias' private key) and constructs the PKCS #7 version
     * of the digital signature as a <code>byte</code> array.
     *
     * @param encryptedDigest
     *            the signed version of the message digest value supplied in
     *            <code>digest</code>. This value will have been previously
     *            obtained by signing <code>digest</code> with the signer's
     *            private key using the signature algorithm specified in
     *            <code>signatureAlgorithm</code>.
     * @param signerCertificateChain
     *            the certificate chain associated with the alias signing the
     *            jar. This value will have been previously obtained as the
     *            result of a call to
     *            {@link java.security.KeyStore#getCertificateChain(String)}
     *            with the signer's alias.
     * @param digestAlgorithm
     *            the name of the message digest algorithm used in the signing
     *            of the jar signature file. Dependent on the signer's key type.
     * @param signatureAlgorithm
     *            the name of the signature algorithm used in the signing
     *            of the jar signature file. Dependent on the signer's key type.
     * @return signature data in the PKCS #7 format which may be used as the
     *         complete contents of the signature block file (e.g. the
     *         <code>.DSA</code> or <code>.RSA</code> file) created for a
     *         given jar signing.
     * @throws IOException
     */
    public static byte[] getSignatureBlockBytes(
            byte[] encryptedDigest, Certificate[] signerCertificateChain,
            String digestAlgorithm, String signatureAlgorithm)
            throws IOException {
        return PKCS7.getASN1DEREncoded(signerCertificateChain, digestAlgorithm,
                signatureAlgorithm, encryptedDigest);
    }
}
