
/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

package com.ibm.oti.vm;

import java.util.Hashtable;

public class AppClassLoader extends AbstractClassLoader {

	private boolean checkingPackageAccess = false;
	private static Hashtable protectionDomainCache = new Hashtable(10);

/**
 * Constructs a new instance of this class after checking
 * with the security manager to see if this is ok.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	SecurityException
 *					if a security manager exists and its checkCreateClassLoader method doesn't
 *					allow creation of new ClassLoaders
 */
public AppClassLoader() {
	this(null);
}

/**
 * Constructs a new instance of this class after checking
 * with the security manager to see if this is ok.  The
 * newly created instance will have specified ClassLoader
 * as its parent.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		parent ClassLoader
 *					the ClassLoader name of the resource to find.
 *
 * @exception	SecurityException
 *					if a security manager exists and its checkCreateClassLoader method doesn't
 *					allow creation of new ClassLoaders
 */
public AppClassLoader(ClassLoader parent) {
	super(parent);
	String classPath = System.getProperty("java.class.path", ".");
	classPath = parsePath(classPath);
	com.ibm.oti.vm.VM.setClassPathImpl(this, classPath);
}

AppClassLoader(boolean ignore) {
	super(null);
}

/**
 * Using a native, call into the VM and ask that the named class
 * be searched for using the java.class.path
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		java.lang.Class
 *					the class or null.
 * @param 		className String
 *					the name of the class to search for.
 * @exception	ClassNotFoundException
 *					if the class is not found.
 */
protected Class findClass (String className) throws ClassNotFoundException {
	Class newClass = com.ibm.oti.vm.VM.findClassOrNull(className, this);
	if (newClass == null) throw new ClassNotFoundException(className);

	int index = VM.getCPIndexImpl(newClass);
	String packageName = getPackageName(newClass);
	if (packageName != null)
		definePackage(packageName, index);
	VM.setPDImpl(newClass, getFilePD(index));
	return newClass;
}

Hashtable getProtectionDomainCache() {
	return protectionDomainCache;
}

/**
 * Overrides the loadClass() of <code>ClassLoader</code>. It calls the security
 * manager's <code>checkPackageAccess()</code> before attempting toload the class.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		java.lang.Class
 *					the Class object.
 * @param 		className String
 *					the name of the class to search for.
 * @param 		resolveClass boolean
 *					indicates if class should be resolved after loading.
 * @exception	ClassNotFoundException
 *					If the class could not be found.
 */
protected synchronized Class loadClass(String className, boolean resolveClass) throws ClassNotFoundException {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null && !checkingPackageAccess) {
		int index = className.lastIndexOf('.');
		if (index > 0) { // skip if class is from a default package
			try {
				checkingPackageAccess = true;
				sm.checkPackageAccess(className.substring(0, index));
			} finally {
				checkingPackageAccess = false;
			}
		}
	}
	return super.loadClass(className, resolveClass);
}
}
