
package com.ibm.oti.vm;

import java.util.Vector;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * Represents the running virtual machine. All VM specific API
 * are implemented on this class.
 * <p>
 * Note that all methods in VM are static. There is no singleton
 * instance which represents the actively running VM.
 */
public final class VM {

	private static Vector shutdownHooks = new Vector();
	private static boolean shutdown = false;
	private static final RuntimePermission shutdownHooksPermission =
		new RuntimePermission("shutdownHooks");
	private static boolean closeJars = false;
	private static boolean deleteOnExit = false;
	private static Vector shutdownClasses = new Vector(4);

/**
 * Prevents this class from being instantiated.
 */
private VM() {
}

/**
 * Answer the class at depth.
 *
 * Notes:
 * 	 1) This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers.
 *
 *	 2) The item at index zero describes the caller of this method.
 *
 */
static final native Class getStackClass(int depth);

/**
 * Returns the ClassLoader of the method (including natives) at the
 * specified depth on the stack of the calling thread. Frames representing
 * the VM implementation of java.lang.reflect are not included in the list.
 *
 * This is not a public method as it can return the bootstrap class
 * loader, which should not be accessed by non-bootstrap classes.
 *
 * Notes: <ul>
 * 	 <li> This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers. </li>
 *
 *	 <li> The item at depth zero is the caller of this method </li>
 * </ul>
 *
 * @param depth the stack depth of the requested ClassLoader
 * @return the ClassLoader at the specified depth
 *
 * @see java.lang.ClassLoader#getStackClassLoader
 */
static final native ClassLoader getStackClassLoader(int depth);

/**
 * Walk the stack and answer the most recent non-null and non-bootstrap
 * ClassLoader on the stack of the calling thread. If no such ClassLoader
 * is found, null is returned.
 *
 * Notes:
 * 	 1) This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers.
 *
 * @return the first non-bootstrap ClassLoader on the stack
 */
static public final native ClassLoader getNonBootstrapClassLoader();

/**
 * Initialize the classloader.
 *
 * @param 		loader ClassLoader
 *					the ClassLoader instance
 * @param 		bootLoader boolean
 *					true for the bootstrap class loader
 */
public final static native void initializeClassLoader(ClassLoader loader, boolean bootLoader);

/**
 * Return whether or not a ClassLoader is the bootstrap class loader.
 *
 * @param 		loader the ClassLoader instance to test
 * @return	true if the specified ClassLoader is the bootstrap ClassLoader
 */
static private final native boolean isBootstrapClassLoader(ClassLoader loader);

/**
 * Native used to find and load a class using the VM
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		java.lang.Class
 *					the class or null.
 * @param 		className String
 *					the name of the class to search for.
 * @param		classLoader
 *					the classloader to do the work
 */
static native Class findClassOrNull(String className, ClassLoader classLoader);

/**
 * Returns the ClassLoader of the method that called the caller.
 * i.e. A.x() calls B.y() calls callerClassLoader(),
 * A's ClassLoader will be returned. Returns null for the
 * bootstrap ClassLoader.
 *
 * @return 		a ClassLoader or null for the bootstrap ClassLoader
 *
 * @throws SecurityException when called from a non-bootstrap Class
 */
public static ClassLoader callerClassLoader() {
	ClassLoader loader = getStackClassLoader(2);
	ClassLoader caller = getStackClassLoader(1);
	if (!isBootstrapClassLoader(caller))
		throw new SecurityException();
	if (isBootstrapClassLoader(loader)) return null;
	return loader;
}

private static ClassLoader caller2ClassLoader() {
	ClassLoader loader = getStackClassLoader(3);
	if (isBootstrapClassLoader(loader)) return null;
	return loader;
}

/**
 * Native used to dump a string to the system console for debugging.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		str String
 *					the String to display
 */
public static native void dumpString(String str);

/**
 * Native used to set the classpath for an OTI implemented classloader
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		classLoader
 *					the classloader to give the path to.
 *				classPath
 *					the string used as the class path.
 */
public static void setClassPathImpl(ClassLoader classLoader, String classPath) {
	setClassPathImpl(classLoader, com.ibm.oti.util.Util.getBytes(classPath));
}
private static native void setClassPathImpl(ClassLoader classLoader, byte[] classPath);

/**
 * Tags a class as hot-swap enabled, this flag is intended for
 * use by OTI implemented class loaders.  This flag implies that
 * when a class is changed in the IDE it's replacement can be found
 * by searching the bootstrap classpath and classpath entries.
 *
 * User implemented classLoaders can set this flag if they respect
 * these limitations.
 *
 * @param 		hotSwapClass
 *					the Class to tag as hot-swap enabled.
 *
 * @author		OTI
 * @version		initial
 *
 * @deprecated	No longer used for hot-swap
 */
public static void enableClassHotSwap(Class hotSwapClass) {
}

/**
 * Sets the ProtectionDomain of the first arg.
 * <p>
 * This method is for internal use only.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		targetClass Class
 *					the class to set the protection domain of.
 *				pd ProtectionDomain
 *					targetClass' new ProtectionDomain.
 *
 * @see			java.lang.Class
 */
static native void setPDImpl(Class targetClass, Object pd);

/**
 * Get the classpath entry that was used to load the class that is the arg.
 * <p>
 * This method is for internal use only.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		targetClass Class
 *					the class to set the classpath of.
 *
 * @see			java.lang.Class
 */
static native int getCPIndexImpl(Class targetClass);

/**
 * Does internal initializaion required by VM.
 *
 * @author		OTI
 * @version		initial
 */
static void initializeVM() {
}

/**
 * Registers a new virtual-machine shutdown hook.
 * This is equivalent to the 1.3 API of the same name.
 *
 * @param		hook	the hook (a Thread) to register
 *
 * @author		OTI
 * @version		initial
 */
public static void addShutdownHook(Thread hook) {
	SecurityManager manager = System.getSecurityManager();
	if (manager != null)
		manager.checkPermission(shutdownHooksPermission);

	synchronized (shutdownHooks) {
		if (shutdown) {
// K01a4 = Already shutting down
			throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K01a4"));
		}
		if (hook == null || shutdownHooks.contains(hook)) {
			// K01a5 = Illegal shutdown hook: {0}
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K01a5", hook));
		}
		shutdownHooks.addElement(hook);
	}
}

/**
 * De-registers a previously-registered virtual-machine shutdown hook.
 * This is equivalent to the 1.3 API of the same name.
 *
 * @param		hook	the hook (a Thread) to de-register
 * @return		true if the hook could be de-registered
 *
 * @author		OTI
 * @version		initial
 */
public static boolean removeShutdownHook(Thread hook) {
	SecurityManager manager = System.getSecurityManager();
	if (manager != null)
		manager.checkPermission(shutdownHooksPermission);

	synchronized (shutdownHooks) {
		if (shutdown) {
// K01a4 = Already shutting down
			throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K01a4"));
		}
		return shutdownHooks.removeElement(hook);
	}
}

/**
 * Called to signal that the com.ibm.oti.net.www.protocol.jar.JarURLConnection
 * class has been loaded and JarURLConnection.closeCachedFiles() should be called
 * on VM shutdown.
 *
 * @see #shutdown()
 */
public static void closeJars() {
	closeJars = true;
}

/**
 * Called to signal that the com.ibm.oti.util.DeleteOnExit class has been loaded
 * and DeleteOnExit.deleteOnExit() should be called on VM shutdown.
 *
 * @see #shutdown()
 */
public static void deleteOnExit() {
	deleteOnExit = true;
}

public static void addShutdownClass(Runnable cl) {
	shutdownClasses.addElement(cl);
}

/**
 * Internal use only. Shutdown the JCL. Do not create new Threads.
 * If shutting down without calling Runtime.exit(),
 * or Runtime.halt(),
 * daemon threads will still be running. If shutting down from a call
 * to System.exit(), JNI DestroyJavaVM(),
 * or Runtime.halt(),
 * all threads will still be running.
 * Called after java.lang.Runtime shutdownHooks have been run,
 * and after any finalization.
 *
 * @see #deleteOnExit()
 * @see #closeJars()
 */
private static void shutdown() {
	for (int i=shutdownClasses.size() - 1; i>=0; i--)
		((Runnable)shutdownClasses.elementAt(i)).run();
	if (closeJars)
		com.ibm.oti.net.www.protocol.jar.JarURLConnection.closeCachedFiles();
	if (deleteOnExit)
		com.ibm.oti.util.DeleteOnExit.deleteOnExit();
}

/**
 * Internal use only.
 * Run the shutdown hooks immediately
 *
 * @author		OTI
 * @version		initial
 */
private static void cleanup() {
	synchronized (shutdownHooks) {
		if (shutdown) return;
		shutdown = true;
	}

	/* start all of the hooks concurrently */
	for (int i = 0; i < shutdownHooks.size(); i++) {
		((Thread)shutdownHooks.elementAt(i)).start();
	}

	/* wait for them all to finish */
	for (int i = 0; i < shutdownHooks.size(); i++) {
		Thread hook = (Thread)shutdownHooks.elementAt(i);
		while (true) {
			try {
				hook.join();
			} catch (InterruptedException e) {
				continue;
			}
			break;
		}
	}
}

static final int CPE_TYPE_UNKNOWN = 0;
static final int CPE_TYPE_DIRECTORY = 1;
static final int CPE_TYPE_JAR = 2;
static final int CPE_TYPE_TCP = 3;
static final int CPE_TYPE_JXE = 4;
static final int CPE_TYPE_UNUSABLE = 5;
static final int CPE_TYPE_PALMDB = 6;

/**
 * Return the type of the specified entry on the class path for a ClassLoader.
 * Valid tyes are:
 * 		CPE_TYPE_UNKNOWN
 *		CPE_TYPE_DIRECTORY
 * 		CPE_TYPE_JAR
 *		CPE_TYPE_TCP - this is obsolete
 *		CPE_TYPE_UNUSABLE
 *
 * @param classLoader the ClassLoader
 * @param cpIndex the index on the class path
 *
 * @return a int which specifies the class path entry type
 */
static final native int getClassPathEntryType(Object classLoader, int cpIndex);

static final native long getJxePointerFromClassPath(Object classLoader, int cpIndex);

/**
 * Maps the class path entry to a Jxe.
 */
static Jxe getJxeFromClassPath(Object classLoader, int cpIndex) {
	long pointer = getJxePointerFromClassPath(classLoader, cpIndex);
	if (pointer == 0) return null;
	Jxe jxe = new Jxe(pointer, false);
	return jxe;
}

/**
 * Returns command line arguments passed to the VM. Internally these are broken into
 * optionString and extraInfo. This only returns the optionString part.
 * <p>
 * @return		a String array containing the optionString part of command line arguments
 *
 * @author		OTI
 * @version		initial
 */
public static native String [] getVMArgs();

/**
 * Return the number of entries on the bootclasspath.
 *
 * @return an int which is the number of entries on the bootclasspath
 */
static native int getClassPathCount();

/**
 * Return the specified bootclasspath entry. Directory entries
 * are terminated with a file separator character.
 *
 * @param index the index of the bootclasspath entry
 *
 * @return a byte array containing the bootclasspath entry
 * 			specified in the vm options
 */
static native byte[] getPathFromClassPath(int index);

/**
 * Loads and links the library specified by the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		libName		the name of the library to load
 *
 * @exception	SecurityException
 *							if the library was not allowed to be loaded
 */
public static synchronized void loadLibrary(String libName) throws java.io.IOException {
	try {
		Runtime.getRuntime().loadLibrary(libName);
	} catch (UnsatisfiedLinkError e) {
		throw new java.io.IOException(e.getMessage());
	}
}

/**
 * This method will cause the GC to do a local collection.
 */
public static native void localGC();

/**
 * This method will cause the GC to do a global collection.
 */
public static native void globalGC();

/**
 * Provides a hint to the virtual machine that it would
 * be useful to attempt to perform any outstanding
 * object finalizations.
 */
public static native void runFinalization();

/**
 * This method will disable finalization on aClass.  This action cannot be undone.
 */
public static void disableFinalization(Class aClass) {
	if (callerClassLoader() != null)
		throw new SecurityException();
	disableFinalizationImpl(aClass);
}

private final static native void disableFinalizationImpl(Class aClass);

/**
 * Answer if native implementations should be used.
 */
public final static boolean useNatives() {
	return useNativesImpl();
}

private final static native boolean useNativesImpl();

/**
 * Returns the number of processors, zero or more.
 * Zero if the number cannot be determined.
 */
public static int availableProcessors() {
	return processorsImpl();
}
private static native int processorsImpl();

/**
 * Enable the JIT and return true only if the vm has been started with
 * the JIT explicitly disabled using -Xjit:noJitUntilEnabled, and enableJIT()
 * has not been called previously.
 */
public static native boolean enableJIT();

/**
 * This method will enable finalization on aClass.
 * Do not call this on a class that does not implement finalize()V.
 */
public static void enableFinalization(Class aClass) {
}

/*
 * Returns NULL, or an array of 2 String objects:
 *    array[0] = proxy host string
 *    array[1] = proxy port string
 */
public static String[] getHttpProxyParms() {
	return getHttpProxyImpl();
}
private static native String[] getHttpProxyImpl();
}
