package com.ibm.oti.xlet;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2006  All Rights Reserved
 */

import java.awt.Container;
import java.security.AccessControlContext;
import java.security.AccessControlException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Hashtable;
import java.net.URL;
import java.net.URLClassLoader;

import javax.microedition.xlet.*;

/**
 * This class represents the Xlet to the XletApplicationManager.
 */
public class XletApplication implements XletContext {

	private Xlet xlet;
	private String name;
	private String path;
	private XletState xletState;
	private Container xletContainer;
	private Hashtable properties = new Hashtable();
	private Object waiter;

	private XletApplicationManager applicationManager;
	private XletCommandPipe asyncCommandPipe;
	private XletCommandDispatcher commandDispatcher;
	private ClassLoader classLoader;
	private AccessControlContext securityContext;

	public XletApplication(String name, String path, String[] args) {
		this.name = name;
		this.path = path;
		this.securityContext = AccessController.getContext();

		// save the Xlet's startup args as a property
		properties.put(XletContext.ARGS, args);
	}

	/**
	 * Returns the current xlet state.
	 * @return XletState
	 */
	public XletState getXletState() {
		return xletState;
	}

	public AccessControlContext getSecurityContext() {
		return securityContext;
	}

	/**
     * Waits until all pending state change
     * requests have been processed by the
     * Xlet.  This is required for programatic
     * testing of the state change code to ensure
     * that the requests have been processed before
     * getState is called.
     */
	public void waitForPendingStateChanges() {
		Object localWaiter = new Object();
		synchronized (this) {
			this.waiter = localWaiter;
		}

		synchronized (localWaiter) {
			if (xletState != XletState.XLET_STATE_DESTROYED) {
				if (asyncCommandPipe.notifyWhenIdle(localWaiter) == false) {
					try {
						localWaiter.wait();
					} catch (InterruptedException e) {
					}
				}
			}
		}

		synchronized (this) {
			this.waiter = null;
		}
	}

	/**
	 * Sets the xlet state.
	 * @param xletState The new xlet state
	 * @return the previous xlet state
	 */
	public synchronized XletState setXletState(XletState newState) {
		XletState previousState = xletState;
		if (previousState != newState) {
			xletState = newState;
			applicationManager.notifyXletStateChange(this, newState);
		}
		return previousState;
	}

	/**
	 * Returns the name of the Xlet.
	 * @return xlet name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the path for the Xlet.
	 * @return xlet path
	 */
	public String getPath() {
		return path;
	}

	/**
	 * Bind this XletApplication to the real Xlet
	 */
	public void bind(Xlet xlet, XletApplicationManager manager) {
		xletState = XletState.start();
		this.xlet = xlet;
		applicationManager = manager;
		asyncCommandPipe = new XletCommandPipe();
		commandDispatcher = new XletCommandDispatcher(xlet, this, asyncCommandPipe);
	}

	/**********************************************************************************
	 * Commands to send to the Xlet.
	 * The Xlet runs in its own thread and the communications are done asynchronously.
	 **********************************************************************************/

	/**
	 * Initialize the Xlet.  Setup xlet and send initXlet.  The XletCommandDispatcher and Xlet are run
	 * on a separate thread.  This is done to isolate the Xlet processing and to allow the Xlet Manager
	 * to manage multiple Xlets.
	 */
	public void initXlet() {
		commandDispatcher.start();
		asyncCommandPipe.writeCommand(XletCommandPipe.XLET_INIT);
	}

	/** Start the Xlet.  The start is asynchronously sent to the Xlet, which runs on its own thread. */
	public void startXlet() {
		asyncCommandPipe.writeCommand(XletCommandPipe.XLET_START);
	}

	/** Pause the Xlet.  The pause is asynchronously sent to the Xlet, which runs on its own thread. */
	public void pauseXlet() {
		asyncCommandPipe.writeCommand(XletCommandPipe.XLET_PAUSE);
	}

	/** Destroy the Xlet.  The destroy is asynchronously sent to the Xlet, which runs on its own thread. */
	public void destroyXlet(boolean unconditional) {
		if (unconditional) {
			asyncCommandPipe.writeCommand(XletCommandPipe.XLET_DESTROY_UNCONDITIONAL);
		} else {
			asyncCommandPipe.writeCommand(XletCommandPipe.XLET_DESTROY);
		}
	}

	/**********************************************************************************
	 * Callbacks from the Xlet, to is XletContext.
	 * These methods are called from the Xlet's thread.
	 **********************************************************************************/

	/**
	 * Callback from the Xlet to request a property value.
	 *
	 * @see javax.microedition.xlet.XletContext#getXletProperty(String)
	 */
	public Object getXletProperty(String key) {
		if (key == null)
			throw new NullPointerException();
		if (key.length() == 0)
			throw new IllegalArgumentException();

		return properties.get(key);
	}

	/**
	 * Callback from the Xlet to notify the manager that it has entered Paused state.
	 *
	 * @see javax.microedition.xlet.XletContext#notifyPaused()
	 */
	public void notifyPaused() {
		xletState.eventXletCalledNotifyPaused(this);
	}

	/**
	 * Callback from the Xlet to request to the manager that it wants to enter Active state.
	 *
	 * @see javax.microedition.xlet.XletContext#resumeRequest()
	 */
	public void resumeRequest() {
		if (xletState == XletState.XLET_STATE_PAUSED) {
			startXlet();
		}
	}

	/**
	 * Callback from the Xlet to notify the manager that it has entered Destroyed state.
	 *
	 * @see javax.microedition.xlet.XletContext#notifyDestroyed()
	 */
	public void notifyDestroyed() {
		xletState.eventXletCalledNotifyDestroyed(this);
	}

	/**
	 * Callback from the Xlet to get its container for putting AWT components.
	 *
	 * @see javax.microedition.xlet.XletContext#getContainer()
	 */
	public Container getContainer() throws UnavailableContainerException {

		if (xletContainer == null) {
			// Create Container for Xlet
			xletContainer = applicationManager.getFrame().createXletContainer(name);
		}
		xletContainer.setVisible(false);
		return xletContainer;
	}

	public ClassLoader getClassLoader() {
		if (classLoader == null) {
			// Create classLoader for xlet.
			//  For security reasons, use null as the parent.  This sets the parent class loader to the
			//  bootstrap class loader; and the Xlet will not have access to the system class loader
			//  and the XletApplicationManager classes.

			final URL[] xletPathURLs = XletApplicationManager.getURLs(getPath());
			classLoader = (URLClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
				public Object run() {
					return new URLClassLoader(xletPathURLs, null);
				}
			});

		}
		return classLoader;
	}

	void cleanup() {
		commandDispatcher.stop();
		synchronized (this) {
			if (waiter != null) {
				synchronized (waiter) {
					waiter.notifyAll();
				}
			}
		}
	}
}
