package com.ibm.oti.xlet;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2005  All Rights Reserved
 */

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.microedition.xlet.UnavailableContainerException;

import com.ibm.oti.awt.MessageStrings;
/**
 * A Window for the Xlet Application Manager to display and control Xlets.
 */
public class XletApplicationManagerFrame extends Frame implements XletListener {

	protected MenuBar mb;
	protected Menu xletViewSubmenu;
	protected MenuItem initMenuItem;
	protected MenuItem startMenuItem;
	protected MenuItem pauseMenuItem;
	protected MenuItem destroyMenuItem;
	protected MenuItem destroyUnconditionalMenuItem;
	protected MenuItem removeMenuItem;

	// Status bar
	protected TextField statusMessageArea;
	// Container
	protected Container xletContainerArea;
	protected CardLayout xletContainerLayout;

	protected XletApplicationManager manager;
	protected XletApplication currentXlet;

	/**
	 * Constructs a Frame for the xletApplicationManager
	 */
	public XletApplicationManagerFrame(XletApplicationManager xletApplicationManager) {

		manager = xletApplicationManager;
		manager.addXletListener(this);

		final Dimension screenDimension = new Dimension(400, 400);

		setTitle(MessageStrings.getString("BBAWTXLET_FRAMETITLE")); //$NON-NLS-1$
		setLayout(new BorderLayout());
		setLocation(0,0);

		//add an event listener to deal with the closing of the window
		addWindowListener(new WindowAdapter() {
			public void windowIconified(WindowEvent e) {
				for (Enumeration en = manager.getXlets() ; en.hasMoreElements() ; ) {
					XletApplication xlet = (XletApplication)en.nextElement();
					if (xlet.getXletState() == XletState.XLET_STATE_ACTIVE)
						xlet.pauseXlet();
				}
			}
			public void windowDeiconified(WindowEvent e) {
				for (Enumeration en = manager.getXlets() ; en.hasMoreElements() ; ) {
					XletApplication xlet = (XletApplication)en.nextElement();
					if (xlet.getXletState() == XletState.XLET_STATE_PAUSED)
						xlet.startXlet();
				}
			}
			public void windowClosing(WindowEvent e) {
				for (Enumeration en = manager.getXlets() ; en.hasMoreElements() ; ) {
					XletApplication xlet = (XletApplication)en.nextElement();
					if (xlet.getXletState() != XletState.XLET_STATE_DESTROYED)
						xlet.destroyXlet(true);
				}

				Window window = e.getWindow();
				if (window != null) {
					window.dispose();
				}
			}
			public void windowClosed(WindowEvent e) {
				// Attempt to take down the VM.
				// Note that the SecurityManager is checked to see whether
				// the running program is allowed to terminate itself. If
				// not allowed, an exception will be thrown.
				// The Personal Basis TCK sets a security manager which will
				// not allow this frame to terminate. As a result, a deadlock
				// will occur within the AddWIndowListener test, which is awaiting
				// notification of the window closed event, unless we catch the
				// exception and allow the event processing to continue.
				try {
					System.exit(0);
				} catch (SecurityException ex) {
					// Sorry, exit not permitted.
				}
			}
		});

		buildMenuBar();
		buildStatusArea();

		// Xlet Container area
		xletContainerLayout = new CardLayout();
		xletContainerArea = new Container() {
			public Dimension getPreferredSize() {
				return screenDimension;
			}
		};
		xletContainerArea.setSize(screenDimension.width, screenDimension.height);
		xletContainerArea.setLayout(xletContainerLayout);
		add(xletContainerArea, BorderLayout.CENTER);

		pack();
		initialize();
	}

	private void buildMenuBar() {
		final Frame parentFrame = this;

		//create the menubar, and add items to it
		mb = new MenuBar();

		// Xlet menu
		Menu menu = new Menu(MessageStrings.getString("BBAWTXLET_MENU_XLET")); //$NON-NLS-1$

		// New... choice
		MenuItem item = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_NEW")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				NewXletDialog newXletDialog = new NewXletDialog(parentFrame, manager);
				newXletDialog.show();
			}
		});
		menu.add(item);

		menu.addSeparator();

		// View submenu
		xletViewSubmenu = new Menu(MessageStrings.getString("BBAWTXLET_MENU_VIEW")); //$NON-NLS-1$
		xletViewSubmenu.setEnabled(false);
		menu.add(xletViewSubmenu);

		menu.addSeparator();

		// Exit choice
		item = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_EXIT")); //$NON-NLS-1$
		item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//post the window-closing event:
				dispatchEvent(new WindowEvent(XletApplicationManagerFrame.this, WindowEvent.WINDOW_CLOSING));
				//If this WINDOW_CLOSING event sends dispose() to the
				//window, dispose() will post a WINDOW_CLOSED event.
			}
		});
		menu.add(item);

		mb.add(menu);

		// Action menu
		menu = new Menu(MessageStrings.getString("BBAWTXLET_MENU_ACTIONS")); //$NON-NLS-1$

		// Init
		initMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_INITIALIZE")); //$NON-NLS-1$
		initMenuItem.setEnabled(false);
		initMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					getCurrentXlet().initXlet();
			}
		});
		menu.add(initMenuItem);

		// Start
		startMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_START")); //$NON-NLS-1$
		startMenuItem.setEnabled(false);
		startMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					getCurrentXlet().startXlet();
			}
		});
		menu.add(startMenuItem);

		// Pause
		pauseMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_PAUSE")); //$NON-NLS-1$
		pauseMenuItem.setEnabled(false);
		pauseMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					getCurrentXlet().pauseXlet();
			}
		});
		menu.add(pauseMenuItem);

		// Destroy
		destroyMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_DESTROY")); //$NON-NLS-1$
		destroyMenuItem.setEnabled(false);
		destroyMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					getCurrentXlet().destroyXlet(false);
			}
		});
		menu.add(destroyMenuItem);

		// Destroy Unconditionally
		destroyUnconditionalMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_DESTROY_UNCOND")); //$NON-NLS-1$
		destroyUnconditionalMenuItem.setEnabled(false);
		destroyUnconditionalMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					getCurrentXlet().destroyXlet(true);
			}
		});
		menu.add(destroyUnconditionalMenuItem);

		menu.addSeparator();

		// Remove
		removeMenuItem = new MenuItem(MessageStrings.getString("BBAWTXLET_MENUITEM_REMOVE")); //$NON-NLS-1$
		removeMenuItem.setEnabled(false);
		removeMenuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
					manager.removeXlet(getCurrentXlet());
			}
		});
		menu.add(removeMenuItem);

		mb.add(menu);

		setMenuBar(mb);
	}

	private void buildStatusArea() {
		// Status message area
		statusMessageArea = new TextField(MessageStrings.getString("BBAWTXLET_TEXTFIELD_NOXLETS")); //$NON-NLS-1$
		statusMessageArea.setEditable(false);
		add(statusMessageArea, BorderLayout.SOUTH);
	}

	private void updateActionMenu(XletApplication xlet) {
		//default values
		initMenuItem.setEnabled(false);
		startMenuItem.setEnabled(false);
		pauseMenuItem.setEnabled(false);
		destroyMenuItem.setEnabled(true);
		destroyUnconditionalMenuItem.setEnabled(true);
		removeMenuItem.setEnabled(false);

		// enable/disable action menu items
		XletState state = xlet.getXletState();
		if (state == XletState.XLET_STATE_LOADED) {
			initMenuItem.setEnabled(true);
		} else if (state == XletState.XLET_STATE_PAUSED) {
			startMenuItem.setEnabled(true);
		} else if (state == XletState.XLET_STATE_ACTIVE) {
			pauseMenuItem.setEnabled(true);
		} else if (state == XletState.XLET_STATE_DESTROYED) {
			destroyMenuItem.setEnabled(false);
			destroyUnconditionalMenuItem.setEnabled(false);
			removeMenuItem.setEnabled(true);
		}
	}

	private void showStatus(String msg) {
		statusMessageArea.setText(msg);
	}

	private void initialize() {
		setVisible(true);
		xletContainerArea.requestFocus();
	}

	public void xletAdded(final XletApplication xlet) {
		// Add xlet to View submenu
		CheckboxMenuItem item = new CheckboxMenuItem(xlet.getName());
		item.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String xletName = (String) e.getItem();
				if (xlet.getName() != xletName) {
					throw new RuntimeException("internal error"); //$NON-NLS-1$
				}
				setXletFocus(xlet);
			}
		});
		xletViewSubmenu.add(item);
		xletViewSubmenu.setEnabled(true);

		// Add the container for the xlet
		try {
			xletContainerArea.add(xlet.getContainer(), xlet.getName());
		} catch (UnavailableContainerException e) {
			throw (new RuntimeException(e.toString()));
		}

		//set the focus to this new xlet
		setXletFocus(xlet);
	}

	public void xletRemoved(XletApplication removedXlet, XletApplication nextXlet) {
		// Remove xlet from View submenu
		for (int i=0; i < xletViewSubmenu.getItemCount(); i++) {
			MenuItem mi = xletViewSubmenu.getItem(i);
			if (mi.getLabel().equals(removedXlet.getName())) {
				xletViewSubmenu.remove(mi);
				if (xletViewSubmenu.getItemCount() == 0){
					xletViewSubmenu.setEnabled(false);
				}
				removeMenuItem.setEnabled(false);
				break;
			}
		}

		setXletFocus(nextXlet);

		// Remove the container for the xlet
		try {
			xletContainerArea.remove(removedXlet.getContainer());
		} catch (UnavailableContainerException e) {
			throw (new RuntimeException(e.toString()));
		}
	}

	public void xletStateChanged(XletApplication xlet, XletState newState) {
		if (currentXlet == xlet) {
			// enable/disable Action menu items
			updateActionMenu(xlet);

			// update status message
			showStatus(MessageStrings.getString("BBAWTXLET_WHICH_XLET") + xlet.getName() + " [" + xlet.getXletState() + "]");	 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
	}

	void setXletFocus(XletApplication xlet) {
		setCurrentXlet(xlet);
		// Select (checkbox) xlet in View submenu
		for (int i = 0; i < xletViewSubmenu.getItemCount(); i++) {
			CheckboxMenuItem mi = (CheckboxMenuItem) xletViewSubmenu.getItem(i);
			if (mi.getLabel().equals(xlet.getName())) {
				mi.setState(true);
			} else {
				mi.setState(false);
			}
		}

		// enable/disable Action menu items
		updateActionMenu(xlet);

		// update status message
		showStatus(MessageStrings.getString("BBAWTXLET_WHICH_XLET") + xlet.getName() + " [" + xlet.getXletState() + "]");	 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

		// Update Title
		setTitle(MessageStrings.getString("BBAWTXLET_FRAME_APPMANAGER") + xlet.getName()); //$NON-NLS-1$

		// Display container for xlet
		xletContainerLayout.show(xletContainerArea, xlet.getName());
		xletContainerArea.requestFocus();
	}

	private XletApplication getCurrentXlet() {
		return currentXlet;
	}

	private void setCurrentXlet(XletApplication xlet) {
		currentXlet = xlet;
	}

	Container createXletContainer(String name) {
		Container container = new Container();

		// Set the size to the usable area of the frame
		container.setSize(xletContainerArea.getSize());

		return container;
	}
}
