package com.ibm.oti.xlet;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2006  All Rights Reserved
 */

import javax.microedition.xlet.Xlet;
import javax.microedition.xlet.XletStateChangeException;

/**
 * The XletCommandDispatcher runs on a separate thread.  One command dispatcher and thread per Xlet.
 * It receives xlet lifecycle commands and dispatches them to the actual Xlet.
 */
public class XletCommandDispatcher implements Runnable {

	private Xlet xlet;			// real Xlet
	private XletApplication xletApplication;		// Xlet wrapper
	private XletCommandPipe commandPipe;
	private boolean done;
	private Thread dispatcherThread;

	/**
	 * Constructor.
	 *
	 * @param xlet   the Xlet to dispatch commands to
	 * @param xletApplication   the XletApplication that is sending the commands
	 * @param commandPipe   the communications pipe used to send xlet commands
	 */
	public XletCommandDispatcher(Xlet xlet, XletApplication xletApplication, XletCommandPipe pipe) {
		this.xlet = xlet;
		this.xletApplication = xletApplication;
		commandPipe = pipe;
		done = false;
	}

	/**
	 * Runs forever, dispatching commands to the Xlet.
	 */
	public void run() {
		int command;
		while (true) {
			synchronized (commandPipe) {
				if (done) return;
				try {
					command = commandPipe.readCommand();
				} catch (InterruptedException ex) {
					// If this thread was interrupted then maybe
					// the dispatcher is done.  Continue the loop to check
					continue;
				}
			}

			dispatchCommand(command);
		}
	}

	/*
	 * Receive and dispatch an xlet command.
	 *
	 * This dispatcher will block those commands that are not allowed for the current Xlet state.
	 * These blocked commands are ignored; per the specification.  These checks are performed by the
	 * dispatcher, since it runs on the same thread as Xlet and the state can be determined
	 * accurately (without race conditions).
	 *
	 * Following the completion of a command or an error, the dispatcher will call back to the
	 * xlet's state machine, with an appropriate event
	 */
	private void dispatchCommand(int command) {
		try {
			switch (command) {
				case XletCommandPipe.XLET_INIT :
					if (xletApplication.getXletState() == XletState.XLET_STATE_LOADED){
						xlet.initXlet(xletApplication);
						xletApplication.getXletState().eventXletInitialized(xletApplication);
					}
					break;
				case XletCommandPipe.XLET_START :
					if (xletApplication.getXletState() == XletState.XLET_STATE_PAUSED){
						xlet.startXlet();
						xletApplication.getXletState().eventXletStarted(xletApplication);
					}
					break;
				case XletCommandPipe.XLET_PAUSE :
					if (xletApplication.getXletState() == XletState.XLET_STATE_ACTIVE){
						xlet.pauseXlet();
						xletApplication.getXletState().eventXletPaused(xletApplication);
					}
					break;
				case XletCommandPipe.XLET_DESTROY :
					if (xletApplication.getXletState() != XletState.XLET_STATE_DESTROYED){
						xlet.destroyXlet(false);
						xletApplication.getXletState().eventXletDestroyed(xletApplication, false);
					}
					break;
				case XletCommandPipe.XLET_DESTROY_UNCONDITIONAL :
					if (xletApplication.getXletState() != XletState.XLET_STATE_DESTROYED){
						try{
							xlet.destroyXlet(true);
						}
						catch(XletStateChangeException ex){}
						xletApplication.getXletState().eventXletDestroyed(xletApplication, true);
					}
					break;
				default :
					throw new RuntimeException("Invalid Xlet command: " + command); //$NON-NLS-1$
			}

		} catch (XletStateChangeException e) {
			/* Callback to xlet's state machine that an XletStateChangeException was thrown by the Xlet.
	 		 * This callback is needed because the Xlet runs on a separate thread and the xlet commands
			 * are run asyncronously.
			 */
			xletApplication.getXletState().eventXletStateChangeException(xletApplication, command, e.toString());

		} catch (Throwable t) {
			/* Callback to xlet's state machine that an Xlet Throwable error was thrown by the Xlet.
			 * This callback is needed because the Xlet runs on a separate thread and the xlet commands
			 * are run asyncronously.
			 */
			t.printStackTrace();
			xletApplication.getXletState().eventXletError(xletApplication, command, t.toString());
		}
	}

	public void start() {
		// Avoid starting the dispatcher thread more than once.
		// This happens when the TCK calls initXlet after the xlet
		// has already started.  If a second thread is allowed to start
		// then the first thread will be lost and will not be properly
		// terminated when the xlet is finally destroyed.
		if (dispatcherThread == null) {
			String name = xletApplication.getName();

			// start the xlet command dispatcher thread
			//    For security reasons, create a new ThreadGroup for all of the Xlet's threads to be part of
			dispatcherThread = new Thread(
				new ThreadGroup(name),
				this,
				"Xlet Dispatcher Thread - " + name); //$NON-NLS-1$

			dispatcherThread.start();
		}
	}

	public void stop() {
		synchronized (commandPipe) {
			done = true;
			if (dispatcherThread != null) commandPipe.interruptIfWaiting(dispatcherThread);
		}
	}
}
