package com.ibm.oti.xlet;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2005  All Rights Reserved
 */

import java.util.Vector;

/**
 * A simple communications pipe for communications between threads.
 * It is used to send Xlet commands between the XletApplicationManager and the XletCommandDispatcher.
 * The Application Manager writes xlet commands to the pipe.  The writes are asyncronous and do
 * not block.  The XletCommandDispatcher runs on a separate thread and reads the commands.
 * If no command is available, the thread will block.
 */
public class XletCommandPipe {

	// Xlet Commands
	public static final int XLET_INIT = 1;
	public static final int XLET_START = 2;
	public static final int XLET_PAUSE = 3;
	public static final int XLET_DESTROY = 4;
	public static final int XLET_DESTROY_UNCONDITIONAL = 5;

	private Vector pipe = new Vector();
	private Object waiter;
	private boolean isIdle=false;

	/**
	 * Reads the next xlet command from this pipe. If no commands are available,
	 * this method blocks until a command is available.
	 */
	public synchronized int readCommand() throws InterruptedException {
		while (pipe.size() == 0) {

			if (waiter != null) {
				synchronized (waiter) {
					waiter.notifyAll();
					waiter = null;
				}
			}

			isIdle = true;
			wait();
			isIdle =  false;
		}
		return ((Integer) pipe.remove(0)).intValue();
	}

	/**
	 * Writes the specified xlet command to the pipe.  This method will not block.
	 */
	public synchronized void writeCommand(int command) {
		pipe.add(new Integer(command));
		if (isIdle) notify();
		// Yield to give the XletDispatcher a chance to issue the command to the Xlet
		Thread.yield();
	}

	/**
	 * Returns a name for the specified command.
	 */
	public static String getCommandName(int command) {
		switch (command) {
			case XLET_INIT :
				return "InitXlet"; //$NON-NLS-1$

			case XLET_START :
				return "StartXlet"; //$NON-NLS-1$

			case XLET_PAUSE :
				return "PauseXlet"; //$NON-NLS-1$

			case XLET_DESTROY :
				return "DestroyXlet"; //$NON-NLS-1$

			case XLET_DESTROY_UNCONDITIONAL :
				return "DestroyUnconditionalXlet"; //$NON-NLS-1$
		}
		return "invalid command"; //$NON-NLS-1$
	}

	public synchronized boolean notifyWhenIdle(Object newWaiter) {
		if (isIdle) return true;
		waiter = newWaiter;
		return false;
	}

	public synchronized void interruptIfWaiting(Thread thread) {
		if (isIdle) thread.interrupt();
	}

}
