package com.ibm.oti.xlet;

import javax.microedition.xlet.ixc.IxcRegistry;

import com.ibm.oti.xlet.ixc.IxcRegistryImpl;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2005  All Rights Reserved
 */

/**
 * This class implements the state machine for an Xlet.  Each state is a private subclass of
 * this class.
 */
public abstract class XletState {

	// Xlet states
	static final XletState_Loaded XLET_STATE_LOADED = new XletState_Loaded();
	static final XletState_Paused XLET_STATE_PAUSED = new XletState_Paused();
	static final XletState_Active XLET_STATE_ACTIVE = new XletState_Active();
	static final XletState_Destroyed XLET_STATE_DESTROYED = new XletState_Destroyed();

	/**
	 * Start the state machine.  Initial state is Loaded.
	 */
	public static XletState start() {
		return XLET_STATE_LOADED;
	}

	/**
	 * Called to enter the state (state is the receiver)
	 */
	public void enter(XletApplication xlet) {
		// update the XletApplication with the new state.
		xlet.setXletState(this);
	}

	/***********************************************************
	 * Events - state machine events
	 *
	 * All events are called from the Xlet's thread.
	 *
	 * The abstract class contains the default behavior.
	 * See the subclass states for overridden behaviors.
	 ***********************************************************/

	/**
	 * Event - after the initXlet method returns successfully
	 *
	 * @param 		xlet 	XletApplication that initialized
	 */
	public void eventXletInitialized(XletApplication xlet) {
		throw new RuntimeException("Invalid Xlet state transition. From state '" + this.toString() + "' with event 'XletInitialized"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Event - after the startXlet method returns successfully
	 *
	 * @param 		xlet 	XletApplication that started
	 */
	public void eventXletStarted(XletApplication xlet) {
		throw new RuntimeException("Invalid Xlet state transition. From state '" + this.toString() + "' with event 'XletStarted"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Event - after the pauseXlet method returns successfully
	 *
	 * @param 		xlet 	XletApplication that paused
	 */
	public void eventXletPaused(XletApplication xlet) {
		throw new RuntimeException("Invalid Xlet state transition. From state '" + this.toString() + "' with event 'XletPaused"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Event - after the destroyXlet method returns successfully
	 *
	 * @param 		xlet 	XletApplication that was destroyed
	 */
	public void eventXletDestroyed(XletApplication xlet, boolean unconditional) {
		// transition to the Destroyed state
		XLET_STATE_DESTROYED.enter(xlet);
	}

	/**
	 * Event - Xlet threw a XletStateChangeException, signaling that a state change failed
	 *
	 * @param 		xlet 	XletApplication that threw XletStateChangeException
	 * @param 		command 	command being executed
	 * @param 		description description of exception
	 */
	public void eventXletStateChangeException(XletApplication xlet, int command, String description) {

		// System.out.println("xlet '" + xlet.getName() + "' threw a XletStateChangeException for command '" + XletCommandPipe.getCommandName(command) + "':" + description);

		if (command == XletCommandPipe.XLET_DESTROY) {
			// Xlet has requested not to be destroyed.  Remain in current state.
		} else {
			xlet.destroyXlet(true);
		}
	}

	/**
	 * Event - Xlet raised an Error
	 *
	 * @param 		xlet 	XletApplication that threw XletStateChangeException
	 * @param 		command 	command being executed
	 * @param 		description 	description of error
	 */
	public void eventXletError(XletApplication xlet, int command, String description) {

		// System.out.println("xlet '" + xlet.getName() + "' had an error for command '" + XletCommandPipe.getCommandName(command) + "':" + description);

		if ((command == XletCommandPipe.XLET_DESTROY) ||
			(command == XletCommandPipe.XLET_DESTROY_UNCONDITIONAL)
			) {
				// Xlet had error on a destroy command.  Force the Xlet into the Destroyed state.
				XLET_STATE_DESTROYED.enter(xlet);
		}
		xlet.destroyXlet(true);
	}

	/**
	 * Event - Xlet called nofifyPaused
	 *
	 * @param 		xlet 	XletApplication
	 */
	public void eventXletCalledNotifyPaused(XletApplication xlet) {
		throw new RuntimeException("Invalid Xlet state transition. From state '" + this.toString() + "' with event 'XletCalledNotifyPaused"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Event - Xlet called nofifyDestroyed
	 *
	 * @param 		xlet 	XletApplication
	 */
	public void eventXletCalledNotifyDestroyed(XletApplication xlet) {
		// Xlet already destroyed itself.  Force xlet into destroyed state.
		XLET_STATE_DESTROYED.enter(xlet);
	}

	public abstract String toString();

	/********************************************
	 * Class representing the Loaded state
	 ********************************************/
	private static class XletState_Loaded extends XletState {

		/**
		 * Event - after the initXlet method returns successfully
		 *
		 * @param 		xlet 	XletApplication that initialized
		 */
		public void eventXletInitialized(XletApplication xlet) {
			XLET_STATE_PAUSED.enter(xlet);
		}

		public String toString() {
			return "Loaded"; //$NON-NLS-1$
		}
	}

	/********************************************
	 * Class representing the Paused state
	 ********************************************/
	private static class XletState_Paused extends XletState {

		/**
		 * Event - Xlet called nofifyPaused
		 *
		 * @param 		xlet 	XletApplication
		 */
		public void eventXletCalledNotifyPaused(XletApplication xlet) {
			//already in Paused state; ignore
		}

		/**
		 * Event - after the startXlet method returns successfully
		 *
		 * @param 		xlet 	XletApplication that started
		 */
		public void eventXletStarted(XletApplication xlet) {
			XLET_STATE_ACTIVE.enter(xlet);
		}

		/**
		 * Event - Xlet threw a XletStateChangeException, signaling that a state change failed
		 *
		 * @param 		xlet 	XletApplication that threw XletStateChangeException
		 * @param 		command 	command being executed
		 * @param 		description description of exception
		 */
		public void eventXletStateChangeException(XletApplication xlet, int command, String description) {
			if (command == XletCommandPipe.XLET_START) {
				// Xlet did not start service, stay in Paused state
				return;
			}
			super.eventXletStateChangeException(xlet, command, description);
		}

		public String toString() {
			return "Paused"; //$NON-NLS-1$
		}
	}

	/********************************************
	 * Class representing the Active state
	 ********************************************/
	private static class XletState_Active extends XletState {

		/**
		 * Event - after the pauseXlet method returns successfully
		 *
		 * @param 		xlet 	XletApplication that paused
		 */
		public void eventXletPaused(XletApplication xlet) {
			XLET_STATE_PAUSED.enter(xlet);
		}

		/**
		 * Event - Xlet called nofifyPaused
		 *
		 * @param 		xlet 	XletApplication
		 */
		public void eventXletCalledNotifyPaused(XletApplication xlet) {
			XLET_STATE_PAUSED.enter(xlet);
		}

		public String toString() {
			return "Active"; //$NON-NLS-1$
		}
	}

	/********************************************
	 * Class representing the Destroyed state
	 ********************************************/
	private static class XletState_Destroyed extends XletState {

		/**
		 * Called when the state is entered
		 */
		public void enter(XletApplication xlet) {
			// Clean up the IXC registry entries owned by this Xlet
			((IxcRegistryImpl)(IxcRegistry.getRegistry(xlet))).unbindAllForXletContext();
			super.enter(xlet);
			xlet.cleanup();
		}

		/**
		 * Event - after the destroyXlet method returns successfully
		 *
		 * @param 		xlet 	XletApplication that was destroyed
		 */
		public void eventXletDestroyed(XletApplication xlet, boolean unconditional) {
			//already in Destroyed state; ignore
	 	}

		/**
		 * Event - Xlet raised an Error
		 *
		 * @param 		xlet 	XletApplication that threw XletStateChangeException
		 * @param 		command 	command being executed
		 * @param 		description description of error
		 */
		public void eventXletError(XletApplication xlet, int command, String description) {
			//in Destroyed state; ignore
	 	}

		/**
		 * Event - Xlet called nofifyDestroyed
		 *
		 * @param 		xlet 	XletApplication
		 */
		public void eventXletCalledNotifyDestroyed(XletApplication xlet) {
			//already in Destroyed state; ignore
		}

		public String toString() {
			return "Destroyed"; //$NON-NLS-1$
		}

		public void eventXletStateChangeException(XletApplication xlet, int command, String description) {
			//already in Destroyed state; ignore
		}
	}
}
