package com.ibm.oti.xlet.ixc;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2006  All Rights Reserved
 */

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.microedition.xlet.XletContext;

import javax.microedition.xlet.ixc.IxcPermission;
import javax.microedition.xlet.ixc.IxcRegistry;
import javax.microedition.xlet.ixc.StubException;

public class IxcRegistryImpl extends IxcRegistry {
	//private static IxcRegistryImpl serverRegistry;
	private static Hashtable nameToStub = new Hashtable();

	private XletContext context;

	static final String LOOKUP_PERMISSION	= "lookup";		//$NON-NLS-1$
	static final String BIND_PERMISSION		= "bind";		//$NON-NLS-1$

	public IxcRegistryImpl(XletContext context) {
		this.context =context;
	}

	public Remote lookup(String name) throws StubException, NotBoundException {
		checkPermission(name, LOOKUP_PERMISSION);

		Remote stub = (Remote) nameToStub.get(name);
		if (stub == null) throw new NotBoundException(name);

		Remote ro =IxcServerTable.getRemoteObjectForStub(stub);
		XletContext context_ro = IxcServerTable.toContext(ro);
		XletContext context_stub = IxcServerTable.toContext(stub);

		// check if this is a remote object created in this registry's context,
		// but has been registered by another context.
		if (context == context_ro) {
			return ro;
		} else {
			// if this is a stub that had been exported by another context,
			// but bound by this registry
			if (context == context_stub) {
				return stub;
			} else {
				// if this is a stub that has been exported and
				// bound by another xletcontext
				return copyStub(name, stub);
			}
		}
	}

	private Remote copyStub(String name, Remote stub) throws StubException {
		Remote result;
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream out = new ObjectOutputStream(baos);
			out.writeObject(stub);
			out.close();

			byte[] objBytes = baos.toByteArray();

			ObjectInputStream in =
				new XletObjectInputStream(new ByteArrayInputStream(objBytes), context);
			result = (Remote) in.readObject();
			in.close();
		} catch (IOException e) {
			throw new StubException("stub could not be created for:" + name, e); //$NON-NLS-1$
		} catch (ClassNotFoundException e) {
			throw new StubException("stub could not be created for:" + name, e); //$NON-NLS-1$
		}

		// record that this context is the importer of this stub
		IxcServerTable.changeImporterContextForStub(result, context);
		return result;
	}

	public void bind(String name, Remote o) throws StubException, AlreadyBoundException {
		checkPermission(name, BIND_PERMISSION);
		synchronized(nameToStub) {
			Remote stub = (Remote)nameToStub.get(name);
			if (stub != null)
				throw new AlreadyBoundException(name);

			stub = IxcServerTable.toStub(o,context);
			nameToStub.put(name, stub);
		}
	}

	/**
	 * Unbinds the remote object with the given name from the registry,
	 * if it has been bound by this XletContext originally.
	 * If the remote object with the given name is bound by another XletContext,
	 * this method will throw an AccessException.
	 *
	 */
	public void unbind(String name) throws NotBoundException, AccessException {
		synchronized(nameToStub) {
			Remote stub = (Remote)nameToStub.get(name);
			if (stub == null) {
				checkPermission(name, BIND_PERMISSION);
				throw new NotBoundException(name);
			}

			// check if this stub had been bound by this registry
			XletContext context_stub = IxcServerTable.toContext(stub);

			if (context!=context_stub) {
				checkPermission(name, BIND_PERMISSION);
				throw new AccessException(name);
			}

			nameToStub.remove(name);
		}
	}

	/**
	 * Unbinds all the remote objects that has been
	 * bound by this XletContext from the registry.
	 * This method is for internal use only.
	 *
	 */
	public void unbindAllForXletContext() {
		synchronized(nameToStub) {
			Enumeration keys = nameToStub.keys();
			while (keys.hasMoreElements()) {
				String name = (String)keys.nextElement();
				Remote stub = (Remote)nameToStub.get(name);
				XletContext context_stub = IxcServerTable.toContext(stub);
				if (context ==context_stub)
					nameToStub.remove(name);
			}

		}
	}

	public void rebind(String name, Remote o) throws StubException, AccessException {
		checkPermission(name, BIND_PERMISSION);
		synchronized(nameToStub) {
			Remote stub = (Remote) nameToStub.get(name);
			if (stub!=null) {
			// check if this stub had been bound by this registry
			XletContext context_stub = IxcServerTable.toContext(stub);
			if (context!=context_stub)
				throw new AccessException(name);
			}

			Remote newStub = IxcServerTable.toStub(o,context);
			nameToStub.put(name, newStub);
		}
	}

	/**
	 * Lists the names of all the remote objects in the registry
	 * that has been bound by all the XletContexts.
	 *
	 */
	public String[] list() {
		ArrayList list = new ArrayList(nameToStub.size());
		synchronized(nameToStub) {
			Enumeration keys = nameToStub.keys();
			while (keys.hasMoreElements()) {
				try {
					String name = (String)keys.nextElement();
					checkPermission(name, LOOKUP_PERMISSION);
					list.add(name);
				} catch (SecurityException e) {
					// NOP - No Permission!
				}
			}
		}
		String[] result = new String[list.size()];
		return (String[])list.toArray(result);
	}

	/**
	 * Lists only the names of the remote objects
	 * which has been bound to the registry by this XletContext.
	 * This methos is for internal use only.
	 */
	public String[] listForXletContext() {
		int count = 0;
		String[] all = new String[nameToStub.size()];
		synchronized(nameToStub) {
			Enumeration keys = nameToStub.keys();
			while (keys.hasMoreElements()) {
				String name = (String)keys.nextElement();
				Remote stub = (Remote)nameToStub.get(name);
				XletContext context_stub = IxcServerTable.toContext(stub);
				if (context ==context_stub) {
					all[count]=name;
					count++;
				}
			}
		}
		String[] result = new String[count];
		for (int i=0; i < count; i++) {
			result [i] = all[i];
		}
		return result;
	}

	public void unbindAll() {
		unbindAllForXletContext();
	}

	void checkPermission(final String name, final String permissionName) {
		final SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkPermission(new IxcPermission(name, permissionName)); //$NON-NLS-1$
		}
	}

}
