package java.applet;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.*;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Locale;

import com.ibm.oti.awt.AppletAudioClip;

import java.io.ObjectInputStream;

/**
 * An applet is a small application intended to be embedded within a
 * larger application.  The behavior of the applet is restricted by the
 * containing application.  The Applet class is the superclass of all applets.
 * Applets interact with their containing environment through the methods
 * provided by this class, and must override certain methods in order to
 * exhibit their own functionality.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			AppletStub
 * @see			AppletContext
 * @see			AudioClip
 */
public class Applet extends Panel {

	static final long serialVersionUID = -5836846270535785031L;

	private transient AppletStub stub;

	/**
	 *  Construct a new Applet.
	 *
	 */
	public Applet() throws HeadlessException {
		super();
	}

	/**
	 *  Sets the stub object of this applet.  This method is
	 *  called automatically by the applet's containing environment,
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		stub	the stub of this applet
	 *
	 */
	public final void setStub(AppletStub stub) {
		this.stub = stub;
	}

	/**
	 * Answers true if the receiver is active and false otherwise.
	 * An applet becomes active just before its <code>start()</code> method
	 * is called and ceases to be active just before its <code>stop()</code>
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		boolean
	 *				whether or not the receiver is active.
	 *
	 */
	public boolean isActive() {
		return (stub == null) ? false : stub.isActive();
	}

	/**
	 * Returns this applet's document base URL.  Let the 'document URL' be the
	 * URL of the document in which this applet was embedded.  Thus, the 'document
	 * base URL' is the part of the document URL excluding the documents name.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		the document base URL of this applet
	 *
	 */
	public URL getDocumentBase() {
		return stub.getDocumentBase();
	}

	/**
	 * Returns this applet's code base URL.  Let the 'code URL' be the
	 * URL indicating the location of this applet's class file.  Thus, the 'code
	 * base URL' is the part of the code URL excluding the applet's class filename.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		the code base URL of this applet
	 *
	 */
	public URL getCodeBase() {
		return stub.getCodeBase();
	}

	/**
	 * Returns the value of the applet parameter named <code>name</code>.
	 * Applet parameters are typically specified in the HTML document containing
	 * the applet.  If this is the case, this method will return the <code>value</code>
	 * attribute of a <code>param</code> tag (nested within this applet's
	 * <code>applet</code> tag) whose <code>name</code> attribute is <code>name</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		name	the name of the parameter whose value is to be returned
	 *
	 * @return		the value of the parameter with the specified name, or <code>null</code>
	 *				if the named parameter does not exist
	 */
	public String getParameter(String name) {
		return stub.getParameter(name);
	}

	/**
	 * Returns this applet's applet context.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		this applet's applet context
	 *
	 * @see			AppletContext
	 */
	public AppletContext getAppletContext() {
		return stub.getAppletContext();
	}

	/**
	 * Requests that this applet be resized.  The containing environment
	 * may ignore this request.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		width	The intended new width for the applet.
	 * @param		height	The intended new height for the applet.
	 */
	public void resize(int width, int height) {
		Dimension s = getSize();
		if ((s.width == width) && (s.height == height)) return;
		super.resize(width, height);
		if (stub != null) {
			stub.appletResize(width, height);
		}
	}

	/**
	 * Requests that this applet be resized.  The containing environment
	 * may ignore this request.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		newSize		The dimension to which the caller intends to resize
	 *						the applet.
	 */
	public void resize(Dimension newSize) {
		resize(newSize.width, newSize.height);
	}

	/**
	 * Requests that the string <code>msg</code> be displayed in the 'status
	 * window'.  Many environments in which an applet can run provide a 'status
	 * window', which is an area always available for the display of text messages.
	 * The environment may ignore this request.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		msg		the text to be displayed in the status window.
	 *						If <code>null</code> the previous message is cleared.
	 */
	public void showStatus(String msg) {
		stub.getAppletContext().showStatus(msg);
	}

	/**
	 * Retrieves an Image object representing the image file located at the URL
	 * <code>url</code>.  This method always returns immediately, and it is only
	 * when the applet attempts to render the image that the image data is loaded.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the absolute URL from which the image should be loaded
	 *
	 * @return		an Image object representing the image to be retrieved
	 */
	public Image getImage(URL url) {
		if ( isValid(url) )
			return stub.getAppletContext().getImage(url);
		else
			return stub.getAppletContext().getImage(getFromJar(url));
	}

	private boolean isValid(URL url) {
		try {
			url.openStream().close() ;	// open and close the stream
			return true ;
		} catch (IOException e) { 		// The url is not valid
			return false ;
		} catch (java.security.AccessControlException e) {	// The url starts with file:/
			return false ;
		}
	}

	/**
	 * Tries to locate a resource in the jar with the same relative path as the
	 * one passed in as the parameter.
	 *
	 * @param imageUrl is the URL of a resource the relative portion of which may be used to locate a resource in the jar
	 * @return the url of the resource in the jar if there is a corresponding resource in the jar, else returns the input URL.
	 */
	private URL getFromJar(URL imageUrl) {
		String relativePath = getRelativePath(imageUrl) ;
		if (relativePath == null) { // the url was not under the codebase, so we can't look in the jar
			return imageUrl ;
		} else {
			// the url was under the codebase, so we get the resource in the jar
			URL jarUrl = getClass().getResource(relativePath) ;
			return (jarUrl == null) ? imageUrl : jarUrl ;
		}
	}

	private String getRelativePath(URL url) { 	// subtract the code base from the url to get the relative path
		String codeBase = normalizeURL(getCodeBase().toString());
		if (codeBase != null && url.toString().indexOf(codeBase) == 0) {
			return url.toString().substring(codeBase.length()) ;
		}
		// Return null to signal that the URL is not under the codebase, so we don't search for the resource in the jar
		return null ;
	}

	/**
	 * removes any '..' or '.' segments from the url
	 * copied from URLStreamHandler.parseURL
	 * @param string
	 * @return
	 */
	private String normalizeURL(String file) {
		if (file == null)
			return null ;
		// modify file if there's any relative referencing
		int dirIndex;
		while ((dirIndex = file.indexOf("/./")) >= 0)
			file = file.substring(0, dirIndex + 1) + file.substring(dirIndex + 3);
		if (file.endsWith("/.")) file = file.substring(0, file.length() - 1);
		while ((dirIndex = file.indexOf("/../")) >= 0) {
			if (dirIndex != 0) {
				file = file.substring(0, file.lastIndexOf('/', dirIndex - 1)) + file.substring(dirIndex + 3);
			} else
				file = file.substring(dirIndex + 3);
		}
		if (file.endsWith("/..") && file.length() > 3)
			file = file.substring(0, file.lastIndexOf('/', file.length() - 4) + 1);
		return file;
	}

	/**
	 * Retrieves an Image object representing the image file located at the URL
	 * specified by the parameters.  <code>name</code> specifies a relative path
	 * including the image filename.  The image URL is composed by considering this
	 * path relative to the absolute URL specified by <code>url</code>.  This method
	 * always returns immediately, and it is only when the applet attempts to
	 * render the image that the image data is loaded.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		a non-<code>null</code> absolute URL
	 * @param		name	a non-<code>null</code> relative path, which,
	 *						when considered relative to <code>url</code>, yields
	 *						the location of the image to be loaded
	 *
	 * @return		an Image object representing the image to be retrieved
	 */
	public Image getImage(URL url, String name) {
		try {
			URL imgURL = new URL(url, name);
			return getImage(imgURL);
		} catch (MalformedURLException e) {
			return null;
		}
	}

	/**
	 * Retrieves an AudioClip object representing the audio clip located at the URL
	 * <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the non-<code>null</code> absolute URL from which the
	 *						audio clip should be loaded
	 *
	 * @return		an AudioClip object representing the audio clip to be retrieved
	 */
	public static final AudioClip newAudioClip(URL url) {
		return new AppletAudioClip(url) ;
	}

	/**
	 * Retrieves an AudioClip object representing the audio clip located at the URL
	 * <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the non-<code>null</code> absolute URL from which the
	 *						audio clip should be loaded
	 *
	 * @return		an AudioClip object representing the audio clip to be retrieved
	 */
	public AudioClip getAudioClip(URL url) {
		if ( isValid(url) )
			return stub.getAppletContext().getAudioClip(url);
		else
			return stub.getAppletContext().getAudioClip(getFromJar(url));
	}

	/**
	 * Retrieves an AudioClip object representing the audio clip located at the URL
	 * specified by the parameters.  <code>name</code> specifies a relative path
	 * including the clip filename.  The clip URL is composed by considering this
	 * path relative to the absolute URL specified by <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		a non-<code>null</code> absolute URL
	 * @param		name	a non-<code>null</code> relative path, which,
	 *						when considered relative to <code>url</code>, yields
	 *						the location of the audioclip to be loaded
	 *
	 * @return		an AudioClip object representing the audio clip to be retrieved
	 */
	public AudioClip getAudioClip(URL url, String name) {
		try {
			URL clipURL = new URL(url, name);
			return getAudioClip(clipURL);
		} catch(MalformedURLException e) {
			return null;
		}
	}

	/**
	 * Returns a possibly <code>null</code> string describing this applet.
	 * This method should be overridden by subclasses of Applet, in order to
	 * provide relevant information.  The default implementation returns
	 * <code>null</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		a possibly <code>null</code> string describing this applet
	 */
	public String getAppletInfo() {
		return null;
	}

	/**
	 * Returns the locale of this applet if it has been set, and otherwise returns
	 * the default locale.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		the locale of this applet, if it has been set, or the default
	 *				locale, if no locale has been set
	 *
	 * @see			Locale
	 */
	public Locale getLocale() {
		try {
			return super.getLocale();
		} catch(IllegalComponentStateException e) {
			return Locale.getDefault();
		}
	}

	/**
	 * Returns a two dimensional array of strings describing the
	 * parameters understood by this applet. This array will be an N by 3
	 * array, where N is the number of different parameters understood by
	 * the applet.  Each of the N groups of three strings describes a parameter.
	 * The first elemtent is the name of the parameter, the second its type,
	 * and the third its meaning.  An example of such an array follows.  If an applet
	 * understood two parameters, one being a URL for an audio clip and one
	 * being an integer between 1 and 50 representing the number of times to repeatedly
	 * play the audio clip, the return array might be:
	 * <code>
	 *	String[][] parmInfo = {
	 *		{"url", "URL", "location of audio clip"},
	 *		{"timesToRepeat","Integer 1-50", "number of times to repeat"}
	 *	};
	 *</code>
	 *
	 * This method should be overridden by subclasses.  The default implementation
	 * returns <code>null</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		information about the parameters understood by this applet
	 *				(see above).  Possibly <code>null</code>
	 *
	 */
	public String[][] getParameterInfo() {
		return null;
	}

	/**
	 * Retrieves and immediately plays the audio clip located at the URL
	 * <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		the non-<code>null</code> absolute URL from which the
	 *						audio clip should be loaded
	 */
	public void play(URL url) {
		play(getAudioClip(url));
	}

	/**
	 * Retrieves and immediately plays the audio clip located at the URL
	 * specified by the parameters.  <code>name</code> specifies a relative path
	 * including the clip filename.  The clip URL is composed by considering this
	 * path relative to the absolute URL specified by <code>url</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @param		url		a non-<code>null</code> absolute URL
	 * @param		name	a non-<code>null</code> relative path, which,
	 *						when considered relative to <code>url</code>, yields
	 *						the location of the audio clip to be loaded
	 */
	public void play(URL url, String name) {
		play(getAudioClip(url, name));
	}

	private void play(AudioClip clip) {
		if (clip == null) return;
		clip.play();
	}

	/**
	 * This method is called by the applet's containing environment just after
	 * the applet is loaded.  Subclasses of Applet should override this method
	 * to perform any work that should be done before <code>start()</code> is
	 * invoked.  The default implementation does nothing.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void init() {
	}

	/**
	 * This method is called by the applet's environment when the applet is made
	 * active.  This may occur repeatedly, but not without an intervening call to
	 * <code>stop()</code>, which represents a transition from an active to an inactive
	 * state.  Also, calls to this method always occur after <code>init()</code>
	 * is called, and before <code>destroy()</code> is called.
	 *
	 * Subclasses generally override this method, the default implementation of which
	 * does nothing.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void start() {
	}

	/**
	 * This method is called by the applet's environment when the applet changes
	 * from an active state to an inactive state.
	 * This may occur repeatedly, but not without an intervening call to
	 * <code>start()</code>, which represents a transition from an inactive to an active
	 * state.  Also, calls to this method always occur after at least one call to
	 * <code>start()</code>, after the calling of <code>init()</code>, and before
	 * the calling of <code>destroy()</code>.
	 *
	 * Subclasses generally override this method, the default implementation of which
	 * does nothing.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void stop() {
	}

	/**
	 * This method is called by the applet's containing environment when it
	 * decides to dispose of the loaded applet and any resources it is
	 * using.  This will only be called when the applet is inactive (i.e. <code>
	 * stop()</code> will be called first).
	 *
	 * Subclasses generally override this method, the default implementation of which
	 * does nothing.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 */
	public void destroy() {
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException, HeadlessException {
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}
}
