package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.util.EventObject;

/**
 * AWTEvent is the is the abstract superclass of all AWT event classes.
 * There are a number of existing subclasses of this class
 * which use IDs less than the value of the field <code>RESERVED_ID_MAX</code>
 * and subclasses added outside of the java.awt.event package should specify
 * an ID larger than the reserved maximum.  This class specifies a number of
 * event masks used by the <code>Component</code> classes in <code>enableEvents()</code>
 * or enabled automatically when a listener is added.
 *
 * @author		IBM
 * @version		initial
 *
 */
public abstract class AWTEvent extends EventObject {

static final long serialVersionUID = -1825314779160409405L;

/**
 * flag to indicate if this event was consumed
 */
protected boolean consumed;

/**
 * MASK id for the receiver.
 */
protected int id;

/**
 * Corresponding EVENT_MASK (Log 2) (see below)
 */
private int eventMaskLog2;
/**
 * Masks for events used by <code>Component</code> classes.
 */
public static final long COMPONENT_EVENT_MASK = 1;
public static final long CONTAINER_EVENT_MASK = 2;
public static final long FOCUS_EVENT_MASK = 4;
public static final long KEY_EVENT_MASK = 8;
public static final long MOUSE_EVENT_MASK = 16;
public static final long MOUSE_MOTION_EVENT_MASK = 32;
public static final long WINDOW_EVENT_MASK = 64;
public static final long ACTION_EVENT_MASK = 128;
public static final long ADJUSTMENT_EVENT_MASK = 256;
public static final long ITEM_EVENT_MASK = 512;
public static final long TEXT_EVENT_MASK = 1024;
public static final long PAINT_EVENT_MASK = 8192;
public static final long INVOCATION_EVENT_MASK = 16384;
public static final long INPUT_METHOD_EVENT_MASK = 2048;
public static final long MOUSE_WHEEL_EVENT_MASK = 131072;
public static final long WINDOW_FOCUS_EVENT_MASK = 524288;

transient boolean focusManagerIsDispatching = false;
transient boolean postProcessed = false;

// We can't have a switch indexed on a long
// so we have to provide the log2 value for each EVENT_MASK
// and use this int as the switch parameter...
static final int COMPONENT_LOG2 = 0;
static final int CONTAINER_LOG2 = 1;
static final int FOCUS_LOG2 = 2;
static final int KEY_LOG2 = 3;
static final int MOUSE_LOG2 = 4;
static final int MOUSE_MOTION_LOG2 = 5;
static final int WINDOW_LOG2 = 6;
static final int ACTION_LOG2 = 7;
static final int ADJUSTMENT_LOG2 = 8;
static final int ITEM_LOG2 = 9;
static final int TEXT_LOG2 = 10;
static final int INPUT_METHOD_LOG2 = 11;
static final int PAINT_LOG2 = 13;
static final int INVOCATION_LOG2 = 14;
static final int MOUSE_WHEEL_LOG2 = 17;
static final int WINDOW_FOCUS_LOG2 = 19;
static final int MAX_LOG2 = 19;

/**
 * Maximum event id mask used in this class.
 */
public static final int RESERVED_ID_MAX = 1999;

/**
 * Constructs a new <code>AWTEvent</code> from
 * a 1.0 event.
 *
 * @author		IBM
 * @version 	initial
 *
 * @param	event	the 1.0 event to construct from
 */
public AWTEvent(Event event) {
	this(event.target, event.id);
}

/**
 * Constructs a new <code>AWTEvent</code> with the
 * specified source and id.
 *
 * @author		IBM
 * @version 	initial
 *
 * @param	source	the source object for the receiver
 * @param	id		the mask id for the receiver
 */
public AWTEvent(Object source, int id) {
	super(source);

	this.id = id;

	switch (id) {
		case ActionEvent.ACTION_PERFORMED:
		case AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED:
		case ItemEvent.ITEM_STATE_CHANGED:
		case TextEvent.TEXT_VALUE_CHANGED:
			consumed = true;
			break;
		default:
			consumed = false;
	}

	this.eventMaskLog2 = AWTEvent.getEventMaskLog2From(this);
}
int getEventMaskLog2() {
	return eventMaskLog2;
}
/**
 * Returns the Event mask (in Log 2) corresponding to the AWTEvent.
 * Return -1 if unknown...
 *
 * @param event	- AWTEvent analyzed
 *
 * @return int -							Long			Log2
 * 					COMPONENT_EVENT_MASK 	= 1				0
 * 					CONTAINER_EVENT_MASK 	= 2				1
 * 					FOCUS_EVENT_MASK 		= 4				2
 * 					KEY_EVENT_MASK 			= 8				3
 * 					MOUSE_EVENT_MASK 		= 16			4
 * 					MOUSE_MOTION_EVENT_MASK = 32			5
 * 					WINDOW_EVENT_MASK 		= 64			6
 * 					ACTION_EVENT_MASK 		= 128			7
 * 					ADJUSTMENT_EVENT_MASK 	= 256			8
 * 					ITEM_EVENT_MASK 		= 512			9
 * 					TEXT_EVENT_MASK 		= 1024			10
 *[IF PPRO11|PBPRO11]
 *					INPUT_METHOD_EVENT_MASK = 2048			11
 *					PAINT_EVENT_MASK 		= 8192			13
 *					INVOCATION_EVENT_MASK 	= 16384			14
 *					MOUSE_WHEEL_EVENT_MASK 	= 131072		17
 *					WINDOW_FOCUS_EVENT_MASK = 524288		19
 *[ENDIF]
 *
 */
static int getEventMaskLog2From(AWTEvent event) {
	int id = event.getID();
	if (	(id >= ActionEvent.ACTION_FIRST) &&
			(id <= ActionEvent.ACTION_LAST)) return ACTION_LOG2;

	if (	(id >= AdjustmentEvent.ADJUSTMENT_FIRST) &&
			(id <= AdjustmentEvent.ADJUSTMENT_LAST)) return ADJUSTMENT_LOG2;

	if (	(id >= ComponentEvent.COMPONENT_FIRST) &&
			(id <= ComponentEvent.COMPONENT_LAST)) return COMPONENT_LOG2;

	if (	(id >= ContainerEvent.CONTAINER_FIRST) &&
			(id <= ContainerEvent.CONTAINER_LAST)) return CONTAINER_LOG2;

	if (	(id >= FocusEvent.FOCUS_FIRST) &&
			(id <= FocusEvent.FOCUS_LAST)) return FOCUS_LOG2;

	if (	(id >= KeyEvent.KEY_FIRST) &&
			(id <= KeyEvent.KEY_LAST)) return KEY_LOG2;

	if (	(id >= MouseEvent.MOUSE_FIRST) &&
			(id <= MouseEvent.MOUSE_LAST)) {

		if (	(id == MouseEvent.MOUSE_DRAGGED) ||
				(id == MouseEvent.MOUSE_MOVED)) return MOUSE_MOTION_LOG2;
		if (	(id == MouseEvent.MOUSE_WHEEL)) return MOUSE_WHEEL_LOG2;
		return MOUSE_LOG2;
	}
	if (	(id >= PaintEvent.PAINT_FIRST) &&
			(id <= PaintEvent.PAINT_LAST)) return PAINT_LOG2;

	if (	(id >= WindowEvent.WINDOW_FIRST) &&
			(id <= WindowEvent.WINDOW_LAST)) {
		if (	(id == WindowEvent.WINDOW_GAINED_FOCUS) ||
				(id == WindowEvent.WINDOW_LOST_FOCUS)) return WINDOW_FOCUS_LOG2;
		return WINDOW_LOG2;
	}

	if (	(id >= InputMethodEvent.INPUT_METHOD_FIRST) &&
			(id <= InputMethodEvent.INPUT_METHOD_LAST)) return INPUT_METHOD_LOG2;

	if (	(id >= InvocationEvent.INVOCATION_FIRST) &&
			(id <= InvocationEvent.INVOCATION_LAST)) return INVOCATION_LOG2;

	if (	(id >= ItemEvent.ITEM_FIRST) &&
			(id <= ItemEvent.ITEM_LAST)) return ITEM_LOG2;

	if (	(id >= TextEvent.TEXT_FIRST) &&
			(id <= TextEvent.TEXT_LAST)) return TEXT_LOG2;

	return -1;
}
/**
 * Indicate that this event should be consumed.  This means it should
 * never be processed by the receiver's source object.
 *
 * @author		IBM
 * @version 	initial
 *
 */
protected void consume() {
	switch (getID()) {
		case KeyEvent.KEY_PRESSED :
		case KeyEvent.KEY_RELEASED :
		case MouseEvent.MOUSE_PRESSED :
		case MouseEvent.MOUSE_RELEASED :
		case MouseEvent.MOUSE_ENTERED :
		case MouseEvent.MOUSE_EXITED :
		case MouseEvent.MOUSE_MOVED :
		case MouseEvent.MOUSE_DRAGGED :
		case MouseEvent.MOUSE_WHEEL :
			this.consumed = true;
	}
}

int convertTo10Key(KeyEvent keyEvent) {
	if (keyEvent.getKeyCode() != KeyEvent.VK_UNDEFINED) {
		switch (keyEvent.getKeyCode()) {
		case KeyEvent.VK_HOME: return Event.HOME;
		case KeyEvent.VK_END:return Event.END;
		case KeyEvent.VK_PAGE_UP:return Event.PGUP;
		case KeyEvent.VK_PAGE_DOWN:return Event.PGDN;
		case KeyEvent.VK_UP:return Event.UP;
		case KeyEvent.VK_DOWN:return Event.DOWN;
		case KeyEvent.VK_LEFT:return Event.LEFT;
		case KeyEvent.VK_RIGHT:return Event.RIGHT;
		case KeyEvent.VK_F1:return Event.F1;
		case KeyEvent.VK_F2:return Event.F2;
		case KeyEvent.VK_F3:return Event.F3;
		case KeyEvent.VK_F4:return Event.F4;
		case KeyEvent.VK_F5:return Event.F5;
		case KeyEvent.VK_F6:return Event.F6;
		case KeyEvent.VK_F7:return Event.F7;
		case KeyEvent.VK_F8:return Event.F8;
		case KeyEvent.VK_F9:return Event.F9;
		case KeyEvent.VK_F10:return Event.F10;
		case KeyEvent.VK_F11:return Event.F11;
		case KeyEvent.VK_F12:return Event.F12;
		case KeyEvent.VK_PRINTSCREEN:return Event.PRINT_SCREEN;
		case KeyEvent.VK_SCROLL_LOCK:return Event.SCROLL_LOCK;
		case KeyEvent.VK_CAPS_LOCK:return Event.CAPS_LOCK;
		case KeyEvent.VK_NUM_LOCK:return Event.NUM_LOCK;
		case KeyEvent.VK_PAUSE:return Event.PAUSE;
		case KeyEvent.VK_INSERT:return Event.INSERT;
		}
	}
	return keyEvent.getKeyChar();
}

//-------------------------------------------------------------------
//
//-------------------------------------------------------------------
Event convertTo10() {
	Object source = getSource();
	int id = getID();

	switch (id) {
		case KeyEvent.KEY_PRESSED :
		case KeyEvent.KEY_RELEASED :
			KeyEvent keyE = (KeyEvent) this;
			if (keyE.isActionKey()) {
				if (id == KeyEvent.KEY_PRESSED)
					id = Event.KEY_ACTION;
				else
					id = Event.KEY_ACTION_RELEASE;
			}
			return new Event(source, keyE.getWhen(), id, 0, 0, convertTo10Key(keyE), keyE.getModifiers(), null);

		case MouseEvent.MOUSE_PRESSED :
		case MouseEvent.MOUSE_RELEASED :
		case MouseEvent.MOUSE_ENTERED :
		case MouseEvent.MOUSE_EXITED :
		case MouseEvent.MOUSE_MOVED :
		case MouseEvent.MOUSE_DRAGGED :
		case MouseEvent.MOUSE_WHEEL :
			if (!(source instanceof Canvas || source instanceof Container))
				return null;
			MouseEvent mouseE = (MouseEvent)this;
			return new Event(source, mouseE.getWhen(), id, mouseE.getX(), mouseE.getY(), 0, mouseE.getModifiers(), null);

		case AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED:
			AdjustmentEvent adjustE = (AdjustmentEvent)this;
			switch (adjustE.getAdjustmentType()) {
				case AdjustmentEvent.UNIT_INCREMENT: id = Event.SCROLL_LINE_UP; break;
				case AdjustmentEvent.UNIT_DECREMENT: id = Event.SCROLL_LINE_DOWN; break;
				case AdjustmentEvent.TRACK: id = Event.SCROLL_ABSOLUTE; break;
				case AdjustmentEvent.BLOCK_INCREMENT: id = Event.SCROLL_PAGE_UP; break;
				case AdjustmentEvent.BLOCK_DECREMENT: id = Event.SCROLL_PAGE_DOWN; break;
				default: return null;
			}
			return new Event(source, id, new Integer(adjustE.getValue()));

		case FocusEvent.FOCUS_GAINED:
		case FocusEvent.FOCUS_LOST:
			return new Event(source, id, null);

		case ActionEvent.ACTION_PERFORMED:
			ActionEvent actionE = (ActionEvent)this;
			return new Event(source, id, actionE.getActionCommand());

		case ItemEvent.ITEM_STATE_CHANGED:
			ItemEvent itemE = (ItemEvent)this;
			id = Event.ACTION_EVENT;
			Object arg = itemE.getItem();

			if (source instanceof List) {
				switch (itemE.getStateChange()) {
					case ItemEvent.SELECTED :
						id = Event.LIST_SELECT;
						break;
					case ItemEvent.DESELECTED :
						id = Event.LIST_DESELECT;
						break;
					default :
						return null;
				}
			} else if (source instanceof Checkbox) {
				arg = new Boolean(itemE.getStateChange() == ItemEvent.SELECTED);
			} else if (source instanceof Choice) {
				arg = ((Choice) source).getItem(((Choice) source).getSelectedIndex());
			}
			return new Event(source, id, arg);

		case WindowEvent.WINDOW_CLOSING:
		case WindowEvent.WINDOW_ICONIFIED:
		case WindowEvent.WINDOW_DEICONIFIED:
			return new Event(source, id, null);

		case ComponentEvent.COMPONENT_MOVED:
			if (!(source instanceof Window))
				return null;
			return new Event(source, Event.WINDOW_MOVED, null);

		default:
			return null;
	}
}

/**
 * Answer the id for the receiver.
 *
 * @author		IBM
 * @version 	initial
 *
 * @return	an int	the id for the receiver.
 */
public int getID() {
	return id;
}

/**
 * Answer whether this event should be consumed.  This means it should
 * never be processed.
 *
 * @author		IBM
 * @version 	initial
 *
 * @return	a boolean	<code>true</code> if this event should be consumed
 *						<code>false</code> otherwise.
 */
protected boolean isConsumed() {
	return consumed;
}

/**
 * Answer a string with the relevant information describing the state
 * of the component.  This will not be <code>null</code>.
 *
 * @author		IBM
 * @version		initial
 *
 * @return	a <code>String</code> 	the param string.
 */
public String paramString() {
	return ""; //$NON-NLS-1$
}

/**
 * Modify the event source
 */
void modifyEventSource(Object source) {
	this.source = source;
}

void setSource(Object source) {
	modifyEventSource(source);
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver's relevant information.
 *
 * @author		IBM
 * @version		initial
 *
 * @return	a String which is a printable representation for the receiver
 */
public String toString() {
	String name = null;
	if (getSource() instanceof Component) {
		name = ((Component)getSource()).getName();
	} else if (getSource() instanceof MenuComponent) {
		name = ((MenuComponent)getSource()).getName();
	}
    if (name == null) {
	     name = getSource().toString();
    }
	return getClass().getName() + "[" + paramString() + "] on " + name;
}

}
