package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2005  All Rights Reserved
 */

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.Hashtable;
import java.util.Set;
import java.util.StringTokenizer;

public class AWTKeyStroke implements Serializable {

	private static final long serialVersionUID = -6430539691155161871L;

	static Hashtable cachedKeyStrokes = new Hashtable();
	static Class factoryClass;
	static Hashtable modifierTable;

	char keyChar;
	int keyCode;
	int modifiers;
	boolean onKeyRelease;

	AWTKeyStroke() {
		this(KeyEvent.CHAR_UNDEFINED, KeyEvent.VK_UNDEFINED, 0, false);
	}

	AWTKeyStroke(char keyChar, int keyCode, int modifiers, boolean onKeyRelease) {
		this.keyChar = keyChar;
		this.keyCode = keyCode;
		this.modifiers = modifiers;
		this.onKeyRelease = onKeyRelease;
	}

	private static void registerSubclass(Class subclass) {
		if (subclass == null) throw new IllegalArgumentException("subclass cannot be null"); //$NON-NLS-1$
		if (!validSubclass(subclass)) throw new ClassCastException("subclass is not derived from AWTKeyStroke"); //$NON-NLS-1$
	    cachedKeyStrokes.clear();
		factoryClass = subclass;
	}

	private static boolean validSubclass(Class subclass) {
		Class keyStrokeClass = null;
		try {
		    keyStrokeClass = Class.forName("java.awt.AWTKeyStroke"); //$NON-NLS-1$
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}

		// Can we instead use AWTKeyStroke.class reference instead
		// of the Class.forName() code?
	    return keyStrokeClass.isAssignableFrom(subclass);
	}

	public static AWTKeyStroke getAWTKeyStroke(char keyChar) {
		return getAWTKeyStroke(new Character(keyChar), 0);
	}

	public static AWTKeyStroke getAWTKeyStroke(Character keyChar, int modifiers) {
		if (keyChar == null) throw new IllegalArgumentException("keyChar cannot be null"); //$NON-NLS-1$
		return getAWTKeyStroke(0, keyChar.charValue(), modifiers, false);
	}

	public static AWTKeyStroke getAWTKeyStroke(int keyCode, int modifiers, boolean released) {
		return getAWTKeyStroke(keyCode, KeyEvent.CHAR_UNDEFINED, modifiers, released);
	}

	public static AWTKeyStroke getAWTKeyStroke(int keyCode, int modifiers) {
		return getAWTKeyStroke(keyCode, modifiers, false);
	}

	public static AWTKeyStroke getAWTKeyStrokeForEvent(KeyEvent keyEvent) {
	    boolean released = false;
	    if (keyEvent.getID() == KeyEvent.KEY_RELEASED) released = true;
		return getAWTKeyStroke(keyEvent.getKeyCode(), keyEvent.getKeyChar(), keyEvent.getModifiers(), released);
	}

	private static AWTKeyStroke getAWTKeyStroke(int keyCode, char keyChar, int modifiers, boolean released) {
	    AWTKeyStroke keyStroke = getKeyStrokeInstance();
	    if (keyCode != KeyEvent.VK_UNDEFINED) {
			keyStroke.keyChar = KeyEvent.CHAR_UNDEFINED;
	    } else {
			keyStroke.keyChar = keyChar;
	    }
		keyStroke.keyCode = keyCode;
		keyStroke.modifiers = adjustModifiers(modifiers);
		keyStroke.onKeyRelease = released;
		AWTKeyStroke cachedKeyStroke = (AWTKeyStroke)cachedKeyStrokes.get(keyStroke);
		if (cachedKeyStroke != null) return cachedKeyStroke;
		cachedKeyStrokes.put(keyStroke, keyStroke);
		return keyStroke;
	}

	public static AWTKeyStroke getAWTKeyStroke(String keyString) {
		if (keyString == null) {
			throw new IllegalArgumentException("String cannot be null"); //$NON-NLS-1$
		}
		StringTokenizer tokenizer = new StringTokenizer(keyString);
		if (!tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$

		Hashtable modifierTags = getModifierTable();
		Set tags = modifierTags.keySet();
		boolean moreModifiers = true;
		String token = null;
		int modifierKeys = 0;
		while (moreModifiers && tokenizer.hasMoreElements()) {
			token = tokenizer.nextToken();
			if (tags.contains(token)) {
				modifierKeys = modifierKeys | ((Integer)modifierTable.get(token)).intValue();
			} else {
				moreModifiers = false;
			}
		}

		if (token.equals("typed")) { //$NON-NLS-1$
			if (!tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
			token = tokenizer.nextToken();
			if (tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
			return getAWTKeyStroke(new Character(token.charAt(0)), modifierKeys);
		}
		boolean released = false;
		if (token.equals("released")) { //$NON-NLS-1$
			released = true;
			if (!tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
			token = tokenizer.nextToken();
		} else if (token.equals("pressed")) { //$NON-NLS-1$
			if (!tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
			token = tokenizer.nextToken();
		}

		if (tokenizer.hasMoreElements()) throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
		int keyCode = getVKValue(token);
		return getAWTKeyStroke(keyCode, modifierKeys, released);
	}

	public final boolean equals(Object object) {
		if (object == this) return true;
		AWTKeyStroke keyStroke;
		if (object instanceof AWTKeyStroke) keyStroke = (AWTKeyStroke)object;
		else return false;
		if (this.keyCode != keyStroke.keyCode) return false;
		if (this.keyChar != keyStroke.keyChar) return false;
		if (this.modifiers != keyStroke.modifiers) return false;
		if (this.onKeyRelease != keyStroke.onKeyRelease) return false;
		return true;
	}

	static private int adjustModifiers(int modifiers) {
	    // account for the old and new modifiers
	    if ((modifiers & InputEvent.ALT_MASK) != 0) modifiers |= InputEvent.ALT_DOWN_MASK;
	    else if ((modifiers & InputEvent.ALT_DOWN_MASK) != 0) modifiers |= InputEvent.ALT_MASK;
	    if ((modifiers & InputEvent.SHIFT_MASK) != 0) modifiers |= InputEvent.SHIFT_DOWN_MASK;
	    else if ((modifiers & InputEvent.SHIFT_DOWN_MASK) != 0) modifiers |= InputEvent.SHIFT_MASK;
	    if ((modifiers & InputEvent.CTRL_MASK) != 0) modifiers |= InputEvent.CTRL_DOWN_MASK;
	    else if ((modifiers & InputEvent.CTRL_DOWN_MASK) != 0) modifiers |= InputEvent.CTRL_MASK;
	    if ((modifiers & InputEvent.META_MASK) != 0) modifiers |= InputEvent.META_DOWN_MASK;
	    else if ((modifiers & InputEvent.META_DOWN_MASK) != 0) modifiers |= InputEvent.META_MASK;
	    if ((modifiers & InputEvent.ALT_GRAPH_MASK) != 0) modifiers |= InputEvent.ALT_GRAPH_DOWN_MASK;
	    else if ((modifiers & InputEvent.ALT_GRAPH_DOWN_MASK) != 0) modifiers |= InputEvent.ALT_GRAPH_MASK;
	    if ((modifiers & InputEvent.BUTTON1_MASK) != 0) {
		    modifiers = modifiers ^ InputEvent.BUTTON1_MASK;
	        modifiers |= InputEvent.BUTTON1_DOWN_MASK;
	    }
	    return modifiers;
	}
	public final char getKeyChar() {
		return keyChar;
	}

	public final int getKeyCode() {
		return keyCode;
	}

	static private AWTKeyStroke getKeyStrokeInstance() {
	    // return a factory instance
		if (factoryClass == null) {
			try {
				factoryClass = Class.forName("java.awt.AWTKeyStroke"); //$NON-NLS-1$
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			}
		}
		AWTKeyStroke keyStroke = null;
		try {
			keyStroke = (AWTKeyStroke)factoryClass.newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return keyStroke;
	}
	public final int getModifiers() {
		return modifiers;
	}

	static private Hashtable getModifierTable() {
		if (modifierTable != null) return modifierTable;
		modifierTable = new Hashtable();
		// TODO add externalized string flags
		modifierTable.put("shift", new Integer(InputEvent.SHIFT_MASK | InputEvent.SHIFT_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("control", new Integer(InputEvent.CTRL_MASK | InputEvent.CTRL_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("ctrl", new Integer(InputEvent.CTRL_MASK | InputEvent.CTRL_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("meta", new Integer(InputEvent.META_MASK | InputEvent.META_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("alt", new Integer(InputEvent.ALT_MASK | InputEvent.ALT_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("altGraph", new Integer(InputEvent.ALT_GRAPH_MASK | InputEvent.ALT_GRAPH_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("button1", new Integer(InputEvent.BUTTON1_DOWN_MASK | InputEvent.BUTTON1_MASK)); //$NON-NLS-1$
		modifierTable.put("button2", new Integer(InputEvent.BUTTON2_DOWN_MASK)); //$NON-NLS-1$
		modifierTable.put("button3", new Integer(InputEvent.BUTTON3_DOWN_MASK)); //$NON-NLS-1$
		return modifierTable;

	}
	static private int getVKValue(String vkString) {
		Class keyEventClass = null;
		try {
			keyEventClass = Class.forName("java.awt.event.KeyEvent"); //$NON-NLS-1$
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		Field vkField = null;
		try {
			vkField = keyEventClass.getField("VK_" + vkString); //$NON-NLS-1$
		} catch (NoSuchFieldException e) {
			throw new IllegalArgumentException("String formatted incorrectly"); //$NON-NLS-1$
		}
		int value = -1;
		try {
			value = vkField.getInt(null);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return value;
	}

	public final int getKeyEventType() {
	    if (getKeyCode() == KeyEvent.VK_UNDEFINED) return KeyEvent.KEY_TYPED;
		if (isOnKeyRelease()) return KeyEvent.KEY_RELEASED;
		else return KeyEvent.KEY_PRESSED;
	}

	public int hashCode() {
		if (keyCode != 0) return keyCode & modifiers;
		else return keyChar & modifiers;
	}

	public final boolean isOnKeyRelease() {
		return onKeyRelease;
	}

	public String toString() {
	    // remove button1 mask for printing
	    int mods = modifiers;
	    if ((mods & InputEvent.BUTTON1_MASK) != 0) mods = mods ^ InputEvent.BUTTON1_MASK;
		StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
		if (keyCode == 0) {
			buffer.append("keyChar "); //$NON-NLS-1$
			buffer.append(KeyEvent.getKeyModifiersText(mods));
			buffer.append(keyChar);
		} else {
			buffer.append("keyCode "); //$NON-NLS-1$
			buffer.append(KeyEvent.getKeyModifiersText(mods));
			buffer.append(KeyEvent.getKeyText(keyCode));
			if (isOnKeyRelease()) {
				buffer.append("-R"); //$NON-NLS-1$
			} else {
				buffer.append("-P"); //$NON-NLS-1$
			}
		}
		return buffer.toString();
	}

	private Object readResolve() throws ObjectStreamException {
		AWTKeyStroke cachedKeyStroke = (AWTKeyStroke)cachedKeyStrokes.get(this);
		if (cachedKeyStroke != null) return cachedKeyStroke;
		cachedKeyStrokes.put(this, this);
		return this;
	}

}
