package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2005  All Rights Reserved
 */

public class BasicStroke implements Stroke {

	public static final int CAP_BUTT = 0;
	public static final int CAP_ROUND = 1;
	public static final int CAP_SQUARE = 2;
	public static final int JOIN_BEVEL = 2;
	public static final int JOIN_MITER = 0;
	public static final int JOIN_ROUND = 1;

	float lineWidth;
	int endCap;
	int lineJoin;
	float miterLimit;

	/**
	 * Create a basic stroke that indicates the desired
	 * line width, end cap, line join, and miter limit
	 * values.
	 *
	 * Restriction from the Personal Profile 1.1 spec:
	 * An implementation need not support the line
	 * attributes end cap, line join, and miter width.
	 * In such a case, attempts to change these line
	 * attributes will fail silently.
	 *
	 * @param width the pen width
	 * @param cap the appearence of the end of line segments
	 * @param join the appearence of the intersection of line segments
	 * @param miterlimit the amount by which a miter join is trimmed
	 */
	public BasicStroke(float width, int cap, int join, float miterlimit) {
		setLineWidth(width);
		setEndCap(cap);
		setLineJoin(join);
		setMiterLimit(miterlimit);
	}

	private void setMiterLimit(float miterlimit) {
		if (getLineJoin() == JOIN_MITER) {
			if (miterlimit < 1) {
				throw new IllegalArgumentException("miterLimit is not valid");
			}
		}
		this.miterLimit = miterlimit;
	}

	private void setLineJoin(int join) {
		switch (join) {
			case JOIN_BEVEL:
			case JOIN_MITER:
			case JOIN_ROUND:
				break;
			default:
				throw new IllegalArgumentException("line join is not valid");
		}
		this.lineJoin = join;
	}

	private void setEndCap(int cap) {
		switch (cap) {
			case CAP_BUTT:
			case CAP_ROUND:
			case CAP_SQUARE:
				break;
			default:
				throw new IllegalArgumentException("end cap is not valid");
		}
		this.endCap = cap;
	}

	private void setLineWidth(float width) {
		if (width < 0) {
			throw new IllegalArgumentException("stroke width cannot be negative");
		}
		this.lineWidth = width;
	}

	public BasicStroke(float width, int cap, int join) {
		this(width, cap, join, 10);
	}

	public BasicStroke(float width) {
		this(width, CAP_SQUARE, JOIN_MITER);
	}

	public BasicStroke() {
		this(1.0f);
	}

	public float getLineWidth() {
		return lineWidth;
	}

	public int getEndCap() {
		return endCap;
	}

	public int getLineJoin() {
		return lineJoin;
	}

	public float getMiterLimit() {
		return miterLimit;
	}

	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + endCap;
		hash = 31 * hash + lineJoin;
		hash = 31 * hash + Float.floatToIntBits(lineWidth);
		hash = 31 * hash + Float.floatToIntBits(miterLimit);
		return hash;
	}

	public boolean equals(Object object) {
		if (object == this)
			return true;
		if (!(object instanceof BasicStroke))
			return false;
		BasicStroke stroke = (BasicStroke) object;
		return (
			endCap == stroke.endCap &&
			lineJoin == stroke.lineJoin &&
			lineWidth == stroke.lineWidth &&
			miterLimit == stroke.miterLimit);
	}

}
