package java.awt;

import java.io.*;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * <code>BorderLayout</code> is a <code>LayoutManager2</code> that
 * lays out the components in five areas: north, south, east, west
 * and center.
 *
 * @author      IBM
 * @version		initial
 *
 * @see		<code>LayoutManager</code>
 *
 */

public class BorderLayout implements LayoutManager2, Serializable {

static final long serialVersionUID = -8658291919501921765L;

// The component that occupies the center area
Component center;

// The component that occupies the east area
Component east;

// The component that occupies the north area
Component north;

// The component that occupies the south area
Component south;

// The component that occupies the west area
Component west;

// The horizontal gap between components
int hgap;

// The vertical gap between components
int vgap;

/**
 * Constants that specify the name of the component in
 * each area.
 */
public static final String CENTER = "Center"; //$NON-NLS-1$
public static final String EAST = "East"; //$NON-NLS-1$
public static final String NORTH = "North"; //$NON-NLS-1$
public static final String SOUTH = "South"; //$NON-NLS-1$
public static final String WEST = "West"; //$NON-NLS-1$

/**
 * Constructs a new <code>BorderLayout</code> with no
 * gaps.
 *
 * @author      IBM
 * @version 	initial
 */
public BorderLayout() {
	this(0, 0);
}

/**
 * Constructs a new <code>BorderLayout</code> with the
 * specified horizontal and vertical gaps.
 *
 * @author      IBM
 * @version 	initial
 *
 * @param	hgap	the horizontal gap
 * @param	vgap	the vertical gap
 */
public BorderLayout(int hgap, int vgap) {
	this.hgap = hgap;
	this.vgap = vgap;
}

/**
 * Adds the specified <code>Component</code> to the receiver
 * using the given name.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		name		the name which identifies the component
 * @param		component	the component to be added
 *
 * @see			<code>Container.add()</code>
 *
 * @deprecated	use <code>addLayoutComponent(Component, Object)</code> instead
 *
 */
public void addLayoutComponent(String name, Component component) {
	addLayoutComponent(component, name);
}

/**
 * Adds the specified <code>Component</code> to the receiver
 * using the given constraints.
 * <p>
 * The constraints for a <code>BorderLayout</code> must be
 * one of the pre-defined strings:
 * 	<code>CENTER</code>
 * 	<code>NORTH</code>
 * 	<code>SOUTH</code>
 * 	<code>EAST</code>
 * 	<code>WEST</code>
 * <p>
 * The value <code>null</code> is a valid constraints and has the same
 * effect as <code>CENTER</code>.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		component	the component to be added
 * @param		constraints	the constraints object for the given component
 *
 * @throws		IllegalArgumentException	if the constraints is invalid
 *
 * @see			<code>Container.add()</code>
 */
public void addLayoutComponent(Component component, Object constraints) {
	if (constraints == null || constraints.equals(CENTER)) center = component;
	else if (constraints.equals(EAST)) east = component;
	else if (constraints.equals(NORTH)) north = component;
	else if (constraints.equals(SOUTH)) south = component;
	else if (constraints.equals(WEST)) west = component;
	else
		throw new IllegalArgumentException();
}

/**
 * Answers the horizontal gap between components.
 *
 * @author      IBM
 * @version		initial
 *
 * @return		an <code>int</code> which is the horizontal gap
 *
 * @see			<code>setHgap(int)</code>
 */
public int getHgap() {
	return hgap;
}

/**
 * Answers the vertical gap between components.
 *
 * @author      IBM
 * @version		initial
 *
 * @return		an <code>int</code> which is the vertaical gap
 *
 * @see			<code>setVgap(int)</code>
 */
public int getVgap() {
	return vgap;
}

/**
 * Answer the preferred vertical alignment for the specified
 * <code>Container</code>.
 *
 * @author      IBM
 * @version		initial
 *
 * @return		a <code>float</code> which is the alignment value
 *
 * @see			<code>Container.getAlignmentX()</code>
 */
public float getLayoutAlignmentX(Container container) {
	return Component.CENTER_ALIGNMENT;
}

/**
 * Answer the preferred horizontal alignment for the specified
 * <code>Container</code>.
 *
 * @author      IBM
 * @version		initial
 *
 * @return		a <code>float</code> which is the alignment value
 *
 * @see			<code>Container.getAlignmentY()</code>
 */
public float getLayoutAlignmentY(Container container) {
	return Component.CENTER_ALIGNMENT;
}

/**
 * Invalidates the layout for the specified <code>Container</code>.  Any
 * cached information should be discarded.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		container	the <code>Container</code> to invalidate the layout
 *
 * @see			<code>Container.invalidate()</code>
 */
public void invalidateLayout(Container container) {
}

/**
 * Lays out the children of the specified <code>Container</code>.
 * <p>
 * The receiver first lays out the children in the <code>NORTH</code> and
 * <code>SOUTH</code> areas if they are present and visible.  Then it lays
 * out the children <code>EAST</code> and <code>WEST</code> areas if they
 * are present and visible.  And finally, it lays out the child in the
 * <code>CENTER</code>.
 * <p>
 * The children in the <code>NORTH</code> and <code>SOUTH</code> will
 * have their preferred height and the width of the container's client area.
 * <p>
 * The children in the <code>EAST</code> and <code>WEST</code> will
 * have their preferred width and the height of the container's client area
 * excluding the area occupied by the children in the <code>NORTH</code>
 * and <code>SOUTH</code>.
 * <p>
 * The child in the <code>CENTER</code> occupies the remaning area.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		container	the <code>Container</code> to be laid out
 *
 * @see			<code>Container.doLayout()</code>
 */
public void layoutContainer(Container container) {
	Dimension size = container.getSize();
	Insets insets = container.getInsets();

	int centerX = insets.left;
	int centerY = insets.top;
	int centerWidth = size.width - insets.left - insets.right;
	int centerHeight = size.height - insets.top - insets.bottom;

	if (north != null && north.isVisible()) {
		Dimension  preferredSize = north.getPreferredSize();
		centerY += preferredSize.height + vgap;
		centerHeight -= preferredSize.height + vgap;
		north.setBounds(
			insets.left,
			insets.top,
			centerWidth,
			preferredSize.height);
	}

	if (south != null && south.isVisible()) {
		Dimension preferredSize = south.getPreferredSize();
		centerHeight -= preferredSize.height + vgap;
		south.setBounds(
			insets.left,
			size.height - insets.bottom - preferredSize.height,
			centerWidth,
			preferredSize.height);
	}

	if (west != null && west.isVisible()) {
		Dimension  preferredSize = west.getPreferredSize();
		centerX += preferredSize.width + hgap;
		centerWidth -= preferredSize.width + hgap;
		west.setBounds(
			insets.left,
			centerY,
			preferredSize.width,
			centerHeight);
	}

	if (east != null && east.isVisible()) {
		Dimension  preferredSize = east.getPreferredSize();
		centerWidth -= preferredSize.width + hgap;
		east.setBounds(
			size.width - insets.right - preferredSize.width,
			centerY,
			preferredSize.width,
			centerHeight);
	}

	if (center != null) {
		center.setBounds(centerX, centerY, centerWidth, centerHeight);
	}
}

/**
 * Calculates the maximum size for the specified <code>Container</code>.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		container	the <code>Container</code> to calculate the maximum size
 *
 * @return		a <code>Dimension</code> which is the maximum size
 *				for the specified <code>Container</code>
 *
 * @see			<code>Container.getMaximumSize()</code>
 */
public Dimension maximumLayoutSize(Container container) {
	return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
}

/**
 * Calculates the minimum size for the specified <code>Container</code>.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		container	the <code>Container</code> to calculate the minimum size
 *
 * @return		a <code>Dimension</code> which is the minimum size
 *				for the specified <code>Container</code>
 *
 * @see			<code>Container.minimumSize()</code>
 */
public Dimension minimumLayoutSize(Container container) {
	Dimension size;
	int width = 0;
	int height = 0;
	if (west != null && west.isVisible()) {
		size = west.getMinimumSize();
		width = size.width + hgap;
		height = size.height;
	}
	if (east != null && east.isVisible()) {
		size = east.getMinimumSize();
		width += size.width + hgap;
		height = Math.max(height, size.height);
	}
	if (center != null && center.isVisible()) {
		size = center.getMinimumSize();
		width += size.width;
		height = Math.max(height, size.height);
	}
	if (north != null && north.isVisible()) {
		size = north.getMinimumSize();
		width = Math.max(width, size.width);
		height += size.height + vgap;
	}
	if (south != null && south.isVisible()) {
		size = south.getMinimumSize();
		width = Math.max(width, size.width);
		height += size.height + vgap;
	}
	Insets insets = container.getInsets();
	return new Dimension(
		width + insets.left + insets.right,
		height + insets.top + insets.bottom);
}

/**
 * Calculates the preferred size for the specified <code>Container</code>.
 * The preferred size is the size with which the container can fully display
 * itself and its children.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		container	the <code>Container</code> to calculate the preferred size
 *
 * @return		a <code>Dimension</code> which is the preferred size
 *				for the specified <code>Container</code>
 *
 * @see			<code>Container.preferredSize()</code>
 */
public Dimension preferredLayoutSize(Container container) {
	Dimension size;
	int width = 0;
	int height = 0;
	if (west != null && west.isVisible()) {
		size = west.getPreferredSize();
		width = size.width + hgap;
		height = size.height;
	}
	if (east != null && east.isVisible()) {
		size = east.getPreferredSize();
		width += size.width + hgap;
		height = Math.max(height, size.height);
	}
	if (center != null && center.isVisible()) {
		size = center.getPreferredSize();
		width += size.width;
		height = Math.max(height, size.height);
	}
	if (north != null && north.isVisible()) {
		size = north.getPreferredSize();
		width = Math.max(width, size.width);
		height += size.height + vgap;
	}
	if (south != null && south.isVisible()) {
		size = south.getPreferredSize();
		width = Math.max(width, size.width);
		height += size.height + vgap;
	}
	Insets insets = container.getInsets();
	return new Dimension(
		width + insets.left + insets.right,
		height + insets.top + insets.bottom);
}

/**
 * Removes the specified <code>Component</code> from the receiver.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		component	the component to be removed
 *
 * @see			<code>Container.remove()</code>
 */
public void removeLayoutComponent(Component component) {
	if (component == center) center = null;
	else if (component == east) east = null;
	else if (component == north) north = null;
	else if (component == south) south = null;
	else if (component == west) west = null;
}

/**
 * Sets the horizontal gap between components.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		hgap	the desired horizontal gap
 *
 * @see			<code>getHgap()</code>
 */
public void setHgap(int hgap) {
	this.hgap = hgap;
}

/**
 * Sets the vertical gap between components.
 *
 * @author      IBM
 * @version		initial
 *
 * @param		vgap	the desired vertical gap
 *
 * @see			<code>getHgap()</code>
 */
public void setVgap(int vgap) {
	this.vgap = vgap;
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver's relevant information.
 *
 * @author      IBM
 * @version		initial
 *
 * @return	a java.lang.String which is a printable representation for the receiver
 */
public String toString() {
	return getClass().getName() + "[hgap=" + getHgap() + ",vgap=" + getVgap() + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
}

}
