package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.io.*;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Button</code> is graphical user interface that
 * displays a text and fires an action when pressed and
 * released.
 *
 * @author      IBM
 * @version		initial
 */
public class Button extends Component {

	static final long serialVersionUID = -8774683716313001058L;

	String label;

	//	The action command of the receiver
	String actionCommand;

	//	The action listener of the receiver
	transient ActionListener actionListener;

	private int buttonSerializedDataVersion;

	/**
	 * Constructs a new <code>Button</code> with a <code>null</code> text.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Button()
	 	throws HeadlessException
	{
		this(null);
	}

	/**
	 * Constructs a new <code>Button</code> with the specified text.
	 *
	 * @param	text	the text displayed by the receiver
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Button(String text)
	 	throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		if (null == text) text = ""; //$NON-NLS-1$
		this.label = text;
	}

	int _createPeer(Container hwParent) {
		return OS.PushButton_New(hwParent.peer, 0);
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	void _setupPeer() {
		super._setupPeer();

		_setPeerLabel(label);
	}

	void buttonCallback() {
		ActionEvent event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, getActionCommand(),
			System.currentTimeMillis(),
		0);
		postEvent(event);
	}

	String classNonlocalizedName() {
		return "button"; //$NON-NLS-1$
	}

	/**
	 * Answer the <code>String</code> that is the text
	 * displayed by the receiver.  Answer <code>null</code> if
	 * no text is specified.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> which is text for the receiver
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Set the <code>String</code> that is the text
	 * diplayed by the receiver.  Use <code>null</code> if
	 * no label is desired.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	text	the text to be displayed by the receiver
	 */
	public void setLabel(String text) {
		if (null == text) text = ""; //$NON-NLS-1$
		String oldLabel = label;
		label = text;
		if (oldLabel.equals(label)) return;

		acquirePeerReadLock();
		try {
			if (peer != 0) _setPeerLabel(label);
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	/*
	 * Updates the native label to match the current label.
	 * This method will always be called with a lock around
	 * the peer
	 */
	void _setPeerLabel(String newLabel) {
		OS.Button_SetLabel(peer, newLabel == null ? "" : newLabel); //$NON-NLS-1$
	}

	/**
	 * Set the <code>String</code> representing the
	 * name of the action command for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	actionCommand	the command name
	 */
	public void setActionCommand(String actionCommand) {
		this.actionCommand = actionCommand;
	}

	/**
	 * Answer the <code>String</code> representing the
	 * name of the action command for the receiver. If
	 * the receiver's action command is not set, the
	 * receiver's label is returned.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> the command name
	 */
	public String getActionCommand() {
//		synchronized (this) {
			return (actionCommand != null) ? actionCommand : getLabel();
//		}
	}

	public ActionListener[] getActionListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(actionListener, ActionListener.class);
		ActionListener[] resultArray = new ActionListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	/**
	 * Adds an <code>ActionListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be added.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void addActionListener(ActionListener actionListener) {
		addEventListener(actionListener);
		this.actionListener = AWTEventMulticaster.add(this.actionListener, actionListener);
	}

	/**
	 * Removes an <code>ActionListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be removed.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void removeActionListener(ActionListener actionListener) {
		this.actionListener = AWTEventMulticaster.remove(this.actionListener, actionListener);
	}

	/**
	* Called when an <code>AWTEvent</code> occurs on the receiver.
	* <p>
	* Events must be enabled in order for a component to receive them.
	* Events can be enabled using <code>enableEvents(long)</code> or by
	* adding listeners for that event.
	* <p>
	* If a component is unable to get sufficient behaviour by adding
	* listeners then it must override this method to do custom event handling.
	* When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	* <p>
	* The default event handled is <code>ActionEvent</code>, the receiver
	* will forward calls to the default event by calling
	* <code>processActionEvent(ActionEvent)</code>.
	*
	* @author      IBM
	* @version		initial
	*
	* @param	awtEvent	the event to be processed
	*/
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ActionEvent) {
			processActionEvent((ActionEvent) awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ActionEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processActionEvent(ActionEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	actionEvent	the event to be processed
	 *
	 * @see		ActionEvent
	 */
	protected void processActionEvent(ActionEvent actionEvent) {
		if (actionListener == null)
			return;

		if ((actionEvent == null) ||
			(actionEvent.getID() == ActionEvent.ACTION_PERFORMED)) {
			actionListener.actionPerformed(actionEvent);
		}
	}

	/**
	 * @see java.awt.Component#isOpaque()
	 */
	public boolean isOpaque() {
		return true;
	}

	boolean isFocusTraversableImpl() {
		return true;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ActionEvent)
			return ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) || actionListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the menu component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() +
			",label=" + getLabel(); //$NON-NLS-1$
	}

	public boolean isLightweight() {
		return false;
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, actionListenerK, actionListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(actionListenerK))
				addActionListener((ActionListener)listener);
		}
	}

}
