package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.*;
import java.util.*;

/**
 * <code>CardLayout</code> is a <code>LayoutManager2</code> that
 * lays out one component at a time.  The displayed component
 * occupies the whole area of its container.
 */
public class CardLayout implements LayoutManager2, Serializable {

	static final long serialVersionUID = -4328196481005934313L;

	// following names are required for serialization
	int       hgap;
	int       vgap;
	Hashtable tab;
	Vector    vector;
	int       currentCard = -1;

	/**
	 * Constructs a new <code>CardLayout</code> with no gaps.
	 * <p> The <code>CardLayout</code> gaps are space left between
	 * the container edges and the current displayed child.
	 */
	public CardLayout() {
		this(0, 0);
	}

	/**
	 * Constructs a new <code>CardLayout</code> with the
	 * specified horizontal and vertical gaps.
	 * <p>
	 * The <code>CardLayout</code> gaps are space left between
	 * the container edges and the current displayed child.
	 *
	 * @param	hgap	the horizontal gap
	 * @param	vgap	the vertical gap
	 */
	public CardLayout(int hgap, int vgap) {
		this.hgap = hgap;
		this.vgap = vgap;
		this.tab = new Hashtable();
	}

	/**
	 * Answers the horizontal gap between components.
	 *
	 * @return		an <code>int</code> which is the horizontal gap
	 */
	public int getHgap() {
		return hgap;
	}

	/**
	 * Sets the horizontal gap between components.
	 *
	 * @param		hgap	the desired horizontal gap
	 */
	public void setHgap(int hgap) {
		this.hgap = hgap;
	}

	/**
	 * Answers the vertical gap between components.
	 *
	 * @return		an <code>int</code> which is the vertaical gap
	 */
	public int getVgap() {
		return vgap;
	}

	/**
	 * Sets the vertical gap between components.
	 *
	 * @param		vgap	the desired vertical gap
	 */
	public void setVgap(int vgap) {
		this.vgap = vgap;
	}

	/**
	 * Adds the specified <code>Component</code> to the receiver
	 * using the given constraints.
	 * <p>
	 * The constraints for a <code>CardLayout</code> must be
	 * a string or <code>null</code>.
	 *
	 * @param		component	the component to be added
	 * @param		constraints	the constraints object for the given component
	 *
	 * @throws		IllegalArgumentException	if the constraints is invalid
	 */
	public void addLayoutComponent(Component component, Object constraints) {
		if (!(constraints instanceof String)) {
			throw new IllegalArgumentException();
		}

		component.setVisible(tab.size() == 0); // Bug 98331
		tab.put(constraints, component);
	}

	/**
	 * Adds the specified <code>Component</code> to the receiver
	 * using the given name.
	 *
	 * @param		name		the name which identifies the component
	 * @param		component	the component to be added
	 *
	 * @deprecated	use <code>addLayoutComponent(Component, Object)</code> instead
	 */
	public void addLayoutComponent(String name, Component component) {
		addLayoutComponent(component, name);
	}

	/**
	 * Removes the specified <code>Component</code> from the receiver.
	 *
	 * @param		component	the component to be removed
	 */
	public void removeLayoutComponent(Component component) {

		// If the current component being removed is visible,
		// then make the next component visible first.
		if (component.isVisible()) {
			Container parent = component.getParent();
			if (parent != null) {
				next(parent);
			}
		}

		Enumeration keys = tab.keys();
		while (keys.hasMoreElements()) {
			Object key = keys.nextElement();
			if (tab.get(key) == component) {
				tab.remove(key);
				break;
			}
		}
	}

	/**
	 * Calculates the preferred size for the specified <code>Container</code>.
	 * The preferred size is the size with which the container can fully display
	 * itself and its children.
	 * <p>
	 * The preferred size of a container laid out using a <code>CardLayout</code>
	 * is the largest preferred size of each individual child.
	 * <p>
	 * The container insets is included.
	 *
	 * @param		container	the <code>Container</code> to calculate the preferred size
	 *
	 * @return		a <code>Dimension</code> which is the preferred size
	 *				for the specified <code>Container</code>
	 */
	public Dimension preferredLayoutSize(Container container) {
		int width       = 0;
		int height      = 0;
		int ncomponents = container.getComponentCount();

		for (int i=0; i<ncomponents; i++) {
			Dimension size = container.getComponent(i).getPreferredSize();

			width  = Math.max(width,  size.width);
			height = Math.max(height, size.height);
		}

		Insets insets = container.getInsets();

		return new Dimension(
			width  + insets.left + insets.right  + (2 * hgap),
			height + insets.top  + insets.bottom + (2 * vgap)
			);
	}

	/**
	 * Calculates the minimum size for the specified <code>Container</code>.
	 * <p>
	 * The minimum size of a container laid out using a <code>CardLayout</code>
	 * is the largest minimum size of each individual child.
	 * <p>
	 * The container insets is included.
	 *
	 * @param		container	the <code>Container</code> to calculate the minimum size
	 *
	 * @return		a <code>Dimension</code> which is the minimum size
	 *				for the specified <code>Container</code>
	 */
	public Dimension minimumLayoutSize(Container container) {
		int width       = 0;
		int height      = 0;
		int ncomponents = container.getComponentCount();

		for (int i=0; i<ncomponents; i++) {
			Dimension size = container.getComponent(i).getMinimumSize();

			width  = Math.max(width,  size.width);
			height = Math.max(height, size.height);
		}

		Insets insets = container.getInsets();

		return new Dimension(
			width  + insets.left + insets.right  + (2 * hgap),
			height + insets.top  + insets.bottom + (2 * vgap));
	}

	/**
	 * Calculates the maximum size for the specified <code>Container</code>.
	 *
	 * @param		container	the <code>Container</code> to calculate the maximum size
	 *
	 * @return		a <code>Dimension</code> which is the maximum size
	 *				for the specified <code>Container</code>
	 */
	public Dimension maximumLayoutSize(Container container) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	/**
	 * Answer the preferred vertical alignment for the specified
	 * <code>Container</code>.
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getLayoutAlignmentX(Container container) {
		return Component.CENTER_ALIGNMENT;
	}

	/**
	 * Answer the preferred horizontal alignment for the specified
	 * <code>Container</code>.
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getLayoutAlignmentY(Container container) {
		return Component.CENTER_ALIGNMENT;
	}

	/**
	 * Invalidates the layout for the specified <code>Container</code>.  Any
	 * cached information should be discarded.
	 *
	 * @param		container	the <code>Container</code> to invalidate the layout
	 */
	public void invalidateLayout(Container container) {
	}

	/**
	 * Lays out the children of the specified <code>Container</code>.
	 * <p>
	 * The <code>CardLayout</code> just updates the bounds of the
	 * current displayed child to be the size of the specified
	 * container excluding the gaps.
	 *
	 * @param		container	the <code>Container</code> to be laid out
	 */
	public void layoutContainer(Container container) {
		int ncomponents = container.getComponentCount();

		if (ncomponents <= 0) return;

		boolean oneIsVisible = false;

		for (int i=0; i<ncomponents; i++) {
			Component component = container.getComponent(i);
			if (component.isVisible()) {
				oneIsVisible = true;
			}

			layoutComponent(container, component, false);
		}

		if (!oneIsVisible) {
			container.getComponent(0).setVisible(true);
		}
	}

	/**
	 * Makes the first component in the specified <code>Container</code>
	 * the current displayed child.
	 *
	 * @param		container	the container
	 */
	public void first(Container container) {
		int ncomponents = container.getComponentCount();

		if (ncomponents <= 0) return;

		for (int i=1; i<ncomponents; i++) {
			Component component = container.getComponent(i);
			component.setVisible(false);
		}
		show(container, container.getComponent(0), true);
	}

	/**
	 * Makes the next component in the specified <code>Container</code>
	 * the current displayed child.
	 * <p>
	 * If the current displayed component is the last child in the
	 * container, the first child will be the current displayed
	 * child.
	 *
	 * @param		container	the container
	 */
	public void next(Container container) {
		int ncomponents = container.getComponentCount();
		if (ncomponents <= 0) return;

		boolean found = false;
		for (int i=0; i<ncomponents; i++) {
			Component component = container.getComponent(i);

			if (component.isVisible()) {
				component.setVisible(false);

				if (!found) {
					found = true;
					show(container, container.getComponent(i == ncomponents - 1 ? 0 : i + 1), true);
					i++;
				}
			}
		}
	}

	/**
	 * Makes the previous component in the specified <code>Container</code>
	 * the current displayed child.
	 * <p>
	 * If the current displayed component is the first child in the
	 * container, the last child will be the current displayed
	 * child.
	 *
	 * @param		container	the container
	 */
	public void previous(Container container) {
		int ncomponents = container.getComponentCount();
		if (ncomponents <= 0) return;

		boolean found = false;
		for (int i=ncomponents - 1; i>=0; i--) {

			Component component = container.getComponent(i);

			if (component.isVisible()) {
				component.setVisible(false);
				if (!found) {
					found = true;
					show(container, container.getComponent(i == 0 ? ncomponents - 1 : i - 1), true);
					i--;
				}
			}
		}
	}

	/**
	 * Makes the last component in the specified <code>Container</code>
	 * the current displayed child.
	 *
	 * @param		container	the container
	 */
	public void last(Container container) {
		int ncomponents = container.getComponentCount();
		if (ncomponents <= 0) return;

		for (int i=0; i<ncomponents - 1; i++) {
			Component component = container.getComponent(i);
			component.setVisible(false);
		}

		show(container, container.getComponent(ncomponents - 1), true);
	}

	/**
	 * Makes the component with the given name in the specified
	 * <code>Container</code> the current displayed child.
	 * <p>
	 * This is a noop, if no child was added to the container
	 * with the specified name.
	 *
	 * @param		container	the container
	 * @param		name		the of the component to be shown.  This
	 * 							the value used to added the component to the
	 *							container
	 */
	public void show(Container container, String name) {
		int ncomponents = container.getComponentCount();
		if (ncomponents <= 0) return;

		Object object = tab.get(name);

		if (object == null) return;

		for (int i=0; i<ncomponents; i++) {
			Component component = container.getComponent(i);
			if (component != object) {
				component.setVisible(false);
			}
		}

		show(container, (Component)object, true);
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[hgap=" + getHgap() + ",vgap=" + getVgap() + "]";  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	/**
	 * Lays out one component.
	 */
	void layoutComponent(Container container, Component component, boolean validate) {

		Dimension size   = container.getSize();
		Insets    insets = container.getInsets();

		component.setBounds(
			insets.left + hgap,
			insets.top  + vgap,
			size.width  - insets.left - insets.right  - (2 * hgap),
			size.height - insets.top  - insets.bottom - (2 * vgap));

		if (validate) component.validate();
	}

	/**
	 * Shows the specified <code>Component</code> and updates
	 * its bounds.  Validates the component if specified.
	 *
	 * @param		container	the container
	 * @param		component	the component to be shown
	 * @param		validate	<code>true</code> if the component should be
	 *							validated, <code>false</code> otherwise
	 *
	 */
	void show(Container container, Component component, boolean validate) {
		component.setVisible(true);

		layoutComponent(container, component, validate);
	}

}
