package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Checkbox</code> is graphical user interface that
 * displays a text and an indicator that can be either selected
 * or not. The indicator is usually rounded or square.
 * <p>
 * A set of checkboxes can belong to a group and only one checkbox
 * in that group will be selected at a given time.
 *
 * @author		IBM
 * @version		initial
 */
public class Checkbox extends Component implements ItemSelectable {

	static final long serialVersionUID = 7270714317450821763L;

	// Field to support TCK serialization tests
	private int checkboxSerializedDataVersion;

	// The text of the receiver
	String label;

	// The state of the receiver
	boolean state;

	// The checkbox group of the receiver
	CheckboxGroup group;

	// The item listener of the receiver
	transient ItemListener itemListener;

	/**
	 * Constructs a new <code>Checkbox</code> with a <code>""</code> text.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Checkbox()
	 	throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Checkbox</code> with the specified text.
	 *
	 * @param	text	the text display by the receiver
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Checkbox(String text)
	 	throws HeadlessException
	{
		this(text, false);
	}

	/**
	 * Constructs a new <code>Checkbox</code> with the specified text and
	 * initial state.
	 *
	 * @param	text	the text displayed by the receiver
	 * @param	state	the initial state for the receiver
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Checkbox(String text, boolean state)
	 	throws HeadlessException
	{
		this(text, state, null);
	}

	/**
	 * Constructs a new <code>Checkbox</code> with the specified text,
	 * initial state and checkbox group.
	 *
	 * @param	text			the text display by the receiver
	 * @param	state			the initial state for the receiver
	 * @param	checkboxGroup	the initial state for the receiver
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Checkbox(String text, boolean state, CheckboxGroup checkboxGroup)
	 	throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		this.label = text;
		this.state = state;
		this.group = checkboxGroup;

		if (group != null)
			group.updateSelectedCheckbox(this, state);
	}

	/**
	 * Constructs a new <code>Checkbox</code> with the specified text,
	 * initial state and checkbox group.
	 *
	 * @param	text			the text displayed by the receiver
	 * @param	checkboxGroup	the initial state for the receiver
	 * @param	state			the initial state for the receiver
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Checkbox(String text, CheckboxGroup checkboxGroup, boolean state)
	 	throws HeadlessException
	{
		this(text, state, checkboxGroup);
	}

	String classNonlocalizedName() {
		return "checkbox"; //$NON-NLS-1$
	}

	void _setupPeer() {
		OS.Button_SetLabel(peer, label == null ? "" : label); //$NON-NLS-1$
		OS.Button_SetSelection(peer, state);

		super._setupPeer();
	}

	int _createPeer(Container hwParent) {
		int parentPeer = hwParent.peer;
		if (group == null) {
			return OS.Checkbox_New(parentPeer, 0);
		} else {
			return OS.RadioButton_New(parentPeer, 0);
		}
	}

	void buttonCallback() {
		postItemEvent(!state);
	}

	/**
	 * @see java.awt.Component#isLightweight()
	 */
	public boolean isLightweight() {
		return false;
	}

	/**
	 * @see java.awt.Component#isOpaque()
	 */
	public boolean isOpaque() {
		return true;
	}

	boolean isFocusTraversableImpl() {
		return true;
	}

	/**
	 * Answer the <code>String</code> that is the text
	 * displayed by the receiver.  Answer <code>null</code> if
	 * no text is specified.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> which is text for the receiver
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Set the <code>String</code> that is the text
	 * diplayed by the receiver.  Use <code>null</code> if
	 * no label is desired.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text to be displayed by the receiver
	 */
	public void setLabel(String text) {
//		synchronized (this) {
			if (this.label == text) {
				return;
			}
			this.label = text;

			acquirePeerReadLock();
			try {
				if (peer != 0) OS.Button_SetLabel(peer, text == null ? "" : text); //$NON-NLS-1$
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	/**
	 * Answer the state of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>boolean</code> representing if the receiver is selected or not
	 */
	public boolean getState() {
		return state;
	}

	/**
	 * Set the state of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	state	<code>true</code> if the receiver is to be selected
	 *					<code>false</code> otherwise
	 */
	public void setState(boolean state) {
//		synchronized (this) {
			if ((group != null) && (group.getSelectedCheckbox() == this) && (!state)) {
				return;
			}
			if (this.state == state) {
				return;
			}
			this.state = state;

			if (group != null)
				group.updateSelectedCheckbox(this, state);

			acquirePeerReadLock();
			try {
				if (peer != 0) OS.Button_SetSelection(peer, state);
			} finally {
				releasePeerReadLock();
			}
//		}
	}

	/**
	 * Answer an array containg the receiver text if it is selected.
	 * If the receiver is not selected, the return value will be
	 * <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Object[]</code> which contains the receiver text or <code>null</code>
	 *
	 * @see	<code>getState()</code>
	 */
	public Object[] getSelectedObjects() {
//		synchronized (this) {
			if (getState())
				return new String[] {getLabel()};

			return null;
//		}
	}

	/**
	 * Answer the <code>CheckboxGroup</code> which the receiver
	 * belongs to.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>CheckboxGroup</code> which is the receiver checkbox group
	 */
	public CheckboxGroup getCheckboxGroup() {
		return group;
	}

	/**
	 * Set the <code>CheckboxGroup</code> which the receiver
	 * belongs to.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	checkboxGroup	the checkbox group for the receiver
	 */
	public void setCheckboxGroup(CheckboxGroup checkboxGroup) {
//		synchronized (this) {
			if (this.group == checkboxGroup) {
				return;
			}

			CheckboxGroup oldGroup = this.group;

			this.group = checkboxGroup;

			if (group != null) {
				// New behavior as of PPro 1.1
				if (group.getSelectedCheckbox() != null) {
					setState(false);
					//return;
				} else {
				group.updateSelectedCheckbox(this, state);
				}
			}

			if (oldGroup != null) {
				oldGroup.updateSelectedCheckbox(this, false);
			}

			if (peer != 0) {
				// If the checkbox is switching from in a group
				// to not in a group, or the other way around then
				// rebuild the peer to reflect the change in the
				// native widget
				if ((oldGroup == null) != (checkboxGroup == null)) recreate();
			}
//	}
}

	/**
	 * Adds a <code>ItemListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be added.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void addItemListener(ItemListener itemListener) {
		if (itemListener == null) return;
		addEventListener(itemListener);
		this.itemListener = AWTEventMulticaster.add(this.itemListener, itemListener);
	}

	/**
	 * Removes an <code>ItemListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be removed.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void removeItemListener(ItemListener itemListener) {
		if (itemListener == null) return;
		this.itemListener = AWTEventMulticaster.remove(this.itemListener, itemListener);
	}

	public ItemListener[] getItemListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(itemListener, ItemListener.class);
		ItemListener[] resultArray = new ItemListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ActionEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processItemEvent(ItemEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ItemEvent) {
			processItemEvent((ItemEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ItemEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processItemEvent(ItemEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	itemEvent	the event to be processed
	 *
	 * @see		ItemEvent
	 */
	protected void processItemEvent(ItemEvent itemEvent) {
		if (itemListener == null)
			return;

		if (itemEvent == null) {
			itemListener.itemStateChanged(itemEvent);
			return;
		}

		switch (itemEvent.getID()) {
			case ItemEvent.ITEM_STATE_CHANGED:
				itemListener.itemStateChanged(itemEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		String labelString = getLabel() == null ? "" : ",label=" + getLabel(); //$NON-NLS-1$ //$NON-NLS-2$
		return super.paramString() +
			labelString +
			",state=" + getState(); //$NON-NLS-1$
	}

	/**
	 * @return
	 */
	// see Bug 99618
	// see Bug 99620
	void postItemEvent(boolean newSelection) {
		// This first check lets the common layer be a bit more flexible about
		// how the peer layer sends events back.  For instance, if the peer layer sends
		// events when the checkbox is selected through API then this check will
		// avoid that event from being sent to the Java listeners.
		if (newSelection == state) return;

		// if the receiver is in a group, it can only be deselected
		// if another radio button in the same group is selected.
//		if (!Util.isGTK()) {  // should revisit this
			if (state && group != null) {
				acquirePeerReadLock();
				try {
					if (peer != 0) OS.Button_SetSelection(peer, true);
				} finally {
					releasePeerReadLock();
				}
				return;
			}
//		}

		state = newSelection;

		if (group != null)
			group.updateSelectedCheckbox(this, newSelection);

		ItemEvent itemEvent = new ItemEvent(
			this, ItemEvent.ITEM_STATE_CHANGED,
			label, state ? ItemEvent.SELECTED : ItemEvent.DESELECTED);
		postEvent(itemEvent);
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ItemEvent)
			return ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) || itemListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, itemListenerK, itemListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(itemListenerK))
				addItemListener((ItemListener)listener);
		}
	}

}
