package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.*;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>CheckBoxMenuItem</code> is the toggleable user interface object
 * which appears in a <code>Menu</code> object.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class CheckboxMenuItem extends MenuItem implements ItemSelectable {

	static final long serialVersionUID = 6190621106981774043L;
	private int checkboxMenuItemSerializedDataVersion;

	boolean state;

	// The item listener of the receiver
	transient ItemListener itemListener;

	/**
	 * Constructs a new <code>CheckBoxMenuItem</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public CheckboxMenuItem()
 	  	throws HeadlessException
{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>CheckBoxMenuItem</code> with a label.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	label	the text label for the receiver
	 */
	public CheckboxMenuItem(String label)
 		throws HeadlessException
{
		this(label, false);
	}

	/**
	 * Constructs a new <code>CheckBoxMenuItem</code> with a label and
	 * a boolean indicating whether it should appear checked or not.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	label	the text label for the receiver
	 * @param	checked	the checked state of the receiver
	 *
	 */
	public CheckboxMenuItem(String textLabel, boolean checked)
	 	throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		label = textLabel;
		state = checked;
	}

	void _addNotify() {
		super._addNotify();
		// "check" the checkbox if state is true
		if (peer != 0) {
			OS.StateMenuItem_SetSelection(peer, state);
		}
	}

	String classNonlocalizedName() {
		return "checkboxmenuitem"; //$NON-NLS-1$
	}

	int _createPeer() {
		return OS.StateMenuItem_New(_getParentPeer(), index, OS.STYLE_CHECK);
	}

	void selectionCallback(int modifiers) {
		postItemEvent(!state);
		super.selectionCallback(modifiers);
	}

	/**
	 * Answer the state of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>boolean</code> representing if the receiver is selected or not
	 */
	public boolean getState() {
		return state;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ItemEvent)
			return ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) || itemListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	/**
	 * Set the state of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	state	<code>true</code> if the receiver is to be selected
	 *				<code>false</code> otherwise
	 */
	public void setState(boolean state) {
		this.state = state;

		if (peer != 0)
			OS.StateMenuItem_SetSelection(peer, state);
	}

	/**
	 * Answer an array with the receiver's label if it is selected
	 * or an array with <code>null</code> if it is not.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return an array of <code>Object</code>	 the selected item labels
	 */
	public synchronized Object[] getSelectedObjects() {
		if (getState())
			return new String[]{ getLabel() };

		return null;
	}

	/**
	 * Adds an <code>ItemListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be added.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void addItemListener(ItemListener itemListener) {
		addEventListener(itemListener);
		this.itemListener = AWTEventMulticaster.add(this.itemListener, itemListener);
	}

	/**
	 * Removes an <code>ItemListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be removed.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void removeItemListener(ItemListener itemListener) {
		this.itemListener = AWTEventMulticaster.remove(this.itemListener, itemListener);
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a menu component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a menu component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ItemEvent</code>, the receiver
	 * will forward calls to the default events by calling
	 * <code>processItemEvent</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ItemEvent) {
			processItemEvent((ItemEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ItemEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processItemEvent(ItemEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	itemEvent	the event to be processed
	 *
	 * @see		ItemEvent
	 */
	protected void processItemEvent(ItemEvent itemEvent) {
		if (itemListener == null)
			return;

		if ((itemEvent == null) ||
		    (itemEvent.getID() == ItemEvent.ITEM_STATE_CHANGED)) {
			itemListener.itemStateChanged(itemEvent);
		}
	}

	/**
	 * @return
	 */
	void postItemEvent(boolean selection) {
		state = selection;

		ItemEvent itemEvent = new ItemEvent(
			this, ItemEvent.ITEM_STATE_CHANGED,
			getLabel(), state ? ItemEvent.SELECTED : ItemEvent.DESELECTED);
		postEvent(itemEvent);
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the menu component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code>	the param string.
	 */
	public String paramString() {
		return super.paramString() + " state=" + getState(); //$NON-NLS-1$
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, Component.itemListenerK, itemListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(Component.itemListenerK))
				addItemListener((ItemListener)listener);
		}
	}

	public ItemListener[] getItemListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(itemListener, ItemListener.class);
		ItemListener[] resultArray = new ItemListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}
}
