package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.*;
import java.util.EventListener;
import java.util.Vector;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Choice</code> is a graphical interface that displays only
 * one item from a list of selectable items.
 * <p>
 * When the <code>Choice</code> is clicked, a popup list will display
 * all selectable items so that the user can select the item to be
 * displayed.
 * <p>
 * Each time the user selects an item, the <code>Choice</code> will
 * fire a item event.
 *
 * @author		IBM
 * @version		initial
 */
public class Choice extends Component implements ItemSelectable {

	static final long serialVersionUID = -4075310674757313071L;
	private int choiceSerializedDataVersion;

	// selectable items of the receiver
	Vector pItems;

	// position of the selected item
	int selectedIndex;

	// The item listener of the receiver
	transient ItemListener itemListener;

	/**
	 * Constructs a new <code>Choice</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Choice()
	  	throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		pItems = new Vector();
		selectedIndex = -1;
	}

	int _createPeer(Container hwParent) {
		return OS.Choice_New(hwParent.peer);
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	void _setupPeer() {

		if (!pItems.isEmpty()) {
			int size = pItems.size();
			for (int i=0; i<size; i++) {
				OS.Choice_AddItem(peer, (String) pItems.elementAt(i), i);
			}
			OS.Choice_SelectItem(peer, selectedIndex == -1 ? 0 : selectedIndex);
		}

		super._setupPeer();
	}

	private void choiceCallback(int index) {
		/*
		 * See bug 99327
		 */
		postItemEvent(index);
	}

	/**
	 * Destroy the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void removeNotify() {
		acquirePeerReadLock();
		try {
			if (peer != 0) selectedIndex = OS.Choice_GetSelectedItemIndex(peer);
		} finally {
			releasePeerReadLock();
		}

		super.removeNotify();
	}

	public boolean isLightweight() {
		return false;
	}

	boolean isFocusTraversableImpl() {
		return true;
	}

	/**
	 * Answers the number of selectable items in the list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the number of items in the list
	 */
	public int getItemCount() {
		return pItems.size();
	}

	String classNonlocalizedName() {
		return "choice"; //$NON-NLS-1$
	}

	/**
	 * Answers the number of selectable items in the list.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the number of items in the list
	 *
	 * @deprecated	use <code>getItemCount()</code> instead
	 */
	public int countItems() {
		return getItemCount();
	}

	/**
	 * Answers the selectable item at the specified position.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the index of the desired item
	 *
	 * @return	an <code>String</code> which is the item at the given position
	 *
	 * @throws	IllegalArgumentException	if there is not an item at the specified position
	 */
	public String getItem(int position) {
		if (pItems == null) {
			return null;
		}
		else {
			return (String)pItems.elementAt(position);
		}
	}

	/**
	 * Adds a new element to the end of the list of selectable items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item	the item to be added
	 *
	 * @throws	NullPointerException	if the new item is <code>null</code>
	 */
	public void add(String item) {
		addItem(item);
	}

	/**
	 * Adds a new element to the end of the list of selectable items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item	the item to be added
	 *
	 * @throws	NullPointerException	if the new item is <code>null</code>
	 *
	 * @deprecated	use <code>add(String)</code> instead
	 */
	public void addItem(String item) {
		if (item == null)
			throw new NullPointerException("cannot add null item to Choice"); //$NON-NLS-1$

		pItems.addElement(item);

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.Choice_AddItem(peer, item, pItems.size() - 1);
				if (OS.Choice_GetSelectedItemIndex(peer) == -1)
					OS.Choice_SelectItem(peer, 0);
			} else if (selectedIndex == -1) {
				selectedIndex = 0;
			}
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	/**
	 * Adds a new element to the list of selectable items at the
	 * specified position. The new item will be added to the end of
	 * the current list, if the specified position is bigger than
	 * its size.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item		the item to be added
	 * @param	position	the position for the new item
	 *
	 * @throws	IllegalArgumentException	if the specifed position is negative
	 */
	public void insert(String item, int position) {
//		synchronized (this) {
			if (position < 0)
				throw new IllegalArgumentException();

			if (item == null)
				throw new NullPointerException("cannot add null item to Choice"); //$NON-NLS-1$

			int n = getItemCount();
			if (position > n)
				position = n;

			pItems.insertElementAt(item, position);

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					int oldIndex = getSelectedIndex();
					OS.Choice_AddItem(peer, item, position);
					if (position <= oldIndex) {
						// JDK resets to 0 in this case
						select(0);
					}
				} else if (position <= selectedIndex) {
					// selectedIndex++;
					// JDK resets to 0 in this case
					selectedIndex = 0;
				} else if (selectedIndex == -1) {
					selectedIndex = position;
				}
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ItemEvent)
			return ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) || itemListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	public boolean isOpaque() {
		return true;
	}

	/**
	 * Removes the first item in the list of selectable items that is
	 * equal to the specified item.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item	the item to be removed
	 *
	 * @throws	IllegalArgumentException	if the specified item does not
	 *										exists in the list
	 */
	public void remove(String item) {
		synchronized (this) {
			int position = pItems.indexOf(item);

			if (position == -1)
				throw new IllegalArgumentException();

			remove(position);
		}
	}

	/**
	 * Removes the item from the receiver's list at the specified
	 * position.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the index of the item to be removed
	 */
	public void remove(int position) {
//		synchronized (this) {
			if (position < 0 || position >= getItemCount())
				throw new ArrayIndexOutOfBoundsException();

			pItems.removeElementAt(position);

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					int oldSel = getSelectedIndex();
					OS.Choice_RemoveItem(peer, position);
					if (oldSel == position) {
						if (getItemCount() > 0) {
							OS.Choice_SelectItem(peer, 0);
						}
					}
				} else {
					if (getItemCount() == 0) {
						selectedIndex = -1;
					} else if (selectedIndex == position) {
						selectedIndex = 0;
					} else if (position <= selectedIndex) {
						selectedIndex--;
					}
				}
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	/**
	 * Removes all items from the receiver's list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 */
	public void removeAll() {
		synchronized (this) {
			pItems = new Vector();

			acquirePeerReadLock();
			try {
				if (peer != 0)
					OS.Choice_RemoveAllItems(peer);
				else
					selectedIndex = -1;
			} finally {
				releasePeerReadLock();
			}
		}
		if (isValid()) invalidate();
	}

	/**
	 * Answers the selected item.  If there is not a selected item,
	 * answers <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the selected item or <code>null</code>
	 */
	public synchronized String getSelectedItem() {
		int position = getSelectedIndex();

		if (position < 0)
			return null;

		return getItem(position);
	}

	/**
	 * Answers an array containg the selected item. If the list of selectable items
	 * is empty, the return value will be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Object[]</code> which contains the selected item or <code>null</code>
	 *
	 * @see	<code>getState()</code>
	 */
	public synchronized Object[] getSelectedObjects() {
		String selection = getSelectedItem();
		if (selection != null)
			return new String[]{ selection };

		return null;
	}

	/**
	 * Answers the position of the selected item.  If there is not a selected
	 * item, answers <code>-1</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the position of the selected item or <code>-1</code>
	 */
	public int getSelectedIndex() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return selectedIndex;

			return OS.Choice_GetSelectedItemIndex(peer);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the item at the specified position to be the receiver's
	 * selected item.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the index of the item to be selected
	 *
	 * @throws	IllegalArgumentEception	if there is not an item at the specified position
	 */
	public synchronized void select(int position) {
		if (position < 0 || position >= getItemCount()) {
			throw new IllegalArgumentException("illegal Choice item position: "+position); //$NON-NLS-1$
		}

		if (position == getSelectedIndex()) {
			return;
		}

		acquirePeerReadLock();
		try {
			if (peer != 0)
				OS.Choice_SelectItem(peer, position);
			else
				selectedIndex = position;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the receiver's seleted item to be the first item in the
	 * list of selectable items that is equal to the specified item.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item	the item to be selected
	 */
	public synchronized void select(String item) {
		int position = pItems.indexOf(item);
		if (position != -1)
			select(position);
	}

	/**
	 * Adds a <code>ItemListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	itemListener	the <code>ItemListener</code> to be added.
	 *
	 * @see		java.awt.event.ItemListener
	 * @see		java.awt.event.ItemEvent
	 */
	public synchronized void addItemListener(ItemListener itemListener) {
		if (itemListener == null) return;
		addEventListener(itemListener);
		this.itemListener = AWTEventMulticaster.add(this.itemListener, itemListener);
	}

	/**
	 * Removes an <code>ItemListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	itemListener	the <code>ItemListener</code> to be removed.
	 *
	 * @see		java.awt.event.ItemListener
	 * @see		java.awt.event.ItemEvent
	 */
	public synchronized void removeItemListener(ItemListener itemListener) {
		if (itemListener == null) return;
		this.itemListener = AWTEventMulticaster.remove(this.itemListener, itemListener);
	}

	public ItemListener[] getItemListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(itemListener, ItemListener.class);
		ItemListener[] resultArray = new ItemListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ActionEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processItemEvent(ItemEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ItemEvent) {
			processItemEvent((ItemEvent) awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ItemEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processItemEvent(ItemEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	itemEvent	the event to be processed
	 *
	 * @see		ItemEvent
	 */
	protected void processItemEvent(ItemEvent itemEvent) {
		if (itemListener == null)
			return;

		if ((itemEvent == null) || (itemEvent.getID() == ItemEvent.ITEM_STATE_CHANGED)) {
			itemListener.itemStateChanged(itemEvent);
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() + ",selected=" + getSelectedItem(); //$NON-NLS-1$
	}

	/**
	 * @return true if the event should be processed
	 */
//	boolean postItemEvent() {
	/*
	 * See bug 99327
	 */
	boolean postItemEvent(int index) {
//		ChoicePeer combo = (ChoicePeer) peer;
//		int selInd = combo.getSelectionIndex();
		if (index == -1) return true;
		ItemEvent itemEvent = new ItemEvent(
			this, ItemEvent.ITEM_STATE_CHANGED,
//			combo.getItem(selInd), ItemEvent.SELECTED);
			getItem(index), ItemEvent.SELECTED);
		postEvent(itemEvent);
		return !itemEvent.isConsumed();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, itemListenerK, itemListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(itemListenerK))
				addItemListener((ItemListener)listener);
		}
	}
}
