package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.color.ColorSpace;
import java.io.Serializable;
import java.util.Hashtable;

import com.ibm.oti.awt.Util;
import com.ibm.ugl.p3ml.OS;

/**
 * A <code>Color</code> represents RGB value.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class Color implements Transparency, Serializable {

	static final long serialVersionUID = 118526816881161077L;

	/**
	 * The basic colors.
	 */
	public static final Color white = new Color(255, 255, 255);
	public static final Color lightGray = new Color(192, 192, 192);
	public static final Color gray = new Color(128, 128, 128);
	public static final Color darkGray = new Color(64, 64, 64);
	public static final Color black = new Color(0, 0, 0);
	public static final Color red = new Color(255, 0, 0);
	public static final Color pink = new Color(255, 175, 175);
	public static final Color orange = new Color(255, 200, 0);
	public static final Color yellow = new Color(255, 255, 0);
	public static final Color green = new Color(0, 255, 0);
	public static final Color magenta = new Color(255, 0, 255);
	public static final Color cyan = new Color(0, 255, 255);
	public static final Color blue = new Color(0, 0, 255);

	public static final Color WHITE = white;
	public static final Color LIGHT_GRAY = lightGray;
	public static final Color GRAY = gray;
	public static final Color DARK_GRAY = darkGray;
	public static final Color BLACK = black;
	public static final Color RED = red;
	public static final Color PINK = pink;
	public static final Color ORANGE = orange;
	public static final Color YELLOW = yellow;
	public static final Color GREEN = green;
	public static final Color MAGENTA = magenta;
	public static final Color CYAN = cyan;
	public static final Color BLUE = blue;

	float falpha;
	float[] frgbvalue;
	float[] fvalue;

	/**
	 * The receiver RGB value and alpha in the default <code>ColorSpace</code>
	 */
	int value;
	transient float[] valueAsPercentages;

	static int RED_INDEX= 0;
	static int GREEN_INDEX= 1;
	static int BLUE_INDEX= 2;
	static int ALPHA= 3;

	static ColorCache colorCache = new ColorCache();
	transient int colorHandle;

	/**
	 * Caches the <code>ColorPeer</code> for a given rgb value.
	 */
	static class ColorCache {
		transient Hashtable cache = new Hashtable();

		/**
		 * Returns the <code>ColorPeer</code> with the given rgb value.
		 * Returns either an existing cached instance or creates a new
		 * one and caches it.
		 *
		 * @param rgb RGB of the desired <code>ColorPeer</code>
		 * @return a handle to the native color resource with the given rgb value
		 */
		public int getColor(int rgb) {
			Integer integerRgb = new Integer(rgb);
			ColorPeerReferenceCount referenceCount = (ColorPeerReferenceCount) cache.get(integerRgb);
			int cachedHandle;

			if (referenceCount != null) {
				cachedHandle = referenceCount.getReference();
			} else {
				int red = (rgb >> 16) & 0xff;
				int green = (rgb >> 8) & 0xff;
				int blue = rgb & 0xff;
				cachedHandle = OS.Color_New(Util.getDisplay().fPeerHandle, red, green, blue);
				referenceCount = new ColorPeerReferenceCount(cachedHandle);
				cache.put(integerRgb, referenceCount);
			}
			return cachedHandle;
		}
		/**
		 * Removes one reference to the <code>ColorPeer</code> with the
		 * given RGB value. Disposes the peer and removes it from the cache
		 * if the last reference was removed.
		 *
		 * @param rgb RGB of the <code>ColorPeer</code> that should be remove
 		 */
		public void removeColor(int rgb) {
			Integer integerRgb = new Integer(rgb);
			ColorPeerReferenceCount referenceCount = (ColorPeerReferenceCount) cache.get(integerRgb);

			if (referenceCount != null) {
				referenceCount.dereference();
				if (referenceCount.hasReference() == false) {
					int cachedHandle = referenceCount.getReference();
					cache.remove(integerRgb);
					OS.Color_Dispose(cachedHandle);
				}
			}
		}
	}

	/**
	 * A <code>ColorPeerReferenceCount</code> keeps a reference count to a
	 * <code>ColorPeer</code> object.
	 */
	static class ColorPeerReferenceCount {
		private int count = 1;
		private int colorHandle;

		/**
		 * Creates a new instance of the receiver.
		 * The reference count will be set to one.
		 *
		 * @param colorHandle the color handle whose references
		 * 	are being tracked
		 */
		public ColorPeerReferenceCount(int colorHandle) {
			this.colorHandle = colorHandle;
		}
		/**
		 * Decreases the reference count by one.
		 */
		public void dereference() {
			count--;
		}
		/**
		 * Returns the color handle and increases
		 * the reference count by one.
		 *
		 * @return the handle to the native color
		 */
		public int getReference() {
			count++;
			return colorHandle;
		}
		/**
		 * Returns whether there are references to the <code>ColorPeer</code>
		 * object.
		 *
		 * @return <code>true</code> if there is at least one reference to the
		 * 	<code>ColorPeer</code>. <code>false</code> if there are no references
		 *  to the <code>ColorPeer</code>.
		 */
		public boolean hasReference() {
			return count > 0;
		}
	}

	/**
	 * Answers the factor used to generate brighter and darker colors.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	component	the RGB component
	 *
	 * @return	an <code>int</code> which is the brighter value for the given component
	 */
	static float factor(int component) {
		switch (component) {
			case 0: return 0;
			case 1: return 0.25f;
			case 2: return 0.5f;
			default: return 0.7f;
		}
	}

	/**
	 * Answers the brighter value for the specified RGB component.  This algorithm
	 * divides the RGB component by a fixed factor and truncate the result.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	component	the RGB component
	 *
	 * @return	an <code>int</code> which is the brighter value for the given component
	 */
	static int brighter(int component) {
		return Math.min(255, (int)(component / factor(component)));
	}

	/**
	 * Answers the darker value for the specified RGB component.  This algorithm
	 * multiples the RGB component by a fixed factor and truncate the result.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	component	the RGB component
	 *
	 * @return	an <code>int</code> which is the darker value for the given component
	 */
	static int darker(int component) {
		return (int)(component * factor(component));
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB components and alpha channel equals to <code>255<code>.
	 * <p>
	 * Color components must be in the range (0 - 255).
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Color(int red, int green, int blue) {
		this(red, green, blue, 0xff);
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB components and alpha channel.
	 * <p>
	 * Color components and alpha channel must be in the range (0 - 255).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @throws	IllegalArgumentException	if any of the components or alpha channel
	 *							are invalid
	 */
	public Color(int red, int green, int blue, int alpha) {
		if (red   < 0 || red   > 255) throw new IllegalArgumentException("Illegal red value: "   + red); //$NON-NLS-1$
		if (green < 0 || green > 255) throw new IllegalArgumentException("Illegal green value: " + green); //$NON-NLS-1$
		if (blue  < 0 || blue  > 255) throw new IllegalArgumentException("Illegal blue value: "  + blue); //$NON-NLS-1$
		if (alpha < 0 || alpha > 255) throw new IllegalArgumentException("Illegal alpha value: " + alpha); //$NON-NLS-1$

		value = (alpha << 24) | (red << 16) | (green << 8) | blue;
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB value.  The color is completely opaque.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Color(int rgb) {
		this(rgb, false);
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB value and alpha channel.  If <code>hasAlpha</code>
	 * is <code>false<code>, the alpha channel is set to <code>255</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Color(int rgba, boolean hasAlpha) {
		if (hasAlpha) {
			value = rgba;
		} else {
			value = rgba | 0xff000000;
		}
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB components and alpha channel equals to <code>1.0<code>.
	 * <p>
	 * Color components must be in the range (0.0 - 1.0).
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Color(float red, float green, float blue) {
		this(red, green, blue, 1.0f);
	}

	/**
	 * Constructs a new <code>Color</code> in the default <code>ColorSpace</code>,
	 * with the given RGB components and alpha channel.
	 * <p>
	 * Color components and alpha channel must be in the range (0.0 - 1.0).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @throws	IllegalArgumentException	if any of the components or alpha channel
	 *							are invalid
	 */
	public Color(float red, float green, float blue, float alpha) {
		if (red   < 0 || red   > 1) throw new IllegalArgumentException("Illegal red value: "   + red); //$NON-NLS-1$
		if (green < 0 || green > 1) throw new IllegalArgumentException("Illegal green value: " + green); //$NON-NLS-1$
		if (blue  < 0 || blue  > 1) throw new IllegalArgumentException("Illegal blue value: "  + blue); //$NON-NLS-1$
		if (alpha < 0 || alpha > 1) throw new IllegalArgumentException("Illegal alpha value: " + alpha); //$NON-NLS-1$

		valueAsPercentages = new float[] {red, green, blue, alpha};

		value =
			((Math.round(alpha * 255) & 0xff) << 24) |
			((Math.round(red   * 255) & 0xff) << 16) |
			((Math.round(green * 255) & 0xff) <<  8) |
			 (Math.round(blue  * 255) & 0xff);
	}

	/**
	 * Answers the red component of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the red component of the receiver
	 * 			in the range (0 - 255)
	 */
	public int getRed() {
		return (getRGB() >> 16) & 0xff;
	}

	/**
	 * Answers the green component of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the green component of the receiver
	 * 			in the range (0 - 255)
	 */
	public int getGreen() {
		return (getRGB() >> 8) & 0xff;
	}

	/**
	 * Answers the blue component of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the blue component of the receiver
	 * 			in the range (0 - 255)
	 *
	 */
	public int getBlue() {
		return getRGB() & 0xff;
	}

	/**
	 * Answers the alpha channel of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the alpha channel of the receiver
	 * 			in the range (0 - 255)
	 *
	 * @see #getRGB
	 * @since JDK1.2
	 */
	public int getAlpha() {
		return (getRGB() >> 24) & 0x00ff;
	}

	/**
	 * Answers the RGB value of the receiver in the default <code>ColorSpace</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the receiver's RGB value
	 */
	public int getRGB() {
		return value;
	}

	/**
	 * Answers a brighter version of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Color</code> which is the brighter version of the receiver
	 */
	public Color brighter() {
		if (equals(Color.black))
			return new Color(3, 3, 3);
		return new Color(brighter(getRed()), brighter(getGreen()), brighter(getBlue()));
	}

	/**
	 * Removes the color peer reference from the cache.
	 */
	void disposePeer() {
		if (colorHandle == 0) return;
		colorCache.removeColor(getRGB());
		colorHandle = 0;
	}

	/**
	 * Answers a darker version of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Color</code> which is the darker version of the receiver
	 */
	public Color darker() {
		return new Color(darker(getRed()), darker(getGreen()), darker(getBlue()));
	}

	/**
	 * Answers an integer hash code for the receiver. Any two
	 * objects which answer <code>true</code> when passed to
	 * <code>equals(Object)</code> must answer the same value for this
	 * method.
	 *
	 * @author	IBM
	 * @version		initial
	 *
	 * @return		an <code>int</code> which is the receiver's hash.
	 *
	 * @see			<code>equals(Object)</code>
	 */
	public int hashCode() {
		return getRGB();
	}

	/**
	 * Compares the argument to the receiver, and answers true
	 * if the argument is a <code>Color</code> and its red, green
	 * and blue components are equals to the receiver's.
	 *
	 * @author	IBM
	 * @version		initial
	 *
	 * @param		object	Object	the object to compare with this dimension
	 *
	 * @return		<code>true</code> if the object is the same as this object
	 *				<code>false</code> otherwise
	 */
	public boolean equals(Object object) {
		if (object == this) return true;
		if (!(object instanceof Color)) return false;
		return getRGB() == ((Color)object).getRGB();
	}

	/**
	 * Removes the color peer reference from the cache.
	 */
	protected void finalize() throws Throwable {
		disposePeer();
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author	IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 *
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
		buffer.append(getClass().getName());
		buffer.append("[r="); //$NON-NLS-1$
		buffer.append(getRed());
		buffer.append(",g="); //$NON-NLS-1$
		buffer.append(getGreen());
		buffer.append(",b="); //$NON-NLS-1$
		buffer.append(getBlue());
		if (getAlpha() != 0xff) {
			buffer.append(",a="); //$NON-NLS-1$
			buffer.append(getAlpha());
		}
		buffer.append("]"); //$NON-NLS-1$
		return buffer.toString();
	}

	/**
	 * Decodes the speficied <code>String</code> in a <code>Color</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Color</code>
	 */
	public static Color decode(String rgbString) throws NumberFormatException {
		return new Color(Integer.decode(rgbString).intValue());
	}

	/**
	 * Answers the <code>Color</code> for the speficied system
	 * property.  Answers <code>null</code>, if the property is
	 * invalid.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the system property
	 *
	 * @return	a <code>Color</code>
	 */
	public static Color getColor(String propertyName) {
		return getColor(propertyName, null);
	}

	/**
	 * Answers the <code>Color</code> for the speficied system
	 * property.  Answers the default value, if the property is
	 * invalid.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the system property
	 * @param	defaultColor	the default color
	 *
	 * @return	a <code>Color</code>
	 */
	public static Color getColor(String propertyName, Color defaultValue) {
		try {
			return decode(System.getProperty(propertyName));
		} catch (Exception e) {
			return defaultValue;
		}
	}

	/**
	 * Answers the <code>Color</code> for the speficied system
	 * property.  Answers the <code>Color</code> with the specified
	 * RGB value, if the property is invalid.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the system property
	 * @param	defaultRGB		the default RGB value
	 *
	 * @return	a <code>Color</code>
	 */
	public static Color getColor(String propertyName, int defaultRGB) {
		return getColor(propertyName, new Color(defaultRGB));
	}

	/**
	 * Converts the specified HSB values to a RBG value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	hue		the hue component
	 * @param	saturation	the saturation component
	 * @param	brightness	the brightness component
	 *
	 * @return	an <code>int</code> which is the converted RGB value
	 */
	public static int HSBtoRGB(float hue, float saturation, float brightness) {
		float s = saturation;
		float v = brightness;

		int x = (int)Math.round(v * 255);

		if (s == 0) {
			return 0xff000000 | (x << 16) | (x << 8) | x;
		} else {
			float h;
			if (hue == 1.0f)
				h = 0f;
			else
				h = (hue - (float)Math.floor(hue)) * 6.0f;

			int i = (int)Math.floor(h);
			float f = h - i;

			int p = (int)Math.round(255 * (v * (1 - s)));
			int q = (int)Math.round(255 * (v * (1 - (s * f))));
			int t = (int)Math.round(255 * (v * (1 - (s * (1 - f)))));
			switch(i){
				case 0: return 0xff000000 | (x << 16) | (t << 8) | p;
				case 1: return 0xff000000 | (q << 16) | (x << 8) | p;
				case 2: return 0xff000000 | (p << 16) | (x << 8) | t;
				case 3: return 0xff000000 | (p << 16) | (q << 8) | x;
				case 4: return 0xff000000 | (t << 16) | (p << 8) | x;
				case 5: return 0xff000000 | (x << 16) | (p << 8) | q;
				default: return 0;
			}
		}
	}

	/**
	 * Fill and return the specified array with the specified RGB values
	 * converted to HSB. Create, fill and return a new array if <code>null</code>
	 * is passed in.
	 * <p>
	 * Color components must be in the range (0 - 255).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	red	the red component
	 * @param	green	the green component
	 * @param	blue	the blue component
	 * @param	hsb	the return value or <code>null</code>
	 *
	 * @return	an <code>float[]</code> containing the converted HSB values
	 *
	 * @throws	IllegalArgumentException	if any of the components are invalid
	 */
	public static float[] RGBtoHSB(int red, int green, int blue, float hsb[]) {
		float r = red / 255f;
		float g = green / 255f;
		float b = blue / 255f;

		float h = 0; // hue
		float s; // saturation
		float v; // brightnees

		float max = Math.max(r, Math.max(g, b));
		float min = Math.min(r, Math.min(g, b));

		// Calculate brightness
		v = max;

		// Calculate saturation
		if (max != 0)
			s = (max - min)/max;
		else
			s = 0;

		// Calculate hue
		if (s != 0) {
			float delta = max - min;
			if (r == max)
				h = (g - b)/delta;
			else if (g == max)
				h = 2 + (b - r)/delta;
			else if (b == max)
				h = 4 + (r - g)/delta;

			h = h * 60f;
			if (h < 0)
				h += 360f;
			h /= 360f;
		}

		// Fill return value
		if (hsb == null)
			return new float[]{h, s, v};
		else {
			hsb[0] = h;
			hsb[1] = s;
			hsb[2] = v;
			return hsb;
		}
	}

	/**
	 * Creates a color from the specified HSB values.  Converts the
	 * HSB components to RGB and creates a new color.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	hue			the hue component
	 * @param	saturation	the saturation component
	 * @param	brightness	the brightness component
	 *
	 * @return	a <code>Color</code>
	 */
	public static Color getHSBColor(float hue, float saturation, float brightness) {
		return new Color(HSBtoRGB(hue, saturation, brightness));
	}

	/**
	 * Fill and return the specified array with the components and alpha
	 * channel of the receiver in the default <code>ColorSpace</code>. Create,
	 * fill and return a new array if <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>float[]</code> containing the receiver's components
	 */
	public float[] getRGBComponents(float[] floatArray) {
		float[] result = floatArray;
		if (result == null) {
			result = new float[4];
		}

		if (valueAsPercentages != null) {
			result[RED_INDEX]= valueAsPercentages[RED_INDEX];
			result[GREEN_INDEX]= valueAsPercentages[GREEN_INDEX];
			result[BLUE_INDEX]= valueAsPercentages[BLUE_INDEX];
			result[ALPHA]= valueAsPercentages[ALPHA];
		} else {
			result[RED_INDEX]= getRed() / 255f;
			result[GREEN_INDEX]= getGreen() / 255f;
			result[BLUE_INDEX]= getBlue() / 255f;
			result[ALPHA]= getAlpha() / 255f;
		}
		return result;
	}

	/**
	 * Fill and return the specified array with the components of
	 * the receiver in the default <code>ColorSpace</code>. Create,
	 * fill and return a new array if <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>float[]</code> containing the receiver's components
	 */
	public float[] getRGBColorComponents(float[] floatArray) {
		float[] result = floatArray;
		if (result == null) {
			result = new float[3];
		}

		if (valueAsPercentages != null) {
			result[RED_INDEX]= valueAsPercentages[RED_INDEX];
			result[GREEN_INDEX]= valueAsPercentages[GREEN_INDEX];
			result[BLUE_INDEX]= valueAsPercentages[BLUE_INDEX];
		} else {
			result[RED_INDEX]= getRed() / 255f;
			result[GREEN_INDEX]= getGreen() / 255f;
			result[BLUE_INDEX]= getBlue() / 255f;
		}
		return result;
	}

	/**
	 * Fill and return the specified array with the components and alpha
	 * channel of the receiver in the specified <code>ColorSpace</code>. Create,
	 * fill and return a new array if <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>float[]</code> containing the receiver's components
	 */
	public float[] getComponents(float[] floatArray) {
		float[] result = floatArray;
		if (result == null) {
			result = new float[4];
		}

		if (valueAsPercentages != null) {
			result[RED_INDEX]= valueAsPercentages[RED_INDEX];
			result[GREEN_INDEX]= valueAsPercentages[GREEN_INDEX];
			result[BLUE_INDEX]= valueAsPercentages[BLUE_INDEX];
			result[ALPHA]= valueAsPercentages[ALPHA];
		} else {
			result[RED_INDEX]= getRed() / 255f;
			result[GREEN_INDEX]= getGreen() / 255f;
			result[BLUE_INDEX]= getBlue() / 255f;
			result[ALPHA]= getAlpha() / 255f;
		}
		return result;
	}

	/**
	 * Fill and return the specified array with the components of
	 * the receiver in its <code>ColorSpace</code>. Create,
	 * fill and return a new array if <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>float[]</code> containing the receiver's components
	 */
	public float[] getColorComponents(float[] floatArray) {
		float[] result = floatArray;
		if (result == null) {
			result = new float[3];
		}

		if (valueAsPercentages != null) {
			result[RED_INDEX]= valueAsPercentages[RED_INDEX];
			result[GREEN_INDEX]= valueAsPercentages[GREEN_INDEX];
			result[BLUE_INDEX]= valueAsPercentages[BLUE_INDEX];
		} else {
			result[RED_INDEX]= getRed() / 255f;
			result[GREEN_INDEX]= getGreen() / 255f;
			result[BLUE_INDEX]= getBlue() / 255f;
		}
		return result;
	}

	/**
	 * Answers the <code>ColorSpace</code> of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>ColorSpace</code> which is the color space of the receiver
	 *
	 */
	public ColorSpace getColorSpace() {
		return ColorSpace.getInstance(ColorSpace.CS_sRGB);
	}

	/**
	 * Answers the transparency of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the receiver's tranparency
	 *
	 * @see	<code>Paint</code>
	 */
	public int getTransparency() {
		if (valueAsPercentages != null) {
			float alpha= valueAsPercentages[ALPHA];
			if (alpha == 1.0)
				return Transparency.OPAQUE;
			else if (alpha == 0.0)
				return Transparency.BITMASK;
			else
				return Transparency.TRANSLUCENT;
		} else {
			int alpha= getAlpha();
			if (alpha == 255)
				return Transparency.OPAQUE;
			else if (alpha == 0)
				return Transparency.BITMASK;
			else
				return Transparency.TRANSLUCENT;
		}
	}

	/**
	 * Answer the peer representation for the color
	 */
	int getPeer() {
		if (colorHandle == 0) {
			colorHandle = colorCache.getColor(getRGB());
		}

		return colorHandle;
	}
}
