package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.awt.im.*;
import java.beans.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import java.util.List;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.metal.Handle;
import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.oti.awt.metal.widgets.Display;
import com.ibm.ugl.eswt.OS;

/**
 * Component is the abstract superclass of all widget
 * classes.  The term widget is used to describe
 * graphical user interface objects.
 * <p>
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public abstract class Component implements ImageObserver, Serializable
											,MenuContainer
{

	static final long serialVersionUID = -7644114512714619750L;

	// The key listener of the receiver
	transient ComponentListener componentListener;

	// The focus listener of the receiver
	transient FocusListener focusListener;

	// The key listener of the receiver
	transient KeyListener keyListener;

	// The mouse listener of the receiver
	transient MouseListener mouseListener;

	// The mouse motion listener of the receiver
	transient MouseMotionListener mouseMotionListener;

	PropertyChangeSupport changeSupport;

	transient InputMethodListener inputMethodListener;

	transient boolean enableInputMethodsCalled = false;

	transient MouseWheelListener mouseWheelListener;

	boolean focusTraversalKeysEnabled;

	// The focus traversal keys. These keys will generate focus traversal behavior
	// for Components for which focus traversal keys are enabled. If a value of null
	// is specified for a traversal key, this Component inherits that traversal key
	// from its parent. If all ancestors of this Component have null specified for
	// that traversal key, then the current KeyboardFocusManager's default traversal
	// key is used.
	Set[] focusTraversalKeys = new Set[] {null, null, null, null};

	// Tracks whether or not the default focusability of the Component has been
	// overridden.
	int isFocusTraversableOverridden = 0;

	boolean focusable;

	boolean ignoreRepaint;

	//	Determines which events are enabled.  If a event type is enabled,
	//	processEvent(AWTEvent) will be called when events of that type
	//	are dispatched by the receiver.  If there is a listener to a
	//	determined event type, processEvent(AWTEvent) will be called
	//	independently if that event type is enabled or not.
	//	see AWTEvent
	long eventMask;

	//	Determines whether Java 1.1 event model should be used
	//	by the receiver or not.
	boolean newEventsOnly;

	private static Hashtable nameCounters = new Hashtable();

	//	The nonlocalized name of the receiver
	String name;

	//	Determines whether the receiver's name was set by calling
	//	<code>setName()</code>
	boolean nameExplicitlySet = false;

	//	Determines whether the receiver is visible
	boolean visible;

	//	Determines whether the receiver is enabled
	boolean enabled;

	//	Determines whether the receiver is properly laid out
	private boolean valid;

	Vector popups = new Vector();

	//	The cursor displayed when the mouse pointer enters receiver
	Cursor cursor;

	//	The background color of the receiver
	Color background;

	//	The foreground color of the receiver
	private Color foreground;

	//	The font of the receiver
	private Font font;

	//	The locale of the receiver
 	Locale locale;

	//  The buttons which have been dragged since they were pressed
	transient int lastDraggedButtons = 0;

	//	The x coordinate of the receiver related to its parent
	int x;

	//	The y coordinate of the receiver related to its parent
	int y;

	//	The width of the receiver
	int width;

	//	The height of the receiver
	int height;

	// These fields are used to cache the minimum
	// and preferred sizes of a component while
	// it is valid.  When the component becomes
	// invalid the sizes are reset to null.
	// In addition to being used to improve performance,
	// these two fields are also required to pass
	// the TCK serialization tests.
	Dimension minSize;
	Dimension prefSize;

	int componentSerializedDataVersion;
	boolean isPacked;

	// The wrapper on the native widget
	transient int peer;

	transient boolean isDisplayable;

	// The parent of the receiver
 	transient Container parent;

	transient ArrayList availableGraphics = new ArrayList();
	transient HashSet activeGraphics = new HashSet();

	static Object TREE_LOCK = new Object();

	static final String componentListenerK = "componentL"; //$NON-NLS-1$
	static final String focusListenerK = "focusL"; //$NON-NLS-1$
	static final String keyListenerK = "keyL"; //$NON-NLS-1$
	static final String mouseListenerK = "mouseL"; //$NON-NLS-1$
	static final String mouseMotionListenerK = "mouseMotionL"; //$NON-NLS-1$
	static final String actionListenerK = "actionL"; //$NON-NLS-1$
	static final String containerListenerK = "containerL"; //$NON-NLS-1$
	static final String itemListenerK = "itemL"; //$NON-NLS-1$
	static final String adjustmentListenerK = "adjustmentL"; //$NON-NLS-1$
	static final String textListenerK = "textL"; //$NON-NLS-1$
	static final String ownedListenerK = "ownedL"; //$NON-NLS-1$
	static final String windowListenerK = "windowL"; //$NON-NLS-1$
	static final String mouseWheelListenerK = "mouseWheelL"; //$NON-NLS-1$
	static final String inputMethodListenerK = "inputMethodL"; //$NON-NLS-1$
	//static final String propertyChangeListenerK = "propertyChangeL"; //$NON-NLS-1$

	// These constants are for the acquire/release methods
	transient volatile int waitingWriters;
	transient Object rwLock;
	transient volatile HashMap readers;
	class MutableInteger {
		int value;
	}

	transient int redrawBlocks;

 transient private static GraphicsReaper reaper = new GraphicsReaper();

	/**
	 * Alignment constants used for layout.
	 *
	 * @see	#getAlignmentX()
	 * @see	#getAlignmentY()
	 */
	public static final float BOTTOM_ALIGNMENT = 1.0f;
	public static final float CENTER_ALIGNMENT = 0.5f;
	public static final float LEFT_ALIGNMENT = 0.0f;
	public static final float RIGHT_ALIGNMENT = 1.0f;
	public static final float TOP_ALIGNMENT = 0.0f;

	/**
	 * Constructs a new <code>Component</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	protected Component() {
		visible = true;
		enabled = true;
		valid = false;
		changeSupport = new PropertyChangeSupport(this);
		ignoreRepaint = false;
		focusable = isFocusableImpl();
		rwLock = new Object();
		readers = new HashMap();
		isDisplayable = false;
	}

	/**
	 * Answer the name of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is the receiver's name
	 */
	public String getName() {
		if (!nameExplicitlySet && (name == null)) {
//			name = classNonlocalizedName() + "-" + nameCounter++;
			String tempName = classNonlocalizedName();
			if (tempName != null) {
				name = _generateNewName(tempName);
			}
		}
		return name;
	}

	private String _generateNewName(String tempName) {
		Integer count = (Integer) nameCounters.get(tempName);
		int currentIndex = 0;
		if (count != null) {
			currentIndex = count.intValue();
		}
		nameCounters.put(tempName, new Integer(currentIndex+1));
		return tempName + currentIndex;
	}

	String classNonlocalizedName() {
		//return "component"; //$NON-NLS-1$
		return null; // Match RI behavior
	}

	/**
	 * Set the name of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	name	the name for the receiver
	 */
	public void setName(String newName) {
		nameExplicitlySet = true;
		name = newName;
	}

	/**
	 * Answer the parent of the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Container which is the receiver's parent
	 */
	public Container getParent() {
		return parent;
	}

	/**
	 * Answer the graphics configuration for this instance.
	 * If no graphics configuration has been assigned to this
	 * component, answer the graphics configuration for the
	 * top level component, else answer null.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>GraphicsConfiguration</code>
	 */
	public GraphicsConfiguration getGraphicsConfiguration() {
		if (getParent() != null) {
			return getParent().getGraphicsConfiguration();
		} else {
			return null;
		}
	}

	/**
	 * Answer an Object which can be synchronized when performing
	 * operations on the component's tree.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.Object to use as a lock
	 */
	public final Object getTreeLock() {
		// The assumption is that if we are in the AWT event loop,
		// return the real TreeLock.  If we aren't in the
		// AWT event loop, but are in the peer event loop, then
		// someone has already grabbed the treelock, so we don't
		// need to grab it again.  If we aren't in either
		// event loop, return the real TreeLock
		if (EventQueue.isDispatchThread()) return TREE_LOCK;
		if (isDisplayThread()) return new Object();
		return TREE_LOCK;
	}

	boolean isDisplayThread() {
		return ((ToolkitImpl)getToolkit()).isDisplayThread();
	}

	/**
	 * Answer the <code>Toolkit</code> for the receiver. The receiver's
	 * frame parent determines which toolkit that is.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Toolkit which is the receiver's toolkit
	 */
	public Toolkit getToolkit() {
		if (parent != null)
			return parent.getToolkit();

		return Toolkit.getDefaultToolkit();
	}

	/**
	 * Answer whether the receiver and its children do not
	 * need to be laid out.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver and its children are valid,
	 *				<code>false</code> if layout is required
	 */
	public boolean isValid() {
		return valid && isDisplayable();
	}

	/**
	 * Answer whether the receiver is displayable, which
	 * means it has an underlying widget created.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is displayable,
	 *				<code>false</code> otherwise
	 *
	 * @since JDK1.2
	 */
	public boolean isDisplayable() {
		return isDisplayable;
	}

	boolean isPeerNull(){
		return peer == 0;
	}

	/**
	 * Answer whether the receiver is visible.  If it is visible
	 * it will be showing on the screen if its parent is showing.
	 * Components that are not top level are created visible,
	 * regardless of whether they are showing on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is visible,
	 *				<code>false</code> otherwise
	 */
	public boolean isVisible() {
		return visible;
	}

	/**
	 * Answer whether the receiver is visible and its parent is
	 * showing.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is showing,
	 *				<code>false</code> otherwise
	 */
	public boolean isShowing() {
		return isVisible() && parent != null && parent.isShowing();
	}

	/**
	 * Answer whether double buffering is used to draw the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is double buffered,
	 *				<code>false</code> otherwise
	 */
	public boolean isDoubleBuffered() {
		return false;
	}

	/**
	 * Answer whether the receiver is enabled and will respond
	 * to events.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is enabled,
	 *				<code>false</code> otherwise
	 */
	public boolean isEnabled() {
		return enabled;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof MouseWheelEvent)
			return ((eventMask & AWTEvent.MOUSE_WHEEL_EVENT_MASK) != 0) || mouseWheelListener != null;
		else
		if (awtEvent instanceof MouseEvent)
			switch (awtEvent.getID()) {
			case MouseEvent.MOUSE_MOVED:
			case MouseEvent.MOUSE_DRAGGED:
				return ((eventMask & AWTEvent.MOUSE_MOTION_EVENT_MASK) != 0) || mouseMotionListener != null;
			default:
				return ((eventMask & AWTEvent.MOUSE_EVENT_MASK) != 0) || mouseListener != null;
			}
		else if (awtEvent instanceof InputMethodEvent)
			return ((eventMask & AWTEvent.INPUT_METHOD_EVENT_MASK) != 0) || inputMethodListener != null;
		else if (awtEvent instanceof KeyEvent)
			return ((eventMask & AWTEvent.KEY_EVENT_MASK) != 0) || keyListener != null;
		else if (awtEvent instanceof FocusEvent)
			return ((eventMask & AWTEvent.FOCUS_EVENT_MASK) != 0) || focusListener != null;
		else if (awtEvent instanceof ComponentEvent)
			return ((eventMask & AWTEvent.COMPONENT_EVENT_MASK) != 0) || componentListener != null;
		else
			return true;
	}

	/**
	 * Enable or disable the receiver according to
	 * the specified value.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		enabled	<code>true</code> if the receiver is to be enabled
	 *						<code>false</code> for disabling the receiver
	 */
	public void setEnabled(boolean enabled) {
		if (this.enabled == enabled) {
			return;
		}

		this.enabled = enabled;

		if ((enabled == false) && (hasFocus())) {
			removeFocus();
		}

		acquirePeerReadLock();
		try {
			if (peer != 0){
				OS.Control_SetEnabled(peer, enabled);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Enable the receiver so that it will respond
	 * to events.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 */
	public void enable() {
		setEnabled(true);
	}

	/**
	 * Enable or disable the receiver according to
	 * the specified value.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		enabled	<code>true</code> if the receiver is to be enabled
	 *						<code>false</code> for disabling the receiver
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 * @see			#setEnabled(boolean)
	 */
	public void enable(boolean enabled) {
		setEnabled(enabled);
	}

	/**
	 * Disable the receiver.  In this state a widget
	 * will typically not respond to events.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 * @see			#setEnabled(boolean)
	 */
	public void disable() {
		setEnabled(false);
	}

	/**
	 * Make the receiver visible according the specified value. If the
	 * receiver's parent is showing, the receiver will show on the screen.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @param	visible	<code>true</code> if the receiver is to be visible,
	 *					<code>false</code> otherwise
	 */
	public void setVisible(boolean visible) {
		if ((visible == false) && (containsFocus())) {
			removeFocus();
		}

		show(visible);
	}

	void setRedraw(boolean redraw) {
		if (redraw) {
			redrawBlocks--;
			if (redrawBlocks == 0) {
				acquirePeerReadLock();
				try {
					if (peer != 0) OS.Control_SetRedraw(peer, true);
				} finally {
					releasePeerReadLock();
				}
			}
		} else {
			redrawBlocks++;
			if (redrawBlocks == 1) {
				acquirePeerReadLock();
				try {
					if (peer != 0) OS.Control_SetRedraw(peer, false);
				} finally {
					releasePeerReadLock();
				}
			}
		}
	}

	/**
	 * Make the receiver visible. If the receiver's parent is showing,
	 * the receiver will show on the screen.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @deprecated	use <code>setVisible(boolean)</code> instead
	 *
	 * @see #setVisible(boolean)
	 */
	public void show() {
		_show(true);
	}

	/**
	 * Make the receiver visible according the specified value. If the
	 * receiver's parent is showing, the receiver will show on the screen.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @param	visible	<code>true</code> if the receiver is to be visible,
	 *					<code>false</code> otherwise
	 *
	 * @deprecated	use <code>setVisible(boolean)</code> instead
	 *
	 * @see #setVisible(boolean)
	 */
	public void show(boolean visible) {
		if (visible) {
			show();
		} else {
			hide();
		}
	}

	/**
	 * Hide the receiver so that it is no longer visible on the screen.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @deprecated	use <code>setVisible(boolean)</code> instead
	 * @see			#setVisible(boolean)
	 */
	public void hide() {
		_show(false);
	}

	void _show(final boolean visible) {
		if (isVisible() == visible) return;
		synchronized (getTreeLock()) {
			this.visible = visible;
			setVisibleImpl(visible, true);

			postEvent(new ComponentEvent(
				this,
				visible ? ComponentEvent.COMPONENT_SHOWN : ComponentEvent.COMPONENT_HIDDEN));

			if ((parent != null) && (parent.isValid())) parent.invalidate();
 		 }
	}

	/**
	 * Returns the foreground color of the widget which has been
	 * set explicitly via the API.  If no foreground color has been
	 * set with setForeground() then this API returns null.
	 */
	Color _getForeground() {
		if (foreground != null)
			return foreground;

		if (parent != null)
			return parent._getForeground();

		return null;
	}

	/**
	 * Answer the foreground <code>Color</code> for the
	 * receiver. If its foreground is not set, answer the
	 * foreground of the receiver's parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Color which is the foreground color
	 */
	public Color getForeground() {
		// If the foreground color was set by the user then this
		// color must be the one that the widget is using
		Color explicitForeground = _getForeground();
		if (explicitForeground != null) return explicitForeground;

		acquirePeerReadLock();
		try {
			// If there was no explicit foreground color and the peer
			// has been created then return whichever color the
			// platform defaults to for this widget
			if (peer != 0) {
				return new Color(OS.Control_GetForegroundRGB(peer));
			}
			else {
				// Lightweights always have (peer == 0), so they need
				// to get the color from their parent.
				if (parent != null) {
					return parent.getForeground();
				}
				else {
					return null;
				}
			}
		} finally {
			releasePeerReadLock();
		}

	}

	/**
	 * Set the foreground <code>Color</code> for the
	 * receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	color the foreground color for the receiver
	 */
	public void setForeground(Color color) {
		if (foreground == color) return;
		if ((color != null) && (color.equals(foreground))) return;

		firePropertyChange("foreground", foreground, color);
		this.foreground = color;

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				// Since the peer has been created the foreground color
				// should be inherited from the parent if the user just
				// changed this color to NULL.
				color = _getForeground();
				OS.Control_SetForeground(peer, color == null ? 0 : color.getPeer());
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer the background <code>Color</code> for the
	 * receiver. If its background is not set, answer the
	 * background of the receiver's parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Color which is the background color
	 */
	public Color getBackground() {
		// If the background color was set by the user then this
		// color must be the one that the widget is using.
		if (background != null) return background;

		acquirePeerReadLock();
		try {
			// If the user did not set the background color and the peer
			// has been created, then return whichever color the
			// platform defaults to for this widget.
			if (peer != 0) {
				return new Color(OS.Control_GetBackgroundRGB(peer));
			}
			else {
				// Lightweights always have (peer == 0), so they need
				// to get the color from their parent.
				if (parent != null) {
					return parent.getBackground();
				}
				else {
					return null;
				}
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Returns the background color of the widget which has been
	 * set explicitly via the API.  If no background color has been
	 * set with setBackground() then it returns the parent's
	 * background color.
	 */
	Color _getBackground() {
		if (background != null)
			return background;

		if (parent != null)
			return parent._getBackground();

		return null;
	}

	/**
	 * Set the background <code>Color</code> for the
	 * receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	color the background color for the receiver
	 */
	public void setBackground(Color color) {
		if (background == color) return;
		if ((color != null) && (color.equals(background))) return;

		firePropertyChange("background", background, color);
		this.background = color;

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				// Since the peer has been created the background color
				// should be inherited from the parent if the user just
				// changed this color to NULL.
				color = _getBackground();
				OS.Control_SetBackground(peer, color == null ? 0 : color.getPeer());
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer the <code>Font</code> used by the receiver.  If
	 * a <code>Font</code> has not been set obtain the font from
	 * the parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Font
	 */
	public Font getFont() {
		// If the font was set by the user then this
		// font must be the one that the widget is using
		Font explicitFont = _getFont();
		if (explicitFont != null) return explicitFont;

		acquirePeerReadLock();
		try {
			// If there was no explicit font and the peer
			// has been created then return whichever font the
			// platform defaults to for this widget
			if (peer != 0) {
				int fontHandle = OS.Control_GetDefaultFont(peer);
				if (fontHandle == 0) throw new RuntimeException("Control_GetDefaultFont returned 0");

				return new Font(new FontPeer(fontHandle));
			}
			else {
				// Lightweights always have (peer == 0), so they need
				// to get the font from their parent.
				if (parent != null) {
					return parent.getFont();
				}
				else {
					return null;
				}
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Returns the font of the widget which has been set explicitly
	 * via the API.  If no font has been set with setFont() then this
	 * null will be returned.
	 */
	Font _getFont() {
		if (font != null)
			return font;

		if (parent != null)
			return parent._getFont();

		return null;
	}

	/**
	 * Set the <code>Font</code> used by the receiver.  If
	 * the specified <code>Font</code> is <code>null</code> obtain the
	 * font from the parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	componentFont	the font for the receiver
	 */
	public void setFont(Font componentFont) {
		if (font == componentFont) {
			return;
		}
		firePropertyChange("font", font, componentFont);
		font = componentFont;
		acquirePeerReadLock();
		try {
			if (peer != 0) {
				componentFont = _getFont();
				FontPeer fontPeer = componentFont == null ? null : componentFont.getPeer();
				OS.Control_SetFont(peer, (fontPeer == null ? 0 : fontPeer.fPeerHandle));
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer the <code>Locale</code> for the receiver or if
	 * one was not specified, answer the parent's.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.util.Locale
	 *
	 * @throws		java.awt.IllegalComponentStateException
	 */
	public Locale getLocale() {
		if (locale == null) {
			if (parent == null)
				throw new IllegalComponentStateException();

			return parent.getLocale();
		}
		return locale;
	}

	/**
	 * Set the Locale for this component.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	locale
	 */
	public void setLocale(Locale locale) {
		this.locale = locale;
	}

	/**
	 * Answer the <code>ColorModel</code> for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.image.ColorModel
	 */
	public ColorModel getColorModel() {
		return getToolkit().getColorModel();
	}

	/**
	 * Answer the <code>Point</code> representing the top-left
	 * corner of the receiver relative to the top-left corner
	 * of its parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Point with the receiver's x, y
	 */
	public Point getLocation() {
		return getLocation(null);
	}

	/**
	 * Fill and return the specified <code>Point</code> with the top-left
	 * corner of the receiver relative to the top-left corner
	 * of its parent.  Create, fill and return a new <code>Point</code> if
	 * <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the point to fill with the receiver's location
	 *
	 * @return		a java.awt.Point with the receiver's x, y
	 *
	 * @since		JDK1.2
	 */
	public Point getLocation(Point point) {
		if (point == null) point = new Point();

		point.x = getX();
		point.y = getY();
		return point;
	}

	/**
	 * Answer the <code>Point</code> representing the top-left
	 * corner of the receiver relative to the top-left corner
	 * of the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Point with the receiver's x, y relative to the screen
	 *
	 * @throws	IllegalComponentStateException	if the receiver is not showing in the
	 *											screen
	 */
	public Point getLocationOnScreen() {
		if (!isShowing())
			throw new IllegalComponentStateException();

		Point point = new Point(0, 0);
		Component temp = this;
		while (temp != null && temp.isLightweight()) {
			point.x += temp.getX();
			point.y += temp.getY();
			temp = temp.parent;
		}
		if (temp != null) {
			point = temp.toDisplay(point);
		}
		return new Point(point.x, point.y);
	}

	/**
	 * Converts the specified point from component relative coordinates
	 * to display relative coordinates.
	 *
	 * @param point the point to be translated (must not be null)
	 * @return the point converted to display relative coordinates
	 */
	Point toDisplay(Point point) {
		Component control = this;
		Point displayPoint = new Point(point);

		while (control != null) {
			displayPoint.x += control.getX();
			displayPoint.y += control.getY();
			control = control.parent;
		}
		return displayPoint;
	}

	/**
	 * Answer the <code>Point</code> representing the top-left
	 * corner of the receiver relative to the top-left corner
	 * of its parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Point with the receiver's x, y
	 *
	 * @deprecated	use <code>getLocation()</code> instead
	 *
	 * @see #getLocation
	 */
	public Point location() {
		return getLocation();
	}

	/**
	 * Set the components x and y coordinates to the specified <code>point</code>,
	 * and move the component to that location relative to the parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	point	the location for the receiver
	 */
	public void setLocation(Point point) {
		setLocation(point.x, point.y);
	}

	/**
	 * Set the components x and y coordinates to the specified values,
	 * and move the component to that location relative to the parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x coordinate to be used
	 * @param	y	the y coordinate to be used
	 *
	 * @deprecated	use <code>setLocation(int, int)</code> instead
	 *
	 * @see #setLocation(int,int)
	 */
	public void move(int x, int y) {
		setBounds(x, y, width, height);
	}

	/**
	 * Set the components x and y coordinates to the specified values,
	 * and move the component to that location relative to the parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x coordinate to be used
	 * @param	y	the y coordinate to be used
	 */
	public void setLocation(int x, int y) {
		move(x, y);
	}

//	Point translateFromPeerCoords(Point point) {
//		Insets insets = getNativeParent().getInsets();
//		point.x += insets.left;
//		point.y += insets.top;
//		return point;
//	}

	boolean shouldGenerateKeyTypedEvent(char keyChar) {
		if (0 == keyChar) return false;

		if (Character.isDefined(keyChar) && !Character.isISOControl(keyChar)) {
			return true;
		}

		switch ((int) keyChar) {
			case KeyEvent.VK_ENTER :
			case KeyEvent.VK_ESCAPE :
			case KeyEvent.VK_TAB :
			case KeyEvent.VK_DELETE :
			case KeyEvent.VK_BACK_SPACE :
				return true;
			default :
				break;
		}

		return false;
	}

	Point translateToPeerCoords(Point point) {
		Container temp = parent;
		while (temp != null && temp.isLightweight()) {
			point.x += temp.getX();
			point.y += temp.getY();
			temp = temp.parent;
		}

		return point;
	}

	/**
	 * Answer a <code>Dimension</code> representing the size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's size
	 */
	public Dimension getSize() {
		return getSize(null);
	}

	/**
	 * Fill and return the specified <code>Dimension</code> with the size of
	 * the receiver.  Create, fill and return a new <code>Dimension</code> if
	 * <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		dimension	the dimension to fill with the receiver's size.
	 *
	 * @return		a java.awt.Dimension with the receiver's width, height
	 *
	 * @since		JDK1.2
	 */
	public Dimension getSize(Dimension dimension) {
		if (dimension == null) dimension = new Dimension();

		dimension.width = width;
		dimension.height = height;
		return dimension;
	}

	/**
	 * Answer a <code>Dimension</code> representing the size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's size
	 *
	 * @deprecated	use <code>getSize()</code> instead
	 *
	 * @see #getSize
	 */
	public Dimension size() {
		return getSize();
	}

	/**
	 * Resize the receiver to the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	width	the width to resize the receiver to.
	 * @param	height	the height to resize the receiver to.
	 */
	public void setSize(int width, int height) {
		resize(width,height);
	}

	/**
	 * Resize the receiver to the specified <code>Dimension</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	dimension	the new size for the receiver
	 */
	public void setSize(Dimension dimension) {
		resize(dimension);
	}

	/**
	 * Resize the receiver to the specified <code>Dimension</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	dimension	the new size for the receiver
	 *
	 * @deprecated	use <code>setSize(Dimension)</code> instead.
	 *
	 * @see #setSize(Dimension)
	 */
	public void resize(Dimension dimension) {
		setSize(dimension.width, dimension.height);
	}

	/**
	 * Resize the receiver to the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	width	the width to resize the receiver to.
	 * @param	height	the height to resize the receiver to.
	 *
	 * @deprecated	use <code>setSize(int,int)</code> instead.
	 *
	 * @see #setSize(int,int)
	 */
	public void resize(int width, int height) {
		setBounds(getX(), getY(), width, height);
	}

	/**
	 * Answer a <code>Rectangle</code> representing the
	 * receiver's bounds.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Rectangle with the receiver's x, y, width, height
	 *
	 * @deprecated	use <code>getBounds()</code> instead
	 * @see			#getBounds()
	 */
	public Rectangle bounds() {
		return getBounds();
	}

	/**
	 * Move and resize the receiver to the specified
	 * <code>Rectangle</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	rectangle		the new location and size for the receiver.
	 */
	public void setBounds(Rectangle rectangle) {
		setBounds(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
	}

	/**
	 * Answer the x coordinate of the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a <code>int</code> which is the receiver's x coordinate
	 *
	 * @since JDK1.2
	 */
	public int getX() {
		return x;
	}

	/**
	 * Answer the y coordinate of the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a <code>int</code> which is the receiver's y coordinate
	 *
	 * @since JDK1.2
	 */
	public int getY() {
		return y;
	}

	/**
	 * Answer the height of the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return		an <code>int</code>	which is the receiver's height
	 *
	 * @since JDK1.2
	 */
	public int getHeight() {
		return height;
	}

	/**
	 * Answer the width of the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a <code>int</code> which is the receiver's width
	 *
	 * @since JDK1.2
	 */
	public int getWidth() {
		return width;
	}

	/**
	 * Answer a <code>Rectangle</code> representing the
	 * receiver's bounds.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Rectangle with the receiver's x, y, width, height
	 *
	 * @see #setBounds(int, int, int, int)
	 * @see #getLocation()
	 * @see #getSize()
	 */
	public Rectangle getBounds() {
		return getBounds(null);
	}

	/**
	 * Fill and return the specified <code>Rectangle</code> with the bounds of
	 * the receiver.  Create, fill and return a new <code>Rectangle</code> if
	 * <code>null</code> is passed in.
	 * <p>
	 * This method can be used to avoid allocating new objects by passing
	 * in the returning value.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		rectangle	the rectangle to fill with the receiver's bounds.
	 *
	 * @return		a java.awt.Rectangle with the receiver's x, y, width, height
	 *
	 * @since JDK1.2
	 */
	public Rectangle getBounds(Rectangle rectangle) {
		if (rectangle == null) rectangle = new Rectangle();

		rectangle.x = x;
		rectangle.y = y;
		rectangle.width = width;
		rectangle.height = height;
		return rectangle;
	}

	/**
	 * Move and resize the receiver to the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x		the x coordinate to move the receiver to.
	 * @param	y		the y coordinate to move the receiver to.
	 * @param	width	the width to resize the receiver to.
	 * @param	height	the height to resize the receiver to.
	 *
	 * @deprecated	use <code>setBounds(int,int,int,int)</code> instead.
	 *
	 * @see #setBounds(int,int,int,int)
	 */
	public void reshape(final int x, final int y, final int width, final int height) {
	//	synchronized (this) {
			reshapeImpl(x,y,width,height);
	//	}
	}

	void reshapeImpl(final int x, final int y, final int width, final int height) {
		synchronized (getTreeLock()) {
			_setBounds(x, y, width, height);
		}
	}

	/**
	 * Move and resize the receiver to the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x		the x coordinate to move the receiver to.
	 * @param	y		the y coordinate to move the receiver to.
	 * @param	width	the width to resize the receiver to.
	 * @param	height	the height to resize the receiver to.
	 */
	public void setBounds(int x, int y, int width, int height) {
		reshape(x,y,width,height);
	}

	void _setBounds(int x, int y, int width, int height) {
		boolean moved = getX() != x || getY() != y;
		boolean resized = getWidth() != width || getHeight() != height;

		if (moved || resized) {
			setBoundsImpl(x, y, width, height, true);

			if (resized) {
				invalidate();
			} else if (moved) {
				if (parent != null)
					parent.invalidate();
			}

			_postEventsOnResize(moved, resized);
		}
	}

	void _postEventsOnResize(boolean moved, boolean resized) {
		if (moved)
			postEvent(new ComponentEvent(this, ComponentEvent.COMPONENT_MOVED));
		if (resized)
			postEvent(new ComponentEvent(this, ComponentEvent.COMPONENT_RESIZED));
	}

	void setBoundsImpl(int x, int y, int width, int height, boolean repaint) {
		synchronized (getTreeLock()) {
			int oldX = getX();
			int oldY = getY();
			int oldWidth = getWidth();
			int oldHeight = getHeight();

			// resize and move events are generated differently for
			// Window vs. non-Window components, so special case caching
			// the bounds values
			cacheBounds(x, y, width, height);

			if (isLightweight() ) {
				if (isShowing() && repaint) {
					parent.repaint(oldX, oldY, oldWidth, oldHeight);
					parent.repaint(x, y, width, height);
				}
			} else {
				acquirePeerReadLock();
				try {
					_setPeerBounds(x, y, width, height);
				} finally {
					releasePeerReadLock();
				}
			}
		}
	}

	/**
	 * Answer whether the receiver is opaque.  Opaque widgets are ones
	 * which draw all pixels in their visible area so that you can not
	 * see through them.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is opaque,
	 *				<code>false</code> otherwise
	 * @since JDK1.2
	 */
	public boolean isOpaque() {
		return false;
	}

	/**
	 * Answer whether the receiver is not written to use an underlying
	 * widget but rather drawn on its heavyweight parent.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is lightweight,
	 *				<code>false</code> otherwise
	 * @since JDK1.2
	 */
	public boolean isLightweight() {
		return isDisplayable && isPeerNull();
	}

	Container getNativeParent() {
		Container temp = parent;
		while (temp != null && temp.isLightweight())
			temp = temp.getParent();
		return temp;
	}

	Dimension getPreferredSizeImpl() {
		return getMinimumSizeImpl();
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	public Dimension getPreferredSize() {
		synchronized (getTreeLock()) {
			// Only return the cached perfSize field when
			// the component is valid because classes
			// might skip calling invalidate when the component
			// is already invalid.  Skipping invalidate()
			// results in the cached value not being reset, so
			// it is only safe to return the cached value if
			// the component is valid.
			if (prefSize != null && valid) return prefSize;
		}

		// Call old method so that older applications
		// continue to work
		Dimension tempSize = preferredSize();
		prefSize = tempSize;
		return tempSize;
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Dimension which is the receiver's preferred size
	 *
	 * @deprecated	use <code>getPreferredSize()</code> instead
	 *
	 * @see #getPreferredSize
	 */
	public Dimension preferredSize() {
		return getPreferredSizeImpl();
	}

	/**
	* Answer a <code>Dimension</code> representing the minimum
	* size of the receiver.
	*
	* @author      IBM
	* @version		initial
	*
	* @return	a java.awt.Dimension which is the minimum size of the receiver
	*/
	public Dimension getMinimumSize() {
		synchronized (getTreeLock()) {
			// Only return the cached minSize field when
			// the component is valid because classes
			// might skip calling invalidate when the component
			// is already invalid.  Skipping invalidate()
			// results in the cached value not being reset, so
			// it is only safe to return the cached value if
			// the component is valid.
			if (minSize != null && valid) return minSize;
		}

		// Call old method so that older applications
		// continue to work
		Dimension tempSize = minimumSize();
		minSize = tempSize;
		return tempSize;
	}

	/**
	* Provides common code for both getMinimumSize() and getPreferredSize() to call.
	* Prevents both method to return the same thing when one method is override.
	*
	* @author      IBM
	* @version		initial
	*
	* @return	a java.awt.Dimension which is the minimum size of the receiver
	*/
	Dimension getMinimumSizeImpl() {
		acquirePeerReadLock();
		try {
			// Necessary to pass the TCK.
			if (!isDisplayable()) return getSize();
			// but if I exist, and I'm a lightweight, then return a minimal size
			// A very small minimum is need to not mess up layouts, which use
			// minimum/preferred sizes in its calculations.
			if (peer == 0) {
				return new Dimension(1,1);
			}

			int[] pointArray = OS.Control_GetMinimumSize(peer);
			return new Dimension(pointArray[OS.INDEX_X],pointArray[OS.INDEX_Y]);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 *
	 * @deprecated	use <code>getMinimumSize()</code> instead
	 *
	 * @see #getMinimumSize
	 */
	public Dimension minimumSize() {
		return getMinimumSizeImpl();
	}

	/**
	 * Answer a <code>Dimension</code> representing the maximum
	 * size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the maximum size of the receiver
	 */
	public Dimension getMaximumSize() {
		return new Dimension(Short.MAX_VALUE, Short.MAX_VALUE);
	}

	/**
	 * Answer the preferred horizontal alignment for
	 * the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getAlignmentX() {
		return CENTER_ALIGNMENT;
	}

	/**
	 * Answer the preferred vertical alignment for
	 * the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getAlignmentY() {
		return CENTER_ALIGNMENT;
	}

	/**
	 * If the receiver has a <code>LayoutManager</code> ask
	 * it to layout the receiver's children.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void doLayout() {
	}

	/**
	 * If the receiver has a <code>LayoutManager</code> ask
	 * it to layout the receivers children.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @deprecated	use <code>doLayout()</code> instead
	 *
	 * @see #doLayout
	 */
	public void layout() {
	}

	/**
	 * Make sure that the receiver and all its children if
	 * applicable are properly laid out.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void validate() {
		// Avoid getting the tree lock if the
		// widget is already valid
		if (valid) return;

		synchronized (getTreeLock()) {
			validateTree();
		}
	}

	void validateTree() {
		valid = true;
	}

	void callValidateTree() {
		validate();
	}

	/**
	 * Mark this component and all its parents to
	 * require that they be laid out.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void invalidate() {
		synchronized (getTreeLock()) {
			valid = false;

			minSize = null;
			prefSize = null;

			if (parent != null && parent.isValid())
				parent.invalidate();
		}
	}

	/**
	 * Answer a <code>Graphics</code> used for drawing
	 * on the receiver or <code>null</code>.  If the component
	 * is a lightweight, the graphics of the first heavyweight
	 * parent is returned.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Graphics for drawing
	 */
	Graphics getGraphicsImpl(Component c){
		if (isLightweight()) {
			if (parent != null) {
				Graphics g = parent.getGraphicsImpl(c);
				if (g != null) {
					return g;
				}
			}
			return null;
		}
		acquirePeerReadLock();
		try {
			if (peer == 0) return null;

			Graphics graphics;
			//if both components are the same, that means getGraphics was
			//called on a heavy weight; therefore, we need to call GraphicsACImpl
			if (this == c) {
				graphics = new GraphicsACImpl(this);
			} else {
				graphics = new LightweightGraphicsACImpl(this, c);
			}
			return graphics;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Graphics</code> used for drawing
	 * on the receiver or <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Graphics for drawing
	 */
	public Graphics getGraphics() {
		return getGraphicsImpl(this);
	}

	void newGraphicsHandle(Handle graphicsHandle) {
		synchronized (activeGraphics) {
			synchronized (availableGraphics) {
				if (availableGraphics.size() == 0) {
					graphicsHandle.value = OS.Control_NewGraphics(peer);
				} else {
					Handle handle = (Handle) availableGraphics.remove(0);
					graphicsHandle.value = handle.value;
				}
			}

			activeGraphics.add(graphicsHandle);
		}
	}

	void releaseGraphics(Handle graphicsHandle) {
		synchronized (activeGraphics) {
			Handle copy;

			// Guard against disposing the peer while a Graphic's peer
			// call is still executing
			synchronized (graphicsHandle) {
				if (graphicsHandle.value == 0) return;
				activeGraphics.remove(graphicsHandle);

				copy = new Handle();
				copy.value = graphicsHandle.value;
				graphicsHandle.value = 0;
			}

			synchronized (availableGraphics) {
				availableGraphics.add(copy);
				if (availableGraphics.size() == 1) {
					reaper.addComponent(this);
				}
			}
		}
	}

	/**
	 * Answer the <code>FontMetrics</code> for the specified font.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		font	the font to get the font metrics for
	 *
	 * @return		a java.awt.FontMetrics for the specified font
	 */
	public FontMetrics getFontMetrics(Font font) {
		return ((ToolkitImpl) getToolkit())._getFontMetrics(font);
	}

	/**
	 * Set the <code>Cursor</code> used by the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	cursor	the new cursor for the receiver or <code>null</code>
	 */
	public synchronized void setCursor(final Cursor cursor) {
		if (this.cursor == cursor) {
			return;
		}

		this.cursor = cursor;
		_setCursor(cursor);
	}

	void _setCursor(final Cursor cursor) {
		Component nativeComponent = this;
		if (0 == nativeComponent.peer) {
			nativeComponent = getNativeParent();
		}

		if (null != nativeComponent) {
			nativeComponent.acquirePeerReadLock();
			try {
				if (0 != nativeComponent.peer) {
					OS.Control_SetCursor(nativeComponent.peer, cursor == null ? 0 : cursor.getPeer());
				}
			} finally {
				nativeComponent.releasePeerReadLock();
			}
		}
	}

	void resetCursor(Component compCursor) {
		class ResetCursorEvent extends AWTEvent implements ActiveEvent {
			Component compCursor;
			public ResetCursorEvent(Object source, int id, Component compCursor) {
				super(source, id);
				this.compCursor = compCursor;
			}
			public void dispatch() {
//				syncExec(new Runnable() {
//					public void run() {
						Component comp = (Component)getSource();
						Cursor cursor = compCursor.getCursor();
						comp.acquirePeerReadLock();
						try {
							if (comp.peer != 0)
								OS.Control_SetCursor(comp.peer, cursor == null ? 0 : cursor.getPeer());
						} finally {
							comp.releasePeerReadLock();
						}
//					}
//				});
			}
		}
		postEvent(new ResetCursorEvent(this, 0, compCursor));
	}

	/**
	 * Answer the <code>Cursor</code> used by the receiver.
	 * If not specified, answer parents, and noone
	 * specifies it, return the default.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Cursor
	 */
	public Cursor getCursor() {
		if (cursor != null)
			return cursor;

		if (parent != null) {
			// possibly don't need the peer check at the Component level
			return parent.getCursor();
		}

		return _getDefaultCursor();
	}

	Cursor _getDefaultCursor() {
		return Cursor.getDefaultCursor();
	}

	Window getWindow() {
		if (null == parent) return null;
		return parent.getWindow();
	}

	/**
	 * Called when regions of this component need to be
	 * redrawn.  The graphics clip region represents the damaged area.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void paint(Graphics graphics) {
	}

	/**
	 * Called when the receiver needs to be redrawn.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void update(Graphics graphics) {
//		if (isShowing()) {
//			if (!isLightweight()) {
//				graphics.clearRect(0, 0, getWidth(), getHeight());
//			}
//			paint(graphics);
//		}
		if ((this instanceof Frame)
			|| (this instanceof Panel)
			|| (this instanceof Canvas)
			|| (this instanceof Dialog)
			|| (this instanceof Window)) {
			graphics.setColor(getBackground());
			graphics.fillRect(0, 0, getWidth(), getHeight());
			graphics.setColor(getForeground());
		}
		paint(graphics);
	}

	/**
	 * Called when this component and all of its children need to be
	 * redrawn.  The graphics clip region represents the total area of
	 * this component.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void paintAll(Graphics graphics) {
		paint(graphics);
	}

	/**
	 * Repaint all of the receiver, <code>update()</code> will be called
	 * as a result.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void repaint() {
		repaint(0);
	}

	/**
	 * Repaint all of the receiver, <code>update()</code> will be called
	 * by the specified maximum time as a result.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		time	the maximum time allow before updating the receiver.
	 */
	public void repaint(long time) {
		repaint(0, 0, 0, getWidth(), getHeight());
	}

	/**
	 * Repaint the portion of the receiver specified, <code>update()</code>
	 * will be called as a result.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x 		the x coordinate of the area to update
	 * @param	y 		the y coordinate of the area to update
	 * @param	width 	the widht of the area to update
	 * @param	height	the height of the area to update
	 */
	public void repaint(int x, int y, int width, int height) {
		repaint(0, x, y, width, height);
	}

	/**
	 * Repaint the portion of the receiver specified, <code>update()</code>
	 * will be called by the specified maximum time as a result.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	time	the maximum time allow before updating the receiver.
	 * @param	x 		the x coordinate of the area to update
	 * @param	y 		the y coordinate of the area to update
	 * @param	width 	the widht of the area to update
	 * @param	height	the height of the area to update
	 */
	public void repaint(long time, int x, int y, int width, int height) {
		if (isLightweight()) {
			if (null == parent) return;

			int thisX = getX();
			int thisY = getY();
			int thisWidth = getWidth();
			int thisHeight = getHeight();
			int transX = x + thisX;
			int transY = y + thisY;
			int left = transX > thisX ? transX : thisX;
			int top = transY > thisY ? transY : thisY;
			int x1 = transX + width;
			int x2 = thisX + thisWidth;
			int y1 = transY + height;
			int y2 = thisY + thisHeight;
			int right = x1 < x2 ? x1 : x2;
			int bottom = y1 < y2 ? y1 : y2;
			parent.repaint(time, left, top, right - left, bottom - top);
		} else {
			acquirePeerReadLock();
			try {
				if (peer != 0)
					postEvent(new PaintEvent(this, PaintEvent.UPDATE, new Rectangle(x, y, width, height)));
			} finally {
				releasePeerReadLock();
			}
		}
	}

	/**
	 * Print this component on the specfied graphics.
	 * The graphics clip region represents what should be printed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void print(Graphics graphics) {
		paint(graphics);
	}

	/**
	 * Print this component and all of its children on the specified
	 * graphics.  The graphics clip region represents the total area of
	 * this component.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void printAll(Graphics graphics) {
		print(graphics);
	}

	/**
	 * Called to deliver image loading information.
	 * Part of the ImageObserver interface.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	image
	 * @param	flags
	 * @param	x
	 * @param	y
	 * @param	width
	 * @param	height
	 */
	public boolean imageUpdate(Image image, int flags, int x, int y, int width, int height) {
		if (flags == FRAMEBITS || flags == ALLBITS) {
			Graphics g = getGraphics();
			if (g == null) return true;
			g.setClip(0, 0, getWidth(), getHeight());
			repaint();
			g.dispose();
		}
		return true;
	}

	/**
	 * Create an image from a <code>ImageProducer</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		imageProducer
	 *
	 * @return		a java.awt.Image
	 */
	public Image createImage(ImageProducer imageProducer) {
		return getToolkit().createImage(imageProducer);
	}

	/**
	 * Create a new image with the specified width and height.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		width	the width for the new image
	 * @param		height	the height for the new image
	 *
	 * @return		a java.awt.Image
	 */
	public Image createImage(int width, int height) {
		return createImage(width, height, getBackground());
	}

	Image createImage(int width, int height, Color fillColor) {
		if (!isDisplayable()) return null;
		if ((width <= 0) || (height <= 0)) {
			throw new IllegalArgumentException("Width ("+width+") and height ("+height+") must be > 0"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return new ImageImpl(this, width, height, fillColor);
	}

	/**
	 * Starts the loading of an image's data.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	image
	 * @param	imageObserver
	 */
	public boolean prepareImage(Image image, ImageObserver imageObserver) {
		return prepareImage(image, -1, -1, imageObserver);
	}

	/**
	 * Starts the loading of an image's data.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	image
	 * @param	width	the width of the image
	 * @param	height	the height of the image
	 * @param	imageObserver
	 */
	public boolean prepareImage(Image image, int width, int height, ImageObserver imageObserver) {
		return getToolkit().prepareImage(image, width, height, imageObserver);
	}

	/**
	 * Cache the bounds values for the Component.  These values
	 * will be checked when the resize/move events are received and
	 * if the values have not changed, the events will not be
	 * processed.
	 */
	void cacheBounds(int x, int y, int width, int height) {
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
	}

	/**
	 * Answer the ImageObserver Image Update Status Bits corresponding to this image.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		image	the image
	 * @param		imageObeserver
	 *
	 * @return		an <code>int</code>
	 */
	public int checkImage(Image image, ImageObserver imageObserver) {
		return checkImage(image, -1, -1, imageObserver);
	}

	/**
	 * Answer the ImageObserver Image Update Status Bits corresponding to this image.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		image	the image
	 * @param		width	the width of the image
	 * @param		height	the height of the image
	 * @param		imageObeserver
	 *
	 * @return		an <code>int</code>
	 */
	public int checkImage(Image image, int width, int height, ImageObserver imageObserver) {
		return getToolkit().checkImage(image, width, height, imageObserver);
	}

	/**
	 * Answer whether the receiver contains the x,y location relative
	 * to its coordinates.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x	the x coordinate to test for
	 * @param		y	the y coordinate to test for
	 *
	 * @return		<code>true</code> if the receiver contains the location
	 *				specified, <code>false</code> otherwise
	 */
	public boolean contains(int x, int y) {
		return x >= 0 && y >= 0 && x < getWidth() && y < getHeight();
	}

	/**
	 * Answer whether the receiver contains the point relative
	 * to its coordinates.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the <code>Point</code> to test for
	 *
	 * @return		<code>true</code> if the receiver contains the location
	 *				specified, <code>false</code> otherwise
	 */
	public boolean contains(Point point) {
		return contains(point.x, point.y);
	}

	/**
	 * Answer whether the receiver contains the x,y location relative
	 * to its coordinates.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x	the x coordinate to test for
	 * @param		y	the y coordinate to test for
	 *
	 * @return		<code>true</code> if the receiver contains the location
	 *				specified, <code>false</code> otherwise
	 *
	 * @deprecated	use <code>contains(int,int)</code> instead
	 * @see			#contains(int, int)
	 */
	public boolean inside(int x, int y) {
		return x >= 0 && y >= 0 && x < getWidth() && y < getHeight();
	}

	/**
	 * Answer the <code>Component</code> at the specified
	 * point.  In the case of a <code>Container</code> if a child
	 * component contains the specified point the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the point relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 */
	public Component getComponentAt(Point point) {
		return getComponentAt(point.x, point.y);
	}

	/**
	 * Answer the <code>Component</code> at the specified
	 * x and y location.  In the case of a <code>Container</code> if a child
	 * component contains the specified x and y the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the x coordinate relative to the receiver
	 * @param		y		the y coordinate relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 */
	public Component getComponentAt(int x, int y) {
		if (contains(x, y))
			return this;

		return null;
	}

	/**
	 * Answer the <code>Component</code> at the specified
	 * x and y location.  In the case of a <code>Container</code> if a child
	 * component contains the specified x and y the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the x coordinate relative to the receiver
	 * @param		y		the y coordinate relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @deprecated	use <code>getComponentAt(int, int)</code> instead
	 *
	 * @see #getComponentAt(int, int)
	 */
	public Component locate(int x, int y) {
		return getComponentAt(x, y);
	}

	/**
	 * Post an event to the queue.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		event	the event to deliver
	 *
	 * @deprecated	use <code>dispatchEvent()</code> instead
	 */
	public void deliverEvent(Event event) {
		postEvent(event);
	}

	/**
	 * Dispatch the specified event to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	awtEvent	the event to be dispatched
	 */
	public final void dispatchEvent(AWTEvent awtEvent) {
		// Don't dispatch the event until the TreeLock can be acquired.
		// This primarily prevents events from being processed while
		// widgets are being added or removed. This should also help to
		// reduce event traffic since remove processing will purge any
		// pending events for widgets that are being removed.
		// *** THIS DOESN'T WORK AS IT CAN RESULT IN A DEADLOCK ***
//		synchronized (getTreeLock()) {
			dispatchEventImpl(awtEvent);
//		}
	}

	void dispatchEventImpl(AWTEvent awtEvent) {
		if (awtEvent instanceof PaintEvent) {
			Rectangle damageRect = ((PaintEvent) awtEvent).getUpdateRect();
			if (damageRect.width <= 0 || damageRect.height <= 0)
				return;
			Graphics g;
			synchronized(getTreeLock()) {
				g = getGraphics();
				if (g == null)
					return;

				g.clipRect(damageRect.x, damageRect.y, damageRect.width, damageRect.height);
			}
			switch (awtEvent.getID()) {
				case PaintEvent.PAINT:
					paint(g);
					break;
				case PaintEvent.UPDATE:
					update(g);
					break;
				default:
			}
			g.dispose();
		}
	    boolean continueProcessing = true;
		// If input methods are enabled, give the input context first chance at dispatching the event
		if ((!enableInputMethodsCalled || ((eventMask & AWTEvent.INPUT_METHOD_EVENT_MASK) != 0)) && (eventMask & AWTEvent.KEY_EVENT_MASK) != 0) {
			getInputContext().dispatchEvent(awtEvent);
			if (awtEvent.isConsumed()) return;
		}

		// all Key, Focus, and Window events that are related to focus should be dispatched to the
		// KeyboardFocusManager first, which will then dispatch them back
		int eventType = awtEvent.getID();
		boolean isKeyEvent = awtEvent instanceof KeyEvent;
		boolean isFocusManagerEvent = isKeyEvent ||
			awtEvent instanceof FocusEvent ||
			(awtEvent instanceof WindowEvent &&
			        ((eventType == WindowEvent.WINDOW_GAINED_FOCUS) ||
					 (eventType == WindowEvent.WINDOW_LOST_FOCUS) ||
					 (eventType == WindowEvent.WINDOW_CLOSED) ||
					 (eventType == WindowEvent.WINDOW_ACTIVATED) ||
					 (eventType == WindowEvent.WINDOW_DEACTIVATED)));
		KeyboardFocusManager kbMgr = KeyboardFocusManager.getCurrentKeyboardFocusManager();
		if (isFocusManagerEvent) {
		    if (awtEvent.focusManagerIsDispatching)  {
		    	if (KeyboardFocusManager.getDebug()) System.out.println("CMPO DISPATCH " + awtEvent.toString());
		    }
		    if (!awtEvent.focusManagerIsDispatching)  {
				// if the KeyboardFocusManager dispatchEvent returns false, then the AWT
				// event dispatcher will attempt to dispatch the event itself.
				boolean processed;
			    processed = kbMgr.dispatchEvent(awtEvent);
			    if (processed) return;
			    // The keyboard focus manager will always return true when key events
			    // are processed, except in the case where key events are ignored due
			    // to no global focus owner.
			    if (awtEvent instanceof KeyEvent) continueProcessing = false;
		    }
		}

		// Do not process key events if no focus owner exists
		if (continueProcessing) {

			if (awtEvent.getID() == KeyEvent.KEY_PRESSED &&
					awtEvent instanceof KeyEvent)
			{
				KeyEvent keyE = (KeyEvent)awtEvent;

				if (keyE.isControlDown() &&
						keyE.isShiftDown() &&
						keyE.getKeyCode() == KeyEvent.VK_F1) {
					Window window = getWindow();
					if (null != window) {
						window.list();
					}
				}

//				else if (Util.isWinCE() &&
//				keyE.isControlDown() &&
//				keyE.isShiftDown() &&
//				keyE.getKeyCode() == KeyEvent.VK_RIGHT) {
//				Util.togglePocketPCMouseMovementEvent();
//				}

				else if (keyE.getKeyCode() == KeyEvent.VK_TAB) {
					// TODO Make sure backtab works
					if (keyE.isControlDown() || !wantsAllKeys())
						transferFocus(!keyE.isShiftDown());
				}
			}
			if (newEventsOnly) {
				if (isEventEnabled(awtEvent))
					processEvent(awtEvent);
			} else {
				Event event10 = awtEvent.convertTo10();
				if (event10 != null)
					postEvent(event10);
			}
		}

		// Must post process a key event after it is handled.  Only post process key events
		// that have not been postProcessed already or consumed.  Key events when no
		// focus owner exists are still post processed.
    	if ((isKeyEvent) && !(awtEvent.postProcessed) && !(awtEvent.isConsumed())) {
    		awtEvent.postProcessed = true;
    		kbMgr.postProcessKeyEvent((KeyEvent)awtEvent);
    	}
	}

	/**
	 * Answer whether the receiver can be used as a canvas
	 */
	boolean isPaintable() {
		return false;
	}

	boolean _shouldAllowPaint() {
		if (getIgnoreRepaint()) return false;
		// If the component is laid out already, it's ready to have paint() called
		return isValid();
	}

	boolean wantsAllKeys() {
		return false;
	}

	/**
	 * Post the event to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		event the event to be posted
	 *
	 * @deprecated	use <code>dispatchEvent(AWTEvent)</code> instead.
	 *
	 * @see #dispatchEvent(AWTEvent)
	 */
	public boolean postEvent(Event event) {
		if (!handleEvent(event)) {
			Component p = parent;
			if (p != null)  {
				Point location = getLocation();
				event.translate(location.x, location.y);
				return p.postEvent(event);
			}
		}
		return false;
	}

	void postEvent(AWTEvent awtEvent) {
		if (KeyboardFocusManager.getDebug()) {
			if ((awtEvent instanceof FocusEvent) || (awtEvent instanceof WindowEvent)) {
				System.out.println("postEvent " + awtEvent.toString());
			}
		}
		getToolkit().getSystemEventQueueImpl().postEvent(awtEvent);
	}

	/**
	 * Adds a <code>ComponentListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		componentListener	the <code>ComponentListener</code> to be added.
	 *
	 * @see			java.awt.event.ComponentListener
	 * @see			java.awt.event.ComponentEvent
	 */
	public synchronized void addComponentListener(ComponentListener componentListener) {
		if (componentListener == null) return;
		addEventListener(componentListener);
		this.componentListener = AWTEventMulticaster.add(this.componentListener, componentListener);
	}

	/**
	 * Adds a <code>FocusListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		focusListener	the <code>FocusListener</code> to be added.
	 *
	 * @see			java.awt.event.FocusListener
	 * @see			java.awt.event.FocusEvent
	 */
	public synchronized void addFocusListener(FocusListener focusListener) {
		if (focusListener == null) return;
		addEventListener(focusListener);
		this.focusListener = AWTEventMulticaster.add(this.focusListener, focusListener);
	}

	/**
	 * Adds a <code>KeyListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		keyListener	the <code>KeyListener</code> to be added.
	 *
	 * @see			java.awt.event.KeyListener
	 * @see			java.awt.event.KeyEvent
	 */
	public synchronized void addKeyListener(KeyListener keyListener) {
		if (keyListener == null) return;
		addEventListener(keyListener);
		this.keyListener = AWTEventMulticaster.add(this.keyListener, keyListener);
	}

	/**
	 * Adds a <code>MouseListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		mouseListener	the <code>MouseListener</code> to be added.
	 *
	 * @see			java.awt.event.MouseListener
	 * @see			java.awt.event.MouseEvent
	 */
	public synchronized void addMouseListener(MouseListener mouseListener) {
		if (mouseListener == null) return;
		addEventListener(mouseListener);
		this.mouseListener = AWTEventMulticaster.add(this.mouseListener, mouseListener);
	}

	/**
	 * Adds a <code>MouseMotionListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		mouseMotionListener	the <code>MouseMotionListener</code> to be added.
	 *
	 * @see			java.awt.event.MouseMotionListener
	 * @see			java.awt.event.MouseEvent
	 */
	public synchronized void addMouseMotionListener(MouseMotionListener mouseMotionListener) {
		if (mouseMotionListener == null) return;
		addEventListener(mouseMotionListener);
		this.mouseMotionListener = AWTEventMulticaster.add(this.mouseMotionListener, mouseMotionListener);
	}

	/**
	 * Removes a <code>ComponentListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		componentListener	the <code>ComponentListener</code> to be removed.
	 *
	 * @see			java.awt.event.ComponentListener
	 * @see			java.awt.event.ComponentEvent
	 */
	public synchronized void removeComponentListener(ComponentListener componentListener) {
	 	this.componentListener = AWTEventMulticaster.remove(this.componentListener, componentListener);
	}

	/**
	 * Removes a <code>FocusListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		focusListener	the <code>FocusListener</code> to be removed.
	 *
	 * @see			java.awt.event.FocusListener
	 * @see			java.awt.event.FocusEvent
	 */
	public synchronized void removeFocusListener(FocusListener focusListener) {
		this.focusListener = AWTEventMulticaster.remove(this.focusListener, focusListener);
	}

	/**
	 * Removes a <code>KeyListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		KeyListener	the <code>KeyListener</code> to be removed.
	 *
	 * @see			java.awt.event.KeyListener
	 * @see			java.awt.event.KeyEvent
	 */
	public synchronized void removeKeyListener(KeyListener keyListener) {
		this.keyListener = AWTEventMulticaster.remove(this.keyListener, keyListener);
	}

	/**
	 * Removes a <code>MouseListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		mouseListener	the <code>MouseListener</code> to be removed.
	 *
	 * @see			java.awt.event.MouseListener
	 * @see			java.awt.event.MouseEvent
	 */
	public synchronized void removeMouseListener(MouseListener mouseListener) {
		this.mouseListener = AWTEventMulticaster.remove(this.mouseListener, mouseListener);
	}

	/**
	 * Removes a <code>MouseMotionListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		mouseMotionListener	the <code>MouseMotionListener</code> to be removed.
	 *
	 * @see			java.awt.event.MouseMotionListener
	 * @see			java.awt.event.MouseEvent
	 */
	public synchronized void removeMouseMotionListener(MouseMotionListener mouseMotionListener) {
		this.mouseMotionListener = AWTEventMulticaster.remove(this.mouseMotionListener, mouseMotionListener);
	}

	/**
	 * Enable the specified events for delivery to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	events	events to enable
	 */
	protected final void enableEvents(long events) {
		newEventsOnly = true;
		eventMask |= events;
	}

	/**
	 * Disable the specified events from being delivered.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	events	events to disable
	 */
	protected final void disableEvents(long events) {
		eventMask &= ~events;
	}

	/**
	 * Combine the current event with the new event into a one event
	 * which can serve for both.  Answer the combined event if the
	 * events can be combined, or null if they cannot.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		currentEvent
	 * @param		newEvent
	 *
	 * @return		a java.awt.AWTEvent
	 *
	 * @since		JDK1.2
	 */
	protected AWTEvent coalesceEvents(AWTEvent currentEvent, AWTEvent newEvent) {
		if (currentEvent instanceof MouseEvent) {
			if (currentEvent.getID() == MouseEvent.MOUSE_MOVED ||
				currentEvent.getID() == MouseEvent.MOUSE_DRAGGED)
					return newEvent;
		} else if (currentEvent instanceof PaintEvent) {
			PaintEvent result = (PaintEvent) currentEvent;
			Rectangle currentRect = result.getUpdateRect();
			Rectangle newRect = ((PaintEvent) newEvent).getUpdateRect();
//			if (currentRect.intersects(newRect)) {
				result.setUpdateRect(currentRect.union(newRect));
				return result;
//			}
		} else if (currentEvent instanceof ComponentEvent) {
			if (currentEvent.getID() == ComponentEvent.COMPONENT_RESIZED) {
					return newEvent;
			}
		}
		return null;
	}

	/**
	 * Called when a <code>ComponentEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processComponentEvent(ComponentEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	componentEvent	the event to be processed
	 *
	 * @see		ComponentEvent
	 */
	protected void processComponentEvent(ComponentEvent componentEvent) {
		if (componentListener == null)
			return;

		switch (componentEvent.getID()) {
			case ComponentEvent.COMPONENT_MOVED:
				componentListener.componentMoved(componentEvent);
				break;
			case ComponentEvent.COMPONENT_RESIZED:
				componentListener.componentResized(componentEvent);
				break;
			case ComponentEvent.COMPONENT_SHOWN:
				componentListener.componentShown(componentEvent);
				break;
			case ComponentEvent.COMPONENT_HIDDEN:
				componentListener.componentHidden(componentEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default events handled are <code>MouseEvent</code>, <code>KeyEvent</code>,
	 * <code>ComponentEvent</code> and <code>FocusEvent</code>.  This method
	 * will forward calls to the default events by calling
	 * 		<code>processMouseEvent(MouseEvent)</code>,
	 * 		<code>processKeyEvent(KeyEvent)</code>,
	 * 		<code>processComponentEvent(ComponentEvent)</code>
	 * 	and	<code>processFocusEvent(FocusEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if ((awtEvent instanceof MouseEvent)) {
			switch (awtEvent.getID()) {
				case MouseEvent.MOUSE_WHEEL:
					processMouseWheelEvent((MouseWheelEvent)awtEvent);
					break;
				case MouseEvent.MOUSE_MOVED:
				case MouseEvent.MOUSE_DRAGGED:
					processMouseMotionEvent((MouseEvent)awtEvent);
					break;
				default:
					processMouseEvent((MouseEvent)awtEvent);
			}
		} else if (awtEvent instanceof KeyEvent) {
			processKeyEvent((KeyEvent)awtEvent);
		} else if (awtEvent instanceof FocusEvent) {
			processFocusEvent((FocusEvent)awtEvent);
		} else if (awtEvent instanceof ComponentEvent) {
			processComponentEvent((ComponentEvent)awtEvent);
		}
		else if (awtEvent instanceof InputMethodEvent) {
			processInputMethodEvent((InputMethodEvent)awtEvent);
		}

	}

	/**
	 * Called when a <code>FocusEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processFocusEvent(FocusEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	focusEvent	the event to be processed
	 *
	 * @see		FocusEvent
	 */
	protected void processFocusEvent(FocusEvent focusEvent) {
	    if (focusListener == null)
			return;

		switch (focusEvent.getID()) {
			case FocusEvent.FOCUS_GAINED:
				focusListener.focusGained(focusEvent);
				break;
			case FocusEvent.FOCUS_LOST:
				focusListener.focusLost(focusEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when a <code>KeyEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processKeyEvent(KeyEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	keyEvent	the event to be processed
	 *
	 * @see		KeyEvent
	 */
	protected void processKeyEvent(KeyEvent keyEvent) {
		if (keyListener == null)
			return;

//		if ((eventMask & AWTEvent.INPUT_METHOD_EVENT_MASK) != 0) {
//			getInputContext().dispatchEvent(keyEvent);
//			if (keyEvent.isConsumed()) return;
//		}

		switch (keyEvent.getID()) {
			case KeyEvent.KEY_TYPED:
				keyListener.keyTyped(keyEvent);
				break;
			case KeyEvent.KEY_PRESSED:
				keyListener.keyPressed(keyEvent);
				break;
			case KeyEvent.KEY_RELEASED:
				keyListener.keyReleased(keyEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when a <code>MouseEvent</code> (of type: <code>MOUSE_CLICKED</code>,
	 * <code>MOUSE_ENTERED</code>, <code>MOUSE_EXITED<code>,
	 * <code>MOUSE_PRESSED</code>, <code>MOUSE_RELEASED</code>) occurs
	 * on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processMouseEvent(MouseEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	mouseEvent	the event to be processed
	 *
	 * @see		MouseEvent
	 */
	protected void processMouseEvent(MouseEvent mouseEvent) {
		if (mouseListener == null)
			return;

		switch (mouseEvent.getID()) {
			case MouseEvent.MOUSE_CLICKED:
				mouseListener.mouseClicked(mouseEvent);
				break;
			case MouseEvent.MOUSE_PRESSED:
				mouseListener.mousePressed(mouseEvent);
				break;
			case MouseEvent.MOUSE_RELEASED:
				mouseListener.mouseReleased(mouseEvent);
				break;
			case MouseEvent.MOUSE_ENTERED:
				mouseListener.mouseEntered(mouseEvent);
				break;
			case MouseEvent.MOUSE_EXITED:
				mouseListener.mouseExited(mouseEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when a <code>MouseEvent</code> (of type: <code>MOUSE_DRAGGED</code>
	 * or <code>MOUSE_MOVED</code>) occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processMouseMotionEvent(MouseEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	mouseEvent	the event to be processed
	 *
	 * @see		MouseEvent
	 */
	protected void processMouseMotionEvent(MouseEvent mouseEvent) {
		if (mouseMotionListener == null)
			return;

		switch (mouseEvent.getID()) {
			case MouseEvent.MOUSE_MOVED:
				mouseMotionListener.mouseMoved(mouseEvent);
				break;
			case MouseEvent.MOUSE_DRAGGED:
				mouseMotionListener.mouseDragged(mouseEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when the receiver gets an event.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processEvent(AWTEvent)</code>
	 *				or EventListeners instead.
	 * @see		EventListener
	 * @see		#processEvent(AWTEvent)
	 */
	public boolean handleEvent(Event event) {
		switch (event.id) {
			case Event.ACTION_EVENT: return action(event, event.arg);
			case Event.GOT_FOCUS: return gotFocus(event, event.arg);
			case Event.LOST_FOCUS: return lostFocus(event, event.arg);
			case Event.KEY_PRESS:
			case Event.KEY_ACTION: return keyDown(event, event.key);
			case Event.KEY_RELEASE:
			case Event.KEY_ACTION_RELEASE: return keyUp(event, event.key);
			case Event.MOUSE_ENTER: return mouseEnter(event, event.x, event.y);
			case Event.MOUSE_EXIT: return mouseExit(event, event.x, event.y);
			case Event.MOUSE_MOVE: return mouseMove(event, event.x, event.y);
			case Event.MOUSE_DRAG: return mouseDrag(event, event.x, event.y);
			case Event.MOUSE_DOWN: return mouseDown(event, event.x, event.y);
			case Event.MOUSE_UP: return mouseUp(event, event.x, event.y);
			default: return false;
		}
	}

	/**
	 * Called when a mouse button is pressed on the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseEvent(MouseEvent)</code>
	 *				or MouseListeners instead.
	 *
	 * @see #processMouseEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseDown(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called when a mouse button is pressed on the receiver and then
	 * the mouse is moved while the button is pressed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseMotionEvent(MouseEvent)</code>
	 *				or MouseMotionListeners instead.
	 *
	 * @see #processMouseMotionEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseDrag(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called when the mouse pointer enters the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseEvent(MouseEvent)</code>
	 *				or MouseListeners instead.
	 *
	 * @see #processMouseEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseEnter(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called when the mouse pointer exits the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseEvent(MouseEvent)</code>
	 *				or MouseListeners instead.
	 *
	 * @see #processMouseEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseExit(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called as the mouse pointer moves over the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseMotionEvent(MouseEvent)</code>
	 *				or MouseMotionListeners instead.
	 *
	 * @see #processMouseMotionEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseMove(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called when a mouse button is released on the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	x		the x coordinate of the mouse when the event occured
	 * @param	y		the y coordinate of the mouse when the event occured
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processMouseEvent(MouseEvent)</code>
	 *				or MouseListeners instead.
	 *
	 * @see #processMouseEvent(java.awt.event.MouseEvent)
	 */
	public boolean mouseUp(Event event, int x, int y) {
		return false;
	}

	/**
	 * Called when a key is pressed when the receiver has keyboard
	 * focus.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	key		the value of key pressed
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processKeyEvent(KeyEvent)</code>
	 *				or KeyListeners instead.
	 *
	 * @see #processKeyEvent(java.awt.event.KeyEvent)
	 */
	public boolean keyDown(Event event, int key) {
		return false;
	}

	/**
	 * Called when a key is released when the receiver has keyboard
	 * focus.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	key		the value of key released
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processKeyEvent(KeyEvent)</code>
	 *				or KeyListeners instead.
	 *
	 * @see #processKeyEvent(java.awt.event.KeyEvent)
	 */
	public boolean keyUp(Event event, int key) {
		return false;
	}

	/**
	 * Called when an action occurs on the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	object	and event specific data object
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use ActionListeners instead.
	 *
	 * @see			java.awt.event.ActionListener
	 */
	public boolean action(Event event, Object object) {
		return false;
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @see			#removeNotify
	 */
	public void addNotify() {
		synchronized (getTreeLock()) {
			if (peer == 0) {
				int newPeer = 0;
				Container hwParent = getNativeParent();
				synchronized (hwParent != null ? hwParent.getTreeLock() : new Object()) {
					newPeer = _createPeer(hwParent);
				}

				// Because the write lock is so prone to deadlock, it is best to do
				// as little as possible inside it.
				acquirePeerWriteLock();
				peer = newPeer;
				releasePeerWriteLock();
				isDisplayable = true;

				if (!isPeerNull()) {
					acquirePeerReadLock();
					try {
						_setupPeer();
					} finally {
						releasePeerReadLock();
					}
				}
			}

			// Although it doesn't make much sense, the
			// tree should be invalidated every time addNotify is called.
			// This happens even if the widget has previously been
			// created.  This is to match the behavior of the Sun JRE.
			invalidate();
		}
	}

	void addEventListener(EventListener eventListener) {
		newEventsOnly = true;
	}

	final void acquirePeerWriteLock() {
		synchronized (rwLock) {
			waitingWriters++;
			while (readers.size() != 0) {
				try {
					rwLock.wait();
				} catch (InterruptedException e) {
					throw new RuntimeException("Thread interrupted while waiting for write access to peer");
				}
			}
		}
	}

	final void acquirePeerReadLock() {
		synchronized (rwLock) {
			Thread currentThread = Thread.currentThread();
			MutableInteger count = (MutableInteger)readers.get(currentThread);

			// If the current thread is already reading then we
			// should not wait till the writers are done
			if (count == null) {
				while (waitingWriters > 0) {
					// If the current thread is the event loop then don't make it block on writers.
					// Locking the display thread is a bad idea because it add the possiblity of deadlock
					// when another thread is waiting for it to finish executing.
					if (Util.getDisplay().isEventLoopThread(currentThread) == false) {
						try {
							rwLock.wait();
						} catch (InterruptedException e) {
							throw new RuntimeException("Thread interrupted while waiting for read access to peer");
						}
					}
				}

				// At this point the writers are all done and
				// this thread needs to be marked as reading.
				// We don't need to increment the count b/c that will
				// happen below
				count = new MutableInteger();
				readers.put(currentThread, count);
			}

			count.value++;
		}
	}

	final void releasePeerReadLock() {
		synchronized (rwLock) {
			Thread currentThread = Thread.currentThread();
			MutableInteger count = (MutableInteger)readers.get(currentThread);
			count.value--;
			if (count.value == 0) {
				readers.remove(currentThread);

				// Only if the current thread was just removed
				// can the size become 0 for the first time
				if (readers.size() == 0) {
					rwLock.notifyAll();
				}
			}
		}
	}

	final void releasePeerWriteLock() {
		synchronized (rwLock) {
			waitingWriters--;
			rwLock.notifyAll();
		}
	}

	/**
	 * Destroy the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @see			#addNotify
	 */
	public void removeNotify() {
		synchronized(getTreeLock()) {
			int oldPeer;

			this.isDisplayable = false;

			// Because the write lock is so prone to deadlock, it is best to do
			// as little as possible inside it.  This block will affectly make
			// the widget look disposed from all other threads/API and after
			// that we can take our time cleaning up the peer, without deadlock
			// concerns
			acquirePeerWriteLock();
			oldPeer = peer;
			peer = 0;
			releasePeerWriteLock();

			if (oldPeer != 0) {
				synchronized (activeGraphics) {
					Iterator iterator = activeGraphics.iterator();
					while (iterator.hasNext()) {
						Handle handle = (Handle) iterator.next();
						// Lock graphics handle prior to disposing it to prevent
						// disposing it at the same time in GraphicsImpl.dispose
						synchronized (handle) {
							if (handle.value != 0) {
								OS.Graphics_Dispose(handle.value);
								handle.value = 0;
							}
						}
					}
					activeGraphics.clear();

					synchronized (availableGraphics) {
						// Clean up all unused graphics objects that are waiting to be
						// picked up again.
						int size = availableGraphics.size();
						for (int i=0;i<size;i++) {
							Handle handle = (Handle)availableGraphics.remove(0);
							OS.Graphics_Dispose(handle.value);
							handle.value = 0;
						}
					}
				}

		//		/*[IF]*/
		//		// The data must be null for a programmatic dispose() to work properly.  If this causes an NPE for you,
		//		// come see me and we'll fix the problem together.  - DLW
		//		/*[ENDIF]*/
				OS.Widget_SetData(oldPeer, null);
				OS.Widget_Dispose(oldPeer);
			}

			getToolkit().getSystemEventQueueImpl().removeEvents(this);
		}
//		System.out.println("Component.removeNotify2");

		getInputContext().removeNotify(this);

		/*widget = null;
		popups = null;
		peer = null;*/
		/*parent = null;
		cursor = null;
		background = null;
		foreground = null;
		font = null;
		locale = null;
		componentListener = null;
		focusListener = null;
		keyListener = null;
		inputMethodListener = null;
		mouseListener = null;
		mouseMotionListener = null;*/
	}

	void recreate() {
		removeNotify();
		addNotify();
	}

	/**
	 * Called when the receiver got focus.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	object	and event specific data object
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processFocusEvent(FocusEvent)</code>
	 *				or FocusListeners instead.
	 *
	 * @see #processFocusEvent(java.awt.event.FocusEvent)
	 */
	public boolean gotFocus(Event event, Object object) {
		return false;
	}

	/**
	 * Called when the receiver lost focus.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	event	the event that occured
	 * @param	object	and event specific data object
	 *
	 * @return		<code>true</code> if the event should not be handled
	 *				by the parent, <code>false</code> otherwise.
	 *
	 * @deprecated	use <code>processFocusEvent(FocusEvent)</code>
	 *				or FocusListeners instead.
	 *
	 * @see #processFocusEvent(java.awt.event.FocusEvent)
	 */
	public boolean lostFocus(Event event, Object object) {
		return false;
	}

	/**
	 * Answer whether the receiver can be traversed by keyboard input
	 * only.  Tab and Shift-Tab are used to traverse from one component
	 * to another.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is traversable,
	 *				<code>false</code> otherwise
	 * @deprecated	use <code>isFocusable()</code> instead
	 */
	public boolean isFocusTraversable() {
	    if (isFocusTraversableOverridden == 1) {
	        // client code has overridden focus traversability
	        return isFocusable();
	    }
		return isFocusTraversableImpl();
	}
	boolean isFocusTraversableImpl() {
		return false;
	}
	public void requestFocus() {
		requestFocus(false);
	}
	/**
	 * If the receiver is visible, ask for keyboard focus.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	boolean implRequestFocus(boolean temporary) {
		if (!isDisplayable() || !isVisible()) return false;

		Window window = getWindow();
		if (null == window) return false;

		FocusManager manager = window.getFocusManager();

		synchronized (manager.getRequestedLock()) {
			if (isLightweight()) {
				Container newHwFocus = getNativeParent();
				if (null == newHwFocus) return false;

				Component oldHwFocus = null, oldFocus = null;

				synchronized (manager) {
					oldFocus = manager.getCurrentFocus();

					if (this == oldFocus) return false;

					newHwFocus.focusComponent = this;
					// Keep track of whether or not this is a temporary focus request,
					// in which case we need to update the temporary field properly when
					// the container generates the focus events for the lightweight.
					if (temporary) newHwFocus.isTemporaryFocusRequest = true;

					if (oldFocus != null) {
						if (oldFocus.isLightweight())
							oldHwFocus = oldFocus.getNativeParent();
						else
							oldHwFocus = oldFocus;

						if (newHwFocus == oldHwFocus) {
							manager.requestFocusEmulated(this, oldFocus, temporary);
						}
					}
				}

				if (newHwFocus != oldHwFocus) {
					// See bug 109901.  Calling _widgetRequestFocus() from
					// here is bad because it modifies focusComponent.
					return manager.requestFocus(newHwFocus);
				}

				return true;
			} else {
				if (manager.getCurrentFocus() == this) return false;
				return _widgetRequestFocus(manager);
			}
		}
	}

	boolean _widgetRequestFocus(FocusManager manager) {
		if (KeyboardFocusManager.getDebug()) System.out.println("_widgetRequestFocus OS.force " + this.getName());
		return manager.requestFocus(this);
	}

	/**
	 * Move keyboard focus to the next traversable component.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void transferFocus() {
		if (KeyboardFocusManager.getDebug()) System.out.println("transfer focus " + this.getName());
	    KeyboardFocusManager fm = KeyboardFocusManager.getCurrentKeyboardFocusManager();
	    fm.focusNextComponent(this);
	}

	/**
	 * Move keyboard focus to the next traversable component.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @deprecated	use <code>transferFocus()</code> instead
	 *
	 * @see #transferFocus
	 */
	public void nextFocus() {
		transferFocus();
	}

	void removeFocus() {
		if (parent != null) {
		    nextFocus();

		    // If the component still has focus
		    // then give the focus to the parent to
		    // avoid leaving a disposed widget with
		    // focus
		    if (containsFocus()) {
		    	parent.requestFocus();
		    }
		}
	}

	boolean transferFocus(boolean next) {
		Container parent = this.parent;
		if (parent != null) {
			int start, end, offset;
			if (next) {
				offset = 1;
				end = parent.ncomponents;
			} else {
				offset = -1;
				end = -1;
			}
			start = parent.indexOf(this) + offset;
			for (int index = start; index != end; index += offset) {
				if (parent.component[index].requestFocusChange(next))
					return true;
			}
			return parent.transferFocus(next);
		}
		return requestFocusChange(next);
	}

	boolean primRequestFocus(boolean next) {
	    if (KeyboardFocusManager.getDebug()) System.out.println("primRequestFocus " + this.getName());
		if (isFocusTraversable() && isEnabled() &&
			isVisible() && isDisplayable())
		{
			requestFocus();
			return true;
		} else
			return false;
	}

	boolean requestFocusChange(boolean next) {
		return primRequestFocus(next);
	}

	boolean containsFocus() {
	    Window w = getWindow();
	    if (w != null) {
	    	FocusManager manager = w.getFocusManager();
	    	synchronized (manager) {
		    	Component c = manager.getCurrentFocus();
		    	while (c != null) {
		    		if (c == this) return true;
		    		c = c.getParent();
		    	}
	    	}
	    }
	    return false;
	}

	/**
	 * Answer whether the receiver has keyboard focus.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver has keyboard focus,
				<code>false</code> otherwise.
	 * @since JDK1.2
	 */
	public boolean hasFocus() {
		Window window = getWindow();
		if (null == window) return false;

		return this == window.getFocusOwner();
	}

	/**
	 * Adds a <code>PopupMenu</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		popupMenu	the <code>PopupMenu</code> to be added.
	 * @see			java.awt.PopupMenu
	 */
	public synchronized void add(PopupMenu popupMenu) {
		if (popupMenu != null) {
			popupMenu.parent = this;
			popups.add(popupMenu);
		}

		acquirePeerReadLock();
		try {
			if ((peer != 0) && (popupMenu.peer == 0)) {
				popupMenu.addNotify();
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * This method can be called by the platform to determine
	 * if the menu is actually available.
	 *
	 * @return one of the INFO_HASMENU_* flags
	 */
	int hasMenuInfo(int x, int y) {
		if ((popups != null) && (popups.size() > 0)) return OS.INFO_HASMENU_YES;
		if (mouseListener != null) return OS.INFO_HASMENU_MAYBE;
		return OS.INFO_HASMENU_NO;
	}

	/**
	 * Remove a <code>MenuComponent</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		menuComponent	the <code>MenuComponent</code> to be removed.
	 *
	 * @see			java.awt.MenuComponent
	 * @see			java.awt.PopupMenu
	 */
	public synchronized void remove(MenuComponent menuComponent) {
		if (menuComponent != null) {
			if (menuComponent.parent == this) {
				menuComponent.parent = null;
			}
			popups.remove(menuComponent);
			menuComponent.removeNotify();
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		StringBuffer buffer = new StringBuffer();
		Rectangle bounds = getBounds();
		buffer.append(getName());
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.x);
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.y);
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.width);
		buffer.append("x"); //$NON-NLS-1$
		buffer.append(bounds.height);
		if (!isVisible())
			buffer.append(",hidden"); //$NON-NLS-1$
		if (!isEnabled())
			buffer.append(",disabled"); //$NON-NLS-1$
		if (!isValid())
			buffer.append(",invalid"); //$NON-NLS-1$
		if (isLightweight())
			buffer.append(",lightweight"); //$NON-NLS-1$
		return buffer.toString();
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[" + paramString() + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to System.out.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void list() {
		list(System.out);
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print stream.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printStream	the stream to print on
	 */
	public void list(PrintStream printStream) {
		list(printStream, 0);
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print writer.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printWriter	the writer to print on
	 *
	 */
	public void list(PrintWriter printWriter) {
		list(printWriter, 0);
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print stream indenting each child level
	 * by the specified number of spaces.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printStream	the stream to print on
	 * @param	indent		the number of indenting spaces for child levels
	 *
	 * @see java.io.PrintStream#println(java.lang.Object)
	 */
	public void list(PrintStream printStream, int indent) {
		indent = indent < 0 ? 0 : indent;
		for (int i=0; i<indent; i++) {
			printStream.print(' ');
		}
		printStream.println(this);
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print writer indenting each child level
	 * by the specified number of spaces.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printWriter	the writer to print on
	 * @param	indent		the number of indenting spaces for child levels
	 *
	 * @see java.io.PrintStream#println(java.lang.Object)
	 */
	public void list(PrintWriter printWriter, int indent) {
		indent = indent < 0 ? 0 : indent;
		for (int i=0; i<indent; i++) {
			printWriter.print(' ');
		}
		printWriter.println(this);
	}

	void _adjustZOrder() {
		if (parent == null) return;
		int position = parent.indexOf(this);
		for (int i = position + 1; i < parent.ncomponents; i++) {
			Component nextComp = parent.component[i];
			if (!nextComp.isLightweight()) {
				nextComp.acquirePeerReadLock();
				try {
					int nextWidget = nextComp.peer;
					if (nextWidget != 0)
						OS.Control_MoveAbove(peer, nextWidget);
					break;
				} finally {
					nextComp.releasePeerReadLock();
				}
			}
		}
	}

	int _createPeer(Container hwParent) {
		return 0;
	}

	/*
	 * Created and initializes the native widget.
	 * This method will NOT be called if the native widget has already been created.
	 * This method will always be called in a context which has write access to the peer
	 */
	void _setupPeer() {
		//addAWTListeners();

		OS.Widget_SetData(peer, this);
		_setPeerBounds(getX(), getY(), getWidth(), getHeight());
		OS.Control_SetEnabled(peer, isEnabled());
		OS.Control_SetVisible(peer, isVisible());

		if (cursor != null) OS.Control_SetCursor(peer, cursor.getPeer());

        // If the user has explicitly set a foreground color
        // then that color should be set on the widget, otherwise
        // whichever color the widget defaults to is correct
        Color color = _getForeground();
        if (color != null) OS.Control_SetForeground(peer, color.getPeer());

		// Set the background color to the correct initial value.
		// This logic has been tweaked to match the behavior of the
		// RI exactly and any modification should be done carefully.
		if (background != null) {
			// If the user has explicitly set a background color
			// of THIS widget then that color should be used.
			OS.Control_SetBackground(peer, background.getPeer());
		} else {
			// If the background color for this widget has not
			// been set by the user then the background color should
			// only be inherited on creation for some widgets depending
			// on the platform behaviour.  For example, on most platfoms
			// it looks nice when a label gets the background color from
			// the parent but usually it is not desired for TextField/TextArea
			// to get the background color
			if (OS.Control_InheritsBackground(peer)) {
				// Check to see if any background color has been set on any
				// widget in the parent chain.
				Color inheritedBackground = _getBackground();
				if (inheritedBackground != null) {
					OS.Control_SetBackground(peer, inheritedBackground.getPeer());
				} else {
					// It is possible that the control would like to inherit the
					// background color from the parent but no color has been
					// set.  In that case nothing needs to be done and the default
					// system color is the right one to use.
				}
			} else {
				// In the case where the background color should not be inherited
				// then no background should be set and the default system color will
				// be used
			}
		}

		// If the user has explicitly set a font then that font
		// should be set on the widget, otherwise whichever font
		// the widget defaults to is correct
		Font font = _getFont();
		if (font != null) OS.Control_SetFont(peer, font.getPeer().fPeerHandle);

		_adjustZOrder();
	}

	/*
	 * Return true if the event was consumed, false otherwise.
	 * PPro KeyEvents are dispatched asynchronosly from this
	 * method, via the EventQueue.  So it is not possible for
	 * this method to know whether or not the event was consumed.
	 * Therefore, this method always returns false so that the
	 * implementation always passes key events on to the native widget.
	 *
	 * Note: UGL always returns keycodes of capital characters.
	 */
	boolean keyCallback(int type, char character, int keycode, int modifiers) {
		//System.out.println("keyCallback("+type+","+character+","+keycode+","+modifiers+")");

		int awtKeycode = keycode;
		char awtCharacter = character;
		boolean consumed = false;

		/*
		 * Assign the AWT keycode and character value for the event.
		 * Some key strokes are expected to have KeyEvent.CHAR_UNDEFINED
		 * for the character value. This is not clearly documented in the
		 * spec, so for consistency it's best to follow the behavior of
		 * the RI and JDK.
		 */

		switch (keycode) {

			case OS.KEYCODE_BACKSPACE:
				awtKeycode = KeyEvent.VK_BACK_SPACE;
				awtCharacter = KeyEvent.VK_BACK_SPACE;
				break;

			case OS.KEYCODE_TAB:
				awtKeycode = KeyEvent.VK_TAB;
				awtCharacter = KeyEvent.VK_TAB;
				if ((modifiers &  OS.MODIFIER_MASK_CONTROL) == OS.MODIFIER_MASK_CONTROL) {
				    consumed = true;
				}
				break;

			case OS.KEYCODE_ENTER:
				awtKeycode = KeyEvent.VK_ENTER;
				awtCharacter = KeyEvent.VK_ENTER;
				break;

			case OS.KEYCODE_ESCAPE:
				awtKeycode = KeyEvent.VK_ESCAPE;
				awtCharacter = KeyEvent.VK_ESCAPE;
				break;

			case OS.KEYCODE_PRINTSCREEN:
				awtKeycode = KeyEvent.VK_PRINTSCREEN;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_PAUSE:
				awtKeycode = KeyEvent.VK_PAUSE;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_HOME:
				awtKeycode = KeyEvent.VK_HOME;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_END:
				awtKeycode = KeyEvent.VK_END;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_PAGE_UP:
				awtKeycode = KeyEvent.VK_PAGE_UP;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_PAGE_DOWN:
				awtKeycode = KeyEvent.VK_PAGE_DOWN;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_INSERT:
				awtKeycode = KeyEvent.VK_INSERT;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_DELETE:
				awtKeycode = KeyEvent.VK_DELETE;
				awtCharacter = KeyEvent.VK_DELETE;
				break;

			case OS.KEYCODE_ARROW_LEFT:
				awtKeycode = KeyEvent.VK_LEFT;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_ARROW_RIGHT:
				awtKeycode = KeyEvent.VK_RIGHT;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_ARROW_UP:
				awtKeycode = KeyEvent.VK_UP;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_ARROW_DOWN:
				awtKeycode = KeyEvent.VK_DOWN;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_SHIFT:
				awtKeycode = KeyEvent.VK_SHIFT;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_ALT:
				awtKeycode = KeyEvent.VK_ALT;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_CONTROL:
				awtKeycode = KeyEvent.VK_CONTROL;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_LOCK_CAPS:
				awtKeycode = KeyEvent.VK_CAPS_LOCK;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_LOCK_NUM:
				awtKeycode = KeyEvent.VK_NUM_LOCK;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_LOCK_SCROLL:
				awtKeycode = KeyEvent.VK_SCROLL_LOCK;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F1:
				awtKeycode = KeyEvent.VK_F1;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F2:
				awtKeycode = KeyEvent.VK_F2;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F3:
				awtKeycode = KeyEvent.VK_F3;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F4:
				awtKeycode = KeyEvent.VK_F4;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F5:
				awtKeycode = KeyEvent.VK_F5;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F6:
				awtKeycode = KeyEvent.VK_F6;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F7:
				awtKeycode = KeyEvent.VK_F7;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F8:
				awtKeycode = KeyEvent.VK_F8;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F9:
				awtKeycode = KeyEvent.VK_F9;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F10:
				awtKeycode = KeyEvent.VK_F10;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F11:
				awtKeycode = KeyEvent.VK_F11;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F12:
				awtKeycode = KeyEvent.VK_F12;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F13:
				awtKeycode = KeyEvent.VK_F13;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F14:
				awtKeycode = KeyEvent.VK_F14;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F15:
				awtKeycode = KeyEvent.VK_F15;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F16:
				awtKeycode = KeyEvent.VK_F16;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F17:
				awtKeycode = KeyEvent.VK_F17;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F18:
				awtKeycode = KeyEvent.VK_F18;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F19:
				awtKeycode = KeyEvent.VK_F19;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F20:
				awtKeycode = KeyEvent.VK_F20;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F21:
				awtKeycode = KeyEvent.VK_F21;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F22:
				awtKeycode = KeyEvent.VK_F22;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F23:
				awtKeycode = KeyEvent.VK_F23;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_F24:
				awtKeycode = KeyEvent.VK_F24;
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case '\'':
				awtKeycode = KeyEvent.VK_QUOTE;
				break;

			case '"':
				awtKeycode = KeyEvent.VK_QUOTEDBL;
				break;

			case '{':
				awtKeycode = KeyEvent.VK_BRACELEFT;
				break;

			case '}':
				awtKeycode = KeyEvent.VK_BRACERIGHT;
				break;

			case '*':
				awtKeycode = KeyEvent.VK_ASTERISK;
				break;

			case '+':
				awtKeycode = KeyEvent.VK_PLUS;
				break;

			case ' ':
				awtKeycode = KeyEvent.VK_SPACE;
				break;

			case '0':
				awtKeycode = KeyEvent.VK_0;
				break;

			case '1':
				awtKeycode = KeyEvent.VK_1;
				break;

			case '2':
				awtKeycode = KeyEvent.VK_2;
				break;

			case '3':
				awtKeycode = KeyEvent.VK_3;
				break;

			case '4':
				awtKeycode = KeyEvent.VK_4;
				break;

			case '5':
				awtKeycode = KeyEvent.VK_5;
				break;

			case '6':
				awtKeycode = KeyEvent.VK_6;
				break;

			case '7':
				awtKeycode = KeyEvent.VK_7;
				break;

			case '8':
				awtKeycode = KeyEvent.VK_8;
				break;

			case '9':
				awtKeycode = KeyEvent.VK_9;
				break;

			case 'A':
				awtKeycode = KeyEvent.VK_A;
				break;

			case 'B':
				awtKeycode = KeyEvent.VK_B;
				break;

			case 'C':
				awtKeycode = KeyEvent.VK_C;
				break;

			case 'D':
				awtKeycode = KeyEvent.VK_D;
				break;

			case 'E':
				awtKeycode = KeyEvent.VK_E;
				break;

			case 'F':
				awtKeycode = KeyEvent.VK_F;
				break;

			case 'G':
				awtKeycode = KeyEvent.VK_G;
				break;

			case 'H':
				awtKeycode = KeyEvent.VK_H;
				break;

			case 'I':
				awtKeycode = KeyEvent.VK_I;
				break;

			case 'J':
				awtKeycode = KeyEvent.VK_J;
				break;

			case 'K':
				awtKeycode = KeyEvent.VK_K;
				break;

			case 'L':
				awtKeycode = KeyEvent.VK_L;
				break;

			case 'M':
				awtKeycode = KeyEvent.VK_M;
				break;

			case 'N':
				awtKeycode = KeyEvent.VK_N;
				break;

			case 'O':
				awtKeycode = KeyEvent.VK_O;
				break;

			case 'P':
				awtKeycode = KeyEvent.VK_P;
				break;

			case 'Q':
				awtKeycode = KeyEvent.VK_Q;
				break;

			case 'R':
				awtKeycode = KeyEvent.VK_R;
				break;

			case 'S':
				awtKeycode = KeyEvent.VK_S;
				break;

			case 'T':
				awtKeycode = KeyEvent.VK_T;
				break;

			case 'U':
				awtKeycode = KeyEvent.VK_U;
				break;

			case 'V':
				awtKeycode = KeyEvent.VK_V;
				break;

			case 'W':
				awtKeycode = KeyEvent.VK_W;
				break;

			case 'X':
				awtKeycode = KeyEvent.VK_X;
				break;

			case 'Y':
				awtKeycode = KeyEvent.VK_Y;
				break;

			case 'Z':
				awtKeycode = KeyEvent.VK_Z;
				break;

			case '-':
				awtKeycode = KeyEvent.VK_MINUS;
				break;

			case '&':
				awtKeycode = KeyEvent.VK_AMPERSAND;
				break;

			case '@':
				awtKeycode = KeyEvent.VK_AT;
				break;

			case ':':
				awtKeycode = KeyEvent.VK_COLON;
				break;

			case ';':
				awtKeycode = KeyEvent.VK_SEMICOLON;
				break;

			case ',':
				awtKeycode = KeyEvent.VK_COMMA;
				break;

			case '$':
				awtKeycode = KeyEvent.VK_DOLLAR;
				break;

			case '!':
				awtKeycode = KeyEvent.VK_EXCLAMATION_MARK;
				break;

			case '=':
				awtKeycode = KeyEvent.VK_EQUALS;
				break;

			case '(':
				awtKeycode = KeyEvent.VK_LEFT_PARENTHESIS;
				break;

			case ')':
				awtKeycode = KeyEvent.VK_RIGHT_PARENTHESIS;
				break;

			case '#':
				awtKeycode = KeyEvent.VK_NUMBER_SIGN;
				break;

			case OS.KEYCODE_KEYPAD_ADD:
				awtKeycode = KeyEvent.VK_ADD;
				break;

			case OS.KEYCODE_KEYPAD_MULTIPLY:
				awtKeycode = KeyEvent.VK_MULTIPLY;
				break;

			case OS.KEYCODE_KEYPAD_SUBTRACT:
				awtKeycode = KeyEvent.VK_SUBTRACT;
				break;

			case OS.KEYCODE_KEYPAD_DIVIDE:
				awtKeycode = KeyEvent.VK_DIVIDE;
				break;

			case OS.KEYCODE_KEYPAD_DECIMAL:
				awtKeycode = KeyEvent.VK_DECIMAL;
				break;

			case OS.KEYCODE_KEYPAD_0:
				awtKeycode = KeyEvent.VK_NUMPAD0;
				break;

			case OS.KEYCODE_KEYPAD_1:
				awtKeycode = KeyEvent.VK_NUMPAD1;
				break;

			case OS.KEYCODE_KEYPAD_2:
				awtKeycode = KeyEvent.VK_NUMPAD2;
				break;

			case OS.KEYCODE_KEYPAD_3:
				awtKeycode = KeyEvent.VK_NUMPAD3;
				break;

			case OS.KEYCODE_KEYPAD_4:
				awtKeycode = KeyEvent.VK_NUMPAD4;
				break;

			case OS.KEYCODE_KEYPAD_5:
				awtKeycode = KeyEvent.VK_NUMPAD5;
				break;

			case OS.KEYCODE_KEYPAD_6:
				awtKeycode = KeyEvent.VK_NUMPAD6;
				break;

			case OS.KEYCODE_KEYPAD_7:
				awtKeycode = KeyEvent.VK_NUMPAD7;
				break;

			case OS.KEYCODE_KEYPAD_8:
				awtKeycode = KeyEvent.VK_NUMPAD8;
				break;

			case OS.KEYCODE_KEYPAD_9:
				awtKeycode = KeyEvent.VK_NUMPAD9;
				break;

			case '[':
				awtKeycode = KeyEvent.VK_OPEN_BRACKET;
				break;

			case ']':
				awtKeycode = KeyEvent.VK_CLOSE_BRACKET;
				break;

			case '_':
				awtKeycode = KeyEvent.VK_UNDERSCORE;
				break;

			case '/':
				awtKeycode = KeyEvent.VK_SLASH;
				break;

			case '\\':
				awtKeycode = KeyEvent.VK_BACK_SLASH;
				break;

			case '.':
				awtKeycode = KeyEvent.VK_PERIOD;
				break;

			case '`':
				awtKeycode = KeyEvent.VK_BACK_QUOTE;
				break;

			case OS.KEYCODE_HELP:
				awtKeycode = KeyEvent.VK_HELP;
				break;

			/*
			 * CBA keys.  Do not use the constant names because
			 * we do not want a dependency on the extension class.
			 */
			case OS.KEYCODE_DEVICE0:
				awtKeycode = 164; // EikCommandButtonGroup.BUTTON1_KEYCODE
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_DEVICE1:
				awtKeycode = 165; // EikCommandButtonGroup.BUTTON2_KEYCODE
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_DEVICE2:
				awtKeycode = 166; // EikCommandButtonGroup.BUTTON3_KEYCODE
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;

			case OS.KEYCODE_DEVICE3:
				awtKeycode = 167; // EikCommandButtonGroup.BUTTON4_KEYCODE
				awtCharacter = KeyEvent.CHAR_UNDEFINED;
				break;
		}

		int awtModifiers = Util.convertToAWTKeyModifiers(modifiers);
		int eventType;
		switch (type) {

			case OS.EVENT_KEY_DOWN:
				eventType = KeyEvent.KEY_PRESSED;
				break;

			case OS.EVENT_KEY_UP:
				eventType = KeyEvent.KEY_RELEASED;
				break;

			default: return true;
		}

		_postKeyEvent(eventType, awtCharacter, awtKeycode, awtModifiers);

		return consumed;
	}

	void paintCallback(int x, int y, int width, int height) {
		if (getIgnoreRepaint()) return;

		postEvent(new PaintEvent(
			this,
			PaintEvent.PAINT,
			new Rectangle(x, y, width, height)));
	}

	void mouseWheelCallback(int type, int x, int y, int modifiers, int scrollAmount, int wheelRotation) {
		int awtModifiers = Util.convertToAWTButtonMasks(modifiers) | Util.convertToAWTKeyModifiers(modifiers);
		int awtType = MouseEvent.MOUSE_WHEEL;
		int clickCount = 0;
		int awtScrollType = 0;
		long time = System.currentTimeMillis();

		switch (type) {
			case OS.EVENT_WHEEL_BLOCK:
				awtScrollType = MouseWheelEvent.WHEEL_BLOCK_SCROLL;
				break;
			case OS.EVENT_WHEEL_UNIT:
				awtScrollType = MouseWheelEvent.WHEEL_UNIT_SCROLL;
				break;
		}
		postEvent(new MouseWheelEvent(this, MouseEvent.MOUSE_WHEEL, time, awtModifiers,
			x, y, clickCount, false, awtScrollType, scrollAmount, wheelRotation));
	}

	void pointerCallback(int type, int x, int y, int buttonState, int modifiers) {

		// While button is really a PPro 1.1 artifact, it's also necessary
		// to maintain it separately for PPro 1.0 to avoid ambiguity with
		// the key modifiers (Alt/Ctrl/Shift).
		int awtButton = Util.convertToAWTButtonMasks(buttonState);
		// Since the button field is ultimately not used for PPro 1.0,
		// make sure it's included in the modifiers.
		int awtModifiers = Util.convertToAWTButtonMasks(modifiers) | Util.convertToAWTKeyModifiers(modifiers) | awtButton;

		int awtType = 0;
		boolean isPopupTrigger = false;
		int clickCount = 0;

		switch (type) {

			case OS.EVENT_POINTER_DOWN:
				awtType = MouseEvent.MOUSE_PRESSED;
				if ((modifiers & OS.POINTER_POPUP_TRIGGER) != 0) {
					isPopupTrigger = true;
				}
				clickCount = 1;
				break;

			case OS.EVENT_POINTER_UP:
				awtType = MouseEvent.MOUSE_RELEASED;
				if ((modifiers & OS.POINTER_POPUP_TRIGGER) != 0) {
					isPopupTrigger = true;
				}
				clickCount = 1;
				break;

			case OS.EVENT_POINTER_MOVE:
				awtType = MouseEvent.MOUSE_MOVED;
				break;

			case OS.EVENT_POINTER_ENTER:
				awtType = MouseEvent.MOUSE_ENTERED;
				break;

			case OS.EVENT_POINTER_EXIT:
				awtType = MouseEvent.MOUSE_EXITED;
				break;

			case OS.EVENT_POINTER_DRAG:
				awtType = MouseEvent.MOUSE_DRAGGED;
				break;

			case OS.EVENT_POINTER_DOUBLECLICK:
				awtType = MouseEvent.MOUSE_PRESSED;
				clickCount = 2;
				break;

			default:

//			case OS.EVENT_POINTER_HOVER:
				// do nothing
				return;
		}

		_postMouseEvent(awtType, x, y, awtButton, awtModifiers, clickCount, isPopupTrigger);
	}

	void focusCallback(int type) {
		int javaType;
		switch (type) {
			case OS.EVENT_FOCUS_IN:
				javaType = FocusEvent.FOCUS_GAINED;
				break;

			case OS.EVENT_FOCUS_OUT:
				javaType = FocusEvent.FOCUS_LOST;
				break;

			default:
				return;
		}

		if (KeyboardFocusManager.getDebug()) {
			String focusType = " gained ";
			if (javaType == FocusEvent.FOCUS_LOST) focusType = " lost ";
			System.out.println("focus callback" + focusType + this.getName());
		}

		boolean temporary = ((type & OS.EVENT_FOCUS_TEMPORARY_FLAG) != 0);

//		FocusEvent event;
//		/*[IF PPRO11|PBPRO11]*/
//		// TODO PPro 1.1 - using getFocusOwner is bogus since it's
//		// not up-to-date yet - instead UGL will give us this info
//		if (javaType == FocusEvent.FOCUS_GAINED) {
//			// this seems to work ok since lost precedes gained
//			event = new FocusEvent(this, javaType, temporary, previouslyFocusedComponent);
//		} else {
//			previouslyFocusedComponent = this;
//			Window parentWindow = getWindow();
//			Component newFocus = null;
//			if (parentWindow != null) {
//				//newFocus = parentWindow.getFocusOwner();
//				// TODO update when we get proper UGL support
//				newFocus = null;
//			}
//			if (!isShowing()) temporary = true;
//			event = new FocusEvent(this, javaType, temporary, newFocus);
//		}
//		postEvent(event);
//		/*[ELSE]
//		event = new FocusEvent(this, javaType, temporary);
//		_postFocusEvent(event);
//		/*[ENDIF]*/

		FocusEvent event;
//		/*[IF PPRO11|PBPRO11]*/
//		if (javaType == FocusEvent.FOCUS_GAINED) {
//			// this seems to work ok since lost precedes gained
//			if (getPreviousLostFocusEvent() != null) {
//				_postFocusEvent(recreateLastFocusEvent());
//			}
//			Component lostComponent = previousLostFocusEvent == null ? null : previousLostFocusEvent.getComponent();
//			if (lostComponent == this) {
//				// This happens when you click away to an OS window and then click back
//				lostComponent = null;
//			}
//			event = new FocusEvent(this, javaType, temporary, lostComponent);
//			setPreviousLostFocusEvent(null);
//		} else {
//			if (!isShowing()) temporary = true;
//			event = new FocusEvent(this, javaType, temporary, KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner());
//			setPreviousLostFocusEvent(event);
//			return;
//		}
//		postEvent(event);
//		/*[ELSE]
		event = new FocusEvent(this, javaType, temporary);
		_postFocusEvent(event);
//		/*[ENDIF]*/
	}

	void _setPeerBounds(int x1, int y1, int width1, int height1) {
		if (peer != 0) {
			Point location = translateToPeerCoords(new Point(x1, y1));

			// See bug 105415 - The origin of a child widget in UGL
			// is outside the bounds of the parent's decorations.  So
			// we need to adjust the position here, because PPro puts
			// children underneath the decorations.
			if (parent != null) {
			    // See bug 106680 for why we adjust by insets
			    /*
			     * It is important that we call getInsetsImpl() here instead
			     * of getInsets(), because the user may override getInsets(),
			     * but we still need to adjust by the OS-dictated insets here.
			     * A good example of why is the TCK Agent-Frame harness. The
			     * entire custom-drawn panel (with the hand-drawn tabs) is positioned
			     * incorrectly, rendering the tabs invisible (off the screen) if
			     * we call getInsets() here.
			     * Also see bug 108776
			     */
			    Insets insets = getNativeParent().getInsetsImpl();
			    location.x -= insets.left;
			    location.y -= insets.top;
			}

			OS.Control_SetBounds(peer, location.x, location.y, width1, height1);
		}
	}

	void setVisibleImpl(boolean visible, boolean repaint) {
		if (isLightweight()) {
			if (null != parent) {
				if (repaint) {
					parent.repaint(getX(), getY(), getWidth(), getHeight());
				}
			}
		} else {
			acquirePeerReadLock();
			try {
				if (peer != 0)
					OS.Control_SetVisible(peer, visible);
			} finally {
				releasePeerReadLock();
			}
		}
	}

	void buttonCallback() {
		// This method is here to provide a common method
		// between Button and Checkbox.  If checkbox and button
		// had a different parent this method would be moved there
	}

	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(componentListenerK))
				addComponentListener((ComponentListener)listener);
			else if (key.equals(focusListenerK))
				addFocusListener((FocusListener)listener);
			else if (key.equals(keyListenerK))
				addKeyListener((KeyListener)listener);
			else if (key.equals(mouseWheelListenerK))
				addMouseWheelListener((MouseWheelListener)listener);
			else if (key.equals(inputMethodListenerK))
				addInputMethodListener((InputMethodListener)listener);
			else if (key.equals(mouseListenerK))
				addMouseListener((MouseListener)listener);
			else if (key.equals(mouseMotionListenerK))
				addMouseMotionListener((MouseMotionListener)listener);
		}

		// transient object fields are not initialized when a object
		// is deserialized.  Some methods depend on object fields being
		// initialized to good defaults.
		rwLock = new Object();
		readers = new HashMap();
		focusable = isFocusableImpl();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, componentListenerK, componentListener);
		AWTEventMulticaster.save(stream, focusListenerK, focusListener);
		AWTEventMulticaster.save(stream, keyListenerK, keyListener);
		AWTEventMulticaster.save(stream, mouseListenerK, mouseListener);
		AWTEventMulticaster.save(stream, mouseMotionListenerK, mouseMotionListener);
		AWTEventMulticaster.save(stream, mouseWheelListenerK, mouseWheelListener);
		AWTEventMulticaster.save(stream, inputMethodListenerK, inputMethodListener);
		stream.writeObject(null);
	}

	void _processDestroyEvent() {
	}

	void _postFocusEvent(FocusEvent focusEvent) {
		Component source = focusEvent.getComponent();
		Window window = source.getWindow();
		if (window != null) {
			FocusManager manager = window.getFocusManager();
			synchronized (manager) {
				if (focusEvent.getID() == FocusEvent.FOCUS_GAINED) {
					manager.focusGained(source);
				} else
				if (focusEvent.getID() == FocusEvent.FOCUS_LOST) {
					manager.focusLost(source);
				}
			}
		}
		source.postEvent(focusEvent);
	}

	void _postKeyEvent(int type, char character, int keyCode, int stateMask) {
		long time = System.currentTimeMillis();

		// TODO i'm sure we need to copy the special key processing from BBAWT
		postEvent(new KeyEvent(this, type, time, stateMask, keyCode, character));
		if (type == KeyEvent.KEY_PRESSED) {
			if (shouldGenerateKeyTypedEvent(character)) {
				postEvent(new KeyEvent(this, KeyEvent.KEY_TYPED, time, stateMask, KeyEvent.VK_UNDEFINED, character));
			}
		}
	}

	static int fgButtonModifierMask =
		InputEvent.BUTTON1_MASK | InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK
		| InputEvent.BUTTON1_DOWN_MASK | InputEvent.BUTTON2_DOWN_MASK | InputEvent.BUTTON3_DOWN_MASK
		;

	void _postMouseEvent(int type, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		long time = System.currentTimeMillis();

		switch (type) {
			case MouseEvent.MOUSE_PRESSED :
				_generateMousePressedEvent(time, x, y, awtButton, awtModifiers, clickCount, popupTrigger);
				break;
			case MouseEvent.MOUSE_RELEASED :
				_generateMouseReleasedEvent(time, x, y, awtButton, awtModifiers, clickCount, popupTrigger);
				break;
			case MouseEvent.MOUSE_MOVED :
				if (this instanceof Container) {
					((Container)this)._generateMouseEnterExitForLightweights(time, x, y);
				}

				_generateMouseMovedEvent(time, x, y, awtModifiers);
				break;
			case MouseEvent.MOUSE_DRAGGED :
				// Although a heavyweight component received the drag event,
				// since there is a grab in affect, the mouse may actually be
				// over a container with lightweights. In this case, allow the
				// container to generate any enter/exit events as needed.
				Component lastEntered = Display.getLastComponentEntered();
				if (lastEntered != null) {
					if (lastEntered.isLightweight()) {
						// If the lastEntered component is a lightweight, then back up
						// to the heavyweight parent. This is necessary in order to be
						// able to determine whether a lightweight has been exited.
						lastEntered = lastEntered.getNativeParent();
					}
					if (lastEntered instanceof Container) {
						Point coords = new Point(x, y);
						translateCoords(this, lastEntered, coords);
						((Container)lastEntered)._generateMouseEnterExitForLightweights(time, coords.x, coords.y);
					}
				}

				_generateMouseDraggedEvent(time, x, y, awtModifiers);
				break;
			case MouseEvent.MOUSE_ENTERED :
				_generateMouseEnteredEvent(time, x, y);
				break;
			case MouseEvent.MOUSE_EXITED :
				_generateMouseExitedEvent(time, x, y);
				break;

			default :
				break;
		}
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 */
	void _generateMouseEnteredEvent(long time, int x, int y) {
		postEvent(new MouseEvent(this, MouseEvent.MOUSE_ENTERED, time, 0, x, y, 0, false));
		Display.logComponentEntered(this);
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 */
	void _generateMouseExitedEvent(long time, int x, int y) {
		postEvent(new MouseEvent(this, MouseEvent.MOUSE_EXITED, time, 0, x, y, 0, false));
		Display.logComponentExited(this);
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 * @param mods
	 */
	void _generateMouseMovedEvent(long time, int x, int y, int mods) {
		postEvent(new MouseEvent(this, MouseEvent.MOUSE_MOVED, time, mods, x, y, 0, false));
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 * @param mods
	 */
	void _generateMouseDraggedEvent(long time, int x, int y, int mods) {
		int buttonMask = (InputEvent.BUTTON1_MASK | InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK |
			InputEvent.BUTTON1_DOWN_MASK | InputEvent.BUTTON2_DOWN_MASK | InputEvent.BUTTON3_DOWN_MASK);

		lastDraggedButtons |= (mods & buttonMask);

		postEvent(new MouseEvent(this, MouseEvent.MOUSE_DRAGGED, time, mods, x, y, 0, false));
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 * @param awtButton
	 * @param awtModifiers
	 * @param clickCount
	 * @param popupTrigger
	 */
	void _generateMousePressedEvent(long time, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		// Only grab on the first press
		if (Display.getLastGrabbedComponentWithEvents() == null) {
			Display.setLastGrabbedComponentWithEvents(this);
		}

		postEvent(new MouseEvent(this, MouseEvent.MOUSE_PRESSED, time, awtModifiers, x, y, clickCount, popupTrigger
				, Util.convertToMouseEventButton(awtButton)
				));
	}

	/**
	 * @param time
	 * @param x
	 * @param y
	 * @param awtButton
	 * @param awtModifiers
	 * @param clickCount
	 * @param popupTrigger
	 */
	void _generateMouseReleasedEvent(long time, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		// If the cursor location is no longer over the component where
		// the grab began, then reset the cursor for the component underneath
		// the cursor location. This is particularly important for lightweights,
		// but can also affect heavyweights in certain scenarios (such as
		// pressing on a lightweight child and then releasing while over the
		// heavyweight parent).
		// Note that the component under the cursor should,
		// and is assumed to be, the last component entered.
		Component lastGrabbed = Display.getLastGrabbedComponentWithEvents();
		Component componentUnderCursor = Display.getLastComponentEntered();
		if ((componentUnderCursor != null) && (componentUnderCursor != lastGrabbed)) {
			Component heavyweight;
			if (componentUnderCursor.isLightweight()) {
				heavyweight = componentUnderCursor.getNativeParent();
			} else {
				heavyweight = componentUnderCursor;
			}
			heavyweight.resetCursor(componentUnderCursor);
		}

		postEvent(new MouseEvent(this, MouseEvent.MOUSE_RELEASED, time, awtModifiers, x, y, clickCount, popupTrigger
				, Util.convertToMouseEventButton(awtButton)
				));

		// If the last component to grab the cursor was this widget, and
		// no buttons have been dragged then generate a click event
		if ((this == lastGrabbed) && ((lastDraggedButtons & awtButton) == 0)) {
			// No drag occurred while button was pressed, so generate a clicked event.
			postEvent(new MouseEvent(this, MouseEvent.MOUSE_CLICKED, time, awtModifiers, x, y, clickCount, false
					, Util.convertToMouseEventButton(awtButton)
					));
		}
		// Clear drag state for released button.
		lastDraggedButtons &= ~awtButton;

		// Only ungrab after all buttons are released
		// Note that the button being released will also be
		// set in the modifiers. See pointerCallback().
		if ((awtModifiers & fgButtonModifierMask) == awtButton) {
			Display.setLastGrabbedComponentWithEvents(null);
		}
	}

	Component findComponentAt(int x, int y) {
		if (getParent() != null){
			return getParent().findComponentAt(x,y);
		} else {
			return getComponentAt(x,y);
		}
	}

	Point translateCoords(Component from, Component to, Point point) {
		if (to == from)
			return point;

		Component fromHW = from;
		int fromX = 0, fromY = 0;
		while (fromHW != null && fromHW.isLightweight()) {
			fromX += fromHW.getX();
			fromY += fromHW.getY();
			fromHW = fromHW.parent;
		}
		Component toHW = to;
		int toX = 0, toY = 0;
		while (toHW != null && toHW.isLightweight()) {
			toX += toHW.getX();
			toY += toHW.getY();
			toHW = toHW.parent;
		}

		if (toHW == fromHW) {
			point.x += fromX - toX;
			point.y += fromY - toY;
			return point;
		}

		Point fromPoint = new Point(fromX, fromY);
		if (fromHW != null && fromHW.peer != 0)
			fromPoint = fromHW.toDisplay(fromPoint);

		Point toPoint = new Point(toX, toY);
		if (toHW != null && toHW.peer != 0)
			toPoint = toHW.toDisplay(toPoint);

		point.x += fromPoint.x - toPoint.x;
		point.y += fromPoint.y - toPoint.y;
		return point;
	}

	public void enableInputMethods(boolean enable) {
		enableInputMethodsCalled = true;
		if (enable) {
			enableEvents(AWTEvent.INPUT_METHOD_EVENT_MASK);
		} else {
			disableEvents(AWTEvent.INPUT_METHOD_EVENT_MASK);
		}
	}

	public boolean isForegroundSet() {
		return foreground != null;
	}

	public boolean isBackgroundSet() {
		return background != null;
	}

	public boolean isFontSet() {
		return font != null;
	}

	public boolean isCursorSet() {
		return cursor != null;
	}

	public VolatileImage createVolatileImage(int width, int height) {
		VolatileImage img = null;
		try {
			img = createVolatileImage(width, height, null);
		} catch (AWTException ae) {
		}
		return img;
	}

	public VolatileImage createVolatileImage(int width, int height, ImageCapabilities imageCapabilities) throws AWTException {
		if (GraphicsEnvironment.isHeadless() || !isDisplayable()) return null;
		if ((width <= 0) || (height <= 0)) {
			throw new IllegalArgumentException("Width ("+width+") and height ("+height+") must be > 0"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return new VolatileImageImpl(this, width, height, getBackground(), imageCapabilities);
	}

	/**
	 * Set whether OS-generated paint requests should be ignored.
	 */
	public void setIgnoreRepaint(boolean ignoreRepaint) {
		this.ignoreRepaint = ignoreRepaint;
	}

	/**
	 * Answer whether OS-generated paint requests should be ignored
	 */
	public boolean getIgnoreRepaint() {
		return ignoreRepaint;
	}

	public ComponentListener[] getComponentListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(componentListener, ComponentListener.class);
		ComponentListener[] result = new ComponentListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	public FocusListener[] getFocusListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(focusListener, FocusListener.class);
		FocusListener[] result = new FocusListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	public KeyListener[] getKeyListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(keyListener, KeyListener.class);
		KeyListener[] result = new KeyListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	public MouseListener[] getMouseListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(mouseListener, MouseListener.class);
		MouseListener[] result = new MouseListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	public MouseMotionListener[] getMouseMotionListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(mouseMotionListener, MouseMotionListener.class);
		MouseMotionListener[] result = new MouseMotionListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	/**
	 * Adds a <code>MouseWheelListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		mouseWheelListener	the <code>MouseWheelListener</code> to be added.
	 *
	 * @see			java.awt.event.MouseWheelListener
	 * @see			java.awt.event.MouseWheelEvent
	 */
	public synchronized void addMouseWheelListener(MouseWheelListener mouseWheelListener) {
		if (mouseWheelListener == null) return;
		addEventListener(mouseWheelListener);
		this.mouseWheelListener = AWTEventMulticaster.add(this.mouseWheelListener, mouseWheelListener);
		if (getParent() != null) {
			getParent().addMouseWheelListener(mouseWheelListener);
		}
	}

	public synchronized void removeMouseWheelListener(MouseWheelListener mouseWheelListener) {
		this.mouseWheelListener = AWTEventMulticaster.remove(this.mouseWheelListener, mouseWheelListener);
		if (getParent() != null) {
			getParent().addMouseWheelListener(mouseWheelListener);
		}
	}

	public MouseWheelListener[] getMouseWheelListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(mouseWheelListener, MouseWheelListener.class);
		MouseWheelListener[] result = new MouseWheelListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	/**
	 * Adds a <code>InputMethodListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		newInputMethodListener	the <code>InputMethodListener</code> to be added.
	 *
	 * @see			java.awt.event.InputMethodListener
	 * @see			java.awt.event.InputMethodEvent
	 */
	public synchronized void addInputMethodListener(InputMethodListener newInputMethodListener) {
		if (newInputMethodListener == null) return;
		addEventListener(newInputMethodListener);
		this.inputMethodListener = AWTEventMulticaster.add(this.inputMethodListener, newInputMethodListener);
	}

	public synchronized void removeInputMethodListener(InputMethodListener inputMethodListener) {
		this.inputMethodListener = AWTEventMulticaster.remove(this.inputMethodListener, inputMethodListener);
	}

	public InputMethodListener[] getInputMethodListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(inputMethodListener, InputMethodListener.class);
		InputMethodListener[] result = new InputMethodListener[listeners.length];
		System.arraycopy(listeners, 0, result, 0, listeners.length);
		return result;
	}

	public InputMethodRequests getInputMethodRequests() {
		// TODO PPro 1.1 - update when input methods support available
		return null;
	}

	public InputContext getInputContext() {
		if (getParent() == null) return null;
		return getParent().getInputContext();
	}

	protected void processMouseWheelEvent(MouseWheelEvent mouseWheelEvent) {
		if (mouseWheelListener == null)
			return;

		switch (mouseWheelEvent.getID()) {
			case MouseWheelEvent.MOUSE_WHEEL:
				mouseWheelListener.mouseWheelMoved(mouseWheelEvent);
			break;
			default:
				return;
		}
	}

	protected void processInputMethodEvent(InputMethodEvent inputMethodEvent) {
		if (inputMethodListener == null) {
			return;
		}

		switch (inputMethodEvent.getID()) {
			case InputMethodEvent.INPUT_METHOD_TEXT_CHANGED:
				inputMethodListener.inputMethodTextChanged(inputMethodEvent);
				break;
			case InputMethodEvent.CARET_POSITION_CHANGED:
				inputMethodListener.caretPositionChanged(inputMethodEvent);
				break;
			default:
				return;
		}
	}

	public boolean isFocusable() {
		return focusable;
	}

	boolean isFocusableImpl() {
	    // By default, every widget is focusable.  If focusability has not
	    // been overridden by the application, the accept method of the
	    // DefaultFocusTraversalPolicy will ignore Labels, Canvases, etc.
	    // during focus traversal.
	    return true;
	}

	boolean isValidTraversalKey(int id) {
	    return ((id == KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS));
	}

	boolean isValidTraversalKeystrokes(int id, Set keystrokes) {
		if (keystrokes == null) return false;
		Iterator iterator = keystrokes.iterator();
		while (iterator.hasNext()) {
			Object next = iterator.next();
			if (next == null) return false;
			AWTKeyStroke keystroke = null;
			if (next instanceof AWTKeyStroke) {
				keystroke = (AWTKeyStroke) next;
			}
			if (keystroke == null) {
			    return false;
			}
			if (keystroke.getKeyEventType() == KeyEvent.KEY_TYPED) return false;
			for (int i=0; i<focusTraversalKeys.length; i++) {
			    if (i == id) continue;
			    Set traversalKeys = focusTraversalKeys[i];
			    if (traversalKeys != null) {
			        if (traversalKeys.contains(keystroke)) return false;
			    }
			}
		}
		Component parent = getParent();
		if (parent != null) return parent.isValidTraversalKeystrokes(id, keystrokes);
		else return true;
	}

	/**
	 * Sets the focusable state of this Component to the specified value.
	 * This value overrides the Component's default focusability.
	 */
	public void setFocusable(boolean focusable) {
		isFocusTraversableOverridden = 1;
	    boolean oldFocusable = this.focusable;
		this.focusable = focusable;
		changeSupport.firePropertyChange("focusable", new Boolean(oldFocusable), new Boolean(focusable));

		if ((focusable == false) && (containsFocus())) {
			removeFocus();
		}
	}

	/**
	 * Sets the focus traversal keys for a given traversal operation for this
	 * Component.  The default values for a Component's focus traversal keys are
	 * implementation-dependent. Sun recommends that all implementations for a
	 * particular native platform use the same default values. The recommendations
	 * for Windows and Unix are:
	 * 		FORWARD_TRAVERSAL_KEYS - TAB and CTRL-TAB on KEY_PRESSED
	 *		BACKWARD_TRAVERSAL_KEYS - SHIFT-TAB and CTRL-SHIFT-TAB on KEY_PRESSED
	 *		UP_CYCLE_TRAVERSAL_KEYS - None
	 *
	 * To disable a traversal key, use an empty Set; Collections.EMPTY_SET is
	 * recommended.
	 *
	 * Using the AWTKeyStroke API, client code can specify on which of two specific
	 * KeyEvents, KEY_PRESSED or KEY_RELEASED, the focus traversal operation will
	 * occur. Regardless of which KeyEvent is specified, however, all KeyEvents
	 * related to the focus traversal key, including the associated KEY_TYPED event,
	 * will be consumed, and will not be dispatched to any Component. It is a runtime
	 * error to specify a KEY_TYPED event as mapping to a focus traversal operation,
	 * or to map the same event to multiple default focus traversal operations.
	 *
	 * If a value of null is specified for the Set, this Component inherits the Set
	 * from its parent. If all ancestors of this Component have null specified for
	 * the Set, then the current KeyboardFocusManager's default Set is used.
	 *
	 * Parameters: id - one of KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
	 * KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS, or KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS,
	 * keystrokes - the Set of AWTKeyStroke for the specified operation
	 *
	 * Throws: IllegalArgumentException - if id is not one of KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
	 * KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS, or KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS,
	 * or if keystrokes contains null, or if any Object in keystrokes is not an AWTKeyStroke, or if
	 * any keystroke represents a KEY_TYPED event, or if any keystroke already maps to another focus
	 * traversal operation for this Component.
	 */
	public void setFocusTraversalKeys(int id, Set keystrokes) {
	    if (!isValidTraversalKey(id)) throw new IllegalArgumentException("invalid focus traversal key identifier");
	    if ((keystrokes != null) && !isValidTraversalKeystrokes(id, keystrokes)) throw new IllegalArgumentException("invalid focus traversal keystrokes");
	    Set oldKeystrokes = focusTraversalKeys[id];
	    focusTraversalKeys[id] = keystrokes;
	    if (id == KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) {
	    	changeSupport.firePropertyChange("forwardFocusTraversalKeys", oldKeystrokes, keystrokes);
	    } else if (id == KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS) {
	    	changeSupport.firePropertyChange("backwardFocusTraversalKeys", oldKeystrokes, keystrokes);
	    } else if (id == KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS) {
	    	changeSupport.firePropertyChange("upCycleFocusTraversalKeys", oldKeystrokes, keystrokes);
	    } else if (id == KeyboardFocusManager.DOWN_CYCLE_TRAVERSAL_KEYS) {
	    	changeSupport.firePropertyChange("downCycleFocusTraversalKeys", oldKeystrokes, keystrokes);
	    }
	}

	/**
	 * Returns the Set of focus traversal keys for a given traversal operation for
	 * this Component. (See setFocusTraversalKeys for a full description of each key.)
	 * If a Set of traversal keys has not been explicitly defined for this Component,
	 * then this Component's parent's Set is returned. If no Set has been explicitly
	 * defined for any of this Component's ancestors, then the current
	 * KeyboardFocusManager's default Set is returned.
	 */
	public Set getFocusTraversalKeys(int id) {
	    if (!isValidTraversalKey(id)) throw new IllegalArgumentException("invalid focus traversal key identifier");

	    Component comp = this;
	    do {
			Set keys = comp.focusTraversalKeys[id];
			if (keys != null) return keys;
			comp = comp.parent;
	    } while (comp != null);

		return getDefaultFocusTraversalKeys(id);
	}

	Set getDefaultFocusTraversalKeys(int id) {
		return KeyboardFocusManager.getCurrentKeyboardFocusManager().getDefaultFocusTraversalKeys(id);
	}

	public boolean areFocusTraversalKeysSet(int id) {
	    if (!isValidTraversalKey(id)) throw new IllegalArgumentException("invalid focus traversal key identifier");
		Set keys = focusTraversalKeys[id];
		return (keys != null);
	}

	/**
	 * Sets whether focus traversal keys are enabled for this Component.
	 * Components for which focus traversal keys are disabled receive key events
	 * for focus traversal keys. Components for which focus traversal keys are
	 * enabled do not see these events; instead, the events are automatically
	 * converted to traversal operations.
	 */
	public void setFocusTraversalKeysEnabled(boolean focusTraversalKeysEnabled) {
		boolean oldEnabled = this.focusTraversalKeysEnabled;
		this.focusTraversalKeysEnabled = focusTraversalKeysEnabled;
		changeSupport.firePropertyChange("focusTraversalKeysEnabled", new Boolean(oldEnabled), new Boolean(focusable));
	}

	/**
	 * Returns whether focus traversal keys are enabled for this Component.
	 * Components for which focus traversal keys are disabled receive key
	 * events for focus traversal keys. Components for which focus traversal
	 * keys are enabled do not see these events; instead, the events are
	 * automatically converted to traversal operations.
	 */
	public boolean getFocusTraversalKeysEnabled() {
		return focusTraversalKeysEnabled;
	}

	/**
	 * Requests that this Component get the input focus, and that this Component's
	 * top-level ancestor become the focused Window. This component must be displayable,
	 * visible, and focusable for the request to be granted. Every effort will be made to
	 * honor the request; however, in some cases it may be impossible to do so. Developers
	 * must never assume that this Component is the focus owner until this Component
	 * receives a FOCUS_GAINED event. If this request is denied because this Component's
	 * top-level Window cannot become the focused Window, the request will be remembered
	 * and will be granted when the Window is later focused by the user.
	 *
	 * This method returns a boolean value. If false is returned, the request is
	 * guaranteed to fail. If true is returned, the request will succeed unless it
	 * is vetoed, or an extraordinary event, such as disposal of the Component's peer,
	 * occurs before the request can be granted by the native windowing system. Again,
	 * while a return value of true indicates that the request is likely to succeed,
	 * developers must never assume that this Component is the focus owner until this
	 * Component receives a FOCUS_GAINED event.
	 *
	 * This method cannot be used to set the focus owner to no Component at all. Use
	 * KeyboardFocusManager.clearGlobalFocusOwner() instead.
	 *
	 * Because the focus behavior of this method is platform-dependent, developers are
	 * strongly encouraged to use requestFocusInWindow when possible.
	 *
	 * Every effort will be made to ensure that FocusEvents generated as a result of
	 * this request will have the specified temporary value. However, because specifying
	 * an arbitrary temporary state may not be implementable on all native windowing
	 * systems, correct behavior for this method can be guaranteed only for lightweight
	 * Components. This method is not intended for general use, but exists instead as a
	 * hook for lightweight Component libraries, such as Swing.
	 *
	 * Returns: false if the focus change request is guaranteed to fail; true if it
	 * is likely to succeed
	 */
	protected boolean requestFocus(boolean temporary) {
	    if (!isDisplayable() || !isFocusable() || !isVisible()) return false;
		if (!isLightweight() && (0 == peer)) return false;
		Window window = getWindow();
		if (null == window) return false;
		// Handle the cross-window focus behavior for ppro1.1.
	    if (!window.isFocused()) window.toFront();
		// TODO PPro 1.1 Focus - Containers can have focus, need to support
		// allowing Constainer to have focus. Currently, calling
		// widgetRequestFocus forces focus to the first child component in
		// the Container.  Bug 109519.
	    if (KeyboardFocusManager.getDebug()) System.out.println("impl request focus " + this.getName());
		return implRequestFocus(temporary); // __widgetRequestFocus() does not handle lightweights, so call this instead
	}

	/**
	 * @see Component#requestFocusInWindow(boolean)
	 */
	public boolean requestFocusInWindow() {
		return requestFocusInWindow(false);
	}

	/**
	 * Request input focus for this Component.  The Component's window must be the focused window.
	 * In order to gain focus, the component must be displayable, visible, and focusable.
	 * Note that this Component will become the focus owner only when it receives a FOCUS_GAINED
	 * event since the focus request may be denied (e.g., vetoed).
	 *
	 * A return value of false indicates that the request is impossible to fulfill. A return value
	 * of true indicates that an attempt to fulfill the request will be honored; however, the
	 * only indication that the request was fulfilled is if this Component receives a FOCUS_GAINED
	 * event.
	 *
	 */
	protected boolean requestFocusInWindow(boolean temporary) {
	    if (!isDisplayable() || !isFocusable() || !isVisible()) return false;
	    if (!getWindow().isFocused()) return false;
	    return requestFocus(temporary);
	}

	/**
	 * Returns the Container which is the focus cycle root of this Component's
	 * focus traversal cycle. Each focus traversal cycle has only a single
	 * focus cycle root and each Component which is not a Container belongs
	 * to only a single focus traversal cycle. Containers which are focus cycle
	 * roots belong to two cycles: one rooted at the Container itself, and one
	 * rooted at the Container's nearest focus-cycle-root ancestor. For such
	 * Containers, this method will return the Container's nearest
	 * focus-cycle-root ancestor.
	 */
	public Container getFocusCycleRootAncestor() {
		Container parent = getParent();
		if (parent == null) return null;
		if (parent.isFocusCycleRoot()) return parent;
		return parent.getFocusCycleRootAncestor();
	}

	/**
	 * Returns whether the specified Container is the focus
	 * cycle root of this Component's focus traversal cycle.
	 */
	public boolean isFocusCycleRoot(Container container) {
		if (container == null) return false;
	    if (!container.isFocusCycleRoot()) return false;
		return isFocusCycleRootImpl(container);
	}

	boolean isFocusCycleRootImpl(Container container) {
		Component[] components = container.getComponents();
		for (int i=0; i<components.length; i++) {
		    Component cmpo = components[i];
		    if (cmpo == this) return true;
		    if (cmpo instanceof Container) {
		        boolean result = isFocusCycleRootImpl((Container)cmpo);
		        if (result) return true;
		    }
		}
		return false;
	}

	/*
	 * Move keyboard focus to the previous traversable component.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void transferFocusBackward() {
	    KeyboardFocusManager.getCurrentKeyboardFocusManager().focusPreviousComponent(this);
	}

	public void transferFocusUpCycle() {
	    KeyboardFocusManager.getCurrentKeyboardFocusManager().upFocusCycle(this);
	}

	public boolean isFocusOwner() {
		return hasFocus();
	}

	public synchronized void addPropertyChangeListener(PropertyChangeListener propertyChangeListener) {
		if (propertyChangeListener == null) return;
		changeSupport.addPropertyChangeListener(propertyChangeListener);
	}

	public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener) {
		changeSupport.removePropertyChangeListener(propertyChangeListener);
	}

	public PropertyChangeListener[] getPropertyChangeListeners() {
		return changeSupport.getPropertyChangeListeners();
	}

	protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		changeSupport.firePropertyChange(propertyName, oldValue, newValue);
	}

	void _postMouseWheelEvent(int type, int x2, int y2, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger, int scrollType, int scrollAmount, int wheelRotation) {
		long time = System.currentTimeMillis();
		_generateMouseWheelEvent(time, x, y, awtButton, awtModifiers, clickCount, popupTrigger, scrollType, scrollAmount, wheelRotation);
	}

	void _generateMouseWheelEvent(long time, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger, int scrollType, int scrollAmount, int wheelRotation) {
		postEvent(new MouseWheelEvent(this, MouseEvent.MOUSE_WHEEL, time, 0, x, y, 0, false, scrollType, scrollAmount, wheelRotation));
	}

	static class GraphicsReaper implements Runnable {

		private Thread thread;

		private List componentsToReap = new ArrayList();

		public void addComponent(Component c) {
			synchronized (componentsToReap) {
				if (componentsToReap.contains(c))
					return;
				componentsToReap.add(c);
				if (thread == null) {
					thread = new Thread(this);
					thread.start();
				}
			}
		}

		public void run() {
			int pos = 0;
			Component next = null;

			while (true) {

				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}

				synchronized (componentsToReap) {
					if (componentsToReap.isEmpty()) {
						// At this point no more components will be
						// processed by this thread so make it
						// look like no thread is currently running
						thread = null;
						break;
					} else {
						if (pos >= componentsToReap.size()) {
							pos = 0;
							continue;
						} else {
							next = (Component) componentsToReap.get(pos++);
						}
					}
				}

				synchronized (next.availableGraphics) {
					if (next.availableGraphics.size() > 0) {
						Handle handle = (Handle) next.availableGraphics.remove(0);
						OS.Graphics_Dispose(handle.value);
					} else {
						synchronized (componentsToReap) {
							componentsToReap.remove(--pos);
						}
					}
				}
			}
		}

	}

}
