package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.io.*;
import java.util.EventListener;
import java.util.Set;

import com.ibm.oti.awt.image.BufferedImageHandler;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.oti.awt.metal.image.ImagePeer;
import com.ibm.oti.awt.metal.widgets.Display;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>Container</code> is the superclass of all widget
 * classes whose instances are components that can hold
 * other components.
 *
 * @author      IBM
 * @version		initial
 */
public class Container extends Component {

	static final long serialVersionUID = 4613797578919906343L;

	// The children of the receiver
	Component[] component;

	// The number of children of the receiver
	int ncomponents;

	// The layout manager of this component
	LayoutManager layoutMgr;

	// The container listener of the receiver
	transient ContainerListener containerListener;

	// Lightweight support
	transient Component focusComponent;
	transient boolean isTemporaryFocusRequest = false;

	transient Insets insets;
	transient Point graphicsOrigin;

	Dimension maxSize;
	Object dispatcher;

	int containerSerializedDataVersion;

	boolean focusCycleRoot = false;
	transient boolean focusPolicySet = false;
	transient FocusTraversalPolicy focusTraversalPolicy;

	/**
	 * Constructs a new <code>Container</code> with specified layout
	 * manager.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	Container(LayoutManager layoutManager) {
		this();
		layoutMgr = layoutManager;
	}

	/**
	 * Constructs a new <code>Container</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Container() {
		super();
	}

	String classNonlocalizedName() {
		//return "container"; //$NON-NLS-1$
		return null; // Match RI behavior
	}

	/**
	 * Answers the number of components added to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> which is the number of components in the receiver
	 *
	 * @deprecated	use <code>getComponentCount()</code> instead
	 *
	 * @see		#getComponentCount
	 */
	public int countComponents() {
		return getComponentCount();
	}

	/**
	 * Answers the component at the specified position.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	position	the index of the desired component
	 *
	 * @return	an <code>Component</code> which is the component at the given position
	 *
	 * @throws	ArrayIndexOutOfBoundsException	if the specified position is smaller than
	 *											<code>0</code> or greater than the number of children
	 */
	public Component getComponent(int position) {
		if (position >= ncomponents)
			throw new ArrayIndexOutOfBoundsException();
		if (component == null)
			throw new ArrayIndexOutOfBoundsException();

		return component[position];
	}

	/**
	 * Answer the <code>Component</code> at the specified
	 * x and y location.  In the case of a <code>Container</code> if a child
	 * component contains the specified x and y the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the x coordinate relative to the receiver
	 * @param		y		the y coordinate relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @see			#getComponentAt(java.awt.Point)
	 * @see			java.awt.Component#contains(int, int)
	 */
	public Component getComponentAt(int x, int y) {
		synchronized (getTreeLock()) {
			if (!contains(x, y))
				return null;

			for (int i=0; i<ncomponents; i++) {
				Component comp = component[i];
				if (comp.contains(x - comp.getX(), y - comp.getY()))
					return comp;
			}

			return this;
		}
	}

	/**
	 * Answer the <code>Component</code> at the
	 * x and y location specified by <code>point</code>.
	 * In the case of a <code>Container</code> if a child
	 * component contains the specified x and y the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point  The location relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @see			#getComponentAt(int, int)
	 */
	public Component getComponentAt(Point point) {
		return super.getComponentAt(point);
	}

	/**
	 * Answers the number of components added to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> which is the number of components in the receiver
	 *
	 * @see		#getComponent(int)
	 */
	public int getComponentCount() {
		return ncomponents;
	}

	/**
	 * Answers the components added to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Component[]</code> which cotains the receiver's components
	 */
	public Component[] getComponents() {
		synchronized (getTreeLock()) {
			Component result[] = new Component[ncomponents];
			if (ncomponents > 0)
				System.arraycopy(component, 0, result, 0, ncomponents);
			return result;
		}
	}

	/**
	 * Answers a <code>Insets</code> representing the size of the
	 * receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Insets</code> which represents the size of the receiver's decorations
	 */
	public Insets getInsets() {
		return getInsetsImpl();
	}

	/**
	 * Answers the size of the decorations at the top, left, bottom
	 * and right of the container.
	 *
	 * @return the size of the decorations at the top, left, bottom
	 * 	and right of the container.
	 */
	Insets getInsetsImpl() {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				return new Insets(0, 0, 0, 0);
			}
			if (insets != null) {
				return insets;
			}

			int[] insetArray = OS.Scrollable_GetInsets(peer);
			insets = new java.awt.Insets(
				insetArray[OS.INDEX_TOP],
				insetArray[OS.INDEX_LEFT],
				insetArray[OS.INDEX_BOTTOM],
				insetArray[OS.INDEX_RIGHT]);
			return insets;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers a <code>Insets</code> representing the size of the
	 * receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Insets</code> which represents the size of the receiver's decorations
	 *
	 * @deprecated	use <code>getInsets()</code> instead
	 *
	 * @see		#getInsets
	 */
	public Insets insets() {
		return getInsets();
	}

	/**
	 * Adds the specified <code>Component</code> to the end
	 * of the receiver's list.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to be added to the receiver
	 *
	 * @return	a Component which is argument passed to this method
	 */
	public Component add(Component component) {
		addImpl(component, null, -1);
		return component;
	}

	/**
	 * Adds the specified <code>Component</code> at the specified
	 * position.  If <code>position</code> is <code>-1</code>, the component
	 * is added to the end of the receiver.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to be added to the receiver
	 * @param	position	the position of the new component
	 *
	 * @return	a Component which is argument passed to this method
	 */
	public Component add(Component component, int position) {
		addImpl(component, null, position);
		return component;
	}

	/**
	 * Adds the specified <code>Component</code> using the specified
	 * constraints object for layout purposes.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to be added to the receiver
	 * @param	constraints	the object representing layout constraints
	 */
	public void add(Component component, Object constraints) {
		addImpl(component, constraints, -1);
	}

	/**
	 * Adds the specified <code>Component</code> at the specified
	 * position.  The specified constraints object is used for layout
	 * purposes.
	 * <p>
	 * If <code>position</code> is <code>-1</code>, the component
	 * is added to the end of the receiver.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to be added to the receiver
	 * @param	constraints	the object representing layout constraints
	 * @param	position	the position of the new component
	 */
	public void add(Component component, Object constraints, int position) {
		addImpl(component, constraints, position);
	}

	/**
	 * Adds the specified <code>Component</code> using the specified
	 * <code>String</code> for layout purposes.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to be added to the receiver
	 * @param	name		the object representing layout constraints
	 */
	public Component add(String name, Component component) {
		addImpl(component, name, -1);
		return component;
	}

	/**
	 * Adds the specified <code>Component</code> at the specified
	 * position.  The specified constraints object is used for layout
	 * purposes.
	 * <p>
	 * If <code>position</code> is <code>-1</code>, the component
	 * is added to the end of the receiver.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 * <p>
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	newComponent	the component to be added to the receiver
	 * @param	constraints		the object representing layout constraints
	 * @param	position		the position of the new component
	 *
	 * @throws	IllegalArgumentException	if the specified position is
	 *										out of range and different from <code>-1</code>
	 */
	protected void addImpl(Component comp, Object constraints, int position) {
		synchronized (getTreeLock()) {
			if (comp instanceof Window)
				throw new IllegalArgumentException("can't add window to a container"); //$NON-NLS-1$

			if (comp == this)
				throw new IllegalArgumentException();

			if (comp instanceof Container && ((Container)comp).isAncestorOf(this))
				throw new IllegalArgumentException();

			if ((position < -1) || (position > ncomponents))
				throw new IllegalArgumentException();

			// Remove the component from its container
			if (comp.parent != null) {
				comp.parent.remove(comp);
			}

			if (component == null) {
				ncomponents = 0;
				component = new Component[4];
			}

			// Need to recheck this since the component array might
			// have changed if comp.parent happens to be the receiver
			if (position > ncomponents)
				throw new IllegalArgumentException();

			comp.parent = this;

			// -1 means end of the list
			if (position == -1)
				position = ncomponents;

			// Grow the list if necessary
			if (ncomponents == component.length) {
				Component newComponents[] = new Component[component.length + 4];
				System.arraycopy(component, 0, newComponents, 0, ncomponents);
				component = newComponents;
			}
			// Move components if this is an insertion
			if (position != ncomponents) {
				System.arraycopy(component, position, component, position + 1, ncomponents - position);
			}
			component[position] = comp;
			ncomponents++;

			// Create the widget if the receiver is displayable
			if (isDisplayable()) {
				comp.addNotify();
			}

			if (layoutMgr != null) {
				if (layoutMgr instanceof LayoutManager2) {
					((LayoutManager2)layoutMgr).addLayoutComponent(comp, constraints);
				} else {
					if (constraints == null || constraints instanceof String) {
						layoutMgr.addLayoutComponent((String)constraints, comp);
					}
				}
			}

			processEvent(new ContainerEvent(this, ContainerEvent.COMPONENT_ADDED, comp));
			//validate();  // bug 98331 --> caused other problems
		}
	}

	/**
	 * Removes the component at the specified position from the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	position	the position of the component to be removed from the receiver
	 */
	public void remove(int position) {
		synchronized (getTreeLock()) {
			if (position < 0 || position >= ncomponents)
				throw new IllegalArgumentException();

			Component comp = component[position];

			if (comp.containsFocus()){
				comp.removeFocus();
			}

			if (comp.isDisplayable())
				comp.removeNotify();

			LayoutManager layout = getLayout();
			if (layout != null) {
				layout.removeLayoutComponent(comp);
			}

			comp.parent = null;
			ncomponents--;
			if (ncomponents == 0) {
				component = null;
			} else {
				if (position != ncomponents) {
					System.arraycopy(component, position + 1, component, position, ncomponents - position);
				}
				component[ncomponents] = null;
			}

			invalidate();
			processEvent(new ContainerEvent(this, ContainerEvent.COMPONENT_REMOVED, comp));
		}
	}

	/**
	 * Answer the index of the specified component.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	component	the component to search for
	 *
	 * @return	an <code>int</code> which is the index of the component or <code>-1</code>
	 *
	 */
	int indexOf(Component comp) {
		for (int i=ncomponents-1; i>=0; i--) {
			if (component[i] == comp) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Removes the specified <code>Component</code> from the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	comp	the component to be removed from the receiver
	 */
	public void remove(Component comp) {
		if (comp == null) throw new NullPointerException();
		synchronized (getTreeLock()) {
			int position = indexOf(comp);
			if (position != -1)
				remove(position);
		}
	}

	/**
	 * Destroy the underlying widget for the receiver and its children.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void removeNotify() {
		synchronized(getTreeLock()) {
			if (component != null) {
				for (int i=0; i < component.length; i++) {
					if (component[i] != null) {
						component[i].removeNotify();
//						System.out.println("Container.removeNotify1");
					}
				}
			}
		}
		super.removeNotify();
	}

	/**
	 * Removes all components from the receiver's list.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void removeAll() {
		synchronized (getTreeLock()) {
			for (int i=ncomponents-1; i>=0; i--) {
				remove(i);
			}
		}
	}

	/**
	 * Answers the <code>LayoutManager</code> which lays out
	 * the children for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>LayoutManager</code> which is the layout manager for the receiver
	 */
	public LayoutManager getLayout() {
		return layoutMgr;
	}

	/**
	 * Sets the <code>LayoutManager</code> which lays out
	 * the children for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	layoutManager	the layout manager for the receiver
	 */
	public void setLayout(LayoutManager layoutManager) {
//		synchronized (this) {
			this.layoutMgr = layoutManager;
			if (isValid()) {
				invalidate();
			}
//		}
	}

	void setBoundsImpl(int x, int y, int width, int height, boolean repaint) {
		super.setBoundsImpl(x, y, width, height, repaint);
		insets = null;

		if (isLightweight()) {
			for (int i=0; i<ncomponents; i++) {
				Component comp = component[i];
				comp.setBoundsImpl(comp.getX(), comp.getY(), comp.getWidth(), comp.getHeight(), false);
			}
		}
	}

	void setVisibleImpl(boolean visible, boolean repaint) {
		super.setVisibleImpl(visible, repaint);

		if (isLightweight()) {
			for (int i=0; i<ncomponents; i++) {
				component[i].setVisibleImpl(visible, false);
			}
		}
	}

	/**
	 * If the receiver has a <code>LayoutManager</code> ask
	 * it to layout the receiver's children.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void doLayout() {
		LayoutManager layout = getLayout();
		if (layout == null) return;

		synchronized (getTreeLock()) {
			setRedraw(false);
			layout.layoutContainer(this);
			setRedraw(true);
		}
	}

	/**
	 * If the receiver has a <code>LayoutManager</code> ask
	 * it to layout the receivers children.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @deprecated	use <code>doLayout()</code> instead
	 *
	 * @see		#doLayout
	 */
	public void layout() {
		doLayout();
	}

	/**
	 * Mark this component and all its parents to
	 * require that they be laid out.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void invalidate() {
		synchronized (getTreeLock()) {
			super.invalidate();

			LayoutManager layout = getLayout();
			if (layout instanceof LayoutManager2) {
				((LayoutManager2)layout).invalidateLayout(this);
			}
		}
	}

	/**
	 * Make sure that the receiver and all its children if
	 * applicable are properly laid out.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void validate() {
		// Avoid getting the tree lock if the
		// widget is already valid
		if (isValid()) return;

		synchronized (getTreeLock()) {
			if (isDisplayable() == false) return;
			setRedraw(false);
			validateTree();
			// The container should be validated even
			// if validateTree was overrideen to do nothing
			super.validateTree();
			setRedraw(true);
		}
	}

	void callValidateTree() {
		if (isValid() == false) {
			validateTree();
		} else {
			super.callValidateTree();
		}
	}

	/**
	 * Validates the receiver's widget tree recursively.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	protected void validateTree() {
		if (isValid()) return;

		layout();
		for (int i=0; i<ncomponents; i++) {
			component[i].callValidateTree();
		}

		// Make sure this widgets gets
		// marked as validated
		super.validateTree();
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	Dimension getPreferredSizeImpl() {
		// Always call super in order to allow the platform an
		// opportunity to describe the preferred size. This is
		// particularly important for Frames where a platform
		// may restrict how small it can be.
		// If there is a layout, also get the preferred size from
		// the layout and return the max of the preferred sizes.
		LayoutManager layout = getLayout();
		Dimension pref = super.getPreferredSizeImpl();
		if (layout == null) {
			return pref;
		}

		Dimension layoutPref = layout.preferredLayoutSize(this);
		return new Dimension(Math.max(layoutPref.width,pref.width),Math.max(layoutPref.height,pref.height));
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	Dimension getMinimumSizeImpl() {
		// Always call super in order to allow the platform an
		// opportunity to describe the minimum size. This is
		// particularly important for Frames where a platform
		// may restrict how small it can be.
		// If there is a layout, also get the minimum size from
		// the layout and return the max of the minimum sizes.
		LayoutManager layout = getLayout();
		Dimension min = super.getMinimumSizeImpl();
		if (layout == null) {
			return min;
		}

		Dimension layoutMin = layout.minimumLayoutSize(this);
		return new Dimension(Math.max(layoutMin.width,min.width),Math.max(layoutMin.height,min.height));
	}

	/**
	 * Answer a <code>Dimension</code> representing the maximum
	 * size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the maximum size of the receiver
	 */
	public Dimension getMaximumSize() {
		LayoutManager layout = getLayout();
		if (!(layout instanceof LayoutManager2))
			return super.getMaximumSize();

		return ((LayoutManager2)layout).maximumLayoutSize(this);
	}

	/**
	 * Answer the preferred horizontal alignment for
	 * the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getAlignmentX() {
		LayoutManager layout = getLayout();
		if (!(layout instanceof LayoutManager2))
			return super.getAlignmentX();

		return ((LayoutManager2)layout).getLayoutAlignmentX(this);
	}

	/**
	 * Answer the preferred vertical alignment for
	 * the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>float</code> which is the alignment value
	 */
	public float getAlignmentY() {
		LayoutManager layout = getLayout();
		if (!(layout instanceof LayoutManager2))
			return super.getAlignmentY();

		return ((LayoutManager2)layout).getLayoutAlignmentY(this);
	}

	boolean lastComponentWithEventsIsLightweightChild() {
		Component c = Display.getLastComponentWithEvents();
		Component nativeParent = (isLightweight() ? getNativeParent() : this);
		return (c != null && c.getNativeParent() == nativeParent);
	}

	Component findLightweightAt(int x, int y) {
		Component child = findComponentAt(x, y);
		if (child != null && child != this && child.isLightweight()) {
			return child;
		}
		return null;
	}

	/**
	 * This method can be called by the platform to determine
	 * if the menu is actually available.
	 *
	 * @return one of the INFO_HASMENU_* flags
	 */
	int hasMenuInfo(int x, int y) {
		Component c = findLightweightAt(x, y);
		if ((c != null) && (c != this)) {
			return c.hasMenuInfo(x - this.x, y - this.y);
		} else {
			return super.hasMenuInfo(x,y);
		}
	}

	void _postMouseEvent(int type, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		// UGL defines incoming pointer events to be relative to the
		// client area.  The incoming pointer event needs to be adjusted
		// to be relative to the window origin like PPro expects.
		// getInsetsImpl() is used instead of getInsets() because
		// getInsets() can be overridden by subclasses but the only
		// thing that needs to be adjusted for is the native insets.
		Insets insets = getInsetsImpl();
		x = insets.left + x;
		y = insets.top + y;

		super._postMouseEvent(type, x, y, awtButton, awtModifiers, clickCount, popupTrigger);
	}

	void _generateMouseEnteredEvent(long time, int x, int y) {
		Component lwChild = findLightweightAt(x, y);
		if (lwChild == null) {
			super._generateMouseEnteredEvent(time, x, y);
		} else {
			Point coords = new Point(x, y);
			translateCoords(this, lwChild, coords);
			lwChild._generateMouseEnteredEvent(time, coords.x, coords.y);
			if (Display.getLastGrabbedComponentWithEvents() == null) {
				resetCursor(lwChild);
			}
		}
		// Make sure that enter events aren't posted twice
		// (synthesized for lightweights during subsequent MouseMove)
		Display.setLastComponentWithEvents(lwChild == null ? this : lwChild);
	}

	void _generateMouseExitedEvent(long time, int x, int y) {
		// Note that the location of the exit event is not guaranteed
		// to be within the bounds of the component for which it occurred.
		// Therefore, the location can't be used reliably to determine
		// whether the exit occurred over a lightweight child.
		// Instead, the assumption is made that the exit is occurring for
		// the last component which was entered as tracked by Display.
		Component lastEntered = Display.getLastComponentEntered();
		if (lastEntered == this) {
			Display.setLastComponentWithEvents(null);
			super._generateMouseExitedEvent(time, x, y);
			return;
		}

		// At this point, the assumption is that if the heavyweight was not the
		// last component entered, yet it is receiving an exit event, then the
		// last component entered must be a lightweight child of this heavyweight.
		// Therefore, generate the exit event for the child. This would handle the
		// case where a lightweight occupies up to the edge of the heavyweight at
		// which the exit is occuring.
		Point exitCoords = new Point(x, y);
		translateCoords(this, lastEntered, exitCoords);
		lastEntered._generateMouseExitedEvent(time, exitCoords.x, exitCoords.y);

		Display.setLastComponentWithEvents(null);
		// If a grab is not currently in affect, then restore the heavyweight's
		// cursor in case it had been modified by the lightweight being exited.
		// This ensures the cursor is correct when the heavyweight is re-entered.
		// Note that it is important to first clear the lastComponentWithEvents.
		// See _setCursor(Cursor) for details.
		if (Display.getLastGrabbedComponentWithEvents() == null) {
			resetCursor(this);
		}
	}

	void _generateMouseMovedEvent(long time, int x, int y, int mods) {
		Component lwChild = findLightweightAt(x, y);
		if (lwChild == null) {
			super._generateMouseMovedEvent(time, x, y, mods);
		} else {
			Point coords = new Point(x, y);
			translateCoords(this, lwChild, coords);
			lwChild._generateMouseMovedEvent(time, coords.x, coords.y, mods);
		}
	}

	void _generateMouseDraggedEvent(long time, int x, int y, int mods) {
		// If this component differs from where the grab began, ensure
		// the event occurs, and is relative to, the grabbed component.
		// While normally the grabbed component would receive events directly,
		// if the grab began on a lightweight, the native events would still
		// be arriving on the heavyweight parent.
		Component lastGrabbed = Display.getLastGrabbedComponentWithEvents();
		if (lastGrabbed == this) {
			super._generateMouseDraggedEvent(time, x, y, mods);
		} else {
			Point coords = new Point(x, y);
			translateCoords(this, lastGrabbed, coords);
			lastGrabbed._generateMouseDraggedEvent(time, coords.x, coords.y, mods);
		}
	}

	void _generateMousePressedEvent(long time, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		Component lwChild = findLightweightAt(x, y);
		if (lwChild == null) {
			super._generateMousePressedEvent(time, x, y, awtButton, awtModifiers, clickCount, popupTrigger);
		} else {
			Point coords = new Point(x, y);
			translateCoords(this, lwChild, coords);
			lwChild._generateMousePressedEvent(time, coords.x, coords.y, awtButton, awtModifiers, clickCount, popupTrigger);
		}
	}

	void _generateMouseReleasedEvent(long time, int x, int y, int awtButton, int awtModifiers, int clickCount, boolean popupTrigger) {
		// If the grabbed component is a lightweight, then the cursor may
		// need to be reset for the heavyweight parent. Essentially, unless
		// the cursor is over the lightweight's heavyweight parent or another
		// lightweight with the same heavyweight parent, the cursor needs to
		// be reset on the heavyweight parent of the grabbed lightweight.
		Component lastGrabbed = Display.getLastGrabbedComponentWithEvents();

		// The lastGrabbed component could be null if a
		// drag event was cancelled and the mouse was
		// moved into this component.  On the release
		// after such an event the lastGrabbed component is
		// null and the release event should be generated for
		// the widget that is at the current location
		if (lastGrabbed == null) {
			lastGrabbed = findComponentAt(x,y);
			if (lastGrabbed == null) return;
		}

		if (lastGrabbed.isLightweight()){
			Container startHW = lastGrabbed.getNativeParent();
			Component componentUnderCursor = Display.getLastComponentEntered();
			Component endHW = null;
			if (componentUnderCursor != null) {
				if (componentUnderCursor.isLightweight()) {
					endHW = componentUnderCursor.getNativeParent();
				} else {
					endHW = componentUnderCursor;
				}
			}
			if (startHW != endHW) {
				startHW.resetCursor(startHW);
			}
		}

		// If this component differs from where the grab began, ensure
		// the event occurs, and is relative to, the grabbed component.
		// While normally the grabbed component would receive the release,
		// if the grab began on a lightweight, the native events would still
		// be arriving on the heavyweight parent.
		if (lastGrabbed == this) {
			super._generateMouseReleasedEvent(time, x, y, awtButton, awtModifiers, clickCount, popupTrigger);
		} else {
			Point coords = new Point(x, y);
			translateCoords(this, lastGrabbed, coords);
			lastGrabbed._generateMouseReleasedEvent(time, coords.x, coords.y, awtButton, awtModifiers, clickCount, popupTrigger);
		}
	}

	void _generateMouseEnterExitForLightweights(long time, int x, int y) {
		Component lwChild = findLightweightAt(x, y);

		// If the event is for the heavyweight and we weren't
		// previously in a lightweight child, there's nothing to do.
		if (!lastComponentWithEventsIsLightweightChild() && (lwChild == null)) {
			Display.setLastComponentWithEvents(this);
			return;
		}

		Component lastComponent = Display.getLastComponentWithEvents();
		if (lastComponent != lwChild) {
			// Generate exit event...
			if (lastComponent != null) {
				// Exit the last component making sure to translate
				// the coordinates relative to that component.
				Point exitCoords = new Point(x, y);
				translateCoords(this, lastComponent, exitCoords);
				postEvent(new MouseEvent(lastComponent, MouseEvent.MOUSE_EXITED, time, 0, exitCoords.x, exitCoords.y, 0, false));
			}

			// Generate enter event...
			Point enterCoords = new Point(x, y);
			Component entered = (lwChild == null ? this : lwChild);
			translateCoords(this, entered, enterCoords);
			postEvent(new MouseEvent(entered, MouseEvent.MOUSE_ENTERED, time, 0, enterCoords.x, enterCoords.y, 0, false));
			Display.logComponentEntered(entered);

			if (Display.getLastGrabbedComponentWithEvents() == null) {
				resetCursor(entered);
			}

			Display.setLastComponentWithEvents(entered);
		}
	}

	void _setCursor(final Cursor cursor) {
		if (lastComponentWithEventsIsLightweightChild()) {
			// If the last known mouse location was over a lightweight
			// child component then don't set the cursor as it will
			// result in the lightweight's cursor also being changed.
			// The cursor will eventually be properly set for the container
			// when the mouse moves out of the lightweight child.
		} else {
			super._setCursor(cursor);
		}
	}

	void _postFocusEvent(FocusEvent focusEvent) {
		Window window = getWindow();
		if (window != null) {
			synchronized (window.getFocusManager()) {
				if (focusComponent != null) {
					if (focusEvent.getID() == FocusEvent.FOCUS_GAINED) {
						focusEvent = new FocusEvent(focusComponent, focusEvent.getID(), isTemporaryFocusRequest, focusEvent.getOppositeComponent());
					} else {
						((AWTEvent) focusEvent).modifyEventSource(focusComponent);
					}
				}

				super._postFocusEvent(focusEvent);
			}
		}
	}

	void _postKeyEvent(int type, char character, int keyCode, int stateMask) {
		if (focusComponent != null) {
			focusComponent._postKeyEvent(type, character, keyCode, stateMask);
		} else {
			super._postKeyEvent(type, character, keyCode, stateMask);
		}
	}

	/**
	 * Called when regions of this component need to be redrawn.
	 * The graphics clip region represents the damaged area.
	 * Subclasses that override this method MUST call super.paint(),
	 * or lightweight children will not be painted.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void paint(Graphics graphics) {
		if (isShowing()) {
			super.paint(graphics);

			paintImpl(graphics);
		}
	}

	void paintImpl(Graphics graphics){
		if (graphics == null) return;
		for (int i = ncomponents - 1; i >= 0; i--) {
			Component comp = component[i];
			if (comp.isLightweight() && comp.isVisible() && comp.isDisplayable()) {
				Rectangle compBounds = comp.getBounds();
				Graphics childGraphics = graphics.create(compBounds.x, compBounds.y, compBounds.width, compBounds.height);
				if (childGraphics != null) {
					if (!childGraphics.getClipBounds().isEmpty()) {
						comp.paintAll(childGraphics);
					}
					childGraphics.dispose();
				}
			}
		}
	}

	/**
	 * Print this component on the specfied graphics.
	 * The graphics clip region represents what should be printed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void print(Graphics graphics) {
	}

	/**
	 * Paint the receiver's children on the specified graphics.
	 * The graphics clip region represents what should be painted.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void paintComponents(Graphics graphics) {
		paintImpl(graphics);
	}

	/**
	 * Print the receiver's children on the specified graphics.
	 * The graphics clip region represents what should be printed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	graphics	the graphics context used for drawing
	 */
	public void printComponents(Graphics graphics) {
		if (component == null) return;
		for (int i = 0; i < component.length; i++) {
			Component comp = component[i];
			if (comp != null) comp.printAll(graphics);
		}
	}

	public void printAll(final Graphics graphics) {
		acquirePeerReadLock();
		try {
			if (peer == 0) return;

			Dimension size = getSize();
			Insets insets = getInsets();
			int imgWidth = size.width - insets.left - insets.right;
			int imgHeight = size.height - insets.top - insets.bottom;

			Image image = createImage(imgWidth, imgHeight);
			IImageImpl bbImage= BufferedImageHandler.createIImageImpl(image);

			ImagePeer imagePeer = bbImage.getImagePeer();
			int imageHandle = imagePeer == null ? 0 : imagePeer.fPeerHandle;
			OS.Control_CopyArea(peer, imageHandle, 0, 0, imgWidth, imgHeight);
			graphics.drawImage(image, 0, 0, null);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Adds an <code>ContainerListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		containerListener	the <code>ContainerListener</code> to be added.
	 *
	 * @see			java.awt.event.ContainerListener
	 * @see			java.awt.event.ContainerEvent
	 */
	public synchronized void addContainerListener(ContainerListener containerListener) {
		addEventListener(containerListener);
		this.containerListener = AWTEventMulticaster.add(this.containerListener, containerListener);
	}

	/**
	 * Removes a <code>ContainerListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		cointainerListener	the <code>ContainerListener</code> to be removed.
	 *
	 * @see			java.awt.event.ContainerListener
	 * @see			java.awt.event.ContainerEvent
	 */
	public synchronized void removeContainerListener(ContainerListener cointainerListener) {
		this.containerListener = AWTEventMulticaster.remove(this.containerListener, containerListener);

	}

	/**
	 * Called when a <code>ContainerEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processContainerEvent(ContainerEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	componentEvent	the event to be processed
	 *
	 * @see		java.awt.event.ContainerEvent
	 */
	protected void processContainerEvent(ContainerEvent containerEvent) {
		if (containerListener == null)
			return;

		switch (containerEvent.getID()) {
			case ContainerEvent.COMPONENT_ADDED:
				containerListener.componentAdded(containerEvent);
				break;
			case ContainerEvent.COMPONENT_REMOVED:
				containerListener.componentRemoved(containerEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ContainerEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processContainerEvent(ContainerEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	awtEvent	the event to be processed
	 *
	 * @see #processContainerEvent(java.awt.event.ContainerEvent)
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ContainerEvent) {
			processContainerEvent((ContainerEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Answer the child <code>Component</code> at the specified
	 * x and y location.  If the child is a <code>Container</code> search
	 * its children.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the x coordinate relative to the receiver
	 * @param		y		the y coordinate relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @since		JDK1.2
	 */
	public Component findComponentAt(int x, int y) {
		if (!isVisible()) return null;
		synchronized (getTreeLock()) {
			if (!contains(x, y))
				return null;

			for (int i=0; i<ncomponents; i++) {
				Component comp = component[i];
				if (comp instanceof Container) {
					Component result = ((Container)comp).findComponentAt(x - comp.getX(), y - comp.getY());
					if (result != null)
						return result;
				} else {
					if ((comp.contains(x - comp.getX(), y - comp.getY())) && (comp.isVisible()))
						return comp;
				}
			}

			return this;
		}
	}

	boolean requestFocusChange(boolean next) {
		if (!(isEnabled() && isVisible() && isDisplayable()))
			return false;
		if (primRequestFocus(next)) {
			return true;
		}
		int start, end, offset;
		if (next) {
			offset = 1;
			start = 0;
			end = ncomponents;
		} else {
			offset = -1;
			start = ncomponents - 1;
			end = -1;
		}
		for (int index = start; index != end; index += offset) {
			if (component[index].requestFocusChange(next))
				return true;
		}
		return false;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ContainerEvent)
			return ((eventMask & AWTEvent.CONTAINER_EVENT_MASK) != 0) || containerListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	/**
	 * Answer the child <code>Component</code> at the specified
	 * point.  If the child is a <code>Container</code> search
	 * its children.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the point relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @since		JDK1.2
	 */
	public Component findComponentAt(Point point) {
		return findComponentAt(point.x, point.y);
	}

	/**
	 * Answer the <code>Component</code> at the specified
	 * x and y location.  In the case of a <code>Container</code> if a child
	 * component contains the specified x and y the child is answered.
	 * <p>
	 * The point specified is relative to the receiver.
	 * <p>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the x coordinate relative to the receiver
	 * @param		y		the y coordinate relative to the receiver
	 *
	 * @return		a java.awt.Component or null
	 *
	 * @deprecated	use <code>getComponentAt(int, int)</code> instead
	 *
	 * @see		#getComponentAt(int, int)
	 */
	public Component locate(int x, int y) {
		return getComponentAt(x, y);
	}

	/**
	 * Answer whether the receiver is the parent of the specified
	 * component directly or indirectly.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is the ancestor of the component,
	 *				<code>false</code> otherwise
	 */
	public boolean isAncestorOf(Component component) {
		if (component == null) return false;
		synchronized (getTreeLock()) {
			Component temp = component.parent;
			while (temp != null) {
				if (temp == this) return true;
				temp = temp.parent;
			}
			return false;
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		LayoutManager layout = getLayout();
		return super.paramString() + ((layout == null) ? "" : (",layout=" + layout.getClass().getName())); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print stream indenting each child level
	 * by the specified number of spaces.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printStream	the stream to print on
	 * @param	indent		the number of indenting spaces for child levels
	 *
	 * @see	#list(PrintStream, int)
	 */
	public void list(PrintStream printStream, int indent) {
		super.list(printStream, indent);

		for (int i=0; i<ncomponents; i++) {
			component[i].list(printStream, indent + 2);
		}
	}

	/**
	 * Print a formatted string representation of the receiver and
	 * its children to the specified print writer indenting each child level
	 * by the specified number of spaces.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	printWriter	the writer to print on
	 * @param	indent		the number of indenting spaces for child levels
	 *
	 * @see	#list(PrintWriter, int)
	 */
	public void list(PrintWriter printWriter, int indent) {
		super.list(printWriter, indent);

		for (int i=0; i<ncomponents; i++) {
			component[i].list(printWriter, indent + 2);
		}
	}

	/**
	 * Create the underlying widget for the receiver and its children.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void addNotify() {
		synchronized (getTreeLock()) {
			super.addNotify();

			for (int i=0; i<ncomponents; i++) {
				component[i].addNotify();
			}
		}
	}

	// See bug 109901.  This call will cancel a lightweight
	// focus owner, so be careful when you call it.
	boolean _widgetRequestFocus(FocusManager manager) {
		synchronized (manager) {
			focusComponent = null;
			isTemporaryFocusRequest = false;
			Component oldHwFocus = null;
			Component oldFocus = manager.getCurrentFocus();
			if (this == oldFocus) return false;

			if (oldFocus != null) {
				if (oldFocus.isLightweight())
					oldHwFocus = oldFocus.getNativeParent();
				else
					oldHwFocus = oldFocus;

				if (oldHwFocus == this) {
					manager.requestFocusEmulated(this, oldFocus, false);
					return true;
				}
			}
		}

		return super._widgetRequestFocus(manager);
	}

	void paintCallback(int x, int y, int width, int height) {
		Insets insets = getInsetsImpl();

		// invalidated rectangle will be in UGL coordinates, relative to
		// insets. Adjust them to be relative to container origin.
		// Fixes bug 109222.
		x += insets.left;
		y += insets.top;

		super.paintCallback(x,y,width,height);
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, containerListenerK, containerListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		for (int i=0; i<ncomponents; i++) {
			component[i].parent = this;
		}
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(containerListenerK))
				addContainerListener((ContainerListener)listener);
		}
	}

	boolean _shouldAllowPaint() {
		// containers shouldn't have to be valid to receive paint events
		if (getIgnoreRepaint()) return false;
		return true;
	}

// TODO PPro 1.1 - much to be implemented
	public ContainerListener[] getContainerListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(containerListener, ContainerListener.class);
		ContainerListener[] resultArray = new ContainerListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	boolean isValidTraversalKey(int id) {
	    return ((id == KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS)) ||
		    	(id == KeyboardFocusManager.DOWN_CYCLE_TRAVERSAL_KEYS);
	}

	public void setFocusTraversalKeys(int id, Set keys) {
	    // Container has one more traversal key id than Component, the method
	    // isValidTraversalKey accounts for this difference, so just call super
		super.setFocusTraversalKeys(id, keys);
	}

	public Set getFocusTraversalKeys(int id) {
	    // Container has one more traversal key id than Component, the method
	    // isValidTraversalKey accounts for this difference, so just call super
		return super.getFocusTraversalKeys(id);
	}

	public boolean areFocusTraversalKeysSet(int id) {
	    // Container has one more traversal key id than Component, the method
	    // isValidTraversalKey accounts for this difference, so just call super
		return super.areFocusTraversalKeysSet(id);
	}

	/**
	 * Returns whether the specified Container is the focus cycle root of this
	 * Container's focus traversal cycle. Each focus traversal cycle has only a
	 * single focus cycle root and each Container which is not a focus cycle root
	 * belongs to only a single focus traversal cycle. Containers which are focus
	 * cycle roots belong to two cycles: one rooted at the Container itself, and
	 * one rooted at the Container's nearest focus-cycle-root ancestor. This
	 * method will return true for both such Containers in this case.
	 */
	public boolean isFocusCycleRoot(Container container) {
	    if (!container.isFocusCycleRoot()) return false;
	    if (container == this) return true;
		return super.isFocusCycleRoot(container);
	}

	public void transferFocusBackward() {
	    // TODO PPro 1.1 Focus - need to determine what more this method needs to
	    // do (i.e., why it is overridden)
		// DLW - probably overridden to deal with lightweights
//		if (focusComponent != null) {
//			// DLW - this could probably blow the whole thing up
//			// but seems like the right thing to do.  Need to ask LK
//			focusComponent.transferFocusBackward();
//			return;
//		}
	    super.transferFocusBackward();
	}

	/**
	 * Sets the focus traversal policy, if the Container is a focus cycle root.
	 * If this Container is not a focus cycle root, the policy will be remembered, but
	 * will not be used or inherited until this Container becomes a focus cycle root.
	 */
	public void setFocusTraversalPolicy(FocusTraversalPolicy policy) {
	    if (policy == null) {
	        focusPolicySet = false;
	        return;
	    }
	    FocusTraversalPolicy oldPolicy = this.focusTraversalPolicy;
	    this.focusTraversalPolicy = policy;
	    focusPolicySet = true;
	    changeSupport.firePropertyChange("focusTraversalPolicy", oldPolicy, policy);
	}

	/**
	 * Returns the focus traversal policy.  Return null if the Container is
	 * not a focus cycle root. If isFocusTraversalPolicySet() is false,
	 * ask this Container's focus-cycle-root ancestor for its traversal policy.
	 */
	public FocusTraversalPolicy getFocusTraversalPolicy() {
	    if (!isFocusCycleRoot()) return null;
	    if (!isFocusTraversalPolicySet()) {
	    	Container root = getFocusCycleRootAncestor();
	    	if ((root == this) || (root == null)) {
	    		return KeyboardFocusManager.getCurrentKeyboardFocusManager().getDefaultFocusTraversalPolicy();
	    	} else {
	    		return root.getFocusTraversalPolicy();
	    	}
	    }
		if (focusTraversalPolicy == null) return KeyboardFocusManager.getCurrentKeyboardFocusManager().getDefaultFocusTraversalPolicy();
		return focusTraversalPolicy;
	}

	/**
	 * Returns whether or not the focus traversal policy has been set.
	 */
	public boolean isFocusTraversalPolicySet() {
		return focusPolicySet;
	}

	public void setFocusCycleRoot(boolean flag) {
		boolean oldFlag = focusCycleRoot;
	    focusCycleRoot = flag;
    	changeSupport.firePropertyChange("focusCycleRoot", new Boolean(oldFlag), new Boolean(focusCycleRoot));
	}

	/**
	 * Returns whether this Container is the focus cycle root for this
	 * this Container's traversal cycle. Only one focus cycle root exists per
	 * traversal cycle, and a non focus cycle root container belongs to only
	 * one traversal cycle.  Focus cycle roots belong to two traversal cycles -
	 * one rooted at the Container, the other rooted at the Container's nearest
	 * focus-cycle-root ancestor.
	 */
	public boolean isFocusCycleRoot() {
		return focusCycleRoot;
	}

	public void transferFocusDownCycle() {
		// DLW - probably overridden to deal with lightweights
	    KeyboardFocusManager.getCurrentKeyboardFocusManager().downFocusCycle(this);
	}
}
