package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2005  All Rights Reserved
 */

import java.io.Serializable;
import java.util.ArrayList;
/**
 * A FocusTraversalPolicy that determines traversal order based on the order of child
 * Components in a Container. From a particular focus cycle root, the policy makes a
 * pre-order traversal of the Component hierarchy, and traverses a Container's children
 * according to the ordering of the array returned by Container.getComponents().
 * Portions of the hierarchy that are not visible and displayable will not be searched.
 *
 * By default, ContainerOrderFocusTraversalPolicy implicitly transfers focus down-cycle.
 * That is, during normal forward focus traversal, the Component traversed after a
 * focus cycle root will be the focus-cycle-root's default Component to focus.
 * This behavior can be disabled using the setImplicitDownCycleTraversal method.
 *
 * By default, methods of this class with return a Component only if it is visible,
 * displayable, enabled, and focusable. Subclasses can modify this behavior by
 * overriding the accept method.
 */
public class ContainerOrderFocusTraversalPolicy extends FocusTraversalPolicy implements Serializable {

	private static final long serialVersionUID = 486933713763926351L;

	// The ContainerOrderFocusTraversalPolicy implicitly transfers focus
    // down-cycle (i.e., during forward traversal, if a focus cycle root is
    // encountered, focus will go down to the cycle root's initial component).
    // Set this flag to false to disable the down-cycle behavior.
    boolean implicitDownCycleTraversal;

	public ContainerOrderFocusTraversalPolicy() {
		implicitDownCycleTraversal = true;
	}

	public Component getComponentAfter(Container focusCycleRoot, Component aComponent) {
		if (focusCycleRoot == null) throw new IllegalArgumentException("focusCycleRoot cannot be null");
		if (aComponent == null) throw new IllegalArgumentException("aComponent cannot be null");
		if (focusCycleRoot == aComponent) {
		    if (!focusCycleRoot.isFocusCycleRoot()) throw new IllegalArgumentException("aComponent is not part of focusCycleRoot");
		} else {
			if (!aComponent.isFocusCycleRoot(focusCycleRoot)) throw new IllegalArgumentException("focusCycleRoot is not a focus cyle root of aComponent");
		}

		// get the components in the focus cycle
		ArrayList list = getPreorderFocusCycle(focusCycleRoot, true);
	    // figure out where to start the search
	    int index = list.indexOf(aComponent);
	    int startIndex = index;
	    if (startIndex == -1) {
	        startIndex = 0;
	    } else {
	        startIndex = startIndex + 1;
		    if (aComponent instanceof Container) {
		        Container cont = (Container)aComponent;
		        if (cont.isFocusCycleRoot() && getImplicitDownCycleTraversal()) {
		            startIndex = startIndex - 1;
		        }
		    }
		}
		for (int i=startIndex; i<list.size(); i++) {
		    Component cmpo = (Component)list.get(i);
		    if (accept(cmpo)) return cmpo;
		    if (cmpo instanceof Container) {
		        Container cont = (Container)cmpo;
		        if (cont.isFocusCycleRoot() && getImplicitDownCycleTraversal()) {
		            Component defCmpo = getDefaultComponent(cont);
		            if (defCmpo != null) return defCmpo;
		            else continue;
		        } else {
		            continue;
		        }
		    }
		}
		// wrap around
		for (int i=0; i<=index; i++) {
		    Component cmpo = (Component)list.get(i);
		    if (accept(cmpo)) return cmpo;
		}
	    return null;
	}
	public Component getComponentBefore(Container focusCycleRoot, Component aComponent) {
		if (focusCycleRoot == null) throw new IllegalArgumentException("focusCycleRoot cannot be null");
		if (aComponent == null) throw new IllegalArgumentException("aComponent cannot be null");
		if (!aComponent.isFocusCycleRoot(focusCycleRoot)) throw new IllegalArgumentException("aComponent is not part of focusCycleRoot");

		// get the components in the focus cycle
		ArrayList list = getPreorderFocusCycle(focusCycleRoot, true);
	    // figure out where to start the search
	    int index = list.indexOf(aComponent);

		for (int i=index-1; i>=0; i--) {
		    Component cmpo = (Component)list.get(i);
		    if (accept(cmpo)) return cmpo;
		}
		for (int i=list.size()-1; i>index; i--) {
		    Component cmpo = (Component)list.get(i);
		    if (accept(cmpo)) return cmpo;
		}

		if ((index != -1) && (accept(aComponent))) return aComponent;

	    return null;
	}

	public Component getFirstComponent(Container focusCycleRoot) {
		if (focusCycleRoot == null) throw new IllegalArgumentException("focusCycleRoot cannot be null");
	    if (!focusCycleRoot.isFocusCycleRoot()) throw new IllegalArgumentException("invalid focusCycleRoot");
//		System.out.println("START CMPOS");
//	    ArrayList cmpos = getPreorderComponents(focusCycleRoot);
//		for (int i=0; i<cmpos.size(); i++) {
//		    Component cmpo = (Component)cmpos.get(i);
//	    	System.out.println(cmpo + " " + accept(cmpo) );
//		}
//		System.out.println("END CMPOS");
	    ArrayList list = getPreorderFocusCycle(focusCycleRoot, true);
		for (int i=0; i<list.size(); i++) {
		    Component cmpo = (Component)list.get(i);
			if (accept(cmpo)) return cmpo;
		}
		return null;
	}

	public Component getLastComponent(Container focusCycleRoot) {
		if (focusCycleRoot == null) throw new IllegalArgumentException("focusCycleRoot cannot be null");
	    if (!focusCycleRoot.isFocusCycleRoot()) throw new IllegalArgumentException("invalid focusCycleRoot");
		ArrayList list = getPreorderFocusCycle(focusCycleRoot, true);
		for (int i=list.size()-1; i>=0; i--) {
		    Component cmpo = (Component)list.get(i);
			if (accept(cmpo)) return cmpo;
		}
		return null;
	}

	public Component getDefaultComponent(Container focusCycleRoot) {
		Component cmpo = getFirstComponent(focusCycleRoot);
		return cmpo;
	}

	/**
	 * Sets whether this ContainerOrderFocusTraversalPolicy transfers focus down-cycle
	 * implicitly. If true, during normal forward focus traversal, the Component
	 * traversed after a focus cycle root will be the focus-cycle-root's default
	 * Component to focus. If false, the next Component in the focus traversal cycle
	 * rooted at the specified focus cycle root will be traversed instead. The default
	 * value for this property is true.
	 */
	public void setImplicitDownCycleTraversal(boolean implicitDownCycleTraversal) {
		this.implicitDownCycleTraversal = implicitDownCycleTraversal;
	}

	/**
	 * Returns whether this ContainerOrderFocusTraversalPolicy transfers focus down-cycle
	 * implicitly. If true, during normal forward focus traversal, the Component
	 * traversed after a focus cycle root will be the focus-cycle-root's default
	 * Component to focus. If false, the next Component in the focus traversal cycle
	 * rooted at the specified focus cycle root will be traversed instead. The default
	 * value for this property is true.
	 */
	public boolean getImplicitDownCycleTraversal() {
		return implicitDownCycleTraversal;
	}

	/* Helper method for debugging. */
	ArrayList getPreorderComponents(Container root) {
	    ArrayList list = new ArrayList();
	    list.add(root);
		Component[] components = root.getComponents();
		for (int i=0; i<components.length; i++) {
		    Component cmpo = components[i];
		    if (cmpo instanceof Container) {
		        list.addAll(getPreorderComponents((Container)cmpo));
		    } else {
		        list.add(cmpo);
		    }
		}
		return list;
	}
	/**
	 * Test for traversal "fitness".  A Component is fit if it is visible,
	 * displayable, enabled and focusable.
	 */
	protected boolean accept(Component component) {
	    if (!component.isVisible()) return false;
	    if (!component.isDisplayable()) return false;
	    if (!component.isFocusable()) return false;
	    if (!component.isEnabled()) return false;
		return true;
	}

}
