package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.Serializable;

import com.ibm.oti.awt.Util;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>Cursor</code> specifies the appearance of the mouse pointer
 * on the screen
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 * @see			java.awt.Component
 */
public class Cursor implements Serializable {

	static final long serialVersionUID = 8028237497568985504L;

	//	Constants that specify cursor types
	public static final int DEFAULT_CURSOR = 0;
	public static final int CROSSHAIR_CURSOR = 1;
	public static final int TEXT_CURSOR = 2;
	public static final int WAIT_CURSOR = 3;
	public static final int SW_RESIZE_CURSOR = 4;
	public static final int SE_RESIZE_CURSOR = 5;
	public static final int NW_RESIZE_CURSOR = 6;
	public static final int NE_RESIZE_CURSOR = 7;
	public static final int N_RESIZE_CURSOR = 8;
	public static final int S_RESIZE_CURSOR = 9;
	public static final int W_RESIZE_CURSOR = 10;
	public static final int E_RESIZE_CURSOR = 11;
	public static final int HAND_CURSOR = 12;
	public static final int MOVE_CURSOR = 13;

	protected static Cursor predefined[];

	// An array of cursor handles.
	static int cursors[];

	// The receiver type
	int type;

	/**
	 * Constructs a new <code>Cursor</code> of the specified
	 * type.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Cursor(int type) {
		if (type < DEFAULT_CURSOR || type > MOVE_CURSOR)
			throw new IllegalArgumentException("illegal cursor type"); //$NON-NLS-1$

		this.type = type;
	}

	/**
	 * Answers a <code>Cursor</code> of the specified type.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		a <code>Cursor</code> which is the default cursor
	 */
	public static Cursor getPredefinedCursor(int type) {
		if (type < DEFAULT_CURSOR || type > MOVE_CURSOR)
			throw new IllegalArgumentException();

		if (predefined == null) predefined = new Cursor[MOVE_CURSOR+1];
		if (predefined[type] == null) predefined[type] = new Cursor(type);
		return predefined[type];
	}

	/**
	 * Answers the default cursor.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		a <code>Cursor</code> which is the default cursor
	 */
	public static Cursor getDefaultCursor() {
		return getPredefinedCursor(DEFAULT_CURSOR);
	}

	/**
	 * Answers the type of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		an <code>int</code> which is the receiver's type
	 */
	public int getType() {
		return type;
	}

	protected void finalize() throws Throwable {
		if (cursors == null) return;

		int cursorHandle = cursors[type];
		if (cursorHandle != 0) {
			cursors[type] = 0;
			OS.Cursor_Dispose(cursorHandle);
		}
	}

	private int convertToNativeCursorStyle(int style) {
		switch (style) {
			case CROSSHAIR_CURSOR:	return OS.CURSOR_CROSSHAIR;
			case DEFAULT_CURSOR:	return OS.CURSOR_ARROW;
			case E_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_E;
			case HAND_CURSOR:		return OS.CURSOR_HAND;
			case MOVE_CURSOR:		return OS.CURSOR_MOVE;
			case N_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_N;
			case NE_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_NE;
			case NW_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_NW;
			case S_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_S;
			case SE_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_SE;
			case SW_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_SW;
			case TEXT_CURSOR:		return OS.CURSOR_TEXT;
			case W_RESIZE_CURSOR:	return OS.CURSOR_RESIZE_W;
			case WAIT_CURSOR:		return OS.CURSOR_WAIT;
			default: return OS.CURSOR_ARROW;
		}
	}

	/**
	 * Answer the peer representation of the cursor
	 */
	int getPeer() {
		if (cursors == null) cursors = new int[MOVE_CURSOR + 1];
		if (cursors[type] == 0) cursors[type] = OS.Cursor_New(Util.getDisplay().fPeerHandle, convertToNativeCursorStyle(type));

		return cursors[type];
	}

	String getCursorName(int type) {
		switch(type) {
			case DEFAULT_CURSOR: 		return "Default Cursor"; //$NON-NLS-1$
			case CROSSHAIR_CURSOR: 		return "Crosshair Cursor"; //$NON-NLS-1$
			case TEXT_CURSOR: 			return "Text Cursor"; //$NON-NLS-1$
			case WAIT_CURSOR: 			return "Wait Cursor"; //$NON-NLS-1$
			case SW_RESIZE_CURSOR: 		return "Southwest Resize Cursor"; //$NON-NLS-1$
			case SE_RESIZE_CURSOR: 		return "Southeast Resize Cursor"; //$NON-NLS-1$
			case NW_RESIZE_CURSOR: 		return "Northwest Resize Cursor"; //$NON-NLS-1$
			case NE_RESIZE_CURSOR: 		return "Northeast Resize Cursor"; //$NON-NLS-1$
			case N_RESIZE_CURSOR: 		return "North Resize Cursor"; //$NON-NLS-1$
			case S_RESIZE_CURSOR: 		return "South Resize Cursor"; //$NON-NLS-1$
			case W_RESIZE_CURSOR: 		return "West Resize Cursor"; //$NON-NLS-1$
			case E_RESIZE_CURSOR: 		return "East Resize Cursor"; //$NON-NLS-1$
			case HAND_CURSOR: 			return "Hand Cursor"; //$NON-NLS-1$
			case MOVE_CURSOR: 			return "Move Cursor"; //$NON-NLS-1$
			default: 					return "unknown cursor"; //$NON-NLS-1$
		}
	}

	public boolean equals(Object object) {
		if (!(object instanceof Cursor)) return false;
		return (type == ((Cursor)object).type);
	}

	public int hashCode() {
		return type;
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[" + getCursorName(getType()) + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}
}
