package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Dialog</code> is a top-level widget that has decorations.
 *
 * @author      IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class Dialog extends Window {

	static final long serialVersionUID = 5920926903803293709L;

	// The title of the receiver
	String title;

	// Determines if the receiver is resizable
	boolean resizable;

	// Determines if the receiver is modal
	boolean modal;
	transient Object modalLock;

	// indicates whether the dialog is undecorated
	boolean undecorated;

	/**
	 * Constructs a new <code>Dialog</code> with the specified
	 * parent.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 *
	 * @throws	IllegalArgumentException	if the parent is <code>null</code>
	 */
	public Dialog(Frame parent) {
		this(parent, ""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Dialog</code> with the specified
	 * parent and title.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 * @param	title	the title displayed in the receiver's decorations.
	 *
	 * @throws	IllegalArgumentException	if the parent is <code>null</code>
	 */
	public Dialog(Frame parent, String title) {
		this(parent, title, false);
	}

	/**
	 * Constructs a new <code>Dialog</code> with the specified
	 * parent and modality.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 * @param	modal	the modality of the dialog
	 *
	 * @throws	IllegalArgumentException	if the parent is <code>null</code>
	 */
	public Dialog(Frame parent, boolean modal) {
		this(parent, "", modal); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Dialog</code> with the specified
	 * parent, title and modality.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 * @param	title	the title displayed in the receiver's decorations.
	 * @param	modal	the modality of the dialog
	 *
	 * @throws	IllegalArgumentException	if the parent is <code>null</code>
	 */
	public Dialog(Frame parent, String title, boolean modal) {
		super(parent);

		this.title = title;
		this.modal = modal;

		if (modal) {
			modalLock = new Object();
		}

		this.resizable = true;
	}

	void _supershow(boolean visible) {
		super._show(visible);
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	void _setupPeer() {
		super._setupPeer();

		_setTitle(title);
		OS.Shell_SetResizable(peer, resizable);
	}

	int _newPeer(Container hwParent) {
		Window parentWindow = hwParent.getWindow();
		return OS.Shell_NewChild(parentWindow.peer, _widgetFlags());
	}

	int _widgetFlags() {
		int style = isModal() ? OS.STYLE_APPLICATION_MODAL : 0;
		if (isUndecorated()) {
			return style | OS.STYLE_NO_TRIM;
		}
		return style | OS.STYLE_TITLE | OS.STYLE_CLOSE | OS.STYLE_BORDER;
	}

	Color getDefaultBackground() {
		return SystemColor.control;
	}

	Color getDefaultForeground() {
		return SystemColor.controlText;
	}

	/*
	 * Make the receiver visible according the specified value. If the
	 * receiver's parent is showing, the receiver will show on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	visible	<code>true</code> if the receiver is to be visible,
	 *					<code>false</code> otherwise
	 */
	void _show(boolean visible) {
		boolean wasVisible = isVisible();
		super._show(visible);

		if (isModal()) {
			if (!wasVisible && visible) {
				waitForDialogClose();
			} else
			if (wasVisible && !visible) {
				synchronized (modalLock) {
					modalLock.notifyAll();
				}
			}
		}
	}

	private void waitForDialogClose() {
		if (EventQueue.isDispatchThread() == false) {
			// If the dialog is being opened from a thread other
			// than the dispatch thread then the thread can simply
			// until the event posting thread delivers the event that
			// the dialog has been closed
			synchronized (modalLock) {
				try {
					modalLock.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		} else {
			// If the dialog is being opened in the event posting
			// thread then we want this method to block but also
			// keep the event loop spinning so that the event
			// that closes the dialog can be delivered
			Toolkit.getCurrentSystemEventQueue().eventLoop(new EventQueue.EventLoop() {
				public boolean isRunning() {
					return isVisible();
				}
			});
		}
	}

	void _setTitle(String newTitle) {
		this.title = newTitle;

		acquirePeerReadLock();
		try {
			if ((peer != 0) && (newTitle != null)) {
				OS.Shell_SetText(peer, newTitle);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer whether the receiver is modal.  A modal dialog will
	 * block the thread that show it on the screen until the dialog
	 * is either hidden or disposed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver is modal,
	 *			<code>false</code> otherwise
	 */
	public boolean isModal() {
		return modal;
	}

	/**
	 * Sets whether the receiver is modal.  A modal dialog will
	 * block the thread that show it on the screen until the dialog
	 * is either hidden or disposed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		modal	<code>true</code> if the receiver is to be modal,
	 *						<code>false</code> otherwise
	 */
	public void setModal(boolean modal) {
		this.modal = modal;
	}

	/**
	 * Answers the title displayed in the receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is the title displayed by the receiver
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Sets the title displayed in the receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	title	the desired title to by displayed by the receiver.
	 */
	public synchronized void setTitle(String title) {
		if (this.title == title) {
			return;
		}
		_setTitle(title);
	}

	/**
	 * Make the receiver invisible, and cause the <code>show</code>
	 * thread to unblock.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void hide() {
		super.hide();
	}

	void _closeWindow() {
		super._closeWindow();
		if ((parent != null) && (parent.isValid())) parent.invalidate();
//		postEvent(new ComponentEvent(
//			this,
//			visible ? ComponentEvent.COMPONENT_SHOWN : ComponentEvent.COMPONENT_HIDDEN));
	}

	/**
	 * Dispose the receiver, and cause the <code>show</code>
	 * thread to unblock.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void dispose() {
		super.dispose();

		// too aggressive when dealing with native dialogs
//		if (parent != null) {
//			Window parentWindow = (Window) parent;
//			parentWindow.toFront();
//			parentWindow.transferFocus();
//		}
	}

	/**
	 * Answer whether the receiver is resizable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver can be resized,
	 *			<code>false</code> otherwise
	 */
	public boolean isResizable() {
		return resizable;
	}

	/**
	 * Sets whether the receiver is resizable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		resizable	<code>true</code> to allow resizing the receiver,
	 *							<code>false</code> otherwise
	 */
	public void setResizable(boolean resizable) {
		if (this.resizable == resizable)
			return;

		this.resizable = resizable;

		acquirePeerReadLock();
		try {
			if (peer != 0)
				OS.Shell_SetResizable(peer, resizable);
		} finally {
			releasePeerReadLock();
		}
	}

	String classNonlocalizedName() {
		return "dialog"; //$NON-NLS-1$
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() +
			",title=" + getTitle() + //$NON-NLS-1$
			",modal=" + isModal() + //$NON-NLS-1$
			",resizable=" + isResizable(); //$NON-NLS-1$
	}

	Point translateToPeerCoords(Point point) {
	// for a dialog, there is not translation required.
		return point;
	}

	public Dialog(Frame parent, String title, boolean modal, GraphicsConfiguration config) {
		this(parent, title, modal);
		this.graphicsConfiguration = config;
	}

	public void setUndecorated(boolean undecorated) {
		if (isDisplayable()) {
			throw new IllegalComponentStateException("Cannot change undecorated state while dialog is displayable");
		}
		this.undecorated = undecorated;
	}

	public boolean isUndecorated() {
		return undecorated;
	}
}
