package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * <code>Dimension</code> is a representation of anything that has a width or height
 * that uses integer precision to store its data.
 * <p>
 * For instance, a dimension is used to represent the size of components.
 * <p>
 *
 * @author      IBM
 * @version		initial
 *
 * @see			java.lang.Cloneable
 */

import java.io.Serializable;

public class Dimension implements Serializable, Cloneable
{

	static final long serialVersionUID = 4723952579491349524L;

	/**
	 * The width of the receiver.
	 */
	public int width;

	/**
	 * The height of the receiver;
	 */
	public int height;

	/**
	 * Constructs a new <code>Dimension</code> with width and height
	 * equals to <code>0</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Dimension() {
		this(0, 0);
	}

	/**
	 * Constructs a new <code>Dimension</code> with width and height
	 * equals to the width and height of <code>dimension</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	dimension	the size of the receiver
	 */
	public Dimension(Dimension dimension) {
		this(dimension.width, dimension.height);
	}

	/**
	 * Constructs a new <code>Dimension</code> with specified
	 * width and height
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	width	the width of the receiver
	 * @param	height	the height of the receiver
	 */
	public Dimension(int width, int height) {
		this.width = width;
		this.height = height;
	}

	/**
	 * Compares the argument to the receiver, and answers true
	 * if the argument is a <code>Dimension</code> and its width
	 * and height are equals to the receiver's width and height.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		object	Object	the object to compare with this dimension
	 *
	 * @return		<code>true</code> if the object is the same as this object
	 *				<code>false</code> otherwise
	 */
	public boolean equals(Object object) {
		if (object == this)
			return true;
		if (!(object instanceof Dimension))
			return false;
		Dimension dimension = (Dimension) object;
		return width == dimension.width && height == dimension.height;
	}

	/**
	 * Generate the hashCode() of this instance.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>int</code> hashCode value
	 *
	 */
	public int hashCode() {
		return width ^ height;
	}

	/**
	 * Answer the size of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a Dimension which has the same width and height compared
	 *				to the receiver
	 */
	public Dimension getSize() {
		return new Dimension(width, height);
	}

	/**
	 * Sets the width and height of the receiver with the
	 * specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		width	the new width for the receiver
	 * @param		height	the new height for the receiver
	 */
	public void setSize(int width, int height) {
		this.width = width;
		this.height = height;
	}

	/**
	 * Sets the width and height of the receiver with the
	 * width and height of the specified dimension.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		dimension	the new size for the receiver
	 */
	public void setSize(Dimension dimension) {
		setSize(dimension.width, dimension.height);
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[width=" + width + ",height=" + height + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	public Object clone() {
		try {
			Dimension clone = (Dimension)super.clone();
			clone.width = width;
			clone.height = height;
			return clone;
		} catch (CloneNotSupportedException e) {return null;}
	}
}
