package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.InputMethodEvent;
import java.awt.event.InvocationEvent;
import java.lang.reflect.InvocationTargetException;
import java.util.EmptyStackException;
import java.util.Stack;

/**
 * <code>EventQueue</code> implements a platform-independent
 * way for queueing AWT events.
 *
 * @author      IBM
 * @version		initial
 */

public class EventQueue {

	Item head;
	Item tail;
	EventDispatchThread dispatchThread;
	Stack queueStack;
	EventQueue parent;
	AWTEvent mostRecentEvent;
	private boolean isRunning;

	private static int count;
	private int id;

	int dependentShells = 0;
	//AWTEventBroadcaster awtEventBroadcaster;

	transient static final private boolean Debug = false;

	static class EventDispatchThread extends Thread implements EventLoop {

		EventQueue eventQueue;

		public EventDispatchThread(EventQueue eventQueue) {
			super("AWT EventDispatchThread-" + eventQueue.id); //$NON-NLS-1$
			this.eventQueue = eventQueue;
		}

		public void run() {
			eventQueue.eventLoop(this);
		}

		public boolean isRunning() {
			return true;
		}

	}

	static class Item {
		AWTEvent awtEvent;
		Item next;
	}

	static interface EventLoop {
		public boolean isRunning();
	}

	static class WakeEvent extends AWTEvent implements ActiveEvent {

		public WakeEvent() {
			super("", 0); //$NON-NLS-1$
		}

		public void dispatch() {
			// NOP
		}

	}

	static class RunnableEvent extends AWTEvent implements ActiveEvent {

		Exception exception;

		public RunnableEvent(Runnable runnable) {
			super(runnable, 0);
		}

		public Exception getException() {
			return exception;
		}

		public void dispatch() {
			synchronized (this) {
				try {
					((Runnable) getSource()).run();
				} catch (Exception e) {
					exception = e;
				} finally {
					notifyAll();
				}
			}
		}

	}

	/**
	 * Invokes the specified <code>Runnable</code> in the
	 * the thread that dispatches events in the receiver.
	 * <p>
	 * The runnable will be invoked after all events in the
	 * receiver have been handled.
	 * <p>
	 * The calling thread will block until the runnable is
	 * invoked.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	runnable	the <code>Runnable</code> to be invoked
	 *
	 * @see		Runnable
	 * @see		java.awt.EventQueue#invokeLater(Runnable)
	 *
	 * @since	JDK1.2
	 */
	public static void invokeAndWait(Runnable runnable) throws InterruptedException, InvocationTargetException {
		EventQueue eventQueue = Toolkit.getCurrentSystemEventQueue();

		if (eventQueue.dispatchThread == Thread.currentThread())
			throw new AWTError("calling from the dispatch thread"); //$NON-NLS-1$

		if (runnable == null) {
			runnable = new Runnable() {
				public void run() {
					throw new NullPointerException();
				}
			};
		}
		RunnableEvent event = new RunnableEvent(runnable);
		synchronized (event) {
			eventQueue.postEvent(event);
			event.wait();
		}

		if (event.getException() != null)
			throw new InvocationTargetException(event.getException());
	}

	/**
	 * Invokes the specified <code>Runnable</code> in the
	 * the thread that dispatches events in the receiver.
	 * <p>
	 * The runnable will be invoked after all events in the
	 * receiver have been handled.
	 * <p>
	 * The calling thread will not block.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	runnable	the <code>Runnable</code> to be invoked
	 *
	 * @see		Runnable
	 * @see		java.awt.EventQueue#invokeAndWait(Runnable)
	 *
	 * @since	JDK1.2
	 */
	public static void invokeLater(Runnable runnable) {
		if (runnable == null) {
			runnable = new Runnable() {
				public void run() {
					throw new NullPointerException();
				}
			};
		}
		RunnableEvent event = new RunnableEvent(runnable);
		Toolkit.getCurrentSystemEventQueue().postEvent(event);
	}

	/**
	 * Answers whether the calling thread is the one that
	 * dispatches events in the recevier.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	<code>true</code> if the calling thread is the one
	 * 			that dispatches events in the receiver. <code>false</code>
	 *			otherwise
	 *
	 * @since	JDK1.2
	 */
	public static boolean isDispatchThread() {
		return Thread.currentThread() == Toolkit.getCurrentSystemEventQueue().dispatchThread;
	}

	/**
	 * Constructs a new <code>EventQueue</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public EventQueue() {
		queueStack = new Stack();
		id = count++;
		this.mostRecentEvent = null;
	}

	/**
	 * Dispatches the specified event.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	awtEvent	the event to dispatch
	 */
	protected void dispatchEvent(AWTEvent awtEvent) {
		Toolkit.getDefaultToolkit().dispatchEventToAWTEventListener(awtEvent);

		Object source = awtEvent.getSource();
		if (awtEvent instanceof ActiveEvent)
			((ActiveEvent)awtEvent).dispatch();
		else if (source instanceof Component)
			((Component)source).dispatchEvent(awtEvent);
		else if (source instanceof MenuComponent)
			((MenuComponent)source).dispatchEvent(awtEvent);
	}

	void eventLoop(EventLoop eventLoop) {
		while (eventLoop.isRunning()) {
			try {
				AWTEvent awtEvent;
				synchronized (this) {
					if (dependentShells > 0) {
						awtEvent = getNextEvent();
					} else {
						awtEvent = getNextEventNoWait();
					}

					if (awtEvent == null) {
						if (dependentShells == 0) {
							isRunning = false;
							return;
						} else {
							continue;
						}
					}
				}
				setMostRecentEvent(awtEvent);
				dispatchEvent(awtEvent);
			} catch (Throwable e) {
				// This catch block used to only handle Expcetions but Throwable is a
				// better options because there should not be any way for an application
				// to cause the event loop to stop except for closing the windows.  See
				// bug 109486 - "dispatchEvent should catch throwable instead of exception"
				// for a specific example
				System.out.println("Throwable caught in event loop: " + e + "; event loop continuing ..."); //$NON-NLS-1$ //$NON-NLS-2$
				e.printStackTrace();
			}
		}
	}

	void incrementShellCount() {
		synchronized (this) {
			dependentShells++;
			ensureThreadRunning();
		}
	}

	void decrementShellCount() {
		synchronized (this) {
			dependentShells--;
		}
	}

	/**
	 * Answers the next event in the receiver.  If there
	 * is an event waiting to be dispatched, it is removed
	 * from the receiver and returned.  Otherwise, the calling
	 * thread will be blocked until there is an event available.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return 	a java.awt.AWTEvent	the next event waiting to be dispatched
	 */
	public AWTEvent getNextEvent() throws InterruptedException {
		synchronized (this) {
			while (true) {
				AWTEvent event = getNextEventNoWait();

				if (event == null) {
					wait();
				} else {
					return event;
				}
			}
		}
	}

	private AWTEvent getNextEventNoWait() {
		while (true) {
			if (head == null) return null;

			AWTEvent awtEvent = head.awtEvent;
			head.awtEvent = null;
			head = head.next;
			if (head == null)
				tail = null;

			if (awtEvent != null) {
				if (Debug) {
					System.out.println(queueSize() + " <- " + awtEvent); //$NON-NLS-1$
				}
				return awtEvent;
			}
		}
	}

	/**
	 * Answers the next event in the receiver, or <code>null</code>,
	 * if there are no events waiting to be dispatched.  The event
	 * is not removed from the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return 	a java.awt.AWTEvent	the next event waiting to be dispatched
	 */
	public AWTEvent peekEvent() {
		synchronized (this) {
			Item item = head;
			while (item != null) {
				if (item.awtEvent != null)
					return item.awtEvent;
				item = item.next;
			}
			return null;
		}
	}

	/**
	 * Answers the next event of the given id in the receiver.
	 * Return <code>null</code> if there are no events waiting to be
	 * dispatched, or if an event with that id cannot be located.
	 * The event is not removed from the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return 	a java.awt.AWTEvent	the next event waiting to be dispatched
	 *			that has the given id
	 */
	public AWTEvent peekEvent(int id) {
		synchronized (this) {
			Item item = head;
			while (item != null) {
				if (item.awtEvent != null && item.awtEvent.getID() == id)
					return item.awtEvent;
				item = item.next;
			}
			return null;
		}
	}

	/**
	 * Stop dispatching events.  Any events waiting to be
	 * dispatched are transferred to the previous event queue.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @throws	EmptyStackException	if there were no previous queue
	 *
	 * @since	JDK1.2
	 */
	protected void pop() throws EmptyStackException {
		synchronized (this) {
			EventQueue root = getRoot();
			EventQueue popped = root.getCurrentEventQueue();
			EventQueue parentQueue= popped.parent;
			if (parentQueue == null) {
				throw new EmptyStackException();
			}

			parentQueue.queueStack.pop();
			EventQueue current = root.getCurrentEventQueue();

			if (popped.head != null) {
				if (current.head == null)
					current.head = popped.head;
				else
					current.tail.next = popped.head;

				current.tail = popped.tail;
				popped.head = popped.tail = null;
				current.ensureThreadRunning();
			}
		}
	}

	EventQueue getRoot() {
		EventQueue root = this;
		while (root.parent != null)
			root = root.parent;

		return root;
	}

	EventQueue getCurrentEventQueue() {
		if (!queueStack.empty())
			return ((EventQueue)queueStack.peek()).getCurrentEventQueue();
		else
			return this;
	}

	/**
	 * Posts the specified event to the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	awtEvent	the event to be posted
	 */
	public void postEvent(AWTEvent awtEvent) {
		if (awtEvent == null) {
			throw new NullPointerException();
		}

		getRoot().getCurrentEventQueue().postEventImpl(awtEvent);
		if (Debug) {
			System.out.println(queueSize() + " -> " + awtEvent); //$NON-NLS-1$
		}
	}

	/**
	 * Dump event queue
	 */
	private synchronized int queueSize() {
		Item item;
		int  count = 0;

		item = head;
		while (item != null) {
			item = item.next;
			count++;
		}

		return count;
	}

	/**
	 * Posts the specified event to the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	awtEvent	the event to be posted
	 */
	void postEventImpl(AWTEvent awtEvent) {
		synchronized (this) {
			Item item = head;
			while (item != null) {
				AWTEvent currentEvent = item.awtEvent;
				if (currentEvent != null) {
					Object source = currentEvent.getSource();
					if (currentEvent.getID() == awtEvent.getID() &&
						source == awtEvent.getSource() &&
							source instanceof Component)
					{
						currentEvent = ((Component) source).coalesceEvents(currentEvent, awtEvent);
						if (currentEvent != null) {
							awtEvent = currentEvent;
							item.awtEvent = null;
						}
						// Only one matching event to coalesce; no need to search further.
						// The currentEvent has already coalesced with previous events.
						break;
					}
				}
				item = item.next;
			}

			if (head == null) {
				head = tail = new Item();
				head.awtEvent = awtEvent;
			} else {
				item = new Item();
				item.awtEvent = awtEvent;
				tail.next = item;
				tail = item;
			}

			ensureThreadRunning();
			notifyAll();
		}
	}

	private void ensureThreadRunning() {
		if (isRunning == false) {
			// Start the event thread up to dispatch
			// the new event
			if ((dispatchThread != null) && (dispatchThread.isAlive())) {
				try {
					dispatchThread.join();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			isRunning = true;
			dispatchThread = new EventDispatchThread(this);
			dispatchThread.start();
		}
	}

	/**
	 * Replaces the receiver with the specified event
	 * queue.  Any events waiting to be dispatched are
	 * transferred to the speficied event queue.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	eventQueue	 the new event queue
	 *
	 * @since	JDK1.2
	 */
	public void push(EventQueue eventQueue) {
		synchronized (this) {
			EventQueue current = getCurrentEventQueue();
			if (current.head != null) {
				if (eventQueue.head == null)
					eventQueue.head = current.head;
				else
					eventQueue.tail.next = current.head;

				eventQueue.tail = current.tail;
				current.head = current.tail = null;
				eventQueue.ensureThreadRunning();
			}

			queueStack.push(eventQueue);
			eventQueue.parent = this;
		}
	}

	void removeEvents(Object source) {
		getRoot().getCurrentEventQueue().removeEventsImpl(source);
	}

	void removeEventsImpl(Object source) {
		//System.out.println("EventQueue.removeEventsImpl1");
		synchronized (this) {
			Item item = head;
			while (item != null) {
				if (item.awtEvent != null && item.awtEvent.getSource() == source)
					item.awtEvent = null;
				item = item.next;
			}
		}
		//System.out.println("EventQueue.removeEventsImpl2");
	}

	void wake() {
		postEventImpl(new WakeEvent());
	}

	/**
	 * Returns the timestamp of the most recent event that had a timestamp,
	 * and that was dispatched from the EventQueue associated with the calling thread.
	 * If an event with a timestamp is currently being dispatched, its timestamp will be returned.
	 * If no events have yet been dispatched, the EventQueue's initialization time will be returned instead.
	 * In the current version of the Java platform SDK, only InputEvents, ActionEvents, and InvocationEvents
	 * have timestamps; however, future versions of the SDK may add timestamps to additional event types.
	 * Note that this method should only be invoked from an application's event dispatching thread.
	 * If this method is invoked from another thread, the current system time (as reported by System.currentTimeMillis())
	 * will be returned instead.
	 */
	public static long getMostRecentEventTime() {
		if (!isDispatchThread()) return System.currentTimeMillis();
		return Toolkit.getCurrentSystemEventQueue()._getMostRecentEventTime();
	}
	long _getMostRecentEventTime() {
		if (mostRecentEvent == null) return System.currentTimeMillis();
		if (mostRecentEvent instanceof ActionEvent) return ((ActionEvent)mostRecentEvent).getWhen();
		if (mostRecentEvent instanceof InputEvent) return ((InputEvent)mostRecentEvent).getWhen();
		if (mostRecentEvent instanceof InputMethodEvent) return ((InputMethodEvent)mostRecentEvent).getWhen();
		if (mostRecentEvent instanceof InvocationEvent) return ((InvocationEvent)mostRecentEvent).getWhen();
		return System.currentTimeMillis();
	}
	/**
	 * Returns the the event currently being dispatched by the EventQueue associated with the calling thread.
	 * This is useful if a method needs access to the event, but was not designed to receive a reference to it as an argument.
	 * Note that this method should only be invoked from an application's event dispatching thread.
	 * If this method is invoked from another thread, null will be returned.
	 */
	public static AWTEvent getCurrentEvent() {
		if (!isDispatchThread()) return null;
//		// Is this correct, or do we need the one currently going through
//		// dispatchEvent?  peekEvent() might be the one after the one we are
//		// supposed to return;
//		return Toolkit.getDefaultToolkit().getSystemEventQueue().
//			getCurrentEventQueue().peekEvent();
		// The previous code was not correct because this API is used to access
		// to the current event (not the next) in case an API was not designed to receive a reference to the AWTEvent
		// who generates the callback..
		return Toolkit.getCurrentSystemEventQueue()._getMostRecentEvent();
	}

	void setMostRecentEvent(AWTEvent awtEvent) {
		this.mostRecentEvent = awtEvent;
	}
	AWTEvent _getMostRecentEvent() {
		return mostRecentEvent;
	}

}
