package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.File;
import java.io.FilenameFilter;

import com.ibm.oti.security.permissions.awt.AWTPermissionEx;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>FileDialog</code> is a modal dialog used to navegate
 * through file systems in order to select a given file.
 *
 * @author      IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */

public class FileDialog extends Dialog {

	static final long serialVersionUID = 5035145889651310422L;

	// The current directory of the receiver
	String dir;

	// The current file of the receiver
	String file;

	// The filename filter of the receiver
	FilenameFilter filter;

	// The mode of the receiver
	int mode;

	transient int fpeer;
	transient int fpeermode;

	/**
	 * Constants used to specify the type of operation that will
	 * be performed in the selected file.
	 */
	public static final int LOAD = 0;
	public static final int SAVE = 1;

	/**
	 * Constructs a new <code>FileDialog</code> with the specified
	 * parent.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 */
	public FileDialog(Frame parent) {
		this(parent, ""); //$NON-NLS-1$
	}

	/**
	* Constructs a new <code>FileDialog</code> with the specified
	* parent and title.
	*
	* @author      IBM
	* @version 	initial
	*
	* @param	parent	the frame which is the parent of the receiver
	* @param	title	the title displayed in the receiver's decorations.
	*/
	public FileDialog(Frame parent, String title) {
		this(parent, title, LOAD);
	}

	/**
	 * Constructs a new <code>FileDialog</code> with the specified
	 * parent, title and mode. The <code>mode</code> value can be one of
	 * <code>LOAD</code>
	 * <code>SAVE</code>
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 * @param	title	the title displayed in the receiver's decorations.
	 * @param	mode	a constant representing the type of operation will
	 *					be performed in the selected file.
	 *
	 * @throws	NullPointerException	if the parent is <code>null</code>
	 */
	public FileDialog(Frame parent, String title, int mode) {
		super(parent, title, true);
		switch (mode) {
			case LOAD :
			case SAVE :
				break;
			default :
				throw new IllegalArgumentException();
		}
		this.mode = mode;
		setLayout(null);

		SecurityManager sm = System.getSecurityManager();
		if (null != sm) {
			sm.checkPermission(AWTPermissionEx.permissionToCreateFileDialog);
		}
	}

	int _widgetFlags() {
		return super._widgetFlags()
			| (getMode() == LOAD ? OS.STYLE_OPEN : OS.STYLE_SAVE);
	}

	int _newPeer(Container hwParent) {
		if (fpeer != 0) return 0;

		int peerParent = hwParent.getWindow().peer;
		fpeer = OS.FileDialog_New(peerParent, _widgetFlags());
		fpeermode = getMode();

		if (getDirectory() == null) {
			setDirectory(System.getProperty("user.dir")); //$NON-NLS-1$
		}

		OS.FileDialog_SetFilterPath(fpeer, getDirectory());
		OS.FileDialog_SetFileName(fpeer, getFile());

		return 0;
	}

	String classNonlocalizedName() {
		return "filedlg"; //$NON-NLS-1$
	}

	void _setTitle(String newTitle) {
		// Overridden from Dialog since two key differences
		// are that we use the fpeer field instead of peer,
		// and the native call is separate

		if (fpeer != 0 && newTitle != null) {
			OS.FileDialog_SetTitle(fpeer, newTitle);
		}
		title = newTitle;
	}

	void _setupPeer() {
		// Do not call super here, because that will eventually
		// call Component._setupPeer() to try and set colors, cursors,
		// and the like, which we don't want (especially since we don't
		// use the 'peer' field on FileDialog
		_setTitle(title);
	}

	/**
	 * Answers the current directory for the receiver.  If a file
	 * has been selected by the user, the return value is the
	 * directory of the selected file.  If no file has been selected,
	 * the return value is the last paramater passed to <code>setDirectory(String)</code>
	 * or <code>null</code>, if the method has not been called.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is the current directory
	 */
	public String getDirectory() {
		return dir;
	}

	/**
	 * Answers the file for the receiver.  If a file
	 * has been selected by the user, the return value is the
	 * selected file.  If no file has been selected, the return
	 * value is the paramater last passed to <code>setFile(String)</code>
	 * or <code>null</code>, if the method has not been called.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is the current directory
	 */
	public String getFile() {
		return file;
	}

	/**
	 * Answer the FilenameFilter.  Currently ignored.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return a <code>FilenameFilter</code>
	 *
	 * @since		JDK1.2
	 */
	public FilenameFilter getFilenameFilter() {
		return filter;
	}

	/**
	 * Answer whether the receiver is modal.  A modal dialog will
	 * block the thread that show it on the screen until the dialog
	 * is either hidden or disposed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver is modal,
	 *			<code>false</code> otherwise
	 */
	public boolean isModal() {
		// According to Personal Profile spec, FileDialog instances
		// are always modal

		return true;
	}

	/**
	 * Answers the mode of the receiver.  The mode can be either
	 * <code>LOAD</code> or <code>SAVE</code> and it represets
	 * the type of operation that will be performed in the
	 * selected file.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> representing the mode of the receiver
	 *
	 * @since		JDK1.2
	 */
	public int getMode() {
		return mode;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		StringBuffer buffer = new StringBuffer();
		Rectangle bounds = getBounds();
		buffer.append(getName());
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.x);
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.y);
		buffer.append(","); //$NON-NLS-1$
		buffer.append(bounds.width);
		buffer.append("x"); //$NON-NLS-1$
		buffer.append(bounds.height);
		if (!isVisible())
			buffer.append(",hidden"); //$NON-NLS-1$
		if (!isEnabled())
			buffer.append(",disabled"); //$NON-NLS-1$
		if (!isValid())
			buffer.append(",invalid"); //$NON-NLS-1$
		if (isLightweight())
			buffer.append(",lightweight"); //$NON-NLS-1$
		buffer.append(",title=" + getTitle()); //$NON-NLS-1$
		buffer.append(",modal=" + isModal()); //$NON-NLS-1$
		buffer.append(",resizable=" + isResizable()); //$NON-NLS-1$
		buffer.append(",directory=" + getDirectory()); //$NON-NLS-1$
		buffer.append(",filename=" + getFile()); //$NON-NLS-1$
		return buffer.toString();
	}

	/**
	 * Sets the directory for the receiver.  The directory specifies
	 * the current directory when the receiver is opened.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	directory	the directory for the receiver
	 */
	public void setDirectory(String directory) {
		this.dir = "".equals(directory) ? null : directory; //$NON-NLS-1$
	}

	/**
	 * Sets the file for the receiver.  The file specifies the
	 * default file when the receiver is opened.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	filename	the default file for the receiver
	 */
	public void setFile(String filename) {
		this.file = "".equals(filename) ? null : filename; //$NON-NLS-1$
	}

	/**
	 * Set the FilenameFilter.  Currently ignored.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	filenameFilter	the filename filter for the receiver
	 *
	 * @since		JDK1.2
	 */
	public void setFilenameFilter(FilenameFilter filenameFilter) {
		this.filter = filenameFilter;
	}

	/**
	 * Sets the mode of the receiver.  The mode can be either
	 * <code>LOAD</code> or <code>SAVE</code> and it represets
	 * the type of operation that will be performed in the
	 * selected file.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	mode	the mode for the receiver
	 *
	 * @throws	IllegalArgumentException	if the mode is not valid
	 *
	 * @since	JDK1.2
	 */
	public void setMode(int mode) {
		switch (mode) {
			case LOAD :
			case SAVE :
				break;
			default :
				throw new IllegalArgumentException();
		}
		this.mode = mode;
	}

	/*
	 * Make the receiver visible according the specified value. If the
	 * receiver's parent is showing, the receiver will show on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	visible	<code>true</code> if the receiver is to be visible,
	 *					<code>false</code> otherwise
	 */
	void _show(boolean visible) {
		// Calling super._show() is a bad idea for FileDialogs because
		// Dialog._show() tries to handle modality issues and it assumes that
		// _setVisible(true) and _setVisible(false) will make the native
		// widget appear and disapear.

		// Since FileDialogs are always modal (according to the class Javadoc
		// on FileDialog) it is best if we avoid the modal code handling in
		// Dialog._show() and call Window._show() instead.
		_supershow(visible);

		// It is not possible to programmatically hide file dialogs.
		if (!visible) return;

		if ((fpeer == 0) || (fpeermode == getMode())) {
			addNotify();
		}

		OS.FileDialog_Open(fpeer);

		// This code used to be executed when the
		// WindowClosing event came into the FileDialog.
		// Since fpeer.open() is always blocking it is simpler
		// if the results are read as soon as it returns and
		// the peer layer does not need to send any events.
		String selectedFile;
		String[] selectedFiles = OS.FileDialog_GetFileNames(fpeer);
		if (selectedFiles == null || selectedFiles.length == 0) {
			selectedFile = null;
		} else {
			selectedFile = selectedFiles[0];
		}
		if (selectedFile != null) {
			// The separator character should not unconditionally
			// be added to the directory.
			String filterPath = OS.FileDialog_GetFilterPath(fpeer);
			if (!filterPath.endsWith(File.separator)) {
				filterPath += File.separatorChar;
			}

			setDirectory(filterPath);
			setFile(selectedFile);
		} else {
			setFile(null);
		}

		_supershow(false);
	}

	public synchronized void dispose() {
		super.dispose();
		if (fpeer == 0) return;
		OS.Widget_Dispose(fpeer);
		fpeer = 0;
	}

	boolean isPeerNull() {
		return fpeer == 0;
	}

}
