package java.awt;

import java.io.Serializable;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * <code>FlowLayout</code> is a <code>LayoutManager</code> that
 * lays out the components in a lines which can be centered,
 * left or right aligned.
 * <p>
 * Children are laid out using their preferred sizes.
 *
 * @author      IBM
 * @version		initial
 *
 * @see		java.awt.LayoutManager
 *
 */

public class FlowLayout implements LayoutManager, Serializable {

	static final long serialVersionUID = -7262534875583282631L;

	// The aligment of the receiver
	int align;

	// The horizontal gap between components
	int hgap;

	// The vertical gap between components
	int vgap;

	/**
	 * Constants that specify alignment.
	 */
	public static final int LEFT = 0;
	public static final int CENTER = 1;
	public static final int RIGHT = 2;

	/**
	 * Constructs a new <code>FlowLayout</code> with centered
	 * lines and horizontal/vertical gaps equals <code>5</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public FlowLayout() {
		this(CENTER);
	}

	/**
	 * Constructs a new <code>FlowLayout</code> with specified
	 * alignment and horizontal/vertical gaps equals <code>5</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	alignment	the desired alignment of the lines
	 */
	public FlowLayout(int alignment) {
		this(alignment, 5, 5);
	}

	/**
	 * Constructs a new <code>FlowLayout</code> with specified
	 * alignment and gaps.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	alignment	the desired alignment of the lines
	 * @param	hgap		the desired horizontal gap
	 * @param	vgap		the desired vertical gap
	 *
	 * @throws	IllegalArgumentException	if the alignment is invalid
	 */
	public FlowLayout(int alignment, int hgap, int vgap) {
		this.align = alignment;
		this.hgap = hgap;
		this.vgap = vgap;
	}

	/**
	 * Adds the specified <code>Component</code> to the receiver
	 * using the given name.
	 * <p>
	 * The receiver does not need to keep the component or name in order
	 * to perform the layout operation.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		name		the name which identifies the component
	 * @param		component	the component to be added
	 *
	 * @see			Container#add(Component)
	 */
	public void addLayoutComponent(String name, Component component) {
	}

	/**
	 * Answers the current alignment of the lines.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		an <code>int</code> which is the current alignment
	 *
	 * @see			#setAlignment(int)
	 */
	public int getAlignment() {
		return align;
	}

	/**
	 * Answers the horizontal gap between components.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		an <code>int</code> which is the horizontal gap
	 *
	 * @see			#setHgap
	 */
	public int getHgap() {
		return hgap;
	}

	/**
	 * Answers the vertical gap between components.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		an <code>int</code> which is the vertaical gap
	 *
	 * @see			#setVgap(int)
	 */
	public int getVgap() {
		return vgap;
	}

	/**
	 * Lays out the children of the specified <code>Container</code>.
	 * <p>
	 * Each child is laid out using its preferred size.  When a
	 * child can not be fully displayed in the current line, a new line
	 * is started.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		container	the <code>Container</code> to be laid out
	 *
	 * @see			Container#doLayout()
	 */
	public void layoutContainer(Container container) {
		Insets insets = container.getInsets();
		int maxX =
			container.getWidth()
				- (insets.left + insets.right + this.getHgap() * 2);
		int ncomponents = container.getComponentCount();
		int startX = 0;
		int y = insets.top + this.getVgap();
		int rowMaxHeight = 0, start = 0;

		//Calculate the starting height of each component in the row
		for (int i = 0; i < ncomponents; i++) {
			Component m = container.getComponent(i);
			if (m.isVisible()) {
				Dimension d = m.getPreferredSize();
				m.setSize(d.width, d.height);
				//Bug fix: Must account for parent height and
				//component height before laying out in case of
				//multiple rows
				if ((startX == 0) || ((startX + d.width) <= maxX)) {
					if (startX > 0) {
						//Must account for hgap in this case
						startX += this.getHgap();
					}
					startX += d.width;
					rowMaxHeight = Math.max(rowMaxHeight, d.height);
				} else {
					//The actual laying out of the components in the row has been moved to this method
					this.alignComponents(
						container,
						insets.left + hgap,
						y,
						maxX - startX,
						rowMaxHeight,
						start,
						i);
					startX = d.width;
					y += this.getVgap() + rowMaxHeight;
					rowMaxHeight = d.height;
					start = i;
				}
			}
		}
		//Lay out the components in the row(s)
		this.alignComponents(
			container,
			insets.left + this.getHgap(),
			y,
			maxX - startX,
			rowMaxHeight,
			start,
			ncomponents);
	}

	/**
	 * Calculates the minimum size for the specified <code>Container</code>.
	 * <p>
	 * The minimum size of a container laid out using a <code>FlowLayout</code>
	 * is the size needed to lay out all children in one line using their minimum
	 * sizes.
	 * <p>
	 * The container insets is included.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		container	the <code>Container</code> to calculate the minimum size
	 *
	 * @return		a <code>Dimension</code> which is the minimum size
	 *				for the specified <code>Container</code>
	 *
	 * @see			Container#minimumSize
	 */
	public Dimension minimumLayoutSize(Container container) {
		int width = 0;
		int height = 0;
		int ncomponents = container.getComponentCount();
		for (int i = 0; i < ncomponents; i++) {
			Component component = container.getComponent(i);
			if (component.isVisible()) {
				Dimension size = component.getMinimumSize();
				if(i > 0) width += hgap;
				width += size.width;
				height = Math.max(height, size.height);
			}
		}

		Insets insets = container.getInsets();
		return new Dimension(
			width + insets.left + insets.right + (2* hgap),
			height + insets.top + insets.bottom + (2 * vgap));
	}

	/**
	 * Calculates the preferred size for the specified <code>Container</code>.
	 * The preferred size is the size with which the container can fully display
	 * itself and its children.
	 * <p>
	 * The preferred size of a container laid out using a <code>FlowLayout</code>
	 * is the size needed to lay out all children in one line using their preferred
	 * sizes.
	 * <p>
	 * The container insets is included.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		container	the <code>Container</code> to calculate the preferred size
	 *
	 * @return		a <code>Dimension</code> which is the preferred size
	 *				for the specified <code>Container</code>
	 *
	 * @see			Container#preferredSize()
	 */
	public Dimension preferredLayoutSize(Container container) {
		int width = 0;
		int height = 0;
		int ncomponents = container.getComponentCount();
		for (int i = 0; i < ncomponents; i++) {
			Component component = container.getComponent(i);
			if (component.isVisible()) {
				Dimension size = component.getPreferredSize();
				if(i > 0) width += hgap;
				width += size.width;
				height = Math.max(height, size.height);
			}
		}

		Insets insets = container.getInsets();
		return new Dimension(
			width + insets.left + insets.right + (2 * hgap),
			height + insets.top + insets.bottom + (2 * vgap));
	}

	/**
	 * Removes the specified <code>Component</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		component	the component to be removed
	 *
	 * @see			Container#remove(Component)
	 */
	public void removeLayoutComponent(Component component) {
	}

	/**
	 * Sets the alignment of the lines.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	alignment	the desired alignment of the lines
	 *
	 * @throws	IllegalArgumentException	if the alignment is invalid
	 *
	 * @see			#getAlignment()
	 */
	public void setAlignment(int alignment) {
		this.align = alignment;
	}

	/**
	 * Sets the horizontal gap between components.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		hgap	the desired horizontal gap
	 *
	 * @see			#getHgap()
	 */
	public void setHgap(int hgap) {
		this.hgap = hgap;
	}

	/**
	 * Sets the vertical gap between components.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		vgap	the desired vertical gap
	 *
	 * @see			#getHgap()
	 */
	public void setVgap(int vgap) {
		this.vgap = vgap;
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[alignment=" + getAlignment() + //$NON-NLS-1$
		",hgap=" + getHgap() + ",vgap=" + getVgap() + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}
	/**
	*	Align the components in this <code>FlowLayout</code>
	*	properly along their X-axis.
	*
	*	<p>
	*	@see java.awt.FlowLayout#layoutContainer(Container)
	*
	*	@param target	The <code>Container</code> for layout
	*	@param x		The unadjusted x coordinate for the component being laid out
	*	@param y		The y coordinate of the component to lay out
	*	@param width	The width of the component
	*	@param height	The height of the component
	*	@param rowStart	Indicates which row in the container contains the component
	*	@param rowEnd	The last row of the container
	*
	**/
	private void alignComponents(
		Container target,
		int x,
		int y,
		int width,
		int height,
		int rowStart,
		int rowEnd) {

		// test cases show Sun JDK works like this ...
		switch (getAlignment()) {
			case 4 :
			case FlowLayout.RIGHT :
				x += width;
				break;
			case FlowLayout.CENTER :
				x += width / 2;
			default :
				break;
		}

		//Determine the row position of each component
		for (int i = rowStart; i < rowEnd; i++) {
			Component m = target.getComponent(i);

			if (m.isVisible()) {
				m.setLocation(x, y + (height - m.getHeight()) / 2);
			} else {
				m.setLocation(
					target.getWidth() - x - m.getWidth(),
					y + (height - m.getHeight()) / 2);
			}

			x += m.getWidth() + hgap;
		}

	}

}
