package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.*;
import java.util.*;
import java.text.AttributedCharacterIterator.Attribute;
import java.awt.font.TextAttribute;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.oti.awt.metal.widgets.Display;
import com.ibm.ugl.eswt.OS;

/**
 * Font is the class used by widgets to describe the
 * font it would like to use.
 * <p>
 * Fonts are described by their font face name (eg. Times New Roman or
 * Times New Roman Bold),and their point size (eg. 10,12 etc...).
 * <p>
 * Java uses unicode font values and converts those values to
 * matching font characters in separate tables for each locale.  So in
 * order for a certain unicode character to be drawn on a widget the following
 * basic steps occur:
 *	- unicode char is used in a font specific operation (eg. <code>drawString()</code>)
 *	- character from unicode font is matched to character in operating system (OS) font
 *	- glyph for the character in the OS font is looked-up and drawn.
 * <code>GraphicsEnvironment.getAllFonts()</code> answers all fonts available
 * from the OS.
 *
 *
 * @author      IBM
 * @version	initial
 *
 * @see	java.io.Serializable
 */
public class Font implements Serializable {

	static final long serialVersionUID = -4206021311591459213L;

	protected String name;
	protected int size;
	protected int style;
	transient int extraUglStyles = 0;
	private Hashtable fRequestedAttributes;

	// This value tracks the font scaling value that is
	// requested by the system variables.  When this value
	// is < 0 it means that the value has not been read yet
	private static double fontScaling = -1;

	// While not used directly, the following field is part
	// of the serialized form for Font as defined by both
	// the Personal Basis and Personal Profile specs.
	private int fontSerializedDataVersion;

	/**
	 * Constants used to represent font styles.
	 */
	public static final int BOLD = 1;
	public static final int ITALIC = 2;
	public static final int PLAIN = 0;
	static final int BOLDITALIC = 3;

	transient FontPeer peer;

	static Map fontCache = new HashMap();

	/**
	 * Constructs a new <code>Font</code> using the specified
	 * name, style and size.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	fontName	the font face name, or a logical name (eg. Dialog)
	 *						which maps to a face name.
	 * @param	fontStyle	the style for the font (eg. <code>Font.BOLD</code>)
	 * @param	fontSize	the point size for the font
	 */
	public Font(String fontName, int fontStyle, int fontSize) {
		name = fontName == null ? "Default" : fontName; //$NON-NLS-1$
		style = fontStyle;
		size = fontSize;
	}

	/**
	 * Constructs a new <code>Font</code> from the peer recieved from the
	 * native layer.
	 */
	Font(FontPeer peer) {
		this(peer.getName(), peer.getStyle(), peer.getSizeInPoints());
		this.extraUglStyles = peer.getExtraStyles();
	}

	/**
	 * Answer a new <code>Font</code> using the specified
	 * <code>String</code> describing the font.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	font	the <code>String</code> describing the font
	 *
	 * @return	a Font	the font described by a string passed in
	 */
	public static Font decode(String font) {
		String defaultFontName = "dialog"; //$NON-NLS-1$
		String fontName = null;
		int style = -1;
		int styleIndex = -1;
		int points = -1;
		int defaultPoints = 12;

		if (font == null) return new Font(defaultFontName, PLAIN, defaultPoints);

		char delimiter = '-';
		if ((font.trim().indexOf('-') == -1)) {
			// if there no dashes, a space is delimiter
			delimiter = ' ';
		}

		String tempToken = font;
		String[] tokens = new String[3];
		int index = -1;
		int j;
		for (j = 0; j < tokens.length; j++) {
			if (j == 2) {
				fontName = tempToken; // if third time through loop, take rest of string as fontName
				break;
			}
			index = tempToken.lastIndexOf(delimiter);
			tokens[j] = tempToken.substring(index + 1).trim();
			if (j == 0) {
				// point value will always be right-most token, if it exists
				if ((tokens[0].length() > 0) && Character.isDigit(tokens[0].charAt(0))) {
					int tempPoints = Integer.parseInt(tokens[0]);
					if (tempPoints != Integer.MIN_VALUE) {
						points = Math.abs(tempPoints);
					}
					if (index == -1) break;
					tempToken = tempToken.substring(0, index);
					continue;
				}
			}
			if (styleIndex == -1) {
				// If we haven't located a style yet, look for one now
				int tempStyle = parseStyleFrom(tokens[j]);
				if ((points != -1) && (j==1) && "".equals(tokens[1])) {
					// if "" is passed in style, make it PLAIN
					tempStyle = PLAIN;
				}
				if (tempStyle != -1) {
					// style was found in input string
					style = tempStyle;
					styleIndex = j;
					if (index == -1) break;  // string has only style & points info
					tempToken = tempToken.substring(0, index);
					continue;
				}
			}
			if (index == -1) break; // no more delimiters, leave loop
			if ((j == 1) && ((style == -1) ^ (points == -1))) {
				if (styleIndex == 0)  {
					// if style has been included, go ahead and make the rest of the
					// string the fontName
					fontName = tempToken;
					break;
				}
				else if (points != -1) {
					// fonts that have no style info should be names with the input
					// name and the points concatenated
					fontName = tempToken; //+ delimiter + tokens[0];
					break;
				}
			}
			tempToken = tempToken.substring(0, index);
			if (fontName != null) break; // if we have a fontName, we're done
		}

		if ((fontName == null) && (styleIndex == j)) {
			if ((style == PLAIN) && (delimiter == ' ')) {
				// fonts with no name, style=PLAIN should be named with "Plain <points>"
				fontName = tokens[styleIndex] + ((points > -1) ? (String.valueOf(delimiter) + points) : "");
			} else {
				fontName = "";
			}
		}
		if (fontName == null) {
			if ((j == 0) && (points != -1)) {
				// if all we have is a number, then the fontName is ""
				fontName = "";
			} else {
				fontName = tokens[j];
			}
		}
		if (style == -1) {
			if ((points == -1) && (delimiter == ' ')) {
				// if the entire string is the name, use it for fontName
				fontName = font.trim();
			}
			style = PLAIN;
		}
		if (points <= 0) points = defaultPoints;

		return new Font(fontName, style, points);
	}

	static int parseStyleFrom(String token) {
		if (token.equalsIgnoreCase("bold")) //$NON-NLS-1$
			return BOLD;
		else if (token.equalsIgnoreCase("bolditalic")) //$NON-NLS-1$
			return BOLD | ITALIC;
		else if (token.equalsIgnoreCase("italic")) //$NON-NLS-1$
			return ITALIC;
		else if (token.equalsIgnoreCase("plain")) //$NON-NLS-1$
			return PLAIN;
		return -1;
	}

	/**
	 * Answer a <code>Font</code> that matches the specified
	 * <code>System</code> property name specified.  The property
	 * name is a combintion of <em>fontname</em>, <em>style</em>
	 * and <em>pointsize</em> as in:
	 *		 - <em>fontname</em>
	 *		 - <em>fontname-style</em>
	 *		 - <em>fontname-pointsize</em>
	 *		 - <em>fontname-style-pointsize</em>
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	property	the system property name
	 *
	 * @return	a Font		a font with for the specified property
	 */
	public static Font getFont(String property) {
		return getFont(property, null);
	}

	/**
	 * Answer a <code>Font</code> that matches the specified
	 * <code>System</code> property name specified.  The property
	 * name is a combintion of <em>fontname</em>, <em>style</em>
	 * and <em>pointsize</em> as in:
	 *		 - <em>fontname</em>
	 *		 - <em>fontname-style</em>
	 *		 - <em>fontname-pointsize</em>
	 *		 - <em>fontname-style-pointsize</em>
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	property	the system property name
	 * @param	font		the font to return if the property was not found
	 *
	 * @return	a Font		a font with for the specified property
	 */
	public static Font getFont(String property, Font font) {
		String fontSpec = System.getProperty(property);
		if (fontSpec != null) {
			return Font.decode(fontSpec);
		}
		return font;
	}

	static String[] getFontList() {
		return new String[] {
			"Dialog",
			"SansSerif",
			"Serif",
			"Monospaced",
			"DialogInput"
		};
	}

	/**
	 * Compares two fonts.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	object		the other font to compare with the receiver.
	 *
	 * @return	a boolean	<code>true</code> if the font
	 *						is the same as the receiver
	 *						code>false</code> otherwise
	 */
	public boolean equals(Object object) {
		if (object == this)
			return true;
		if (!(object instanceof Font))
			return false;
		Font font = (Font) object;
		boolean namesEqual = false;
		if (getName() == null) {
			namesEqual = (font.getName() == null);
		} else {
			namesEqual = getName().equals(font.getName());
		}
		return namesEqual
			&& getSize() == font.getSize()
			&& getStyle() == font.getStyle()
			&& extraUglStyles == font.extraUglStyles
			;
	}

	/**
	 * Perform clean up for the receiver as it has become garbage.
	 * Called by the virtual machine when there are no longer any (non-weak)
	 * references to the receiver. Clean up the underlying OS font.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @exception		Throwable
	 */
	protected void finalize() throws Throwable {
		if (peer == null)
			return;

		EventQueue.invokeLater(new Runnable() {
			public void run() {
				synchronized (fontCache) {
					if (peer.removeReference() == 0) {
						fontCache.remove(Font.this);
						peer.dispose();
					}
				}
				peer = null;
			}
		});
	}

	/**
	 * Answer the family name for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code>	the receiver's family name
	 */
	public String getFamily() {

		String[] fontFamilies = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
		for (int i=0;i<fontFamilies.length;i++) {
			String family = fontFamilies[i];
			if (name.equalsIgnoreCase(family)) {
				return name;
			}
		}

		// If the code gets this far it is because the
		// name of the font specified in the constructor does
		// not match one of the font families.  Since a matching family
		// is REQUIRED by the TCK a guaranteed match can be returned

		return "dialog";
	}

	/**
	 * Answer a logical name for this font.  This is not the same
	 * as a face name which is the name for the font in the underlying OS.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>String</code>	the logical name for the receiver
	 */
	public String getName() {
		return name;
	}

	/**
	 * Answer the point size for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code>	the receiver's point size
	 */
	public int getSize() {
		return size;
	}

	/**
	 * Answer a <em>style</em> used to draw characters for the receiver.
	 * This will be one of the following:
	 *		- BOLD
	 *		- BOLD+ITALIC
	 *		- ITALIC
	 *		- PLAIN
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code>	the style for the receiver
	 */
	public int getStyle() {
		return (isPlain() ? PLAIN : style);
	}

	/**
	 * Answers an integer hash code for the receiver. Any two
	 * objects which answer <code>true</code> when passed to
	 * <code>.equals</code> must answer the same value for this
	 * method.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		int	the receiver's hash.
	 *
	 * @see			#equals
	 */
	public int hashCode() {
		return getName().hashCode() ^ getStyle() ^ getSize()
			^ extraUglStyles
			;
	}

	/**
	 * Answers <code>true</code> if the receiver has
	 * BOLD style.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a boolean	<code>true</code> if the receiver is BOLD
	 *							<code>false</code> otherwise.
	 */
	public boolean isBold() {
		return (style == BOLD || style == BOLDITALIC);
	}

	/**
	 * Answers <code>true</code> if the receiver has
	 * ITALIC style.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a boolean	<code>true</code> if the receiver is ITALIC
	 *							<code>false</code> otherwise.
	 */
	public boolean isItalic() {
		return (style == ITALIC || style == BOLDITALIC);
	}

	/**
	 * Answers <code>true</code> if the receiver has
	 * PLAIN style.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a boolean	<code>true</code> if the receiver is PLAIN
	 *							<code>false</code> otherwise.
	 */
	public boolean isPlain() {
		return style < BOLD || style > BOLDITALIC;
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		String styleString = "plain"; //$NON-NLS-1$
		if (isBold() && isItalic()) {
			styleString = "bolditalic"; //$NON-NLS-1$
		} else if (isBold()) {
			styleString = "bold"; //$NON-NLS-1$
		} else if (isItalic()) {
			styleString = "italic"; //$NON-NLS-1$
		}

		return getClass().getName() + "[" + "family=" + getFamily() + "," + "name=" + getName() + ",style=" + styleString + ",size=" + getSize() + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
	}

	private int scaleFontSize(int size) {
		// If it hasn't been checked before, check to
		// see if font scaling was requested.
		if (fontScaling < 0) {
			String value = System.getProperty("com.ibm.ive.awt.FontScaling");
			if (value == null) {
				fontScaling = 1;
			} else {
				// Check to see if the given value is a valid
				// (positive) double.
				try {
					fontScaling = Double.parseDouble(value);
				} catch (NumberFormatException exp) {
					System.err.println("The requested font scaling '" + value + "' is not a valid double");
					fontScaling = 1;
				}

				// Make sure the requested value is positive
				if (fontScaling < 0) {
					System.err.println("Ignoring requested font scaling (" + fontScaling + ") because it is negative");
					fontScaling = 1;
				}
			}
		}

		return (int)(size*fontScaling);
	}

	FontPeer getPeer() {
		if (peer == null) {
			// Try to find a cached font peer that matches this font
			synchronized (fontCache) {
				FontPeer cachedPeer = (FontPeer)fontCache.get(this);
				if (cachedPeer == null) {
					// Create a dummy font object to be the key for the
					// cache.  A dummy is used because we don't want to keep
					// a reference to an application's font ( which would affect
					// garbage collection )

					Font cacheKey = new Font(getName(), getStyle(), size);
					cacheKey.extraUglStyles = extraUglStyles;

					cachedPeer = _getPeer();
					fontCache.put(cacheKey, cachedPeer);
				}
				cachedPeer.addReference();
				peer = cachedPeer;
			}
		}

		return peer;
	}

	FontPeer _getPeer() {
		Display d = Util.getDisplay();
		// Avoid passing negative font sizes down
		// to the native code.  0 is the smallest supported
		// size in UGL
		int actualSize = Math.max(getSize(), 0);
		actualSize = scaleFontSize(actualSize);

		boolean isLogicalFont = true;
		int logicalConstant=-1;

		// Convert 'logical' font names to platform font names by asking
		// the platform
		if (name.equalsIgnoreCase("Default")) {
			logicalConstant = OS.FONT_LOGICAL_DEFAULT;
		} else
		if (name.equalsIgnoreCase("Dialog")) {
			logicalConstant = OS.FONT_LOGICAL_DIALOG;
		} else
		if (name.equalsIgnoreCase("SansSerif")) {
			logicalConstant = OS.FONT_LOGICAL_SANSSERIF;
		} else
		if (name.equalsIgnoreCase("Serif")) {
			logicalConstant = OS.FONT_LOGICAL_SERIF;
		} else
		if (name.equalsIgnoreCase("Monospaced")) {
			logicalConstant = OS.FONT_LOGICAL_MONOSPACED;
		} else
		if (name.equalsIgnoreCase("DialogInput")) {
			logicalConstant = OS.FONT_LOGICAL_DIALOGINPUT;
		} else {
			isLogicalFont = false;
		}

		int peerStyle = getStyle();
		// This allows us to handle underline & strikethrough per PPro 1.1 requirements of Graphics.drawString()
		peerStyle |= extraUglStyles;

		if (isLogicalFont) {
			return new FontPeer(d, logicalConstant, actualSize, peerStyle);
		} else {
			return new FontPeer(d, name, actualSize, peerStyle);
		}
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}

	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
	}

	public Font(Map attributes) {
		super();
		// Default font!
		this.name = "Default"; //$NON-NLS-1$
		this.style = PLAIN;
		this.size = 12;
		if (attributes == null) return;
		Font font = (Font)attributes.get(TextAttribute.FONT);
		if (font != null) {
			this.name = font.getName();
			this.style = font.getStyle();
			this.size = font.getSize();
			return;
		}
		String family = (String) attributes.get(TextAttribute.FAMILY);
		if (family != null) this.name = family;

		Float floatSize = (Float) attributes.get(TextAttribute.SIZE);
		if (floatSize != null) this.size = (int)floatSize.floatValue();

		Float floatWeight = (Float) attributes.get(TextAttribute.WEIGHT);
		if (floatWeight != null) {
			if (floatWeight.floatValue() >= TextAttribute.WEIGHT_BOLD.floatValue()) {
				this.style |= Font.BOLD;
			}
		}
		Float floatPosture = (Float) attributes.get(TextAttribute.POSTURE);
		if ((floatPosture != null) && (!(TextAttribute.POSTURE_REGULAR.equals(floatPosture)))) {
			this.style |= Font.ITALIC;
		}
	}

	public static Font getFont(Map attributes) {
		Font font = (Font)attributes.get(TextAttribute.FONT);
		if (font != null) return font;
		return new Font(attributes);
	}

	public Map getAttributes() {
		if (fRequestedAttributes == null) {
			fRequestedAttributes = new Hashtable();
			fRequestedAttributes.put(TextAttribute.FONT, this);
			fRequestedAttributes.put(TextAttribute.FAMILY, name);
			fRequestedAttributes.put(TextAttribute.SIZE, new Float(size));
			fRequestedAttributes.put(TextAttribute.WEIGHT, ((style & Font.BOLD) != 0)?TextAttribute.WEIGHT_BOLD:TextAttribute.WEIGHT_REGULAR);
			fRequestedAttributes.put(TextAttribute.POSTURE, ((style & Font.ITALIC) != 0)?TextAttribute.POSTURE_OBLIQUE:TextAttribute.POSTURE_REGULAR);
		}
		return fRequestedAttributes;
	}

	public Attribute[] getAvailableAttributes() {
        return new Attribute[]{
        		TextAttribute.FONT,
                TextAttribute.FAMILY,
                TextAttribute.SIZE,
                TextAttribute.WEIGHT,
                TextAttribute.POSTURE
            };
	}
}
