package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.Serializable;

/**
 * This class provides information for many aspects of
 * a given font and screen combination.  For example character
 * widths, and font ascent and descent.
 * <p>
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public abstract class FontMetrics implements Serializable {

	static final long serialVersionUID = 1681126225205050147L;

	protected Font font;

	/**
	 * Constructs a new <code>FontMetrics</code> on the
	 * specified font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	metricsFont the font to answer metrics for
	 */
	protected FontMetrics(Font metricsFont) {
		font = metricsFont;
	}

	/**
	 * Answer the receiver's font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Font	the receiver's font
	 */
	public Font getFont() {
		return font;
	}

	/**
	 * Answer the leading for the receiver's font.
	 * This is the measure of the distance that should be
	 * left as empty space between the top of characters on
	 * one line and the bottom of characters on another line.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the leading space for the receiver's font
	 */
	public int getLeading() {
		return getHeight() - getDescent() - getAscent();
	}

	/**
	 * Answer the ascent for the receiver's font.
	 * This is the measure of the distance
	 * from the baseline of the font to the top of the highest
	 * alphanumeric character.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the ascent of the receiver's font
	 */
	public int getAscent() {
		return getHeight() - getDescent() - getLeading();
	}

	/**
	 * Answer the descent for the receiver's font.
	 * This is the measure of the distance
	 * from the baseline of the font to the bottom of the
	 * alphanumeric characters with descenders (eg: 'g').
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the descent of the receiver's font
	 */
	public int getDescent() {
		return getHeight() - getAscent() - getLeading();
	}

	/**
	 * Answer the height for the receiver's font.
	 * This is the measure of the total distance
	 * from the baseline of one line of text to the baseline
	 * of the next line.  That is the sum of the ascent,
	 * descent and leading.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the height of the receiver's font
	 */
	public int getHeight() {
		return getAscent() + getDescent() + getLeading();
	}

	/**
	 * Answer the maximum ascent of a character in the receiver's font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the max ascent of a character in the receiver's font
	 */
	public int getMaxAscent() {
		return getAscent();
	}

	/**
	 * Answer the maximum descent of a character in the receiver's font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the max descent of a character in the receiver's font
	 */
	public int getMaxDescent() {
		return getDescent();
	}

	/**
	 * Answer the maximum descent of a character in the receiver's font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the max descent of a character in the receiver's font
	 *
	 * @deprecated	use <code>getMaxDescent</code> instead.
	 */
	public int getMaxDecent() {
		return getMaxDescent();
	}

	/**
	 * Answer the maximum advance width of a character in the receiver's font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int	the max advance of a character in the receiver's font
	 */
	public int getMaxAdvance() {
		int maxAdvance = 0;
		int[] widths = getWidths();
		for (int i = 0; i < widths.length; i++)
			maxAdvance = Math.max(maxAdvance, widths[i]);
		return maxAdvance;
	}

	/**
	 * Answer the advance width of the specified character.
	 * The width value is specified in points.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	character	the character to measure
	 *
	 * @return	an int	the width of the character in points
	 */
	public int charWidth(int character) {
		return charWidth((char)character);
	}

	/**
	 * Answer the advance width of the specified character.
	 * The width value is specified in points.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	character	the character to measure
	 *
	 * @return	an int	the width of the character in points
	 */
	public int charWidth(char character) {
		if (character < 256 && character >= 0)
			return getWidths()[character];

		return stringWidth(String.valueOf(character));
	}

	/**
	 * Answer the advance width of the specified string.
	 * The width value is specified in points.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	widthString	the string to measure
	 *
	 * @return	an int	the width of the string in points
	 */
	public int stringWidth(String widthString) {
		if (widthString == null)
			throw new NullPointerException();

		int size = widthString.length();
		char[] chars = new char[size];
		widthString.getChars(0, size, chars, 0);
		return charsWidth(chars, 0, size);
	}

	/**
	 * Answer the advance width of the specified chars in the specified range.
	 * The width value is specified in points.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	charArray	the chars to measure
	 * @param	start		the index in the char array to start at
	 * @param	length		the number of chars to measure starting at <em>start</em>
	 *
	 * @return	an int	the width of the chars in points
	 */
	public int charsWidth(char[] charArray, int start, int length) {
		return stringWidth(new String(charArray, start, length));
	}

	/**
	 * Answer the advance width of the specified bytes in the specified range.
	 * The width value is specified in points.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	byteArray	the bytes to measure
	 * @param	start		the index in the byte array to start at
	 * @param	length		the number of bytes to measure starting at <em>start</em>
	 *
	 * @return	an int	the width of the bytes in points
	 */
	public int bytesWidth(byte[] byteArray, int start, int length) {
		return stringWidth(new String(byteArray, start, length));
	}

	/**
	 * Answer an array of <code>int</code> of length 256 representing the
	 * character advance widths of the first 256 characters using the receiver's
	 * font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an array of <code>int</code>	the widths of the first 256 chars
	 */
	public int[] getWidths() {
		int[] widths = new int[256];
		for (int i = 0; i < 256; i++)
			widths[i] = charWidth(i);
		return widths;
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[" + getFont().toString() + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

}
