package java.awt;

import java.io.*;

import com.ibm.oti.awt.metal.graphics.FontPeer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */
class FontMetricsImpl extends FontMetrics {

	transient int ascent;
	transient int descent;
	transient int leading;
	transient int height;
	transient int[] widths;

	/**
	 * Constructs a new <code>FontMetrics</code> on the
	 * specified font.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @param	font	the font to answer metrics for
	 */
	protected FontMetricsImpl(Font font) {
		super(font);
		initialize();
	}

	public static FontMetricsImpl createFromFont(Font font) {
		return new FontMetricsImpl(font);
	}

	/**
	 * Answer the advance width of the specified character.
	 * The width value is specified in points.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @param	character	the character to measure
	 *
	 * @return	an int	the width of the character in points
	 */
	public int charWidth(char character) {
		if (character <= 255 && character >= 0) {
			return getWidths()[character];
		}

		return getFont().getPeer().charWidth(character);
	}

	/**
	 * Answer the ascent for the receiver's font.
	 * This is the measure of the distance
	 * from the baseline of the font to the top of the highest
	 * alphanumeric character.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an int	the ascent of the receiver's font
	 */
	public int getAscent() {
		return ascent;
	}

	/**
	 * Answer the descent for the receiver's font.
	 * This is the measure of the distance
	 * from the baseline of the font to the bottom of the
	 * alphanumeric characters with descenders (eg: 'g').
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an int	the descent of the receiver's font
	 */
	public int getDescent() {
		return descent;
	}

	/**
	 * Answer the height for the receiver's font.
	 * This is the measure of the total distance
	 * from the baseline of one line of text to the baseline
	 * of the next line.  That is the sum of the ascent,
	 * descent and leading.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an int	the height of the receiver's font
	 */
	public int getHeight() {
		return height;
	}

	/**
	 * Answer the leading for the receiver's font.
	 * This is the measure of the distance that should be
	 * left as empty space between the top of characters on
	 * one line and the bottom of characters on another line.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an int	the leading space for the receiver's font
	 */
	public int getLeading() {
		return leading;
	}

	/**
	 * Answer the maximum advance width of a character in the receiver's font.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an int	the max advance of a character in the receiver's font
	 */
	public int getMaxAdvance() {
		return super.getMaxAdvance();
	}

	/**
	 * Answer an array of <code>int</code> of length 256 representing the
	 * character advance widths of the first 256 characters using the receiver's
	 * font.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @return	an array of <code>int</code>	the widths of the first 256 chars
	 */
	public synchronized int[] getWidths() {
		if (widths != null)
			return widths;

		widths = new int[256];
		for (int i = 0; i < 256; i++) {
			widths[i] = getFont().getPeer().charWidth((char)i);
		}
		return widths;
	}

	public boolean equals(Object object) {
		if (object == this) return true;
		if (!(object instanceof FontMetricsImpl)) return false;
		FontMetricsImpl fm = (FontMetricsImpl)object;
		return getFont().equals(fm.getFont());
	}

	public int hashCode() {
		return getFont().hashCode();
	}

	void initialize() {
		FontPeer peer = getFont().getPeer();

		ascent = peer.getAscent();
		descent = peer.getDescent();
		leading = peer.getLeading();

		height = super.getHeight();
	}

	private void readObject(ObjectInputStream stream)
		throws IOException, ClassNotFoundException
	{
		stream.defaultReadObject();
		initialize();
	}

	/**
	 * Answer the advance width of the specified string.
	 * The width value is specified in points.
	 *
	 * @author        IBM
	 * @version 	initial
	 *
	 * @param	widthString	the string to measure
	 *
	 * @return	an int	the width of the string in points
	 */
	public int stringWidth(String widthString) {
		if (widthString == null)
			throw new NullPointerException();

		if ("".equals(widthString)) return 0; //$NON-NLS-1$

		int width = getFont().getPeer().stringWidth(widthString);
		return width;
	}
}
