package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.image.BufferedImageHandler;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.oti.awt.metal.widgets.*;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>Frame</code> is a top-level widget that has decorations.
 *
 * @author      IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class Frame extends Window
				   implements MenuContainer
{

	static final long serialVersionUID = 2673458971256075116L;

	// The title of the receiver
	String title;

	// The menu bar of the receiver
	java.awt.MenuBar menuBar;

	// The icon of the receiver
	Image icon;

	// Determines whether the receiver is resizable
	boolean resizable;

	// Determine whether the receiver is <code>NORMAL</code> or <code>ICONIFIED</code>
	int state;

	boolean mbManagement; //used for Motif only

	private int frameSerializedDataVersion; // for PPro

	/**
	 * Constants that specify the states of a frame.
	 */
	public static final int NORMAL = 0;
	public static final int ICONIFIED = 1;

	// Constants that specify cursor types
	public static final int CROSSHAIR_CURSOR = Cursor.CROSSHAIR_CURSOR;
	public static final int DEFAULT_CURSOR = Cursor.DEFAULT_CURSOR;
	public static final int E_RESIZE_CURSOR = Cursor.E_RESIZE_CURSOR;
	public static final int HAND_CURSOR = Cursor.HAND_CURSOR;
	public static final int MOVE_CURSOR = Cursor.MOVE_CURSOR;
	public static final int N_RESIZE_CURSOR = Cursor.N_RESIZE_CURSOR;
	public static final int NE_RESIZE_CURSOR = Cursor.NE_RESIZE_CURSOR;
	public static final int NW_RESIZE_CURSOR = Cursor.NW_RESIZE_CURSOR;
	public static final int S_RESIZE_CURSOR = Cursor.S_RESIZE_CURSOR;
	public static final int SE_RESIZE_CURSOR = Cursor.SE_RESIZE_CURSOR;
	public static final int SW_RESIZE_CURSOR = Cursor.SW_RESIZE_CURSOR;
	public static final int TEXT_CURSOR = Cursor.TEXT_CURSOR;
	public static final int W_RESIZE_CURSOR = Cursor.W_RESIZE_CURSOR;
	public static final int WAIT_CURSOR = Cursor.WAIT_CURSOR;

	boolean undecorated;

	/**
	 * Constructs a new <code>Frame</code> with an empty title.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Frame()
	throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Frame</code> with an empty title,
	 * and the specified graphics configuration.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param config the graphics configuration
	 */
	public Frame(GraphicsConfiguration config) {
		this("", config); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Frame</code> with the specified title
	 * and graphics configuration.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	title	the desired title for the new frame
	 * @param	config	the graphics configuration
	 */
	public Frame(String title, GraphicsConfiguration config) {
		super(config);

		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		_setTitle(title);

		this.resizable = true;

		this.state = NORMAL;
	}

	/**
	 * Constructs a new <code>Frame</code> with the specified title.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	title	the desired title for the new frame
	 */
	public Frame(String title)
		throws HeadlessException
	{
		this(title, null);
	}

	void _setupPeer() {
		// Must create the frame peer before the menubar
		super._setupPeer();

		_setTitle(title);

		// TODO What about platforms that do support setState()?
		//frame.setMinimized(state == ICONIFIED);
		// DLW - commenting out until we decide what to do
		//initWindowIcon();

		if (menuBar != null) {
			menuBar._addNotify();
			OS.Shell_SetMenuBar(peer, menuBar.peer);
		}

		OS.Shell_SetResizable(peer, resizable);
	}

	int _createPeer(Container hwParent) {
		Display d = Util.getDisplay();
		Toolkit.getDefaultToolkit().getSystemEventQueueImpl().incrementShellCount();
		Util.addFrame(this);
		return OS.Shell_New(d.fPeerHandle, _widgetFlags());
	}

	int _widgetFlags() {
		if (isUndecorated()) {
			return OS.STYLE_NO_TRIM;
		}
		return OS.STYLE_CLOSE | OS.STYLE_TITLE | OS.STYLE_MIN | OS.STYLE_MAX;
	}

	/**
	 * Destroy the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void removeNotify() {
		synchronized (getTreeLock()) {
			if (peer == 0) return;

			Util.removeFrame(this);
			_cleanUpPeer();
			super.removeNotify();

			Toolkit.getDefaultToolkit().getSystemEventQueueImpl().decrementShellCount();
		}
	}

	void _cleanUpPeer() {
		if (OS.Shell_GetMinimized(peer))
			state = ICONIFIED;
		else
			state = NORMAL;

		if (menuBar != null)
			menuBar.removeNotify();
	}

	void _setTitle(String newTitle) {
		this.title = newTitle;

		acquirePeerReadLock();
		try {
			if ((peer != 0) && (newTitle != null))
				OS.Shell_SetText(peer, newTitle);
		} finally {
			releasePeerReadLock();
		}
	}

	void _postMoveEvent(int newX, int newY) {
		acquirePeerReadLock();
		try {
			if (peer != 0)
				// Ignore all move and resize events while a frame is minimized
				if (OS.Shell_GetMinimized(peer)) return;
		} finally {
			releasePeerReadLock();
		}

		super._postMoveEvent(newX, newY);
	}

	void _postResizeEvent(int newWidth, int newHeight) {
		acquirePeerReadLock();
		try {
			if (peer != 0)
				// Ignore all move and resize events while a frame is minimized
				if (OS.Shell_GetMinimized(peer)) return;
		} finally {
			releasePeerReadLock();
		}

		super._postResizeEvent(newWidth, newHeight);
	}

	/**
	 * Answers the title displayed in the receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is the title displayed by the receiver
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Sets the title displayed in the receiver's decorations.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	title	the desired title to by displayed by the receiver.
	 */
	public synchronized void setTitle(String title) {
		String newTitle = (title == null ? "" : title); //$NON-NLS-1$
		if (this.title == newTitle) {
			return;
		}
		this.title = newTitle;

		acquirePeerReadLock();
		try {
			if (peer != 0)
				OS.Shell_SetText(peer, newTitle);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the image that is displayed when the receiver is
	 * iconic.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Image which is displayed when the receiver is iconic
	 */
	public Image getIconImage() {
		return icon;
	}

	/**
	 * Sets the image that is displayed when the receiver is
	 * iconic.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	image	the desired image or <code>null</code> to use a
	 *					default image
	 */
	public synchronized void setIconImage(final Image image) {
		if (this.icon == image) {
			return;
		}
		this.icon = image;

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				if (image instanceof Image) {
					IImageImpl imageImpl= BufferedImageHandler.createIImageImpl(image);
					OS.Shell_SetImage(peer, imageImpl.getImagePeer().fPeerHandle);
				} else {
					OS.Shell_SetImage(peer, 0);
				}
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the receiver's <code>MenuBar</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.MenuBar or <code>null</code> if the menu bar is not set
	 */
	public MenuBar getMenuBar() {
		return menuBar;
	}

	/**
	 * Sets the <code>MenuBar</code> for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	newMenuBar	the desired menu bar or <code>null</code> to remove
	 *					the current menu bar
	 */
	public void setMenuBar(MenuBar newMenuBar) {
		if (newMenuBar == menuBar) return;

		Insets oldInsets = getInsetsImpl();
		try {
			graphicsOrigin = null;

			if (menuBar != null)
				remove(menuBar, false);

			menuBar = newMenuBar;

			if (newMenuBar == null) {
				return;
			}

			if ((menuBar.parent != null) && (menuBar.parent != this)) {
				menuBar.parent.remove(menuBar);
			}

			menuBar.parent = this;

			acquirePeerReadLock();
			try {
				if (peer == 0)
					return;

				menuBar._addNotify();
				OS.Shell_SetMenuBar(peer, menuBar.peer);
			} finally {
				releasePeerReadLock();
			}
		} finally {
			/* Flush cached insets.  Do this in a finally
			 * block to ensure things get cleaned up properly
			 * after a change in the menu bar status.
			 */
			insets = null;

			Insets newInsets = getInsetsImpl();
			if (newInsets.equals(oldInsets) == false) {
				invalidate();
				validate();
			}
		}
	}

	/**
	 * Answer whether the receiver is resizable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver can be resized,
	 *			<code>false</code> otherwise
	 */
	public boolean isResizable() {
		return resizable;
	}

	/**
	 * Sets whether the receiver is resizable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		resizable	<code>true</code> to allow resizing the receiver,
	 *							<code>false</code> otherwise
	 */
	public void setResizable(boolean resizable) {
		if (this.resizable == resizable)
			return;

		this.resizable = resizable;

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.Shell_SetResizable(peer, resizable);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the state of the receiver.  The state can be one of
	 * <code>NORMAL</code> or <code>ICONIFIED</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	state	the constants representing the desired state
	 *
	 * @since	JDK1.2
	 */
	public synchronized void setState(final int state) {
		// Calling getState() is a better option than
		// referencing the this.state variable directly
		// since getState can either use the cached value
		// or ask the peer, whichever is more appropriate.
		if (getState() == state) {
			return;
		}

		// Silently fail if new state is not supported
		if (!Toolkit.getDefaultToolkit().isFrameStateSupported(state)) {
			return;
		}
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				this.state = state;
				return;
			}

			switch (state) {
				case NORMAL:
					OS.Shell_SetMinimized(peer, false);
					OS.Shell_SetMaximized(peer, false);
					break;
				default:
					OS.Shell_SetMinimized(peer, true);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the current state of the receiver.  The state can be one of
	 * <code>NORMAL</code> or <code>ICONIFIED</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> representing the state of the receiver
	 *
	 * @since	JDK1.2
	 */
	public int getState() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return state;

			if (OS.Shell_GetMinimized(peer))
				return ICONIFIED;
			return NORMAL;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Remove a <code>MenuComponent</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		menuComponent	the <code>MenuComponent</code> to be removed.
	 *
	 * @see			java.awt.MenuBar
	 * @see			java.awt.PopupMenu
	 */
	public void remove(final MenuComponent menuComponent) {
//		if (menuComponent != menuBar) {
//			super.remove(menuComponent);
//			return;
//		}

		if (menuComponent == null) {
			if (menuBar != null) return;
			throw new NullPointerException();
		}

		remove(menuComponent, true);
	}

	void remove(final MenuComponent menuComponent, boolean validate) {
		if (menuComponent != menuBar) {
			return;
		}

		menuBar = null;

		Insets insetsBefore = getInsetsImpl();
		menuComponent.removeNotify();
		if (validate) {
			insets = null;
			if (getInsetsImpl().equals(insetsBefore) == false) {
				invalidate();
				validate();
			}
		}

		menuComponent.parent = null;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() + (isResizable() ? ",resizable" : "") + ",title=" + getTitle(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	/**
	 * Set the <code>Cursor</code> used by the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	cursorType	the cursor type for the receiver
	 *
	 * @deprecated	use <code>setCursor(Cursor)</code>
	 */
	public synchronized void setCursor(int cursorType) {
		setCursor(Cursor.getPredefinedCursor(cursorType));
	}

	/**
	 * Answers the type of the cursor being used by the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> representing the cursor used by the receiver
	 *
	 * @deprecated	use <code>getCursor().getType()</code> instead
	 */
	public int getCursorType() {
		return getCursor().getType();
	}

	String classNonlocalizedName() {
		return "frame"; //$NON-NLS-1$
	}

	/**
	 * Answer whether the receiver is undecorated.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param undecorated whether the frame is undecorated
	 * @throws IllegalComponentStateException if the frame's peer has been created
	 */
	public void setUndecorated(boolean undecorated) {
		if (isDisplayable()) {
			throw new IllegalComponentStateException("Cannot change undecorated state while frame is displayable"); //$NON-NLS-1$
		}
		this.undecorated = undecorated;
	}

	/**
	 * Answer whether the receiver is undecorated.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	<code>true</code> if the receiver has no decorations,
	 *			<code>false</code> otherwise
	 */
	public boolean isUndecorated() {
		return undecorated;
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException, HeadlessException {
		// TODO PPro 1.1 Javadoc says we need to deal with icons, but not sure that
		// we do, since the comment seems to apply only to Swing
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}
}
