package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.image.ImageObserver;
import java.text.AttributedCharacterIterator;

/**
 * <code>Graphics</code> is an abstract base class for a graphics context
 * used to paint on a drawing surface (such as a screen, image, or printer).
 *
 * @author		IBM
 * @version		initial
 */
public abstract class Graphics {

	/**
	 * Creates a new instance of a graphics context.  Applications cannot
	 * invoke this constructor directly, and should instead request a graphics
	 * context from a component or image instance.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see	java.awt.Component#getGraphics()
	 * @see	java.awt.Image#getGraphics()
	 */
	protected Graphics() {
	}

	/**
	 * Answer a new <code>Graphics</code> instance that is a copy of the
	 * receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>Graphics</code> which is a copy of the receiver
	 */
	public abstract Graphics create();

	/**
	 * Answer a new <code>Graphics</code> instance that is a copy of the
	 * receiver, translating the new graphics context and setting a
	 * clipping area.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x		the x value for translation
	 * @param	y		the y value for translation
	 * @param	width	the width of the new clipping area
	 * @param	height	the height of the new clipping area
	 *
	 * @return	a <code>Graphics</code> which is a copy of the receiver
	 *		that has been translated and has a new clipping area
	 */
	public Graphics create(int x, int y, int width, int height) {
		Graphics graphics = create();

		if (graphics != null) {
			graphics.translate(x, y);
			graphics.clipRect(0, 0, width, height);
		}
		return graphics;
	}

	/**
	 * Add to the receiver's origin by the point (x, y).
	 */
	public abstract void translate(int x, int y);

	/**
	 * Answer the color currently being used for graphics
	 * operations.
	 */
	public abstract Color getColor();

	/**
	 * Set the color to be used for future graphics operations.
	 */
	public abstract void setColor(Color color);

	/**
	 * Set the receiver to overwrite the destination with
	 * the current color.
	 */
	public abstract void setPaintMode();

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void setXORMode(Color xorColor);

	/**
	 * Answer the font used for drawing text.
	 */
	public abstract Font getFont();

	/**
	 * Set the font used for drawing text.
	 */
	public abstract void setFont(Font font);

	/**
	 * Answer the metrics of the current font.
	 */
	public FontMetrics getFontMetrics() {
		return getFontMetrics(getFont());
	}

	/**
	 * Answer the font metrics for the Font represented by
	 * <code>font</code>.
	 */
	public abstract FontMetrics getFontMetrics(Font font);

	/**
	 * Answer the rectangle representing the area being clipped to for
	 * graphics operations.
	 */
	public abstract Rectangle getClipBounds();

	/**
	 * Reduces the clipping area to the intersection of the specified
	 * rectangle and the current clipping area of the graphics context.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x		the x coordinate of the rectangle
	 * @param	y		the y coordinate of the rectangle
	 * @param	width	the width of the rectangle
	 * @param	height	the height of the rectangle
	 */
	public abstract void clipRect(int x, int y, int width, int height);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void setClip(int x, int y, int width, int height);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract Shape getClip();

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void setClip(Shape shape);

	/**
	 * Copies an area of the drawing surface to another location.  The
	 * pixels of the indicated rectangular area are copied to a location
	 * indicated by the input offsets.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x		the x coordinate of the source rectangle
	 * @param	y		the y coordinate of the source rectangle
	 * @param	width	the width of the source rectangle
	 * @param	height	the height of the source rectangle
	 * @param	dx		the horizontal distance from the source
	 * @param	dy		the vertical distance from the source
	 */
	public abstract void copyArea(int x, int y, int width, int height, int dx, int dy);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void drawLine(int x1, int y1, int x2, int y2);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void fillRect(int x, int y, int width, int height);

	//------------------------------------------------------------
	/**
	 *
	 */
	public void drawRect(int x, int y, int width, int height) {
		drawLine(x, y, x, y + height);
		drawLine(x, y, x + width, y);
		drawLine(x, y + height, x + width, y + height);
		drawLine(x + width, y, x + width, y + height);
	}

	/**
	 * Clear the indicated rectangular area.  The area is filled with the
	 * current background color of the drawing surface.  If width or height
	 * is less than or equal to 0, the method returns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x		the x coordinate of the rectangle
	 * @param	y		the y coordinate of the rectangle
	 * @param	width	the width of the rectangle
	 * @param	height	the height of the rectangle
	 */
	public abstract void clearRect(int x, int y, int width, int height);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight);

	/**
	 * Draws a 3D border around the specified rectangular area.  The
	 * border is drawn raised or lowered, depending on the value of
	 * the <code>raised</code> argument.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x		the x coordinate of the rectangle
	 * @param	y		the y coordinate of the rectangle
	 * @param	width	the width of the rectangle
	 * @param	height	the height of the rectangle
	 * @param	raised	whether the border is drawn raised or lowered
	 */
	public void draw3DRect(int x, int y, int width, int height, boolean raised) {
		if (width < 0 || height < 0)
			return;

		int cornerX = x + width;
		int cornerY = y + height;

		Color originalColor = getColor();
		Color darkerColor = originalColor.darker();
		Color brighterColor = originalColor.brighter();

		if (width == 0 && height == 0) {
			if (raised) setColor(brighterColor);
			else setColor(darkerColor);
			drawLine(x, y, x, y);
		} else if (width == 0) {
			if (raised) setColor(brighterColor);
			else setColor(darkerColor);
			drawLine(x, cornerY, x, cornerY);

			if (raised) setColor(darkerColor);
			else setColor(brighterColor);
			drawLine(x, y, x, y);
		} else if (height == 0) {
			if (raised) setColor(brighterColor);
			else setColor(darkerColor);
			drawLine(x, y, x, y);

			if (raised) setColor(darkerColor);
			else setColor(brighterColor);
			drawLine(cornerX, y, cornerX, y);
		} else {
			if (raised) setColor(brighterColor);
			else setColor(darkerColor);
			drawLine(x, y, cornerX - 1, y);
			drawLine(x, y, x, cornerY);

			if (raised) setColor(darkerColor);
			else setColor(brighterColor);
			drawLine(x + 1, cornerY, cornerX, cornerY);
			drawLine(cornerX, y, cornerX, cornerY);
		}

		setColor(originalColor);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void fill3DRect(int x, int y, int width, int height, boolean raised) {
		if (width <= 0 || height <= 0)
			return;

		int cornerX = x + width - 1;
		int cornerY = y + height - 1;

		Color originalColor = getColor();
		Color darkerColor = originalColor.darker();
		Color brighterColor = originalColor.brighter();

		if (raised) {
			fillRect(x + 1, y + 1, width - 2, height - 2);
			setColor(brighterColor);
		} else {
			setColor(darkerColor);
			fillRect(x + 1, y + 1, width - 2, height - 2);
		}
		if (width > 1)
			drawLine(x, y, cornerX - 1, y);
		drawLine(x, y, x, cornerY);

		if (raised) setColor(darkerColor);
		else setColor(brighterColor);
		if (width > 1)
			drawLine(x + 1, cornerY, cornerX, cornerY);
		drawLine(cornerX, y, cornerX, cornerY);

		setColor(originalColor);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void drawOval(int x, int y, int width, int height);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void fillOval(int x, int y, int width, int height);

	/**
	 * Draws an elliptical arc on the graphics context.  The indicated rectangle
	 * reflects the bounding box around the arc, and the angle arguments
	 * indicate the starting and ending position for drawing the arc.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x			the x coordinate of the bounding rectangle
	 * @param	y			the y coordinate of the bounding rectangle
	 * @param	width		the width of the bounding rectangle
	 * @param	height		the height of the bounding rectangle
	 * @param	startAngle	the angle indicating the starting point for drawing
	 * @param	arcAngle	the angle indicating the end point for drawing
	 *
	 * @see		#fillArc
	 */
	public abstract void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void drawPolyline(int xpoints[], int ypoints[], int npoints);

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void drawPolygon(int xpoints[], int ypoints[], int npoints);

	//------------------------------------------------------------
	/**
	 *
	 */
	public void drawPolygon(Polygon polygon) {
		drawPolygon(polygon.xpoints, polygon.ypoints, polygon.npoints);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public abstract void fillPolygon(int xpoints[], int ypoints[], int npoints);

	//------------------------------------------------------------
	/**
	 *
	 */
	public void fillPolygon(Polygon polygon) {
		fillPolygon(polygon.xpoints, polygon.ypoints, polygon.npoints);
	}

	/**
	 * Draws a string on the graphics context.  The baseline represents
	 * the lower left corner of the string (not the upper left as with most
	 * other methods).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	string		the string to be drawn
	 * @param	x			the x location of the baseline of the string
	 * @param	y			the y location of the baseline of the string
	 *
	 * @see		#drawBytes
	 * @see		#drawChars
	 * @see		FontMetrics#getAscent()
	 */
	public abstract void drawString(String string, int x, int y);

	/**
	 * Draws an array of characters on the graphics context.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	data		the array of characters
	 * @param	offset		the index of first character to draw
	 * @param	length		the number of characters to draw
	 * @param	x			the x location of the baseline of the string
	 * @param	y			the y location of the baseline of the string
	 *
	 * @see		#drawBytes
	 * @see		#drawString
	 * @see		FontMetrics
	 */
	public void drawChars(char data[], int offset, int length, int x, int y) {
		drawString(new String(data, offset, length), x, y);
	}

	/**
	 * Draws an array of bytes as characters on the graphics context.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	data		the array of bytes
	 * @param	offset		the index of first character to draw
	 * @param	length		the number of characters to draw
	 * @param	x			the x location of the baseline of the string
	 * @param	y			the y location of the baseline of the string
	 *
	 * @see		#drawChars
	 * @see		#drawString
	 * @see		FontMetrics
	 */
	public void drawBytes(byte data[], int offset, int length, int x, int y) {
		drawString(new String(data, offset, length), x, y);
	}

	/**
	 * Draws the available amount of the indicated image.  The image is drawn
	 * at position <i>(x, y)</i> from the top left corner of the graphics
	 * context.  When transparent pixels are present, the previously existing
	 * pixels at those locations in the graphics context are preserved.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	x				the x location where drawing begins
	 * @param	y				the y location where drawing begins
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int x, int y, ImageObserver imageObserver);

	/**
	 * Draws the available amount of the indicated image and scales the image
	 * if necessary.
	 *
	 * The image is drawn at position <i>(x, y)</i> from the top left corner
	 * of the graphics context, and scaled to fit within the specified
	 * <code>width</code> and <code>height</code> (if necessary).  When
	 * transparent pixels are present, the previously existing pixels at
	 * those locations in the graphics context are preserved.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	x				the x location where drawing begins
	 * @param	y				the y location where drawing begins
	 * @param	width			the width of the scaled image
	 * @param	height			the height of the scaled image
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int x, int y, int width, int height, ImageObserver imageObserver);

	/**
	 * Draws the available amount of the indicated image.  The image is drawn
	 * at position <i>(x, y)</i> from the top left corner of the graphics
	 * context.  The <code>background</code> argument specifies the transparent
	 * pixel color.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	x				the x location where drawing begins
	 * @param	y				the y location where drawing begins
	 * @param	background		the color used for transparent pixels
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int x, int y, Color background, ImageObserver imageObserver);

	/**
	 * Draws the available amount of the indicated image and scales the image
	 * if necessary.
	 *
	 * The image is drawn at position <i>(x, y)</i> from the top left corner
	 * of the graphics context, and scaled to fit within the specified
	 * <code>width</code> and <code>height</code> (if necessary).  The
	 * <code>background</code> argument specifies the transparent pixel color.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	x				the x location where drawing begins
	 * @param	y				the y location where drawing begins
	 * @param	width			the width of the scaled image
	 * @param	height			the height of the scaled image
	 * @param	background		the color used for transparent pixels
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int x, int y, int width, int height, Color background, ImageObserver imageObserver);

	/**
	 * Draws a subimage from the input <code>image</code> to an indicated
	 * area of the graphics context area.
	 *
	 * The image is scaled to fit within the specified area (if necessary).
	 * When transparent pixels are present, the previously existing pixels at
	 * those locations in the graphics context are preserved.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	dstX1			the <i>x</i> coordinate of the top left corner
	 *							of the destination rectangle
	 * @param	dstY1			the <i>y</i> coordinate of the top left corner
	 *							of the destination rectangle
	 * @param	dstX2			the <i>x</i> coordinate of the bottom right
	 *							corner of the destination rectangle
	 * @param	dstY2			the <i>y</i> coordinate of the bottom right
	 *							corner of the destination rectangle
	 * @param	srcX1			the <i>x</i> coordinate of the top left corner
	 *							of the source rectangle
	 * @param	srcY1			the <i>y</i> coordinate of the top left corner
	 *							of the source rectangle
	 * @param	srcX2			the <i>x</i> coordinate of the bottom right
	 *							corner of the source rectangle
	 * @param	srcY2			the <i>y</i> coordinate of the bottom right
	 *							corner of the source rectangle
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int dstX1, int dstY1, int dstX2, int dstY2, int srcX1, int srcY1, int srcX2, int srcY2, ImageObserver imageObserver);

	/**
	 * Draws a subimage from the input <code>image</code> to an indicated
	 * area of the graphics context area.
	 *
	 * The image is scaled to fit within the specified area (if necessary).  The
	 * <code>background</code> argument specifies the transparent pixel color.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	image			the image being drawn onto the context
	 * @param	dstX1			the <i>x</i> coordinate of the top left corner
	 *							of the destination rectangle
	 * @param	dstY1			the <i>y</i> coordinate of the top left corner
	 *							of the destination rectangle
	 * @param	dstX2			the <i>x</i> coordinate of the bottom right
	 *							corner of the destination rectangle
	 * @param	dstY2			the <i>y</i> coordinate of the bottom right
	 *							corner of the destination rectangle
	 * @param	srcX1			the <i>x</i> coordinate of the top left corner
	 *							of the source rectangle
	 * @param	srcY1			the <i>y</i> coordinate of the top left corner
	 *							of the source rectangle
	 * @param	srcX2			the <i>x</i> coordinate of the bottom right
	 *							corner of the source rectangle
	 * @param	srcY2			the <i>y</i> coordinate of the bottom right
	 *							corner of the source rectangle
	 * @param	background		the color used for transparent pixels
	 * @param	imageObserver	object that is sent drawing update messages
	 *
	 * @see		Image
	 * @see		ImageObserver
	 */
	public abstract boolean drawImage(Image image, int dstX1, int dstY1, int dstX2, int dstY2, int srcX1, int srcY1, int srcX2, int srcY2, Color background, ImageObserver imageObserver);

	/**
	 * Disposes of the receiver and releases any associated system
	 * resources.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see #finalize()
	 */
	public abstract void dispose();

	//------------------------------------------------------------
	/**
	 *
	 */
	public void finalize() {
		dispose();
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is a printable representation for the
	 *			receiver
	 */
	public String toString() {
		return getClass().getName() +
			"[color=" + getColor() + //$NON-NLS-1$
			",font=" + getFont() + //$NON-NLS-1$
			",clip=" + getClip() + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	//------------------------------------------------------------
	/**
	 *
	 * @deprecated
	 */
	public Rectangle getClipRect() {
		return getClipBounds();
	}

	/**
	 * Answers whether the passed in area intersects with the clipping area
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	boolean whether the indicated area intersects with the clip area
	 *
	 * @since	JDK1.2
	 */
	public boolean hitClip(int x, int y, int width, int height) {
		return getClipBounds().intersects(new Rectangle(x, y, width, height));
	}

	/**
	 * Return the current clipping area boundary
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rectangle Rectangle into which the clip area bounds will be copied
	 *
	 * @return	Rectangle modified <code>rectangle</code> argument containing
	 *			clip area
	 *
	 * @since	JDK1.2
	 */
	public Rectangle getClipBounds(Rectangle rectangle) {
		Rectangle clip = getClipBounds();
		if (clip != null)
			rectangle.setBounds(clip);
		return rectangle;
	}

	public abstract void drawString(AttributedCharacterIterator attributedcharacteriterator0, int int1, int int2);

}
