package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageObserver;
import java.awt.image.PixelGrabber;
import java.awt.image.RGBImageFilter;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.ugl.p3ml.OS;

class GraphicsACImpl extends GraphicsImpl {

	AlphaComposite composite;
	GraphicsConfigurationImpl graphicsConfiguration;

	BufferedImage tOffScreenImage;
	Graphics tOffScreenGraphics;
	int tBackgroundPixel;

	int logicalRule;
	int logicalAlpha;

	GraphicsACImpl(Component component) {
		super(component);
		this.composite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER);
		this.graphicsConfiguration = (GraphicsConfigurationImpl) component.getGraphicsConfiguration();
	}

	GraphicsACImpl(GraphicsACImpl graphics) {
		super(graphics);
		this.composite = graphics.composite;
		this.graphicsConfiguration = graphics.graphicsConfiguration;
	}

	GraphicsACImpl(Component component, IImageImpl iImage) {
		super(component, iImage);
		this.composite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER);
		if (component != null) {
			this.graphicsConfiguration = (GraphicsConfigurationImpl) component.getGraphicsConfiguration();
		}
	}

	GraphicsACImpl(GraphicsConfiguration graphicsConfiguration, IImageImpl iImage) {
		super(iImage.getComponent(), iImage);
		this.composite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER);
		this.graphicsConfiguration = (GraphicsConfigurationImpl) graphicsConfiguration;
	}

	public Graphics create() {
		return new GraphicsACImpl(this);
	}

	/**
	 * @see Graphics2D#getComposite()
	 */
	public Composite getComposite() {
		return composite;
	}

	/**
	 * @see Graphics2D#setComposite(Composite)
	 */
	public void setComposite(Composite comp) {
		if (!(comp instanceof AlphaComposite)) {
			throw new IllegalArgumentException("Composite must be instance of AlphaComposite"); //$NON-NLS-1$
		}

		composite = (AlphaComposite) comp;

		updateNativeColor(getColor());
	}

	private void setLogicalRule(Color color) {
		if (composite == null) {
			if (color.getAlpha() == 255) {
				// If the color does not have an alpha
				// then the normal graphics primitives should
				// work correctly
				logicalRule = -1;
			} else {
				logicalRule = AlphaComposite.SRC_OVER;
				logicalAlpha = color.getAlpha();
			}
		} else {
			if (composite.getRule() == AlphaComposite.SRC) {
				// In the SRC rule, we can adjust the native color
				// to be the correct blended color and
				// draw directly to the graphics as usual
				logicalRule = -1;
			} else {
				logicalRule = composite.getRule();
				logicalAlpha = (int)(color.getAlpha() * composite.getAlpha());
			}
		}
		if ((logicalRule == AlphaComposite.SRC_OVER) && (logicalAlpha == 0xFF)) {
			// If the rule is always SRC_OVER, an infinite loop could occur without this
			logicalRule = -1;
		}
	}

	Color getAdjustedColor(Color newColor) {
		setLogicalRule(newColor);

		if (composite != null) {
			if (composite.getRule() == AlphaComposite.CLEAR) {
				return new Color(0,0,0,0);
			} else
			if (composite.getRule() == AlphaComposite.SRC) {
				float compAlpha = composite.getAlpha();

				// This if statement is totally bogus according to
				// the spec, but the TCK tests that this is the correct
				// behavior.
				if (compAlpha != 0.0f) compAlpha = 1.0f;
				return getPreMultipliedColor(newColor, compAlpha);
			}
		}

		return super.getAdjustedColor(newColor);
	}

	/**
	 * @see Graphics2D#getDeviceConfiguration()
	 */
	public GraphicsConfiguration getDeviceConfiguration() {
		return graphicsConfiguration;
	}

	private Image createAdjustedImage(Image image, final AlphaComposite rule) {
		RGBImageFilter filter = new RGBImageFilter() {
			public int filterRGB(int x, int y, int rgb) {
				int alpha = (rgb >> 24) & 0x000000FF;
				alpha = Math.round(alpha * rule.getAlpha());
				return (rgb & 0x00FFFFFF) | (alpha << 24);
			}
		};

		FilteredImageSource source = new FilteredImageSource(image.getSource(), filter);
		return Toolkit.getDefaultToolkit().createImage(source);
	}

	// The following methods override the standard behavior in order
	// to handle alpha compositing when a composite has been set.

	public boolean drawImage(Image image, int x, int y, int width, int height, Color background, ImageObserver imageObserver) {
		// Use the AlphaComposite rule directly to determine the blending
		// because logicalAlpha and logicalRule are based on the current color's
		// alpha.  Image drawing should not affected by the current color, and JRE
		// only blends based on the current rule as well.
		final AlphaComposite rule = composite;
		if (rule == null || rule.getRule() != rule.SRC_OVER || rule.getAlpha() == 1.0f) {
			return super.drawImage(image, x, y, width, height, background, imageObserver);
		}

		Image newImage = createAdjustedImage(image, rule);
		return super.drawImage(newImage, x, y, width, height, background, imageObserver);
	}

	public boolean drawImage(Image image, int x, int y, Color background, ImageObserver imageObserver) {
		// Use the AlphaComposite rule directly to determine the blending
		// because logicalAlpha and logicalRule are based on the current color's
		// alpha.  Image drawing should not affected by the current color, and JRE
		// only blends based on the current rule as well.
		final AlphaComposite rule = composite;
		if (rule == null || rule.getRule() != rule.SRC_OVER || rule.getAlpha() == 1.0f) {
			return super.drawImage(image, x, y, background, imageObserver);
		}

		Image newImage = createAdjustedImage(image, rule);
		return super.drawImage(newImage, x, y, background, imageObserver);
	}

	public boolean drawImage(
			Image image,
			int dstX1,
			int dstY1,
			int dstX2,
			int dstY2,
			int srcX1,
			int srcY1,
			int srcX2,
			int srcY2,
			Color background,
			ImageObserver imageObserver) {
		// Use the AlphaComposite rule directly to determine the blending
		// because logicalAlpha and logicalRule are based on the current color's
		// alpha.  Image drawing should not affected by the current color, and JRE
		// only blends based on the current rule as well.
		final AlphaComposite rule = composite;
		if (rule == null || rule.getRule() != rule.SRC_OVER || rule.getAlpha() == 1.0f) {
			return super.drawImage(image, dstX1, dstY1, dstX2, dstY2, srcX1, srcY1, srcX2, srcY2, background, imageObserver);
		}

		Image newImage = createAdjustedImage(image, rule);
		return super.drawImage(newImage, dstX1, dstY1, dstX2, dstY2, srcX1, srcY1, srcX2, srcY2, background, imageObserver);
	}

	public void draw3DRect(int x, int y, int width, int height, boolean raised) {
		if ((width < 0) || (height < 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.draw3DRect(x, y, width, height, raised);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.draw3DRect(0, 0, width, height, raised);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
		if ((width <= 0) || (height <= 0)) return;
		if (arcAngle == 0) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawArc(x, y, width, height, startAngle, arcAngle);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.drawArc(0, 0, width, height, startAngle, arcAngle);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void drawLine(int x1, int y1, int x2, int y2) {
		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawLine(x1, y1, x2, y2);
			return;
		}

		Rectangle rect= new Rectangle(
			(x2 < x1 ? x2 : x1),
			(y2 < y1 ? y2 : y1),
			Math.abs(x2-x1),
			Math.abs(y2-y1));

		// Prepare the off-screen image.
		createOffScreenImage(rect.width, rect.height);
		// Render to the off-screen image.
		tOffScreenGraphics.drawLine(x1-rect.x, y1-rect.y, x2-rect.x, y2-rect.y);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(rect.x, rect.y, rect.width, rect.height);
	}

	public void drawOval(int x, int y, int width, int height) {
		if ((width < 0) || (height < 0)) return;
		if ((width == 0) && (height == 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawOval(x, y, width, height);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.drawOval(0, 0, width, height);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void drawPolygon(int xpoints[], int ypoints[], int npoints) {
		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawPolygon(xpoints, ypoints, npoints);
			return;
		}

		Polygon p= new Polygon(xpoints, ypoints, npoints);
		Rectangle bounds= p.getBounds();

		// Prepare the off-screen image.
		createOffScreenImage(bounds.width, bounds.height);
		// Render to the off-screen image.
		p.translate(-bounds.x, -bounds.y);
		tOffScreenGraphics.drawPolygon(p.xpoints, p.ypoints, p.npoints);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(bounds.x, bounds.y, bounds.width, bounds.height);
	}

	public void drawPolyline(int xpoints[], int ypoints[], int npoints) {
		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawPolyline(xpoints, ypoints, npoints);
			return;
		}

		Polygon p= new Polygon(xpoints, ypoints, npoints);
		Rectangle bounds= p.getBounds();

		// Prepare the off-screen image.
		createOffScreenImage(bounds.width, bounds.height);
		// Render to the off-screen image.
		p.translate(-bounds.x, -bounds.y);
		tOffScreenGraphics.drawPolyline(p.xpoints, p.ypoints, p.npoints);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(bounds.x, bounds.y, bounds.width, bounds.height);
	}

	public void drawRect(int x, int y, int width, int height) {
		if ((width < 0) || (height < 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawRect(x, y, width, height);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width+1, height+1);
		// Render to the off-screen image.
		tOffScreenGraphics.drawRect(0, 0, width, height);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width+1, height+1);
	}

	public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
		if ((width <= 0) || (height <= 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawRoundRect(x, y, width, height, arcWidth, arcHeight);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.drawRoundRect(0, 0, width, height, arcWidth, arcHeight);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void drawString(String string, int x, int y) {
		if (string == null) throw new NullPointerException();
		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.drawString(string, x, y);
			return;
		}
		FontMetrics fm= getFontMetrics();
		int width= fm.stringWidth(string);
		int height= fm.getHeight();

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.drawString(string, 0, height-fm.getDescent());

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y-fm.getAscent(), width, height);
	}

	public void fill3DRect(int x, int y, int width, int height, boolean raised) {
		if ((width <= 0) || (height <= 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fill3DRect(x, y, width, height, raised);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.fill3DRect(0, 0, width, height, raised);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
		if ((width <= 0) || (height <= 0)) return;
		if (arcAngle == 0) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fillArc(x, y, width, height, startAngle, arcAngle);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.fillArc(0, 0, width, height, startAngle, arcAngle);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void fillOval(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fillOval(x, y, width, height);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.fillOval(0, 0, width, height);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x, y, width, height);
	}

	public void fillPolygon(int xpoints[], int ypoints[], int npoints) {
		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fillPolygon(xpoints, ypoints, npoints);
			return;
		}

		Polygon p= new Polygon(xpoints, ypoints, npoints);
		Rectangle bounds= p.getBounds();

		// Prepare the off-screen image.
		createOffScreenImage(bounds.width, bounds.height);
		// Render to the off-screen image.
		p.translate(-bounds.x, -bounds.y);
		tOffScreenGraphics.fillPolygon(p.xpoints, p.ypoints, p.npoints);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(bounds.x, bounds.y, bounds.width, bounds.height);
	}

	public void fillRect(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fillRect(x, y, width, height);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.fillRect(0, 0, width, height);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x,y, width, height);
	}

	public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
		if ((width <= 0) || (height <= 0)) return;

		if (logicalRule != AlphaComposite.SRC_OVER) {
			super.fillRoundRect(x, y, width, height, arcWidth, arcHeight);
			return;
		}

		// Prepare the off-screen image.
		createOffScreenImage(width, height);
		// Render to the off-screen image.
		tOffScreenGraphics.fillRoundRect(0, 0, width, height, arcWidth, arcHeight);

		// Apply the affected regions of the off-screen
		// image based on the indicated compositing rule.
		applyCompositingRule(x,y, width, height);
	}

	// The following methods are generalized utility methods.

	void createOffScreenImage(int width, int height) {
		// Create an off-screen image.
		tOffScreenImage= graphicsConfiguration.createCompatibleImage(width, height);

		// Select a background color other than the active color.  It is important that
		// the color be checked without considering alpha values because it is possible
		// the background color will be set without alpha when drawing to the offscreen
		// image
		Color background= Color.white;
		if ((background.getRGB() & 0x00FFFFFF) == (getColor().getRGB() & 0x00FFFFFF))
			background= Color.black;

		// Fill the background of the off-screen image.
		tOffScreenGraphics= tOffScreenImage.getGraphics();
		tOffScreenGraphics.setColor(background);
		tOffScreenGraphics.fillRect(0, 0, width, height);

		// Save the background pixel.
		int[] getPixel = new int[1];
		tOffScreenImage.getRGB(0,0,1,1, getPixel, 0,1);
		tBackgroundPixel = getPixel[0];
		// Restore the color and font to be used for rendering.
		tOffScreenGraphics.setColor(new Color(getColor().getRGB()));
		tOffScreenGraphics.setFont(getFont());
	}

	void applyCompositingRule(int targetX, int targetY, int width, int height) {
		// Fetch the pixels in the image data. If the color differs from
		// the default background, then apply the compositing rule.
		// Otherwise set the alpha to full transparent.

		int[] finalRGBs = new int[width*height];

		int[] pixels= new int[width];
		for (int y=0; y<height; y++) {
			tOffScreenImage.getRGB(0,y,width,1, pixels, 0, width);
			for (int x=0; x<width; x++) {
				if (pixels[x] == tBackgroundPixel) {
					finalRGBs[y*width + x] = 0x00000000;
				} else {
					finalRGBs[y*width + x] = (pixels[x] & 0x00FFFFFF) | (logicalAlpha << 24);
				}
			}
		}

		int imageHandle = OS.Image_NewFromIntData(Util.getDisplay().fPeerHandle, width, height, finalRGBs, 32, -1, -1, null);
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle != 0) OS.Graphics_DrawImage(handle, imageHandle, targetX+originX, targetY+originY);
		}
		OS.Image_Dispose(imageHandle);

		// Clean up.
		tOffScreenImage.flush();
		tOffScreenGraphics.dispose();

		tOffScreenImage= null;
		tOffScreenGraphics= null;
	}

	/**
	 * Blend the specified color with the current logical alpha.
	 *
	 * @param color the color to blend
	 * @return the blended color
	 */
	Color getPreMultipliedColor(Color color, float alpha) {
		double alphaFactor = alpha * (color.getAlpha() / 255d);
		int r = (int)(color.getRed() * alphaFactor);
		int g = (int)(color.getGreen() * alphaFactor);
		int b = (int)(color.getBlue() * alphaFactor);
		return new Color(r,g,b);
	}

}
