package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * <code>GraphicsDevice</code> is an abstract base class,
 * concrete subclasses of which represent available devices that
 * can be the target of drawing commands (e.g. a screen, a printer,
 * or an image buffer).
 *
 * @author		IBM
 * @version		initial
 */
public abstract class GraphicsDevice {

	public static final int TYPE_RASTER_SCREEN = 0;
	public static final int TYPE_PRINTER = 1;
	public static final int TYPE_IMAGE_BUFFER = 2;

	protected GraphicsDevice() {
		super();
	}

	/**
	 * Answer all the configurations available for this device.
	 */
	public abstract GraphicsConfiguration[] getConfigurations();

	/**
	 * Answer the default configuration of this graphics device.
	 */
	public abstract GraphicsConfiguration getDefaultConfiguration();

	/**
	 * Answer a string that uniquely identifies the receiver.
	 */
	public abstract String getIDstring();

	/**
	 * Answer the type of graphics device of the receiver.
	 *
	 * @return TYPE_RASTER_SCREEN, TYPE_PRINTER or TYPE_IMAGE_BUFFER.
	 */
	public abstract int getType();

	/**
	 * Returns true if the native windowing system supports
	 * a native full screen "exclusive" mode.  This means that
	 * other windows cannot overlap the window currently indicated
	 * as full screen, and appear below that window in z-order.
	 */
	public abstract boolean isFullScreenSupported();

	/**
	 * Enter full screen or windowed mode.  If the argument
	 * is null, then normal windowed mode is assumed.  Otherwise,
	 * the window argument is placed into exclusive full screen
	 * mode, meaning that the window is at the top of the z-order
	 * and has no other windows overlapping it, as it fills the
	 * full bounds of the device.
	 *
	 * If isFullscreenSupported() returns false, then exclusive
	 * full screen mode is emulated.
	 *
	 * @param window the Window to be made full screen, or null
	 * to disable full screen mode
	 *
	 * @see #isFullScreenSupported()
	 */
	public abstract void setFullScreenWindow(Window window);

	/**
	 * Answer the window currently in full screen mode, or null
	 * if no window if the device is not in full screen mode
	 *
	 * @return the window currently in full screen mode
	 */
	public abstract Window getFullScreenWindow();

	/**
	 * Answer the number of available bytes for accelerated
	 * memory for this device.  This data may allow an application
	 * to make efficient use of resources when allocating images.
	 * A negative return value indicates an unlimited amount of
	 * memory available.
	 *
	 * @return the number of available accelerated memory bytes
	 * @see java.awt.image.VolatileImage#flush()
	 * @see ImageCapabilities#isAccelerated()
	 */
	public abstract int getAvailableAcceleratedMemory();
}
