package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.image.ImageObserver;
import java.awt.font.TextAttribute;
import java.text.AttributedCharacterIterator;
import java.util.Map;

import com.ibm.oti.awt.GraphicsHandler;
import com.ibm.oti.awt.image.BufferedImageHandler;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.oti.awt.metal.Handle;
import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.oti.awt.metal.image.ImagePeer;
import com.ibm.ugl.eswt.OS;

abstract class GraphicsImpl extends Graphics2D {

	static {
		GraphicsHandler.setGraphicsFactory(new GraphicsFactory());
	}

	Handle peerHandle = new Handle();
	Component component;
	IImageImpl image;
	int originX;
	int originY;
	Color color;
	Color xorColor;
	Font font;
	FontMetrics fontMetrics;
	Rectangle clip;
	Stroke stroke;

	private Color nativeColor; // This holds onto the color being used by the
	                           // native graphics object.  If the color is not
							   // held onto then the GC might dispose of the color
							   // while it is being used by the OS.

	static void reference() {
		// NOP
	}

	GraphicsImpl(Component component) {
		this.component = component;
		if (component instanceof Container) {
			Insets insets = ((Container)component).getInsetsImpl(); //bug 108776
			this.originX = -insets.left;
			this.originY = -insets.top;
		}
		setColor(component.getForeground());
		setFont(component.getFont());
		setStroke(new BasicStroke());
	}

	GraphicsImpl(GraphicsImpl graphics) {
		this.component = graphics.component;
		this.image = graphics.image;
		this.originX = graphics.originX;
		this.originY = graphics.originY;
		setColor(graphics.color);
		if (graphics.xorColor != null) {
			setXORMode(graphics.xorColor);
		}
		setFont(graphics.font);
		setClip(graphics.clip);
		setStroke(graphics.stroke);
	}

	GraphicsImpl(Component component, IImageImpl image) {
		this.component = component;
		this.image = image;
		if (component != null) {
			setColor(component.getForeground());
			setFont(component.getFont());
		}
		setStroke(new BasicStroke());
	}

	public void clearRect(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return;

		Color background = getDefaultBackground();

		if (background == null) {
			background = Color.white;
		}

		synchronized (peerHandle) {
   			int handle = getPeer();
   			if (handle == 0) return;
			int colorHandle = background.getPeer();
   			OS.Graphics_FillBackground(handle, colorHandle, addWithoutOverflow(originX, x), addWithoutOverflow(originY, y), width, height);
		}
	}

	Color getDefaultBackground() {
		if (component != null) {
			return component.getBackground();
		} else {
			return null;
		}
	}

	public void clipRect(int x, int y, int width, int height) {
		if ((width < 0) || (height < 0)) {
			x = y = width = height = 0;
		}

		Rectangle newClip = new Rectangle(x + originX, y + originY, width, height);
		if (clip == null)
			setClip(newClip);
		else
			setClip(clip.intersection(newClip));
	}

	public void copyArea(int x, int y, int width, int height, int dx, int dy) {
		if (x < 0) {
			width += x;
			dx    += x;
			x      = 0;
		}

		if (y < 0) {
			height += y;
			dy     += y;
			y       = 0;
		}

		if ((width <= 0) || (height <= 0)) return;
		if ((dx == 0) && (dy == 0)) return;

		synchronized (peerHandle) {
   			int handle = getPeer();
   			if (handle == 0) return;
   			OS.Graphics_CopyArea(handle, originX + x, originY + y, width, height, originX + x + dx, originY + y + dy);
		}
	}

	public void dispose() {
		if (image != null) {
			ImagePeer imagePeer = image.getPrimitiveImagePeer();
			if (imagePeer != null) {
				// Don't dispose the imagepeer and the graphics at the same time.
				// Graphics could be disposed twice otherwise.
				synchronized (imagePeer) {
					// Resetting image/component here will abort any active
					// graphics primitive that does not have a graphics handle yet.
					image = null;
					component = null;
					imagePeer.disposeGraphicsHandle(peerHandle);
				}
			}
		} else {
			Component comp = component;
			if (comp != null) {
				// Resetting image/component here will abort any active
				// graphics primitive that does not have a graphics handle yet.
				image = null;
				component = null;

				comp.releaseGraphics(peerHandle);
			}
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to the spec:
	 * "The resulting rectangle covers an area that is width + 1 pixels wide
	 * by height + 1 pixels tall."
	 */
	public void draw3DRect(int x, int y, int width, int height, boolean raised) {
		if ((width < 0) || (height < 0)) return;

		synchronized (peerHandle) {
   			int handle = getPeer();
   			if (handle == 0) return;

			x = originX + x;
			y = originY + y;

			Color originalColor = getAdjustedColor(color);
			Color darkerColor = getAdjustedColor(color.darker());
			Color brighterColor = getAdjustedColor(color.brighter());

   			if (raised) OS.Graphics_SetForeground(handle, brighterColor.getPeer());
   			else OS.Graphics_SetForeground(handle, darkerColor.getPeer());

			int trX = x + width;
			int trY = y;

			int blX = x;
			int blY = y + height;
   			OS.Graphics_DrawLine(handle, x, y, trX-1, trY);

			// don't render on top of previous line (XOR mode compatibility)
   			OS.Graphics_DrawLine(handle, x, y+1, blX, blY);

			if (raised) {
    			OS.Graphics_SetForeground(handle, darkerColor.getPeer());
			} else {
    			OS.Graphics_SetForeground(handle, brighterColor.getPeer());
			}

   			OS.Graphics_DrawLine(handle, trX, blY, trX, trY);
			// don't render on top of previous line (XOR mode compatibility)
   			OS.Graphics_DrawLine(handle, trX-1, blY, blX+1, blY);

   			OS.Graphics_SetForeground(handle, originalColor.getPeer());
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to the spec:
	 * "The center of the arc is the center of the rectangle whose origin is (x,y)
	 * and whose size is specified by the width and height arguments.
	 * The resulting arc covers an area width + 1 pixels wide by height + 1 pixels tall."
	 */
	public void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
		if (arcAngle == 0) return;
		if ((width <= 0) || (height <= 0)) return;
		if ((width == 1) && (height == 1)) {
			drawLine(x,y,x,y);
			return;
		}

		synchronized (peerHandle) {
   			int handle = getPeer();
   			if (handle == 0) return;
   			OS.Graphics_DrawArc(handle, originX + x, originY + y, width + 1, height + 1, startAngle, arcAngle);
		}
	}

	public boolean drawImage(
		Image image,
		int x,
		int y,
		Color background,
		ImageObserver imageObserver) {

		boolean result = false;
		IImageImpl iImage= BufferedImageHandler.createIImageImpl(image);
		if (iImage != null) {
			synchronized (image) {
				ImagePeer imagePeer = iImage.getImagePeer(imageObserver);
				if (imagePeer != null) {
					Rectangle r = imagePeer.getBounds();
					synchronized (peerHandle) {
      					int handle = getPeer();
      					if (handle == 0) return false;

						if (background != null)
							OS.Graphics_FillBackground(handle, background.getPeer(), originX + x, originY + y, r.width, r.height);
						OS.Graphics_DrawImage(handle, imagePeer.fPeerHandle, originX + x, originY + y);
					}
					result = true;
				}
			}
		}

		return result;
	}

	public boolean drawImage(Image image,
		int x,
		int y,
		ImageObserver imageObserver)
	{
		return drawImage(image, x, y, null, imageObserver);
	}

	public boolean drawImage(
		Image image,
		int x,
		int y,
		int width,
		int height,
		Color background,
		ImageObserver imageObserver) {

		boolean result = false;
		IImageImpl iImage= BufferedImageHandler.createIImageImpl(image);
		if (iImage != null) {
			synchronized (image) {
				ImagePeer imagePeer = iImage.getImagePeer(imageObserver);
				if (imagePeer != null) {
					Rectangle r = imagePeer.getBounds();
					if (width == 0 || height == 0)
						return true;
					if (width < 0 && height < 0) {
						width = r.width;
						height = r.height;
					} else if (width < 0) {
						width = r.width * height / r.height;
					} else if (height < 0) {
						height = r.height * width / r.width;
					}
					synchronized (peerHandle) {
						int handle = getPeer();
						if (handle == 0) return false;

						if (background != null)
							OS.Graphics_FillBackground(handle, background.getPeer(), originX + x, originY + y, width, height);
						OS.Graphics_DrawImageRegion(handle, imagePeer.fPeerHandle, 0, 0, r.width, r.height, originX + x, originY + y, width, height);
					}
					result = true;
				}
			}
		}

		return result;
	}

	public boolean drawImage(
		Image image,
		int x,
		int y,
		int width,
		int height,
		ImageObserver imageObserver) {
			return drawImage(image, x, y, width, height, null, imageObserver);
	}

	public boolean drawImage(
		Image image,
		int dstX1,
		int dstY1,
		int dstX2,
		int dstY2,
		int srcX1,
		int srcY1,
		int srcX2,
		int srcY2,
		Color background,
		ImageObserver imageObserver) {

		if ((dstX1 == dstX2) || (dstY1 == dstY2)) return true;

		boolean result = false;
		IImageImpl iImage= BufferedImageHandler.createIImageImpl(image);
		if (iImage != null) {
			synchronized (image) {
				ImagePeer imagePeer = iImage.getImagePeer(imageObserver);
				if (imagePeer != null) {
					ImagePeer drawnImage = imagePeer;

					int srcX = srcX1;
					int srcY = srcY1;
					int srcW = srcX2 - srcX1;
					int srcH = srcY2 - srcY1;
					int dstX = dstX1;
					int dstY = dstY1;
					int dstW = dstX2 - dstX1;
					int dstH = dstY2 - dstY1;

					if ((0 == srcW) || (0 == srcH) || (0 == dstW) || (0 == dstH))  return true;

					// Needs a Horizontal flip
					// if the sign of the source width is not equal to the sign of the destination width
					boolean hFlip = (srcW / ((srcW > 0) ? srcW : -srcW)) != (dstW / ((dstW > 0) ? dstW : -dstW));
					// Needs a Vertical flip
					// if the sign of the source height is not equal to the sign of the destination height
					boolean vFlip = (srcH / ((srcH > 0) ? srcH : -srcH)) != (dstH / ((dstH > 0) ? dstH : -dstH));

					if (srcW < 0) {
						srcX = srcX + srcW;
						srcW = -srcW;
					}
					if (srcH < 0) {
						srcY = srcY + srcH;
						srcH = -srcH;
					}

					// Once normalized, if the source rectangle does not
					// intersect the image bounds, there is nothing to draw.
					Rectangle imageRect = imagePeer.getBounds();
					Rectangle sourceRect = new Rectangle(srcX, srcY, srcW, srcH);
					Rectangle newRect = sourceRect.intersection(imageRect);
					if (newRect.isEmpty()) return true;

					if (dstW < 0) {
						dstX = dstX + dstW;
						dstW = -dstW;
					}
					if (dstH < 0) {
						dstY = dstY + dstH;
						dstH = -dstH;
					}

					// scale the destination bounds if the source bounds
					// are outside the image bounds.
					if (srcX + srcW > imageRect.width || srcX < imageRect.x) {
						float rel = (float)dstW / srcW;
						// scale destination width to match actual source
						// bounds (clipped to image bounds)
						dstW = (int)(rel * newRect.width);
						if (dstW == 0) return true;
						if (srcX < imageRect.x) {
							int insetX = imageRect.x - srcX;
							dstX += (int)(rel * insetX);
						}
					}
					if (srcY + srcH > imageRect.height || srcY < imageRect.y) {
						float rel = (float)dstH / srcH;
						// scale destination height to match actual source
						// bounds (clipped to image bounds)
						dstH = (int)(rel * newRect.height);
						if (dstH == 0) return true;
						if (srcY < imageRect.y) {
							int insetY = imageRect.y - srcY;
							dstY += (int)(rel * insetY);
						}
					}

					if (hFlip || vFlip) {
						drawnImage = imagePeer.flip(hFlip, vFlip, newRect.x, newRect.y, newRect.width, newRect.height);
						// If flipping occurred, only the flipped sub-image will be
						// returned. We want to render the entire flipped image.
						newRect.x = 0;
						newRect.y = 0;
					}

					synchronized (peerHandle) {
						int handle = getPeer();
						if (handle == 0) return false;

						if (background != null)
							OS.Graphics_FillBackground(handle, background.getPeer(), originX + dstX, originY + dstY, dstW, dstH);

							OS.Graphics_DrawImageRegion(handle, drawnImage.fPeerHandle,
								newRect.x, newRect.y, newRect.width, newRect.height,
								originX + dstX, originY + dstY, dstW, dstH);
					}

					// If flipping occurred, clean up the temporary image peer.
					if (drawnImage != imagePeer) {
						drawnImage.dispose();
					}

					result = true;
				}
			}
		}

		return result;
	}

	public boolean drawImage(
		Image image,
		int dstX1,
		int dstY1,
		int dstX2,
		int dstY2,
		int srcX1,
		int srcY1,
		int srcX2,
		int srcY2,
		ImageObserver imageObserver) {
		return drawImage(image, dstX1, dstY1, dstX2, dstY2, srcX1, srcY1, srcX2, srcY2, null, imageObserver);
	}

	public void drawLine(int x1, int y1, int x2, int y2) {
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_DrawLine(handle, originX + x1, originY + y1, originX + x2, originY + y2);
		}
	}

	/*
	 *  (non-Javadoc)
	 * @see java.awt.Graphics#drawRect(int, int, int, int)
	 * According to the spec:
	 * "Draws the outline of an oval. The result is a circle or
	 * ellipse that fits within the rectangle specified by the x, y, width
	 * and height arguments. The oval covers an area that is width + 1 pixels
	 * wide and height + 1 pixels tall."
	 */
	public void drawOval(int x, int y, int width, int height) {
		if ((width < 0) || (height < 0)) return;
		if ((width == 0) && (height == 0)) return;

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_DrawOval(handle, originX + x, originY + y, width + 1, height + 1);
		}
	}

	public void drawPolygon(int xpoints[], int ypoints[], int npoints) {
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			int[] points = translate(xpoints, ypoints, npoints);
			OS.Graphics_DrawPolygon(handle, points, points.length);
		}
	}

	public void drawPolyline(int xpoints[], int ypoints[], int npoints) {
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			int[] points = translate(xpoints, ypoints, npoints);
			OS.Graphics_DrawPolyline(handle, points, points.length);
		}
	}

	/*
	 *  (non-Javadoc)
	 * @see java.awt.Graphics#drawRect(int, int, int, int)
	 * According to the spec:
	 * "The left and right edges of the rectangle are at x and x + width.
	 * The top and bottom edgs are at y and y + height."
	 */
	public void drawRect(int x, int y, int width, int height) {
		if ((width < 0) || (height < 0)) return;
		if ((width == 0) && (height == 0)) {
			drawLine(x,y,x,y);
			return;
		}
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_DrawRectangle(handle, originX + x, originY + y, width + 1, height + 1);
		}
	}

	/*
	 *  (non-Javadoc)
	 * @see java.awt.Graphics#drawRect(int, int, int, int)
	 * According to the spec:
	 * "The left and right edges of the rectangle are at x and x + width.
	 * The top and bottom edgs are at y and y + height."
	 */
	public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
		if ((width <= 0) || (height <= 0)) return;
		if ((width == 1) && (height == 1)) {
			drawLine(x,y,x,y);
			return;
		}

		if (arcWidth < 0) arcWidth = 0;
		if (arcHeight < 0) arcHeight = 0;

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_DrawRoundRectangle(handle, originX + x, originY + y, width + 1, height + 1, arcWidth, arcHeight);
		}
	}

//	public void drawString(AttributedCharacterIterator iterator, int x, int y) {
//		// attributes currently ignored
//		StringBuffer buf = new StringBuffer();
//		while (iterator.getIndex() < iterator.getEndIndex()) {
//			buf.append(iterator.next());
//		}
//		drawString(buf.toString(), x, y);
//	}

	public void drawString(String string, int x, int y) {
		if (string == null) throw new NullPointerException();

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;

			int ascent = 0;
			int leading = 0;
			if (font != null) {
				FontPeer peer = font.getPeer();
				ascent = peer.getAscent();
				leading = peer.getLeading();
			}
			// Graphics_DrawString expects y as top of bounding rectangle
			// PPro specs y as baseline
			OS.Graphics_DrawString(handle, string, originX + x, originY + y - (ascent + leading), OS.STRING_TRANSPARENT);
		}
	}

	public void fill3DRect(int x, int y, int width, int height, boolean raised) {
		if ((width <= 0) || (height <= 0)) return;

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;

			x = originX + x;
			y = originY + y;

			Color originalColor = getAdjustedColor(color);
			Color darkerColor = getAdjustedColor(color.darker());

			if (!raised) {
				OS.Graphics_SetForeground(handle, darkerColor.getPeer());
				OS.Graphics_SetBackground(handle, darkerColor.getPeer());
			}
			// draw filled rectangle such that it doesn't overlap with lower/right
			// lines (XOR mode compatibility)
			OS.Graphics_FillRectangle(handle, x,y,width-1,height-1);
			OS.Graphics_DrawLine(handle, x,y+height-1,x,y+height-1);

			if (raised) {
				OS.Graphics_SetForeground(handle, darkerColor.getPeer());
			} else {
				OS.Graphics_SetForeground(handle, originalColor.getPeer());
				OS.Graphics_SetBackground(handle, originalColor.getPeer());
			}

			OS.Graphics_DrawLine(handle, x+width-1,y, x+width-1, y+height-1);
			OS.Graphics_DrawLine(handle, x+width-2,y+height-1, x+1, y+height-1);

			if (raised) {
				OS.Graphics_SetForeground(handle, originalColor.getPeer());
			}
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to the spec:
	 * "The center of the arc is the center of the rectangle whose origin is (x,y)
	 * and whose size is specified by the width and height arguments.
	 * The resulting arc covers an area width + 1 pixels wide by height + 1 pixels tall."
	 */
	public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
		if (arcAngle == 0) return;
		if ((width <= 0) || (height <= 0)) return;
		if ((width == 1) && (height == 1)) {
			drawLine(x,y,x,y);
			return;
		}

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_FillArc(handle, originX + x, originY + y, width + 1, height + 1, startAngle, arcAngle);
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to the spec:
	 * "Fills an oval bounded by the specified rectangle with the current
	 * color"
	 * (This means it is 1 pixel wider and taller than the width and height,
	 * as in drawOval)
	 */
	public void fillOval(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return;
		if ((width == 1) && (height == 1)) {
			drawLine(x,y,x,y);
			return;
		}

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_FillOval(handle, originX + x, originY + y, width + 1, height + 1);
		}
	}

	public void fillPolygon(int xpoints[], int ypoints[], int npoints) {
		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			int[] points = translate(xpoints, ypoints, npoints);
			OS.Graphics_FillPolygon(handle, points, points.length);
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to spec:
	 * "The left and right edges of the rectangle are at x and x + width - 1.
	 * The top and bottom edges are at y and y + height - 1. The resulting
	 * rectangle covers an area <width> pixels wide and <height> pixels tall."
	 */
	public void fillRect(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return;

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_FillRectangle(handle, originX + x, originY + y, width, height);
		}
	}

	/*
	 *  (non-Javadoc)
	 * According to the spec:
	 * "The left and right edges of the rectangle are at x and x + width - 1.
	 * The top and bottom edgs are at y and y + height - 1."
	 */
	public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
		if ((width <= 0) || (height <= 0)) return;
		if ((width == 1) && (height == 1)) {
			drawLine(x,y,x,y);
			return;
		}

		if (arcWidth < 0) arcWidth = 0;
		if (arcHeight < 0) arcHeight = 0;

		synchronized (peerHandle) {
			int handle = getPeer();
			if (handle == 0) return;
			OS.Graphics_FillRoundRectangle(handle, originX + x, originY + y, width, height, arcWidth, arcHeight);
		}
	}

	public Shape getClip() {
		return getClipBounds();
	}

	public Rectangle getClipBounds() {
		if (clip == null) return null;

		return new Rectangle(clip.x - originX, clip.y - originY, clip.width, clip.height);
	}

	public Color getColor() {
		return color == null ? Color.black : color;
	}

	public Font getFont() {
		if (null != font) return font;

		setFont(Font.decode("")); //$NON-NLS-1$
		return font;
	}

	public FontMetrics getFontMetrics() {
		if (fontMetrics == null)
			fontMetrics= getFontMetrics(getFont());

		return fontMetrics;
	}

	public FontMetrics getFontMetrics(Font font) {
		return new FontMetricsImpl(font);
	}

	public boolean hitClip(int x, int y, int width, int height) {
		if (clip == null) return false;

		int newX = 0;
		int newY = 0;

		newX = addWithoutOverflow(x, originX);
		newY = addWithoutOverflow(y, originY);

		return clip.intersects(new Rectangle(newX, newY, width, height));
	}

	/**
	 * Returns <code>true</code> if the receiver can be rendered to.
	 *
	 * @return <code>true</code> if the receiver can be rendered to.
	 *         <code>false</code> if render operations should be no-ops
	 */
	int getPeer() {
		if (peerHandle.value != 0)
			return peerHandle.value;

		if ((component == null) && (image == null))
			return 0;

		if (image != null) {
			ImagePeer imagePeer = image.getImagePeer();
			imagePeer.newGraphicsHandle(peerHandle);
		} else {
			component.acquirePeerReadLock();
			try {
				if (component.peer != 0) {
					component.newGraphicsHandle(peerHandle);
				}
			} finally {
				component.releasePeerReadLock();
			}
		}

		if (peerHandle.value != 0) {
			if (nativeColor == null) nativeColor = getAdjustedColor(getColor());
			int colorPeer = nativeColor.getPeer();
			if (colorPeer != 0) {
				OS.Graphics_SetBackground(peerHandle.value, colorPeer);
				OS.Graphics_SetForeground(peerHandle.value, colorPeer);
			}

			setClipImpl(clip);

			if (font != null)
				OS.Graphics_SetFont(peerHandle.value, font.getPeer().fPeerHandle);

			OS.Graphics_SetXORMode(peerHandle.value, xorColor == null ? false : true);

			BasicStroke basicStroke = (BasicStroke) stroke;
			OS.Graphics_SetLineWidth(peerHandle.value, (int) basicStroke.getLineWidth());
			OS.Graphics_SetLineNodeStyle(peerHandle.value, peerEndCapStyle(basicStroke), peerLineJoinStyle(basicStroke), (int) basicStroke.getMiterLimit());
		}

		return peerHandle.value;
	}

	public void setClip(int x, int y, int width, int height) {
		setClip(new Rectangle(x + originX, y + originY, width, height));
	}

	void setClip(Rectangle clip) {
		if (this.clip == clip) return;

		this.clip = clip;
		setClipImpl(clip);
	}

	void setClipImpl(Rectangle clip){
		synchronized (peerHandle) {
			if (peerHandle.value == 0) return;

			if (clip != null) {
				OS.Graphics_SetClipping(peerHandle.value,
						clip.x, clip.y,
						clip.width < 0 ? 0 : clip.width,
						clip.height < 0 ? 0 : clip.height);
			} else {
				OS.Graphics_ClearClipping(peerHandle.value);
			}
		}
	}

	public void setClip(Shape shape) {
		if (shape == null) {
			setClip(null);
		} else if (shape instanceof Rectangle) {
			Rectangle newClip = (Rectangle) shape;
			setClip(newClip.x, newClip.y, newClip.width, newClip.height);
		} else {
			throw new IllegalArgumentException();
		}
	}

	public void setColor(Color color) {
		if (color == null) color = Color.black;

		if (this.color == color) return;
		if (color.equals(this.color)) return; // Assuming this is a valid optimization
		this.color = color;

		updateNativeColor(color);
	}

	void updateNativeColor(Color newColor) {
		// Get the adjusted color before checking the peer handle
		// so that the native color (and logicalAlpha) will be ready
		// then the graphics peer is finally created
		newColor = getAdjustedColor(newColor);
		if (newColor.equals(nativeColor)) return;
		nativeColor = newColor;

		synchronized (peerHandle) {
			if (peerHandle.value == 0) return;

			int colorPeer = newColor.getPeer();
			if (colorPeer != 0) {
				OS.Graphics_SetBackground(peerHandle.value, colorPeer);
				OS.Graphics_SetForeground(peerHandle.value, colorPeer);
			}
		}
	}

	Color getAdjustedColor(Color newColor) {
		if (xorColor != null) {
			return new Color(newColor.value ^ xorColor.value);
		} else {
			return newColor;
		}
	}

	public void setFont(Font font) {
		if (font == null) font = Font.decode(""); //$NON-NLS-1$

		if (this.font == font) return;
		if (font.equals(this.font)) return; // Assuming this is a valid optimization
		this.font = font;
		this.fontMetrics = null;

		if (font != null) {
			synchronized (peerHandle) {
				if (peerHandle.value == 0) return;
				OS.Graphics_SetFont(peerHandle.value, font.getPeer().fPeerHandle);
			}
		}
	}

	public void setPaintMode() {
		if (xorColor == null) return;

		xorColor = null;

		updateNativeColor(color);

		synchronized (peerHandle) {
			if (peerHandle.value == 0) return;

			OS.Graphics_SetXORMode(peerHandle.value, false);
		}
	}

	public void setXORMode(Color xorColor) {
		if (xorColor == null) throw new IllegalArgumentException();

		if (this.xorColor == xorColor) return;

		this.xorColor = xorColor;
		synchronized (peerHandle) {
			// Don't return immediately if the peer handle
			// is 0 so that the native color will be
			// updated below
			if (peerHandle.value != 0) {
				OS.Graphics_SetXORMode(peerHandle.value, true);
			}
		}

		// Keep the native color up to date so that
		// when the peer is created it will be ready
		// to use
		updateNativeColor(color);
	}

	public void translate(int x, int y) {
		this.originX += x;
		this.originY += y;
	}

	int[] translate(int xpoints[], int ypoints[], int npoints) {
		int points[] = new int[npoints * 2];
		int count = 0;
		for (int i=0; i<npoints; i++) {
			points[count++] = addWithoutOverflow(xpoints[i], originX);
			points[count++] = addWithoutOverflow(ypoints[i], originY);
		}
		return points;
	}

	static int addWithoutOverflow(int value1, int value2) {
		long result = (long) value1 + (long) value2;
		return (int) Math.max(Integer.MIN_VALUE, Math.min(result, Integer.MAX_VALUE));
	}

	public void drawString(AttributedCharacterIterator iterator, int x, int y) {
		Font oldFont = getFont();
		Color oldForeground = getColor();
		char current = iterator.current();
		Graphics currentGraphics = this;
		Image offScreen = null;
//		if (component != null) {
//			// Create an offscreen transparent image so that the painting
//			// of the text happens all at once rather than animated.
//			offScreen = component.createImage(component.getWidth(), component.getHeight(), 0x00ffffff);
//			currentGraphics = offScreen.getGraphics();
//		}
		while (iterator.getIndex() < iterator.getEndIndex()) {
			Map attrs = iterator.getAttributes();
			Color newColor = (Color) attrs.get(TextAttribute.FOREGROUND);
			if (newColor != null) {
				currentGraphics.setColor(newColor);
			} else {
				currentGraphics.setColor(oldForeground);
			}
			Font newFont = (Font) attrs.get(TextAttribute.FONT);
			if (newFont == null) {
				// FONT attribute overrides any FAMILY, SIZE, WEIGHT, or POSTURE setting
				String family = (String) attrs.get(TextAttribute.FAMILY);
				if (family == null) {
					family = oldFont.getFamily();
				}
				Float size = (Float) attrs.get(TextAttribute.SIZE);
				if (size == null) {
					size = new Float(oldFont.getSize());
				}
				Float weight = (Float) attrs.get(TextAttribute.WEIGHT);
				int style = Font.PLAIN;
				if (weight != null) {
					if (weight.floatValue() >= TextAttribute.WEIGHT_BOLD.floatValue()) {
						style |= Font.BOLD;
					}
				}
				Float posture = (Float) attrs.get(TextAttribute.POSTURE);
				if ((posture != null) && (!(TextAttribute.POSTURE_REGULAR.equals(posture)))) {
					style |= Font.ITALIC;
				}
				newFont = new Font((family == null ? "" : family) , style , (int) size.floatValue());
			} else {
				// Create a copy of the font so the setFont() call will be
				// effective each time the extraUglStyles are changed (otherwise
				// the peer never changes)
				newFont = new Font(newFont.getFamily(), newFont.getStyle(), newFont.getSize());
			}
			// Set underline & strikethrough outside font == null block,
			// since it seems J2SE lets you change those styles even if
			// a FONT attribute is in effect
			newFont.extraUglStyles = 0;
			Integer underline = (Integer) attrs.get(TextAttribute.UNDERLINE);
			if (TextAttribute.UNDERLINE_ON.equals(underline)) {
				// Might consider drawing the line ourselves in the common
				// layer instead of using actual peer font attributes.
				// A bit problematic to calculate correct line width for
				// underline though.  Same concern for strikethrough.
				newFont.extraUglStyles |= OS.FONT_STYLE_UNDERLINED;
			}
			Boolean strikethrough = (Boolean) attrs.get(TextAttribute.STRIKETHROUGH);
			if (TextAttribute.STRIKETHROUGH_ON.equals(strikethrough)) {
				newFont.extraUglStyles |= OS.FONT_STYLE_STRIKETHROUGH;
			}
			currentGraphics.setFont(newFont);
			FontMetrics fm = currentGraphics.getFontMetrics(newFont);
			currentGraphics.drawString(current + "", x, y);
			x += fm.charWidth(current);
			current = iterator.next();
		}
		if (offScreen != null) {
			drawImage(offScreen, 0, 0, null);
		}
		setFont(oldFont);
		setColor(oldForeground);
	}

	public void setStroke(Stroke stroke) {
		if (stroke == null) {
			throw new IllegalArgumentException("null Stroke"); //$NON-NLS-1$
		}
		if (!(stroke instanceof BasicStroke)) {
			throw new IllegalArgumentException("custom stroke not supported"); //$NON-NLS-1$
		}
		if (stroke.equals(this.stroke)) return;
		this.stroke = stroke;

		synchronized (peerHandle) {
			if (peerHandle.value == 0) return;
			BasicStroke basicStroke = (BasicStroke) stroke;
			OS.Graphics_SetLineWidth(peerHandle.value, (int) basicStroke.getLineWidth());
			OS.Graphics_SetLineNodeStyle(peerHandle.value, peerEndCapStyle(basicStroke), peerLineJoinStyle(basicStroke), (int) basicStroke.getMiterLimit());
		}
	}

	protected int peerLineJoinStyle(BasicStroke basicStroke) {
		switch (basicStroke.getLineJoin()) {
			case BasicStroke.JOIN_ROUND:
				return OS.LINE_JOIN_ROUND;

			case BasicStroke.JOIN_BEVEL:
				return OS.LINE_JOIN_BEVEL;

			default: // JOIN_MITER
				return OS.LINE_JOIN_MITER;
		}
	}

	protected int peerEndCapStyle(BasicStroke basicStroke) {
		switch (basicStroke.getEndCap()) {
			case BasicStroke.CAP_BUTT:
				return OS.LINE_CAP_BUTT;

			case BasicStroke.CAP_ROUND:
				return OS.LINE_CAP_ROUND;

			default: // CAP_SQUARE
				return OS.LINE_CAP_SQUARE;
		}
	}

	public Stroke getStroke() {
		return stroke;
	}
}
