package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.image.*;

/**
 * <code>Image</code> is the abstract superclass of all
 * classes that represent a graphical image.
 *
 * @author		IBM
 * @version		initial
 */
public abstract class Image {

	/**
	 * Constants that specify the scaling algorithm.
	 */
	public static final int SCALE_DEFAULT = 1;
	public static final int SCALE_FAST = 2;
	public static final int SCALE_SMOOTH = 4;
	public static final int SCALE_REPLICATE = 8;
	public static final int SCALE_AREA_AVERAGING = 16;

	public static final Object UndefinedProperty = new Object();

	/**
	 * Answers the width of the receiver if it is known.  Otherwise
	 * answers -1 and the specified <code>ImageObserver</code> will
	 * be notified when this value is available.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *							notification
	 *
	 * @return	an <code>int</code> which is the receiver's width or <code>-1</code>
	 */
	public abstract int getWidth(ImageObserver imageObserver);

	/**
	 * Answers the height of the receiver if it is known.  Otherwise
	 * answers -1 and the specified <code>ImageObserver</code> will
	 * be notified when this value is available.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *							notification
	 *
	 * @return	an <code>int</code> which is the receiver's height or <code>-1</code>
	 */
	public abstract int getHeight(ImageObserver imageObserver);

	/**
	 * Answers the <code>ImageProducer</code> that generates the
	 * pixels for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>ImageProducer</code> which is the receiver's source
	 */
	public abstract ImageProducer getSource();

	/**
	 * Answers a <code>Graphics</code> used for drawing
	 * on the receiver or <code>null</code> if the receiver
	 * is not a off-screen image.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Graphics for drawing
	 */
	public abstract Graphics getGraphics();

	/**
	 * Answers the property specified by the given name if it is
	 * known.  Otherwise answers <code>null</code> and the specified
	 * <code>ImageObserver</code> will be notified when this value
	 * is available.
	 * <p>
	 * If the property is not defined for the receiver, answers
	 * <code>UndefinedProperty</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the desired property
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *								notification
	 *
	 * @return	an <code>Object</code> which is the property value or <code>null</code>
	 */
	public abstract Object getProperty(String propertyName, ImageObserver imageObserver);

	/**
	 * Answers the scaled version of the receiver.  If <code>width</code>
	 * or <code>height</code> is negative, uses a value that keeps
	 * the aspect ratio of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	width	the width of the new image
	 * @param	height	the height of the new image
	 * @param	hints	a flag specifying the scaling algorithm
	 *
	 * @return	an <code>Image</code> which is the scaled
	 */
	public Image getScaledInstance(int width, int height, int hints) {
		if (width < 0) {
			if (height < 0)
				width = getWidth(null);
			else
				width = (getWidth(null) * height) / getHeight(null);
		}
		if (height < 0) {
			if (width < 0)
				height = getHeight(null);
			else
				height = (getHeight(null) * width) / getWidth(null);
		}

		ImageFilter filter;
		if ((hints & SCALE_SMOOTH) > 0 || (hints & SCALE_AREA_AVERAGING) > 0)
			filter = new AreaAveragingScaleFilter(width, height);
		else
			filter = new ReplicateScaleFilter(width, height);

		return Toolkit.getDefaultToolkit().createImage(
			new FilteredImageSource(getSource(), filter));
	}

	/**
	 * Releases all resources used by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public abstract void flush();
}
