package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.image.*;
import java.net.URL;
import java.util.Hashtable;
import java.util.Vector;

import com.ibm.oti.awt.*;
import com.ibm.oti.awt.image.*;
import com.ibm.oti.awt.metal.image.ImagePeer;

class ImageImpl extends Image implements IImageImpl {

	ImageProducer fProducer;
	Vector fObservers;

	Component fComponent;
	ColorModel fColorModel;
	ImagePeer tImagePeer;
	Hashtable properties;

	int fStatus;

	ImageImpl() {
		fObservers= new Vector();
		fStatus = IImageImpl.NOT_STARTED;
	}

	public ImageImpl(byte[] data, int offset, int length) {
		this();
		fProducer= new ByteArrayImageProducer(data, offset, length);
	}

	public ImageImpl(URL url) {
		this();
		fProducer= new URLImageProducer(url);
	}

	public ImageImpl(String filename) {
		this();
		fProducer= new FileImageProducer(filename);
	}

	public ImageImpl(ImageProducer imageProducer) {
		this();
		fProducer= imageProducer;
	}

	public ImageImpl(Component component, int width, int height, Color fillColor) {
		this();
		fComponent= component;
		tImagePeer = ImagePeer.createImmutableCopy(Util.getDisplay(), width, height, fillColor);
		fStatus = IImageImpl.FINISHED;
		fProducer = new OffScreenImageProducer(tImagePeer);
	}

	public void setProperties(Hashtable properties) {
		this.properties = properties;
	}

	/**
	 * @see java.awt.Image#getWidth(ImageObserver)
	 */
	public int getWidth(ImageObserver imageObserver) {
		ImagePeer imagePeer= getImagePeer(imageObserver);
		if (imagePeer == null)
			return -1;
		Rectangle peerBounds = imagePeer.getBounds();
		return peerBounds.width;
	}

	/**
	 * @see java.awt.Image#getHeight(ImageObserver)
	 */
	public int getHeight(ImageObserver imageObserver) {
		ImagePeer imagePeer= getImagePeer(imageObserver);
		if (imagePeer == null)
			return -1;
		Rectangle peerBounds = imagePeer.getBounds();
		return peerBounds.height;
	}

	/**
	 * @see java.awt.Image#getSource()
	 */
	public ImageProducer getSource() {
		return fProducer;
	}

	public Component getComponent() {
		return fComponent;
	}

	/**
	 * @see java.awt.Image#getGraphics()
	 */
	public Graphics getGraphics() {
		try {
			if (fComponent != null) {
				return new GraphicsACImpl(fComponent, this);
			} else {
				throw new UnsupportedOperationException("getGraphics() only valid for images created with createImage(w, h)"); //$NON-NLS-1$
			}
		} catch(IllegalArgumentException e) {
			return null;
		}
	}

	/**
	 * @see java.awt.Image#getProperty(String, ImageObserver)
	 */
	public Object getProperty(String propertyName, ImageObserver imageObserver) {
		ImagePeer imagePeer= getImagePeer(imageObserver);
		if (imagePeer != null) {
			if (properties != null) {
				if (properties.containsKey(propertyName)) {
					Object value= properties.get(propertyName);
					if (value == null && imageObserver != null) {
						if (!fObservers.contains(imageObserver)) {
							fObservers.addElement(imageObserver);
						}
					}
					return value;
				}
			}

			// Since the image peer is non-null, any property not in
			// the hashtable should be 'Undefined'
			return UndefinedProperty;
		}

		// If the peer doesn't exist then null should be returned
		// to show that "the properties for this image are not yet known"
		return null;
	}

	/**
	 * Perform clean up for the receiver as it has become garbage.
	 * Called by the virtual machine when there are no longer any (non-weak)
	 * references to the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @exception	Throwable
	 */
	protected void finalize() throws Throwable {
		// This was removed in fixing bug# 108929. Removing the delay ensures
		// resources are released in a more timely  manner (eg. an application
		// does a System.gc() during a paint event).
		// There are two reasons why code should not be run on the finalize thread:
		// 1. Any exceptions that are thrown are eaten by the vm so never seen
		// 2. If the finalize thread runs too long, the VM will shut down the thread.
//		EventQueue.invokeLater(new Runnable() {
//			public void run() {
//				// Always call flush() for subclasses
//				flush();
//				_flush(false);
//			}
//		});

		try {
			flush();
			_flush(false);
		} catch (Throwable t) {
			System.out.println("Exception caught during finalization: " + t); //$NON-NLS-1$
			t.printStackTrace();
		}
	}

	/**
	 * @see java.awt.Image#flush()
	 */
	public void flush() {
		// Force even off-screen images to be flushed.
		_flush(true);
	}

	synchronized void _flush(boolean checkForNonFlushables) {
		// Per the spec, flushing an off-screen image does nothing
		// (unless it's being flushed as a result of having been finalized)
		if ((checkForNonFlushables) && (fComponent != null)) return;

		disposeImage();
		setStatus(IImageImpl.NOT_STARTED);
	}

	//--------------------------------------------------------------------------------
	public ImagePeer getImagePeer() {
		return getImagePeer(null);
	}

	public ImagePeer getImagePeer(ImageObserver imageObserver) {
		if ((fStatus == IImageImpl.NOT_STARTED) ||
			(fStatus == IImageImpl.LOADING) ||
		    (fStatus == IImageImpl.FRAME_DONE)) {
			if (imageObserver != null) {
				if (!fObservers.contains(imageObserver)) {
					fObservers.addElement(imageObserver);
				}
			}
			if (fStatus == IImageImpl.NOT_STARTED) {
				AbstractImageProducer.buildImageFromProducer(fProducer, BufferedImageHandler.createIImageImpl(this));
			}
		} else
		if (fStatus == IImageImpl.ERRORED) {
			// When an image has errored, the peer will always return null.
			// When the image peer returns null, the imageObserver needs to
			// be notified in some way.  Since the error state is already known
			// the observer can be called directly to explain the null return value.
			//
			// This notifying behavior might seem weird, but it makes the use of
			// the observer argument consistently used when the peer is null, and the
			// behavior is checked by the PPro 1.1 TCK.
			if (imageObserver != null) imageObserver.imageUpdate(this, ImageObserver.ERROR, 0, 0, 0, 0);
		}

		return tImagePeer;
	}

	public ImagePeer getPrimitiveImagePeer() {
		return tImagePeer;
	}

	public void setImagePeer(ImagePeer imagePeer) {
		// Dispose any existing image before replacing.
		disposeImage();

		tImagePeer= imagePeer;
	}

	public int getStatus() {
		return fStatus;
	}

	public void setStatus(int status) {
		fStatus= status;
	}

	public void disposeImage() {
		if (tImagePeer == null) return;

		fColorModel= null;
		tImagePeer.dispose();
		tImagePeer= null;
	}

	public void notifyObservers(int infoflags, int x, int y, int width, int height) {
		for (int i=fObservers.size() - 1; i>=0; i--) {
			ImageObserver observer= (ImageObserver) fObservers.elementAt(i);
			boolean keepNotifing= observer.imageUpdate(this, infoflags, x, y, width, height);
			if (!keepNotifing)
				fObservers.removeElement(observer);
		}
	}
}
