package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2005  All Rights Reserved
 */

import java.awt.event.*;
import java.beans.*;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.ibm.oti.awt.Util;

public abstract class KeyboardFocusManager implements KeyEventDispatcher, KeyEventPostProcessor {

	public static final int FORWARD_TRAVERSAL_KEYS = 0;
	public static final int BACKWARD_TRAVERSAL_KEYS = 1;
	public static final int UP_CYCLE_TRAVERSAL_KEYS = 2;
	public static final int DOWN_CYCLE_TRAVERSAL_KEYS = 3;

	static final boolean FocusDebug = Util.getBooleanProperty("FocusDebug", false);

	String[] focusTraversalPropertyNames = new String[] {
	        "forwardDefaultFocusTraversalKeys",
	        "backwardDefaultFocusTraversalKeys",
	        "upCycleDefaultFocusTraversalKeys",
	        "downCycleDefaultFocusTraversalKeys"
	};

	static KeyboardFocusManager currentKeyboardFocusManager = new KeyboardFocusManagerImpl();
	FocusTraversalPolicy defaultFocusTraversalPolicy;
	Set[] defaultFocusTraversalKeys = new Set[] {null, null, null, null};
	transient Set[] platformTraversalKeys = new Set[] {null, null, null, null};

	LinkedList keyEventDispatchers = new LinkedList();
	LinkedList keyEventPostProcessors = new LinkedList();
	PropertyChangeSupport propertyChangeSupporter = new PropertyChangeSupport(this);
	VetoableChangeSupport vetoableChangeSupporter = new VetoableChangeSupport(this);

	Component focusOwner;
	Component permanentFocusOwner;
	Window focusedWindow;
	Window activeWindow;
	Container currentFocusCycleRoot;

	KeyboardFocusManager() {
	    defaultFocusTraversalPolicy = new DefaultFocusTraversalPolicy();
	    Set keys = new HashSet();
	    keys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, 0, false));
	    keys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, InputEvent.CTRL_DOWN_MASK, false));
	    platformTraversalKeys[FORWARD_TRAVERSAL_KEYS] = keys;
	    keys = new HashSet();
	    keys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, InputEvent.SHIFT_DOWN_MASK | InputEvent.CTRL_DOWN_MASK, false));
	    keys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, InputEvent.SHIFT_DOWN_MASK, false));
	    platformTraversalKeys[BACKWARD_TRAVERSAL_KEYS] = keys;
	    keys = Collections.EMPTY_SET;
	    platformTraversalKeys[UP_CYCLE_TRAVERSAL_KEYS] = keys;
	    platformTraversalKeys[DOWN_CYCLE_TRAVERSAL_KEYS] = keys;
	}

	static boolean getDebug() {
		return FocusDebug;
	}
	public static KeyboardFocusManager getCurrentKeyboardFocusManager() {
	    return currentKeyboardFocusManager;
	}

	public Component getFocusOwner() {
	    // Only return the focusOwner if the focusOwner is in the same context as the
	    // calling thread.
	    if (!isSecure()) return null;
		return focusOwner;
	}

	Component getGlobalFocusOwner() throws SecurityException {
		checkSecurity();
		return focusOwner;
	}

	void setGlobalFocusOwner(Component component) {
	    if (component != null) {
	        if (!component.isFocusable()) return;
	    }
	    if (getDebug()) {
	    	if (component == null) System.out.println("set global focus owner null");
	    	else System.out.println("set global focus owner " + component.getName());
	    }
	    Component oldFocusOwner = focusOwner;
	    try {
	        fireVetoableChange("focusOwner", oldFocusOwner, component);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    focusOwner = component;
	    Container root = null;
	    if (component == null) {
	    	root = null;
	    } else {
	    	root = component.getFocusCycleRootAncestor();
	    }
	    setGlobalCurrentFocusCycleRoot(root);
	    fireFocusPropertyChange("focusOwner", oldFocusOwner, focusOwner);
	}
	private void checkSecurity() {
	    // TODO PPro Focus 1.1 - investigate why this fails sometimes
	    if (this != KeyboardFocusManager.getCurrentKeyboardFocusManager()) {
	        throw new SecurityException("this KeyboardFocusManager is not installed in the current thread's context");
	    }
	}
	private boolean isSecure() {
		try {
			checkSecurity();
		} catch (SecurityException e) {
			return false;
		}
		return true;
	}
	/**
	 * Clears the global focus owner at both the Java and native levels. If a focusOwner
	 * exists, send the Component a permanent FOCUS_LOST event. This operation
	 * does not change the focusedWindow or the activeWindow.
	 */
	public void clearGlobalFocusOwner() {
		Util.getDisplay().clearFocus();
	}

	public Component getPermanentFocusOwner() {
		return permanentFocusOwner;
	}

	Component getGlobalPermanentFocusOwner() throws SecurityException {
		checkSecurity();
		return permanentFocusOwner;
	}

	void setGlobalPermanentFocusOwner(Component component) {
//	    if (component == null) return;
	    Component oldOwner = permanentFocusOwner;
	    try {
	        fireVetoableChange("permanentFocusOwner", oldOwner, component);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    permanentFocusOwner = component;
	    fireFocusPropertyChange("permanentFocusOwner", oldOwner, permanentFocusOwner);
	}

	public Window getFocusedWindow() {
	    // Only return the focusedWindow if the focusedWindow is in the same context as the
	    // calling thread.
	    if (!isSecure()) return null;
		return focusedWindow;
	}

	Window getGlobalFocusedWindow() throws SecurityException {
		checkSecurity();
		return focusedWindow;
	}

	void setGlobalFocusedWindow(Window window) {
//	    if (window == null) return;
	    if (window != null) {
	    	if (!window.isFocusableWindow()) return;
	    }
	    if (getDebug()) {
	    	if (window == null) System.out.println("set global focus window null");
	    	else System.out.println("set global focus window " + window.getName());
	    }
	    Window oldFocusedWindow = focusedWindow;
	    try {
	        fireVetoableChange("focusedWindow", oldFocusedWindow, window);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    focusedWindow = window;
	    fireFocusPropertyChange("focusedWindow", oldFocusedWindow, focusedWindow);
	}

	public Window getActiveWindow() {
	    // Only return the activeWindow if the activeWindow is in the same context as the
	    // calling thread.
	    if (!isSecure()) return null;
		return activeWindow;
	}

	public Container getCurrentFocusCycleRoot() {
	    // Only return the currentFocusCycleRoot if the currentFocusCycleRoot is in the same context as the
	    // calling thread.
	    if (!isSecure()) return null;
		return currentFocusCycleRoot;
	}

	Window getGlobalActiveWindow() throws SecurityException {
		checkSecurity();
		return activeWindow;
	}

	void setGlobalActiveWindow(Window window) {
	    Window oldActiveWindow = activeWindow;
	    try {
	        fireVetoableChange("activeWindow", oldActiveWindow, window);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    activeWindow = window;
	    fireFocusPropertyChange("activeWindow", oldActiveWindow, activeWindow);
	}

	public FocusTraversalPolicy getDefaultFocusTraversalPolicy() {
	    if (defaultFocusTraversalPolicy == null) {
	        defaultFocusTraversalPolicy = new DefaultFocusTraversalPolicy();
	    };
	    return defaultFocusTraversalPolicy;
	}

	public void setDefaultFocusTraversalPolicy(FocusTraversalPolicy focusTraversalPolicy) {
	    if (defaultFocusTraversalPolicy.equals(focusTraversalPolicy)) return;
	    FocusTraversalPolicy oldFocusTraversalPolicy = defaultFocusTraversalPolicy;
	    try {
	        fireVetoableChange("defaultFocusTraversalPolicy", oldFocusTraversalPolicy, focusTraversalPolicy);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    defaultFocusTraversalPolicy = focusTraversalPolicy;
	    firePropertyChange("defaultFocusTraversalPolicy", oldFocusTraversalPolicy, focusTraversalPolicy);
	}

	public void setDefaultFocusTraversalKeys(int id, Set keys) {
	    if (!isValidTraversalKey(id)) throw new IllegalArgumentException("invalid focus traversal key identifier");
	    if (!isValidTraversalKeystrokes(id, keys)) throw new IllegalArgumentException("invalid focus traversal keystrokes");
	    Set oldKeys = defaultFocusTraversalKeys[id];
	    try {
	        fireVetoableChange(focusTraversalPropertyNames[id], oldKeys, keys);
	    } catch (PropertyVetoException e) {
	        return;
	    }
	    defaultFocusTraversalKeys[id] = keys;
	    firePropertyChange(focusTraversalPropertyNames[id], oldKeys, keys);
	}

	public Set getDefaultFocusTraversalKeys(int id) {
	    if ((id != KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) &&
	            (id != KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS) &&
	            (id != KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS) &&
	            (id != KeyboardFocusManager.DOWN_CYCLE_TRAVERSAL_KEYS)) throw new IllegalArgumentException ("invalid focus traversal key identifier");
	    Set keys = defaultFocusTraversalKeys[id];
	    if (keys == null) keys = platformTraversalKeys[id];
	    return keys;
	}

	Container getGlobalCurrentFocusCycleRoot() throws SecurityException {
		checkSecurity();
		return currentFocusCycleRoot;
	}

	void setGlobalCurrentFocusCycleRoot(Container cycleRoot) {
	    if (cycleRoot == null) return;
	    Container oldRoot = currentFocusCycleRoot;
	    currentFocusCycleRoot = cycleRoot;
	    fireFocusPropertyChange("currentFocusCycleRoot", oldRoot, currentFocusCycleRoot);
	}

	public void addPropertyChangeListener(PropertyChangeListener propertyChangeListener) {
	    if (propertyChangeListener == null) return;
	    propertyChangeSupporter.addPropertyChangeListener(propertyChangeListener);
	}

	public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener) {
	    if (propertyChangeListener == null) return;
	    propertyChangeSupporter.removePropertyChangeListener(propertyChangeListener);
	}

	public PropertyChangeListener[] getPropertyChangeListeners() {
	    return propertyChangeSupporter.getPropertyChangeListeners();
	}

	void fireFocusPropertyChange(String property, Object oldValue, Object newValue) {
		PropertyChangeListener[] listeners = getPropertyChangeListeners();
		PropertyChangeEvent lostEvent = null;
		PropertyChangeEvent gainedEvent = null;
		if (oldValue != null) {
			lostEvent = new PropertyChangeEvent(this, property, oldValue, null);
		}
		if (newValue != null) {
			gainedEvent = new PropertyChangeEvent(this, property, null, newValue);
		}
		for (int i=0; i < listeners.length; i++) {
			PropertyChangeListener listener = listeners[i];
			if (lostEvent != null) listener.propertyChange(lostEvent);
			if (gainedEvent != null) listener.propertyChange(gainedEvent);
		}
	}
	void firePropertyChange(String property, Object oldValue, Object newValue) {
	    propertyChangeSupporter.firePropertyChange(property, oldValue, newValue);
	}

	public void addVetoableChangeListener(VetoableChangeListener vetoableChangeListener) {
	    if (vetoableChangeListener == null) return;
	    vetoableChangeSupporter.addVetoableChangeListener(vetoableChangeListener);
	}

	public void removeVetoableChangeListener(VetoableChangeListener vetoableChangeListener) {
	    if (vetoableChangeListener == null) return;
	    vetoableChangeSupporter.removeVetoableChangeListener(vetoableChangeListener);
	}

	public VetoableChangeListener[] getVetoableChangeListeners() {
	    return vetoableChangeSupporter.getVetoableChangeListeners();
	}

	void fireVetoableChange(String property, Object oldValue, Object newValue) throws PropertyVetoException {
	    vetoableChangeSupporter.fireVetoableChange(property, oldValue, newValue);
	}

	public void addKeyEventDispatcher(KeyEventDispatcher keyEventDispatcher) {
	    if (keyEventDispatcher == null) return;
	    keyEventDispatchers.add(keyEventDispatcher);
	}

	public void removeKeyEventDispatcher(KeyEventDispatcher keyEventDispatcher) {
	    if (keyEventDispatcher == null) return;
	    keyEventDispatchers.remove(keyEventDispatcher);
	}

	List getKeyEventDispatchers() {
		return keyEventDispatchers;
	}

	public void addKeyEventPostProcessor(KeyEventPostProcessor keyEventPostProcessor) {
	    if (keyEventPostProcessor == null) return;
	    keyEventPostProcessors.add(keyEventPostProcessor);
	}

	public void removeKeyEventPostProcessor(KeyEventPostProcessor keyEventPostProcessor) {
	    if (keyEventPostProcessor == null) return;
	    keyEventPostProcessors.remove(keyEventPostProcessor);
	}

	List getKeyEventPostProcessors() {
		return keyEventPostProcessors;
	}

	public abstract boolean dispatchEvent(AWTEvent awtevent0);

	public final void redispatchEvent(Component component, AWTEvent awtEvent) {
		if (getDebug()) System.out.println("redispatch event " + awtEvent.toString());
		awtEvent.focusManagerIsDispatching = true;
	    component.dispatchEvent(awtEvent);
	    awtEvent.focusManagerIsDispatching = false;
	}

	public abstract boolean dispatchKeyEvent(KeyEvent keyevent0);

	public abstract boolean postProcessKeyEvent(KeyEvent keyevent0);

	public abstract void processKeyEvent(Component component0, KeyEvent keyevent1);

	public abstract void focusNextComponent(Component component0);

	public abstract void focusPreviousComponent(Component component0);

	public abstract void upFocusCycle(Component component0);

	public abstract void downFocusCycle(Container container0);

	public final void focusNextComponent() {
	    Component owner = getGlobalFocusOwner();
	    if (owner == null) return;
	    focusNextComponent(owner);
	}

	public final void focusPreviousComponent() {
	    Component owner = getGlobalFocusOwner();
	    if (owner == null) return;
	    focusPreviousComponent(owner);
	}

	public final void upFocusCycle() {
	    Component owner = getGlobalFocusOwner();
	    if (owner == null) return;
	    upFocusCycle(owner);
	}

	public final void downFocusCycle() {
	    Component owner = getGlobalFocusOwner();
	    if (owner == null) return;
	    if (owner instanceof Container) {
	        Container cont = (Container)owner;
		    if (!cont.isFocusCycleRoot()) return;
		    downFocusCycle(cont);
	    }
	}

	boolean isValidTraversalKey(int id) {
	    return ((id == KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) ||
		    	(id == KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS) ||
    			(id == KeyboardFocusManager.DOWN_CYCLE_TRAVERSAL_KEYS));
	}

	boolean isValidTraversalKeystrokes(int id, Set keystrokes) {
		if (keystrokes == null) return false;
		Iterator iterator = keystrokes.iterator();
		while (iterator.hasNext()) {
			Object next = iterator.next();
			if (next == null) return false;
			AWTKeyStroke keystroke = null;
			if (next instanceof AWTKeyStroke) {
				keystroke = (AWTKeyStroke) next;
			}
			if (keystroke == null) {
			    return false;
			}
			if (keystroke.getKeyEventType() == KeyEvent.KEY_TYPED) return false;
			for (int i=0; i<defaultFocusTraversalKeys.length; i++) {
			    if (i == id) continue;
			    Set traversalKeys = defaultFocusTraversalKeys[i];
			    if (traversalKeys != null) {
			        if (traversalKeys.contains(keystroke)) return false;
			    }
			}
		}
		return true;
	}
}
