package java.awt;
/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2006  All Rights Reserved
 */

import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.util.List;

/**
 * The only KeyboardFocusManager for Personal Profile applications.
 * In J2SE, others can be configured, but not in Personal Profile.
 * Focus traversal is done in response to a Component's focus traversal
 * keys, and using a Container's FocusTraversalPolicy.
 */

class KeyboardFocusManagerImpl extends KeyboardFocusManager {
	public KeyboardFocusManagerImpl() {
	    super();
	}

	/**
	 * This method is called by the AWT event dispatcher requesting that the
	 * current KeyboardFocusManager dispatch the specified event on its behalf.
	 * DefaultKeyboardFocusManagers dispatch all FocusEvents, all WindowEvents
	 * related to focus, and all KeyEvents. These events are dispatched based
	 * on the KeyboardFocusManager's notion of the focus owner and the focused
	 * and active Windows, sometimes overriding the source of the specified
	 * AWTEvent. If this method returns false, then the AWT event dispatcher will
	 * attempt to dispatch the event itself.
	 */
	public boolean dispatchEvent(AWTEvent awtEvent) {
		if (getDebug()) System.out.println("dispatch event " + awtEvent.toString());
		if (awtEvent instanceof FocusEvent) {
		   return dispatchFocusEvent((FocusEvent)awtEvent);
		} else if (awtEvent instanceof KeyEvent) {
			List dispatchers = getKeyEventDispatchers();
			for (int i=0; i < dispatchers.size(); i++) {
				KeyEventDispatcher dispatcher = (KeyEventDispatcher)dispatchers.get(i);
				boolean handled = dispatcher.dispatchKeyEvent((KeyEvent) awtEvent);
				if (handled) return true;
			}
			if (awtEvent.isConsumed()) return true;
			return dispatchKeyEvent((KeyEvent)awtEvent);
		} else if (awtEvent instanceof WindowEvent) {
		    return dispatchWindowEvent((WindowEvent)awtEvent);
		} else {
		    return false;
		}
	}
	private boolean dispatchFocusEvent(FocusEvent focusEvent) {
	    Component source = (Component)focusEvent.getSource();
	    int id = focusEvent.getID();
	    Window window = source.getWindow();
	    if (window == null) return true;
	    if (id == FocusEvent.FOCUS_GAINED) {
	    	if (!window.isFocusableWindow()) return true;
			setGlobalFocusOwner(source);
			// change was vetoed
			if (getGlobalFocusOwner() != source) return true;
			if (!focusEvent.isTemporary()) {
				setGlobalPermanentFocusOwner(source);
				// change was vetoed
				if (getGlobalPermanentFocusOwner() != source) return true;
			}
			window.setMostRecentFocusOwner(source);
			redispatchEvent(source, focusEvent);
			return true;
	    } else if (id == FocusEvent.FOCUS_LOST){
	    	// we dispatch focus lost events to ensure the correct
	    	// focus order according to the spec, so ignore focus lost
	    	// events if the source does not have focus
			if (getGlobalFocusOwner() != source) return true;
	        setGlobalFocusOwner(null);
			// change was vetoed
			if (getGlobalFocusOwner() != null) return true;
			redispatchEvent(source, focusEvent);
		    return true;
	    }
	    return false;
	}
	/**
	 * Called by dispatchEvent if no other KeyEventDispatcher in the dispatcher chain
	 * dispatched the KeyEvent, or if no other KeyEventDispatchers are registered. If
	 * the event has not been consumed, its target is enabled, and the focus owner is
	 * not null, this method dispatches the event to its target. This method will
	 * also subsequently dispatch the event to all registered KeyEventPostProcessors.
	 * After all this operations are finished, the event is passed to peers for processing.
	 * In all cases, this method returns true, since KeyboardFocusManagerImpl is
	 * designed so that neither dispatchEvent, nor the AWT event dispatcher, should
	 * take further action on the event in any situation.
	 */
	public boolean dispatchKeyEvent(KeyEvent keyEvent) {
		Component c = getGlobalFocusOwner();

		// return false since key events are post processed even if key
		// events are being ignored due to no global focus owner
		if (c == null) return false;

		if (c.getFocusTraversalKeysEnabled()) {
			// handle traversal
			AWTKeyStroke key = AWTKeyStroke.getAWTKeyStrokeForEvent(keyEvent);
			if (c.getFocusTraversalKeys(FORWARD_TRAVERSAL_KEYS).contains(key)) {
				focusNextComponent(c);
				return true;
			}
			if (c.getFocusTraversalKeys(BACKWARD_TRAVERSAL_KEYS).contains(key)) {
				focusPreviousComponent(c);
				return true;
			}
			if (c.getFocusTraversalKeys(UP_CYCLE_TRAVERSAL_KEYS).contains(key)) {
				upFocusCycle(c);
				return true;
			}

			if (c instanceof Container) {
				if (c.getFocusTraversalKeys(DOWN_CYCLE_TRAVERSAL_KEYS).contains(key)) {
					downFocusCycle((Container)c);
					return true;
				}
			}
		}

		// keyEvents must be retargeted to the focus owner, the peer layer
		// is unaware of any lightweight components and keyEvents will arrive
		// from the peer layer targeted to the focus owner's heavyweight container,
		// not the focus owner
		if (keyEvent.getSource() != c) keyEvent.setSource(c);
		redispatchEvent(c, keyEvent);
		// always return true
	    return true;
	}

	private boolean dispatchWindowEvent(WindowEvent windowEvent) {
	    // TODO PPro 1.1 Focus - Not sure if this code is complete

	    // Focus manager needs to dispatch the keyEvent based on the current
	    // state of focus owner, focused window, and active windows.  Sometimes
	    // it may be necessary to override the specified AWTEvent.
	    Window window = (Window)windowEvent.getSource();
	    WindowEvent event = null;
		switch (windowEvent.getID()) {
			case WindowEvent.WINDOW_CLOSED:
				// TODO Opposite field for window event that is created.
				// Dispatch events in order to ensure that the events occur in the specified
				// order.
				event = new WindowEvent(window, WindowEvent.WINDOW_DEACTIVATED, null);
				dispatchEvent(event);
				redispatchEvent(window, windowEvent);
				return true;
			case WindowEvent.WINDOW_ACTIVATED:
			    Window active = window;
				boolean redispatch = false;
				redispatch = (window instanceof Frame) || (window instanceof Dialog);
				if (redispatch) {
			    	setGlobalActiveWindow(active);
			    	// change was vetoed
			    	if (getGlobalActiveWindow() != active) return true;
			    	// redispatch the activated event
			    	redispatchEvent(active, windowEvent);
			    }
				if (window.isFocusableWindow() && window.isFocusable()) {
					event = new WindowEvent(active, WindowEvent.WINDOW_GAINED_FOCUS, windowEvent.getOppositeWindow());
					dispatchEvent(event);
				}
				return true;
			case WindowEvent.WINDOW_DEACTIVATED:
			    Window deactive = window;
				if (window.isFocusableWindow() && window.isFocusable()) {
				    event = new WindowEvent(window, WindowEvent.WINDOW_LOST_FOCUS, windowEvent.getOppositeWindow());
				    dispatchEvent(event);
				}
				redispatch = false;
				redispatch = (window instanceof Frame) || (window instanceof Dialog);
				if (redispatch) {
			    	redispatchEvent(deactive, windowEvent);
			    }
				return true;
			case WindowEvent.WINDOW_LOST_FOCUS:
				// dispatch the focusEvent so that events come in the proper order
				Component focus = window.getFocusOwner();
				// TODO set opposite field
				if (focus != null) {
					FocusEvent focusEvent = new FocusEvent(window.getFocusOwner(), FocusEvent.FOCUS_LOST, false, null);
					dispatchEvent(focusEvent);
				}
				redispatchEvent(window, windowEvent);
				return true;
			case WindowEvent.WINDOW_GAINED_FOCUS:
		    	if (!window.isFocusableWindow()) return true;
				setGlobalFocusedWindow(window);
				// change was vetoed
				if (getGlobalFocusedWindow() != window) return true;
				redispatchEvent(window, windowEvent);
				return true;
			default:
				return false;
		}
	}
	/**
	 * This method will be called by dispatchKeyEvent. It will handle any unconsumed
	 * KeyEvents that map to an AWT MenuShortcut by consuming the event and activating
	 * the shortcut.
	 */
	public boolean postProcessKeyEvent(KeyEvent keyEvent) {
		// key event has been handled by the focus owner, post process
		// the event
		List dispatchers = getKeyEventPostProcessors();
		for (int i=0; i < dispatchers.size(); i++) {
			KeyEventPostProcessor dispatcher = (KeyEventPostProcessor)dispatchers.get(i);
			boolean handled = dispatcher.postProcessKeyEvent(keyEvent);
			if (handled) return true;
		}
		// TODO PPro 1.1 Focus - do anything for post processing?
		return true;
	}
	/**
	 * This method initiates a focus traversal operation if and only if the KeyEvent
	 * represents a focus traversal key for the specified focusedComponent. It is
	 * expected that focusedComponent is the current focus owner, although this need
	 * not be the case. If it is not, focus traversal will nevertheless proceed as if
	 * focusedComponent were the focus owner.
	 * Parameters: cmpo - the Component that is the basis for a focus traversal operation
	 * if the specified event represents a focus traversal key for the Component,
	 * evt - the event that may represent a focus traversal key.
	 */
	public void processKeyEvent(Component cmpo, KeyEvent evt) {
	    // if there is no focus owner, discard the key events
	    if (getGlobalFocusOwner() == null) return;
		AWTKeyStroke key = AWTKeyStroke.getAWTKeyStrokeForEvent(evt);
		if (cmpo.isValidTraversalKey(FORWARD_TRAVERSAL_KEYS)) {
			if (cmpo.getFocusTraversalKeys(FORWARD_TRAVERSAL_KEYS).contains(key)) {
				focusNextComponent(cmpo);
			}
		}
		if (cmpo.isValidTraversalKey(BACKWARD_TRAVERSAL_KEYS)) {
			if (cmpo.getFocusTraversalKeys(BACKWARD_TRAVERSAL_KEYS).contains(key)) {
				focusPreviousComponent();
			}
		}
		if (cmpo.isValidTraversalKey(UP_CYCLE_TRAVERSAL_KEYS)) {
			if (cmpo.getFocusTraversalKeys(UP_CYCLE_TRAVERSAL_KEYS).contains(key)) {
				upFocusCycle();
			}
		}
		if (cmpo.isValidTraversalKey(DOWN_CYCLE_TRAVERSAL_KEYS)) {
			if (cmpo.getFocusTraversalKeys(DOWN_CYCLE_TRAVERSAL_KEYS).contains(key)) {
				downFocusCycle();
			}
		}
	}

	/**
	 * Focuses the Component before prevCmpo, typically based on a FocusTraversalPolicy.
	 * Parameters: prevCmpo - the Component that is the basis for the focus traversal operation
	 */
	public void focusPreviousComponent(Component prevCmpo) {
	    if (currentFocusCycleRoot != null) {
	    	FocusTraversalPolicy fp = currentFocusCycleRoot.getFocusTraversalPolicy();
	    	Component cmpo = fp.getComponentBefore(currentFocusCycleRoot, prevCmpo);
	    	if (cmpo == null) cmpo = fp.getDefaultComponent(currentFocusCycleRoot);
	    	if (cmpo != null) cmpo.requestFocusInWindow();
	    }
	}
	/**
	 * Focuses the Component after nextCmpo, typically based on a FocusTraversalPolicy.
	 * Parameters: nextCmpo - the Component that is the basis for the focus traversal operation
	 */
	public void focusNextComponent(Component nextCmpo) {
	    if (KeyboardFocusManager.getDebug()) System.out.print("focusNextComponent ");
	    Container cycleRoot = nextCmpo.getFocusCycleRootAncestor();
	    if (cycleRoot != null) {
		    FocusTraversalPolicy fp = cycleRoot.getFocusTraversalPolicy();
		    Component cmpo = fp.getComponentAfter(cycleRoot, nextCmpo);
		    if (cmpo == null) cmpo = fp.getDefaultComponent(cycleRoot);
			if (KeyboardFocusManager.getDebug()) System.out.println("focus next cmpo " + cmpo.getName());
			if (cmpo != null) cmpo.requestFocusInWindow();
		}
	}
	/**
	 * Moves the focus up one focus traversal cycle. Typically, the focus owner is set
	 * to upCmpo's focus cycle root, and the current focus cycle root is set to the
	 * new focus owner's focus cycle root. If, however, upCmpo's focus cycle root
	 * is a Window, then the focus owner is set to the focus cycle root's default
	 * Component to focus, and the current focus cycle root is unchanged.
	 * Parameters: upCmpo - the Component that is the basis for the focus traversal operation
	 */
	public void upFocusCycle(Component upCmpo) {
		Container newFocus = upCmpo.getFocusCycleRootAncestor();
		if (newFocus != null) newFocus.requestFocusInWindow();
		// TODO PPro1.1 Focus -  this logic doesn't match the jdk behavior,
		// upFocusCycle will set currentRootCycle to a frame so the
		// comment above seems wrong
//		if (root instanceof Window) {
//			Component focus = root.getFocusTraversalPolicy().getDefaultComponent(root);
//			setGlobalFocusOwner(focus);
//			focus.requestFocusInWindow();
//		} else {
//			setWindowFocus(root);
//		}
	}

	/**
	 * Moves the focus down one focus traversal cycle. If downContainer is a focus cycle root,
	 * then the focus owner is set to downContainer's default Component to focus, and
	 * the current focus cycle root is set to downContainer. If downContainer is not a focus
	 * cycle root, then no focus traversal operation occurs.
	 * Parameters: downContainer - the Container that is the basis for the focus traversal operation
	 */
	public void downFocusCycle(Container downContainer) {
		if (!downContainer.isFocusCycleRoot()) return;
		Component focus = downContainer.getFocusTraversalPolicy().getDefaultComponent(downContainer);
		if (focus != null) focus.requestFocusInWindow();
	}

}
