package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Label</code> is a graphical interface that displays static
 * single-lined text with no borders.
 * <p>
 * The text can be displayed in three position related to the <code>Label</code>
 * displayable area: aligned to the left border, centered or aligned to the right
 * border.
 *
 * @author		IBM
 * @version		initial
 */
public class Label extends Component {

	static final long serialVersionUID = 3094126758329070636L;

	// The text of the receiver
	String text;

	// The text alignment of the receiver
	int alignment;

	/**
	 * Indicates that the label should be left justified.
	 */
	public static final int LEFT = 0;

	/**
	 * Indicates that the label should be centered.
	 */
	public static final int CENTER = 1;

	/**
	 * Indicates that the label should be right justified.
	 */
	public static final int RIGHT = 2;

	/**
	 * Constructs a new <code>Label</code> with a <code>null</code> text.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Label()
		throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Label</code> with the specified text.
	 * The default alignment is <code>LEFT</code>.
	 *
	 * @param	text	the text displayed by the receiver
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Label(String text)
		throws HeadlessException
	{
		this(text, LEFT);
	}

	/**
	 * Constructs a new <code>Label</code> with the specified text
	 * and alignment.
	 *
	 * @param	text		the text displayed by the receiver
	 * @param	alignment	the alignment of the text
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Label(String text, int alignment)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		this.text = text;
		switch (alignment) {
			case LEFT:
			case CENTER:
			case RIGHT:
				break;
			default:
				throw new IllegalArgumentException();
		}
		this.alignment = alignment;
	}

	int _createPeer(Container hwParent) {
		return OS.Label_New(hwParent.peer, OS.STYLE_MIDDLE);
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	void _setupPeer() {
		OS.Label_SetText(peer, text == null ? "" : text); //$NON-NLS-1$
		OS.Label_SetAlignment(peer, getNativeAlignment());

		super._setupPeer();
	}

	private int getNativeAlignment() {
		switch (alignment) {
			case Label.LEFT:	return OS.STYLE_LEAD;
			case Label.CENTER:	return OS.STYLE_CENTER;
			case Label.RIGHT:	return OS.STYLE_TRAIL;
			default: return OS.STYLE_LEAD;
		}
	}

	public boolean isLightweight() {
		return false;
	}

	public boolean isOpaque() {
		return true;
	}

	/**
	 * Answers the receiver's text alignment. The return value is one of:
	 * <code>LEFT</code>
	 * <code>CENTER</code>
	 * <code>RIGHT</code>
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is text alignment for the receiver
	 */
	public int getAlignment() {
		return alignment;
	}

	/**
	 * Set the receiver's text alignment. The specified value has to
	 * be one of:
	 * <code>LEFT</code>
	 * <code>CENTER</code>
	 * <code>RIGHT</code>
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	alignment	the text alignment for the receiver
	 *
	 * @throws	IllegalArgumentException	if the specified alignment is invalid
	 */
	public void setAlignment(int alignment) {
		switch (alignment) {
			case LEFT:
			case CENTER:
			case RIGHT:
				break;
			default:
				throw new IllegalArgumentException();
		}

//		synchronized (this) {
			if (this.alignment == alignment) {
				return;
			}
			this.alignment = alignment;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.Label_SetAlignment(peer, getNativeAlignment());
				}
			} finally {
				releasePeerReadLock();
			}
//		}
	}

	Dimension getMinimumSizeImpl() {
		/*
		 * Without this buffer, we fit the label around the text almost perfectly.
		 * The trouble is that some apps (like the TCK, for example) assume that
		 * there's a little bit of buffer room to work with.
		 * Specifically, without this buffer room, the count-down timer label on the
		 * TCK doesn't show up right until you force a layout.
		 */
		Dimension dim = super.getMinimumSizeImpl();
		if (peer != 0) {
			dim.width += 14;
			dim.height += 8;
		}

		return dim;
	}

	/**
	 * Answer the <code>String</code> that is the text
	 * displayed by the receiver.  Answer <code>null</code> if
	 * no text is specified.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> which is text for the receiver
	 */
	public String getText() {
		return text;
	}

	/**
	 * Set the <code>String</code> that is the text
	 * diplayed by the receiver.  Use <code>null</code> if
	 * no label is desired.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text to be displayed by the receiver
	 */
	public void setText(String text) {
//		synchronized (this) {
			if (this.text == text) return;
			if ((text != null) && (text.equals(this.text))) return;

			this.text = text;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.Label_SetText(peer, text == null ? "" : text); //$NON-NLS-1$
				}
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() +
			",text=" + getText() + //$NON-NLS-1$
			",alignment=" + getAlignment(); //$NON-NLS-1$
	}

	String classNonlocalizedName() {
		return "label"; //$NON-NLS-1$
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException, HeadlessException {
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}
}
