package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.io.*;
import java.util.EventListener;
import java.util.Vector;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>List</code> is a graphical interface that displays a list
 * of selectable items.
 * <p>
 * Items in the list can be either selected or deselected by clicking
 * on them.
 * <p>
 * Each time the user selects an item, the <code>List</code> will
 * fire a item event and each time the user double-clicks on an item
 * in the list, an action will be fired.
 *
 * @author		IBM
 * @version		initial
 */
public class List extends Component implements ItemSelectable {

	static final long serialVersionUID = -3304312411574666869L;
	private static final int DefaultMinRows = 4;

	// The items of the receiver
	Vector items;

	// The position of the selected items
	int selected[];

	// Determines wheter the receiver allows multiple selection
	boolean multipleMode;

	// The number of rows of the receiver
	int rows;

	// The visible index of the receiver
	int visibleIndex;

	// The item listener of the receiver
	transient ItemListener itemListener;

	// The action listener of the receiver
	transient ActionListener actionListener;

	int listSerializedDataVersion;

	/**
	 * Constructs a new <code>List</code> with four visible items and
	 * with single selection mode.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public List()
		throws HeadlessException
	{
		this(0);
	}

	/**
	 * Constructs a new <code>List</code> with the specified number of
	 * visible items and single selection mode.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rows	the number of visible items
	 */
	public List(int rows)
		throws HeadlessException
	{
		this((rows == 0) ? 4 : rows, false);
	}

	/**
	 * Constructs a new <code>List</code> with the specified number of
	 * visible items and the specified selection mode.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rows			the number of visible items
	 * @param	multipleMode	<code>true</code> if the receiver is to be in multiple
	 * 							selection mode, <code>false</code> for single selection mode
	 */
	public List(int rows, boolean multipleMode)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		this.rows = ((rows == 0) ? 4 : rows);
		this.multipleMode = multipleMode;
		this.visibleIndex = -1;
		this.items = new Vector();
	}

	int _createPeer(Container hwParent) {
		return OS.List_New(hwParent.peer, widgetStyle());
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	void _setupPeer() {
		int size = items.size();
		for (int i = 0; i < size; i++) {
			OS.List_Add(peer, (String) items.elementAt(i), i);
		}
		if (selected != null) {
			OS.List_SelectIndices(peer, selected, selected.length);
			selected = null;
		}
		// initiate peer listbox multiplemode
		OS.List_SetMultipleMode(peer, multipleMode);

		super._setupPeer();
	}

	/*
	 * Handle two event types.
	 */
	private void listCallback(int index, int type) {
		AWTEvent event;
		switch (type) {

			case OS.EVENT_LIST_SELECTION:
				event = new ItemEvent(
						this, ItemEvent.ITEM_STATE_CHANGED,
						new Integer(index), this.isSelected(index) ? ItemEvent.SELECTED : ItemEvent.DESELECTED);
				break;

			case OS.EVENT_LIST_ACTION:
				event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, getItem(index),
						System.currentTimeMillis(),
					0);
				break;
			default: return;
		}
		postEvent(event);
	}

	/**
	 * Destroy the underlying widget for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 */
 	public void removeNotify() {
		synchronized (getTreeLock()) {
			if (peer == 0) return;

			if (OS.List_GetSelectionCount(peer) > 0)
				selected = OS.List_GetSelectionIndices(peer);
			super.removeNotify();
		}
	}

	int widgetStyle() {
		return OS.STYLE_BORDER | OS.STYLE_H_SCROLL | OS.STYLE_V_SCROLL;
	}

	public boolean isLightweight() {
		return false;
	}

	public boolean isOpaque() {
		return true;
	}

	boolean isFocusTraversableImpl() {
		return true;
	}

	/**
	 * Answers the number of items in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the number of items in the receiver
	 */
	public int getItemCount() {
		return items.size();
	}

	/**
	 * Answers the number of items in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the number of items in the receiver
	 *
	 * @deprecated	use <code>getItemCount()</code> instead
	 */
	public int countItems() {
		return getItemCount();
	}

	/**
	 * Answers the item at the specified position.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the index of the desired item
	 *
	 * @return	an <code>String</code> which is the item at the given position
	 */
	public String getItem(int position) {
		if (position < 0 || position > getItemCount())
			throw new ArrayIndexOutOfBoundsException();

		return (String)items.elementAt(position);
	}

	/**
	 * Answers the items in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String[]</code> which cotains the receiver's items
	 */
	public String[] getItems() {
		String result[] = new String[items.size()];
		items.copyInto(result);
		return result;
	}

	/**
	 * Adds the specified item to the end of the receiver's list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item	the item to be added
	 */
	public void add(String item) {
		items.addElement(item == null ? "" : item); //$NON-NLS-1$

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.List_Add(peer, item == null ? "" : item, items.size() - 1); //$NON-NLS-1$
			}
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	/**
	 * Adds the specified item to the end of the receiver's list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item		the item to be added
	 *
	 * @deprecated	use <code>add(String)</code> instead
	 */
	public void addItem(String item) {
		add(item);
	}

	/**
	 * Adds the specified item at the specified position of the receiver's
	 * list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item		the item to be added
	 * @param	position	the position of the new item
	 *
	 * @throws	IllegalArgumentException	if the specified position is less than <code>-1</code>
	 */
	public void add(String item, int position) {
		if (position < 0 || position > getItemCount())
			position = getItemCount();

		items.insertElementAt(item == null ? "" : item, position); //$NON-NLS-1$

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.List_Add(peer, item == null ? "" : item, position); //$NON-NLS-1$
			}
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	/**
	 * Adds the specified item at the specified position of the receiver's
	 * list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	item		the item to be added
	 * @param	position	the position of the new item
	 *
	 * @deprecated	use <code>add(String, int)</code> instead
	 */
	public void addItem(String item, int position) {
		add(item, position);
	}

	/**
	 * Replaces the item at the specified position with the
	 * specified item.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	newItem	the new item
	 *
	 * @throws	ArrayIndexOutOfBoundsException if the position is smaller
	 *			than 0 or greater than the number of items in the receiver
	 */
	public void replaceItem(String newItem, int position) {
		if (position < 0 || position >= getItemCount())
			throw new ArrayIndexOutOfBoundsException();

		if (newItem == null) newItem = "";

		items.setElementAt(newItem, position);

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.List_SetItem(peer, position, newItem);
			}
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	/**
	 * Removes all items from the receiver's list.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void removeAll() {
		items = new Vector();

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.List_RemoveAll(peer);
			}
		} finally {
			releasePeerReadLock();
		}
		if (isValid()) invalidate();
	}

	String classNonlocalizedName() {
		return "list"; //$NON-NLS-1$
	}

	/**
	 * Removes all items from the receiver's list.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @deprecated	use <code>removeAll()</code> instead
	 *
	 */
	public void clear() {
		removeAll();
	}

	/**
	 * Removes the first item in the receiver's list that is
	 * equal to the specified item.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	item	the item to be removed
	 *
	 * @throws	IllegalArgumentException	if the specified item does not exists in the list
	 */
	public synchronized void remove(String item) {
		int position = items.indexOf(item);

		if (position == -1)
			throw new IllegalArgumentException();

		remove(position);
	}

	/**
	 * Removes the item from the receiver's list at the specified
	 * position.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	position	the index of the item to be removed
	 *
	 * @throws	ArrayIndexOutOfBoundsException if the position is smaller
	 *			than 0 or greater than the number of items in the receiver
	 */
	public void remove(int position) {
//		synchronized (this) {
			if (position < 0 || position >= getItemCount())
				throw new ArrayIndexOutOfBoundsException();

			items.removeElementAt(position);

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.List_Remove(peer, position);
				}
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	/**
	 * Removes the item from the receiver's list at the specified
	 * position.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the index of the item to be removed
	 *
	 * @deprecated	use <code>remove(int)</code> instead
	 */
	public void delItem(int position) {
		remove(position);
	}

	/**
	 * Answers the position of the selected item for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the position of the selected item or
	 *			<code>-1</code> if there is not a selected item or if there is
	 *			more than one selected item.
	 */
	public int getSelectedIndex() {
//		synchronized (this) {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				if (selected == null) return -1;
				if (isMultipleMode() && selected.length != 1) return -1;
				return selected[0];
			}

			if (isMultipleMode()) {
				return (OS.List_GetSelectionCount(peer) == 1) ? OS.List_GetSelectionIndices(peer)[0] : -1;
			}
			return OS.List_GetSelectionIndex(peer);
		} finally {
			releasePeerReadLock();
		}
//		}
	}

	/**
	 * Answers the positions of the selected items for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	an <code>int[]</code> which contains the position of each
	 *			selected item.
	 */
	public int[] getSelectedIndexes() {
//		synchronized (this) {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				if (selected == null) return new int[0];
				return (int[])selected.clone();
			}

			return OS.List_GetSelectionIndices(peer);
		} finally {
			releasePeerReadLock();
		}
//		}
	}

	/**
	 * Answers the selected item for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is the selected item or <code>null</code>
	 *			if there is not a selected item
	 */
	public synchronized String getSelectedItem() {
		int position = getSelectedIndex();

		if (position < 0) return null;

		return getItem(position);
	}

	/**
	 * Answers the selected items for the receiver.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @return	a non-null <code>String[]</code> which contains the selected items
	 */
	public synchronized String[] getSelectedItems() {
		int[] selectedIndices = getSelectedIndexes();
		int size = selectedIndices.length;
		String selectedItems[] = new String[size];
		for (int i = 0; i < size; i++) {
			selectedItems[i] = (String)items.elementAt(selectedIndices[i]);
		}
		return selectedItems;
	}

	/**
	 * Answers the selected items for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a non-null <code>Object[]</code> which contains the selected items
	 */
	public Object[] getSelectedObjects() {
		return getSelectedItems();
	}

	/**
	 * Selects a item.  If the item at the specified position is
	 * deselected, it is selected.  If the item at the specified
	 * position is selected, it remains selected.  Indices that
	 * are out of range are ignored.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	position	the position of the item to be selected
	 */
	public synchronized void select(int position) {
		if ((position >= countItems()) || (position < 0)) {
			return;
		}

		acquirePeerReadLock();
		try {
			if (peer == 0) {
				if (selected == null) {
					selected = new int[1];
					selected[0] = position;
				} else {
					if (isMultipleMode()) {
						if (isSelected(position)) return;
						int size = selected.length;
						int newSelected[] = new int[size + 1];
						System.arraycopy(selected, 0, newSelected, 0, size);
						newSelected[size] = position;
						selected = newSelected;
					} else {
						selected[0] = position;
					}
				}
				return;
			}

			OS.List_Select(peer, position);
			/*
			 * This isn't quite right, but it's a short-term improvement.
			 * If we simply call select(), then the list item is not scrolled
			 * into view (if it needs to be). By calling setTopItem, we make
			 * sure that the newly selected item is made visible. However, if the
			 * item we're selecting is already scrolled into view, we will set it
			 * as the top item in the list, which we don't really want to do.
			 * A potentially better way to do this would be to add a parameter to
			 * List_Select() that tells whether we should scroll the current item into
			 * view (we don't want to on eSWT), and if it's true, take advantage
			 * of the native API that we call to select the item, which, at least
			 * on Windows, does it for us
			 * (in other words, don't reset the top index, like we currently do)
			 */
			OS.List_SetTopItemIndex(peer, position);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Deselects a item.  If the item at the specified position is
	 * selected, it is deselected.  If the item at the specified
	 * position is not selected, it remains deselected.  Indices that
	 * are out of range are ignored.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	position	the position of the item to be deselected
	 */
	public synchronized void deselect(int position) {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				if (selected == null) return;

				if (isMultipleMode()) {
					int size = selected.length;
					for (int i = size-1; i>=0; i--) {
						if (selected[i] == position) {
							int newSelected[] = new int[size - 1];
							System.arraycopy(selected, 0, newSelected, 0, i);
							System.arraycopy(selected, i + 1, newSelected, i, size - i - 1);
							selected = newSelected;
							break;
						}
					}
				} else {
					if (selected[0] == position)
						selected = null;
				}
				return;
			}
			// According to the spec, if the index is out of range, or if the
			// item at that index is not selected, then the request is ignored.
			if (OS.List_IsSelected(peer, position)) {
				OS.List_Deselect(peer, position);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer whether the item at the specified position is selected.
	 *
	 * @author      IBM
	 * @version	initial
	 *
	 * @param	position	the position of the item to be checked
	 *
	 * @return	<code>true</code> if the item is selected,
	 *			<code>false</code> otherwise
	 */
	public boolean isIndexSelected(int position) {
		int[] selectedIndexes = getSelectedIndexes();
		int size = selectedIndexes.length;
		for (int i = 0; i < size; i++) {
			if (selectedIndexes[i] == position)
				return true;
		}
		return false;
	}

	/**
	 * Answer whether the item at the specified position is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	position	the position of the item to be checked
	 *
	 * @return	<code>true</code> if the item is selected,
	 *			<code>false</code> otherwise
	 *
	 * @deprecated	use <code>isIndexSelected(int)</code> instead
	 */
	public boolean isSelected(int position) {
		return isIndexSelected(position);
	}

	/**
	 * Answer the number of visible items for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>int</code> which is the number of visible items
	 */
	public int getRows() {
		return rows;
	}

	/**
	 * Answer whether the receiver allows multiple selected items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	<code>true</code> if the receiver allows multiple selection,
	 *			<code>false</code> otherwise
	 */
	public boolean isMultipleMode() {
		return multipleMode;
	}

	/**
	 * Answer whether the receiver allows multiple selected items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		<code>true</code> if the receiver allows multiple selection,
	 *				<code>false</code> otherwise
	 *
	 * @deprecated	use <code>isMultipleMode()</code> instead
	 */
	public boolean allowsMultipleSelections() {
		return isMultipleMode();
	}

	/**
	 * Sets whether the receiver allows multiple selected items.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		multipleMode	<code>true</code> to allow multiple selected items,
	 *								<code>false</code> otherwise
	 */
	public void setMultipleMode(boolean multipleMode) {
		if (this.multipleMode == multipleMode)
			return;

		this.multipleMode = multipleMode;

		acquirePeerReadLock();
		try {
			if (peer != 0){
				OS.List_SetMultipleMode(peer, multipleMode);
			}
		} finally {
			releasePeerReadLock();
		}
		// some platforms may use checkboxes for multiselect lists, so
		// invalidate the List since its size may change
		if (isValid()) invalidate();
	}

	/**
	 * Sets whether the receiver allows multiple selected items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		multipleMode	<code>true</code> to allow multiple selected items,
	 *								<code>false</code> otherwise
	 *
	 * @deprecated	use <code>setMultipleMode(boolean)</code) instead
	 */
	public void setMultipleSelections(boolean multipleSelections) {
		setMultipleMode(multipleSelections);
	}

	/**
	 * Answers the position of the item that was last used when
	 * calling <code>makeVisible(int)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the last value passed to
	 *			<code>makeVisible(int)</code> or <code>-1</code> if the
	 *			was never called
	 */
	public int getVisibleIndex() {
		return visibleIndex;
	}

	/**
	 * Scrolls the receiver in order to ensure that the item at the
	 * specified position is displayed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	position	the position of the item to be displayed
	 */
	public void makeVisible(int position) {
		// Caching the value of the visible index before it is checked
		// for being a valid index seems wrong, since calls to getVisibleIndex()
		// will return values that are not valid positions but the Sun JRE has
		// this same behavior so it should be left alone.
		visibleIndex = position;
		if (position < 0 || position >= getItemCount())
			return;

		acquirePeerReadLock();
		try {
			if (peer != 0)
				OS.List_SetTopItemIndex(peer, position);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of items.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	public Dimension getPreferredSize(int rows) {
		return getMinimumSize(rows);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rows	the number of visible items
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 *
	 * @deprecated	use <code>getPreferredSize(int)</code> instead
	 */
	public Dimension preferredSize(int rows) {
		return getPreferredSize(rows);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	Dimension getPreferredSizeImpl() {
		int rows = getRows();
		if (rows <= 0) rows = DefaultMinRows;
		return getPreferredSize(rows);
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of visible items.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	public Dimension getMinimumSize(int rows) {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				return getSize();
			}

			int[] point = OS.List_GetMinimumSize(peer,rows);
			return new Dimension(point[OS.INDEX_X],point[OS.INDEX_Y]);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of visible items.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rows	the number of visible items
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 *
	 * @deprecated	use <code>getMinimumSize()</code> instead
	 */
	public Dimension minimumSize(int rows) {
		return getMinimumSize(rows);
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
 	 * size of the receiver.
 	 *
 	 * @author      IBM
 	 * @version		initial
	 *
 	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	Dimension getMinimumSizeImpl() {
		int rows = getRows();
		if (rows <= 0) rows = DefaultMinRows;
		return getMinimumSize(rows);
	}

	/**
	 * Adds a <code>ItemListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be added.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void addItemListener(ItemListener itemListener) {
		addEventListener(itemListener);
		this.itemListener = AWTEventMulticaster.add(this.itemListener, itemListener);
	}

	/**
	 * Removes an <code>ItemListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		itemListener	the <code>ItemListener</code> to be removed.
	 *
	 * @see			java.awt.event.ItemListener
	 * @see			java.awt.event.ItemEvent
	 */
	public synchronized void removeItemListener(ItemListener itemListener) {
		this.itemListener = AWTEventMulticaster.remove(this.itemListener, itemListener);
	}

	/**
	 * Adds an <code>ActionListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be added.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void addActionListener(ActionListener actionListener) {
		addEventListener(actionListener);
		this.actionListener = AWTEventMulticaster.add(this.actionListener, actionListener);
	}

	/**
	 * Removes an <code>ActionListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be removed.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void removeActionListener(ActionListener actionListener) {
		this.actionListener = AWTEventMulticaster.remove(this.actionListener, actionListener);
	}

	/**
	 * Called when an <code>ActionEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processActionEvent(ActionEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	actionEvent	the event to be processed
	 *
	 * @see		ActionEvent
	 */
	protected void processActionEvent(ActionEvent actionEvent) {
		if (actionListener == null)
			return;

		if ((actionEvent == null) || (actionEvent.getID() == ActionEvent.ACTION_PERFORMED)) {
			actionListener.actionPerformed(actionEvent);
		}
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default events handled are <code>ActionEvent</code> and <code>ItemEvent</code>,
	 * the receiver will forward calls to the default event by calling
	 * <code>processActionEvent(ActionEvent)</code>.
	 * <code>processItemEvent(ItemEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ItemEvent) {
			processItemEvent((ItemEvent)awtEvent);
		} else if (awtEvent instanceof ActionEvent) {
			processActionEvent((ActionEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ItemEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processItemEvent(ItemEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	itemEvent	the event to be processed
	 *
	 * @see		ItemEvent
	 */
	protected void processItemEvent(ItemEvent itemEvent) {
		if (itemListener == null)
			return;

		if ((itemEvent == null) || (itemEvent.getID() == ItemEvent.ITEM_STATE_CHANGED)) {
			itemListener.itemStateChanged(itemEvent);
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() +
			",selected=" + getSelectedItem(); //$NON-NLS-1$
	}

	/**
	 * Removes items in the specified range from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	from	the position of the first item to be removed
	 * @param	to		the position of the last item to be removed
	 *
	 * @deprecated	no replacement
	 */
	public synchronized void delItems(final int from, final int to) {
		if ((from < 0) || (to < 0) || (from >= items.size()) || (to >= items.size())) {
			throw new ArrayIndexOutOfBoundsException();
		}

		for (int i = from; i <= to; i++){
			remove(from);
		}
	}

/*
 	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ItemEvent)
			return ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) || itemListener != null;
		else if (awtEvent instanceof ActionEvent)
			return ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) || actionListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	boolean isFocusTraversableImpl() {
		return true;
	}
*/

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, itemListenerK, itemListener);
		AWTEventMulticaster.save(stream, actionListenerK, actionListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
		throws IOException, ClassNotFoundException
		, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();

		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(itemListenerK))
				addItemListener((ItemListener)listener);
			else if (key.equals(actionListenerK))
				addActionListener((ActionListener)listener);
		}
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ActionEvent)
			return ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) || actionListener != null;
		if (awtEvent instanceof ItemEvent)
			return ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) || itemListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	public ItemListener[] getItemListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(itemListener, ItemListener.class);
		ItemListener[] resultArray = new ItemListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	public ActionListener[] getActionListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(actionListener, ActionListener.class);
		ActionListener[] resultArray = new ActionListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}
}
