package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.*;

import com.ibm.oti.awt.image.*;

/**
 * The <code>MediaTracker</code> class is used to track the
 * fetching of one or more media objects.  IDs can be assigned
 * to these objects so that the media tracker can be queried
 * by ID for their status.  Also, groups of media objects can be tracked
 * by assigning the same ID to more than one object.
 * <p>
 * Status is kept by a media tracker in one of the following states:
 * 		ABORTED
 *		COMPLETE
 *		ERRORED
 *		LOADING
 * <p>
 *
 * @author      IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */

public class MediaTracker implements Serializable {

	static final long serialVersionUID = -483174189758638095L;

	// static variables representing state of media fetching
	public static final int ABORTED = 2;
	public static final int COMPLETE = 8;
	public static final int ERRORED = 4;
	public static final int LOADING = 1;

	// first item in the list of objects being tracked
	transient
	MediaEntry head;

	// target Component which media object is fetched for
	Component target;

	/**
	 * Constructs a new <code>MediaTracker</code> to
	 * track media object loading for the specified
	 * <code>Component</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	component	the component to track media
	 *						object loading
	 *
	 */
	public MediaTracker(Component component) {
		target = component;
	}

	/**
	 * Add an image to the list of objects whose loading are
	 * being tracked.  Use the specified id when answering status
	 * of the image provided.  The size of the image will be in its
	 * default.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	image	the image being loaded
	 * @param	id		the integer ID of objects tracked
	 *
	 */
	public void addImage(Image image, int id) {
		if (image == null) return;
		add(new MediaEntry(image, id));
	}

	//	Add a <code>MediaEntry</code> into the
	//	linked list of entries.
	 private synchronized void add(MediaEntry me) {
		 if (head == null) {
			 head = me;
		 } else {
			 MediaEntry end = head;
			 while (end.next != null)
				 end = end.next;

			 end.next = me;
			 me.previous = end;
		 }
	 }

	/**
	 * Add an image to the list of objects whose loading are
	 * being tracked.  Use the specified id when answering status
	 * of the image provided.  The size of the image will be
	 * the width and height specified.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	image	the image being loaded
	 * @param	id		the integer ID of objects tracked
	 * @param	width	the width of the image
	 * @param	height	the height of the image
	 *
	 */
	public void addImage(Image image, int id, int width, int height) {
		add(new MediaEntry(image, id, width, height));
	}

	/**
	 * Answer whether all media objects being tracked by
	 * the receiver are no longer loading.  Send <code>isErrorAny</code>
	 * or <code>isErrorID</code> to check for errors in loading.

	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	a <code>boolean</code> which is true if
	 *			all images are loaded/aborted/ or have hit errors while
	 *			loading, answer false otherwise
	 */
	public synchronized boolean checkAll() {
		MediaEntry me = head;

		while (me != null) {
			if (me.getStatus() <= LOADING)
				return false;
			me = me.next;
		}

		return true;
	}

	/**
	 * Answer whether all media objects being tracked by
	 * the receiver are no longer loading.  If <code>startLoad</code>
	 * is true then start loading objects which have not
	 * yet started loading.  Send <code>isErrorAny</code>
	 * or <code>isErrorID</code> to check for errors in loading.

	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	startLoad	a boolean which is true if loading
	 *						should be initiated on objects that
	 *						are not yet loading, false otherwise
	 *
	 * @return	a <code>boolean</code> which is true if
	 *			all images are loaded/aborted/ or have hit errors while
	 *			loading, answer false otherwise
	 */
	public boolean checkAll(boolean startLoad) {
		if (startLoad) startLoading();
		return checkAll();
	}

	/**
	 * Answer a <code>boolean</code> indicating whether
	 * any errors have occured while loading the media
	 * objects tracked by the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	a <code>boolean</code> which is true if
	 *			any errors have been hit, answer false
	 *			otherwise
	 *
	 */
	public synchronized boolean isErrorAny() {
		return (getErrorsAny() != null);
	}

	/**
	 * Answer an <code>Array</code> of media objects
	 * that have encountered errors while loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	an <code>Array</code> of media objects that have
	 * 			encountered errors while loading
	 *			answer <code>null</code> if none
	 *
	 */
	public synchronized Object[] getErrorsAny() {
		MediaEntry me = head;
		Vector v = new Vector();

		while (me != null) {
			if (0 != (me.getStatus() & ERRORED)) v.addElement(me.image);
			me = me.next;
		}

		if (0 == v.size()) return null;

		Object[] result = new Object[v.size()];
		v.copyInto(result);
		return result;
	}

	/**
	 * Start loading all media objects tracked by the receiver
	 * and return when all have been loaded or an error occurs
	 * while loading or scaling.
	 * Send <code>isErrorAny</code> or <code>isErrorID</code>
	 * to check for errors in loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @throws	<code>InterruptedException</code> if this thread
	 * 			was interrupted by another thread.
	 */
	public void waitForAll() throws InterruptedException {
		while (0 != (LOADING & statusAll(true))) {
			Thread.sleep(250);
		}
	}

	/**
	* Start loading all media objects tracked by the receiver
	* and return when all have been loaded or an error occurs
	* while loading or scaling or the time specified
	* by <code>timeMillis</code> has expired.
	* Send <code>isErrorAny</code> or <code>isErrorID</code>
	* to check for errors in loading.
	*
	* @author      IBM
	* @version 	initial
	*
	* @param	timeMillis	the time to wait for objects to load
	*						before returning.
	*
	* @return	a <code>boolean</code> which is true if loading
	*			was successful and false otherwise.
	*
	* @throws	<code>InterruptedException</code> if this thread
	* 			was interrupted by another thread.
	*/
   public boolean waitForAll(long timeMillis) throws InterruptedException {
		if (0 == timeMillis) {
			waitForAll();
			return true;
		}

		while ((timeMillis > 0) && (0 != (LOADING & statusAll(true)))) {
			Thread.sleep(250);
			timeMillis -= 250;
		}

		return statusAll(false) == COMPLETE;
   }

	/**
	 * Answer the bitwise OR of the loading state of all
	 * media objects being tracked by the receiver.
	 * <p>
	 * The possible loading states for a media object are:
	 *	ABORTED, COMPLETE, ERRORED and LOADING.  A media object
	 * can also have a status of 0 if loading has not yet
	 * started.
	 * <p>
	 * If <code>startLoad</code> is true then start loading
	 * objects which have not yet started loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	startLoad	a boolean which is true if loading
	 *						should be initiated on objects that
	 *						are not yet loading, false otherwise
	 *
	 * @return	an <code>int</code> which the bitwise OR of the
	 *			loading state for all media objects tracked
	 *
	 */
	public synchronized int statusAll(boolean startLoad) {
		if (startLoad) startLoading();

		if (head == null)
			return 0;
		int status;
		int totalStatus = 0;

		MediaEntry end = head;
		while (end != null) {
			status = end.getStatus();
			if (status == -1) {
				status = LOADING;
			}
			totalStatus = totalStatus | status;
			end = end.next;
		}
		return totalStatus;
	}

	/**
	* Answer whether all media objects being tracked by
	* the specified ID are no longer loading.
	* Send <code>isErrorAny</code>
	* or <code>isErrorID</code> to check for errors in loading.
	*
	* @author      IBM
	* @version 	initial
	*
	* @param	id		the integer ID of objects tracked
	*
	* @return	a <code>boolean</code> which is true if
	*			all images tagged with the specified ID
	*			are loaded/aborted/ or have hit errors while
	*			loading, answer false otherwise
	*/
	public synchronized boolean checkID(int id) {
		MediaEntry me = head;

		while (me != null) {
			if (me.id == id && me.getStatus() <= LOADING)
				return false;
			me = me.next;
		}

		return true;
	}

	/**
	* Answer whether all media objects being tracked by
	* the specified ID are no longer loading.  If <code>startLoad</code>
	* is true then start loading objects which have not
	* yet started loading.  Send <code>isErrorAny</code>
	* or <code>isErrorID</code> to check for errors in loading.
	*
	* @author      IBM
	* @version 	initial
	*
	* @param	id		the integer ID of objects tracked
	* @param	startLoad	a boolean which is true if loading
	*						should be initiated on objects that
	*						are not yet loading, false otherwise
	*
	* @return	a <code>boolean</code> which is true if
	*			all images tagged with the specified ID
	*			are loaded/aborted/ or have hit errors while
	*			loading, answer false otherwise
	*
	*/
	public boolean checkID(int id, boolean startLoad) {
		if (startLoad)
			startLoading(id);
		return checkID(id);
	}

	/**
	* Answer a <code>boolean</code> indicating whether
	* any errors have occured while loading the media
	* objects tracked by the receiver for the specified ID.
	*
	* @author      IBM
	* @version 	initial
	*
	* @param	id		the integer ID of objects tracked
	*
	* @return	a <code>boolean</code> which is true if
	*			any errors have been hit, answer false
	*			otherwise
	*
	*/
	public synchronized boolean isErrorID(int id) {
		return (getErrorsID(id) != null);
	}

	/**
	 * Answer an <code>Array</code> of media objects tagged
	 * with the specified ID, that have encountered errors
	 * while loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	id		the integer ID of objects tracked
	 *
	 * @return	an <code>Array</code> of media objects that have
	 * 			encountered errors while loading
	 *			answer <code>null</code> if none
	 *
	 */
	public synchronized Object[] getErrorsID(int id) {
		MediaEntry me = head;
		Vector v = new Vector();

		while (me != null) {
			if ((me.id == id) && (0 != (me.getStatus() & ERRORED))) v.addElement(me.image);
			me = me.next;
		}

		if (0 == v.size()) return null;

		Object[] result = new Object[v.size()];
		v.copyInto(result);
		return result;
	}

	/**
	 * Start loading all media objects tracked by the receiver
	 * tagged with the specified ID, and return when all have
	 * been loaded or an error occurs while loading or scaling.
	 * Send <code>isErrorAny</code> or <code>isErrorID</code>
	 * to check for errors in loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	id		the integer ID of objects tracked
	 *
	 * @throws	<code>InterruptedException</code> if this thread
	 * 			was interrupted by another thread.
	 */
	public void waitForID(int id) throws InterruptedException {
		while (0 != (LOADING & statusID(id, true))) {
			Thread.sleep(250);
		}
	}

	/**
	 * Start loading all media objects tracked by the receiver
	 * specified by ID, return when all have been loaded or an error occurs
	 * while loading or scaling or the time specified
	 * by <code>timeMillis</code> has expired.
	 * Send <code>isErrorAny</code> or <code>isErrorID</code>
	 * to check for errors in loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	id		the integer ID of objects tracked
	 * @param	timeMillis	the time to wait for objects to load
	 *						before returning.
	 *
	 * @return	a <code>boolean</code> which is true if loading
	 *			was successful and false otherwise.
	 *
	 * @throws	<code>InterruptedException</code> if this thread
	 * 			was interrupted by another thread.
	 */
	public boolean waitForID(int id, long timeMillis) throws InterruptedException {
		if (0 == timeMillis) {
			waitForID(id);
			return true;
		}

		while ((timeMillis > 0) && (0 != (LOADING & statusID(id, true)))) {
			Thread.sleep(250);
			timeMillis -= 250;
		}

		return statusID(id, false) == COMPLETE;
	}

	/**
	 * Answer whether the bitwise OR of the loading state of all
	 * media objects specified by ID, being tracked by the receiver.
	 * <p>
	 * The possible loading states for a media object are:
	 *	ABORTED, COMPLETE,ERRORED and LOADING.
	 * <p>
	 * If <code>startLoad</code> is true then start loading
	 * objects which have not yet started loading.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	id		the integer ID of objects tracked
	 * @param	startLoad	a boolean which is true if loading
	 *						should be initiated on objects that
	 *						are not yet loading, false otherwise
	 *
	 * @return	an <code>int</code> which the bitwise OR of the
	 *			loading state for some media objects tracked
	 *
	 */
	public synchronized int statusID(int id, boolean startLoad) {
		if (startLoad) startLoading(id);

		if (head == null)
			return 0;
		int status;
		int totalStatus = 0;

		MediaEntry end = head;
		while (end != null) {
			if (end.id == id) {
				status = end.getStatus();
				if (status == -1) {
					status = LOADING;
				}
				totalStatus = totalStatus | status;
			}
			end = end.next;
		}
		return totalStatus;

	}

	private void load(MediaEntry me) {
		Image image  = me.image;
		//ImageProducer source = image.getSource();

		if (me.getStatus() == 0) {
			IImageImpl iImageImpl = BufferedImageHandler.createIImageImpl(image);
			// Note that the accessor will create the image if it doesn't exist yet.
			iImageImpl.getImagePeer();
		}
	}

	private synchronized void startLoading() {
		if (head == null)
			return;

		MediaEntry end = head;
		while (end != null) {
			load(end);

			end = end.next;
		}
	}

	private synchronized void startLoading(int index) {
		if (head == null)
			return;

		MediaEntry end = head;
		while (end != null) {
			if (end.id == index) {
				load(end);
			}

			end = end.next;
		}
	}

	//	Remove all media entries from the
	//	linked list of entries
	private synchronized void remove(MediaEntry me) {
		MediaEntry entry = head;
		while (entry != null) {
			MediaEntry next = entry.next;
			if (entry == me) {
				if (entry == head) {
					head = entry.next;
					entry.next = null;
					if (head != null) {
						head.previous = null;
					}
				} else if (entry.next == null) {
					entry.previous.next = null;
					entry.previous = null;
				} else {
					entry.previous.next = entry.next;
					entry.next.previous = entry.previous;
					entry.next = null;
					entry.previous = null;
				}
			}
			entry = next;
		}
	}

	//	Remove all media entries from the
	//	linked list of entries with the specified image
	private synchronized void remove(Image image) {
		if (head == null)
			return;

		MediaEntry entry = head;
		while (entry != null) {
			MediaEntry next = entry.next;
			if (entry.image == image)
				remove(entry);
			entry = next;
		}
	}

	//	Remove all media entries from the
	//	linked list of entries with the specified image
	//	and id.
	private synchronized void remove(Image image, int id) {
		if (head == null)
			return;

		MediaEntry entry = head;
		while (entry != null) {
			MediaEntry next = entry.next;
			if ((entry.image == image) && (entry.id == id))
				remove(entry);
			entry = next;
		}
	}

	//	Remove all media entries from the
	//	linked list of entries with the specified image
	private synchronized void remove(
		Image image, int id, int width, int height) {
		if (head == null)
			return;

		MediaEntry entry = head;
		while (entry != null) {
			MediaEntry next = entry.next;
			if (entry.image == image
				&& entry.id == id
				&& entry.width == width
				&& entry.height == height)
				remove(entry);
			entry = next;
		}
	}

	/**
	 * Remove the specified image from the list of objects
	 * whose loading is being tracked by the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	image	the image being loaded
	 *
	 */
	public void removeImage(Image image) {
		remove(image);
	}

	/**
	 * Remove the specified image tagged with the specified
	 * ID from the list of objects
	 * whose loading is being tracked by the receiver.
	 * This method ignores the size and removes images that
	 * match the specified ID.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	image	the image being loaded
	 * @param	id		the integer ID of objects tracked
	 *
	 */
	public void removeImage(Image image, int id) {
		remove(image, id);
	}

	/**
	 * Remove the specified image tagged with the specified
	 * ID and with the specified scale width and height,
	 * from the list of objects whose loading is being tracked
	 * by the receiver.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	image	the image being loaded
	 * @param	id		the integer ID of objects tracked
	 * @param	width	the width of the image
	 * @param	height	the height of the image

	 *
	 */
	public void removeImage(Image image, int id, int width, int height) {
		remove(image, id, width, height);
	}

}
