package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>Menu</code> is the selectable user interface object
 * which appears in a <code>MenuBar</code> object or in another
 * <code>Menu</code> as a cascading menu.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class Menu extends MenuItem implements MenuContainer {

	static final long serialVersionUID = -8809584163345499784L;

	boolean isHelpMenu = false;
	Vector items = new Vector();
	boolean tearOff = false;

	transient int menuPeer;

	int menuSerializedDataVersion;

	/**
	 * Constructs a new <code>Menu</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Menu()
		throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>Menu</code> with a label.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textLabel	the text label for the receiver
	 */
	public Menu(String textLabel)
		throws HeadlessException
	{
		this(textLabel, false);
	}

	/**
	 * Constructs a new <code>Menu</code> with a label and
	 * and specifies if it can be torn off.  On some platforms
	 * menus can be torn off their parent and placed elsewhere
	 * on the screen.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textLabel	the text label for the receiver
	 * @param	tearOff		the tear off property for the receiver
	 *
	 */
	public Menu(String textLabel, boolean tearOff)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		label = textLabel;
		this.tearOff = tearOff;
	}

	void _addNotify() {
		if ((peer == 0) && (parent != null) && (parent instanceof MenuComponent)) {

			peer = OS.CascadeMenuItem_New(_getParentPeer(), -1);
			menuPeer = OS.DropDownMenu_New(getOwningWindow().peer);
			OS.CascadeMenuItem_SetMenu(peer, menuPeer);

			OS.AbstractMenuItem_SetEnabled(peer, enabled);
			OS.Widget_SetData(peer, this);
			if (label != null) {
				setText();
			}
			for (int i = 0; i < items.size(); i++)
				 ((MenuItem) (items.elementAt(i)))._addNotify();

			if (font != null) {
				_widgetSetFont(font);
			}
		}
	}

	public void removeNotify() {
		if (menuPeer != 0) {
			for (int i = 0; i < items.size(); i++)
				 ((MenuItem) (items.elementAt(i))).removeNotify();

			OS.Widget_SetData(menuPeer, null);
			OS.Widget_Dispose(menuPeer);
			menuPeer = 0;

			super.removeNotify();
		}
	}

	/**
	 * Answer whether the receiver can be torn off its parent
	 * and placed elsewhere on the screen.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		<code>true</code> if the receiver is a tear off menu,
	 *				<code>false</code> otherwise
	 */
	public boolean isTearOff() {
		return tearOff;
	}

	/**
	 * Answer the number of <code>MenuItem</code> objects contained
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int representing the number of menu items in the receiver
	 */
	public int getItemCount() {
		return items.size();
	}

	/**
	 * Answer the number of items contained
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int representing the number of items in the receiver
	 *
	 * @deprecated	use <code>getItemCount()</code> instead
	 */
	public int countItems() {
		return getItemCount();
	}

	/**
	 * Answers the <code>MenuItem</code> at the specified
	 * at index.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		index	the index of the menu item to return
	 *
	 * @return		a java.awt.MenuItem	the menu item at the index specified
	 *
	 * @see			java.awt.MenuItem
	 */
	public MenuItem getItem(int index) {
		return (MenuItem) (items.elementAt(index));
	}

	/**
	 * Adds a <code>MenuItem</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		menuItem	the <code>MenuItem</code> to be added.
	 *
	 * @return		a <code>MenuItem</code> the MenuItem added
	 *
	 * @see			java.awt.MenuItem
	 */
	public MenuItem add(MenuItem menuItem) {
		synchronized (getTreeLock()) {
			if (null != menuItem.parent) {
				menuItem.parent.remove(menuItem);
				if (menuPeer != 0) {
					menuItem.removeNotify();
				}
			}

			menuItem.parent = this;
			items.addElement(menuItem);
		}
		if (menuPeer != 0)
			menuItem.addNotify();
		return menuItem;
	}

	/**
	 * Adds a <code>MenuItem</code> to the receiver with
	 * the specified label.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		label	the label for the item to be added.
	 */
	public void add(String textLabel) {
		add(new MenuItem(textLabel));
	}

	/**
	 * Insert a <code>MenuItem</code> into the receiver at the
	 * specified index.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	menuItem the menu item to insert into the receiver.
	 * @param	index	the position to insert the menu item at.
	 */
	public void insert(MenuItem menuItem, int index) {
		if (index < 0)
			throw new IllegalArgumentException();

		synchronized (getTreeLock()) {
			if (menuItem.parent != null) {
				removeFromParent(menuItem);
			}
			menuItem.parent = this;
			int tempIndex = index;
			if (index < 0)
				tempIndex = 0;
			if (index > items.size())
				tempIndex = items.size();
			menuItem.index = tempIndex;
			items.insertElementAt(menuItem, tempIndex);
			if (menuPeer != 0)
				menuItem._addNotify();
			menuItem.index = -1;
		}
	}

	/**
	 * Insert a <code>MenuItem</code> with the label specified into
	 * the receiver at the specified index.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	label	the label for the menu item to insert into the receiver.
	 * @param	index	the position to insert the menu item at.
	 */
	public void insert(String textLabel, int index) {
		insert(new MenuItem(textLabel), index);
	}

	/**
	 * Adds a separator to the receiver after what is currently
	 * the last item in the <code>Menu</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void addSeparator() {
		synchronized (getTreeLock()) {
			MenuItem menuItem;
			menuItem = new MenuItem(MenuItem.separatorLabel());
			menuItem.parent = this;
			items.addElement(menuItem);
			if (menuPeer != 0)
				menuItem.addNotify();
		}
	}

	/**
	 * Insert a separator into the receiver at the specified index.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	index	the position to insert the separator at.
	 */
	public void insertSeparator(int index) {
		insert(new MenuItem(MenuItem.separatorLabel()), index);
	}

	/**
	 * Remove a <code>MenuItem</code> at the specified index
	 * from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		index	the index of the menu item to be removed.
	 *
	 */
	public void remove(int index) {
		remove((MenuComponent) items.elementAt(index));
	}

	/**
	 * Remove a <code>MenuComponent</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		menuComponent	the <code>MenuComponent</code> to be removed.
	 *
	 * @see			java.awt.MenuComponent
	 */
	public void remove(MenuComponent menuComponent) {
		if (menuComponent == null)
			return;

		synchronized (getTreeLock()) {
			menuComponent.removeNotify();
			items.remove(menuComponent);
		}
	}

	/**
	 * Remove all items from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see			java.awt.MenuComponent
	 */
	public void removeAll() {
		synchronized (getTreeLock()) {
			for (int i = 0; i < items.size(); i++)
				 ((MenuItem) (items.elementAt(i))).removeNotify();
			items.removeAllElements();
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the menu component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	public String paramString() {
		return super.paramString() + " isHelpMenu=" + isHelpMenu + " tearOff=" + isTearOff(); //$NON-NLS-1$ //$NON-NLS-2$
	}

	void removeFromParent(MenuItem menuItem) {
		if (menuItem.parent instanceof Menu) {
			((Menu) menuItem.parent).remove(menuItem);
		} else {
			menuItem.parent.remove(menuItem);
		}
		if (menuPeer != 0) {
			menuItem.removeNotify();
		}
	}

	/*
	 * If the receiver's shortcut matches <code>menuShortcut</code>, answer
	 * the receiver.  Otherwise answer the first child for which a match
	 * occurs.  If no match found, answer <code>null</code>.
	 */
	MenuItem getShortcutMenuItem(MenuShortcut menuShortcut) {
		MenuItem item = super.getShortcutMenuItem(menuShortcut);
		if (item != null) return item;
		for (int i = 0; i < getItemCount(); i++) {
			item = getItem(i).getShortcutMenuItem(menuShortcut);
			if (item != null) return item;
		}
		return null;
	}

	/*
	 * Answer an enumeration containing the receiver's shortcut &
	 * shortcuts for its children
	 */
	Enumeration getAllShortcuts() {
		Vector shortcutItems = new Vector();
//		if (getShortcut() != null) {
//			shortcutItems.add(getShortcut());
//		}
		for (int i=0; i < getItemCount(); i++) {
			for (Enumeration e = getItem(i).getAllShortcuts(); e.hasMoreElements(); ) {
				shortcutItems.addElement(e.nextElement());
			}
		}
		return shortcutItems.elements();
	}

	String classNonlocalizedName() {
		return "menu"; //$NON-NLS-1$
	}

	/*
	 * This does basically the same thing as the parent class but
	 * it is subclassed because it references the peer, which is not
	 * compatible with the peer that the MenuItem is expecting
	 */
	void _widgetSetFont(Font font) {
		// Super is called so that the owning
		// cascade item is updated with the new
		// font as well.
		super._widgetSetFont(font);

		if (menuPeer != 0) {
			FontPeer fontPeer =  font.getPeer();
			OS.Menu_SetFont(menuPeer, (fontPeer == null ? 0 : fontPeer.fPeerHandle));
		}

		// Ensure all the childrens' fonts are also updated, per the spec
		for (Iterator iter = items.iterator(); iter.hasNext();) {
			MenuItem element = (MenuItem) iter.next();
			// must call getFont() to ensure that menu item's font is used if present
			element._widgetSetFont(element.getFont());
		}
	}

	String _shortcutString() {
		if (getParent() instanceof MenuBar) return ""; //$NON-NLS-1$
		return super._shortcutString();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}

	private void readObject(ObjectInputStream stream)
		throws IOException, ClassNotFoundException
		, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}

}
