package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.ugl.eswt.OS;

/**
 * <code>MenuBar</code> is the user interface object which typically
 * appears as a bar along the top of a <code>Frame</code> and
 * contains selectable menu items.
 *
 * @author		IBM
 * @version		initial
 *
 */
public class MenuBar extends MenuComponent implements MenuContainer {

	static final long serialVersionUID = -4930327919388951260L;

	Vector menus;
	Menu helpMenu;

	int menuBarSerializedDataVersion;

	//	Used to generate the nonlocalized name of the receiver. If
	//	the receiver's name has not been explicitly set, it will
	//	be initialized to <code>classNonlocalizedName() + nameCounter</code>
	static int nameCounter;

	/**
	 * Constructs a new <code>MenuBar</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public MenuBar()
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		menus = new Vector();
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void addNotify() {
		_addNotify();
	}

	void _addNotify() {
		if (peer == 0) {
			peer = OS.MenuBar_New(((Frame)parent).peer);
			if (font != null) {
				_widgetSetFont(font);
			}
			for (int i = 0; i < menus.size(); i++)
				((Menu)(menus.elementAt(i)))._addNotify();
		}
	}

	public void removeNotify() {
		Frame parent = (Frame) getParent();

		if (peer != 0) {
			for (int i=0; i < menus.size(); i++) {
				((MenuComponent)menus.elementAt(i)).removeNotify();
			}

			OS.Shell_SetMenuBar(parent.peer, 0);
			super.removeNotify();
			peer = 0;
		}
	}

	void _widgetSetFont(Font font) {
		if (peer != 0) {
			FontPeer fontPeer = font.getPeer();
			OS.Menu_SetFont(peer, (fontPeer == null ? 0 : fontPeer.fPeerHandle));
		}
	}

	/**
	 * Answers the <code>Menu</code> which is the help menu
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		a <code>Menu</code> the help menu
	 *
	 * @see			java.awt.Menu
	 */
	public Menu getHelpMenu() {
		return helpMenu;
	}

	/**
	 * Set the <code>Menu</code> which is the help menu
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		a <code>Menu</code> to be the help menu
	 *
	 * @see			java.awt.Menu
	 */
	public void setHelpMenu(Menu menu) {
		synchronized (getTreeLock()) {
			if (helpMenu != null) {
				helpMenu.isHelpMenu = false;
				remove(helpMenu);
			}
			helpMenu = menu;
			if (helpMenu != null) {
				helpMenu.isHelpMenu = true;
				add(menu);
			}
		}
	}

	/**
	 * Adds a <code>Menu</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		menu	the <code>Menu</code> to be added.
	 *
	 * @return		a java.awt.Menu the menu added
	 *
	 * @see			java.awt.Menu
	 */
	public Menu add(Menu menu) {
		synchronized (getTreeLock()) {
			if(menu.parent != null) {
				menu.parent.remove(menu);
			}
			menu.parent = this;
			menus.addElement(menu);
		}
		if (peer != 0) {
			menu.addNotify();
		}
		return menu;
	}

	/**
	 * Remove a <code>MenuComponent</code> at the specified index
	 * from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		index	the index of the menu component to be removed.
	 *
	 */
	public void remove(int index) {
		remove((MenuComponent)menus.elementAt(index));
	}

	/**
	 * Remove a <code>MenuComponent</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		menuComponent	the <code>MenuComponent</code> to be removed.
	 *
	 * @see			java.awt.MenuComponent
	 */
	public void remove(MenuComponent menuComponent) {
		menuComponent.removeNotify();
		menus.remove(menuComponent);
	}

	/**
	 * Answer the number of <code>Menu</code> objects contained
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int representing the number of menus in the receiver
	 */
	public int getMenuCount() {
		return menus.size();
	}

	/**
	 * Answer the number of <code>Menu</code> objects contained
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int representing the number of menus in the receiver
	 *
	 * @deprecated	use <code>getMenuCount()</code> instead
	 */
	public int countMenus() {
		return getMenuCount();
	}

	/**
	 * Answers the <code>Menu</code> at the specified
	 * at index.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		index	the index of the menu to return
	 *
	 * @return		a <code>Menu</code> the menu at the index specified
	 *
	 * @see			java.awt.Menu
	 */
	public Menu getMenu(int index) {
		return (Menu)menus.elementAt(index);
	}

	/**
	 * Answer an <code>Enumeration</code> of all the shortcuts for
	 * the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>Enumeration</code> the list of shortcuts
	 */
	public synchronized Enumeration shortcuts() {
		Vector shortcutItems = new Vector();

		for (int i = 0; i < menus.size(); i++) {
			MenuItem mi = (MenuItem)menus.elementAt(i);
			for (Enumeration e = mi.getAllShortcuts(); e.hasMoreElements(); ) {
				shortcutItems.addElement(e.nextElement());
			}
		}

		return shortcutItems.elements();
	}

	/**
	 * Answer the <code>MenuItem</code> associated with the
	 * specified <code>MenuShortcut</code>, or <code>null</code> if
	 * one was not found.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	menuShortcut	the menu shortcut for the menu item to find
	 *
	 * @return	a <code>MenuItem</code> the menu item associated
	 *			with the menu shortcut
	 */
	public MenuItem getShortcutMenuItem(MenuShortcut menuShortcut) {
		if (menuShortcut == null) return null;
		synchronized (getTreeLock()) {
			for (int i = 0; i < menus.size(); i++) {
				MenuItem mi = ((MenuItem)menus.elementAt(i)).getShortcutMenuItem(menuShortcut);
				if (mi != null) return mi;
			}
		}
		return null;
	}

	/**
	 * Delete the specified shortcut from the receiver.
	 * Removes all children of the receiver which are using
	 * the specified <code>MenuShortcut</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	menuShortcut	the shortcut to remove.
	 */
	public void deleteShortcut(MenuShortcut menuShortcut) {
		synchronized (getTreeLock()) {
			MenuItem menuItem = getShortcutMenuItem(menuShortcut);
			while (menuItem != null) {
				menuItem.deleteShortcut();
				menuItem = getShortcutMenuItem(menuShortcut);
			}
		}
	}

	String classNonlocalizedName() {
		return "menubar"; //$NON-NLS-1$
	}

	public String getName() {
		if (!nameExplicitlySet && (name == null)) {
//			name = classNonlocalizedName() + "-" + nameCounter++;
			String tempName = classNonlocalizedName();
			if (tempName != null) {
				name = tempName + nameCounter++;
			}
		}
		return name;
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
//		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
//		String key = null;
//		while ((key = (String)stream.readObject()) != null) {
//			Object listener = stream.readObject();
//		}
	}

}
