package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;

import com.ibm.ugl.eswt.OS;

/**
 * <code>MenuComponent</code> is the abstract superclass of
 * all menu widget classes.  The term widget is used to describe
 * graphical user interface objects.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public abstract class MenuComponent implements Serializable {

	static final long serialVersionUID = -4536902356223894379L;

	Font font;
	String name;
	boolean nameExplicitlySet;
	boolean newEventsOnly;

	transient MenuContainer parent;

	// The wrapper on the native widget
	transient int peer;

	// Subclasses should override to provide concrete impl
	void _addNotify() {
	}

	/**
	 * Constructs a new <code>MenuComponent</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public MenuComponent()
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
	}

	/**
	 * Answer the name of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is the receiver's name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Set the name of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	name	the name for the receiver
	 */
	public void setName(String name) {
		this.nameExplicitlySet = true;
		this.name = name;
	}

	/**
	 * Answer the parent of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.MenuContainer which is the receiver's parent
	 */
	public MenuContainer getParent() {
		return parent;
	}

	/**
	 * Answer the <code>Font</code> used by the receiver.  If
	 * a <code>Font</code> has not been set answer <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		a java.awt.Font
	 */
	public Font getFont() {
		if (font != null)
			return font;

		if (parent != null)
			return parent.getFont();

		return null;
	}

	/**
	 * Set the <code>Font</code> used by the receiver.  The
	 * specified <code>Font</code> is also used for children of
	 * the receiver when they do not explicitly set their own font.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	font	the font for the receiver
	 */
	public void setFont(Font font) {
		this.font = font;

		_widgetSetFont(font);
	}

	void _widgetSetFont(Font font) {
	}

	Window getOwningWindow() {
		if (parent == null) return null;

		if (parent instanceof Component) {
			return ((Component)parent).getWindow();
		}

		if (parent instanceof MenuComponent) {
			return ((MenuComponent)parent).getOwningWindow();
		}

		return null;
	}

	/**
	 * Destroy the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void removeNotify() {
//		synchronized (getTreeLock()) {
		if (peer != 0) {
			OS.Widget_SetData(peer, null);
			OS.Widget_Dispose(peer);
			peer = 0;
		}
//		}
	}

	/**
	 * Post the event to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		event the event to be posted
	 *
	 * @deprecated	use <code>dispatchEvent(AWTEvent)</code> instead.
	 */
	public boolean postEvent(Event event) {
		if (parent != null)
			return parent.postEvent(event);

		return false;
	}

	void postEvent(AWTEvent awtEvent) {
		getToolkit().getSystemEventQueueImpl().postEvent(awtEvent);
	}

	/**
	 * Dispatch the specified event to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be dispatched
	 */
	public final void dispatchEvent(AWTEvent awtEvent) {
		MenuComponent processingMenuComponent;

		processingMenuComponent = findMenuComponentForEvent(this, awtEvent);

		if (null != processingMenuComponent) {
			awtEvent.modifyEventSource(processingMenuComponent);
			processingMenuComponent.processEvent(awtEvent);
		} else {
			Event event10 = awtEvent.convertTo10();
			if (event10 != null) postEvent(event10);
		}
	}

	/**
	 * Find a MenuComponent that can handle this event, or return null
	 */
	private MenuComponent findMenuComponentForEvent(MenuComponent menuComponent, AWTEvent awtEvent) {
		while (menuComponent != null) {
			if (menuComponent.newEventsOnly) {
				if (menuComponent.isEventEnabled(awtEvent)) {
					return menuComponent;
				}
			}

			MenuContainer menuContainer = menuComponent.getParent();
			if (menuContainer instanceof MenuComponent) {
				menuComponent = (MenuComponent) menuContainer;
			} else {
				menuComponent = null;
			}
		}

		return null;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		return false;
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a menu component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a menu component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * Subclasses of <code>MenuComponent</code> will reimplement this to
	 * handle events that it wants to handle.  By default <code>MenuComponent</code>
	 * will do nothing.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		// do nothing for MenuComponent?
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the menu component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return nameExplicitlySet ? getName() : ""; //$NON-NLS-1$
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[" + paramString() + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Answer a synchronizable object against the component tree.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.Object to use as a lock
	 */
	protected final Object getTreeLock() {
		return Component.TREE_LOCK;
	}

	Toolkit getToolkit() {
		// if (parent != null)
		//	return parent.getToolkit();

		return Toolkit.getDefaultToolkit();
	}

	String classNonlocalizedName() {
		return "menucomponent"; //$NON-NLS-1$
	}

	private void readObject(ObjectInputStream stream)
		throws IOException, ClassNotFoundException
		, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}

}
