package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.EventListener;
import java.util.Vector;

import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.metal.graphics.FontPeer;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>MenuItem</code> is the selectable user interface object
 * which appears in a <code>Menu</code> object.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class MenuItem extends MenuComponent {

	static final long serialVersionUID = -21757335363267194L;

	//	Used to generate the nonlocalized name of the receiver. If
	//	the receiver's name has not been explicitly set, it will
	//	be initialized to <code>classNonlocalizedName() + nameCounter</code>
	static int nameCounter;

	String actionCommand;
	boolean enabled;
	long eventMask = 0;
	String label;
	MenuShortcut shortcut;

	// The action listener of the receiver
	transient ActionListener actionListener;
	// used to remember the index of the widget during creation only
	transient int index;

	int menuItemSerializedDataVersion;

	private static final int MOD_MASK = OS.MODIFIER_MASK_CONTROL << 16 | OS.MODIFIER_MASK_SHIFT << 16;

	/*
	 * Set the text in the receiver's underlying widget.  Add
	 * in text for the shortcut if necessary.
	 */
	void setText() {
		if (peer != 0) {
			_widgetSetText(label, _shortcutString());
		}
	}

	String _shortcutString() {
		// Returns a concise string which should be displayed in full by the MenuItemPeer.
		return (shortcut == null) ? "" : (shortcut.toString()); //$NON-NLS-1$
	}

	void _widgetSetText(String newLabel, String shortcutText) {
		if (_isSeparator()) return;
		OS.LabeledMenuItem_SetText(peer, newLabel, shortcutText);
	}

	void _widgetSetAccelerator(int accel) {
		if (_isSeparator()) return;
		int mods = accel & MOD_MASK;
		int key = Util.convertAWTToNativeKeycode(accel & 0x0000FFFF);
		if (key != -1) {
		    OS.LabeledMenuItem_SetAccelerator(peer, mods | key);
		} else {
		    OS.LabeledMenuItem_SetAccelerator(peer, 0);
		}
	}

	void setAccelerator() {
		if (peer == 0)
			return;

		if (shortcut != null) {
			int accel = OS.MODIFIER_MASK_CONTROL << 16;
			if (shortcut.usesShiftModifier())
				accel = accel | OS.MODIFIER_MASK_SHIFT << 16;
			accel = accel | (char) shortcut.getKey();
			_widgetSetAccelerator(accel);
		} else
			_widgetSetAccelerator(0);
	}

	/**
	 * Constructs a new <code>MenuItem</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public MenuItem()
		throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>MenuItem</code> with a label.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textLabel	the text label for the receiver
	 */
	public MenuItem(String textLabel)
		throws HeadlessException
	{
		this(textLabel, null);
	}

	/**
	 * Constructs a new <code>MenuItem</code> with a label and
	 * a shortcut.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textLabel		the text label for the receiver
	 * @param	menuShortcut	the menu shortcut for the receiver
	 *
	 */
	public MenuItem(String textLabel, MenuShortcut menuShortcut)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		label = textLabel;
		shortcut = menuShortcut;
		enabled = true;
		index = -1;
	}

	public void addNotify() {
		_addNotify();
	}

	void _addNotify() {
		if (peer != 0)
			return;
		peer = _createPeer();
		if (!_isSeparator() && label != null) {
			setText();
		}

		OS.AbstractMenuItem_SetEnabled(peer, enabled);
		if (shortcut != null) {
			setAccelerator();
		}

		OS.Widget_SetData(peer, this);
		// Must call getFont() instead of direct field access to ensure font is inherited from
		// parent, if the parent has a non-default font
		Font newFont = getFont();
		if (newFont != null) {
			_widgetSetFont(newFont);
		}
	}

	boolean _isSeparator() {
		return (label != null) && (label.equals(MenuItem.separatorLabel()));
	}

	void _widgetSetFont(Font font) {
		if (_isSeparator()) return;

		if (peer != 0) {
			FontPeer fontPeer = font.getPeer();
			OS.LabeledMenuItem_SetFont(peer, (fontPeer == null ? 0 : fontPeer.fPeerHandle));
		}
	}

	int _createPeer() {
		if (_isSeparator()) {
			return OS.SeparatorItem_New(_getParentPeer(), index);
		} else {
			return OS.PushMenuItem_New(_getParentPeer(), index);
		}
	}

	void selectionCallback(int modifiers) {
		int pproModifiers = 0;
		if ((modifiers & OS.MODIFIER_MASK_ALT) == OS.MODIFIER_MASK_ALT) pproModifiers |= ActionEvent.ALT_MASK;
		if ((modifiers & OS.MODIFIER_MASK_CONTROL) == OS.MODIFIER_MASK_CONTROL) pproModifiers |= ActionEvent.CTRL_MASK;
		if ((modifiers & OS.MODIFIER_MASK_SHIFT) == OS.MODIFIER_MASK_SHIFT) pproModifiers |= ActionEvent.SHIFT_MASK;

		ActionEvent event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, getActionCommand(),
			System.currentTimeMillis(),
		modifiers);
		postEvent(event);
	}

	int _getParentPeer() {
		if (parent instanceof Menu) {
			return ((Menu)parent).menuPeer;
		}
		return ((MenuComponent)parent).peer;
	}

	public String getName() {
		if (!nameExplicitlySet && (name == null)) {
//			name = classNonlocalizedName() + "-" + nameCounter++;
			String tempName = classNonlocalizedName();
			if (tempName != null) {
				name = tempName + nameCounter++;
			}
		}
		return name;
	}

	/**
	 * Answer the <code>String</code> that is the
	 * label for the receiver.  Answer <code>null</code> if
	 * no label is specified.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> the label for the receiver
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Set the <code>String</code> that is the
	 * label for the receiver.  Use <code>null</code> if
	 * no label is desired.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text the label for the receiver
	 */
	public synchronized void setLabel(String text) {
		label = text;
		setText();
	}

	/**
	 * Answer whether the receiver is enabled and will respond
	 * to events.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return		<code>true</code> if the receiver is enabled,
	 *				<code>false</code> otherwise
	 */
	public boolean isEnabled() {
		return enabled;
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ActionEvent)
			return ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) || actionListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	/**
	 * Enable or disable the receiver according to
	 * the specified value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	enabled	<code>true</code> if the receiver is to be enabled
	 *					<code>false</code> for disabling the receiver
	 */
	public synchronized void setEnabled(boolean enabled) {
		this.enabled = enabled;

		if (peer != 0) {
			OS.AbstractMenuItem_SetEnabled(peer, enabled);
		}
	}

	/**
	 * Enable the receiver so that it will respond
	 * to events.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 */
	public void enable() {
		setEnabled(true);
	}

	/**
	 * Enable or disable the receiver according to
	 * the specified value.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	enabled	<code>true</code> if the receiver is to be enabled
	 *					<code>false</code> for disabling the receiver
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 */
	public void enable(boolean enabled) {
		setEnabled(enabled);
	}

	/**
	 * Disable the receiver.  In this state a widget
	 * will typically not respond to events.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @deprecated	use <code>setEnabled(boolean)</code> instead
	 */
	public void disable() {
		setEnabled(false);
	}

	/**
	 * Answer the <code>MenuShortcut</code> associated with the
	 * receiver, or <code>null</code> if there is none.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>MenuShortcut</code> the menu shortcut associated
	 *			with the receiver
	 */
	public MenuShortcut getShortcut() {
		return shortcut;
	}

	/**
	 * Set the <code>MenuShortcut</code> associated with the
	 * receiver.  This will overwrite previous shortcuts.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	menuShortcut the menu shortcut to be associated
	 *			with the receiver
	 */
	public synchronized void setShortcut(MenuShortcut menuShortcut) {
		shortcut = menuShortcut;
		setText();
		setAccelerator();
	}

	/**
	 * Delete menu shortcuts from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public synchronized void deleteShortcut() {
		shortcut = null;

		if (peer != 0) {
			setText();
			setAccelerator();
		}
	}

	/**
	 * Enable the specified events for delivery to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	events	events to enable
	 */
	protected final void enableEvents(long events) {
		newEventsOnly = true;
		eventMask |= events;
	}

	/**
	 * Disable the specified events from being delivered.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	events	events to disable
	 */
	protected final void disableEvents(long events) {
		eventMask &= ~events;
	}

	/**
	 * Set the <code>String</code> representing the
	 * name of the action command for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	a <code>String</code> the command name
	 */
	public void setActionCommand(String command) {
		actionCommand = command;
	}

	/**
	 * Answer the <code>String</code> representing the
	 * name of the action command for the receiver. If
	 * the receiver's action command is not set, the
	 * receiver's label is returned.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> the command name
	 */
	public String getActionCommand() {
		if (actionCommand != null) return actionCommand;

		return getLabel();
	}

	/**
	 * Adds an <code>ActionListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	actionListener	the <code>ActionListener</code> to be added.
	 *
	 * @see		java.awt.event.ActionListener
	 * @see		java.awt.event.ActionEvent
	 */
	public synchronized void addActionListener(ActionListener actionListener) {
		addEventListener(actionListener);
		this.actionListener = AWTEventMulticaster.add(this.actionListener, actionListener);
	}

	void addEventListener(EventListener eventListener) {
		newEventsOnly = true;
	}

	/**
	 * Removes a <code>ActionListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	actionListener	the <code>ActionListener</code> to be removed.
	 *
	 * @see		java.awt.event.ActionListener
	 * @see		java.awt.event.ActionEvent
	 */
	public synchronized void removeActionListener(ActionListener actionListener) {
		if (actionListener == null) return;
		this.actionListener = AWTEventMulticaster.remove(this.actionListener, actionListener);
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a menu component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a menu component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ActionEvent</code>, the receiver
	 * will forward calls to the default events by calling
	 * <code>processActionEvent(FocusEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ActionEvent) {
			processActionEvent((ActionEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ActionEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a menu component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a menu component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processActionEvent(ActionEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	actionEvent	the event to be processed
	 *
	 * @see		ActionEvent
	 */
	protected void processActionEvent(ActionEvent actionEvent) {
		if (actionListener == null)
			return;

		actionListener.actionPerformed(actionEvent);
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the menu component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	public String paramString() {
		String paramString = ((getName() != null) ? getName() : "") + ",label=" + getLabel(); //$NON-NLS-1$ //$NON-NLS-2$
		if (shortcut != null)
			paramString = paramString + ",shortcut=" + shortcut.paramString(); //$NON-NLS-1$
		return paramString;
	}

	/*
	 * Answer an enumeration containing the receiver's shortcut
	 */
	Enumeration getAllShortcuts() {
		Vector shortcutItems = new Vector();
		if (getShortcut() != null) {
			shortcutItems.add(getShortcut());
		}
		return shortcutItems.elements();
	}

	/*
	 * Answer the receiver if it has a shortcut equal to <code>menuShortcut</code>
	 * Otherwise answer <code>null</code>.
	 */
	MenuItem getShortcutMenuItem(MenuShortcut menuShortcut) {
		return menuShortcut.equals(getShortcut()) ? this : null;
	}

	/**
	 * Answer the string used as the label for menu items that are separators
	 */
	static String separatorLabel() {
		return "-"; //$NON-NLS-1$
	}

	String classNonlocalizedName() {
		return "menuitem"; //$NON-NLS-1$
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, Component.actionListenerK, actionListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(Component.actionListenerK))
				addActionListener((ActionListener)listener);
		}
	}

	public ActionListener[] getActionListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(actionListener, ActionListener.class);
		ActionListener[] resultArray = new ActionListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}
}
