package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

import java.awt.event.KeyEvent;
import java.io.Serializable;

/**
 * MenuShortcut is the class used to describe a keyboard shortcut
 * for <code>MenuItem</code> instances.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class MenuShortcut implements Serializable {

	static final long serialVersionUID = 143448358473180225L;

	int key = KeyEvent.CHAR_UNDEFINED;
	boolean usesShift = false;

	/**
	 * Constructs a new <code>MenuShortcut</code>.
	 *
	 * @author	IBM
	 * @version 	initial
	 *
	 * @param	shortcutKey	the key pressed to trigger this menu
	 *
	 */
	public MenuShortcut(int shortcutKey) {
		this(shortcutKey, false);
	}

	/**
	 * Constructs a new <code>MenuShortcut</code>.
	 * If <code>shift</code> is specified then shift
	 * must be pressed in combination with the shortcut key
	 * to trigger the menu.
	 *
	 * @author	IBM
	 * @version 	initial
	 *
	 * @param	shortcutKey	the key pressed to trigger this menu
	 * @param	shift		<code>true</code> if shift is part of the shortcut,
	 *						<code>false</code> otherwise.
	 */
	public MenuShortcut(int shortcutKey, boolean shift) {
		key = shortcutKey;
		//	if (key >= 0) {
		//		if (Character.isLowerCase((char)key)) {
		//			char upper = Character.toUpperCase((char)key);
		//			key = (int)upper;
		//		}
		//	}
		usesShift = shift;
	}

	/**
	 * Answer the key specified as the shortcut key
	 * for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an int the shortcut key for the receiver
	 */
	public int getKey() {
		return key;
	}

	/**
	 * Answer whether the receiver uses the shift key
	 * modifier or not.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	boolean	<code>true</code> if the menuShortcut
	 *					uses the shift modifier and
	 *					<code>false</code> otherwise
	 */
	public boolean usesShiftModifier() {
		return usesShift;
	}

	/**
	 * Compares the argument to the receiver, and answers true
	 * if they represent the <em>same</em> object using a class
	 * specific comparison. The implementation in MenuShortcut
	 * answers true only if the argument uses the same key and
	 * shift modifier.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	menuShortcut	the other MenuShortcut
	 *					to compare with the receiver.
	 *
	 * @return	boolean		<code>true</code> if the menuShortcut
	 *					is the same as the receiver
	 *					<code>false</code> otherwise
	 */
	public boolean equals(MenuShortcut menuShortcut) {
		if (menuShortcut == null)
			return false;
		if (this == menuShortcut)
			return true;
		if ((key == menuShortcut.getKey()) && (usesShift == menuShortcut.usesShiftModifier()))
			return true;
		return false;
	}

	/**
	 * Compares the argument to the receiver, and answers true
	 * if they represent the <em>same</em> object using a class
	 * specific comparison. The implementation in MenuShortcut
	 * answers true only if the argument uses the same key and
	 * shift modifier.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	object	the other MenuShortcut
	 *					to compare with the receiver.
	 *
	 * @return	boolean	<code>true</code> if the menuShortcut
	 *					is the same as the receiver
	 *					<code>false</code> otherwise
	 */
	public boolean equals(Object object) {
		if (object == null)
			return false;
		if (this == object)
			return true;
		if (!(object instanceof MenuShortcut))
			return false;
		return equals((MenuShortcut)object);
	}

	/**
	 * Answers an integer hash code for the receiver. Any two
	 * objects which answer <code>true</code> when passed to
	 * <code>.equals</code> must answer the same value for this
	 * method.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	int	the receiver's hash.
	 *
	 * @see		#equals
	 */
	public int hashCode() {
		if (usesShift)
			return key + 1;
		else
			return key;
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		String out = Toolkit.getProperty("AWT.control", "Ctrl") + "+";  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		if (usesShift)
			out = out + Toolkit.getProperty("AWT.shift", "Shift") + "+";  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

		return out + KeyEvent.getKeyText(key);
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		String paramString = toString();  //$NON-NLS-1$
		if (usesShift)
			paramString = paramString + ",usesShiftModifier";  //$NON-NLS-1$

		return paramString;
	}

}
