package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.Serializable;

/**
 * <code>Point</code> represents a specific coordinate location that uses int
 * precision to store its data.
 * <p>
 * For instance, a point is used to represent the location of components.
 * <p>
 *
 * @author IBM
 * @version initial
 *
 * @see java.io.Serializable
 */

public class Point implements Serializable, Cloneable
{

	static final long serialVersionUID = -5276940640259749850L;

	/**
	 * The x coordinate of the receiver.
	 */
	public int x;

	/**
	 * The y coordinate of the receiver.
	 */
	public int y;

	/**
	 * Constructs a new <code>Point</code> with x and y equals to
	 * <code>0</code>.
	 *
	 * @author IBM
	 * @version initial
	 */
	public Point() {
		this(0, 0);
	}

	/**
	 * Constructs a new <code>Point</code> with x and y equals to the x and y
	 * of <code>point</code>.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param point
	 *            the new coordinates of the receiver
	 */
	public Point(Point point) {
		this(point.x, point.y);
	}

	/**
	 * Constructs a new <code>Point</code> with specified x and y coordinates.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param x
	 *            the x coordinate of the receiver
	 * @param y
	 *            the y coordinate of the receiver
	 */
	public Point(int x, int y) {
		this.x = x;
		this.y = y;
	}

	/**
	 * Compares the argument to the receiver, and answers true if the argument
	 * is a <code>Point</code> and its x and y are equals to the receiver's x
	 * and y.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param object
	 *            Object the object to compare with the receiver
	 *
	 * @return <code>true</code> if the object is the same as this object
	 *         <code>false</code> otherwise
	 */
	public boolean equals(Object object) {
		if (object == this)
			return true;
		if (!(object instanceof Point))
			return false;
		Point point = (Point) object;
		return x == point.x && y == point.y;
	}

	/**
	 * Generate the hashCode() of this instance.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @return <code>int</code> hashCode value
	 *
	 */
	public int hashCode() {
		return x ^ y;
	}

	/**
	 * Answer the location of the receiver.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @return a Point which has the same x and y coordinates compared to the
	 *         receiver
	 */
	public Point getLocation() {
		return new Point(x, y);
	}

	/**
	 * Sets the x and y coordinates of the receiver with the specified values.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param x
	 *            the new x coordinate for the receiver
	 * @param y
	 *            the new y coordinate for the receiver
	 */
	public void move(int x, int y) {
		setLocation(x, y);
	}

	/**
	 * Sets the x and y coordinates of the receiver with the specified values.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param x
	 *            the new x coordinate for the receiver
	 * @param y
	 *            the new y coordinate for the receiver
	 */
	public void setLocation(int x, int y) {
		this.x = x;
		this.y = y;
	}

	/**
	 * Sets the x and y coordinates of the receiver with the x and y coordinates
	 * of the specified point.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param point
	 *            the new location for the receiver
	 */
	public void setLocation(Point point) {
		setLocation(point.x, point.y);
	}

	/**
	 * Answers a string containing a concise, human-readable description of the
	 * receiver's relevant information.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @return a java.lang.String which is a printable representation for the
	 *         receiver
	 */
	public String toString() {
		return getClass().getName() + "[x=" + x + ",y=" + y + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	/**
	 * Adds the specified x and y to the receiver's coordinates.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @param x
	 *            the value to be added to the receiver's x coordinate
	 * @param y
	 *            the value to be added to the receiver's y coordinate
	 */
	public void translate(int x, int y) {
		this.x += x;
		this.y += y;
	}

	public Object clone() {
		try {
			Point clone = (Point) super.clone();
			clone.x = x;
			clone.y = y;
			return clone;
		} catch (CloneNotSupportedException e) {
			return null;
		}
	}

}
