package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.Serializable;

/**
 * <code>Polygon</code> encapsulates a multi-sided 2D geometric
 * region.  The region is determined by x and y pairs which
 * represent the vertices of the polygon.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 * @see			Shape
 */
public class Polygon implements Shape, Serializable {

	static final long serialVersionUID = -6460061437900069969L;

	/**
	 * The x coordinates of the vertices
	 */
	public int xpoints[];

	/**
	 * The y coordinates of the vertices
	 */
	public int ypoints[];

	/**
	 * The number of vertices
	 */
	public int npoints;

	/**
	 * The bounds of the polygon. This is the smallest rectangle
	 * that includes all of the vertices of the polygon.
	 */
	protected Rectangle bounds;

	static int ARRAY_INCREMENT = 4;

	/**
	 * Constructs an empty <code>Polygon</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Polygon() {
		this(new int[ARRAY_INCREMENT], new int[ARRAY_INCREMENT], 0);
	}

	/**
	 * Constructs a new <code>Polygon</code> with the specified
	 * parameters.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Polygon(int[] xpoints, int[] ypoints, int npoints) {
		if (npoints < 0)
			throw new NegativeArraySizeException();
		if (npoints > Math.min(xpoints.length, ypoints.length))
			throw new IndexOutOfBoundsException();

		this.xpoints = (int[]) xpoints.clone();
		this.ypoints = (int[]) ypoints.clone();
		this.npoints = npoints;
	}

	/**
	 * Adds the specified x and y values to the receiver's
	 * x and y coordinates.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x	the value to be added to the receiver's x coordinate
	 * @param	y	the value to be added to the receiver's y coordinate
	 */
	public void translate(int x, int y) {
		for (int i=0; i<npoints; i++) {
			xpoints[i] += x;
			ypoints[i] += y;
		}

		if (bounds != null) {
			bounds.translate(x, y);
		}
	}

	/**
	 * Adds the point determined by the specified coordinates
	 * to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x	the x coordinate of the new vertice
	 * @param	y	the y coordinate of the new vertice
	 */
	public void addPoint(int x, int y) {
		if (npoints == xpoints.length) {
			int temp[] = new int[npoints + ARRAY_INCREMENT];
			System.arraycopy(xpoints, 0, temp, 0, npoints);
			xpoints = temp;
		}

		if (npoints == ypoints.length) {
			int temp[] = new int[npoints + ARRAY_INCREMENT];
			System.arraycopy(ypoints, 0, temp, 0, npoints);
			ypoints = temp;
		}

		xpoints[npoints] = x;
		ypoints[npoints] = y;
		npoints++;

		if (bounds != null) {
			bounds.add(x, y);
		}
	}

	/**
	 * Answers a <code>Rectangle</code> representing the bounds
	 * of the receiver.
	 * <p>
	 * The bounds of the receiver is the smallest rectangle that
	 * contains all vertices of the polygon.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Rectangle which is the receiver's bounds
	 */
	public Rectangle getBounds() {
		if (npoints == 0) return new Rectangle(0,0,0,0);

		if (bounds == null) {
			// The mins and maxs cannot be defaulted to be 0 since
			// the wrong bounds would be created if all the values
			// were greater than 0 or less than 0
			int minX = xpoints[0];
			int minY = ypoints[0];
			int maxX = xpoints[0];
			int maxY = ypoints[0];
			for (int i=1; i<npoints; i++) {
				int x = xpoints[i];
				int y = ypoints[i];
				if (x < minX) minX = x;
				if (y < minY) minY = y;
				if (x > maxX) maxX = x;
				if (y > maxY) maxY = y;
			}
			bounds = new Rectangle(minX, minY, maxX - minX, maxY - minY);
		}
		// Return a copy. Fix for bug 101289.
		return new Rectangle(bounds);
	}

	/**
	 * Answers a <code>Rectangle</code> representing the bounds
	 * of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Rectangle which is the receiver's bounds
	 *
	 * @deprecated	use <code>getBounds()</code> instead
	 */
	public Rectangle getBoundingBox() {
		return getBounds();
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	point	the point to be checked
	 *
	 * @return	<code>true</code> if the point is inside the receiver's boundary,
	 *			<code>false</code> otherwise
	 */
	public boolean contains(Point point) {
		return contains(point.x, point.y);
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x	the x coordinate of the point
	 * @param	y	the y coordinate of the point
	 *
	 * @return	<code>true</code> if the point is inside the receiver's boundary,
	 *			<code>false</code> otherwise
	 */
	public boolean contains(int x, int y) {
		if (!getBounds().contains(x, y))
			return false;

		boolean contains = false;
		for (int i = 0, j = npoints-1; i < npoints; j = i++) {
			if ((((ypoints[i] <= y) && (y < ypoints[j])) ||
			     ((ypoints[j] <= y) && (y < ypoints[i]))) &&
			    (x < (xpoints[j] - xpoints[i]) * (y - ypoints[i]) * 1.0 / (ypoints[j] - ypoints[i]) + xpoints[i]))
				contains = !contains;
		}
		return contains;
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	x	the x coordinate of the point
	 * @param	y	the y coordinate of the point
	 *
	 * @return		<code>true</code> if the point is inside the receiver's boundary,
	 *				<code>false</code> otherwise
	 *
	 * @deprecated	use <code>contains(int, int)</code> instead
	 */
	public boolean inside(int x, int y) {
		return contains(x, y);
	}

	/**
	 * Reset the receiver to represent an empty polygon.
	 * The point arrays remain unchanged.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void reset() {
		this.npoints = 0;
		invalidate();
	}

	/**
	 * Flush internal data.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void invalidate() {
		this.bounds = null;
	}
}
