package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import com.ibm.ugl.eswt.OS;

/**
 * A <code>PopupMenu</code> is the selectable user interface object
 * which is opened over another widget on the screen.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class PopupMenu extends Menu {

	static final long serialVersionUID = -4620452533522760060L;

	/**
	 * Constructs a new <code>PopupMenu</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public PopupMenu()
		throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>PopupMenu</code> with a label.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textLabel	the text label for the receiver
	 */
	public PopupMenu(String textLabel)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		label = textLabel;
	}

	void _addNotify() {
		if (parent != null) {
			if (parent instanceof MenuComponent) {
				super._addNotify();
				return;
			}
			Component hwParent = (Component)parent;
			if (hwParent.isLightweight())
				hwParent = hwParent.getNativeParent();
			menuPeer = OS.PopupMenu_New(hwParent.getWindow().peer);
			//addSWTListeners();
			OS.Widget_SetData(menuPeer, this);
			for (int i = 0; i < items.size(); i++)
				((MenuItem)(items.elementAt(i)))._addNotify();
		}
	}

	String classNonlocalizedName() {
		return "popupmenu"; //$NON-NLS-1$
	}

	/**
	 * Show the receiver on the screen, using the specified coordinates
	 * relative to the owning component specified.
	 * The owning component is not necessarily the receivers parent, but
	 * must be in the same hierarchy of parent/children.
	 *
	 * In order to show the receiver both the parent and the owner passed
	 * in must be visible on the screen.
	 * If the receiver is currently part of a menu (i.e. child of a
	 * <code>MenuComponent</code> then it can not be used as a popup.
	 * <p>
	 * This method is typically called due to an action by the user,
	 * for example pressing the right mouse button.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	owner	the widget the receiver is opened for
	 * @param	x		the x coordinate to open the receiver at
	 * @param	y		the y coordinate to open the receiver at
	 *
	 * @exceptions	IllegalArgumentException	thrown if parent is not a <code>Component</code>
	 */
	public void show(final Component owner, final int x, final int y) {
		if (!(parent instanceof Component)) {
			throw new IllegalArgumentException();
		}

		if (menuPeer == 0) {
			addNotify();
		}

		Point p = new Point(x,y);

		boolean adjustPoint = true; // default case is for non-Windows
		Window owningWindow = owner.getWindow();
		if (owningWindow == owner || owningWindow.getParent() instanceof Window) {
			// Window and its subclasses
			adjustPoint = false;
		}

		if (adjustPoint) {
			// Remap the owner-relative coordinates into the parent's coordinate space.
			p = owningWindow.translateCoords(owner, owningWindow, p);

			// Now adjust for the insets.  See Component.setPeerBounds() for more info.
			Insets insets = owningWindow.getInsetsImpl();
			p.x -= insets.left;
			p.y -= insets.top;
		}

		OS.PopupMenu_Show(menuPeer, p.x, p.y);
	}
}
