package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.Serializable;

/**
 * <code>Rectangle</code> represents a coordinate area
 * that uses int precision to store its data.
 * <p>
 * For instance, a rectangle is used to represent the bounds of components.
 * <p>
 * @author      IBM
 * @version		initial
 *
 * @see			java.io.Serializable
 */
public class Rectangle implements Serializable, Shape, Cloneable
{

	static final long serialVersionUID = -4345857070255674764L;

	/**
	 * The x coordinate of the receiver.
	 */
	public int x;

	/**
	 * The y coordinate of the receiver.
	 */
	public int y;

	/**
	 * The width of the receiver.
	 */
	public int width;

	/**
	 * The height of the receiver.
	 */
	public int height;

	/**
	 * Constructs a new <code>Rectangle</code> with x, y, width
	 * and height equals to <code>0</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public Rectangle() {
		this(0, 0, 0, 0);
	}

	/**
	 * Constructs a new <code>Rectangle</code> with x and y equals
	 * to <code>0</code> and the specified dimension.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	dimension	the size of the new rectangle
	 */
	public Rectangle(Dimension dimension) {
		this(dimension.width, dimension.height);
	}

	/**
	 * Constructs a new <code>Rectangle</code> with x and y equals
	 * to <code>0</code> and the specified dimension.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	width	the width of the new rectangle
	 * @param	height	the height of the new rectangle
	 */
	public Rectangle(int width, int height) {
		this(0, 0, width, height);
	}

	/**
	 * Constructs a new <code>Rectangle</code> with the specified
	 * x, y, width and height.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	x		the x coordinate of the new rectangle
	 * @param	y		the y coordinate of the new rectangle
	 * @param	width	the width of the new rectangle
	 * @param	height	the height of the new rectangle
	 */
	public Rectangle(int x, int y, int width, int height) {
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
	}

	/**
	 * Constructs a new <code>Rectangle</code> with the specified
	 * origin and empty size.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	point		the origin of the new rectangle
	 */
	public Rectangle(Point point) {
		this(point.x, point.y, 0, 0);
	}

	/**
	 * Constructs a new <code>Rectangle</code> with the specified
	 * origin and size.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	point		the origin of the new rectangle
	 * @param	dimension	the size of the new rectangle
	 */
	public Rectangle(Point point, Dimension dimension) {
		this(point.x, point.y, dimension.width, dimension.height);
	}

	/**
	 * Constructs a new <code>Rectangle</code> with x, y,
	 * width and height equals to the x, y, width and height
	 * of <code>Rectangle</code>.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	point	the new coordinates of the receiver
	 */
	public Rectangle(Rectangle rectangle) {
		this(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
	}

	/**
	 * Resizes the receiver so that the specified point is inside
	 * of the receiver's boundary.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x coordinate of the point to be added to the receiver
	 * @param	y	the y coordinate of the point to be added to the receiver
	 */
	public void add(int x, int y) {
		if (x < this.x) {
			// Safely calculate the new width so that
			// the value will overflow to 0 instead of
			// wrapping around into a negative value.  This is
			// the behavior that the PPro 1.1 TCK expects
			long newWidth = (long)width + this.x - x;
			if (newWidth > Integer.MAX_VALUE) width = 0;
			else width = (int)newWidth;

			this.x = x;
		} else if (x > this.x + width) {
			width = x - this.x;
		}

		if (y < this.y) {
			// Safely calculate the new height so that
			// the value will overflow to 0 instead of
			// wrapping around into a negative value.  This is
			// the behavior that the PPro 1.1 TCK expects
			long newHeight = (long)height + this.y - y;
			if (newHeight > Integer.MAX_VALUE) height = 0;
			else height = (int)newHeight;

			this.y = y;
		} else if (y > this.y + height) {
			height = y - this.y;
		}
	}

	/**
	 * Resizes the receiver so that the specified point is inside
	 * of the receiver's boundary.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	point	the point to be added to the receiver
	 */
	public void add(Point point) {
		add(point.x, point.y);
	}

	/**
	 * Adds a rectangle to this rectangle. The resulting rectangle is the
	 * union of the two rectangles.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	r	the rectangle
	 */
	public void add(Rectangle rectangle) {
		int left = x < rectangle.x ? x : rectangle.x;
		int top = y < rectangle.y ? y : rectangle.y;
		int corner1 = x + width;
		int corner2 = rectangle.x + rectangle.width;
		int right = corner1 > corner2 ? corner1 : corner2;
		corner1 = y + height;
		corner2 = rectangle.y + rectangle.height;
		int bottom = corner1 > corner2 ? corner1 : corner2;
		this.x = left;
		this.y = top;
		this.width = right-left;
		this.height = bottom-top;
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x coordinate of the point
	 * @param	y	the y coordinate of the point
	 *
	 * @return		<code>true</code> if the point is inside the receiver's boundary,
	 *				<code>false</code> otherwise
	 */
	public boolean contains(int x, int y) {
		if ((width <= 0) || (height <= 0)) return false;
		if ((x < this.x) || (y < this.y)) return false;
		long right= (long) this.x + width;
		long bottom= (long) this.y + height;
		return ((x < right) && (y < bottom));
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the point to be checked
	 *
	 * @return		<code>true</code> if the point is inside the receiver's boundary,
	 *				<code>false</code> otherwise
	 */
	public boolean contains(Point point) {
		return contains(point.x, point.y);
	}

	/**
	 * Answers whether the receiver contains the specified rectangle
	 * completely.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x		the x coordinate of the rectangle
	 * @param	y		the y coordinate of the rectangle
	 * @param	width	the width of the rectangle
	 * @param	height	the height of the rectangle
	 *
	 * @return		<code>true</code> if the rectangle is completely inside the
	 * 				receiver's boundary, <code>false</code> otherwise
	 *
	 * @since	JDK1.2
	 */
	public boolean contains(int x, int y, int width, int height) {
		if ((width <= 0) || (height <= 0)) return false;
		if ((this.width <= 0) || (this.height <= 0)) return false;
		return contains(x, y) && contains(x + width - 1, y + height - 1);
	}

	/**
	 * Answers whether the receiver contains the specified rectangle
	 * completely.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		rectangle	the point to be checked
	 *
	 * @return		<code>true</code> if the rectangle is completely inside the
	 * 				receiver's boundary, <code>false</code> otherwise
	 *
	 * @since		JDK1.2
	 */
	public boolean contains(Rectangle rectangle) {
		return contains(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
	}

	/**
	 * Compares the argument to the receiver, and answers true
	 * if the argument is a <code>Rectangle</code> and its x,
	 * y, width and height are equals to the receiver's x, y,
	 * width and height.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		object	Object	the object to compare with this dimension
	 *
	 * @return		<code>true</code> if the object is the same as this object
	 *				<code>false</code> otherwise
	 */
	public boolean equals (Object object) {
		if (object == this) return true;
		if (!(object instanceof Rectangle)) return false;
		Rectangle rectangle = (Rectangle)object;
		return x == rectangle.x && y == rectangle.y &&
			width == rectangle.width && height == rectangle.height;
	}

	/**
	 * Generate the hashCode() of this instance.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>int</code> hashCode value
	 *
	 */
	public int hashCode() {
		return x ^ y ^ width ^ height;
	}

	/**
	 * Answers a <code>Rectangle</code> representing the bounds
	 * of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.awt.Rectangle which is the receiver's bounds
	 */
	public Rectangle getBounds() {
		return new Rectangle(x, y, width, height);
	}

	/**
	 * Answer the location of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>Point</code> which is the origin of the receiver
	 */
	public Point getLocation() {
		return new Point(x, y);
	}

	/**
	 * Answer the dimension of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>Dimension</code> which is the size of the receiver
	 */
	public Dimension getSize() {
		return new Dimension(width, height);
	}

	/**
	 * Resizes the receiver so that the resulting by the specified amounts.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	horizontal
	 * @param	vertical
	 */
	public void grow(int horizontal, int vertical) {
		x = x - horizontal;
		y = y - vertical;
		width += 2 * horizontal;
		height += 2 * vertical;
	}

	/**
	 * Answers whether the receiver contains the specified point.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x coordinate of the point
	 * @param	y	the y coordinate of the point
	 *
	 * @return		<code>true</code> if the point is inside the receiver's boundary,
	 *				<code>false</code> otherwise
	 *
	 * @deprecated	use <code>contains(int, int)</code> instead
	 */
	public boolean inside(int x, int y) {
		return contains(x, y);
	}

	/**
	 * Answers a new rectangle which is the intersection of the
	 * receiver and the specified rectangle.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	rectangle	the rectangle to be intersected with the receiver
	 */
	public Rectangle intersection(Rectangle rectangle) {
		if (this == rectangle) return new Rectangle (x, y, width, height);

		int left = x > rectangle.x ? x : rectangle.x;
		int top = y > rectangle.y ? y : rectangle.y;
		long right= Math.min((long) x + width, (long) rectangle.x + rectangle.width);
		long bottom= Math.min((long) y + height, (long) rectangle.y + rectangle.height);

		return new Rectangle (
			(left > right) ? 0 : left,
			(top > bottom) ? 0 : top,
			(right < left) ? 0 : (int) (Math.min(right - left, Integer.MAX_VALUE)),
			(bottom < top) ? 0 : (int) (Math.min(bottom - top, Integer.MAX_VALUE)));
	}

	/**
	 * Answers whether the receiver intersects the specified rectangle.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		rectangle	the rectangle to be checked
	 *
	 * @return		<code>true</code> if the rectangle intersects the receiver,
	 *				<code>false</code> otherwise
	 */
	public boolean intersects(Rectangle rectangle) {
		return (rectangle == this) || (
			(rectangle.x < ((long)x + width)) && (rectangle.y < ((long)y + height)) &&
			(((long)rectangle.x + rectangle.width) > x) && (((long)rectangle.y + rectangle.height) > y));
	}

	/**
	 * Answers whether the receiver is empty.  The receiver is
	 * empty if its width and height are both smaller or equals
	 * to <code>0</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is empty,
	 *				<code>false</code> otherwise
	 */
	public boolean isEmpty() {
		return width <= 0 || height <= 0;
	}

	/**
	 * Sets the x and y coordinates of the receiver with the
	 * specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x	the new x coordinate for the receiver
	 * @param		y	the new y coordinate for the receiver
	 *
	 * @deprecated	use <code>setLocation(int, int)</code> instead
	 */
	public void move(int x, int y) {
		setLocation(x, y);
	}

	/**
	 * Sets x, y, width and height of the receiver with the
	 * specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the new x coordinate for the receiver
	 * @param		y		the new y coordinate for the receiver
	 * @param		width	the new width for the receiver
	 * @param		hieght	the new height for the receiver
	 *
	 * @deprecated	use <code>setBounds(int, int, int, int)</code> instead
	 */
	public void reshape(int x, int y, int width, int height) {
		setBounds(x, y, width, height);
	}

	/**
	 * Sets width and height of the receiver with the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		width	the new width for the receiver
	 * @param		hieght	the new height for the receiver
	 *
	 * @deprecated	use <code>setSize(int, int)</code> instead
	 */
	public void resize(int width, int height) {
		setSize(width, height);
	}

	/**
	 * Sets x, y, width and height of the receiver with the
	 * specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x		the new x coordinate for the receiver
	 * @param		y		the new y coordinate for the receiver
	 * @param		width	the new width for the receiver
	 * @param		hieght	the new height for the receiver
	 */
	public void setBounds(int x, int y, int width, int height) {
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
	}

	/**
	 * Sets x, y, width and height of the receiver with the
	 * specified rectangle.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		rectangle the new x, y, width and height for the receiver
	 */
	public void setBounds(Rectangle rectangle) {
		setBounds(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
	}

	/**
	 * Sets the x and y coordinates of the receiver with the
	 * specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x	the new x coordinate for the receiver
	 * @param		y	the new y coordinate for the receiver
	 */
	public void setLocation(int x, int y) {
		this.x = x;
		this.y = y;
	}

	/**
	 * Sets the x and y coordinates of the receiver with the
	 * specified point.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		point	the new x and y coordinate for the receiver
	 */
	public void setLocation(Point point) {
		setLocation(point.x, point.y);
	}

	/**
	 * Sets width and height of the receiver with the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		width	the new width for the receiver
	 * @param		hieght	the new height for the receiver
	 */
	public void setSize(int width, int height) {
		this.width = width;
		this.height = height;
	}

	/**
	 * Sets width and height of the receiver with the specified dimension.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		dimension	the new size for the receiver
	 */
	public void setSize(Dimension dimension) {
		setSize(dimension.width, dimension.height);
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		return getClass().getName() + "[x=" + x + ",y=" + y + ",width=" + width + ",height=" + height + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	}

	/**
	 * Adds the specified x and y values to the receiver's x and y
	 * coordinates.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the value to be added to the receiver's x coordinate
	 * @param	y	the value to be added to the receiver's y coordinate
	 */
	public void translate(int x, int y) {
		this.x += x;
		this.y += y;
	}

	/**
	 * Answers a new rectangle which is the union of the
	 * receiver and the specified rectangle.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	rectangle	the rectangle to be intersected with the receiver
	 */
	public Rectangle union(Rectangle rectangle) {
		int left = x < rectangle.x ? x : rectangle.x;
		int top = y < rectangle.y ? y : rectangle.y;
		int corner1 = x + width;
		int corner2 = rectangle.x + rectangle.width;
		int right = corner1 > corner2 ? corner1 : corner2;
		corner1 = y + height;
		corner2 = rectangle.y + rectangle.height;
		int bottom = corner1 > corner2 ? corner1 : corner2;
		return new Rectangle (left, top, right - left, bottom - top);
	}

	public Object clone() {
		try {
			Rectangle clone = (Rectangle) super.clone();
			clone.x = x;
			clone.y = y;
			clone.width = width;
			clone.height = height;
			return clone;
		} catch (CloneNotSupportedException e) {
			return null;
		}
	}

}
