package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */
import java.awt.event.MouseWheelEvent;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.awt.event.ComponentEvent;
import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.metal.widgets.Display;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>ScrollPane</code> is a container that might have scrollbars,
 * and has exactly one child component.
 *
 * @author      IBM
 * @version		initial
 */
public class ScrollPane extends Container {

	static final long serialVersionUID = 7956609840827222915L;

//	   The scrollbar display policy of the receiver
	int scrollbarDisplayPolicy;

//	   The vertical scrollbar of the receiver
	transient
	ScrollPaneAdjustable vAdjustable;

//	   The horizontal scrollbar of the receiver
	transient
	ScrollPaneAdjustable hAdjustable;

	boolean wheelScrollingEnabled;

	/**
	 * Constants that specify how scrollbars are displayed.
	 */
	public static final int SCROLLBARS_AS_NEEDED = 0;
	public static final int SCROLLBARS_ALWAYS = 1;
	public static final int SCROLLBARS_NEVER = 2;

	/**
	 * Constructs a new <code>ScrollPane</code> with scrollbars
	 * display policy equals SCROLLBARS_AS_NEEDED.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public ScrollPane()
		throws HeadlessException
	{
		this(SCROLLBARS_AS_NEEDED);
	}

	/**
	 * Constructs a new <code>ScrollPane</code> with the specified
	 * scrollbars display policy.
	 *
	 * @author      IBM
	 * @version 	initial
	 */
	public ScrollPane(int scrollbarDisplayPolicy)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		switch (scrollbarDisplayPolicy) {
			case SCROLLBARS_ALWAYS:
			case SCROLLBARS_AS_NEEDED:
			case SCROLLBARS_NEVER:
				break;
			default:
				throw new IllegalArgumentException();
		}
		this.scrollbarDisplayPolicy = scrollbarDisplayPolicy;

		this.hAdjustable = new ScrollPaneAdjustable(this, Adjustable.HORIZONTAL);
		this.vAdjustable = new ScrollPaneAdjustable(this, Adjustable.VERTICAL);

		// See ScrollPane spec Javadoc
		this.width = 100;
		this.height = 100;

		// Original behavior
		this.width = getMinimumSize().width;
		this.height = getMinimumSize().height;
		this.wheelScrollingEnabled = true;
	}

	/**
	 * Adds the specified <code>Component</code> at the specified
	 * position.  The specified constraints object is used for layout
	 * purposes.
	 * <p>
	 * If <code>position</code> is <code>-1</code>, the component
	 * is added to the end of the receiver.
	 * <p>
	 * The order a component is added to the receiver's list
	 * determines its front-to-back z-order.
	 * <p>
	 * A <code>ScrollPane</code> can have only one component at a time.  If
	 * the receiver already has a child, this child is removed before adding
	 * the specified component.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	newComponent	the component to be added to the receiver
	 * @param	constraints		the object representing layout constraints
	 * @param	position		the position of the new component
	 *
	 * @throws	IllegalArgumentException	if the specified position is
	 *										out of range and different from <code>-1</code>
	 */
	protected final void addImpl(Component comp, Object constraints, int position) {
		synchronized (getTreeLock()) {
			if (getComponentCount() > 0) {
				remove(0);
			}

			super.addImpl(comp, constraints, position);
		}
	}

	Insets getInsetsImpl() {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				return new Insets(0, 0, 0, 0);
			}
			// always go to the peer layer and get the fresh insets since they
			// may change if SCROLLBARS_AS_NEEDED is used.
			int[] insets = OS.Scrollable_GetInsets(peer);
			return new java.awt.Insets(
				insets[OS.INDEX_TOP],
				insets[OS.INDEX_LEFT],
				insets[OS.INDEX_BOTTOM],
				insets[OS.INDEX_RIGHT]);
		} finally {
			releasePeerReadLock();
		}
	}

	int _createPeer(Container hwParent) {
		return OS.Composite_New(hwParent.peer, widgetStyle());
	}

	void _setupPeer() {
		if (scrollbarDisplayPolicy != ScrollPane.SCROLLBARS_NEVER) {

			int horizontalPeer = OS.Scrollable_GetHorizontalBar(peer);
			int verticalPeer = OS.Scrollable_GetVerticalBar(peer);

			((ScrollPaneAdjustable)getHAdjustable()).setPeer(horizontalPeer);
			((ScrollPaneAdjustable)getVAdjustable()).setPeer(verticalPeer);

			if (scrollbarDisplayPolicy != ScrollPane.SCROLLBARS_ALWAYS) {
				OS.Scrollable_SetAdjustableVisible(peer, horizontalPeer, false);
				OS.Scrollable_SetAdjustableVisible(peer, verticalPeer, false);
			}
			else {
				OS.Scrollable_SetAdjustableEnabled(peer, horizontalPeer, false);
				OS.Scrollable_SetAdjustableEnabled(peer, verticalPeer, false);
			}
		}

		super._setupPeer();
	}

	int widgetStyle() {
		return OS.STYLE_BORDER | ((getScrollbarDisplayPolicy() == SCROLLBARS_NEVER) ? 0 : (OS.STYLE_H_SCROLL | OS.STYLE_V_SCROLL));
	}

	/**
	 * Answer whether the receiver can be used as a canvas
	 */
	boolean isPaintable() {
		return true;
	}

	public boolean isLightweight() {
		return false;
	}

	public boolean isOpaque() {
		return true;
	}

	/**
	 * Answers the receiver's scrollbar display policy. The return
	 * value is one of:
	 * <code>SCROLLBARS_ALWAYS</code>
	 * <code>SCROLLBARS_AS_NEEDED</code>
	 * <code>SCROLLBARS_NEVER</code>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> which is the scrollbars display policy of the receiver
	 */
	public int getScrollbarDisplayPolicy() {
		return scrollbarDisplayPolicy;
	}

	/**
	 * Answers a <code>Dimension</code> representing the size of the
	 * client area of the receiver.  The client area is the area of
	 * the receiver excluding the borders and scrollbars.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Dimension</code> which represents the client area size
	 */
	public Dimension getViewportSize() {
		if (peer == 0) {
			// Match RI behavior
			return getSize();
		}

		Rectangle clientArea = getBounds();
		Insets insets = getInsets();
		clientArea.width -= insets.left + insets.right;
		clientArea.height -= insets.top + insets.bottom;
		return new Dimension(clientArea.width, clientArea.height);
	}

	Dimension getMinimumSizeImpl() {
		Dimension prefSize = getViewportSize();
		Insets insets = getInsets();
		prefSize.width  += insets.left + insets.right;
		prefSize.height += insets.top + insets.bottom;
		return prefSize;
	}

	/**
	 * Answers the height of the horizontal scrollbar.  The return value
	 * is <code>0</code>, if the receiver is not displayable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> which represents the height of the horizontal scrollbar
	 */
	public int getHScrollbarHeight() {
		acquirePeerReadLock();
		try {
			int scrollBar = hAdjustable.peer;
			if ((scrollBar != 0) && (peer != 0)) {
				int rect[] = OS.Scrollable_GetAdjustableBounds(peer, scrollBar);
				return rect[OS.INDEX_HEIGHT];
			} else
				return 0;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the width of the vertical scrollbar.  The return value
	 * is <code>0</code>, if the receiver is not displayable.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an <code>int</code> which represents the width of the vertical scrollbar
	 */
	public int getVScrollbarWidth() {
		acquirePeerReadLock();
		try {
			int scrollBar = vAdjustable.peer;
			if ((scrollBar != 0) && (peer != 0)) {
				int rect[] = OS.Scrollable_GetAdjustableBounds(peer, scrollBar);
				return rect[OS.INDEX_WIDTH];
			} else
				return 0;
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers a <code>Adjustable</code> representing the vertical
	 * scrollbar of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Adjustable</code> which represents the vertical scrollbar
	 */
	public Adjustable getVAdjustable() {
		return vAdjustable;
	}

	/**
	 * Answers a <code>Adjustable</code> representing the horizontal
	 * scrollbar of the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Adjustable</code> which represents the horizontal scrollbar
	 */
	public Adjustable getHAdjustable() {
		return hAdjustable;
	}

	/**
	 * Scrolls the receiver's client area to the specified values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the position of the horizontal scrollbar
	 * @param	y	the position of the vertical scrollbar
	 */
	public void setScrollPosition(int x, int y) {
//		synchronized (this) {
			if (getComponentCount() == 0) {
				throw new NullPointerException("child is null"); //$NON-NLS-1$
			}

			// From spec:
			//   x and y are forced to set within legal boundaries.
			//   Legal boundaries are >=0 and <=component size - viewport size
			//   These are so the view stays within the child's borders
			// This is true regardless of the scrollbar display policy
			Component component = getComponent(0);
			Dimension prefSize = component.getPreferredSize();
			Dimension viewSize = getViewportSize();

			int posX = Math.max(Math.min(x,prefSize.width - viewSize.width),0);
			int posY = Math.max(Math.min(y,prefSize.height - viewSize.height),0);

			getHAdjustable().setValue(posX);
			getVAdjustable().setValue(posY);
//		}
	}

	/**
	 * Scrolls the receiver's client area to the specified point.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	point	a point representing the horizontal and vertical
	 *					scrollbar positions
	 */
	public void setScrollPosition(Point point) {
		setScrollPosition(point.x, point.y);
	}

	/**
	 * Answers the scroll position of the receiver.
	 *
	 * @return a <code>Point</code> which represents the scroll position of the viewport
	 */
	public Point getScrollPosition() {
		if (getComponentCount() == 0) throw new NullPointerException("child is null"); //$NON-NLS-1$
		return new Point(getHAdjustable().getValue(), getVAdjustable().getValue());
	}

	/**
	 * A <code>ScrollPane</code> lays out its child internally.  The
	 * specified layout manager is ignored.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	layoutManager	the layout manager for the receiver
	 */
	public final void setLayout(LayoutManager layoutManager) {
		throw new java.awt.AWTError("ScrollPane controls layout"); //$NON-NLS-1$
	}

	/**
	 * Ensure that the receiver child is at its preferred size.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void doLayout() {

		if (getComponentCount() == 0) {
			return;
		}

		acquirePeerReadLock();
		try {
			if (peer == 0) {
				// Match RI behavior.  Set the child's size
				// to its preferred size, without shrinking it.
				Component child = getComponent(0);
				Dimension childSize = child.getPreferredSize();
				boolean changed = false;
				if (childSize.width < this.width) {
					childSize.width = this.width;
					changed = true;
				}
				if (childSize.height < this.height) {
					childSize.height = this.height;
					changed = true;
				}
				if (changed) {
					child.setSize(childSize.width,childSize.height);
				}
				return;
			}

			Component component = getComponent(0);
			Dimension prefSize = component.getPreferredSize();

			setRedraw(false);

			// clientAreaSize cannot be determined until the visiblity of
			// the adjustables has been set since it affects the insets.
			Dimension clientAreaSize;

			int horizontalPeer, verticalPeer;
			int policy = getScrollbarDisplayPolicy();
			switch (policy) {

				case SCROLLBARS_NEVER:
					// The adjustables have no peer in this case.
					clientAreaSize = getViewportSize();
					break;

				case SCROLLBARS_ALWAYS:
					horizontalPeer = OS.Scrollable_GetHorizontalBar(peer);
					verticalPeer = OS.Scrollable_GetVerticalBar(peer);

					OS.Scrollable_SetAdjustableVisible(peer, horizontalPeer, true);
					OS.Scrollable_SetAdjustableVisible(peer, verticalPeer, true);
					clientAreaSize = getViewportSize();

					// Enable the scrollbars if necessary
					OS.Scrollable_SetAdjustableEnabled(peer, horizontalPeer, prefSize.width > clientAreaSize.width);
					OS.Scrollable_SetAdjustableEnabled(peer, verticalPeer, prefSize.height > clientAreaSize.height);
					break;

				case SCROLLBARS_AS_NEEDED:
				default:
					horizontalPeer = OS.Scrollable_GetHorizontalBar(peer);
					verticalPeer = OS.Scrollable_GetVerticalBar(peer);

					OS.Scrollable_SetAdjustableVisible(peer, horizontalPeer, false);
					OS.Scrollable_SetAdjustableVisible(peer, verticalPeer, false);
					clientAreaSize = getViewportSize();

					// Check if the horizontal scrollbar is needed.
					if (prefSize.width > clientAreaSize.width) {
						//hScroll.setVisibleAmount(clientAreaSize.width);
						OS.Scrollable_SetAdjustableVisible(peer, horizontalPeer, true);
						clientAreaSize = getViewportSize();
					}
					// Check if the vertical scrollbar is needed.
					if (prefSize.height > clientAreaSize.height) {
						//vScroll.setVisibleAmount(clientAreaSize.height);
						OS.Scrollable_SetAdjustableVisible(peer, verticalPeer, true);
						clientAreaSize = getViewportSize();
						// Check if the horizontal scrollbar is needed again since the
						// client area changes after showing the vertical scrollbar.
						if (prefSize.width > clientAreaSize.width) {
							//hScroll.setVisibleAmount(clientAreaSize.width);
							OS.Scrollable_SetAdjustableVisible(peer, horizontalPeer, true);
							clientAreaSize = getViewportSize();
						}
					}
			}

			Insets scrollPaneInsets = getInsets();
			component.setBounds(
				- hAdjustable.getValue() + scrollPaneInsets.left,
				- vAdjustable.getValue() + scrollPaneInsets.top,
				Math.max(prefSize.width, clientAreaSize.width),
				Math.max(prefSize.height, clientAreaSize.height));

			// reset Scroll Position, if required
			Point scrollPosition = getScrollPosition();
			switch (policy) {

				case SCROLLBARS_NEVER:
					break;

				case SCROLLBARS_ALWAYS:
				case SCROLLBARS_AS_NEEDED:
				default:
					// check horizontal position
					if (clientAreaSize.width > prefSize.width) {
						scrollPosition.x = 0;
					} else if (prefSize.width - scrollPosition.x < clientAreaSize.width) {
						scrollPosition.x = Math.max(0,clientAreaSize.width + scrollPosition.x - prefSize.width);
					}

					// check vertical position
					if (clientAreaSize.height > prefSize.height) {
						scrollPosition.y = 0;
					} else if (prefSize.height - scrollPosition.y < clientAreaSize.height) {
						scrollPosition.y = Math.max(0,clientAreaSize.height + scrollPosition.y - prefSize.height);
					}
			}

			// update horizontal adjustments
			hAdjustable.setValues(
				scrollPosition.x,
				Math.min(clientAreaSize.width, prefSize.width),
				hAdjustable.getMinimum(), // no change
				prefSize.width);
			hAdjustable.setBlockIncrement(clientAreaSize.width);

			// update Vertical adjustments
			vAdjustable.setValues(
				scrollPosition.y,
				Math.min(clientAreaSize.height, prefSize.height),
				vAdjustable.getMinimum(), // no change
				prefSize.height);
			vAdjustable.setBlockIncrement(clientAreaSize.height);

			setRedraw(true);

		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	public String paramString() {
		return super.paramString() + ",scrollbarsPolicy=" + getScrollbarDisplayPolicy(); //$NON-NLS-1$
	}

	String classNonlocalizedName() {
		return "scrollpane"; //$NON-NLS-1$
	}

	protected void processMouseWheelEvent(MouseWheelEvent mouseWheelEvent) {
		if (isWheelScrollingEnabled()) {
			super.processMouseWheelEvent(mouseWheelEvent);
		}
	}

	public void setWheelScrollingEnabled(boolean wheelScrollingEnabled) {
		this.wheelScrollingEnabled = wheelScrollingEnabled;
	}

	public boolean isWheelScrollingEnabled() {
		return wheelScrollingEnabled;
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException, HeadlessException {
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		this.hAdjustable = new ScrollPaneAdjustable(this, Adjustable.HORIZONTAL);
		this.vAdjustable = new ScrollPaneAdjustable(this, Adjustable.VERTICAL);
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}

	void scrollYTo(int newY) {
		if (getComponentCount() == 0) return;
		Component component = getComponent(0);
		int currentY = component.getY();
		if (newY == currentY) return;
		if (!component.isLightweight()) {
			//component.setLocation(component.getX(), newY);
			component.acquirePeerReadLock();
			try {
				component.y = newY;
				component._setPeerBounds(component.getX(), newY, component.getWidth(), component.getHeight());
			} finally {
				component.releasePeerReadLock();
			}
		} else {
			GraphicsImpl graphics = (GraphicsImpl)getGraphics();
			if (graphics == null) return;

			int imageHandler = 0;
			acquirePeerReadLock();
			try {
				int offset = newY - currentY;
				Insets insets = getInsets();
				Dimension size = getSize();
				int insideWidth = size.width - insets.left - insets.right;
				int insideHeight = size.height - insets.top - insets.bottom;
				Rectangle copiedArea, repaintedArea;
				Point target;
				if (offset < 0) {
					copiedArea = new Rectangle(0, -offset, insideWidth, insideHeight+offset);
					target = new Point(0, 0);
					repaintedArea = new Rectangle(0, insideHeight+offset, insideWidth, -offset);
				} else {
					copiedArea = new Rectangle(0, 0, insideWidth, insideHeight-offset);
					target = new Point(0, offset);
					repaintedArea = new Rectangle(0, 0, insideWidth, offset);
				}

				Display currentDisplay = Util.getDisplay();
				synchronized (graphics.peerHandle) {
					int graphicsHandler = ((GraphicsImpl)graphics).getPeer();//OS.Control_NewGraphics(peer);
					// 1- Copy/Paste the area still visible after the scroll
					if (copiedArea.height > 0) {
						imageHandler = OS.Image_NewFromSize(currentDisplay.fPeerHandle, copiedArea.width, copiedArea.height, Color.black.getRGB());
						OS.Control_CopyArea(peer, imageHandler, copiedArea.x, copiedArea.y, copiedArea.width, copiedArea.height);
						OS.Graphics_DrawImage(graphicsHandler, imageHandler, target.x, target.y);
					}
					// 2- Paint the area which was invisble before the scroll
					Color bgColor = getBackground();
					if (bgColor != null) {
						int bgColorHandler = bgColor.getPeer();
						OS.Graphics_SetBackground(graphicsHandler, bgColorHandler);
						OS.Graphics_FillRectangle(graphicsHandler, repaintedArea.x, repaintedArea.y, repaintedArea.width, repaintedArea.height);
					}
				}

				// At this point with have to use the java.awt.Graphics to let it handle the offsets of the components...
				component.y = newY;
				graphics.setClip(repaintedArea.x+insets.left, repaintedArea.y+insets.top, repaintedArea.width, repaintedArea.height);
				Rectangle compBounds = component.getBounds();
				Graphics childGraphics = graphics.create(compBounds.x, compBounds.y, compBounds.width, compBounds.height);
				if (childGraphics != null) {
					if (!childGraphics.getClipBounds().isEmpty()) {
						component.paintAll(childGraphics);
					}
					childGraphics.dispose();
				}

				// Refresh the position of all heayweight children
				if (component instanceof Container) {
					updateHeavyweightChildren((Container)component);
				}
			} finally {
				if (imageHandler != 0) OS.Image_Dispose(imageHandler);
				graphics.dispose();
				releasePeerReadLock();
			}
		}
		component.postEvent(new ComponentEvent(component, ComponentEvent.COMPONENT_MOVED));
		invalidate();

	}
	void scrollXTo(int newX) {
		if (getComponentCount() == 0) return;
		Component component = getComponent(0);
		int currentX = component.getX();
		if (newX == currentX) return;
		if (!component.isLightweight()) {
			//component.setLocation(newX, component.getY());
			component.acquirePeerReadLock();
			try {
				component.x = newX;
				component._setPeerBounds(newX, component.getY(), component.getWidth(), component.getHeight());
			} finally {
				component.releasePeerReadLock();
			}
		} else {
			GraphicsImpl graphics = (GraphicsImpl)getGraphics();
			if (graphics == null) return;

			int imageHandler = 0;
			acquirePeerReadLock();
			try {
				int offset = newX - currentX;
				Insets insets = getInsets();
				Dimension size = getSize();
				int insideWidth = size.width - insets.left - insets.right;
				int insideHeight = size.height - insets.top - insets.bottom;
				Rectangle copiedArea, repaintedArea;
				Point target;
				if (offset < 0) {
					copiedArea = new Rectangle(-offset, 0, insideWidth+offset, insideHeight);
					target = new Point(0, 0);
					repaintedArea = new Rectangle(insideWidth+offset, 0, -offset, insideWidth);
				} else {
					copiedArea = new Rectangle(0, 0, insideWidth-offset, insideHeight);
					target = new Point(offset, 0);
					repaintedArea = new Rectangle(0, 0, offset, insideHeight);
				}

				Display currentDisplay = Util.getDisplay();
				synchronized (graphics.peerHandle) {
					int graphicsHandler = ((GraphicsImpl)graphics).getPeer();
					// 1- Copy/Paste the area still visible after the scroll
					if (copiedArea.width > 0) {
						imageHandler = OS.Image_NewFromSize(currentDisplay.fPeerHandle, copiedArea.width, copiedArea.height, Color.black.getRGB());
						OS.Control_CopyArea(peer, imageHandler, copiedArea.x, copiedArea.y, copiedArea.width, copiedArea.height);
						OS.Graphics_DrawImage(graphicsHandler, imageHandler, target.x, target.y);
					}
					// 2- Paint the area which was invisble before the scroll
					Color bgColor = getBackground();
					if (bgColor != null) {
						int bgColorHandler = bgColor.getPeer();
						OS.Graphics_SetBackground(graphicsHandler, bgColorHandler);
						OS.Graphics_FillRectangle(graphicsHandler, repaintedArea.x, repaintedArea.y, repaintedArea.width, repaintedArea.height);
					}
				}

				component.x = newX;
				// At this point with have to use the java.awt.Graphics to let it handle the offsets of the components...
				graphics.setClip(repaintedArea.x+insets.left, repaintedArea.y+insets.top, repaintedArea.width, repaintedArea.height);
				Rectangle compBounds = component.getBounds();
				Graphics childGraphics = graphics.create(compBounds.x, compBounds.y, compBounds.width, compBounds.height);
				if (childGraphics != null) {
					if (!childGraphics.getClipBounds().isEmpty()) {
						component.paintAll(childGraphics);
					}
					childGraphics.dispose();
				}

				// Refresh the position of all heayweight children
				if (component instanceof Container) {
					updateHeavyweightChildren((Container)component);
				}
			} finally {
				if (imageHandler != 0) OS.Image_Dispose(imageHandler);
				graphics.dispose();
				releasePeerReadLock();
			}
		}
		component.postEvent(new ComponentEvent(component, ComponentEvent.COMPONENT_MOVED));
		invalidate();
	}

	private void updateHeavyweightChildren(Container cont) {
		for (int i=0;i<cont.ncomponents;i++) {
			Component c = cont.component[i];
			if (c.isLightweight() == false) {
				c.acquirePeerReadLock();
				try {
					c._setPeerBounds(c.x, c.y, c.getWidth(), c.getHeight());
				} finally {
					c.releasePeerReadLock();
				}

				// If the child is heavyweight then we are
				// done with it, even it if is a container
				continue;
			} else
			if (c instanceof Container) {
				updateHeavyweightChildren((Container)c);
			}
		}
	}
}
