package java.awt;

import java.awt.event.*;

import com.ibm.ugl.eswt.OS;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * ScrollPaneAdjustable is a special type of Scrollbar which
 * is only used on a ScrollPane.  It overrides three methods
 * of Scrollbar (setMinimum, setMaximum, setVisibleAmount)
 * to throw AWTError.  It is not valid to call these methods
 * because doing so would put the ScrollPane into an invalid
 * state.
 */
class ScrollPaneAdjustable extends Scrollbar {

	private transient ScrollPane scrollPane;

	static final long serialVersionUID = -3359745691033257079L;

	ScrollPaneAdjustable(ScrollPane scrollPane, int orientation) {
		super(orientation);
		this.scrollPane = scrollPane;
	}

	/*
	 * The original ScrollPaneAdjustable was not a Component, so
	 * it did not need any addNotify behavior.
	 */
	void _addNotify() {
	}

	int getPeer() {
		return peer;
	}

	void setPeer(int peer) {
		this.peer = peer;
		this.isDisplayable = true;
		OS.Widget_SetData(peer, this);

		// Synchronize our state with the peer
		syncWithPeer(peer);
		// Two sets missing from the inherited code...
		OS.Adjustable_SetUnitIncrement(peer, lineIncrement);
		OS.Adjustable_SetBlockIncrement(peer, pageIncrement);
	}

	void syncWithPeer(int peer) {
		int peerValue = OS.Adjustable_GetValue(peer);
		int peerVisible = OS.Adjustable_GetVisibleAmount(peer);
		int peerMinimum = OS.Adjustable_GetMinimum(peer);
		int peerMaximum = OS.Adjustable_GetMaximum(peer);
		setValues(peerValue,peerVisible,peerMinimum,peerMaximum);
	}

	/**
	 * Answer the visible amount value for the receiver.
	 * This is the proportion of the receiver that is
	 * visibly represented.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the visible amount
	 */
	public int getVisibleAmount() {
		if (peer == 0) {
			// Tested by TCK.
			return 0;
		}
		return OS.Adjustable_GetVisibleAmount(peer);
	}

	/**
	 * Set the value for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	value	the value
	 */
	public void setValue(int newValue) {

		if (setValueField(newValue)) {
			// Update the peer.
			if (scrollPane.getScrollbarDisplayPolicy() != ScrollPane.SCROLLBARS_NEVER) {
				if (peer == 0) return;
				OS.Adjustable_SetValue(peer, value);
			}

			// Post an AdjustmentEvent for our listeners.
            // Scroll pane layout is updated asynchronously in dispatchEventImpl.
			AdjustmentEvent event = new AdjustmentEvent(
				this,
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
				AdjustmentEvent.TRACK,
				value);

			postEvent(event);
		}
	}

	/**
	 * The user changed the peer's value.
	 * @see EventPoster#postAdjustmentEvent(Object, int, int)
	 */
	void peerValueChanged(int newValue) {
        setValueField(newValue);
	}

	/*
	 * Update the ScrollPane.
	 */
	private void updateScrollPane() {
		Insets scrollPaneInsets = scrollPane.getInsets();
		if (getOrientation() == Scrollbar.HORIZONTAL) {
			scrollPane.scrollXTo(-value + scrollPaneInsets.left);
		} else {
			scrollPane.scrollYTo(-value + scrollPaneInsets.top);
		}

	}

	/*
	 * Set value after bounds checking the new value.
	 * Return true if the value changed, false otherwise.
	 */
	private boolean setValueField(int newValue) {
		// Bounds checking.
		if (newValue < minimum) newValue = minimum;
		if (newValue > maximum) newValue = maximum;

		if (value == newValue) return false;

		value = newValue;
		return true;
	}

	/* (non-Javadoc)
	 * @see java.awt.Component#dispatchEventImpl(java.awt.AWTEvent)
	 */
	void dispatchEventImpl(AWTEvent awtEvent) {
		if (awtEvent instanceof AdjustmentEvent) {
            // update scroll pane layout asynchronously. Fixes bug 115169.
            updateScrollPane();
			if (adjustmentListener != null) {
				adjustmentListener.adjustmentValueChanged((AdjustmentEvent)awtEvent);
			}
		}
		else {
			super.dispatchEventImpl(awtEvent);
		}
	}

	/**
	 * Set the minimum value for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	minimum	the minimum
	 */
	public void setMinimum(int minimum) {
		throw new AWTError("invalid method"); //$NON-NLS-1$
	}

	/**
	 * Set the maximum value for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	maximum	the maximum
	 */
	public void setMaximum(int maximum) {
		throw new AWTError("invalid method"); //$NON-NLS-1$
	}

	/**
	 * Set the visible amount for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	visibleAmount	the visible amount
	 */
	public void setVisibleAmount(int visibleAmount) {
		throw new AWTError("invalid method"); //$NON-NLS-1$
	}

	public String toString() {
		StringBuffer sb = new StringBuffer(""); //$NON-NLS-1$
		sb.append(getClass().getName());
		sb.append('[');
		sb.append(getOrientation() == VERTICAL ? "vertical" : "horizontal"); //$NON-NLS-1$ //$NON-NLS-2$
		sb.append(",["); //$NON-NLS-1$
		sb.append(getMinimum() + ".." + getMaximum() + "],"); //$NON-NLS-1$ //$NON-NLS-2$
		sb.append("val="); //$NON-NLS-1$
		sb.append(getValue());
		sb.append(",vis="); //$NON-NLS-1$
		sb.append(getVisibleAmount());
		sb.append(",unit="); //$NON-NLS-1$
		sb.append(getUnitIncrement());
		sb.append(",block="); //$NON-NLS-1$
		sb.append(getBlockIncrement());
		sb.append("]"); //$NON-NLS-1$
		return sb.toString();
	}

}
