package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.io.*;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>Scrollbar</code> is graphical user interface that
 * allows the user to scroll through a large amount of data that
 * can be represented by a range of values. For instance, scrollbars
 * are used to scroll text in a text editor.
 *
 * @author		IBM
 * @version		initial
 */
public class Scrollbar extends Component implements Adjustable {

	static final long serialVersionUID = 8451667562882310543L;

	/**
	 * Constants that specify the orientation of a scrollbar.
	 */
	public static final int HORIZONTAL = 0;
	public static final int VERTICAL = 1;

	// Determines the orientation of the receiver;
	int orientation;

	// The unit and block increment of the receiver
	int lineIncrement;
	int pageIncrement;

	// The minimum and maximum values of the receiver
	int minimum;
	int maximum;

	// The current value of the receiver
	int value;

	// The visible amount of the receiver
	int visibleAmount;

	// The adjustment listener of the receiver
	transient AdjustmentListener adjustmentListener;

	int scrollbarSerializedDataVersion;

	/**
	 * Constructs a new <code>Scrollbar</code> using the default values.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public Scrollbar()
		throws HeadlessException
	{
		this(VERTICAL);
	}

	/**
	 * Constructs a new <code>Scrollbar</code> with the specified
	 * orientation.  The orientation can be either <code>VERTICAL</code>
	 * or <code>HORIZONTAL</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	orientation	the desired orientation for the receiver
	 *
	 * @thorws	IllegalArgumentException	if the orientation is not valid
	 */
	public Scrollbar(int orientation)
		throws HeadlessException
	{
		this(orientation, 0, 10, 0, 100);
	}

	/**
	 * Constructs a new <code>Scrollbar</code> with the specified
	 * values.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	orientation		the orientation for the receiver
	 * @param	value			the intial value for the receiver
	 * @param	visibleAmount	the visible value for the receiver
	 * @param	minimumValue	the minimum value for the receiver
	 * @param	maximumValue	the maximum value for the receiver
	 *
	 * @thorws	IllegalArgumentException	if the orientation is not valid
	 */
	public Scrollbar(int orientation, int value, int visibleAmount, int minimumValue, int maximumValue)
		throws HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		switch (orientation) {
			case VERTICAL:
			case HORIZONTAL:
				break;
			default:
				throw new IllegalArgumentException();
		}
		this.orientation = orientation;
//		this.value = value;
//		this.visibleAmount = visibleAmount;
//		this.minimum = minimumValue;
//		this.maximum = maximumValue;
		this.lineIncrement = 1;
		this.pageIncrement = 10;
		setValues(value, visibleAmount, minimumValue, maximumValue);
	}

	int _createPeer(Container hwParent) {
		return OS.Scrollbar_New(hwParent.peer);
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	void _setupPeer() {
		OS.Scrollbar_SetOrientation(peer, getNativeOrienatation());
		OS.Adjustable_SetMaximum(peer, maximum);
		OS.Adjustable_SetMinimum(peer, minimum);
		OS.Adjustable_SetValue(peer, value);
		OS.Adjustable_SetUnitIncrement(peer, lineIncrement);
		OS.Adjustable_SetBlockIncrement(peer, pageIncrement);
		OS.Adjustable_SetVisibleAmount(peer, visibleAmount);

		super._setupPeer();
	}

	private int getNativeOrienatation() {
		switch (orientation) {
			case Scrollbar.HORIZONTAL: return OS.STYLE_HORIZONTAL;
			case Scrollbar.VERTICAL: return OS.STYLE_VERTICAL;
			default: return OS.STYLE_HORIZONTAL;
		}
	}

	private void scrollbarCallback(int value, int actionType) {
		int pproActionType;

		switch (actionType) {

			case OS.SCROLLBAR_ACTION_UNIT_DECREMENT:
				pproActionType = AdjustmentEvent.UNIT_DECREMENT;
				break;

			case OS.SCROLLBAR_ACTION_UNIT_INCREMENT:
				pproActionType = AdjustmentEvent.UNIT_INCREMENT;
				break;

			case OS.SCROLLBAR_ACTION_BLOCK_DECREMENT:
				pproActionType = AdjustmentEvent.BLOCK_DECREMENT;
				break;

			case OS.SCROLLBAR_ACTION_BLOCK_INCREMENT:
				pproActionType = AdjustmentEvent.BLOCK_INCREMENT;
				break;

			case OS.SCROLLBAR_ACTION_TRACK:
				pproActionType = AdjustmentEvent.TRACK;
				break;

			default:
				// OS.SCROLLBAR_ACTION_HOME
				// OS.SCROLLBAR_ACTION_END
				return;
		}

		if (value == getValue()) return;
		peerValueChanged(value);

		AdjustmentEvent event = new AdjustmentEvent(this, AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, pproActionType, value);
		postEvent(event);
	}

	/* (non-Javadoc)
	 * @see java.awt.Component#coalesceEvents(java.awt.AWTEvent, java.awt.AWTEvent)
	 */
	protected AWTEvent coalesceEvents(AWTEvent currentEvent, AWTEvent newEvent) {
		AWTEvent prevEvent = super.coalesceEvents(currentEvent, newEvent);
		if (prevEvent != null) return prevEvent;

		if (currentEvent instanceof AdjustmentEvent) {
			AdjustmentEvent currAdjEvent = (AdjustmentEvent)currentEvent;
			AdjustmentEvent newAdjEvent = (AdjustmentEvent)newEvent;
			if ((currAdjEvent.getAdjustmentType() == AdjustmentEvent.TRACK) &&
			    (newAdjEvent.getAdjustmentType() == AdjustmentEvent.TRACK)) {
				return newEvent;
			}
		}

		return null;
	}

	/**
	 * @see java.awt.Component#isLightweight()
	 */
	public boolean isLightweight() {
		return false;
	}

	/**
	 * @see java.awt.Component#isOpaque()
	 */
	public boolean isOpaque() {
		return true;
	}

	String classNonlocalizedName() {
		return "scrollbar"; //$NON-NLS-1$
	}

	/**
	 * Answers the orientation of the receiver.  The orientation can be
	 * either <code>VERTICAL</code> or <code>HORIZONTAL</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the orientation of the receiver
	 *
	 * @see #setOrientation
	 */
	public int getOrientation() {
		return orientation;
	}

	/**
	 * Sets the orientation of the receiver.  The orientation can be
	 * either <code>VERTICAL</code> or <code>HORIZONTAL</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	orientation	the orientation of the receiver
	 *
	 * @throws	java.awt.IllegalArgumentException
	 *
	 * @see #getOrientation
	 */
	public void setOrientation(int orientation) {
		if (((orientation != Scrollbar.HORIZONTAL) && (orientation != Scrollbar.VERTICAL)))
			throw new IllegalArgumentException();

//		synchronized (this) {
			if (this.orientation == orientation)
				return;

			this.orientation = orientation;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.Scrollbar_SetOrientation(peer, getNativeOrienatation());
				}
			} finally {
				releasePeerReadLock();
			}
			if (isValid()) invalidate();
//		}
	}

	/**
	 * Answers the current value for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the current value for the receiver
	 *
	 * @see #setValue
	 */
	public int getValue() {
		return value;
	}

	/**
	 * Sets the current value for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	value	the new value for the receiver
	 *
	 * @see #getValue
	 */
	public void setValue(int value) {
		setValues(value, visibleAmount, minimum, maximum);
	}

	/**
	 * Answers the minimum value allowed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the minimum value allowed by the receiver
	 *
	 * @see #getValue
	 * @see #getMaximum
	 */
	public int getMinimum() {
		return minimum;
	}

	/**
	 * Sets the minimum value allowed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	minimumValue	the minimum value allowed by the receiver
	 *
	 * @see #getMinimum
	 */
	public void setMinimum(int minimumValue) {
		setValues(value, visibleAmount, minimumValue, maximum);
	}

	/**
	 * Answers the maximum value allowed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the maximum value allowed by the receiver
	 *
	 * @see #getValue
	 * @see #getMinimum
	 */
	public int getMaximum() {
		return maximum;
	}

	/**
	 * Sets the maximum value allowed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	 maximumValue	the maximum value allowed by the receiver
	 *
	 * @see #getMaximum
	 */
	public void setMaximum(int newMaximum) {
		int newMinimum = minimum;
		int newValue = value;
		int newVisibleAmount = visibleAmount;
		if (newMaximum <= newMinimum) {
			newMinimum = newMaximum - 1;
			newValue = newMinimum;
			newVisibleAmount = 1;
		}
		setValues(newValue, newVisibleAmount, newMinimum, newMaximum);
	}

	/**
	 * Answers the visible amount for the receiver.  The visible amount
	 * typically corresponds to the ratio between the size of the visible
	 * data and the total data being scrolled by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the visible amount for the receiver
	 *
	 * @see #setVisibleAmount
	 */
	public int getVisibleAmount() {
		return visibleAmount;
	}

	/**
	 * Answers the visible amount for the receiver.  The visible amount
	 * typically corresponds to the ratio between the size of the visible
	 * data and the total data being scrolled by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the visible amount for the receiver
	 *
	 * @deprecated	use <code>getVisibleAmount()</code> instead
	 *
	 * @see #setVisible
	 */
	public int getVisible() {
		return getVisibleAmount();
	}

	/**
	 * Sets the visible amount for the receiver.  The visible amount
	 * typically corresponds to the ratio between the size of the visible
	 * data and the total data being scrolled by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	visibleAmount	the visible amount for the receiver
	 */
	public void setVisibleAmount(int visibleAmount) {
		setValues(value, visibleAmount, minimum, maximum);
	}

	/**
	 * Sets the unit increment for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	unitIncrement	the unit increment for the receiver
	 *
	 * @see #getUnitIncrement
	 */
	public void setUnitIncrement(int unitIncrement) {
//		synchronized (this) {
			if (this.lineIncrement == unitIncrement) {
				return;
			}
			this.lineIncrement = unitIncrement;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.Adjustable_SetUnitIncrement(peer, unitIncrement);
				}
			} finally {
				releasePeerReadLock();
			}
//		}
	}

	/**
	 * Sets the unit increment for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	lineIncrement	the unit increment for the receiver
	 *
	 * @deprecated	use <code>setUnitIncrement(int)</code> instead
	 *
	 * @see #setUnitIncrement
	 */
	public synchronized void setLineIncrement(int lineIncrement) {
		setUnitIncrement(lineIncrement);
	}

	/**
	 * Answers the unit increment for the receiver.  The default
	 * value of the unit increment is <code>1</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the unit increment for the receiver
	 *
	 * @see #setUnitIncrement
	 */
	public int getUnitIncrement() {
		return lineIncrement;
	}

	/**
	 * Answers the unit increment for the receiver.  The default
	 * value of the unit increment is <code>1</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the unit increment for the receiver
	 *
	 * @deprecated	use <code>getUnitIncrement()</code> instead
	 *
	 * @see #getUnitIncrement
	 */
	public int getLineIncrement() {
		return getUnitIncrement();
	}

	/**
	 * Sets the block increment for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	blockIncrement	the block increment for the receiver
	 *
	 * @see #getBlockIncrement
	 */
	public void setBlockIncrement(int blockIncrement) {
//		synchronized (this) {
			if (this.pageIncrement == blockIncrement) {
				return;
			}
			this.pageIncrement = blockIncrement;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.Adjustable_SetBlockIncrement(peer, blockIncrement);
				}
			} finally {
				releasePeerReadLock();
			}
//		}
	}

	/**
	 * Sets the block increment for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	pageIncrement	the block increment for the receiver
	 *
	 * @deprecated	use <code>setBlockIncrement(int)</code> instead
	 *
	 * @see #setBlockIncrement
	 */
	public synchronized void setPageIncrement(int pageIncrement) {
		setBlockIncrement(pageIncrement);
	}

	/**
	 * Answers the block increment for the receivers.  The default
	 * value of the block increment is <code>10</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the block increment for the receiver
	 *
	 * @see #setBlockIncrement
	 */
	public int getBlockIncrement() {
		return pageIncrement;
	}

	/**
	 * Answers the block increment for the receivers.  The default
	 * value of the block increment is <code>10</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> representing the block increment for the receiver
	 *
	 * @deprecated	use <code>getBlockIncrement()</code> instead
	 *
	 * @see #getBlockIncrement
	 */
	public int getPageIncrement() {
		return getBlockIncrement();
	}

	 /**
	 * Sets the current value, visible amount, minimum and maximum
	 * values for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	value			the new current value for the receiver
	 * @param	visibleAmount	the visible amount for the receiver
	 * @param	minimumValue	the minimum value allowed by the receiver
	 * @param	maximumValue	the maximum value allowed by the receiver
	 */
	public synchronized void setValues(int newValue, int newVisibleAmount, int newMinimum, int newMaximum) {
		if (newMaximum <= newMinimum) {
			newMaximum = newMinimum + 1;
			newValue = newMinimum;
			newVisibleAmount = 1;
		} else {
			if (newVisibleAmount <= 0) {
				newVisibleAmount = 1;
			} else if (newVisibleAmount > (newMaximum - newMinimum)) {
				int tempVisibleAmount = (newMaximum - newMinimum);
				if (tempVisibleAmount >= 0) newVisibleAmount = tempVisibleAmount;
			}

			if (newValue < newMinimum) {
				newValue = newMinimum;
			} else if (newValue > newMaximum - newVisibleAmount) {
				if (newMaximum - newVisibleAmount < 0) {
					newValue = 0;
				} else {
					newValue = newMaximum - newVisibleAmount;
				}
			}
		}

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				if (maximum != newMaximum) OS.Adjustable_SetMaximum(peer, newMaximum);
				if (minimum != newMinimum) OS.Adjustable_SetMinimum(peer, newMinimum);
				if (visibleAmount != newVisibleAmount) OS.Adjustable_SetVisibleAmount(peer, newVisibleAmount);
				if (value != newValue) OS.Adjustable_SetValue(peer, newValue);
			}
		} finally {
			releasePeerReadLock();
		}

		this.value = newValue;
		this.visibleAmount = newVisibleAmount;
		this.minimum = newMinimum;
		this.maximum = newMaximum;
	}

	/*
	 * The user changed the peer's value.
	 * See EventPoster.postAdjustmentEvent.
	 */
	void peerValueChanged(int newValue) {
		//setValue(newValue);

		/*
		 * Since newValue originated in the peer, it is guaranteed
		 * to be in the correct range, so there is no need to
		 * update any other attributes.
		 */
		value = newValue;
	}

	/**
	 * Adds a <code>AdjustmentListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		adjustmentListener	the <code>AdjustmentListener</code> to be added.
	 *
	 * @see			java.awt.event.AdjustmentListener
	 * @see			java.awt.event.AdjustmentEvent
	 */
	public synchronized void addAdjustmentListener(AdjustmentListener adjustmentListener) {
		addEventListener(adjustmentListener);
		this.adjustmentListener = AWTEventMulticaster.add(this.adjustmentListener, adjustmentListener);
	}

	/**
	 * Removes an <code>AdjustmentListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		adjustmentListener	the <code>AdjustmentListener</code> to be removed.
	 *
	 * @see			java.awt.event.AdjustmentListener
	 * @see			java.awt.event.AdjustmentEvent
	 */
	public synchronized void removeAdjustmentListener(AdjustmentListener adjustmentListener) {
		this.adjustmentListener = AWTEventMulticaster.remove(this.adjustmentListener, adjustmentListener);
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>AdjustmentEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processAdjustmentEvent(AdjustmentEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof AdjustmentEvent) {
			processAdjustmentEvent((AdjustmentEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>AdjustmentEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processAdjustmentEvent(AdjustmentEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	adjustmentEvent	the event to be processed
	 *
	 * @see		AdjustmentEvent
	 */
	protected void processAdjustmentEvent(AdjustmentEvent adjustmentEvent) {
		if (adjustmentEvent != null) {
			value = adjustmentEvent.getValue();
		}
		if (adjustmentListener == null)
			return;

		if ((adjustmentEvent == null) ||
		    (adjustmentEvent.getID() == AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED)) {
			adjustmentListener.adjustmentValueChanged(adjustmentEvent);
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return
			super.paramString() +
			",val=" + getValue() + //$NON-NLS-1$
			",vis=" + getVisibleAmount() + //$NON-NLS-1$
			",min=" + getMinimum() + //$NON-NLS-1$
			",max=" + getMaximum() + //$NON-NLS-1$
			"," + ((getOrientation() == VERTICAL) ? "vert" : "horiz"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof AdjustmentEvent)
			return ((eventMask & AWTEvent.ADJUSTMENT_EVENT_MASK) != 0) || adjustmentListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, adjustmentListenerK, adjustmentListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(adjustmentListenerK))
				addAdjustmentListener((AdjustmentListener)listener);
		}
	}

	public AdjustmentListener[] getAdjustmentListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(adjustmentListener, AdjustmentListener.class);
		AdjustmentListener[] resultArray = new AdjustmentListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}
}
