package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.IOException;
import java.io.ObjectInputStream;
import java.awt.event.KeyEvent;
import java.util.HashSet;
import java.util.Set;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>TextArea</code> is a multi-line text editor.
 *
 * @author		IBM
 * @version		initial
 */
public class TextArea extends TextComponent {

	static final long serialVersionUID = 3692302836626095722L;

	// The number of rows of the receiver
	int rows;

	// The number of columns of the receiver
	int columns;

	// Determines which scrollbars are displayed when needed
	int scrollbarVisibility;

	//	The serialized data version of the receiver
	int textAreaSerializedDataVersion;

	/**
	 * Constants for specifing which scrollbars are visible.
	 */
	public static final int SCROLLBARS_BOTH = 0;
	public static final int SCROLLBARS_VERTICAL_ONLY = 1;
	public static final int SCROLLBARS_HORIZONTAL_ONLY = 2;
	public static final int SCROLLBARS_NONE = 3;

	/**
	 * Constructs a new <code>TextArea</code> with an empty text.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public TextArea()
		throws HeadlessException
	{
		this(""); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>TextArea</code> with specified text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text displayed by the receiver
	 */
	public TextArea(String text)
		throws HeadlessException
	{
		this(text, 0, 0);
	}

	/**
	 * Constructs a new <code>TextArea</code> with an empty text and
	 * specified number of visible rows and columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	rows	the desired number of rows
	 * @param	columns	the desired number of columns
	 */
	public TextArea(int rows, int columns)
		throws HeadlessException
	{
		this("", rows, columns); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>TextArea</code> with the specified
	 * text, number of visible rows and columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text displayed by the receiver
	 * @param	rows	the desired number of rows
	 * @param	columns	the desired number of columns
	 */
	public TextArea(String text, int rows, int columns)
		throws HeadlessException
	{
		this(text, rows, columns, SCROLLBARS_BOTH);
	}

	/**
	 * Constructs a new <code>TextArea</code> with the specified
	 * text, number of visible rows and columns. The scrollbars
	 * will be displayed when needed according to the given value of
	 * <code>visibleScrollbars</code>, each can be one of
	 * <code>SCROLLBARS_BOTH</code>
	 * <code>SCROLLBARS_VERTICAL_ONLY</code>
	 * <code>SCROLLBARS_HORIZONTAL_ONLY</code>
	 * <code>SCROLLBARS_NONE</code>
	 * <p>
	 * By default, both scrollbars will be displayed when needed.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text				the text displayed by the receiver
	 * @param	rows				the desired number of rows
	 * @param	columns				the desired number of columns
	 * @param	visibleScrollbars
	 */
	public TextArea(String text, int rows, int columns, int visibleScrollbars)
		throws HeadlessException
	{
		super(text);
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}

		switch (visibleScrollbars) {
			case SCROLLBARS_BOTH:
			case SCROLLBARS_VERTICAL_ONLY:
			case SCROLLBARS_HORIZONTAL_ONLY:
			case SCROLLBARS_NONE:
				break;
			default:
				visibleScrollbars = SCROLLBARS_BOTH;
		}
		this.scrollbarVisibility = visibleScrollbars;

		this.rows = Math.max(rows, 0);
		this.columns = Math.max(columns, 0);
	}

	/**
	 * Inserts the specified text to the text being displayed by
	 * the receiver at the specified position. If <code>position</code>
	 * is greater than the size of the current text, the specified
	 * text will be appended to the current text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	insertText	the text to be inserted
	 * @param	position	the position where the text is to be inserted
	 */
	public void insert(String insertText, int position) {
		acquirePeerReadLock();
		try {
			if (peer != 0 && insertText == null) {
				return;
			}

			insertText = String.valueOf(insertText);

			if (peer == 0) {
				// Only throw StringIndexOutOfBoundsExceptions when the peer
				// has not been created.  This behavior matches the reference implementation.
				int size = (text == null) ? 0 : text.length();
				if ((position < 0) || (position > size)) {
					throw new StringIndexOutOfBoundsException();
				}

				StringBuffer buffer = new StringBuffer(size + insertText.length());
				buffer.append(text.substring(0, position));
				buffer.append(insertText);
				buffer.append(text.substring(position, size));

				text = buffer.toString();
			} else {
				int size = OS.AbstractTextComponent_GetCharCount(peer);

				// On the windows JRE inserting text to a negative position will
				// cause the text to be appended to the end, only if the peer has already been created
				if (position < 0) position = size;

				// On the windwos JRE and the reference implementation inserting text to
				// a position greater than the size of the current text will append the text to the
				// end of the text
				if (position > size) position = size;

				OS.AbstractTextComponent_Replace(peer, insertText, position, position);
				postTextChangedEvent();
			}
		} finally {
			releasePeerReadLock();
		}

		if (isValid()) invalidate();
	}

	/**
	 * Inserts the specified text to the text being displayed by
	 * the receiver at the specified position. If <code>position</code>
	 * is greater than the size of the current text, the specified
	 * text will be appended to the current text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	insertText	the text to be inserted
	 * @param	position	the position where the text is to be inserted
	 *
	 * @deprecated	use <code>insert(String)</code> instead
	 */
	public synchronized void insertText(String insertText, int position) {
		insert(insertText, position);
	}

	/**
	 * Appends the specified text to the current text displayed by
	 * the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text to be appended
	 */
	public void append(String text) {
		acquirePeerReadLock();
		try {
			if (peer != 0 && text == null) {
				return;
			}

			if (peer == 0)
				this.text = this.text + text;
			else {
				int pos = OS.AbstractTextComponent_GetCharCount(peer);
				OS.AbstractTextComponent_Replace(peer, text, pos, pos);
				postTextChangedEvent();
			}
		} finally {
			releasePeerReadLock();
		}

		if (isValid()) invalidate();
	}

	int _createPeer(Container hwParent) {
		return OS.TextArea_New(hwParent.peer, widgetStyle());
	}

		int widgetStyle() {
			int style = 0;
			switch (scrollbarVisibility) {
				case SCROLLBARS_VERTICAL_ONLY:
					style |= OS.STYLE_V_SCROLL | OS.STYLE_WRAP;
					break;
				case SCROLLBARS_HORIZONTAL_ONLY:
					style |= OS.STYLE_H_SCROLL;
					break;
				case SCROLLBARS_NONE:
					style |= OS.STYLE_WRAP;
					break;
				default:
					style |= OS.STYLE_H_SCROLL | OS.STYLE_V_SCROLL;
			}
			return super.widgetStyle() | style;
		}

	boolean wantsAllKeys() {
		return true;
	}

	/**
	 * Appends the specified text to the text being displayed by
	 * the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text to be appended
	 *
	 * @deprecated	use <code>append(String)</code> instead
	 */
	public synchronized void appendText(String text) {
		append(text);
	}

	/**
	 * Replaces the portion of text being displayed by the receiver
	 * at the specified range with the specified text.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	replaceText		the new text
	 * @param	startPosition	the position of the first character that is to be replaced
	 * @param	endPosition		the position of the last character that is to be replaced
	 */
	public void replaceRange(String replaceText, int startPosition, int endPosition) {
		replaceText(replaceText, startPosition, endPosition);
	}

	/**
	 * Replaces the portion of text being displayed by the receiver
	 * at the specified range with the specified text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	replaceText 	the new text
	 * @param	startPosition	the position of the first character that is to be replaced
	 * @param	endPosition		the position of the last character that is to be replaced
	 *
	 * @deprecated	use <code>replaceRange(String, int, int)</code> instead
	 */
	public synchronized void replaceText(final String replaceText, final int startPosition, final int endPosition) {
		String theText = getText();
		int size = (null == theText ? 0 : theText.length());

		acquirePeerReadLock();
		try {
			if ((startPosition < 0)
				|| ((peer == 0) && (startPosition > size))) {
				throw new StringIndexOutOfBoundsException("String index out of range: " + startPosition); //$NON-NLS-1$
			}
			if ((endPosition < 0) || ((peer == 0) && (endPosition > size))) {
				throw new StringIndexOutOfBoundsException("String index out of range: " + endPosition); //$NON-NLS-1$
			}

			_replaceText(replaceText + "", startPosition, endPosition); //$NON-NLS-1$
		} finally {
			releasePeerReadLock();
		}

		if (isValid()) invalidate();
	}

	void _replaceText(String replaceText, int startPosition, int endPosition) {
		if (peer == 0) {
			int size = text.length();

			StringBuffer buffer = new StringBuffer(
					size + replaceText.length() - (endPosition - startPosition));
			buffer.append(text.substring(0, startPosition));
			buffer.append(replaceText);
			buffer.append(text.substring(endPosition, size));

			text = buffer.toString();
		} else {
			int size = OS.AbstractTextComponent_GetCharCount(peer);

			if ((startPosition < 0) || (startPosition > size)) startPosition = size;
			if ((endPosition < 0) || (endPosition > size)) endPosition = size;

			OS.AbstractTextComponent_Replace(peer, replaceText, startPosition, endPosition);
			postTextChangedEvent();
		}
	}

	/**
	 * Answer the number of visible rows for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>int</code> which is the number of visible rows
	 */
	public int getRows() {
		return rows;
	}

	/**
	 * Sets the number of visible rows for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param 	rows	the desired number of rows for the receiver
	 */
	public void setRows(int rows) {
		if (rows < 0)
			throw new IllegalArgumentException();

		if (this.rows == rows) {
			return;
		}
		this.rows = rows;
		if (isValid()) invalidate();
	}

	/**
	 * Answers the number of visible columns for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which the number of visible columns
	 */
	public int getColumns() {
		return columns;
	}

	/**
	 * Sets the number of visible columns for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param 	columns	the desired number of columns for the receiver
	 */
	public void setColumns(int columns) {
		if (columns < 0)
			throw new IllegalArgumentException();

		if (this.columns == columns) {
			return;
		}
		this.columns = columns;
		if (isValid()) invalidate();
	}

	/**
	 * Answer a constants which represents which scrolbars are displayed
	 * when needed. The return value can be one of
	 *
	 * <code>SCROLLBARS_BOTH</code>
	 * <code>SCROLLBARS_VERTICAL_ONLY</code>
	 * <code>SCROLLBARS_HORIZONTAL_ONLY</code>
	 * <code>SCROLLBARS_NONE</code>
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>int</code> which represents which scrolbars are displayed
	 */
	public int getScrollbarVisibility() {
		return scrollbarVisibility;
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of rows and
	 * columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	public Dimension getPreferredSize(int rows, int columns) {
		return getMinimumSize(rows, columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of rows and
	 * columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 *
	 * @deprecated	use <code>getPreferredSize(int, int)</code> instead
	 */
	public Dimension preferredSize(int rows, int columns) {
		return getPreferredSize(rows, columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	Dimension getPreferredSizeImpl() {
		return getPreferredSize(
			getRows() <= 0 ? 4 : getRows(),
			getColumns() <= 0 ? 40 : getColumns());
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of rows and columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of rows and columns.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	public Dimension getMinimumSize(final int rows, final int columns) {
		acquirePeerReadLock();
		try {
			if (peer == 0) return getSize();

			int[] point = OS.TextArea_GetMinimumSize(peer, rows, columns);
			return new Dimension(point[OS.INDEX_X],point[OS.INDEX_Y]);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of rows and columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 *
	 * @deprecated	use <code>getMinimumSize(int, int)</code> instead
	 *
	 */
	public Dimension minimumSize(int rows, int columns) {
		return getMinimumSize(rows, columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	Dimension getMinimumSizeImpl() {
		return getMinimumSize(
			getRows() <= 0 ? 4 : getRows(),
			getColumns() <= 0 ? 40 : getColumns());
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return
			super.paramString() +
			",rows=" + getRows() + //$NON-NLS-1$
			",columns=" + getColumns() + //$NON-NLS-1$
			",scrollbars=" + getScrollbarVisibility(); //$NON-NLS-1$
	}

	String classNonlocalizedName() {
		return "text"; //$NON-NLS-1$
	}

	Set getDefaultFocusTraversalKeys(int id) {
		Set s = KeyboardFocusManager.getCurrentKeyboardFocusManager().getDefaultFocusTraversalKeys(id);

		if (id == KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS) {
			AWTKeyStroke tabDown = AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, 0, false);
			if (s.contains(tabDown)) {
				Set newS = new HashSet(s);
				newS.remove(tabDown);
				s = newS;
			}
		}

		return s;
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException, HeadlessException {
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
	}
}
