package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.event.*;
import java.io.*;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * <code>TextComponent</code> is the abstract superclass of widget
 * classes that allow text editing.
 *
 * @author		IBM
 * @version		initial
 *
 */
public class TextComponent extends Component {

	static final long serialVersionUID = -2214773872412987419L;

	// Determines if the receiver is editable
	boolean editable;

	// The position of the last character selected
	int selectionEnd;

	// The position of the first character selected
	int selectionStart;

	// The text of the receiver
	String text;

	// The text listener of the receiver
	transient protected TextListener textListener;

	//	The serialized data version of the receiver
	int textComponentSerializedDataVersion;

	// checkForEnableIM    -  Mandatory serialized field
	// backgroundSetByClientCode -  Mandatory serialized field
	boolean checkForEnableIM;
	boolean backgroundSetByClientCode;

	TextComponent(String text) {
		this.text = (text == null) ? "" : text; //$NON-NLS-1$
		this.editable = true;
		this.selectionStart = 0;
		this.selectionEnd = 0;
		this.checkForEnableIM = false;
		this.backgroundSetByClientCode = false;
	}

	/**
	 * Create the underlying widget for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 */
	void _setupPeer() {
		if (!editable) OS.AbstractTextComponent_SetEditable(peer, false);

		if ((text != null) && (!"".equals(text))) { //$NON-NLS-1$
			OS.AbstractTextComponent_SetText(peer, text);
		}
		if ((selectionStart != 0) || (selectionEnd != 0)) {
			OS.AbstractTextComponent_SetSelection(peer, selectionStart, selectionEnd);
		}

		text = null;

		super._setupPeer();
	}

	void textModifyCallback() {
		postTextChangedEvent();
	}

	void postTextChangedEvent() {
		TextEvent event = new TextEvent(this, TextEvent.TEXT_VALUE_CHANGED);
		postEvent(event);
	}

	Cursor _getDefaultCursor() {
		return Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR);
	}

	public Cursor getCursor() {
		if (cursor != null)
			return cursor;

		// Unlike the Components, TextComponents should always
		// default to TEXT_CURSOR if one has not been set by the user.
		// Parent cursors are not considered no matter what the state of
		// the parent
		return _getDefaultCursor();
	}

	int _getTextLength() {
		acquirePeerReadLock();
		try {
			return (peer == 0) ? text.length() : OS.AbstractTextComponent_GetCharCount(peer);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Adds an <code>TextListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		textListener	the <code>TextListener</code> to be added.
	 *
	 * @see			java.awt.event.TextListener
	 * @see			java.awt.event.TextEvent
	 */
	public synchronized void addTextListener(TextListener textListener) {
		addEventListener(textListener);
		this.textListener = AWTEventMulticaster.add(this.textListener, textListener);
	}

	/**
	 * Answers the position of the text insertion caret for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>int</code> which is the position of the text insertion caret
	 *
	 * @see #setCaretPosition
	 */
	public synchronized int getCaretPosition() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return 0;

			int[] point = OS.AbstractTextComponent_GetSelection(peer);
			return point[OS.INDEX_X];
		} finally {
			releasePeerReadLock();
		}
	}

	public Color getBackground() {
		// Always answer background if set by setBackground
		if (background != null)
			return background;
		// PPro 1.1 added condition that non-editable texts use control color
		if (!isEditable() && !backgroundSetByClientCode) return SystemColor.control;

		return super.getBackground();
	}

	/**
	 * Answers the <code>String</code> representing the portion of text
	 * that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is the selected text
	 *
	 * @see #select
	 */
	public String getSelectedText() {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				if (text == null) {
					return null;
				}
				else {
					return text.substring(
						Math.min(selectionStart, text.length()),
						Math.min(selectionEnd, text.length()));
				}
			}

			return OS.AbstractTextComponent_GetSelectionText(peer);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the position of the last character that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>int</code> which is the position of the last character selected
	 *
	 * @see #setSelectionEnd
	 * @see #getSelectionStart
	 */
	public int getSelectionEnd() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return selectionEnd;

			int[] point = OS.AbstractTextComponent_GetSelection(peer);
			return point[OS.INDEX_Y];
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the position of the first character that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>int</code> which is the position of the first character selected
	 *
	 * @see #setSelectionStart
	 * @see #getSelectionEnd
	 */
	public synchronized int getSelectionStart() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return selectionStart;

			int[] point = OS.AbstractTextComponent_GetSelection(peer);
			return point[OS.INDEX_X];
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the text displayed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is the text displayed by the receiver
	 *
	 * @see #setText
	 */
	public synchronized String getText() {
		acquirePeerReadLock();
		try {
			if (peer == 0)
				return text;

			return OS.AbstractTextComponent_GetText(peer);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answers the <code>String</code> representing the portion of text
	 * that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.lang.String which is the selected text
	 *
	 * @see #select
	 */
	public boolean isEditable() {
		return editable;
	}

	public boolean isLightweight() {
		return false;
	}

	public boolean isOpaque() {
		return true;
	}

	boolean isFocusTraversableImpl() {
		return true;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		return super.paramString() +
			",editable=" + isEditable() + //$NON-NLS-1$
			",selection=" + getSelectedText(); //$NON-NLS-1$
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>TextEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processTextEvent(TextEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof TextEvent) {
			processTextEvent((TextEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>TextEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behavior by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processTextEvent(TextEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	textEvent	the event to be processed
	 *
	 * @see		java.awt.event.TextEvent
	 * @see		#enableEvents
	 */
	protected void processTextEvent(TextEvent textEvent) {
		if (textListener == null)
			return;

		switch (textEvent.getID()) {
			case TextEvent.TEXT_VALUE_CHANGED:
				textListener.textValueChanged(textEvent);
				break;
			default:
				return;
		}
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(textListenerK))
				addTextListener((TextListener)listener);
		}
	}

	/**
	 * Removes an <code>TextListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		textListener	the <code>TextListener</code> to be removed.
	 *
	 * @see			java.awt.event.TextListener
	 * @see			java.awt.event.TextEvent
	 */
	public synchronized void removeTextListener(TextListener textListener) {
		this.textListener = AWTEventMulticaster.remove(this.textListener, textListener);
	}

	/**
	 * Sets the selected portion of text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	newSelectionStart	the position of the first character to be selected
	 * @param	newSelectionEnd		the position of the last character to be selected
	 *
	 * @see #setSelectionStart(int)
	 * @see #setSelectionEnd(int)
	 * @see #selectAll
	 */
	public synchronized void select(int newSelectionStart, int newSelectionEnd) {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				setSelectionStart(newSelectionStart);
				setSelectionEnd(newSelectionEnd);
			} else {
				OS.AbstractTextComponent_SetSelection(peer, newSelectionStart, newSelectionEnd);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the whole text displayed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see #select
	 */
	public synchronized void selectAll() {
		acquirePeerReadLock();
		try {
			if (peer == 0) {
				selectionStart = 0;
				selectionEnd = text.length();
			} else {
				OS.AbstractTextComponent_SelectAll(peer);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the position of the text insertion caret for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	caretPosition the desired caret position
	 *
	 * @throws	IllegalArgumentException		if the specified position is negative
	 * @throws	IllegalComponentStateException	if the receiver is not displayable
	 *
	 * @see #getCaretPosition
	 */
	public synchronized void setCaretPosition(int caretPosition) {
		if (caretPosition < 0) throw new IllegalArgumentException();

		acquirePeerReadLock();
		try {
			if (peer == 0)
				throw new IllegalComponentStateException();

			OS.AbstractTextComponent_SetCaretPosition(peer, caretPosition);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets whether the receiver is editable.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		editable	<code>true</code> to allow text editing,
	 *							<code>false</code> otherwise
	 *
	 * @see #isEditable
	 */
	public synchronized void setEditable(boolean editable) {
		if (this.editable == editable)
			return;

		this.editable = editable;

		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.AbstractTextComponent_SetEditable(peer, editable);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the position of the last character that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	newSelectionEnd	the position of the last character selected
	 *
	 * @see #getSelectionEnd
	 * @see #setSelectionStart
	 */
	public synchronized void setSelectionEnd(int newSelectionEnd) {
		/* Selection criteria:
		selection start is 0 if less than 0
		selection start is lesser of given value or length of text
		selection end is equal to selection start if given value is lesser than selection start
		selection end is lesser of given value or text length */

		// compute length and start regardless of peer state
		int length = _getTextLength();
		int newSelectionStart = Math.max(0, Math.min(getSelectionStart(), length));
		newSelectionEnd = Math.max(newSelectionStart,Math.min(newSelectionEnd, length));

		acquirePeerReadLock();
		try {
			if (peer == 0) {
				selectionStart = newSelectionStart;
				selectionEnd = newSelectionEnd;
			}
			else {
				OS.AbstractTextComponent_SetSelection(peer, newSelectionStart, newSelectionEnd);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof TextEvent)
			return ((eventMask & AWTEvent.TEXT_EVENT_MASK) != 0) || textListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	/**
	 * Sets the position of the first character that is selected.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	newSelectionStart	the position of the first character selected
	 *
	 * @see #getSelectionStart
	 * @see #setSelectionEnd
	 */
	public synchronized void setSelectionStart(int newSelectionStart) {
		// Selection criteria:
		// selection start is 0 if less than 0
		// selection start is lesser of given value or length of text
		// selection end is equal to selection start if given value is lesser than selection start
		// selection end is lesser of given value or text length

		// compute length and end regardless of peer state
		int length = _getTextLength();
		newSelectionStart = Math.max(0,Math.min(newSelectionStart,length));
		int newSelectionEnd = Math.max(newSelectionStart,Math.min(getSelectionEnd(), length));

		acquirePeerReadLock();
		try {
			if (peer == 0) {
				selectionStart = newSelectionStart;
				selectionEnd = newSelectionEnd;
			}
			else {
				OS.AbstractTextComponent_SetSelection(peer, newSelectionStart, newSelectionEnd);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Sets the text displayed by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	text	the text to be displayed or <code>null</code> for
	 *					an empty text
	 *
	 * @see #getText
	 */
	public synchronized void setText(String text) {
		if (text == null) text = ""; //$NON-NLS-1$

		if (text.equals(this.text)) {
			return;
		}

		acquirePeerReadLock();
		try {
			if (peer == 0) {
				this.text = text;
			} else {
				OS.AbstractTextComponent_SetText(peer, text);
				postTextChangedEvent();
			}
		} finally {
			releasePeerReadLock();
		}

		if (isValid()) invalidate();
	}

	int widgetStyle() {
		return OS.STYLE_BORDER;
	}

	public void removeNotify() {
		acquirePeerReadLock();
		try {
			if (peer != 0) {
				text = OS.AbstractTextComponent_GetText(peer);
				int[] point = OS.AbstractTextComponent_GetSelection(peer);
				selectionStart = point[OS.INDEX_X];
				selectionEnd = point[OS.INDEX_Y];
			}
		} finally {
			releasePeerReadLock();
		}

		super.removeNotify();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, textListenerK, textListener);
		stream.writeObject(null);
	}

	public void setBackground(Color color) {
		backgroundSetByClientCode = true;
		super.setBackground(color);
	}

	public void enableInputMethods(boolean enable) {
		this.checkForEnableIM = enable;
		super.enableInputMethods(enable);
	}

	public TextListener[] getTextListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(textListener, TextListener.class);
		TextListener[] resultArray = new TextListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	void dispatchEventImpl(AWTEvent awtEvent) {
		if (checkForEnableIM) {
			getInputContext().dispatchEvent(awtEvent);
		}
		super.dispatchEventImpl(awtEvent);
	}

}
