package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.*;
import java.util.EventListener;

import com.ibm.ugl.eswt.OS;

/**
 * A <code>TextField</code> is a single-line text editor.
 *
 * @author		IBM
 * @version		initial
 */
public class TextField extends TextComponent {

	static final long serialVersionUID = -2966288784432217853L;

	// The echo char of the receiver
	char echoChar;

	// The number of columns of the receiver
	int columns;

	// The action listener of the receiver
	transient ActionListener actionListener;

	int textFieldSerializedDataVersion;

	/**
	 * Constructs a new <code>TextField</code> with an empty text.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public TextField()
		throws HeadlessException
	{
		this("", 0); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>TextField</code> with specified text.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		text	the text displayed by the receiver
	 */
	public TextField(String text)
		throws HeadlessException
	{
		this(text, (text == null) ? 0 : text.length());
	}

	/**
	 * Constructs a new <code>TextField</code> with an empty text and
	 * the specified number of columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		columns	the desired number of columns
	 */
	public TextField(int columns)
		throws HeadlessException
	{
		this("", columns); //$NON-NLS-1$
	}

	/**
	 * Constructs a new <code>TextField</code> with the specified
	 * text and number of columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		text	the text displayed by the receiver
	 * @param		columns	the desired number of columns
	 */
	public TextField(String text, int columns)
		throws HeadlessException
	{
		super(text);
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}

		this.columns = Math.max(columns, 0);
	}

	int _createPeer(Container hwParent) {
		return OS.TextField_New(hwParent.peer, widgetStyle());
	}

	/**
	 * Answer the receiver's echo charater.  The echo character is the
	 * character that is displayed when the user enters text or the text
	 * is changed programatically.  If the echo character is '\0', the
	 * text will be displayed normally.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>char</code> which is the receiver echo character
	 */
	public char getEchoChar() {
		return echoChar;
	}

	/**
	 * Sets the echo character for the receiver.  The echo character is the
	 * character that is displayed when the user enters text or the text
	 * is changed programatically.  If the echo character is '\0', the
	 * text will be displayed normally.
	 *
	 * Note: some platforms support a limited set of echo characters, so
	 * the <code>echoChar</code> value is considered a hint.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param 	echoChar the desired echo character
	 */
	public void setEchoChar(char echoChar) {
//		synchronized (this) {
			if (this.echoChar == echoChar) {
				return;
			}
			this.echoChar = echoChar;

			acquirePeerReadLock();
			try {
				if (peer != 0) {
					OS.AbstractTextComponent_SetEchoChar(peer, echoChar);
				}
			} finally {
				releasePeerReadLock();
			}
//		}
	}

	/**
	 * Overrides _addNotify() from TextComponent so that the echoChar will
	 * be properly set when the TextField becomes visible.
	 */
	void _setupPeer() {
		super._setupPeer();

		OS.AbstractTextComponent_SetEchoChar(peer, echoChar);
	}

	boolean keyCallback(int type, char character, int keycode, int modifiers) {
		if (keycode == OS.KEYCODE_ENTER) {
			ActionEvent event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, getText(),
				System.currentTimeMillis(),
			modifiers);
			postEvent(event);
		}
		return super.keyCallback(type,character,keycode,modifiers);
	}

	/**
	 * Sets the echo character for the receiver.  The echo character is the
	 * character that is displayed when the user enters text or the text
	 * is changed programatically.  If the echo character is '\0', the
	 * text will be displayed normally.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param 	echoChar the desired echo character
	 *
	 * @deprecated	use <code>setEchoChar(char)</code> instead
	 */
	public void setEchoCharacter(char echoCharacter) {
		setEchoChar(echoCharacter);
	}

	/**
	 * Answer whether the receiver's echo charater is set.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	<code>true</code> if the receiver's echo charater is set,
	 *			<code>false</code> otherwise
	 */
	public boolean echoCharIsSet() {
		return getEchoChar() != '\0';
	}

	/**
	 * Answers the number of visible columns for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return 	an <code>int</code> which the number of visible columns
	 */
	public int getColumns() {
		return columns;
	}

	/**
	 * Sets the number of visible columns for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param 	columns the desired number of columns for the receiver
	 */
	public synchronized void setColumns(int columns) {
		if (columns < 0)
			throw new IllegalArgumentException();

		if (this.columns == columns) {
			return;
		}
		this.columns = columns;
		if (isValid()) invalidate();
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	public Dimension getPreferredSize(int columns) {
		return getMinimumSize(columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display the specified number of columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 *
	 * @deprecated	use <code>getPreferredSize(int)</code> instead
	 */
	public Dimension preferredSize(int columns) {
		return getPreferredSize(columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the size with which
	 * the receiver can fully display itself and its children if applicable.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the receiver's preferred size
	 */
	Dimension getPreferredSizeImpl() {
		return getPreferredSize(getColumns());
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of columns.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	public Dimension getMinimumSize(final int columns) {
		acquirePeerReadLock();
		try {
			if (peer == 0) return getSize();
			int[] point = OS.TextField_GetMinimumSize(peer, columns);
			return new Dimension(point[OS.INDEX_X],point[OS.INDEX_Y]);
		} finally {
			releasePeerReadLock();
		}
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver in order to have the specified number
	 * of columns.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 *
	 * @deprecated	use <code>getMinimumSize(int)</code> instead
	 */
	public Dimension minimumSize(int columns) {
		return getMinimumSize(columns);
	}

	/**
	 * Answer a <code>Dimension</code> representing the minimum
	 * size of the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Dimension which is the minimum size of the receiver
	 */
	Dimension getMinimumSizeImpl() {
		return getMinimumSize(getColumns());
	}

	/**
	 * Adds an <code>ActionListener</code> to the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be added.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void addActionListener(ActionListener actionListener) {
		addEventListener(actionListener);
		this.actionListener = AWTEventMulticaster.add(this.actionListener, actionListener);
	}

	/**
	 * Removes an <code>ActionListener</code> from the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param		actionListener	the <code>ActionListener</code> to be removed.
	 *
	 * @see			java.awt.event.ActionListener
	 * @see			java.awt.event.ActionEvent
	 */
	public synchronized void removeActionListener(ActionListener actionListener) {
		this.actionListener = AWTEventMulticaster.remove(this.actionListener, actionListener);
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>ActionEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processActionEvent(ActionEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof ActionEvent) {
			processActionEvent((ActionEvent)awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>ActionEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processActionEvent(ActionEvent)</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	actionEvent	the event to be processed
	 *
	 * @see		ActionEvent
	 */
	protected void processActionEvent(ActionEvent actionEvent) {
		if (actionListener == null)
			return;

		if ((actionEvent == null)
		    || (actionEvent.getID() == ActionEvent.ACTION_PERFORMED)) {
			actionListener.actionPerformed(actionEvent);
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the component.  This will not be <code>null</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	protected String paramString() {
		if (!echoCharIsSet())
			return super.paramString();

		return super.paramString() + ",echoChar=" + getEchoChar(); //$NON-NLS-1$
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof ActionEvent)
			return ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) || actionListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	String classNonlocalizedName() {
		return "textfield"; //$NON-NLS-1$
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, actionListenerK, actionListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(actionListenerK))
				addActionListener((ActionListener)listener);
		}
	}

	public ActionListener[] getActionListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(actionListener, ActionListener.class);
		ActionListener[] resultArray = new ActionListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}
}
