package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2003, 2006  All Rights Reserved
 */

import java.awt.datatransfer.Clipboard;
import java.awt.image.*;
import java.io.FilePermission;
import java.io.IOException;
import java.net.SocketPermission;
import java.net.URL;
import java.security.AccessController;
import java.security.Permission;
import java.security.PrivilegedAction;

import com.ibm.oti.awt.IToolkitImpl;
import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.image.BufferedImageHandler;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.oti.awt.metal.widgets.Display;

class ToolkitImpl extends java.awt.Toolkit implements IToolkitImpl {
	Clipboard clipboard;
	ColorModel colorModel;
	Display display;
	EventQueue eventQueue;

	com.ibm.oti.awt.image.ImageCache imageCache;

	static {
		Util.reference();
		GraphicsImpl.reference();
	}

	// This method is included so that other classes can trigger
	// the static initializer of this class
	static void reference() {
		// NOP
	}

	/**
	 * Constructor for ToolkitImpl.
	 */
	public ToolkitImpl() {
		imageCache = new com.ibm.oti.awt.image.ImageCache();

		eventQueue = new EventQueue();

		// Create a display object and assume that any threads needed to use
		// the display will be started when this method exits
		display = Display.openDisplay();

		if (display == null) {
			GraphicsEnvironment.getLocalGraphicsEnvironment().setHeadless(true);
		}

		AccessController.doPrivileged(new PrivilegedAction() {
			public Object run() {
				Runtime.getRuntime().addShutdownHook(new Thread("Display shutdown hook") {
					// Need to clean up if someone does a System.exit() without disposing of the frame
					// Bug 98903
					public void run() {
						Util.disposeAllFrames();
						display.dispose();
					}
				});
				return null;
			}
		});
	}

	public synchronized Display getDisplay() {
		return display;
	}

	/**
	 * @see java.awt.Toolkit#getScreenSize()
	 */
	public Dimension getScreenSize() {
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		Rectangle bounds = getDisplay().getBounds();
		return new Dimension(bounds.width, bounds.height);
	}

	Insets _getScreenInsets() {
		return getDisplay().getInsets();
	}

	/**
	 * @see java.awt.Toolkit#getScreenResolution()
	 */
	public int getScreenResolution() {
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		return getDisplay().getDPI().y;
	}

	/**
	 * @see java.awt.Toolkit#getColorModel()
	 */
	public ColorModel getColorModel() {
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		if (colorModel == null) {
			colorModel = ColorModel.getRGBdefault();

//			final ImageData[] imageDatas = new ImageData[1];
//			BBToolkit.syncExec(new Runnable() {
//				public void run() {
//					org.eclipse.swt.graphics.Image image = new org.eclipse.swt.graphics.Image(getDisplay(), 1, 1);
//					imageDatas[0] = image.getImageData();
//					image.dispose();
//				}
//			});
//
//			ImageData imageData = imageDatas[0];
//			if (imageData == null)
//					return null;
//
//			PaletteData palette = imageData.palette;
//			if (palette.isDirect) {
//				colorModel = new DirectColorModel(imageData.depth, palette.redMask, palette.greenMask, palette.blueMask);
//			} else {
//				RGB[] colors = palette.getRGBs();
//				int size = colors.length;
//				byte[] reds = new byte[size];
//				byte[] greens = new byte[size];
//				byte[] blues = new byte[size];
//				for (int i=0; i<size; i++) {
//					RGB color = colors[i];
//					reds[i] = (byte)color.red;
//					greens[i] = (byte)color.green;
//					blues[i] = (byte)color.blue;
//				}
//
//				int tPixel;
//
//				tPixel = imageData.transparentPixel;
//				if (tPixel >= size) tPixel = -1;
//				colorModel = new IndexColorModel(imageData.depth, size, reds, greens, blues, tPixel);
//			}
		}
		return colorModel;
	}

	/**
	 * @see java.awt.Toolkit#getFontList()
	 */
	public String[] getFontList() {
		return Font.getFontList();
	}

	public java.awt.FontMetrics getFontMetrics(Font font) {
		return _getFontMetrics(font);
	}

	FontMetricsImpl _getFontMetrics(Font font) {
		return java.awt.FontMetricsImpl.createFromFont(font);
	}

	/**
	 * @see java.awt.Toolkit#sync()
	 */
	public void sync() {
		// NOP
	}

	/**
	 * @see java.awt.Toolkit#getImage(String)
	 */
	public Image getImage(String fileName) {
		if (null == fileName) throw new NullPointerException("fileName must not be null"); //$NON-NLS-1$
		checkFilePermissions(fileName);
		Image image;

		image = imageCache.getImage(""+fileName); //$NON-NLS-1$
		if (null != image) return image;

		image = createImage(fileName);

		imageCache.putImage(""+fileName,image); //$NON-NLS-1$

		return image;
	}

	/**
	 * @see java.awt.Toolkit#getImage(URL)
	 */
	public Image getImage(URL url) {
		if (null == url) throw new NullPointerException("url must not be null"); //$NON-NLS-1$
		checkURLPermission(url);
		String name = null;
		Image  image;

		if (null != url) name = url.toExternalForm();

		image = imageCache.getImage(""+name); //$NON-NLS-1$
		if (null != image) return image;

		image = createImage(url);

		imageCache.putImage(""+name,image); //$NON-NLS-1$
		return image;
	}

	void flushImageCache() {
		imageCache.flush();
	}

	private void checkFilePermissions(String fileName) {
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkRead(fileName);
		}
	}

	private void checkURLPermission(URL url) {
		SecurityManager sm = System.getSecurityManager();
		if ((url != null) && (sm != null)) {
			try {
				Permission perm = url.openConnection().getPermission();
				if (perm instanceof FilePermission) {
					sm.checkRead(perm.getName());
				} else if (perm instanceof SocketPermission) {
						sm.checkConnect(url.getHost(), url.getPort());
				} else {
					sm.checkPermission(perm);
				}
			} catch (IOException e) {
				sm.checkConnect(url.getHost(), url.getPort());
			}
		}
	}

	/**
	 * @see java.awt.Toolkit#createImage(String)
	 */
	public Image createImage(String fileName) {
		checkFilePermissions(fileName);
		return new ImageImpl(fileName);
	}

	/**
	 * @see java.awt.Toolkit#createImage(URL)
	 */
	public Image createImage(URL url) {
		checkURLPermission(url);
		return new ImageImpl(url);
	}
	/**
	 * @see java.awt.Toolkit#createImage(ImageProducer)
	 */
	public Image createImage(ImageProducer imageProducer) {
		return new ImageImpl(imageProducer);
	}

	/**
	 * @see java.awt.Toolkit#createImage(byte[], int, int)
	 */
	public Image createImage(byte[] data, int offset, int length) {
		return new ImageImpl(data, offset, length);
	}

	/**
	 * @see java.awt.Toolkit#prepareImage(Image, int, int, ImageObserver)
	 */
	public boolean prepareImage(Image image, int width, int height, ImageObserver imageObserver) {
		IImageImpl iImage= BufferedImageHandler.createIImageImpl(image);
		if (iImage != null) {
			if (iImage.getPrimitiveImagePeer() != null)
				return true;

			iImage.getImagePeer(imageObserver);
		}
		return false;
	}

	/**
	 * @see java.awt.Toolkit#checkImage(Image, int, int, ImageObserver)
	 */
	public int checkImage(Image image, int width, int height, ImageObserver imageObserver) {
		IImageImpl iImage= BufferedImageHandler.createIImageImpl(image);
		if (iImage != null) {
			if (iImage.getPrimitiveImagePeer() != null)
				return ImageObserver.ALLBITS;
		}
		return 0;
	}

	/**
	 * @see java.awt.Toolkit#beep()
	 */
	public void beep() {
		getDisplay().beep();
	}

	/**
	 * @see java.awt.Toolkit#getSystemClipboard()
	 */
	public synchronized Clipboard getSystemClipboard() {
		if (GraphicsEnvironment.isHeadless()) {
			throw new HeadlessException();
		}
		final SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkSystemClipboardAccess();
		}

		if (clipboard == null)
			clipboard = new ClipboardImpl();
		return clipboard;
	}

	public boolean isDisplayThread() {
		return display.isValidPeerThread();
	}

	/**
	 * @see java.awt.Toolkit#getSystemEventQueueImpl()
	 */
	protected EventQueue getSystemEventQueueImpl() {
		if (eventQueue == null)
			eventQueue = new EventQueue();
		return eventQueue;
	}
}
