package java.awt;
/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2006  All Rights Reserved
 */

import java.awt.image.BufferedImage;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.VolatileImage;
import java.util.Hashtable;

import com.ibm.oti.awt.GraphicsHandler;
import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.image.IImageImpl;
import com.ibm.oti.awt.image.VolatileImageProducer;
import com.ibm.oti.awt.metal.image.ImagePeer;
import com.ibm.oti.awt.metal.image.VolatileImagePeer;

class VolatileImageImpl extends VolatileImage implements IImageImpl {
	ImageCapabilities fImageCapabilities;
	GraphicsConfiguration fGraphicsConfiguration;
	Component fComponent;
	ImageProducer fProducer;
	VolatileImagePeer fImagePeer;
	int fWidth, fHeight;
	int fStatus;

	boolean fContentsLost;

	VolatileImageImpl() {
		super();
		this.fContentsLost = false;
		this.fStatus = IImageImpl.FINISHED;
	}

	VolatileImageImpl(Component component, int width, int height, Color fillColor, ImageCapabilities imageCapabilities) {
		this(component.getGraphicsConfiguration(), width, height, imageCapabilities, fillColor);
		this.fComponent = component;
	}

	VolatileImageImpl(GraphicsConfiguration gc, int width, int height, ImageCapabilities imageCapabilities, Color fillColor) {
		this();
		this.fComponent = null;
		this.fGraphicsConfiguration = gc;
		this.fWidth = width;
		this.fHeight = height;
		this.fImageCapabilities = imageCapabilities;
		this.fImagePeer = VolatileImagePeer.createVolatileImage(Util.getDisplay(), fWidth, fHeight, fillColor);
		this.fProducer = new VolatileImageProducer(fImagePeer, fGraphicsConfiguration.getColorModel(), width, height);
	}

	public BufferedImage getSnapshot() {
		BufferedImage snapshot = fGraphicsConfiguration.createCompatibleImage(getWidth(), getHeight());
		Graphics gc = snapshot.getGraphics();
		gc.drawImage(this, 0, 0, null);
		gc.dispose();
		return snapshot;
	}

	/**
	 * Answers the width of the receiver.
	 *
	 * @return int the width of the receiver
	 */
	public int getWidth() {
		return fWidth;
	}

	/**
	 * Answers the width of the receiver.
	 *
	 * @param imageObserver
	 *            ignored
	 *
	 * @return an <code>int</code> which is the receiver's width.
	 */
	public int getWidth(ImageObserver imageObserver) {
		return fWidth;
	}

	/**
	 * Answers the height of the receiver.
	 *
	 * @return int the height of the receiver
	 */
	public int getHeight() {
		return fHeight;
	}

	/**
	 * Answers the height of the receiver.
	 *
	 * @param imageObserver
	 *            ignored
	 *
	 * @return an <code>int</code> which is the receiver's height
	 */
	public int getHeight(ImageObserver imageObserver) {
		return fHeight;
	}

	public ImagePeer getImagePeer() {
		return fImagePeer;
	}

	VolatileImagePeer getVolatileImagePeer() {
		return (VolatileImagePeer)getImagePeer();
	}

	public ImagePeer getImagePeer(ImageObserver imageObserver) {
		return getImagePeer();
	}

	public void setProperties(Hashtable properties) {
		// This method should never be called
	}

	/**
	 * Answers a new <code>Graphics2D</code> instance used for drawing on the
	 * receiver.
	 *
	 * @author IBM
	 * @version initial
	 *
	 * @return a java.awt.Graphics2D for drawing
	 */
	public Graphics2D createGraphics() {
		try {
			if (fComponent != null) {
				return (Graphics2D) new GraphicsACImpl(fComponent, this);
			} else {
				return (Graphics2D) new GraphicsACImpl(fGraphicsConfiguration, this);
			}
		} catch (IllegalArgumentException e) {
			return null;
		}

	}

	public int validate(GraphicsConfiguration gc) {
		if ((gc != null) && (gc.getDevice() != fGraphicsConfiguration.getDevice())) {
			return IMAGE_INCOMPATIBLE;
		} else if (contentsLost() || (fImagePeer == null)) {
			getVolatileImagePeer().restoreDrawing();
			return IMAGE_RESTORED;
		} else {
			return IMAGE_OK;
		}
	}

	public boolean contentsLost() {
		return fContentsLost || (peerLostContent());
	}
	/*
	 * We must ask to the peer because the peer layer doesn't
	 * provide a callback informating about the lost.
	 */
	boolean peerLostContent() {
		return (fImagePeer == null)?false:fImagePeer.contentsLost();
	}

	public ImageCapabilities getCapabilities() {
		return fImageCapabilities;
	}

	/**
	 * Answers the property specified by the given name if it is
	 * known.  Otherwise answers <code>null</code>.
	 * <p>
	 * If the property is not defined for the receiver, answers
	 * <code>UndefinedProperty</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the desired property
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *								notification
	 *
	 * @return	an <code>Object</code> which is the property value or <code>null</code>
	 */
	public Object getProperty(String propertyName, ImageObserver imageObserver) {
//		if (fConsumer != null) {
//			Hashtable properties = fConsumer.getProperties();
//			if (properties != null) {
//				if (properties.containsKey(propertyName)) {
//					return properties.get(propertyName);
//				}
//			}
//		}

		return UndefinedProperty;
	}

	/**
	 * Releases all resources used by the receiver.
	 */
	public void flush() {
		disposeImage();
		// Even if the imagepeer was not created
		// the JDK after a flush returns true for a contentsLost()
		this.fContentsLost = true;
	}

	// IImageImpl apis
	public void disposeImage() {
		if (fImagePeer == null) return;
		fImagePeer.dispose();
		this.fImagePeer = null;
		this.fContentsLost = true;
	}

	public Component getComponent() {
		return fComponent;
	}

	public int getStatus() {
		return fStatus;
	}
	public void setStatus(int status) {
		fStatus = status;
	}

	public ImagePeer getPrimitiveImagePeer() {
		return fImagePeer;
	}

	public void notifyObservers(int infoflags, int x, int y, int width, int height) {
		// NOP
	}

	public void setImagePeer(ImagePeer imagePeer) {
		// Dispose any existing image before replacing.
		disposeImage();

		this.fImagePeer = (VolatileImagePeer)imagePeer;
		this.fContentsLost = (imagePeer == null);
	}
}
