package java.awt;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.event.*;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Locale;
import java.util.Vector;

import java.awt.event.WindowFocusListener;
import java.awt.im.InputContext;
import java.util.EventListener;
import java.util.Set;

import com.ibm.oti.awt.MessageStrings;
import com.ibm.oti.awt.Util;
import com.ibm.ugl.eswt.OS;

/**
 * A <code>Window</code> is a top-level widget that has no decorations.
 *
 * @author      IBM
 * @version		initial
 */
public class Window extends Container {

	static final long serialVersionUID = 4497834738069338734L;

	static final int OPEN = 1;
	static final int ACTIVE = 2;
	static final int ICONIC = 4;

	int windowSerializedDataVersion;
	String warningString;
	transient Vector windows;
	WindowListener windowListener;
	WindowFocusListener windowFocusListener;
	boolean focusableWindowState = true;
	transient Component mostRecentFocusOwner;
	transient InputContext inputContext;
	private static OppositeFieldDelay delayThread = null;
	static {
		delayThread = new OppositeFieldDelay();
		Thread t = new Thread(delayThread);
		t.setDaemon(true);
		t.start();
	}
	FocusManager focusMgr;
	int state;
	transient GraphicsConfiguration graphicsConfiguration = null;

	Window(GraphicsConfiguration config) {
		super(new BorderLayout());
		focusMgr = new FocusManager(this);
		visible = false;
		graphicsConfiguration = config;
		initWarningString();
	}

	/**
	 * Constructs a new <code>Window</code> with the specified
	 * parent.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the frame which is the parent of the receiver
	 *
	 * @throws	IllegalArgumentException	if the specified parent is <code>null</code>
	 */
	public Window(Frame parent) {
		this((Window)parent);
	}

	/**
	 * Constructs a new <code>Window</code> with the specified
	 * parent.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the window which is the parent of the receiver
	 *
	 * @throws	IllegalArgumentException	if the specified parent is <code>null</code>
	 */
	public Window(Window parent) {
		this(parent, GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration());
	}

	/**
	 * Constructs a new <code>Window</code> with the specified
	 * parent and graphics configuration.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	parent	the window which is the parent of the receiver
	 * @param	config	the graphics configuration
	 *
	 * @throws	IllegalArgumentException	if the specified parent is <code>null</code>
	 */
	public Window(Window parent, GraphicsConfiguration config) {
		this(config);

		if (GraphicsEnvironment.isHeadless()) {
			throw new IllegalArgumentException("Cannot create this instance in headless mode");
		}
		if (parent == null)
			throw new IllegalArgumentException();

		this.parent = parent;
		parent.addWindow(this);
	}

	/**
	 * @see java.awt.Component#isLightweight()
	 */
	public boolean isLightweight() {
		return false;
	}

	/**
	 * @see java.awt.Component#isOpaque()
	 */
	public boolean isOpaque() {
		return true;
	}

	/**
		 * Ignores resizes and moves in full screen behavior
		 *
		 * @author      IBM
		 * @version		initial
		 *
		 * @param	x		the x coordinate to move the receiver to.
		 * @param	y		the y coordinate to move the receiver to.
		 * @param	width	the width to resize the receiver to.
		 * @param	height	the height to resize the receiver to.
		 */
		public void setBounds(int x, int y, int width, int height) {
			if (!Util.useFullScreenWindowBehavior()) {
				super.setBounds(x, y, width, height);
			}
		}

	/**
	 * Sets the receiver's size to its preferred size and lays out
	 * its children.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void pack() {
		synchronized (getTreeLock()) {
			if (peer == 0)
				addNotify();

			invalidate();
			setSize(getPreferredSize());
			validate();
		}
	}

	/**
	 * Make the receiver visible. If the receiver's parent is showing,
	 * the receiver will show on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @see #setVisible(boolean)
	 */
	public void show() {
		// Re-implemented on this class, since inherited impl is deprecated
		super.show();
	}

	/**
	 * Hide the receiver so that it is no longer visible on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @see			#setVisible(boolean)
	 */
	public void hide() {
		if (null != windows) {
			for (int i=0; i<windows.size(); i++) {
				Window window = (Window) windows.elementAt(i);
				window.hide();
			}
		}

		super.hide();
	}

	void paintCallback(int x, int y, int width, int height) {
		// TODO Auto-generated method stub
		super.paintCallback(x, y, width, height);

		// Draw a warning string at the bottom of the Window if one should
		// be displayed
		if (warningString != null) {
			int font = OS.Control_GetDefaultFont(peer);
			int descent = OS.Font_GetDescent(font);
			int fontHeight = OS.Font_GetAscent(font) + OS.Font_GetLeading(font) + descent;

			int[] insets = OS.Scrollable_GetInsets(peer);
			int[] bounds = OS.Control_GetBounds(peer);

			int yPos = bounds[OS.INDEX_HEIGHT] - insets[OS.INDEX_BOTTOM] - insets[OS.INDEX_TOP] - fontHeight;

			int graphics = OS.Control_NewGraphics(peer);
			OS.Graphics_SetBackground(graphics, Color.yellow.getPeer());
			OS.Graphics_FillRectangle(graphics, 0, yPos-descent, bounds[OS.INDEX_WIDTH], fontHeight+descent);

			OS.Graphics_SetBackground(graphics, Color.black.getPeer());
			OS.Graphics_DrawString(graphics, warningString, 0, yPos-descent, OS.STRING_TRANSPARENT);

			OS.Graphics_Dispose(graphics);
		}
	}

	Insets getInsetsImpl() {
		Insets insets = super.getInsetsImpl();
		if (warningString != null) {
			acquirePeerReadLock();
			try {
				if (peer != 0) {
					int font = OS.Control_GetDefaultFont(peer);
					int fontHeight = OS.Font_GetAscent(font) + OS.Font_GetDescent(font) + OS.Font_GetLeading(font);

					insets = (Insets)insets.clone();
					insets.bottom += fontHeight;
				}
			} finally {
				releasePeerReadLock();
			}
		}
		return insets;
	}

	protected void finalize() throws Throwable {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				dispose();
			}
		});
	}

	/**
	 * Destroy the receiver, its children and all resources used
	 * by them.  The receiver can still be recreated.
	 *
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void dispose() {
		synchronized (getTreeLock()) {
			delayThread.postQueuedDeactivatedEvent();
			if (windows != null) {
				for (java.util.Enumeration enumeration = ((Vector) windows.clone()).elements(); enumeration.hasMoreElements();) {
					Window child = (Window) enumeration.nextElement();
					child.dispose();
				}
			}

			if (!isPeerNull()) { // For FileDialog whose peer is a FileDialogPeer
//			if ( peer != null ) {
				_closeWindow();

				removeNotify();
			}

			postEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSED));
		}
	}

	void _closeWindow() {
		setVisible(false);
		if (peer != 0) {
			OS.Shell_Close(peer);
		}
	}
	/**
	 * Brings the receiver to the front of the screen.  The receiver
	 * is displayed in front of all windows on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void toFront() {
		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.Control_MoveAbove(peer, 0);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	synchronized void removeWindow(Window window) {
		if (windows!= null && windows.removeElement(window)) {
			if (windows.size() == 0) {
				windows = null;
			}
		}
	}

	public void removeNotify() {

		if (null != windows) {
			for (int i=0; i<windows.size(); i++) {
				Window window;

				window = (Window) windows.elementAt(i);
				// This allows removeNotify to be called multiple times without
				// impacting cleanup events such as WINDOW_CLOSED
				if (!window.isPeerNull()) {
					window.removeNotify();
				}
			}
		}

//		if (peer != null) {
//			((WindowPeer) peer).aboutToDispose();
//		}

		super.removeNotify();

		//focusMgr = null;
		//windows = null;
		//windowListener = null;
	}

	void _setPeerBounds(int x1, int y1, int width1, int height1) {
		// Never translate the coordinates because Window location is
		// relative to the Display.
		if (peer != 0) {
			OS.Control_SetBounds(peer, x1,y1, width1, height1);
		}
	}

	/**
	 * Sends the receiver to the back of the screen.  The receiver
	 * is displayed behind all windows on the screen.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public void toBack() {
		acquirePeerReadLock();
		try {
			if (peer != 0) {
				OS.Control_MoveBelow(peer, 0);
			}
		} finally {
			releasePeerReadLock();
		}
	}

	void adjustZOrder() {
		// NOP
	}

	/**
	 * Answer the <code>Toolkit</code> for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Toolkit which is the receiver's toolkit
	 */
	public Toolkit getToolkit() {
		return Toolkit.getDefaultToolkit();
	}

	private void initWarningString() {
		warningString = null;

		SecurityManager securityMgr = System.getSecurityManager();
		if (null == securityMgr) return;
		if (!securityMgr.checkTopLevelWindow(this)) {
			warningString = (String) AccessController.doPrivileged(new PrivilegedAction() {
				public Object run() {
					// We must return a default value != null do create a clear
					// distinction between:
					// 	- the value 'null' ==> the window is "secure" and
					//	- a String not null ==> the window is insecure!
					return System.getProperty("awt.appletWarning", MessageStrings.getString("BBAWTUnsignedJavaAppletWindow")); //$NON-NLS-1$  //$NON-NLS-2$
				}
			});
		}
	}

	/**
	 * Answer the warning string.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String
	 */
	public final String getWarningString() {
		return warningString;
	}

	Window getWindow() {
		return this;
	}

	/**
	 * Answers the <code>Locale</code> for the receiver or if
	 * one was not specified, answer the default locale.
	 *
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.util.Locale
	 */
	public Locale getLocale() {
		if (locale != null)
			return locale;

		return Locale.getDefault();
	}

	/**
	 * Adds a <code>WindowListener</code> to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		windowListener	the <code>WindowListener</code> to be added.
	 *
	 * @see			java.awt.event.WindowListener
	 * @see			java.awt.event.WindowEvent
	 */
	public synchronized void addWindowListener(WindowListener windowListener) {
		addEventListener(windowListener);
		this.windowListener = AWTEventMulticaster.add(this.windowListener, windowListener);
	}

	/**
	 * Removes a <code>WindowListener</code> from the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		windowListener	the <code>WindowListener</code> to be removed.
	 *
	 * @see			java.awt.event.WindowListener
	 * @see			java.awt.event.WindowEvent
	 */
	public synchronized void removeWindowListener(WindowListener windowListener) {
		this.windowListener = AWTEventMulticaster.remove(this.windowListener, windowListener);
	}

	/**
	 * Called when an <code>AWTEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call <code>super.processEvent(AWTEvent)</code>.
	 * <p>
	 * The default event handled is <code>WindowEvent</code>, the receiver
	 * will forward calls to the default event by calling
	 * <code>processWindowEvent(FocusEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	awtEvent	the event to be processed
	 */
	protected void processEvent(AWTEvent awtEvent) {
		if (!isEventEnabled(awtEvent)) return;
		if (awtEvent instanceof WindowEvent) {
		    int id = awtEvent.getID();
		    if ((id == WindowEvent.WINDOW_GAINED_FOCUS) || (id == WindowEvent.WINDOW_LOST_FOCUS)) {
		        processWindowFocusEvent((WindowEvent) awtEvent);
		    }
			processWindowEvent((WindowEvent) awtEvent);
		} else {
			super.processEvent(awtEvent);
		}
	}

	/**
	 * Called when an <code>WindowEvent</code> occurs on the receiver.
	 * <p>
	 * Events must be enabled in order for a component to receive them.
	 * Events can be enabled using <code>enableEvents(long)</code> or by
	 * adding listeners for that event.
	 * <p>
	 * If a component is unable to get sufficient behaviour by adding
	 * listeners then it must override this method to do custom event handling.
	 * When overriding it should call
	 * <code>super.processWindowEvent(WindowEvent)</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	windowEvent	the event to be processed
	 *
	 * @see		WindowEvent
	 */
	protected void processWindowEvent(WindowEvent windowEvent) {
		if (windowListener == null)
			return;
		switch (windowEvent.getID()) {
			case WindowEvent.WINDOW_OPENED:
				windowListener.windowOpened(windowEvent);
				break;
			case WindowEvent.WINDOW_CLOSING:
				windowListener.windowClosing(windowEvent);
				break;
			case WindowEvent.WINDOW_CLOSED:
				windowListener.windowClosed(windowEvent);
				break;
			case WindowEvent.WINDOW_ICONIFIED:
				windowListener.windowIconified(windowEvent);
				break;
			case WindowEvent.WINDOW_DEICONIFIED:
				windowListener.windowDeiconified(windowEvent);
				break;
			case WindowEvent.WINDOW_ACTIVATED:
				windowListener.windowActivated(windowEvent);
				break;
			case WindowEvent.WINDOW_DEACTIVATED:
				windowListener.windowDeactivated(windowEvent);
				break;
			default:
				return;
		}
	}

	void dispatchEventImpl(AWTEvent awtEvent) {
		if (awtEvent instanceof ComponentEvent) {
			if (awtEvent.getID() == ComponentEvent.COMPONENT_RESIZED) {
				invalidate();
				validate();
			}
		}
		if (awtEvent instanceof WindowEvent) {
			switch (awtEvent.getID()) {
				case WindowEvent.WINDOW_OPENED:
				    state |= Window.OPEN;
					break;
				case WindowEvent.WINDOW_CLOSED:
				    state &= ~Window.OPEN;
					break;
				case WindowEvent.WINDOW_ACTIVATED:
				    state |= Window.ACTIVE;
					break;
				case WindowEvent.WINDOW_DEACTIVATED:
				    state &= ~Window.ACTIVE;
					break;
				case WindowEvent.WINDOW_ICONIFIED:
				    state |= Window.ICONIC;
					break;
				case WindowEvent.WINDOW_DEICONIFIED:
				    state &= ~Window.ICONIC;
					break;
				default:
					break;
			}
		}
		super.dispatchEventImpl(awtEvent);
	}

	/**
	 * Answers the child of the receiver that has keyboard focus.  If
	 * the receiver is not the active window, the return value is
	 * <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>Component</code> which is the focus owner or
	 * 			<code>null</code>
	 */
	public Component getFocusOwner() {
	    if (isFocused() && isFocusableWindow()) {
	    	return KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
	    } else {
	        return null;
	    }
	}

	FocusManager getFocusManager() {
		return focusMgr;
	}

	/**
	 * Post the event to the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		event the event to be posted
	 *
	 * @deprecated	use <code>dispatchEvent(AWTEvent)</code> instead.
	 *
	 * @see 		Component#dispatchEvent(java.awt.AWTEvent)
	 */
	public boolean postEvent(Event event) {
		// Do not propagate event up in the hierarchy
		return handleEvent(event);
	}
	/**
	 *	If the widget is created, do nothing - the OS will generate resize/move
	 *  events for Windows in this instance.
	 */
	void _postEventsOnResize(boolean moved, boolean resized) {
		if (!isDisplayable()) super._postEventsOnResize(moved, resized);
	}

	/**
	 * Answer whether the receiver is visible and its parent is
	 * showing.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		<code>true</code> if the receiver is showing,
	 *				<code>false</code> otherwise
	 */
	public boolean isShowing() {
		return isVisible();
	}

	boolean isEventEnabled(AWTEvent awtEvent) {
		if (awtEvent instanceof WindowEvent) {
			boolean focusEnabled = ((eventMask & AWTEvent.WINDOW_FOCUS_EVENT_MASK) != 0) || windowFocusListener != null;
			if (focusEnabled) return true;
		}
		if (awtEvent instanceof WindowEvent)
			return ((eventMask & AWTEvent.WINDOW_EVENT_MASK) != 0) || windowListener != null;
		else
			return super.isEventEnabled(awtEvent);
	}

	public GraphicsConfiguration getGraphicsConfiguration() {
		// Although the PPro spec indicates that this method can return null,
		// tests against the RI show that Windows return the default configuration
		// even if null was explicitly provided via the constructor.
		if (graphicsConfiguration != null) {
			return graphicsConfiguration;
		} else {
			return GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration();
		}
	}
	private void resizeCallback(int width, int height) {
		_postResizeEvent(width, height);	// proceed with posting the resize event
	}
	private void moveCallback(int x, int y) {
		_postMoveEvent(x, y);
	}

	void shellCallback(int code) {
		int windowEventCode = 0;
		Window opposite = null;
		switch (code) {
			case OS.EVENT_SHELL_ACTIVATED: {
				windowEventCode = WindowEvent.WINDOW_ACTIVATED;
				if (delayThread.getDeactivatedEvent() != null) {
					opposite = delayThread.getDeactivatedEvent().getWindow();
				} else {
					opposite = KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusedWindow();
				}
				break;
			}
			case OS.EVENT_SHELL_DEACTIVATED: {
				windowEventCode = WindowEvent.WINDOW_DEACTIVATED;
				opposite = KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusedWindow();
				break;
			}
			case OS.EVENT_SHELL_CLOSING: windowEventCode = WindowEvent.WINDOW_CLOSING; break;
			case OS.EVENT_SHELL_ICONIFIED: windowEventCode = WindowEvent.WINDOW_ICONIFIED; break;
			case OS.EVENT_SHELL_DEICONIFIED: windowEventCode = WindowEvent.WINDOW_DEICONIFIED; break;
			default: return;
		}

		WindowEvent event = new WindowEvent(this, windowEventCode, opposite);
		if (windowEventCode == WindowEvent.WINDOW_DEACTIVATED) {
			delayThread.queueDeactivatedEvent(event);
			return;
		} else if (windowEventCode == WindowEvent.WINDOW_ACTIVATED) {
			// this seems to work ok since deactivated precedes activated
			postLastDeactivatedEvent(this);
			if (opposite == this) {
				// This happens when you click away to an OS window and then click back
				event = new WindowEvent(this, windowEventCode, null);
			}
		}
		if (KeyboardFocusManager.getDebug()) {
			String type = " activated ";
			if (code == WindowEvent.WINDOW_DEACTIVATED) type = " deactivated ";
			System.out.println("shell callback" + type + this.getName());
		}
		postEvent(event);
	}

//
//	void makeFullScreenIfNecessary() {
//		if (!Util.useFullScreenWindowBehavior()) {
//			return; // not necessary
//		}
//		WindowPeer windowPeer = (WindowPeer) peer;
//		windowPeer.makeFullScreenWindow();
//
//		// update common layer information from peer
//		Rectangle peerBounds = windowPeer.getBounds();
//		this.x = peerBounds.x;
//		this.y = peerBounds.y;
//		this.width = peerBounds.width;
//		this.height = peerBounds.height;
//	}

	int _createPeer(Container hwParent) throws AWTError {
		synchronized (hwParent.getTreeLock()) {
			if (hwParent.peer == 0) {
				hwParent.addNotify();
			}

			return _newPeer(hwParent);
		}
	}

	int _newPeer(Container hwParent) {
		return OS.Shell_NewChild(hwParent.peer, OS.STYLE_NO_TRIM);
	}

	void removeFocus() {
		// Removing focus will happen automaticly by the OS
		// when it needs to for windows, dialogs and frames
	}

	boolean transferFocus(boolean next) {
		return requestFocusChange(next);
	}

	void transferFocusToImmediateChild(boolean next) {
		Component cmpo = getFocusTraversalPolicy().getInitialComponent(this);
		if (cmpo == null) cmpo = this;
		cmpo.requestFocus();
	}

	void _show(boolean visible) {

			if (visible && (peer == 0))
				addNotify();

			// Validate the widget after all
			// of the underlying native widgets
			// have been created but before
			// the setVisible() call so that there
			// is no flashing on the initial layout
			if (visible) validate();

			if (visible != isVisible()) {
				super._show(visible);
			}

			if (visible) {
				toFront();
				if ((state & OPEN) == 0) {
					// When the Window is opened for the first time the first
					// focusable widget should get focus.  However, if the
					// application registers and event listener for something
					// like window activated or window focused then the application
					// might request focus before the code gets here.  To avoid
					// transfering focus away from the application requested controls
					// only transfer focus if no control has focus.  The PPro 1.1 TCK
					// test EventDelivery_lw_0004 tests for this behavior
					FocusManager manager = getFocusManager();
					synchronized (manager.getRequestedLock()) {
						if (manager.getLastRequested() == null) {
							transferFocusToImmediateChild(true);
						}
					}

					postEvent(new WindowEvent(this, WindowEvent.WINDOW_OPENED));
				}
			}
	}

	void _postResizeEvent(int width, int height) {
		if ((this.width == width) && (this.height == height)) return;
		this.width = width;
		this.height = height;

		// When frame is resized, the menu bar height can change,
		// thus breaking our cached insets value.  Fix is to
		// drop our cached value so that it will be re-calculated
		// when next needed
		insets = null;

		ComponentEvent event = new ComponentEvent(this, ComponentEvent.COMPONENT_RESIZED);
		postEvent(event);
	}

	synchronized void addWindow(Window window) {
		if (windows == null)
			windows = new Vector();

		windows.addElement(window);
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		AWTEventMulticaster.save(stream, windowListenerK, windowListener);
		stream.writeObject(null);
	}

	private void readObject(ObjectInputStream stream)
	throws IOException, ClassNotFoundException
	, HeadlessException
	{
		if (GraphicsEnvironment.isHeadless()) throw new HeadlessException();
		stream.defaultReadObject();
		String key = null;
		while ((key = (String)stream.readObject()) != null) {
			Object listener = stream.readObject();
			if (key.equals(windowListenerK))
				addWindowListener((WindowListener)listener);
		}
	}

	/**
	 * @param x
	 * @param y
	 */
	void _postMoveEvent(int x, int y) {
		// do not post the move event if the location of the
		// Component has not changed
		if ((this.x == x) && (this.y == y)) return;
		this.x = x;
		this.y = y;
		ComponentEvent event = new ComponentEvent(this, ComponentEvent.COMPONENT_MOVED);
		postEvent(event);
	}

	String classNonlocalizedName() {
		return "win"; //$NON-NLS-1$
	}
	/**
	 * Do not cache bounds when the native widget exists for a Window since
	 * the OS will return the size via the native resize/move callbacks.  If
	 * the native widget does not exist, the OS cannot give us the widget
	 * size/location, so cache the bounds in this instance.
	 */
	void cacheBounds(int x, int y, int width, int height) {
		if (!isDisplayable()) super.cacheBounds(x, y, width, height);
	}

	public InputContext getInputContext() {
		if (inputContext == null) {
			inputContext = InputContext.getInstance();
		}
		return inputContext;
	}

	public
	Window getOwner() {
		return (Window) getParent();
	}

	Component getRefocusComponent() {
		Window owner = getOwner();
		// search the owner chain for the component to give focus to
		if (owner != null) {
			Component cmpo = owner.getMostRecentFocusOwner();
			if (cmpo != null) {
				return cmpo;
			} else {
				return owner.getRefocusComponent();
			}
		}
		return null;
	}
	public void addWindowFocusListener(WindowFocusListener windowFocusListener) {
		addEventListener(windowListener);
		this.windowFocusListener = AWTEventMulticaster.add(this.windowFocusListener, windowFocusListener);
	}

	public void removeWindowFocusListener(WindowFocusListener windowFocusListener) {
		this.windowFocusListener = AWTEventMulticaster.remove(this.windowFocusListener, windowFocusListener);
	}

	public WindowListener[] getWindowListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(windowListener, WindowListener.class);
		WindowListener[] resultArray = new WindowListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	public WindowFocusListener[] getWindowFocusListeners() {
		EventListener[] listeners = AWTEventMulticaster.getListeners(windowFocusListener, WindowFocusListener.class);
		WindowFocusListener[] resultArray = new WindowFocusListener[listeners.length];
		System.arraycopy(listeners, 0, resultArray, 0, listeners.length);
		return resultArray;
	}

	protected void processWindowFocusEvent(WindowEvent windowEvent) {
	    if (windowFocusListener == null)
			return;
		switch (windowEvent.getID()) {
			case WindowEvent.WINDOW_GAINED_FOCUS:
			    windowFocusListener.windowGainedFocus(windowEvent);
				break;
			case WindowEvent.WINDOW_LOST_FOCUS:
			    windowFocusListener.windowLostFocus(windowEvent);
				break;
			default:
				return;
		}
	}

	/**
	 * Returns the child Component of this Window that will receive the focus
	 * when this Window is focused. If this Window is not focused, then the
	 * child Component that most recently requested focus will be returned.
	 * If no child Component has ever requested focus, and this is a
	 * focusable Window, then this Window's initial focusable Component is
	 * returned. If no child Component has ever requested focus, and this is
	 * a non-focusable Window, null is returned.
	 */
	public Component getMostRecentFocusOwner() {
	    if (isFocused()) {
	        // the window has focus
	        return getFocusOwner();
	    }
	    if (!isFocusableWindow()) return null;
	    return this.mostRecentFocusOwner;
	}
	void setMostRecentFocusOwner(Component mostRecentFocusOwner) {
	    this.mostRecentFocusOwner = mostRecentFocusOwner;
	}
	/**
	 * Returns whether this Window is active. Only a Frame or a Dialog
	 * may be active. If the focused Window is a Frame or a Dialog it
	 * is also the active Window. Otherwise, the active Window is the
	 * first Frame or Dialog that is an owner of the focused Window.
	 */
	public boolean isActive() {
        Window window = KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow();
        if (window == null) return false;
	    if ((this instanceof Frame)
	    	|| (this instanceof Dialog)
	    	) {
	        return this == window;
	    } else {
	        return this == window.getOwner();
	    }
	}
	/**
	 * Returns whether this Window is focused. If there exists a focus owner, the
	 * focused Window is the Window that is, or contains, that focus owner. If
	 * there is no focus owner, then no Window is focused.
	 */
	public boolean isFocused() {
	   	return ((this == KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusedWindow()) & isShowing());
	}

	public Set getFocusTraversalKeys(int id) {
	    if (!isValidTraversalKey(id)) throw new IllegalArgumentException("invalid focus traversal key identifier");
		Set keys = focusTraversalKeys[id];
		if (keys != null) return keys;
		return getDefaultFocusTraversalKeys(id);
	}

	public final void setFocusCycleRoot(boolean rootValue) {
	    // no-op, windows must always be focus cycle roots
	}

	public final boolean isFocusCycleRoot() {
	    // windows must always be focus cycle roots
		return true;
	}

	public final Container getFocusCycleRootAncestor() {
		return null;
	}

	/**
	 * Returns whether this Window can become the focused Window, that
	 * is, whether this Window or any of its subcomponents can become
	 * the focus owner. For a Frame or Dialog to be focusable, its
	 * focusable Window state must be set to true. For a Window which
	 * is not a Frame or Dialog to be focusable, its focusable Window
	 * state must be set to true, its nearest owning Frame or Dialog
	 * must be showing on the screen, and it must contain at least one
	 * Component in its focus traversal cycle. If any of these
	 * conditions is not met, then neither this Window nor any of its
	 * subcomponents can become the focus owner.
	 */
	public final boolean isFocusableWindow() {
	    if ((this instanceof Frame)
	    	|| (this instanceof Dialog)
	    	) {
	        return getFocusableWindowState();
	    } else {
			Window owner = getOwner();
			if (getFocusableWindowState() && (owner != null) && owner.isShowing()) {
			    Component cmpo = getFocusTraversalPolicy().getFirstComponent(this);
			    return cmpo != null;
			}
			return false;
	    }
	}

	public boolean getFocusableWindowState() {
		return focusableWindowState;
	}

	public void setFocusableWindowState(boolean focusableWindowState) {
		KeyboardFocusManager manager = KeyboardFocusManager.getCurrentKeyboardFocusManager();
		boolean refocus = this.focusableWindowState && !focusableWindowState && (manager.getFocusedWindow() == this);
		// Only Dialogs and Frames will ever be the active windows.  For windows see if the owner is active.
		boolean gainFocus = false;
		if ((this instanceof Dialog) || (this instanceof Frame))
		{
			gainFocus = !this.focusableWindowState && focusableWindowState && (manager.getGlobalActiveWindow() == this);
		} else {
			gainFocus = !this.focusableWindowState && focusableWindowState && (manager.getGlobalActiveWindow() == this.getOwner());
		}
		boolean oldState = this.focusableWindowState;
		this.focusableWindowState = focusableWindowState;
		changeSupport.firePropertyChange("focusableWindowState", new Boolean(oldState), new Boolean(focusableWindowState));
		if (refocus) {
			// If the focused Window is made non-focusable, then attempt to focus
			// the most recently focused Component of the Window's owner.
			Component cmpo = getRefocusComponent();
			if (cmpo != null) {
				cmpo.requestFocus();
			}
		}
		// If we have an active window that is made focusable, send the window gained focus
		// event.
		if (gainFocus) {
			WindowEvent event = new WindowEvent(this, WindowEvent.WINDOW_GAINED_FOCUS, manager.getFocusedWindow());
			manager.dispatchEvent(event);
		}
	}

	static void postLastDeactivatedEvent(Window window) {
		delayThread.activatedWindow(window);
	}

	static class OppositeFieldDelay implements Runnable {

		private WindowEvent deactivatedEvent;
		public final Object lock = new Object();

		public void run() {
			try {
				while (true) {
					synchronized (lock) {
						lock.wait(1000); // Waiting for notify from allowQueuedDeactivatedEventToPost()
						postQueuedDeactivatedEvent();
						lock.notifyAll(); // Notify waiter in waitForDeactivatedEventToFinishPosting()
					}
				}
			} catch (InterruptedException e) {
			}
		}

		public WindowEvent getDeactivatedEvent() {
			synchronized (this) {
				return deactivatedEvent;
			}
		}

		private void postQueuedDeactivatedEvent() {
			// If one has been queued, we will post to the AWT event queue
			WindowEvent event = getDeactivatedEvent();
			if (event == null) return;  // Early return if nothing to post
			Window source = event.getWindow();
			if (source == event.getOppositeWindow()) {
				event = new WindowEvent(source, event.getID(), null);
			}
			source.postEvent(event);
			setDeactivatedEvent(null);
		}

		public void queueDeactivatedEvent(WindowEvent event) {
			// A deactivated event has come in from the OS
			postQueuedDeactivatedEvent();
			setDeactivatedEvent(event);
		}

		public void activatedWindow(Window window) {
			// An activated event has come in from the OS.
			// Clean up and post the queued deactivated event.
			WindowEvent deactivatedEvent = getDeactivatedEvent();
			if (deactivatedEvent == null) return;  // Nothing queued, we can leave
			// Set the opposite field correctly with the newly activated window
			deactivatedEvent = new WindowEvent(deactivatedEvent.getWindow(), deactivatedEvent.getID(), window);
			setDeactivatedEvent(deactivatedEvent);
			allowQueuedDeactivatedEventToPost();
			waitForDeactivatedEventToFinishPosting();
		}

		private void waitForDeactivatedEventToFinishPosting() {
			synchronized (lock) {
				try {
					lock.wait(); // Wait for deactivate event to post
				} catch (InterruptedException e) {
				}
			}
		}

		private void allowQueuedDeactivatedEventToPost() {
			synchronized (lock) {
				lock.notifyAll();  // Notifying waiter in run()
			}
		}

		private void setDeactivatedEvent(WindowEvent event) {
			synchronized (this) {
				deactivatedEvent = event;
			}
		}
	}

}
