package java.awt.datatransfer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;

/**
 *
 */
public class DataFlavor implements Externalizable, Cloneable {

	static final long   serialVersionUID         = 8367026044764648243L;
	static final String serializedObjectMimeType = "application/x-java-serialized-object"; //$NON-NLS-1$

	Class  representationClass;
	String mimeType;
	String humanPresentableName;

	/**
	 *
	 */
	public static final DataFlavor stringFlavor;

	/**
	 *
	 * @deprecated
	 */
	public static final DataFlavor plainTextFlavor;

	/**
	 *
	 */
	static {
		stringFlavor    = new DataFlavor(String.class, "Unicode String"); //$NON-NLS-1$
		plainTextFlavor = new DataFlavor("text/plain; charset=unicode", "Plain Text"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 *
	 */
	public DataFlavor() {
		super();
	}

	/**
	 *
	 */
	public DataFlavor(Class representationClass, String humanPresentableName) {
		super();

		this.mimeType            = serializedObjectMimeType + "; class="+representationClass.getName(); //$NON-NLS-1$
		this.representationClass = representationClass;

		initializeHumanPresentableName(humanPresentableName);
	}

	/**
	 *
	 */
	public DataFlavor(String mimeType, String humanPresentableName) {
		super();

		int slash = mimeType.indexOf('/');
		if ((slash == -1) || (slash != mimeType.lastIndexOf('/'))) {
			throw new IllegalArgumentException("failed to parse:"+mimeType); //$NON-NLS-1$
		}

		this.mimeType = mimeType;
		initializeRepresentationClass();
		initializeHumanPresentableName(humanPresentableName);
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
		buffer.append(getClass().getName());
		buffer.append("[representationClass="); //$NON-NLS-1$
		buffer.append(getRepresentationClass().getClass().getName());
		buffer.append(",mimeType="); //$NON-NLS-1$
		buffer.append(getMimeType());
		buffer.append("]"); //$NON-NLS-1$
		return buffer.toString();
	}

	/**
	 *
	 */
	public String getMimeType() {
		return mimeType;
	}

	/**
	 * Answer the part of the mime type up to the slash ('/') character
	 */
	public String getPrimaryType() {
		String humanName = extractMimeType(mimeType);
		int slash = humanName.lastIndexOf('/');
		return humanName.substring(0, slash);
	}

	/**
	 * Answer the part of the mime type after the slash ('/') and before the semicolon (';')
	 */
	public String getSubType() {
		String humanName = extractMimeType(mimeType);
		int slash = humanName.lastIndexOf('/');
		return humanName.substring(slash+1);
	}

	/**
	 *
	 */
	public Class getRepresentationClass() {
		return representationClass;
	}

	/**
	 *
	 */
	public String getHumanPresentableName() {
		return humanPresentableName;
	}

	/**
	 *
	 */
	public void setHumanPresentableName(String humanPresentableName) {
		//if (humanPresentableName == null) return;
		this.humanPresentableName = humanPresentableName;
	}

	/**
	 *
	 */
	public boolean equals(Object object) {
		if (object instanceof DataFlavor)
			return equals((DataFlavor) object);
		return false;
	}

	/**
	 *
	 */
	public boolean equals(DataFlavor dataFlavor) {
		if (dataFlavor == this) return true;
		if (dataFlavor == null) return false;

		return isMimeTypeEqualIncludingCharset(dataFlavor) &&
			getRepresentationClass() == dataFlavor.getRepresentationClass();
	}

	/**
	 *
	 */
	public int hashCode() {
		return getMimeType().hashCode() + getRepresentationClass().hashCode();
	}

	/**
	 *
	 */
	public boolean match(DataFlavor dataFlavor) {
		if (dataFlavor == null) return false;
		return equals((DataFlavor) dataFlavor);
	}

	/**
	 *
	 */
	public boolean isMimeTypeEqual(String mimeTypeArg) {
		return extractMimeType(getMimeType()).equals(extractMimeType(mimeTypeArg));
	}

	/**
	 *
	 */
	public final boolean isMimeTypeEqual(DataFlavor dataFlavor) {
		return isMimeTypeEqual(dataFlavor.getMimeType());
	}

	/**
	 *
	 */
	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeUTF(mimeType);
		out.writeUTF(humanPresentableName);
		out.writeUTF(representationClass.getName());
	}

	/**
	 *
	 */
	public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
		mimeType             = in.readUTF();
		humanPresentableName = in.readUTF();
		String repClassName  = in.readUTF();

		if (repClassName.length() != 0) {
			representationClass = Class.forName(repClassName);
		}
	}

	/**
	 *
	 */
	public Object clone() throws CloneNotSupportedException {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			return null;
		}
	}

	/**
	 *
	 * @deprecated
	 */
	protected String normalizeMimeTypeParameter(String name, String value) {
		return value;
	}

	/**
	 *
	 * @deprecated
	 */
	protected String normalizeMimeType(String mimeType) {
		return mimeType;
	}

	/**
	 *
	 */
	private void initializeHumanPresentableName(String humanPresentableName) {
		// If the name is null, we set it to the actual mime type
		setHumanPresentableName(
			humanPresentableName == null ? extractMimeType(mimeType) : humanPresentableName
			);
	}

	/**
	 *
	 */
	private String extractMimeType(String mimeTypeArg) {
		// Answer the mime type from the input string, removing the
		// optional args from the end
		int semicolon = mimeTypeArg.indexOf(';');
		if (semicolon != -1) {
			mimeTypeArg = mimeTypeArg.substring(0, semicolon);
		}
		return mimeTypeArg;
	}

	/**
	 *
	 */
	private void initializeRepresentationClass() {
		String repClassName = "java.io.InputStream"; //$NON-NLS-1$
		int classLoc = mimeType.indexOf("class="); //$NON-NLS-1$
		if (classLoc == -1) {
			if (mimeType.startsWith(serializedObjectMimeType)) {
				throw new IllegalArgumentException("no representation class specified for:"+mimeType); //$NON-NLS-1$
			}
			mimeType += "; class=" + repClassName; //$NON-NLS-1$
		} else {
			String rest = mimeType.substring(classLoc+"class=".length()); //$NON-NLS-1$
			int semiColon = rest.indexOf(';');
			if (semiColon == -1) {
				repClassName = rest;
			} else {
				repClassName = rest.substring(0, semiColon);
			}

		}

		try {
			this.representationClass = Class.forName(repClassName); //$NON-NLS-1$
		} catch (ClassNotFoundException e) {
			throw new IllegalArgumentException("could not locate representationClass "+repClassName); //$NON-NLS-1$
		}
	}

	/**
	 *
	 */
	private boolean isMimeTypeEqualIncludingCharset(DataFlavor df2) {
		boolean initEqual = isMimeTypeEqual(df2);
		if (!initEqual) return false;
		if (extractMimeType(getMimeType()).equals("text/plain")) { //$NON-NLS-1$
			return getMimeParameter("charset").equals(df2.getMimeParameter("charset")); //$NON-NLS-1$ //$NON-NLS-2$
		}
		return initEqual;
	}

	/**
	 *
	 */
	private String getMimeParameter(String name) {
		int parameterIndex = getMimeType().indexOf(name+"="); //$NON-NLS-1$
		if (parameterIndex == -1) return null;
		int valueIndex = parameterIndex + (name+"=").length(); //$NON-NLS-1$
		String remaining = getMimeType().substring(valueIndex);
		int nextSemiColon = remaining.indexOf(';');
		if (nextSemiColon == -1) nextSemiColon = remaining.length();
		String result = remaining.substring(0, nextSemiColon);
		return result;
	}

}
