package java.awt.datatransfer;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.IOException;
import java.io.StringReader;

/**
 * This class implements a simple way of transferring a
 * <code>java.lang.String</code>.
 *
 * @see java.awt.datatransfer.Transferable
 */
public class StringSelection implements Transferable, ClipboardOwner {

	String contents;

	/**
	 * Constructs a new <code>StringSelection</code> which
	 * contains the specifed <code>String</code>.
	 *
	 * @param	contents	the receiver's contents
	 */
	public StringSelection(String contents) {
		this.contents = contents;
	}

	/**
	 * Answers the data transfered using the speficied <code>DataFlavor</code>.
	 * <p>
	 * The receiver supports the following data flavors:
	 * 	<code>DataFlavor.plainTextFlavor</code>
	 * 	<code>DataFlavor.stringFlavor</code>
	 *
	 * @param	dataFlavor	the desired flavor to retrieve the data transfered
	 *
	 * @return	an <code>Object</code> which is the data transfered
	 *
	 * @see		DataFlavor
	 */
	public synchronized Object getTransferData(DataFlavor dataFlavor)
		throws UnsupportedFlavorException, IOException {
		if (dataFlavor.equals(DataFlavor.stringFlavor))
			return contents;
		else if (dataFlavor.equals(DataFlavor.plainTextFlavor))
			return new StringReader(contents);
		else
			throw new UnsupportedFlavorException(dataFlavor);
	}

	/**
	 * Answers all the data flavors supported by the receiver.
	 * <p>
	 * The receiver supports the following data flavors:
	 * 	<code>DataFlavor.plainTextFlavor</code>
	 * 	<code>DataFlavor.stringFlavor</code>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>DataFlavor</code> array containing all the data flavors
	 * 				support
	 *
	 * @see		DataFlavor
	 */
	public synchronized DataFlavor[] getTransferDataFlavors() {
		return new DataFlavor[] {
			DataFlavor.stringFlavor,
			DataFlavor.plainTextFlavor };
	}

	/**
	 * Answers whether the specified <code>DataFlavor</code>
	 * is supported.
	 * <p>
	 * The receiver supports the following data flavors:
	 * 	<code>DataFlavor.plainTextFlavor</code>
	 * 	<code>DataFlavor.stringFlavor</code>
	 *
	 * @param	dataFlavor	the <code>DataFlavor</code> to be checked
	 *
	 * @return	<code>true</code> if the <code>DataFlavor</code> is supported,
	 *			<code>false</code> otherwise
	 *
	 * @see		DataFlavor
	 */
	public boolean isDataFlavorSupported(DataFlavor dataFlavor) {
		return (dataFlavor.equals(DataFlavor.stringFlavor)
			|| dataFlavor.equals(DataFlavor.plainTextFlavor));
	}

	/**
	 * Called when the clipboard contents has been altered and
	 * the receiver is no longer the owner of the specified
	 * clipboard.
	 *
	 * @param	clipboard	 the clipboard which is no longer owned by receiver
	 * @param	transferable the contents that the receiver stored in the clipboard
	 *
	 * @see		Clipboard
	 */
	public void lostOwnership(Clipboard clipboard, Transferable transferable) {
		// Do nothing
	}

}
