package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Component;
import java.awt.Toolkit;

/**
 * This class is the superclass for classes which represent
 * events which model input on a <code>Component</code>.
 *
 * @author      IBM
 * @version		initial
 *
 */
public abstract class InputEvent extends ComponentEvent {

	static final long serialVersionUID = -2482525981698309786L;

	// The keyboard and mouse modifiers state
	int modifiers;

	// The time when the receiver was created
	long when;

	// Input event modifiers
	public static final int ALT_GRAPH_MASK = 32;
	public static final int ALT_MASK = 8;
	public static final int BUTTON1_MASK = 16;
	public static final int BUTTON2_MASK = 8;
	public static final int BUTTON3_MASK = 4;
	public static final int CTRL_MASK = 2;
	public static final int META_MASK = 4;
	public static final int SHIFT_MASK = 1;

	static final int InitialMasks_MASK = 0x3F;		// = SHIFT_MASK | CTRL_MASK | META_MASK | ALT_MASK |
													//	 BUTTON1_MASK | BUTTON2_MASK | BUTTON3_MASK |
													//	 ALT_GRAPH_MASK;
	public static final int SHIFT_DOWN_MASK = 64;
	public static final int CTRL_DOWN_MASK = 128;
	public static final int META_DOWN_MASK = 256;
	public static final int ALT_DOWN_MASK = 512;
	public static final int BUTTON1_DOWN_MASK = 1024;
	public static final int BUTTON2_DOWN_MASK = 2048;
	public static final int BUTTON3_DOWN_MASK = 4096;
	public static final int ALT_GRAPH_DOWN_MASK = 8192;

	static final int ExtendedMasks_MASK = 0x3FC0; 	// = SHIFT_DOWN_MASK | CTRL_DOWN_MASK | META_DOWN_MASK | ALT_DOWN_MASK |
													//   BUTTON1_DOWN_MASK | BUTTON2_DOWN_MASK | BUTTON3_DOWN_MASK |
													//	 ALT_GRAPH_DOWN_MASK;

	/**
	 * Constructs a new <code>InputEvent</code> using the
	 * specified id.  The receiver will be an event that occured
	 * in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source	the source object of the receiver
	 * @param	id		the id of the receiver
	 *
	 */
	InputEvent(Component source, int id, long when, int modifiers) {
		super(source, id);

		this.when = when;
		this.modifiers = modifiers;
	}

	/**
	 * Indicate that this event should be consumed.  This means it should
	 * never be processed by the receiver's source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 */
	public void consume() {
		this.consumed = true;
	}

	/**
	 * Answer the modifier for the receiver.
	 * This will be one of the following:
	 *	<code>ALT_GRAPH_MASK</code>
	 *	<code>ALT_MASK</code>
	 *	<code>BUTTON1_MASK</code>
	 *	<code>BUTTON2_MASK</code>
	 *	<code>BUTTON3_MASK</code>
	 *	<code>CTRL_MASK</code>
	 *	<code>META_MASK</code>
	 *	<code>SHIFT_MASK</code>
	 *
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the modifier
	 */
	public int getModifiers() {
		return modifiers & InitialMasks_MASK;
	}

	/**
	 * Answer the time that the event represented by
	 * the receiver occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a long	the time of the event
	 */
	public long getWhen() {
		return when;
	}

	/**
	 * Answer the whether the Alt key was pressed when the event
	 * that the receiver represents occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if Alt was pressed
	 *						<code>false</code> otherwise
	 */
	public boolean isAltDown() {
		boolean isDown = (getModifiers() & ALT_MASK) != 0;
		isDown |= (getModifiersEx() & ALT_DOWN_MASK) != 0;
		return isDown;
	}

	/**
	 * Answer the whether the Alt-Graph key was pressed when the event
	 * that the receiver represents occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if Alt-Graph was pressed
	 *						<code>false</code> otherwise
	 *
	 * @since		JDK1.2
	 */
	public boolean isAltGraphDown() {
		boolean isDown = (getModifiers() & ALT_GRAPH_MASK) != 0;
		isDown |= (getModifiersEx() & ALT_GRAPH_DOWN_MASK) != 0;
		return isDown;
	}

	/**
	 * Answer whether this event should be consumed.  This means it should
	 * never be processed.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	a boolean	<code>true</code> if this event should be consumed
	 *						<code>false</code> otherwise.
	 */
	public boolean isConsumed() {
		return super.isConsumed();
	}

	/**
	 * Answer the whether the Ctrl key was pressed when the event
	 * that the receiver represents occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if Ctrl was pressed
	 *						<code>false</code> otherwise
	 */
	public boolean isControlDown() {
		boolean isDown = (getModifiers() & CTRL_MASK) != 0;
		isDown |= (getModifiersEx() & CTRL_DOWN_MASK) != 0;
		return isDown;
	}

	/**
	 * Answer the whether the Meta key was pressed when the event
	 * that the receiver represents occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if Meta was pressed
	 *						<code>false</code> otherwise
	 */
	public boolean isMetaDown() {
		boolean isDown = (getModifiers() & META_MASK) != 0;
		isDown |= (getModifiersEx() & META_DOWN_MASK) != 0;
		return isDown;
	}

	/**
	 * Answer the whether the Shift key was pressed when the event
	 * that the receiver represents occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if Shift was pressed
	 *						<code>false</code> otherwise
	 */
	public boolean isShiftDown() {
		boolean isDown = (getModifiers() & SHIFT_MASK) != 0;
		isDown |= (getModifiersEx() & SHIFT_DOWN_MASK) != 0;
		return isDown;
	}

	/**
	 * Answer the extended modifier for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public int getModifiersEx() {
		return modifiers & ExtendedMasks_MASK;
	}

	public static String getModifiersExText(int modifiers) {
	    StringBuffer buf = new StringBuffer("");

	    if ((modifiers & CTRL_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.control", "Ctrl"));
            buf.append('+');
	    }
	    if ((modifiers & ALT_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.alt", "Alt"));
            buf.append('+');
	    }
	    if ((modifiers & SHIFT_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.shift", "Shift"));
            buf.append('+');
	    }
	    if ((modifiers & META_DOWN_MASK) != 0) {
	    	buf.append(Toolkit.getProperty("AWT.meta", "Meta"));
	    	buf.append('+');
	    }
	    if ((modifiers & ALT_GRAPH_DOWN_MASK) != 0) {
	    	buf.append(Toolkit.getProperty("AWT.altGraph", "Alt Graph"));
	    	buf.append('+');
	    }
	    if ((modifiers & BUTTON1_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button1", "Button1"));
            buf.append('+');
        }
	    if ((modifiers & BUTTON2_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button2", "Button2"));
            buf.append('+');
        }
	    if ((modifiers & BUTTON3_DOWN_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button3", "Button3"));
            buf.append('+');
        }

	    if (buf.length() != 0) {
	    	buf.deleteCharAt(buf.length() - 1); // We remove the extra "+"!
	    }

		return buf.toString();
	}

}
