package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2005  All Rights Reserved
 */

import java.awt.*;
import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.font.TextHitInfo;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.AttributedCharacterIterator;
import java.text.CharacterIterator;

/**
 * This class represents text editing events which occur
 * during input of characters that need to be composed (Japanese for example).
 * It may not be possible to map every character to a single key
 * on the keyboard and in this case characters are composed useing
 * intput method editors (IME's).
 * <p>
 * There are typically two ways to do character composition.  Either the
 * a component can act as a composition window and the characters are composed
 * in the component (in-place editing) or a separate window (IME window) or
 * text editor is used and the characters are composed in that
 * window and then the completed characters are sent to the component
 * being edited.
 * <p>
 * <code>InputMethodEvents</code> are the means of communicating this action
 * of character composition to a component.
 *
 * @author		OTI
 * @version		initial
 *
 */

public class InputMethodEvent extends AWTEvent {

	static final long serialVersionUID = 4727190874778922661L;

	public static final int CARET_POSITION_CHANGED = 1101;
    public static final int INPUT_METHOD_FIRST = 1100;
    public static final int INPUT_METHOD_LAST = 1101;
    public static final int INPUT_METHOD_TEXT_CHANGED = 1100;

	long when;
	transient private TextHitInfo caretPosition;
	transient private TextHitInfo visiblePosition;
	private AttributedCharacterIterator text;
	private int committedCharacterCount;

	public InputMethodEvent(Component source, int id, long when,
			AttributedCharacterIterator text, int committedCharacterCount,
			TextHitInfo caretPosition, TextHitInfo visiblePosition) {

		super(source, id);
		if 	((id < INPUT_METHOD_FIRST) ||
			 (id > INPUT_METHOD_LAST)) throw new IllegalArgumentException("id is out of the range");
		if ((id == CARET_POSITION_CHANGED) && (text != null)) throw new IllegalArgumentException("text should be null when id == CARET_POSITION_CHANGED");
		int length = (text == null) ? 0 : text.getEndIndex() - text.getBeginIndex();
		if ((committedCharacterCount < 0) || (committedCharacterCount > length)) throw new IllegalArgumentException("committedCharacterCount is outside of the range");

		this.committedCharacterCount = committedCharacterCount;
		this.when = when;
		this.text = text;
		this.caretPosition = caretPosition;
		this.visiblePosition = visiblePosition;
	}

	public InputMethodEvent(Component source, int id, AttributedCharacterIterator text, int committedCharacterCount, TextHitInfo caret, TextHitInfo visiblePosition) {
	    this(source, id, 0, text, committedCharacterCount, caret, visiblePosition);
	}

	public InputMethodEvent(Component source, int id, TextHitInfo caret, TextHitInfo visiblePosition) {
	    this(source, id, (AttributedCharacterIterator) null, 0, caret, visiblePosition);
	}

	public AttributedCharacterIterator getText() {
		return text;
	}

	public int getCommittedCharacterCount() {
		return committedCharacterCount;
	}

	public TextHitInfo getCaret() {
		return caretPosition;
	}

	public TextHitInfo getVisiblePosition() {
		return visiblePosition;
	}

	/**
	 * Indicate that this event should be consumed.  This means it should
	 * never be processed by the receiver's source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 */
	public void consume() {
		this.consumed = true;
	}
	/**
	 * Answer whether this event should be consumed.  This means it should
	 * never be processed.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @return	a boolean	<code>true</code> if this event should be consumed
	 *						<code>false</code> otherwise.
	 */
	public boolean isConsumed() {
		return super.isConsumed();
	}

	public long getWhen() {
		return when;
	}

	public String paramString() {
		/*
		 * Returns a parameter string identifying this event.
		 * This method is useful for event-logging and for debugging.
		 * It contains the event ID in text form,
		 * the characters of the committed and composed text separated by "+",
		 * the number of committed characters, the caret, and the visible position.
		 */
		StringBuffer buffer = new StringBuffer(super.paramString());

		switch (getID()) {
			case CARET_POSITION_CHANGED:
				buffer.append("CARET_POSITION_CHANGED");
				break;
			case INPUT_METHOD_TEXT_CHANGED:
				buffer.append("INPUT_METHOD_TEXT_CHANGED");
				break;
			default:
				buffer.append("unknown");
		}
		if (text == null) {
			buffer.append("no text");
		} else {
			// Print the committed text
			buffer.append('\"');
			char c = text.first();
			for (int i = 0; i < committedCharacterCount; i++) {
				buffer.append(c);
				c = text.next();
			}
			buffer.append('\"');
			buffer.append('+');
			buffer.append('\"');
			while (c != CharacterIterator.DONE) {
				buffer.append(c);
				c = text.next();
			}
			buffer.append('\"');
		}

		buffer.append(",committedCharacterCount=");
		buffer.append(committedCharacterCount);

		if (caretPosition == null) {
			buffer.append(",no caret");
		} else {
			buffer.append(",caretPosition=");
			buffer.append(caretPosition);
		}
		if (visiblePosition == null) {
			buffer.append(",no visiblePosition");
		} else {
			buffer.append(",visiblePosition=");
			buffer.append(visiblePosition);
		}

		return buffer.toString();
	}

	private void readObject(ObjectInputStream stream) throws ClassNotFoundException, IOException {
		// init the when field
		stream.defaultReadObject();
		if (when == 0) when = EventQueue.getMostRecentEventTime();
	}

	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
	}
}
