package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.Component;
import java.awt.Toolkit;

/**
 * This class represents events which model a keyboard actions
 * on a <code>Component</code>.
 * This event is propagated to all <code>KeyListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addKeyListener</code> method, key events occurring
 * on that component will be delivered to the key listener via
 * the sending of one of the following methods:
 * 	- <code>keyPressed</code>
 * 	- <code>keyReleased</code>
 * 	- <code>keyTyped</code>
 * @author      IBM
 * @version		initial
 *
 */
public class KeyEvent extends InputEvent {

	static final long serialVersionUID = -2352130953028126954L;

	// Unicode value the key or combination of keys that was pressed or released
	char keyChar;

	// Identifier of the key that was pressed or released
	int keyCode;

	// constant for the undefined character
	public static final char CHAR_UNDEFINED = 0xFFFF;

	// Key event ID constants
	public static final int KEY_FIRST = 400;
	public static final int KEY_LAST = 402;
	public static final int KEY_PRESSED = 401;
	public static final int KEY_RELEASED = 402;
	public static final int KEY_TYPED = 400;

	/**
	 * Constructs a new <code>KeyEvent</code> using the
	 * specified id, time, modifiers and keycode.  The receiver
	 * will be an event that occurred in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 * @param	time		the time the event occurred
	 * @param	modifiers	the modifiers for the receiver
	 * @param	keyCode		the key code or VK_UNDEFINED if it does not apply
	 *
	 * @deprecated
	 */
	public KeyEvent(Component source, int id, long time, int modifiers, int keyCode) {
		this(source, id, time, modifiers, keyCode, (char) keyCode);
	}

	/**
	 * Constructs a new <code>KeyEvent</code> using the
	 * specified id, time, modifiers, keycode and keychar.  The receiver
	 * will be an event that occurred in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 * @param	time		the time the event occurred
	 * @param	modifiers	the modifiers for the receiver
	 * @param	keyCode		the key code or VK_UNDEFINED if it does not apply
	 * @param	keyChar		the Unicode character or CHAR_UNDEFINED if it does not apply
	 */
	public KeyEvent(Component source, int id, long time, int modifiers, int keyCode, char keyChar) {
		super(source, id, time, modifiers);
	    if ((id == KeyEvent.KEY_TYPED) && (keyChar == KeyEvent.CHAR_UNDEFINED)) throw new IllegalArgumentException("invalid keychar");
	    if ((id == KeyEvent.KEY_TYPED) && (keyCode != KeyEvent.VK_UNDEFINED)) throw new IllegalArgumentException("invalid keycode");

		this.keyCode = keyCode;
		this.keyChar = keyChar;
	}

	/**
	 * Answer a <code>String</code> which represents the integer
	 * modifiers specified.  For example if the modifier CTRL_MASK
	 * is specified, the return value will be "Ctrl".
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the modifier string
	 */
	public static String getKeyModifiersText(int modifiers) {
		StringBuffer buffer = new StringBuffer();
		if ((modifiers & META_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.meta", "Meta")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & CTRL_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.control", "Ctrl")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & ALT_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.alt", "Alt")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & SHIFT_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.shift", "Shift")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & ALT_GRAPH_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.altGraph", "Alt Graph")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & BUTTON1_MASK) != 0) {
			buffer.append(Toolkit.getProperty("AWT.button1", "Button1")); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append('+');
		}
		if ((modifiers & BUTTON2_MASK) != 0) {
			// Do not handle as BUTTON2_MASK conflicts with ALT_MASK.
			//buffer.append(Toolkit.getProperty("AWT.button2", "Button2")); //$NON-NLS-1$ //$NON-NLS-2$
			//buffer.append('+');
		}
		if ((modifiers & BUTTON3_MASK) != 0) {
			// Do not handle as BUTTON3_MASK conflicts with META_MASK.
			//buffer.append(Toolkit.getProperty("AWT.button3", "Button3")); //$NON-NLS-1$ //$NON-NLS-2$
			//buffer.append('+');
		}
		if (buffer.length() != 0)
			buffer.setLength(buffer.length() - 1);
		return buffer.toString();
	}

	/**
	 * Answer a <code>String</code> which represents the integer
	 * key code specified.  For example if the key code VK_A was
	 * is specified, the return value will be "A".
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the modifier string
	 */
	public static String getKeyText(int keyCode) {
		if ((keyCode >= VK_0 && keyCode <= VK_9) || (keyCode >= VK_A && keyCode <= VK_Z)) {
			return String.valueOf((char) keyCode);
		}
		if (",;/=[]\\.".indexOf((char) keyCode) >=0) { //$NON-NLS-1$
			return String.valueOf((char) keyCode);
		}
		String keyText = _getKeyText(keyCode);
		if (keyText == null) {
			String keyCodeString = (keyCode < 0) ? Integer.toString(keyCode) : Integer.toHexString(keyCode);
			return Toolkit.getProperty("AWT.unknown", "Unknown keyCode") + ": 0x" + keyCodeString; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return keyText;
	}

	/**
	 * Answer the Unicode key character for the receiver, or CHAR_UNDEFINED
	 * if one does not exist for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a char	the Unicode character
	 */
	public char getKeyChar() {
		return keyChar;
	}

	/**
	 * Answer the key code for the receiver, or VK_UNDEFINED
	 * if one does not exist for the receiver.  KEY_TYPED events
	 * do not have a keycode as they represent a different type
	 * of key event.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the key code
	 */
	public int getKeyCode() {
		return keyCode;
	}

	/**
	 * Answer the whether the key in the receiver is an action
	 * key.  Action keys are defined in the class <code>Event</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if the key is an
	 *						action key <code>false</code> otherwise
	 */
	public boolean isActionKey() {
		switch(getKeyCode()) {
			case VK_HOME:
			case VK_END:
			case VK_PAGE_UP:
			case VK_PAGE_DOWN:
			case VK_UP:
			case VK_DOWN:
			case VK_LEFT:
			case VK_RIGHT:
			case VK_F1:
			case VK_F2:
			case VK_F3:
			case VK_F4:
			case VK_F5:
			case VK_F6:
			case VK_F7:
			case VK_F8:
			case VK_F9:
			case VK_F10:
			case VK_F11:
			case VK_F12:
			case VK_PRINTSCREEN:
			case VK_SCROLL_LOCK:
			case VK_CAPS_LOCK:
			case VK_NUM_LOCK:
			case VK_PAUSE:
			case VK_INSERT:
				return true;
			default:
				return false;
		}
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case KEY_TYPED:
				buffer.append("KEY_TYPED"); //$NON-NLS-1$
				break;
			case KEY_PRESSED:
				buffer.append("KEY_PRESSED"); //$NON-NLS-1$
				break;
			case KEY_RELEASED:
				buffer.append("KEY_RELEASED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",keyCode="); //$NON-NLS-1$
		buffer.append(getKeyCode());
		if (getKeyChar() != 0) {
			buffer.append(",keyChar="); //$NON-NLS-1$
			buffer.append(formatKeyText(getKeyChar()));
		} else {
			buffer.append(",keyChar="); //$NON-NLS-1$
			buffer.append(formatKeyText((char) getKeyCode()));
		}
		String modifiersText = getKeyModifiersText(getModifiers());
		if (modifiersText.length() > 0) {
			buffer.append(",modifiers="); //$NON-NLS-1$
			buffer.append(modifiersText);
		}
		return buffer.toString();
	}

	String formatKeyText(char key) {
		String keyText = _getKeyText(getKeyCode());
		if (keyText == null) {
			switch ((int) getKeyChar()) {
				case KeyEvent.VK_ENTER :
				case KeyEvent.VK_ESCAPE :
				case KeyEvent.VK_TAB :
				case KeyEvent.VK_DELETE :
				case KeyEvent.VK_BACK_SPACE :
					return _getKeyText(getKeyChar());
				default :
					return "\'" + key + "\'"; //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		return keyText;
	}

	/**
	 * Set the key character for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	a <code>char</code>	the keyChar
	 */
	public void setKeyChar(char keyChar) {
		this.keyChar = keyChar;
	}

	/**
	 * Set the key code for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	an <code>int</code>	the keyCode
	 */
	public void setKeyCode(int keyCode) {
		this.keyCode = keyCode;
	}

	/**
	 * Set the modifiers for the receiver.
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	an <code>int</code>	the modifiers
	 */
	public void setModifiers(int modifiers) {
		this.modifiers = modifiers;
	}

	/**
	 * Set the source of the receiver.
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	an <code>Object</code>	the new source
	 */
	public void setSource(Object source) {
		this.source = source;
	}

	// key code constants
	public static final int VK_0 = 48;
	public static final int VK_1 = 49;
	public static final int VK_2 = 50;
	public static final int VK_3 = 51;
	public static final int VK_4 = 52;
	public static final int VK_5 = 53;
	public static final int VK_6 = 54;
	public static final int VK_7 = 55;
	public static final int VK_8 = 56;
	public static final int VK_9 = 57;
	public static final int VK_A = 65;
	public static final int VK_ACCEPT = 30;
	public static final int VK_ADD = 107;
	public static final int VK_AGAIN = 65481;
	public static final int VK_ALL_CANDIDATES = 256;
	public static final int VK_ALPHANUMERIC = 240;
	public static final int VK_ALT = 18;
	public static final int VK_ALT_GRAPH = 65406;
	public static final int VK_AMPERSAND = 150;
	public static final int VK_ASTERISK = 151;
	public static final int VK_AT = 512;
	public static final int VK_B = 66;
	public static final int VK_BACK_QUOTE = 192;
	public static final int VK_BACK_SLASH = 92;
	public static final int VK_BACK_SPACE = 8;
	public static final int VK_BRACELEFT = 161;
	public static final int VK_BRACERIGHT = 162;
	public static final int VK_C = 67;
	public static final int VK_CANCEL = 3;
	public static final int VK_CAPS_LOCK = 20;
	public static final int VK_CIRCUMFLEX = 514;
	public static final int VK_CLEAR = 12;
	public static final int VK_CLOSE_BRACKET = 93;
	public static final int VK_CODE_INPUT = 258;
	public static final int VK_COLON = 513;
	public static final int VK_COMMA = 44;
	public static final int VK_COMPOSE = 65312;
	public static final int VK_CONTROL = 17;
	public static final int VK_CONVERT = 28;
	public static final int VK_COPY = 65485;
	public static final int VK_CUT = 65489;
	public static final int VK_D = 68;
	public static final int VK_DEAD_ABOVEDOT = 134;
	public static final int VK_DEAD_ABOVERING = 136;
	public static final int VK_DEAD_ACUTE = 129;
	public static final int VK_DEAD_BREVE = 133;
	public static final int VK_DEAD_CARON = 138;
	public static final int VK_DEAD_CEDILLA = 139;
	public static final int VK_DEAD_CIRCUMFLEX = 130;
	public static final int VK_DEAD_DIAERESIS = 135;
	public static final int VK_DEAD_DOUBLEACUTE = 137;
	public static final int VK_DEAD_GRAVE = 128;
	public static final int VK_DEAD_IOTA = 141;
	public static final int VK_DEAD_MACRON = 132;
	public static final int VK_DEAD_OGONEK = 140;
	public static final int VK_DEAD_SEMIVOICED_SOUND = 143;
	public static final int VK_DEAD_TILDE = 131;
	public static final int VK_DEAD_VOICED_SOUND = 142;
	public static final int VK_DECIMAL = 110;
	public static final int VK_DELETE = 127;
	public static final int VK_DIVIDE = 111;
	public static final int VK_DOLLAR = 515;
	public static final int VK_DOWN = 40;
	public static final int VK_E = 69;
	public static final int VK_END = 35;
	public static final int VK_ENTER = 10;
	public static final int VK_EQUALS = 61;
	public static final int VK_ESCAPE = 27;
	public static final int VK_EURO_SIGN = 516;
	public static final int VK_EXCLAMATION_MARK = 517;
	public static final int VK_F = 70;
	public static final int VK_F1 = 112;
	public static final int VK_F10 = 121;
	public static final int VK_F11 = 122;
	public static final int VK_F12 = 123;
	public static final int VK_F13 = 61440;
	public static final int VK_F14 = 61441;
	public static final int VK_F15 = 61442;
	public static final int VK_F16 = 61443;
	public static final int VK_F17 = 61444;
	public static final int VK_F18 = 61445;
	public static final int VK_F19 = 61446;
	public static final int VK_F2 = 113;
	public static final int VK_F20 = 61447;
	public static final int VK_F21 = 61448;
	public static final int VK_F22 = 61449;
	public static final int VK_F23 = 61450;
	public static final int VK_F24 = 61451;
	public static final int VK_F3 = 114;
	public static final int VK_F4 = 115;
	public static final int VK_F5 = 116;
	public static final int VK_F6 = 117;
	public static final int VK_F7 = 118;
	public static final int VK_F8 = 119;
	public static final int VK_F9 = 120;
	public static final int VK_FINAL = 24;
	public static final int VK_FIND = 65488;
	public static final int VK_FULL_WIDTH = 243;
	public static final int VK_G = 71;
	public static final int VK_GREATER = 160;
	public static final int VK_H = 72;
	public static final int VK_HALF_WIDTH = 244;
	public static final int VK_HELP = 156;
	public static final int VK_HIRAGANA = 242;
	public static final int VK_HOME = 36;
	public static final int VK_I = 73;
	public static final int VK_INPUT_METHOD_ON_OFF = 263;
	public static final int VK_INSERT = 155;
	public static final int VK_INVERTED_EXCLAMATION_MARK = 518;
	public static final int VK_J = 74;
	public static final int VK_JAPANESE_HIRAGANA = 260;
	public static final int VK_JAPANESE_KATAKANA = 259;
	public static final int VK_JAPANESE_ROMAN = 261;
	public static final int VK_K = 75;
	public static final int VK_KANA = 21;
	public static final int VK_KANA_LOCK = 262;
	public static final int VK_KANJI = 25;
	public static final int VK_KATAKANA = 241;
	public static final int VK_KP_DOWN = 225;
	public static final int VK_KP_LEFT = 226;
	public static final int VK_KP_RIGHT = 227;
	public static final int VK_KP_UP = 224;
	public static final int VK_L = 76;
	public static final int VK_LEFT = 37;
	public static final int VK_LEFT_PARENTHESIS = 519;
	public static final int VK_LESS = 153;
	public static final int VK_M = 77;
	public static final int VK_META = 157;
	public static final int VK_MINUS = 45;
	public static final int VK_MODECHANGE = 31;
	public static final int VK_MULTIPLY = 106;
	public static final int VK_N = 78;
	public static final int VK_NONCONVERT = 29;
	public static final int VK_NUM_LOCK = 144;
	public static final int VK_NUMBER_SIGN = 520;
	public static final int VK_NUMPAD0 = 96;
	public static final int VK_NUMPAD1 = 97;
	public static final int VK_NUMPAD2 = 98;
	public static final int VK_NUMPAD3 = 99;
	public static final int VK_NUMPAD4 = 100;
	public static final int VK_NUMPAD5 = 101;
	public static final int VK_NUMPAD6 = 102;
	public static final int VK_NUMPAD7 = 103;
	public static final int VK_NUMPAD8 = 104;
	public static final int VK_NUMPAD9 = 105;
	public static final int VK_O = 79;
	public static final int VK_OPEN_BRACKET = 91;
	public static final int VK_P = 80;
	public static final int VK_PAGE_DOWN = 34;
	public static final int VK_PAGE_UP = 33;
	public static final int VK_PASTE = 65487;
	public static final int VK_PAUSE = 19;
	public static final int VK_PERIOD = 46;
	public static final int VK_PLUS = 521;
	public static final int VK_PREVIOUS_CANDIDATE = 257;
	public static final int VK_PRINTSCREEN = 154;
	public static final int VK_PROPS = 65482;
	public static final int VK_Q = 81;
	public static final int VK_QUOTE = 222;
	public static final int VK_QUOTEDBL = 152;
	public static final int VK_R = 82;
	public static final int VK_RIGHT = 39;
	public static final int VK_RIGHT_PARENTHESIS = 522;
	public static final int VK_ROMAN_CHARACTERS = 245;
	public static final int VK_S = 83;
	public static final int VK_SCROLL_LOCK = 145;
	public static final int VK_SEMICOLON = 59;
	public static final int VK_SEPARATER = 108;
	public static final int VK_SEPARATOR = VK_SEPARATER;
	public static final int VK_SHIFT = 16;
	public static final int VK_SLASH = 47;
	public static final int VK_SPACE = 32;
	public static final int VK_STOP = 65480;
	public static final int VK_SUBTRACT = 109;
	public static final int VK_T = 84;
	public static final int VK_TAB = 9;
	public static final int VK_U = 85;
	public static final int VK_UNDEFINED = 0;
	public static final int VK_UNDERSCORE = 523;
	public static final int VK_UNDO = 65483;
	public static final int VK_UP = 38;
	public static final int VK_V = 86;
	public static final int VK_W = 87;
	public static final int VK_X = 88;
	public static final int VK_Y = 89;
	public static final int VK_Z = 90;

	static String _getKeyText(int keyCode) {
		switch (keyCode) {
			case VK_ACCEPT: return Toolkit.getProperty("AWT.accept", "Accept"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ADD: return Toolkit.getProperty("AWT.add", "NumPad +"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ALT: return Toolkit.getProperty("AWT.alt", "Alt"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_BACK_QUOTE: return Toolkit.getProperty("AWT.backQuote", "Back Quote"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_BACK_SPACE: return Toolkit.getProperty("AWT.backSpace", "Backspace"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CAPS_LOCK: return Toolkit.getProperty("AWT.capsLock", "Caps Lock"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CLEAR: return Toolkit.getProperty("AWT.clear", "Clear"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_COLON: return Toolkit.getProperty("AWT.colon", "Colon"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CONTROL: return Toolkit.getProperty("AWT.control", "Ctrl"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CONVERT: return Toolkit.getProperty("AWT.convert", "Convert"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DECIMAL: return Toolkit.getProperty("AWT.decimal", "NumPad ."); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DELETE: return Toolkit.getProperty("AWT.delete", "Delete"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DIVIDE: return Toolkit.getProperty("AWT.divide", "NumPad /"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DOLLAR: return Toolkit.getProperty("AWT.dollar", "Dollar"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DOWN: return Toolkit.getProperty("AWT.down", "Down"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_END: return Toolkit.getProperty("AWT.end", "End"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ENTER: return Toolkit.getProperty("AWT.enter", "Enter"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ESCAPE: return Toolkit.getProperty("AWT.escape", "Escape"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_EURO_SIGN: return Toolkit.getProperty("AWT.euro", "Euro"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F10: return Toolkit.getProperty("AWT.f10", "F10"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F11: return Toolkit.getProperty("AWT.f11", "F11"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F12: return Toolkit.getProperty("AWT.f12", "F12"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F1: return Toolkit.getProperty("AWT.f1", "F1"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F2: return Toolkit.getProperty("AWT.f2", "F2"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F3: return Toolkit.getProperty("AWT.f3", "F3"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F4: return Toolkit.getProperty("AWT.f4", "F4"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F5: return Toolkit.getProperty("AWT.f5", "F5"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F6: return Toolkit.getProperty("AWT.f6", "F6"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F7: return Toolkit.getProperty("AWT.f7", "F7"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F8: return Toolkit.getProperty("AWT.f8", "F8"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F9: return Toolkit.getProperty("AWT.f9", "F9"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_FINAL: return Toolkit.getProperty("AWT.final", "Final"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_HELP: return Toolkit.getProperty("AWT.help", "Help"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_HOME: return Toolkit.getProperty("AWT.home", "Home"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_INPUT_METHOD_ON_OFF: return Toolkit.getProperty("AWT.inputMethodOnOff", "Input Method On Off"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_INSERT: return Toolkit.getProperty("AWT.insert", "Insert"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_INVERTED_EXCLAMATION_MARK: return Toolkit.getProperty("AWT.invertedExclamationMark", "Inverted Exclamation Mark"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KANA: return Toolkit.getProperty("AWT.kana", "Kana"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KANA_LOCK: return Toolkit.getProperty("AWT.kanaLock", "Kana Lock"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KANJI: return Toolkit.getProperty("AWT.kanji", "Kanji"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_LEFT: return Toolkit.getProperty("AWT.left", "Left"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_META: return Toolkit.getProperty("AWT.meta", "Meta"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_MODECHANGE: return Toolkit.getProperty("AWT.modechange", "Mode Change"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_MULTIPLY: return Toolkit.getProperty("AWT.multiply", "Multiply"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_NONCONVERT: return Toolkit.getProperty("AWT.noconvert", "No Convert"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_NUMPAD0: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-0"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD1: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-1"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD2: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-2"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD3: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-3"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD4: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-4"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD5: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-5"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD6: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-6"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD7: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-7"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD8: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-8"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUMPAD9: return Toolkit.getProperty("AWT.numpad", "Numpad") + "-9"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			case VK_NUM_LOCK: return Toolkit.getProperty("AWT.numLock", "Num Lock"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PAUSE: return Toolkit.getProperty("AWT.pause", "Pause"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PAGE_DOWN: return Toolkit.getProperty("AWT.pgdn", "Page Down"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PAGE_UP: return Toolkit.getProperty("AWT.pgup", "Page Up"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PRINTSCREEN: return Toolkit.getProperty("AWT.printScreen", "Print Screen"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_QUOTE: return Toolkit.getProperty("AWT.quote", "Quote"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_RIGHT: return Toolkit.getProperty("AWT.right", "Right"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_SCROLL_LOCK: return Toolkit.getProperty("AWT.scrollLock", "Scroll Lock"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_SEPARATER: return Toolkit.getProperty("AWT.separater", "NumPad ,"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_SHIFT: return Toolkit.getProperty("AWT.shift", "Shift"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_SPACE: return Toolkit.getProperty("AWT.space", "Space"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_SUBTRACT: return Toolkit.getProperty("AWT.subtract", "NumPad -"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_TAB: return Toolkit.getProperty("AWT.tab", "Tab"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_UP: return Toolkit.getProperty("AWT.up", "Up"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_AGAIN: return Toolkit.getProperty("AWT.again", "Again"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ALL_CANDIDATES: return Toolkit.getProperty("AWT.allCandidates", "All Candidates"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ALPHANUMERIC: return Toolkit.getProperty("AWT.alphanumeric", "Alphanumeric"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ALT_GRAPH: return Toolkit.getProperty("AWT.altGraph", "Alt Graph"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_AMPERSAND: return Toolkit.getProperty("AWT.amp", "Ampersand"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ASTERISK: return Toolkit.getProperty("AWT.asterisk", "Asterisk"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_AT: return Toolkit.getProperty("AWT.at", "At"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_BRACELEFT: return Toolkit.getProperty("AWT.leftBrace", "Left Brace"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_BRACERIGHT: return Toolkit.getProperty("AWT.rightBrace", "Right Brace"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CANCEL: return Toolkit.getProperty("AWT.cancel", "Cancel"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CIRCUMFLEX: return Toolkit.getProperty("AWT.circumflex", "Circumflex"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CODE_INPUT: return Toolkit.getProperty("AWT.codeInput", "Code Input"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_COMPOSE: return Toolkit.getProperty("AWT.compose", "Compose "); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_COPY: return Toolkit.getProperty("AWT.copy", "Copy"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_CUT: return Toolkit.getProperty("AWT.cut", "Cut"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_ABOVEDOT: return Toolkit.getProperty("AWT.deadAboveDot", "Dead Above Dot"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_ABOVERING: return Toolkit.getProperty("AWT.deadAboveRing", "Dead Above Ring"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_ACUTE: return Toolkit.getProperty("AWT.deadAcute", "Dead Acute"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_BREVE: return Toolkit.getProperty("AWT.deadBreve", "Dead Breve"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_CARON: return Toolkit.getProperty("AWT.deadCaron", "Dead Caron"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_CEDILLA: return Toolkit.getProperty("AWT.deadCedilla", "Dead Cedilla"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_CIRCUMFLEX: return Toolkit.getProperty("AWT.deadCircumflex", "Dead Circumflex"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_DIAERESIS: return Toolkit.getProperty("AWT.deadDiaeresis", "Dead Diaeresis"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_DOUBLEACUTE: return Toolkit.getProperty("AWT.deadDoubleAcute", "Dead Double Acute"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_GRAVE: return Toolkit.getProperty("AWT.deadGrave", "Dead Grave"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_IOTA: return Toolkit.getProperty("AWT.deadIota", "Dead Iota"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_MACRON: return Toolkit.getProperty("AWT.deadMacron", "Dead Macron"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_OGONEK: return Toolkit.getProperty("AWT.deadOgonek", "Dead Ogonek"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_SEMIVOICED_SOUND: return Toolkit.getProperty("AWT.deadSemivoicedSound", "Dead Semivoiced Sound"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_TILDE: return Toolkit.getProperty("AWT.deadTilde", "Dead Tilde"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_DEAD_VOICED_SOUND: return Toolkit.getProperty("AWT.deadVoicedSound", "Dead Voiced Sound"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_EXCLAMATION_MARK: return Toolkit.getProperty("AWT.exclamationMark", "Exclamation Mark"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F13: return Toolkit.getProperty("AWT.f13", "F13"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F14: return Toolkit.getProperty("AWT.f14", "F14"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F15: return Toolkit.getProperty("AWT.f15", "F15"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F16: return Toolkit.getProperty("AWT.f16", "F16"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F17: return Toolkit.getProperty("AWT.f17", "F17"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F18: return Toolkit.getProperty("AWT.f18", "F18"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F19: return Toolkit.getProperty("AWT.f19", "F19"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F20: return Toolkit.getProperty("AWT.f20", "F20"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F21: return Toolkit.getProperty("AWT.f21", "F21"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F22: return Toolkit.getProperty("AWT.f22", "F22"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F23: return Toolkit.getProperty("AWT.f23", "F23"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_F24: return Toolkit.getProperty("AWT.f24", "F24"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_FIND: return Toolkit.getProperty("AWT.find", "Find"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_FULL_WIDTH: return Toolkit.getProperty("AWT.fullWidth", "Full-Width"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_GREATER: return Toolkit.getProperty("AWT.greater", "Greater"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_HALF_WIDTH: return Toolkit.getProperty("AWT.halfWidth", "Half-Width"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_HIRAGANA: return Toolkit.getProperty("AWT.hiragana", "Hiragana"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_JAPANESE_HIRAGANA: return Toolkit.getProperty("AWT.japaneseHiragana", "Japanese Hiragana"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_JAPANESE_KATAKANA: return Toolkit.getProperty("AWT.japaneseKatakana", "Japanese Katakana"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_JAPANESE_ROMAN: return Toolkit.getProperty("AWT.japaneseRoman", "Japanese Roman"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KATAKANA: return Toolkit.getProperty("AWT.katakana", "Katakana"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KP_DOWN: return Toolkit.getProperty("AWT.down", "Down"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KP_LEFT: return Toolkit.getProperty("AWT.left", "Left"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KP_RIGHT: return Toolkit.getProperty("AWT.right", "Right"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_KP_UP: return Toolkit.getProperty("AWT.up", "Up"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_LEFT_PARENTHESIS: return Toolkit.getProperty("AWT.leftParenthesis", "Left Parenthesis"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_LESS: return Toolkit.getProperty("AWT.less", "Less"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_MINUS: return Toolkit.getProperty("AWT.minus", "Minus"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_NUMBER_SIGN: return Toolkit.getProperty("AWT.numberSign", "Number Sign"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PASTE: return Toolkit.getProperty("AWT.paste", "Paste"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PLUS: return Toolkit.getProperty("AWT.plus", "Plus"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PREVIOUS_CANDIDATE: return Toolkit.getProperty("AWT.previousCandidate", "Previous Candidate"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_PROPS: return Toolkit.getProperty("AWT.props", "Props"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_QUOTEDBL: return Toolkit.getProperty("AWT.doubleQuote", "Double Quote"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_RIGHT_PARENTHESIS: return Toolkit.getProperty("AWT.rightParenthesis", "Right Parenthesis"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_ROMAN_CHARACTERS: return Toolkit.getProperty("AWT.romanCharacters", "Roman Characters"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_STOP: return Toolkit.getProperty("AWT.stop", "Stop"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_UNDERSCORE: return Toolkit.getProperty("AWT.underscore", "Underscore"); //$NON-NLS-1$ //$NON-NLS-2$
			case VK_UNDO: return Toolkit.getProperty("AWT.undo", "Undo");           //$NON-NLS-1$ //$NON-NLS-2$
			default:
				return null;
		}
	}
}
