package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.Component;
import java.awt.Point;
import java.awt.Toolkit;

/**
 * This class represents events which model a mouse action
 * on a <code>Component</code>.
 * This event is propagated to all <code>MouseListeners</code> and
 * <code>MouseMotionListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addMouseListener</code> method, mouse events occurring
 * on that component will be delivered to the appropriate mouse
 * listener via the sending of one of the following methods:
 * 	Sent to MouseListeners
 * 	- <code>mouseClicked</code>
 * 	- <code>mouseEntered</code>
 * 	- <code>mouseExited</code>
 * 	- <code>mousePressed</code>
 * 	- <code>mouseReleased</code>
 *
 *	Sent to MouseMotionListeners
 * 	- <code>mouseDragged</code>
 * 	- <code>mouseMoved</code>
 *
 * @author      IBM
 * @version		initial
 *
 */
public class MouseEvent extends InputEvent {

	static final long serialVersionUID = -991214153494842848L;

	// The x coordinate of the mouse pointer when the receiver was generated
	int x;

	// The y coordinate of the mouse pointer when the receiver was generated
	int y;

	// Determines wheter a <code>PopupMenu</code> should be shown
	boolean popupTrigger;

	// The number of times this event has happend in a certain amount of time
	int clickCount;

	// The mouse button that was clicked
	int button;

	// Mouse event ID constants
	public static final int MOUSE_CLICKED = 500;
	public static final int MOUSE_DRAGGED = 506;
	public static final int MOUSE_ENTERED = 504;
	public static final int MOUSE_EXITED = 505;
	public static final int MOUSE_FIRST = 500;
	public static final int MOUSE_LAST = 507;
	public static final int MOUSE_MOVED = 503;
	public static final int MOUSE_PRESSED = 501;
	public static final int MOUSE_RELEASED = 502;
	public static final int MOUSE_WHEEL = 507;
	public static final int NOBUTTON = 0;
	public static final int BUTTON1 = 1;
	public static final int BUTTON2 = 2;
	public static final int BUTTON3 = 3;

	/**
	 * Constructs a new <code>MouseEvent</code> using the
	 * specified id, time, modifiers, x, y, clickCount, and popupTrigger.
	 * The receiver will be an event that occured in the specified
	 * source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		 the source object of the receiver
	 * @param	id			 the id of the receiver
	 * @param	time		 the time the event occured
	 * @param	modifiers	 the modifiers for the receiver
	 * @param	x			 the x coordinate of the receiver
	 * @param	y			 the y coordinate of the receiver
	 * @param	clickCount	 the number of times this event has happend in
	 * 						 a certain amount of time
	 * @param	popupTrigger <code>true</code> if the receiver is a
	 *						 popup trigger, <code>false</code> otherwise
	 *
	 */
	public MouseEvent(Component source, int id, long time, int modifiers, int x, int y, int clickCount, boolean popupTrigger) {
		super(source, id, time, modifiers);

		this.x = x;
		this.y = y;
		this.clickCount = clickCount;
		this.popupTrigger = popupTrigger;
	}

	/**
	 * Answer the click count for the receiver.  This is the number of
	 * times the mouse has been clicked repeatedly in a certain interval of
	 * time between clicks.  This is for support of options like double and
	 * triple click etc....
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the click count
	 */
	public int getClickCount() {
		return clickCount;
	}

	/**
	 * Answer a <code>Point</code> representing the x and y coordinates
	 * of the receiver.  The location of the mouse pointer at the time the
	 * event occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Point	the location of the event
	 */
	public Point getPoint() {
		return new Point(getX(), getY());
	}

	/**
	 * Answer the x coordinate of the receiver.  The horizontal location
	 * of the mouse pointer at the time the event occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the x location of the event
	 */
	public int getX() {
		return x;
	}

	/**
	 * Answer the y coordinate of the receiver.  The vertical location
	 * of the mouse pointer at the time the event occured.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the y location of the event
	 */
	public int getY() {
		return y;
	}

	/**
	 * Answer the whether the receiver is a popup trigger.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if the receiver is a
	 *						popup trigger <code>false</code> otherwise
	 */
	public boolean isPopupTrigger() {
		return popupTrigger;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case MOUSE_CLICKED:
				buffer.append("MOUSE_CLICKED"); //$NON-NLS-1$
				break;
			case MOUSE_PRESSED:
				buffer.append("MOUSE_PRESSED"); //$NON-NLS-1$
				break;
			case MOUSE_RELEASED:
				buffer.append("MOUSE_RELEASED"); //$NON-NLS-1$
				break;
			case MOUSE_ENTERED:
				buffer.append("MOUSE_ENTERED"); //$NON-NLS-1$
				break;
			case MOUSE_EXITED:
				buffer.append("MOUSE_EXITED"); //$NON-NLS-1$
				break;
			case MOUSE_MOVED:
				buffer.append("MOUSE_MOVED"); //$NON-NLS-1$
				break;
			case MOUSE_DRAGGED:
				buffer.append("MOUSE_DRAGGED"); //$NON-NLS-1$
				break;
			case MOUSE_WHEEL:
				buffer.append("MOUSE_WHEEL"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
				break;
		}
		buffer.append(",("); //$NON-NLS-1$
		buffer.append(getX());
		buffer.append(","); //$NON-NLS-1$
		buffer.append(getY());
		buffer.append(")");
		buffer.append(",button="); //$NON-NLS-1$
		buffer.append(getButton());
		buffer.append(",mods="); //$NON-NLS-1$
		buffer.append(getModifiers());
		buffer.append(",clickCount="); //$NON-NLS-1$
		buffer.append(getClickCount());
		return buffer.toString();
	}

	/**
	 * Translate the location of the receiver by adding the
	 * specified x and y values.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param	x	the x translation value
	 * @param	y	the y translation value
	 */
	public synchronized void translatePoint(int x, int y) {
		this.x += x;
		this.y += y;
	}

	public MouseEvent(Component source, int id, long when, int modifiers,
			int x, int y, int clickCount, boolean popupTrigger, int button) {
		this(source, id, when, modifiers, x, y, clickCount, popupTrigger);
		setButton(button);
	}

	private void setButton(int button) {
        switch (button) {
            case BUTTON1:
            case BUTTON2:
            case BUTTON3:
            case NOBUTTON:
                break;
            default:
                throw new IllegalArgumentException(
                        "unknown button for mouse event");
        }
		this.button = button;
    }

    public int getButton() {
		return button;
	}

	public static String getMouseModifiersText(int mods) {
	    StringBuffer buf = new StringBuffer("");
	    if ((mods & InputEvent.CTRL_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.control", "Ctrl"));
            buf.append('+');
	    }

	    if ((mods & InputEvent.ALT_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.alt", "Alt"));
            buf.append('+');
	    }

	    if ((mods & InputEvent.SHIFT_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.shift", "Shift"));
            buf.append('+');
	    }

	    if ((mods & InputEvent.META_MASK) != 0) {
           buf.append(Toolkit.getProperty("AWT.meta", "Meta"));
           buf.append('+');
	    }

	    if ((mods & InputEvent.ALT_GRAPH_MASK) != 0) {
	    	buf.append(Toolkit.getProperty("AWT.altGraph", "Alt Graph"));
	    	buf.append('+');
	    }

	    if ((mods & InputEvent.BUTTON1_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button1", "Button1"));
        } else if ((mods & InputEvent.BUTTON2_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button2", "Button2"));
        } else if ((mods & InputEvent.BUTTON3_MASK) != 0) {
            buf.append(Toolkit.getProperty("AWT.button3", "Button3"));
        } else if (buf.length() != 0) {
	    	buf.deleteCharAt(buf.length() - 1); // We remove the extra '+'!
	    }

		return buf.toString();
	}
}
