package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class is a convenience class written as a
 * <code>WindowListener</code> with empty implementations
 * for all callback methods.
 * <p>
 * The purpose of this class is to allow subclasses to implement
 * only those methods that they are interested in listening to
 * without having to re-implement empty versions of the remaining
 * callback methods.  WindowListeners must normally implement
 * the following methods:
 * 	- <code>windowActivated</code>
 * 	- <code>windowClosed</code>
 * 	- <code>windowClosing</code>
 * 	- <code>windowDeactivated</code>
 * 	- <code>windowDeiconified</code>
 * 	- <code>windowIconified</code>
 * 	- <code>windowOpened</code>
 *
 * <p>
 *
 * @see	<code>WindowListener</code>
 */
public abstract class WindowAdapter implements WindowListener, WindowFocusListener {

	/**
	 * Constructs a new <code>WindowAdapter</code>.
	 */
	public WindowAdapter() {
	}

	/**
	 * Called when a window is activated.  This means the window
	 * comes to the forground and it or one of its children is
	 * receiving keyboard input.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowActivated(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is closed.  This means the window
	 * will no longer be visible on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowClosed(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is closing.  This means the window
	 * may still be visible on the screen but a user is attempting
	 * to close it.  This could be using the mouse or keyboard
	 * or other input device.
	 * <p>
	 * If a listener does not dispose the window in response to receiving
	 * this event the window will not be closed.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowClosing(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is de-activated.  This means the window
	 * is no longer the forground window and another window or
	 * or one of its children is receiving keyboard input.
	 *
	 * @author		OTI
	 * @version		initial
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowDeactivated(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is de-iconified.  This means the window
	 * is no longer iconified but will appear drawn in its regular
	 * state on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowDeiconified(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is iconified.  This means the window
	 * is no longer drawn in its regular state on the screen but
	 * will appear as an icon.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowIconified(WindowEvent windowEvent){
	}

	/**
	 * Called when a window is initially opened.  This means the window
	 * will now be visible on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowOpened(WindowEvent windowEvent){
	}

	public void windowGainedFocus(WindowEvent windowEvent) {
	}

	public void windowLostFocus(WindowEvent windowEvent) {
	}
}
