package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.util.EventListener;

/**
 * This interface should be implemented by all classes
 * which wish to receive window events.  Another way to
 * listen to <code>WindowEvents</code> is to extend the
 * <code>WindowAdapter</code> class which implements this
 * interface, and override the methods of interest.
 * <p>
 * When a <code>WindowListener</code> is added to a
 * <code>Component</code> using the <code>addWindowListener</code>
 * method, window events occurring on that component will be
 * delivered to the window listener via the sending of one of
 * the following methods:
 * 	- <code>windowActivated</code>
 * 	- <code>windowClosed</code>
 * 	- <code>windowClosing</code>
 * 	- <code>windowDeactivated</code>
 * 	- <code>windowDeiconified</code>
 * 	- <code>windowIconified</code>
 * 	- <code>windowOpened</code>
 */
public interface WindowListener extends EventListener {

	/**
	 * Called when a window is activated.  This means the window
	 * comes to the forground and it or one of its children is
	 * receiving keyboard input.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowActivated(WindowEvent windowEvent);

	/**
	 * Called when a window is closed.  This means the window
	 * will no longer be visible on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowClosed(WindowEvent windowEvent);

	/**
	 * Called when a window is closing.  This means the window
	 * may still be visible on the screen but a user is attempting
	 * to close it.  This could be using the mouse or keyboard
	 * or other input device.
	 * <p>
	 * If a listener does not dispose the window in response to receiving
	 * this event the window will not be closed.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowClosing(WindowEvent windowEvent);

	/**
	 * Called when a window is de-activated.  This means the window
	 * is no longer the forground window and another window or
	 * or one of its children is receiving keyboard input.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowDeactivated(WindowEvent windowEvent);

	/**
	 * Called when a window is de-iconified.  This means the window
	 * is no longer iconified but will appear drawn in its regular
	 * state on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowDeiconified(WindowEvent windowEvent);

	/**
	 * Called when a window is iconified.  This means the window
	 * is no longer drawn in its regular state on the screen but
	 * will appear as an icon.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowIconified(WindowEvent windowEvent);

	/**
	 * Called when a window is initially opened.  This means the window
	 * will now be visible on the screen.
	 *
	 * @param	windowEvent	the event that occured
	 *
	 * @see		<code>WindowEvent</code>
	 */
	public void windowOpened(WindowEvent windowEvent);

}
