package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.*;
import java.util.*;

import com.ibm.oti.awt.GraphicsHandler;
import com.ibm.oti.awt.Util;
import com.ibm.oti.awt.image.*;
import com.ibm.oti.awt.metal.image.ImagePeer;

/**
 * <code>BufferedImage</code> extends the <code>Image</code>
 * class to provide access to the underlying image data.
 *
 * @author      IBM
 * @version		initial
 */
public class BufferedImage extends Image {

	ImageProducer fProducer;
	Vector fObservers;
	IImageImpl fImageWrapper;

	Component fComponent;
	GraphicsConfiguration fGraphicsConfiguration;
//	ColorModel fColorModel;
	ImagePeer tImagePeer;

	static public final int TYPE_CUSTOM = 0;
	static public final int TYPE_INT_RGB = 1;
	static public final int TYPE_INT_ARGB = 2;
	static public final int TYPE_INT_ARGB_PRE = 3;
	static public final int TYPE_INT_BGR = 4;
	static public final int TYPE_USHORT_565_RGB = 8;
	static public final int TYPE_USHORT_555_RGB = 9;
	static public final int TYPE_BYTE_BINARY = 12;
	static public final int TYPE_BYTE_INDEXED = 13;

	static {
		BufferedImageHandler.setImageFactory(new BufferedImageFactory());
	}

	BufferedImage() {
		fObservers = new Vector();
	}

	BufferedImage(GraphicsConfiguration graphicsConfiguration, int width, int height) {
		this();
		this.fComponent = null;
		this.fGraphicsConfiguration = graphicsConfiguration;
		this.tImagePeer = ImagePeer.createImmutableCopy(Util.getDisplay(), width, height, null);
		this.fProducer = new OffScreenImageProducer(tImagePeer);
	}

	/**
	 * Answer the type of image represented by the receiver.
	 *
	 * @return the type of the image
	 */
	public int getType() {
		if (getColorModel() instanceof DirectColorModel) {
			DirectColorModel dcm = (DirectColorModel) getColorModel();
			if ((dcm.redMask == 0xff0000)
				&& (dcm.greenMask == 0xff00)
				&& (dcm.blueMask == 0xff)) {
				if (dcm.hasAlpha()) {
					if (dcm.isAlphaPremultiplied())
						return TYPE_INT_ARGB_PRE;
					else
						return TYPE_INT_ARGB;
				} else {
					return TYPE_INT_RGB;
				}
			}

			if ((dcm.blueMask == 0xff0000)
				&& (dcm.greenMask == 0xff00)
				&& (dcm.redMask == 0xff))
				return TYPE_INT_BGR;
			if ((dcm.redMask == 0xf800)
				&& (dcm.greenMask == 0x07e0)
				&& (dcm.blueMask == 0x001f))
				return TYPE_USHORT_565_RGB;
			if ((dcm.redMask == 0x7c00)
				&& (dcm.greenMask == 0x03e0)
				&& (dcm.blueMask == 0x001f))
				return TYPE_USHORT_555_RGB;
		} else if (getColorModel() instanceof IndexColorModel) {
			IndexColorModel icm = (IndexColorModel) getColorModel();
			int depth = icm.getPixelSize();
			if (((depth ==  1) ||
				(depth == 2) ||
				(depth == 4)) && (!icm.hasAlpha())) {
				return TYPE_BYTE_BINARY;
			} else {
				return TYPE_BYTE_INDEXED;
			}
		}

		return TYPE_CUSTOM;
	}

	/**
	 *
	 */
	public ColorModel getColorModel() {
		ImagePeer imagePeer = getImagePeer();
		if (imagePeer != null) {
			return imagePeer.getColorModel();
		} else {
			return null;
		}
	}

	/**
	 *
	 */
	public int getRGB(int x, int y) {
        if ((x < 0) || (y < 0) || (x >= getWidth()) || (y >= getHeight())) throw new ArrayIndexOutOfBoundsException("Coordinate out of bounds!("+x+", "+y+")");
		ImagePeer imagePeer = getImagePeer();
		if (imagePeer == null) return -1;

		int[] rgb = new int[1];
		imagePeer.getRGB(x, y, 1, 1, rgb, 0, 1);
		return rgb[0];
	}

	/**
	 *
	 */
	public int[] getRGB(int startX, int startY, int width, int height, int[] rgbArray, int offset, int scansize) {
        if ((startX < 0) || (startY < 0) || (startX+width > getWidth()) || (startY+height > getHeight())) throw new ArrayIndexOutOfBoundsException("Coordinate out of bounds!("+startX+", "+startY+", "+width+", "+height+")");
		ImagePeer imagePeer = getImagePeer();
		if (imagePeer == null) return null;

		int result[] = (rgbArray == null ? new int[(height * width) + offset] : rgbArray);
		imagePeer.getRGB(startX, startY, width, height, result, offset, scansize);
		return result;
	}

	/**
	 *
	 */
	public void setRGB(int x, int y, int rgb) {
        if ((x < 0) || (y < 0) || (x >= getWidth()) || (y >= getHeight())) throw new ArrayIndexOutOfBoundsException("Coordinate out of bounds!");
		ImagePeer imagePeer = getImagePeer();
		if (imagePeer == null) return;

		imagePeer.setRGB(x, y, rgb);
	}

	/**
	 *
	 */
	public void setRGB(int startX, int startY, int width, int height, int[] rgbArray, int offset, int scansize) {
        if ((startX < 0) || (startY < 0) || (startX+width > getWidth()) || (startY+height > getHeight())) throw new ArrayIndexOutOfBoundsException("Coordinate out of bounds!");
		ImagePeer imagePeer = getImagePeer();
		if (imagePeer == null) return;

		for (int y = startY; y < (startY + height); y++) {
			for (int x = startX; x < (startX + width); x++) {
				int pixel = rgbArray[offset + (y - startY) * scansize + (x - startX)];
				imagePeer.setRGB(x, y, pixel);
			}
		}
	}

	/**
	 * Answers the width of the receiver if it is known.  Otherwise
	 * answers -1.
	 *
	 * @return int the width of the receiver
	 */
	public int getWidth() {
		return getWidth(null);
	}

	/**
	 * Answers the height of the receiver if it is known.  Otherwise
	 * answers -1.
	 *
	 * @return int the height of the receiver
	 */
	public int getHeight() {
		return getHeight(null);
	}

	/**
	 * Answers the width of the receiver if it is known.  Otherwise
	 * answers -1 and the specified <code>ImageObserver</code> will
	 * be notified when this value is available.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *							notification
	 *
	 * @return	an <code>int</code> which is the receiver's width or <code>-1</code>
	 */
	public int getWidth(ImageObserver imageObserver) {
		ImagePeer imagePeer = getImagePeer(imageObserver);
		if (imagePeer == null)
			return -1;
		Rectangle peerBounds = imagePeer.getBounds();
		return peerBounds.width;
	}

	/**
	 * Answers the height of the receiver if it is known.  Otherwise
	 * answers -1 and the specified <code>ImageObserver</code> will
	 * be notified when this value is available.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *							notification
	 *
	 * @return	an <code>int</code> which is the receiver's height or <code>-1</code>
	 */
	public int getHeight(ImageObserver imageObserver) {
		ImagePeer imagePeer = getImagePeer(imageObserver);
		if (imagePeer == null)
			return -1;
		Rectangle peerBounds = imagePeer.getBounds();
		return peerBounds.height;
	}

	/**
	 * Answers the <code>ImageProducer</code> that generates the
	 * pixels for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	an <code>ImageProducer</code> which is the receiver's source
	 */
	public ImageProducer getSource() {
		return fProducer;
	}

	/**
	 * Answers the property specified by the given name if it is
	 * known.  Otherwise answers <code>null</code> and the specified
	 * <code>ImageObserver</code> will be notified when this value
	 * is available.
	 * <p>
	 * If the property is not defined for the receiver, answers
	 * <code>UndefinedProperty</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	propertyName	the name of the desired property
	 * @param	imageObserver	the <code>ImageObserver</code> used for asynchronous
	 *								notification
	 *
	 * @return	an <code>Object</code> which is the property value or <code>null</code>
	 */
	public Object getProperty(String propertyName, ImageObserver imageObserver) {
//		if (fConsumer != null) {
//			Hashtable properties = fConsumer.getProperties();
//			if (properties != null) {
//				if (properties.containsKey(propertyName)) {
//					Object value = properties.get(propertyName);
//					if (value == null && imageObserver != null) {
//						if (!fObservers.contains(imageObserver)) {
//							fObservers.addElement(imageObserver);
//						}
//					}
//					return value;
//				}
//			}
//		}

		return UndefinedProperty;
	}

	/**
	 * Answers the property specified by the given name if it is
	 * known.  Otherwise answers <code>null</code>.
	 * <p>
	 * If the property is not defined for the receiver, answers
	 * <code>UndefinedProperty</code>.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param	property	the name of the desired property
	 *
	 * @return	an <code>Object</code> which is the property value or <code>null</code>
	 */
	public Object getProperty(String property) {
		return getProperty(property, null);
	}

	/**
	 * Perform clean up for the receiver as it has become garbage.
	 * Called by the virtual machine when there are no longer any (non-weak)
	 * references to the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @exception	Throwable
	 */
	protected void finalize() throws Throwable {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				// Always call flush() for subclasses
				flush();
				_flush(false);
			}
		});
	}

	/**
	 * Releases all resources used by the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 */
	public void flush() {
		// Force even off-screen images to be flushed.
		_flush(true);
	}

	synchronized void _flush(boolean checkForNonFlushables) {
		if (checkForNonFlushables) {
			// Per the spec, flushing an off-screen image does nothing
			// (unless it's being flushed as a result of having been finalized)
			if (fProducer instanceof OffScreenImageProducer) return;
		}

		disposeImage();

		fObservers = new Vector();
	}

	/**
	 * Answers a <code>Graphics</code> used for drawing
	 * on the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Graphics for drawing
	 */
	public Graphics getGraphics() {
		return createGraphics();
	}

	IImageImpl getImageWrapper() {
		if (fImageWrapper == null) {
			fImageWrapper = BufferedImageHandler.createIImageImpl(this);
		}
		return fImageWrapper;
	}

	/**
	 * Answers a new <code>Graphics2D</code> instance used for drawing
	 * on the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return	a java.awt.Graphics2D for drawing
	 */
	public Graphics2D createGraphics() {
		try {
			if (fComponent != null) {
				return (Graphics2D) GraphicsHandler.createGraphics(fComponent, getImageWrapper());
			} else if (fGraphicsConfiguration != null) {
				return (Graphics2D) GraphicsHandler.createGraphics(fGraphicsConfiguration, getImageWrapper());
			} else {
				throw new IllegalAccessError("getGraphics() only valid for images created with createImage(w, h)"); //$NON-NLS-1$
			}
		} catch (IllegalArgumentException e) {
			return null;
		}

	}

	/**
	 * Answers a new subimage which shares the same
	 * data as the receiver for the specified region.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @param		x	the x position of the start of the region
	 * @param		y	the y position of the start of the region
	 * @param		w	the width of the region
	 * @param		h	the height of the region
	 *
	 * @return		a <code>BufferedImage</code>
	 *
	 * @exception	RasterFormatException when the specified region is not contained within the receiver.
	 */
	public BufferedImage getSubimage(int x, int y, int w, int h) {
		if (x < 0 || y < 0 || w < 0 || h < 0)
			throw new RasterFormatException(""); //$NON-NLS-1$
		if ((x + w) > getWidth() || (y + h) > getHeight())
			throw new RasterFormatException(""); //$NON-NLS-1$

		return new SharedSubimage(this, x, y, w, h);
	}

	/**
	 * Answers a string containing a concise, human-readable
	 * description of the receiver's relevant information.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String which is a printable representation for the receiver
	 */
	public String toString() {
		String infoString = ""; //$NON-NLS-1$
		infoString += super.toString();
		infoString += ": type = " + getType(); //$NON-NLS-1$
		infoString += " " + getColorModel(); //$NON-NLS-1$
		infoString += " " + getSource(); //$NON-NLS-1$
		return infoString;
	}

	/**
	 * Answer all recognized property names or <code>null</code> if there are none.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a <code>String[]</code> containing recognized property names
	 * 				or <code>null</code> if there are none
	 */
	public String[] getPropertyNames() {
//		if (fConsumer != null) {
//			Hashtable properties = fConsumer.getProperties();
//			if (properties != null) {
//				Set keys = properties.keySet();
//				if (keys.size() > 0)
//					return (String[]) keys.toArray(new String[keys.size()]);
//			}
//		}

		return null;
	}

	ImagePeer getImagePeer() {
		return getImagePeer(null);
	}

	ImagePeer getImagePeer(ImageObserver imageObserver) {
		if (tImagePeer == null) {
			if (imageObserver != null) {
				if (!fObservers.contains(imageObserver)) {
					fObservers.addElement(imageObserver);
				}
			}
			AbstractImageProducer.buildImageFromProducer(fProducer, getImageWrapper());
		}

		return tImagePeer;
	}

	ImagePeer getPrimitiveImagePeer() {
		return tImagePeer;
	}

	void setImagePeer(ImagePeer imagePeer) {
		// Dispose any existing image before replacing.
		disposeImage();

		tImagePeer = imagePeer;
	}

	Component getComponent() {
		return fComponent;
	}

	void disposeImage() {
	   if (tImagePeer == null) return;
		doDispose();
	}

	void doDispose() {

//		fColorModel = null;
		tImagePeer.dispose();
		tImagePeer = null;
	}

	void notifyObservers(int infoflags, int x, int y, int width, int height) {
		for (int i = fObservers.size() - 1; i >= 0; i--) {
			ImageObserver observer = (ImageObserver) fObservers.elementAt(i);
			boolean keepNotifing = observer.imageUpdate(this, infoflags, x, y, width, height);
			if (!keepNotifing)
				fObservers.removeElement(observer);
		}
	}

}
