package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Transparency;
import java.awt.color.ColorSpace;

/**
 * A <code>ColorModel</code> helps convert pixel values
 * into colors.
 *
 * @author		IBM
 * @version		initial
 *
 * @see			<code>IndexColorModel</code>
 * @see			<code>DirectColorModel</code>
 */
public abstract class ColorModel implements Transparency {

	/**
	 * The bit size of pixel values with this color model.
	 */
	protected int pixel_bits;

	boolean hasAlpha = false;
	boolean isAlphaPremultiplied = false;

	int[] bits;
	ColorSpace colorSpace;
	int transparency;
	int transferType;

	static ColorModel RgbDefault;

	static {
		RgbDefault = new DirectColorModel(32, 0xff0000, 0xff00, 0xff, 0xff000000);
	}

	void init(int pixelBits, int[] bits, ColorSpace colorSpace, boolean hasAlpha, boolean isAlphaPremultiplied, int transparency, int transferType) {
		if (pixelBits < 1)
			throw new IllegalArgumentException();
		if (bits != null && (bits.length < (colorSpace.getNumComponents() + (hasAlpha ? 1 : 0))))
			throw new IllegalArgumentException();
		if ((transparency != Transparency.OPAQUE) &&
		    (transparency != Transparency.BITMASK) &&
		    (transparency != Transparency.TRANSLUCENT))
			throw new IllegalArgumentException();

		this.pixel_bits = pixelBits;
		this.bits = bits;
		this.colorSpace = colorSpace;
		this.hasAlpha = hasAlpha;
		this.isAlphaPremultiplied = isAlphaPremultiplied;
		this.transparency = transparency;
		this.transferType = transferType;
	}

	void setComponentSize(int[] bits) {
		this.bits = bits;
	}

	/**
	 * Constructs a new <code>ColorModel</code>, specifying
	 * the pixel bit size
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param pixelBits the bit size of pixel values
	 */
	public ColorModel(int pixelBits) {
		ColorSpace colorSpace = ColorSpace.getInstance(ColorSpace.CS_sRGB);

		int transferType;
		if (pixelBits > 32)
			transferType = DataBuffer.TYPE_UNDEFINED;
		else if (pixelBits > 16)
			transferType = DataBuffer.TYPE_INT;
		else if (pixelBits > 8)
			transferType = DataBuffer.TYPE_USHORT;
		else
			transferType = DataBuffer.TYPE_BYTE;

		init(pixelBits, null, colorSpace, true, false, Transparency.TRANSLUCENT, transferType);
	}

	/**
	 * Constructs a new <code>ColorModel</code> that translates pixel
	 * values to color/alpha components.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param pixelBits the bit size of pixel values
	 * @param bits array indicating the number of bits per color and alpha component
	 * @param colorSpace the ColorSpace
	 * @param hasAlpha indicates whether alpha information is present
	 * @param isAlphaPremultiplied indicates whether color samples are assumed to be premultiplied by the alpha samples
	 * @param transparency indicates which alpha values can be represented by this color model
	 * @param transferType the type of the array used to represent pixel values
	 */
	protected ColorModel(int pixelBits, int[] bits, ColorSpace colorSpace, boolean hasAlpha, boolean isAlphaPremultiplied, int transparency, int transferType) {
		init(pixelBits,
		     bits,
		     colorSpace,
		     hasAlpha,
		     isAlphaPremultiplied,
		     transparency,
		     transferType);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public static ColorModel getRGBdefault() {
		return RgbDefault;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final boolean hasAlpha() {
		return hasAlpha;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final boolean isAlphaPremultiplied() {
		return isAlphaPremultiplied;
	}

	/**
	 * Answer the transfer type of the receiver.
	 * The transfer type is the type of primitive array
	 * used to represent pixel values as arrays.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return int
	 */
	public final int getTransferType() {
		return transferType;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public int getPixelSize() {
		return pixel_bits;
	}

	/**
	 * Answer the number of bits for the specified
	 * color/alpha component. Color components are indexed
	 * in the order specified by the ColorSpace. Typically,
	 * this order reflects the name of the color space type.
	 * For example, for TYPE_RGB, index 0 corresponds to red,
	 * index 1 to green, and index 2 to blue. If this ColorModel
	 * supports alpha, the alpha component corresponds to the
	 * index following the last color component.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param int0
	 * @return int
	 */
	public int getComponentSize(int index) {
		return getComponentSize()[index];
	}

	/**
	 * Answer an array containing the number of bits per
	 * color/alpha component.
	 *
	 * The array contains the color
	 * components in the order specified by the ColorSpace.
	 * If an alpha component exists, the number of bits for
	 * it is also included (in the last array position).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return int[] an array of the number of bits per
	 * 					color/alpha component
	 */
	public int[] getComponentSize() {
		return bits;
	}

	/**
	 * Answer the transparency value, which is one
	 * of:
	 * OPAQUE, BITMASK, or TRANSLUCENT.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see Transparency#getTransparency()
	 */
	public int getTransparency() {
		return transparency;
	}

	/**
	 * Answer the number of components, including alpha,
	 * in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return int total number of components
	 */
	public int getNumComponents() {
		return getNumColorComponents() + (hasAlpha() ? 1 : 0);
	}

	/**
	 * Answer the number of color components in the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see ColorSpace#getNumComponents()
	 * @return int number of color components
	 */
	public int getNumColorComponents() {
		return getColorSpace().getNumComponents();
	}

	/**
	 * Answer the red component of a pixel value.
	 *
	 * @param pixel
	 * @return int red color component
	 */
	public abstract int getRed(int pixel);

	/**
	 * Answer the green component of a pixel value.
	 *
	 * @param pixel
	 * @return int green color component
	 */
	public abstract int getGreen(int pixel);

	/**
	 * Answer the blue component of a pixel value.
	 *
	 * @param pixel
	 * @return int blue color component
	 */
	public abstract int getBlue(int pixel);

	/**
	 * Answer the alpha component of a pixel value.
	 *
	 * @param pixel
	 * @return int alpha transparency component
	 */
	public abstract int getAlpha(int pixel);

	/**
	 * Answer a pixel value converted to the default
	 * color model.
	 *
	 * @param pixel
	 * @return int red color component
	 */
	public int getRGB(int pixel) {
		return (getAlpha(pixel) << 24) |
			(getRed(pixel) << 16) |
			(getGreen(pixel) << 8) |
			(getBlue(pixel));
	}

	/**
	 * Tests if the specified Object is an instance of ColorModel
	 * and if it equals this ColorModel.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see Object#equals(Object)
	 */
	public boolean equals(Object object) {
		if (object == this) {
			return true;
		}
		if (!(object instanceof ColorModel)) {
			return false;
		}
		ColorModel colorModel = (ColorModel)object;
		return getPixelSize() == colorModel.getPixelSize();
	}

	/**
	 * Answer the hash code for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see Object#hashCode()
	 */
	public int hashCode() {
		return getPixelSize();
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final ColorSpace getColorSpace() {
		if (colorSpace != null) {
			return colorSpace;
		}
		return ColorSpace.getInstance(ColorSpace.CS_sRGB);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void finalize() {
		this.colorSpace = null;
	}

	/**
	 * Answer a string representation for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see Object#toString()
	 */
	public String toString() {
		return super.toString();
	}

}
