package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Transparency;
import java.awt.color.ColorSpace;

/**
 * A <code>DirectColorModel</code> is a model in which
 * the color information is contained within the pixel
 * values.
 *
 * @author      IBM
 * @version		initial
 */
public class DirectColorModel extends ColorModel {

	int alphaMask;
	int redMask;
	int greenMask;
	int blueMask;

	int alphaShift;
	int redShift;
	int greenShift;
	int blueShift;

	int alphaMax;
	int redMax;
	int greenMax;
	int blueMax;

	int countBits(int mask, int bit) {
		int count = 0;
		while (mask != 0 && ((mask & 1) == bit)) {
			count++;
			mask = mask >>> 1;
		}
		return count;
	}

	void checkMask(int mask, int shift, int count, String maskName) {
		long maxMask = (1l << getPixelSize()) - 1;
		if (mask > maxMask)
			throw new IllegalArgumentException(maskName + " mask overflows: " + Integer.toHexString(mask)); //$NON-NLS-1$

		if ((mask >>> (shift + count)) > 0)
			throw new IllegalArgumentException(maskName + " mask not contiguos: " + Integer.toHexString(mask)); //$NON-NLS-1$
	}

	/**
	 * Tests if the specified Object is an instance of DirectColorModel
	 * and if it equals this DirectColorModel.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see ColorModel#equals(Object)
	 */
	public boolean equals(Object object) {
		if (super.equals(object) == false) {
			return false;
		}
		if ((object instanceof DirectColorModel) == false) {
			return false;
		}
		DirectColorModel colorModel = (DirectColorModel) object;

		if (alphaMask != colorModel.alphaMask ||
			redMask != colorModel.redMask ||
			greenMask != colorModel.greenMask ||
			blueMask != colorModel.blueMask) {
			return false;
		}
		if (alphaShift != colorModel.alphaShift ||
			redShift != colorModel.redShift ||
			greenShift != colorModel.greenShift ||
			blueShift != colorModel.blueShift) {
			return false;
		}
		if (alphaMax != colorModel.alphaMax ||
			redMax != colorModel.redMax ||
			greenMax != colorModel.greenMax ||
			blueMax != colorModel.blueMax) {
			return false;
		}
		return true;
	}

	/**
	 * Answer the hash code for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see ColorModel#hashCode()
	 */
	public int hashCode() {
		return super.hashCode() ^ alphaMask ^ redMask ^ greenMask ^ blueMask ^ alphaShift ^ redShift ^ greenShift ^ blueShift ^ alphaMax ^ redMax ^ greenMax ^ blueMax;
	}

	void initAlphaInfo() {
		int alpha= getAlpha(0);
		if (alpha == 255) {
			this.hasAlpha = false;
			this.transparency = Transparency.OPAQUE;
		} else {
			this.hasAlpha = true;
			this.transparency = Transparency.TRANSLUCENT;
		}
	}

	void init(int redMask, int greenMask, int blueMask, int alphaMask) {
		if (redMask == 0 || greenMask == 0 || blueMask == 0)
			throw new IllegalArgumentException("rgb masks can not be 0"); //$NON-NLS-1$

		this.alphaMask = alphaMask;
		this.redMask = redMask;
		this.greenMask = greenMask;
		this.blueMask = blueMask;

		alphaShift = countBits(alphaMask, 0);
		int alphaCount = countBits(alphaMask >>> alphaShift, 1);
		checkMask(alphaMask, alphaShift, alphaCount, "alpha"); //$NON-NLS-1$
		alphaMax = (1 << alphaCount) - 1;

		redShift = countBits(redMask, 0);
		int redCount = countBits(redMask >>> redShift, 1);
		checkMask(redMask, redShift, redCount, "red"); //$NON-NLS-1$
		redMax = (1 << redCount) - 1;

		greenShift = countBits(greenMask , 0);
		int greenCount = countBits(greenMask >>> greenShift, 1);
		checkMask(greenMask, greenShift, greenCount, "green"); //$NON-NLS-1$
		greenMax = (1 << greenCount) - 1;

		blueShift = countBits(blueMask, 0);
		int blueCount = countBits(blueMask >>> blueShift, 1);
		checkMask(blueMask, blueShift, blueCount, "blue"); //$NON-NLS-1$
		blueMax = (1 << blueCount) - 1;

		initAlphaInfo();

		if (alphaCount > 0) {
			setComponentSize(new int[] {redCount, greenCount, blueCount, alphaCount});
		} else {
			setComponentSize(new int[] {redCount, greenCount, blueCount});
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public DirectColorModel(int nbits, int redMask, int greenMask, int blueMask) {
		this(nbits, redMask, greenMask, blueMask, 0);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public DirectColorModel(int nbits, int redMask, int greenMask, int blueMask, int alphaMask) {
		super(nbits);

		init(redMask, greenMask, blueMask, alphaMask);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public DirectColorModel(ColorSpace space, int nbits, int redMask, int greenMask, int blueMask, int alphaMask, boolean isAlphaPremultiplied, int transferType) {
		super(nbits,
		      null,
		      space,
		      false, // this will be properly set during init
		      isAlphaPremultiplied,
		      Transparency.OPAQUE, // this will be properly set during init
		      transferType);

		if (space.getType() != ColorSpace.TYPE_RGB)
			throw new IllegalArgumentException();

		init(redMask, greenMask, blueMask, alphaMask);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getRedMask() {
		return redMask;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getGreenMask() {
		return greenMask;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getBlueMask() {
		return blueMask;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getAlphaMask() {
		return alphaMask;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getRed(int pixel) {
		int r = (pixel & redMask) >>> redShift;
		if (redMax != 0)
			r = r * 255 / redMax;

		if (isAlphaPremultiplied()) {
			int alpha= getAlpha(pixel);
			r = (alpha == 0 ? 0 : r * 255 / alpha);
		}
		return r;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getGreen(int pixel) {
		int g = (pixel & greenMask) >>> greenShift;
		if (greenMax != 0)
			g = g * 255 / greenMax;

		if (isAlphaPremultiplied()) {
			int alpha= getAlpha(pixel);
			g = (alpha == 0 ? 0 : g * 255 / alpha);
		}
		return g;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getBlue(int pixel) {
		int b = (pixel & blueMask) >>> blueShift;
		if (blueMax != 0)
			b = b * 255 / blueMax;

		if (isAlphaPremultiplied()) {
			int alpha= getAlpha(pixel);
			b = (alpha == 0 ? 0 : b * 255 / alpha);
		}
		return b;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getAlpha(int pixel) {
		if (alphaMask == 0)
			return 255;

		int a = (pixel & alphaMask) >>> alphaShift;
		if (alphaMax != 0)
			a = a * 255 / alphaMax;
		return a;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getRGB(int pixel) {
		return super.getRGB(pixel);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
		buffer.append("DirectColorModel: rmask="); //$NON-NLS-1$
		buffer.append(Integer.toHexString(getRedMask()));
		buffer.append(" gmask="); //$NON-NLS-1$
		buffer.append(Integer.toHexString(getGreenMask()));
		buffer.append(" bmask="); //$NON-NLS-1$
		buffer.append(Integer.toHexString(getBlueMask()));
		buffer.append(" amask="); //$NON-NLS-1$
		buffer.append(Integer.toHexString(getAlphaMask()));
		return buffer.toString();
	}

}
