package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.awt.Transparency;

/**
 * An <code>IndexColorModel</code> helps convert pixel values into colors.  These
 * pixel values are indexes into the color palette.
 */
public class IndexColorModel extends ColorModel {

	byte[] alphas;
	byte[] reds;
	byte[] greens;
	byte[] blues;

	int transPixel;

	private void checkArgs(int nbits, int size) {
		if ((nbits < 1) || (nbits > 16) || (size < 1))
			throw new IllegalArgumentException();
	}

	/**
	 * Tests if the specified Object is an instance of IndexColorModel
	 * and if it equals this IndexColorModel.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see ColorModel#equals(Object)
	 */
	public boolean equals(Object object) {
		if (super.equals(object) == false) {
			return false;
		}
		if ((object instanceof IndexColorModel) == false) {
			return false;
		}
		IndexColorModel colorModel = (IndexColorModel) object;

		if (transPixel != colorModel.transPixel) {
			return false;
		}
		if (arrayEquals(alphas, colorModel.alphas) == false) {
			return false;
		}
		if (arrayEquals(reds, colorModel.reds) == false) {
			return false;
		}
		if (arrayEquals(greens ,colorModel.greens) == false) {
			return false;
		}
		if (arrayEquals(blues, colorModel.blues) == false) {
			return false;
		}
		return true;
	}

	private boolean arrayEquals(byte[] array1, byte[] array2) {
		if (array1 == null && array2 == null) {
			return true;
		}
		if (array1 == null || array2 == null) {
			return false;
		}
		if (array1.length != array2.length) {
			return false;
		}
		for (int i = 0; i < array1.length; i++) {
			if (array1[i] != array2[i]) {
				return false;
			}
		}
		return true;
	}
	/**
	 * Answer the hash code for the receiver.
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @see ColorModel#hashCode()
	 */
	public int hashCode() {
		int alphasHash = 0;
		int redsHash = 0;
		int greensHash = 0;
		int bluesHash = 0;

		for (int i = 0; i < alphas.length; i++) {
			alphasHash += alphas[i];
		}
		for (int i = 0; i < reds.length; i++) {
			redsHash = redsHash * 2 + reds[i];
		}
		for (int i = 0; i < greens.length; i++) {
			greensHash = greensHash * 3 + greens[i];
		}
		for (int i = 0; i < blues.length; i++) {
			bluesHash = bluesHash * 4 + blues[i];
		}

		return super.hashCode() ^ transPixel ^ alphasHash ^ redsHash ^ greensHash ^ bluesHash;
	}

	private void init(int size, byte reds[], byte greens[], byte[] blues) {
		this.reds = new byte[size];
		System.arraycopy(reds, 0, this.reds, 0, size);

		this.greens = new byte[size];
		System.arraycopy(greens, 0, this.greens, 0, size);

		this.blues = new byte[size];
		System.arraycopy(blues, 0, this.blues, 0, size);
	}

	private void init(int size, byte reds[], byte greens[], byte[] blues, byte alphas[]) {
		this.reds = new byte[size];
		System.arraycopy(reds, 0, this.reds, 0, size);

		this.greens = new byte[size];
		System.arraycopy(greens, 0, this.greens, 0, size);

		this.blues = new byte[size];
		System.arraycopy(blues, 0, this.blues, 0, size);

		if (alphas != null) {
			this.alphas = new byte[size];
			System.arraycopy(alphas, 0, this.alphas, 0, size);
		}
	}

	void initTransparencyFields(int defaultTransPixel) {
		this.transPixel = defaultTransPixel;
		this.transparency = Transparency.OPAQUE;
		this.hasAlpha = false;
		if (alphas != null) {
			for (int i=0;i<alphas.length;i++) {
				if (alphas[i] != (byte)0xFF) {
					this.hasAlpha = true;
					if (alphas[i] == (byte)0x00) {
						if (transparency == Transparency.OPAQUE) transparency = Transparency.BITMASK;
						if (transPixel == -1) transPixel = i;
					} else {
						this.transparency = Transparency.TRANSLUCENT;
					}
				}
			}

			if (hasAlpha == false) {
				alphas = null;
			}
		}
	}

	/**
	 * Constructs a new <code>IndexColorModel</code>
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @param nbits
	 * @param size
	 * @param reds
	 * @param greens
	 * @param blues
	 */
	public IndexColorModel(int nbits, int size, byte reds[], byte greens[], byte[] blues) {
		super(nbits);

		checkArgs(nbits, size);

		init(size, reds, greens, blues, null);
		initTransparencyFields(-1);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public IndexColorModel(int nbits, int size, byte reds[], byte greens[], byte[] blues, int transPixel) {
		super(nbits);

		checkArgs(nbits, size);

		if ((transPixel >= 0) && (transPixel < size)) {
			this.alphas = new byte[size];
			for (int i=0;i<size;i++) alphas[i] = (byte)0xFF;
			this.alphas[transPixel] = 0x00;

			this.hasAlpha = true;
			this.transPixel = transPixel;
			this.transparency = Transparency.BITMASK;
		} else {
			this.hasAlpha = false;
			this.transPixel = -1;
			this.transparency = Transparency.OPAQUE;
		}

		init(size, reds, greens, blues);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public IndexColorModel(int nbits, int size, byte reds[], byte greens[], byte[] blues, byte alphas[]) {
		super(nbits);

		checkArgs(nbits, size);

		init(size, reds, greens, blues, alphas);
		initTransparencyFields(-1);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public IndexColorModel(int nbits, int size, byte rgbs[], int offset, boolean hasAlpha) {
		this(nbits, size, rgbs, offset, hasAlpha, -1);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public IndexColorModel(int nbits, int size, byte rgbs[], int offset, boolean hasAlpha, int transPixel) {
		super(nbits);

		checkArgs(nbits, size);

		transPixel = ((transPixel >= 0) && (transPixel < size)) ? transPixel : -1;

		if (hasAlpha == false) {
			this.transPixel = transPixel;
			if (transPixel != -1) {
				this.alphas = new byte[size];
				for (int i=0;i<size;i++) alphas[i] = (byte)0xFF;
				alphas[transPixel] = 0x00;

				this.hasAlpha = true;
				this.transparency = Transparency.BITMASK;
			} else {
				this.hasAlpha = false;
				this.transparency = Transparency.OPAQUE;
			}
		} else {
			this.alphas = new byte[size];
		}

		this.reds = new byte[size];
		this.greens = new byte[size];
		this.blues = new byte[size];

		int entryLength= (hasAlpha ? 4 : 3);
		for (int i=0; i<size; i++) {
			int off = offset + (i * entryLength);
			reds[i] = rgbs[off];
			greens[i] = rgbs[off + 1];
			blues[i] = rgbs[off + 2];
			if (hasAlpha) {
				alphas[i] = rgbs[off + 3];
			}
		}

		if (hasAlpha) {
			initTransparencyFields(transPixel);
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getMapSize() {
		return reds.length;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getTransparentPixel() {
		return transPixel;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final void getReds(byte reds[]) {
		System.arraycopy(this.reds, 0, reds, 0, getMapSize());
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final void getGreens(byte greens[]) {
		System.arraycopy(this.greens, 0, greens, 0, getMapSize());
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final void getBlues(byte blues[]) {
		System.arraycopy(this.blues, 0, blues, 0, getMapSize());
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final void getAlphas(byte alphas[]) {
		int size = getMapSize();
		if (this.alphas != null) {
			System.arraycopy(this.alphas, 0, alphas, 0, size);
		} else {
			for (int i=0;i<size;i++) {
				alphas[i] = (byte)0xFF;
			}
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getRed(int pixel) {
		try {
			return reds[pixel] & 0xff;
		} catch (ArrayIndexOutOfBoundsException e) {
			return 0;
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getGreen(int pixel) {
		try {
			return greens[pixel] & 0xff;
		} catch (ArrayIndexOutOfBoundsException e) {
			return 0;
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getBlue(int pixel) {
		try {
			return blues[pixel] & 0xff;
		} catch (ArrayIndexOutOfBoundsException e) {
			return 0;
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getAlpha(int pixel) {
		if (hasAlpha == false) return 0xFF;
		try {
			return alphas[pixel] & 0xff;
		} catch (ArrayIndexOutOfBoundsException e) {
			return 0;
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public final int getRGB(int pixel) {
		return super.getRGB(pixel);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void finalize() {
	}

	/**
	 * Answer an array containing the number of bits per
	 * color/alpha component.
	 *
	 * The array contains the color
	 * components in the order specified by the ColorSpace.
	 * If an alpha component exists, the number of bits for
	 * it is also included (in the last array position).
	 *
	 * @author	IBM
	 * @version	initial
	 *
	 * @return int[] an array of the number of bits per
	 * 					color/alpha component
	 */
	public int[] getComponentSize() {
		if (hasAlpha) return new int[] {8,8,8,8};
		else return new int[] {8,8,8};
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
		buffer.append("IndexColorModel: pixelBits="); //$NON-NLS-1$
		buffer.append(getPixelSize());
		buffer.append(" numComponents="); //$NON-NLS-1$
		buffer.append(getNumComponents());
		buffer.append(" color space="); //$NON-NLS-1$
		buffer.append(getColorSpace());
		buffer.append(" transparency="); //$NON-NLS-1$
		buffer.append(getTransparency());
		buffer.append(" has alpha="); //$NON-NLS-1$
		buffer.append(hasAlpha());
		buffer.append(" isAlphaPre="); //$NON-NLS-1$
		buffer.append(isAlphaPremultiplied());
		return buffer.toString();
	}

}
