package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.util.Hashtable;
import java.util.Vector;

//------------------------------------------------------------
/**
 *
 */
public class MemoryImageSource implements ImageProducer {

	int width;
	int height;
	int offset;
	int scanline;
	Object pixels;
	ColorModel colorModel;
	Hashtable properties;

	Vector consumers;
	boolean animated;
	boolean fullBuffers;

	void init(ColorModel colorModel, Object pixels, int offset, int scanline) {
		this.colorModel = colorModel;
		this.pixels = pixels;
		this.offset = offset;
		this.scanline = scanline;
	}

	void init(int width, int height, ColorModel colorModel, Object pixels, int offset, int scanline, Hashtable properties) {
		init(colorModel, pixels, offset, scanline);
		this.width = width;
		this.height = height;
		this.properties = (properties == null) ? new Hashtable() : properties;
		this.consumers = new Vector();
	}

	void init(ImageConsumer ic) {
		ic.setDimensions(width, height);
		ic.setProperties(properties);
		ic.setColorModel(colorModel);
		int hints;
		if (animated) {
			if (fullBuffers)
				hints = ImageConsumer.TOPDOWNLEFTRIGHT | ImageConsumer.COMPLETESCANLINES;
			else
				hints = ImageConsumer.RANDOMPIXELORDER;
		} else {
			hints = ImageConsumer.TOPDOWNLEFTRIGHT | ImageConsumer.COMPLETESCANLINES |
				ImageConsumer.SINGLEPASS | ImageConsumer.SINGLEFRAME;
		}
		ic.setHints(hints);
		setPixels(ic, 0, 0, width, height, true);
	}

	synchronized void setPixels(int x, int y, int w, int h, boolean singleFrameNotify) {
		for (int i=0; i<consumers.size(); i++)
			setPixels((ImageConsumer)consumers.elementAt(i), x, y, w, h, singleFrameNotify);
	}

	void setPixels(ImageConsumer ic, int x, int y, int w, int h, boolean singleFrameNotify) {
		if(w < 0 || h < 0) return;
		int pixelArrayOffset = (y * scanline + x) + offset;
		if (pixels instanceof int[])
			ic.setPixels(x, y, w, h, colorModel, (int[])pixels, pixelArrayOffset, scanline);
		else
			ic.setPixels(x, y, w, h, colorModel, (byte[])pixels, pixelArrayOffset, scanline);

		if (animated) {
			if (singleFrameNotify)
				ic.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
		} else {
			ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, ColorModel colorModel, byte[] pixels, int offset, int scanline) {
		init(width, height, colorModel, pixels, offset, scanline, null);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, ColorModel colorModel, byte[] pixels, int offset, int scanline, Hashtable properties) {
		init(width, height, colorModel, pixels, offset, scanline, properties);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, ColorModel colorModel, int[] pixels, int offset, int scanline) {
		init(width, height, colorModel, pixels, offset, scanline, null);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, ColorModel colorModel, int[] pixels, int offset, int scanline, Hashtable properties) {
		init(width, height, colorModel, pixels, offset, scanline, properties);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, int[] pixels, int offset, int scanline) {
		init(width, height, ColorModel.getRGBdefault(), pixels, offset, scanline, null);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public MemoryImageSource(int width, int height, int[] pixels, int offset, int scanline, Hashtable properties) {
		init(width, height, ColorModel.getRGBdefault(), pixels, offset, scanline, properties);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void addConsumer(ImageConsumer imageConsumer) {
		if (isConsumer(imageConsumer))
			return;

		if (animated)
			consumers.addElement(imageConsumer);

		init(imageConsumer);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized boolean isConsumer(ImageConsumer imageConsumer) {
		return consumers.contains(imageConsumer);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void removeConsumer(ImageConsumer imageConsumer) {
		consumers.removeElement(imageConsumer);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void startProduction(ImageConsumer imageConsumer) {
		addConsumer(imageConsumer);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void requestTopDownLeftRightResend(ImageConsumer imageConsumer) {
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void setAnimated(boolean animated) {
		this.animated = animated;

		if (!animated) {
			for (int i=0; i<consumers.size(); i++)
				((ImageConsumer)consumers.elementAt(i)).imageComplete(ImageConsumer.STATICIMAGEDONE);
			consumers.removeAllElements();
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void setFullBufferUpdates(boolean fullBuffers) {
		this.fullBuffers = fullBuffers;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void newPixels() {
		if (!animated) return;
		setPixels(0, 0, width, height, true);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void newPixels(int x, int y, int width, int height) {
		newPixels(x, y, width, height, true);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void newPixels(int x, int y, int w, int h, boolean singleFrameNotify) {
		if (!animated) return;
		if (fullBuffers)
			setPixels(0, 0, width, height, singleFrameNotify);
		else {
			if (x < 0) {
				w = w + x;
				x = 0;
			}
			if (y < 0) {
				h = h + y;
				y = 0;
			}
			if ((x + w) > width) {
				w = width - x;
			}
			if ((y + h) > height) {
				h = height - y;
			}
			setPixels(x, y, w, h, singleFrameNotify);
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void newPixels(byte[] newPixels, ColorModel newCM, int offset, int scanline) {
		init(newCM, newPixels, offset, scanline);
		newPixels();
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void newPixels(int[] newPixels, ColorModel newCM, int offset, int scanline) {
		init(newCM, newPixels, offset, scanline);
		newPixels();
	}

}
