package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Image;
import java.awt.Rectangle;
import java.util.Hashtable;

//------------------------------------------------------------
/**
 *
 */
public class PixelGrabber implements ImageConsumer {

	ImageProducer producer;
	Rectangle rect;
	int offset;
	int scanline;
	int[] pixelsInInt;
	byte[] pixelsInByte;
	ColorModel lastColorModel;
	boolean forceRGB;
	int status;

	//------------------------------------------------------------
	/**
	 *
	 */
	public PixelGrabber(Image image, int x, int y, int width, int height, int[] pixels, int offset, int scanline) {
		this(image.getSource(), x, y, width, height, pixels, offset, scanline);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public PixelGrabber(ImageProducer imageProducer, int x, int y, int width, int height, int[] pixels, int offset, int scanline) {
		this.producer = imageProducer;
		this.rect = new Rectangle(x, y, width, height);
		this.pixelsInInt = pixels;
		this.offset = offset;
		this.scanline = scanline;
		this.lastColorModel = ColorModel.getRGBdefault();
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public PixelGrabber(Image image, int x, int y, int width, int height, boolean forceRGB) {
		this(image, x, y, width, height, null, 0, width);
		this.forceRGB = forceRGB;
		if (!forceRGB) this.lastColorModel = null;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void startGrabbing() {
		producer.startProduction(this);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void abortGrabbing() {
		imageComplete(ImageConsumer.IMAGEABORTED);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public boolean grabPixels() throws InterruptedException {
		return grabPixels(0);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized boolean grabPixels(long timeout) throws InterruptedException {
		status &= ~(ImageObserver.ALLBITS | ImageObserver.FRAMEBITS);

		startGrabbing();

		if (timeout < 0) {
			// If timeout < 0, then there is no waiting.
		} else if (timeout == 0) {
			// Wait until all pixels are delivered.
			while (true) {
				if (((status & ImageObserver.ALLBITS) > 0 ||
				     (status & ImageObserver.FRAMEBITS) > 0))
					return true;

				if (((status & ImageObserver.ERROR) > 0 ||
				     (status & ImageObserver.ABORT) > 0))
					return false;

				wait(500);
			}
		} else if (timeout > 0) {
			// Allow up to the timeout for the pixels to be delivered.
			if (!((status & ImageObserver.ALLBITS) > 0 ||
			      (status & ImageObserver.FRAMEBITS) > 0))
				wait(timeout);
		}

		return (status & ImageObserver.ALLBITS) > 0 ||
			(status & ImageObserver.FRAMEBITS) > 0;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public int getStatus() {
		return status;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public int getWidth() {
		return (rect.width < 0) ? -1 : rect.width;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public int getHeight() {
		return (rect.height < 0) ? -1 : rect.height;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public Object getPixels() {
		return (pixelsInInt != null) ? (Object)pixelsInInt : (Object)pixelsInByte;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized ColorModel getColorModel() {
//		if (pixelsInInt != null)
//			return ColorModel.getRGBdefault();
		return lastColorModel;
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void setDimensions(int width, int height) {
		status |= ImageObserver.WIDTH | ImageObserver.HEIGHT;
		if (rect.width < 0)
			rect.width = width - rect.x;

		if (scanline < 0)
			scanline = rect.width;

		if (rect.height < 0)
			rect.height = height - rect.y;

		if (forceRGB && pixelsInInt == null)
			this.pixelsInInt = new int[rect.width * rect.height];
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void setHints(int hints) {
		// NOP
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void setProperties(Hashtable properties) {
		// NOP
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void setColorModel(ColorModel colorModel) {
        // The color model passed in here is only a hint. It indicates
        // what color model most of the pixels will be delivered in
        // when calling setPixels.
        // If the same color model is used for every call to setPixels
        // then pixels are stored in that color model. If more than one
        // color model is used the pixels are stored in the default RGB
        // color model.
	}
	//------------------------------------------------------------
	/**
	 *
	 */
	public void setPixels(int x, int y, int width, int height, ColorModel colorModel, byte[] pixs, int off, int scan) {
		Rectangle inters = rect.intersection(new Rectangle(x, y, width, height));
		if (inters.isEmpty()) return;

		status |= ImageObserver.SOMEBITS;

		if (pixelsInInt == null) {
			if (pixelsInByte == null) {
				// We can stores the bytes directly in a byte[]: pixelsInByte
				this.pixelsInByte = new byte[rect.width * rect.height];
				this.lastColorModel = colorModel;
			}
			// At the point pixelsInByte != null and pixelsInInt == null
			if (lastColorModel == colorModel) {
				// We can directly copy pixs[i] into the pixelsInByte
				copyPixelArea(inters, pixs, pixelsInByte, y, off, scan);
				return;
			} else {
				// Convert index bytes to RGB int to switch to another ColorModel
				// ==> (pixelsInInt != null)
				convertAllBytesIntoRGBInts();
			}
		}
		// At the point pixelsInInt != null and the pixs[i] have not been copied into pixelsInInt
		convertPixelAreaIntoRGB(inters, pixs, colorModel, y, off, scan);
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public void setPixels(int x, int y, int width, int height, ColorModel colorModel, int[] pixs, int off, int scan) {
		Rectangle inters = rect.intersection(new Rectangle(x, y, width, height));
		if (inters.isEmpty()) return;

		this.status |= ImageObserver.SOMEBITS;
		// Even if the (lastColorModel == colorModel) we must convert the bytes to ints
		// because we will store the pixs in an int[]
		if (pixelsInByte != null) {
			// Convert all bytes to RGB int to be able to add new ints
			convertAllBytesIntoRGBInts();
		}
		// At this point pixelsInByte == null!
		if (pixelsInInt == null) {
			this.pixelsInInt = new int[rect.width * rect.height];
			this.lastColorModel = colorModel;
		}
		// At this point pixelsInInt != null!
		if (lastColorModel == colorModel) {
			// We can directly copy pixs int into the pixelsInInt
			copyPixelArea(inters, pixs, pixelsInInt, y, off, scan);
		} else {
			if (lastColorModel != ColorModel.getRGBdefault()) {
				// We must convert the existing pixelsInInt into RGB values!
				convertAllIntsIntoRGBInts();
			}
			// At this point the pixelsInInt contains only RGB values
			if (colorModel != ColorModel.getRGBdefault()) {
				// Now we can convert each pixs[i] to RGB and store them into pixelsInInt
				convertPixelAreaIntoRGB(inters, pixs, colorModel, y, off, scan);
			} else {
				/// The new pixels are RGN pixels so we can directly copy them into the pixelsInInt
				copyPixelArea(inters, pixs, pixelsInInt, y, off, scan);
			}
		}
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public synchronized void imageComplete(int flags) {
		if (flags == ImageConsumer.STATICIMAGEDONE)
			status |= ImageObserver.ALLBITS;
		else if (flags == ImageConsumer.SINGLEFRAMEDONE)
			status |= ImageObserver.FRAMEBITS;
		else if (flags == ImageConsumer.IMAGEABORTED)
			status |= ImageObserver.ABORT;
		else if (flags == ImageConsumer.IMAGEERROR)
			status |= ImageObserver.ERROR;
		notifyAll();
	}

	//------------------------------------------------------------
	/**
	 *
	 */
	public int status() {
		return getStatus();
	}
	//------------------------------------------------------------
	/*
	 * Converts the array of bytes (pixelsInByte) into RGB values
	 * and stores the result into pixelsInInt;
	 */
	private final void convertAllBytesIntoRGBInts() {
		// If the (pixelsInByte != null) ==> (pixelsInInt == null)
		int len = rect.width * rect.height;
		this.pixelsInInt = new int[len];
		for (int i = 0; i < len; i++) {
			pixelsInInt[i] = lastColorModel.getRGB(pixelsInByte[i] & 0xFF);
		}
		this.pixelsInByte = null;
		// Now the current ColorModel is the RGB ColorModel!
		this.lastColorModel = ColorModel.getRGBdefault();
	}
	/*
	 * Converts the array of ints (pixelsInInt) into RGB values
	 * and stores the result into pixelsInInt;
	 */
	private final void convertAllIntsIntoRGBInts() {
		int len = pixelsInInt.length;
		for (int i = 0; i < len; i++) {
			pixelsInInt[i] = lastColorModel.getRGB(pixelsInInt[i]);
		}
		// Now the current ColorModel is the RGB ColorModel!
		this.lastColorModel = ColorModel.getRGBdefault();
	}
	private final void convertPixelAreaIntoRGB(Rectangle area, byte[] pixs, ColorModel colorModel, int y, int off, int scan) {
		int xend = area.x + area.width;
		int yend = area.y + area.height;
		for (int cy = area.y; cy < yend; cy++) {
			int dstOff = (cy - rect.y) * scanline  + offset - rect.x;
			int srcOff = (cy - y) * scan + off;
			for (int cx = area.x; cx < xend; cx++) {
				pixelsInInt[dstOff + cx] = colorModel.getRGB(pixs[srcOff + cx] & 0xFF);
			}
		}
		// At the point the ColorModel for pixelsInInt is the RGB ColorModel
		this.lastColorModel = ColorModel.getRGBdefault();

	}
	private final void convertPixelAreaIntoRGB(Rectangle area, int[] pixs, ColorModel colorModel, int y, int off, int scan) {
		int xend = area.x + area.width;
		int yend = area.y + area.height;
		for (int cy = area.y; cy < yend; cy++) {
			int dstOff = (cy - rect.y) * scanline  + offset - rect.x;
			int srcOff = (cy - y) * scan + off;
			for (int cx = area.x; cx < xend; cx++) {
				pixelsInInt[dstOff + cx] = colorModel.getRGB(pixs[srcOff + cx]);
			}
		}
		// At the point the ColorModel for pixelsInInt is the RGB ColorModel
		this.lastColorModel = ColorModel.getRGBdefault();

	}
	private void copyPixelArea(Rectangle area, Object arraySource, Object arrayDestination, int y, int off, int scan) {
		int xend = area.x + area.width;
		int yend = area.y + area.height;
		for (int cy = area.y; cy < yend; cy++) {
			int dstOff = (cy - rect.y) * scanline  + offset - rect.x;
			int srcOff = (cy - y) * scan + off;
			System.arraycopy(arraySource, srcOff + area.x, arrayDestination, dstOff + area.x, xend - area.x);
		}
		return;

	}

}
