package java.awt.image;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Graphics2D;

import com.ibm.oti.awt.image.SubimageProducer;
import com.ibm.oti.awt.metal.image.ImagePeer;

/**
 * Personal Profile specs subimages as sharing the same data as the image they
 * was derived from. Changing one affects the other. By subclassing we don't
 * duplicate the data. We merely "remap" the origin and extent to represent the
 * subportion of the image.
 *
 * @author IBM
 * @version initial
 *
 * @see java.awt.image.BufferedImage#getSubimage(int, int, int, int)
 */
class SharedSubimage extends BufferedImage {

	int fX;
	int fY;
	int fWidth;
	int fHeight;
	BufferedImage fSourceImage;

	SharedSubimage(BufferedImage sourceImage, int x, int y, int width, int height) {
		super();

		this.fX = x;
		this.fY = y;
		this.fWidth = width;
		this.fHeight = height;
		this.fSourceImage = sourceImage;
		// These 2 data are necessary to build the assocated ImagePeer.
		this.fComponent = sourceImage.fComponent;
		this.fGraphicsConfiguration = sourceImage.fGraphicsConfiguration;

		this.fProducer = new SubimageProducer(this, sourceImage, x, y, width, height);
	}

	public ColorModel getColorModel() {
		// We don't really need to redirect this API because
		// the superclass accessed to the ColorModel through the ImagePeer
		// but, because this Class recalculate the image peer each time
		// we request it, it is better to avoid to recalculate for the ColorModel...
		return fSourceImage.getColorModel();
	}

	public int getHeight(ImageObserver observer) {
		return fHeight;
	}

	public int getRGB(int x, int y) {
		// Offset into the original source image.
		return fSourceImage.getRGB(x + fX, y + fY);
	}

	public int[] getRGB(int startX, int startY, int w, int h, int[] rgbArray, int offset, int scansize) {
		// Throw IllegalArgumentException if the arguments "leak" outside of the
		// subimage.
		if (startX < 0 || startY < 0 || w < 0 || h < 0)
			throw new IllegalArgumentException();
		if ((startX + w) > fWidth || (startY + h) > fHeight)
			throw new IllegalArgumentException();

		return fSourceImage.getRGB(startX + fX, startY + fY, w, h, rgbArray, offset, scansize);
	}

	public int getWidth(ImageObserver observer) {
		return fWidth;
	}

	public void setRGB(int x, int y, int rgb) {
		// Offset into the original source image.
		fSourceImage.setRGB(x + fX, y + fY, rgb);
	}

	public void setRGB(int startX, int startY, int w, int h, int[] rgbArray, int offset, int scansize) {
		fSourceImage.setRGB(startX + fX, startY + fY, w, h, rgbArray, offset, scansize);
	}

	ImagePeer getImagePeer() {
		// Dispose the ImagePeer to be able to recalculate it
		if (tImagePeer != null) doDispose();
		return super.getImagePeer();
	}

	/* (non-Javadoc)
	 * @see java.awt.Image#getGraphics()
	 */
	public Graphics2D createGraphics() {
		Graphics2D graphics = fSourceImage.createGraphics();
		graphics.translate(fX, fY);
		graphics.setClip(0, 0, fWidth, fHeight);
		return graphics;
	}
}
