package java.beans;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.Image;
import java.awt.Toolkit;
import java.net.URL;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * This interface represents an environment containing a collection of
 * loaded applets, called an applet context.  Applets can obtain their
 * applet context by a call to their <code>getAppletContext()</code> method,
 * then use the returned <code>AppletContext</code> object to query and
 * manipulate their applet context.
 *
 * @see		java.applet.AppletStub, java.applet.Applet
 *
 */
class BeanAppletContext implements java.applet.AppletContext {

	private Applet applet;

	/**
	 * Constructs a new instance of this class with its applet set to the
	 * argument.
	 *
	 * @param		applet Applet
	 *				the single applet to associate with this context.
	 */
	BeanAppletContext(Applet applet) {
		this.applet = applet;
	}

	/**
	 * Retrieves, by name, one of the other applets occurring in this
	 * applet context.  If more than one applet with the specified name
	 * exist, one of them is returned.  If no applet with the specified
	 * name exists, this method returns <code>null</code>.
	 *
	 * @param		name String
	 *				the name of the applet to be retrieved (this
	 *				is usually specified as a parameter in the HTML
	 *				<code>applet</code> tag)
	 *
	 * @return		Applet
	 *				A reference to an applet with the specified
	 *				name occurring in the context, or <code>null</code>
	 *				if no such applet exists.
	 */
	public Applet getApplet(String name) {
		if (name == null)
			return applet;
		else
			return null;
	}

	/**
	 * Returns an Enumeration of all the applets occurring in this
	 * applet context. The references in the Enumeration are of type
	 * <code>Applet</code>.  The Enumeration may contain <code>null</code>
	 * references.
	 *
	 * @return		Enumeration
	 *				A non-null Enumeration of all the applets occurring in
	 *				this applet context. Some elements in the Enumeration
	 *				may be <code>null</code>.
	 */
	public Enumeration getApplets() {
		class AppletEnumeration implements Enumeration {
			private Applet a;
			void setApplet(Applet a) {
				this.a = a;
			}
			public boolean hasMoreElements() {
				return a != null;
			}
			public Object nextElement() {
				if (a == null)
					throw new NoSuchElementException();
				else {
					Object answer = a;
					a = null;
					return answer;
				}
			}
		};
		AppletEnumeration answer = new AppletEnumeration();
		answer.setApplet(applet);
		return answer;
	}

	/**
	 * Retrieves an AudioClip object representing the audio clip
	 * located at the URL <code>url</code>.
	 *
	 * @param		url URL
	 *				the non-<code>null</code> absolute URL from
	 *				which the audio clip should be loaded.
	 *
	 * @return		AudioClip
	 *				an AudioClip object representing the audio
	 *				clip to be retrieved.
	 */
	public AudioClip getAudioClip(URL url) {
		// No audio support currently available. Here we
		// answer a dummy object since this makes most beans
		// happier than getting back null.
		return new BeanAudioClip();
	}

	/**
	 * Retrieves an Image object representing the image file located
	 * at the URL <code>url</code>.  This method always returns immediately,
	 * and it is only when the applet attempts to render the image that the
	 * image data is loaded.
	 *
	 * @param		url URL
	 *				the absolute URL from which the image should be loaded.
	 *
	 * @return		Image
	 *				an Image object representing the image to be retrieved.
	 */
	public Image getImage(URL url) {
		return Toolkit.getDefaultToolkit().getImage(url);
	}

	/**
	 * Requests that a specified document be displayed in place of the document
	 * currently being displayed within the applet context. Such requests may be
	 * ignored by the applet context.
	 *
	 * @param		url URL
	 *				The non-<code>null</code> URL of the document to
	 *				be displayed.
	 */
	public void showDocument(URL url) {
	}

	/**
	 * Requests that a specified document be displayed in a specified target window
	 * or HTML frame.  The <code>target</code> parameter specifies the name of the
	 * frame or window; the frame or window usually acquires this name in the HTML or
	 * script code underlying the documents being displayed in the applet context.
	 * Such requests may be ignored by the applet context.
	 *
	 * @param		url URL
	 *				The non-<code>null</code> URL of the document to be displayed.
	 * @param   	target String
	 *				The name of the frame or window in which to display the document.
	 */
	public void showDocument(URL url, String target) {
	}

	/**
	 * Requests that the string <code>msg</code> be displayed in the 'status
	 * window'.  Many environments in which an applet can run provide a 'status
	 * window', which is an area always available for the display of text messages.
	 * The environment may ignore this request.
	 *
	 * @param		msg String
	 *				the text to be displayed in the status window.
	 *				If <code>null</code> the previous message is cleared.
	 */
	public void showStatus(String p1) {
	}

}
