package java.beans;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.applet.Applet;
import java.applet.AppletContext;
import java.io.*;
import java.net.URL;

/**
 * This class is the front end to the beans support.
 */
public class Beans {

	/**
	 * This constructor does nothing useful; it should have been made private.
	 */
	public Beans() {
	}

	/**
	 * Instantiates a new instance of the bean given a class loader
	 * (which may be null to indicate that the system class loader
	 * should be used) and the name of the bean.
	 *
	 * @param		classLoader the ClassLoader to load the instance from
	 * @param		beanName    the name of the class to instantiate
	 * @return		the newly created bean
	 *
	 * @exception	IOException
	 *				If the bean was serialized and the
	 *				.ser file could not be read.
	 * @exception	ClassNotFoundException
	 *				If no class representing the bean could
	 *				be found.
	 */
	public static Object instantiate(ClassLoader classLoader, String beanName) throws IOException, ClassNotFoundException {
		Object beanInstance = null;
		Class  beanClass   = null;

		// First, look for a serialized bean with a matching name.
		String resourceName = beanName.replace('.', '/') + ".ser"; //$NON-NLS-1$
		if (classLoader == null)
			classLoader = ClassLoader.getSystemClassLoader();
		InputStream resourceStream = classLoader.getResourceAsStream(resourceName);
		if (resourceStream != null) {
			ObjectInputStream ois =
				new ObjectInputStream(resourceStream);
			try {
				beanInstance = ois.readObject();
			} catch (IOException e) {
				throw new IOException("corrupted bean"); //$NON-NLS-1$
			}
			ois.close();
			if (beanInstance != null) {
				beanClass = beanInstance.getClass();
			}
		}

		URL resource = null;
		// If we haven't got a beanInstance yet, then we try to
		// find a matching class and create a new instance of it.
		if (beanInstance == null) {
			resourceName = beanName.replace('.', '/') + ".class"; //$NON-NLS-1$
			resource = classLoader.getResource(resourceName);
			beanClass = classLoader.loadClass(beanName);
			try {
				beanInstance = beanClass.newInstance();
			} catch (Exception ex) {
				throw new ClassNotFoundException(
					"Unable to create bean instance (" + beanName +  //$NON-NLS-1$
					") reason:" + ex); //$NON-NLS-1$
			}
		}

		// Must have a bean instance by now.
		if (beanInstance == null)
			throw new ClassNotFoundException (
				"Unable to create bean instance (" + beanName + ")."); //$NON-NLS-1$ //$NON-NLS-2$

		// If it's not an applet, return it.
		if (!Applet.class.isAssignableFrom(beanClass)) {
			return beanInstance;
		}

		// It's an applet.
		Applet applet = (Applet) beanInstance;

		// No luck. Try to get it running in some reasonable way.

		AppletContext appletcontext = new BeanAppletContext(applet);

		// If we couldn't get a URL for the source, then we can't
		// compute the paths for code and doc. For now, just leave
		// them empty. If we did get a URL for the source, then
		// strip off the filename part and use it for the code base.
		// It's not clear to me yet what the document base should be,
		// so I'm just going to leave it the same as the code base.
		// This seems to match the JDK behavior.
		URL code = null;
		URL doc = null;
		if (resource != null) {
			String str = resource.toExternalForm();
			code = new URL(str.substring(0, str.length() - resourceName.length()));
			doc = code;
		}
		BeanAppletStub stub = new BeanAppletStub(appletcontext, code, doc);
		applet.setStub(stub);

		if (resourceName.endsWith(".class")) { //$NON-NLS-1$
			applet.init();
		}

		stub.activate();

		return beanInstance;
	}

	/**
	 * Answers true if we are running a design time application
	 * building tool, and false if we are in a runtime environment.
	 * For Personal Profile, always return false.
	 *
	 * @return		boolean 	the design time flag
	 */
	public static boolean isDesignTime() {
		return false;
	}

	/**
	 * Answers true if there is a graphic user-interface
	 * available for beans to display in, and false if beans
	 * should not attempt to display anything.
	 * For Personal Profile, always return true.
	 *
	 * @return		boolean		the GUI available flag
	 */
	public static boolean isGuiAvailable() {
		return true;
	}

}
