package java.beans;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.*;

/**
 * PropertyChangeSupport instances manage the details of handling bound properties.
 */
public class PropertyChangeSupport implements Serializable {

	static final long serialVersionUID = 6401253773779951803L;

	Object           source;
	transient Vector allListeners;
	Hashtable        children;
	int              propertyChangeSupportSerializedDataVersion;

	/**
	 * Constructs a new instance of this class with its source set to the argument.
	 */
	public PropertyChangeSupport(Object source) {
		super();

		if (source == null) throw new NullPointerException();

		this.source = source;

		initialize();
	}

	/**
	 * Adds the argument to the list of all listeners. Whenever
	 * any property is changed this listener will be notified.
	 *
	 * @param pCListener the listener to add
	 */
	public void addPropertyChangeListener(PropertyChangeListener pCListener) {
		synchronized (allListeners) {
			allListeners.addElement(pCListener);
		}
	}

	/**
	 * Removes the argument from the list of all listeners. That is,
	 * those which would be notified when any property is changed.
	 *
	 * @param pCListener the listener to remove
	 */
	public void removePropertyChangeListener(PropertyChangeListener pCListener) {
		synchronized (allListeners) {
			allListeners.removeElement(pCListener);
		}
	}

	/**
	 * Distributes an Object valued property change event to all
	 * appropriate listeners. Does nothing if the change event
	 * does not represent a real change.
	 *
	 * @param pName  the name property to fire an event for
	 * @param oldVal the old value of the property
	 * @param newVal the new value of the property
	 */
	public void firePropertyChange (String pName, Object oldVal, Object newVal) {
		if ((oldVal != null) && (newVal != null) && (oldVal.equals(newVal))) {
			return;
		}
		PropertyChangeEvent evt = new PropertyChangeEvent(
			source,
			pName,
			oldVal,
			newVal);
		firePropertyChange(evt);
	}

	/**
	 *
	 */
	private void initialize() {
		allListeners = new Vector(10);
		if (children == null) {
			children = new Hashtable(10);
		}
	}

	/**
	 *
	 */
	private void firePropertyChange (PropertyChangeEvent evt) {
		if (evt == null) return;
		Object oldval = evt.getOldValue();
		Object newval = evt.getNewValue();
		if (oldval != newval || oldval == null) {
			synchronized (allListeners) {
				Enumeration enumeration = allListeners.elements();
				while (enumeration.hasMoreElements())
					((PropertyChangeListener) enumeration.nextElement()).propertyChange(evt);
				if (evt.getPropertyName() != null) {
					Vector v = (Vector) children.get(evt.getPropertyName());
					if (v == null) return;
					enumeration = v.elements();
					while (enumeration.hasMoreElements())
						((PropertyChangeListener) enumeration.nextElement()).propertyChange(evt);
				}
			}
		}
	}

	/**
	 *
	 */
	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		initialize();
		Object listener = null;
		while ((listener = (Object)stream.readObject()) != null) {
			if (listener instanceof PropertyChangeListener)
				addPropertyChangeListener((PropertyChangeListener)listener);
		}
	}

	/**
	 *
	 */
	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		for (Enumeration enumeration = allListeners.elements(); enumeration.hasMoreElements();) {
			Object listener = enumeration.nextElement();
			if (listener instanceof Serializable) {
				stream.writeObject(listener);
			}
		}
		stream.writeObject(null);
	}

	public PropertyChangeListener[] getPropertyChangeListeners() {
		PropertyChangeListener[] listenerArray = new PropertyChangeListener[allListeners.size()];
		allListeners.toArray(listenerArray);
		return listenerArray;
	}
}
