package java.beans;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.*;

/**
 * VetoableChangeSupport instances manage the details of handling constrained properties.
 */
public class VetoableChangeSupport implements Serializable {

	static final long serialVersionUID = -5090210921595982017L;
	Object            source;
	transient Vector  allListeners;
	Hashtable         children;
	int               vetoableChangeSupportSerializedDataVersion;

	/**
	 * Constructs a new instance of this class with it's source set to the argument.
	 */
	public VetoableChangeSupport (Object source) {
		super();

		if (source == null) throw new NullPointerException();

		this.source = source;

		initialize();
	}

	/**
	 * Adds the argument to the list of all listeners.
	 * Whenever any property is changed this listener will be notified.
	 *
	 * @param vCListener the listener to add
	 */
	public void addVetoableChangeListener (VetoableChangeListener vCListener) {
		synchronized (allListeners) {
			allListeners.addElement(vCListener);
		}
	}

	/**
	 * Removes the argument from the list of all listeners.
	 * That is, those which would be notified when any property is changed.
	 *
	 * @param vCListener the listener to remove
	 */
	public void removeVetoableChangeListener (VetoableChangeListener vCListener) {
		synchronized (allListeners) {
			allListeners.removeElement(vCListener);
		}
	}

	/**
	 * Distributes an Object valued property change event to all appropriate listeners.
	 * Does nothing if the change event does not represent a real change.
	 *
	 * @param pName  the name property to fire an event for
	 * @param oldVal the old value of the property
	 * @param newVal the new value of the property
	 */
	public void fireVetoableChange (String pName, Object oldVal, Object newVal) throws PropertyVetoException {
		if ((oldVal != null) && (newVal != null) && (oldVal.equals(newVal))) {
			return;
		}

		try {
			PropertyChangeEvent evt = new PropertyChangeEvent(source,pName,oldVal,newVal);
			_fireVetoableChange(evt, false);
		} catch (PropertyVetoException ex) {
			PropertyChangeEvent evt = new PropertyChangeEvent(source,pName,newVal,oldVal);
			_fireVetoableChange(evt, true);
			throw ex;
		}

	}

	/**
	 *
	 */
	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		initialize();
		Object listener = null;
		while ((listener = (Object)stream.readObject()) != null) {
			if (listener instanceof VetoableChangeListener)
				addVetoableChangeListener((VetoableChangeListener)listener);
		}
//		for (Enumeration enum = children.elements(); enum.hasMoreElements(); ) {
//			Vector element = (Vector) enum.nextElement();
//			for (Enumeration enum2 = element.elements(); enum2.hasMoreElements(); ) {
//				addVetoableChangeListener((VetoableChangeListener) enum2.nextElement());
//			}
//
//		}
	}

	/**
	 *
	 */
	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.defaultWriteObject();
		for (Enumeration enumeration = allListeners.elements(); enumeration.hasMoreElements();) {
			Object listener = enumeration.nextElement();
			if (listener instanceof Serializable) {
				stream.writeObject(listener);
			}
		}
		stream.writeObject(null);
	}

	/**
	 *
	 */
	private void initialize() {
		allListeners = new Vector(10);
		if (children == null) {
			children = new Hashtable(10);
		}
	}

	/**
	 *
	 */
	private void _fireVetoableChange(PropertyChangeEvent evt, boolean quiet) throws PropertyVetoException {
		if (evt == null) return;
		Object oldval = evt.getOldValue();
		Object newval = evt.getNewValue();
		if (oldval != newval || oldval == null) {
			synchronized (allListeners) {
				Enumeration enumeration = allListeners.elements();
				while (enumeration.hasMoreElements())
					_vetoCallback(evt, enumeration, quiet);
				if (evt.getPropertyName() != null) {
					Vector v = (Vector) children.get(evt.getPropertyName());
					if (v == null) return;
					enumeration = v.elements();
					while (enumeration.hasMoreElements())
						_vetoCallback(evt, enumeration, quiet);
				}
			}
		}
		return;
	}

	/**
	 *
	 */
	private void _vetoCallback(PropertyChangeEvent evt, Enumeration enumeration, boolean quiet) throws PropertyVetoException {
		if (quiet) {
			try {
				_primVetoCallback(evt, enumeration);
			} catch (PropertyVetoException ex) {
			}
		} else {
			_primVetoCallback(evt, enumeration);
		}
	}

	/**
	 *
	 */
	private void _primVetoCallback(PropertyChangeEvent evt, Enumeration enumeration) throws PropertyVetoException {
		((VetoableChangeListener) enumeration.nextElement()).vetoableChange(evt);
	}

	public VetoableChangeListener[] getVetoableChangeListeners() {
		VetoableChangeListener[] listenerArray = new VetoableChangeListener[allListeners.size()];
		allListeners.toArray(listenerArray);
		return listenerArray;
	}
}
