package java.io;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * FileInputStream is a class for reading bytes from a file. This class may also be
 * used with other InputStreams, ie: BufferedInputStream, to read data from a
 * file with buffering.
 *
 * @author		OTI
 * @version		initial
 *
 * @see 		FileOutputStream
 */
public class FileInputStream extends InputStream
{
	/**
	 * The FileDescriptor representing this FileInputStream.
	 */
	FileDescriptor fd;

	// Fill in the JNI id caches
	private static native void oneTimeInitialization();
	static {
		oneTimeInitialization();
	}

/**
 * Constructs a new FileInputStream on the File <code>file</code>.  If the
 * file does not exist, the <code>FileNotFoundException</code> is thrown.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file		the File on which to stream reads.
 *
 * @throws 		FileNotFoundException If the <code>file</code> is not found.
 *
 * @see			java.lang.SecurityManager#checkRead(FileDescriptor)
 * @see			java.lang.SecurityManager#checkRead(String)
 * @see			java.lang.SecurityManager#checkRead(String, Object)
 */
public FileInputStream(File file) throws FileNotFoundException {
	super();
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkRead(file.getPath());
	fd = new FileDescriptor();
	if (openImpl(file.properPath(true)) != 0)
		throw new FileNotFoundException(file.getPath());
}

/**
 * Constructs a new FileInputStream on the FileDescriptor <code>fd</code>.  The
 * file must already be open, therefore no <code>FileNotFoundException</code> will be thrown.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		fd			the FileDescriptor on which to stream reads.
 *
 * @see			java.lang.SecurityManager#checkRead(FileDescriptor)
 * @see			java.lang.SecurityManager#checkRead(String)
 * @see			java.lang.SecurityManager#checkRead(String, Object)
 */
public FileInputStream(FileDescriptor fd) {
	super();
	if (fd != null) {
		SecurityManager security = System.getSecurityManager();
		if (security != null)
			security.checkRead(fd);
		this.fd = fd;
	} else throw new NullPointerException();
}

/**
 * Constructs a new FileInputStream on the file named <code>fileName</code>.  If the
 * file does not exist, the <code>FileNotFoundException</code> is thrown. The <code>fileName</code>
 * may be absolute or relative to the System property <code>"user.dir"</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		fileName	the file on which to stream reads.
 *
 * @throws 		FileNotFoundException If the <code>fileName</code> is not found.
 */
public FileInputStream(String fileName) throws FileNotFoundException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkRead(fileName);
	fd = new FileDescriptor();
	if (openImpl(new File(fileName).properPath(true)) != 0)
		throw new FileNotFoundException(fileName);
}

/**
 * Answers a int representing then number of bytes that are available
 * before this InputStream will block.  This method always returns the
 * size of the file minus the current position.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		the number of bytes available before blocking.
 *
 * @throws 		IOException	If an error occurs in this stream.
 */
public native int available() throws IOException;

/**
 * Close the FileInputStream.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 		IOException	If an error occurs attempting to close this FileInputStream.
 */
public void close() throws IOException {
	closeImpl();
}

private native void closeImpl() throws IOException;

/**
 * This method ensures that all resources for this file are released
 * when it is about to be garbage collected.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 		IOException	If an error occurs attempting to finalize this FileInputStream.
 */
protected void finalize() throws IOException {
	if (this.fd != null) close();
}

/**
 * Answers the FileDescriptor representing the operating system resource for
 * this FileInputStream.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the FileDescriptor for this FileInputStream.
 *
 * @throws 		IOException	If an error occurs attempting to get the FileDescriptor of this FileInputStream.
 */
public final FileDescriptor getFD() throws IOException {
	if (fd != null) return fd;
	throw new IOException();
}

private native int openImpl(byte [] fileName);

/**
 * Reads a single byte from this FileInputStream and returns the result as
 * an int.  The low-order byte is returned or -1 of the end of stream was
 * encountered.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		the byte read or -1 if end of stream.
 *
 * @throws 		IOException If the stream is already closed or another IOException occurs.
 */
public int read() throws IOException {
	if (fd != null) return readByteImpl(getFD().descriptor);
	throw new IOException();
}

private native int readByteImpl(long descriptor ) throws IOException;

/**
 * Reads bytes from the FileInputStream and stores them in byte array <code>buffer</code>. Answer the
 * number of bytes actually read or -1 if no bytes were read and end of stream was
 * encountered.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the byte array in which to store the read bytes.
 * @return 		the number of bytes actually read or -1 if end of stream.
 *
 * @throws 		IOException If the stream is already closed or another IOException occurs.
 */
public int read(byte[] buffer) throws IOException {
	return read(buffer, 0, buffer.length);
}

/**
 * Reads at most <code>count</code> bytes from the FileInputStream and stores them in byte
 * array <code>buffer</code> starting at <code>offset</code>. Answer the number of
 * bytes actually read or -1 if no bytes were read and end of stream was encountered.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the byte array in which to store the read bytes.
 * @param		offset		the offset in <code>buffer</code> to store the read bytes.
 * @param		count		the maximum number of bytes to store in <code>buffer</code>.
 * @return 		the number of bytes actually read or -1 if end of stream.
 *
 * @throws 		IOException If the stream is already closed or another IOException occurs.
 */
public int read(byte[] buffer, int offset, int count) throws IOException {
	if (fd != null) return readImpl(buffer, offset, count, getFD().descriptor);
	throw new IOException();
}

private native int readImpl(byte[] buffer, int offset, int count, long descriptor) throws IOException;

/**
 * Skips <code>count</code> number of bytes in this FileInputStream.  Subsequent
 * <code>read()</code>'s will not return these bytes unless <code>reset()</code>
 * is used.  This method may perform multiple reads to read <code>count</code>
 * bytes.  This default implementation reads <code>count</code> bytes into a temporary
 * buffer.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		count		the number of bytes to skip.
 * @return		the number of bytes actually skipped.
 *
 * @throws 		IOException If the stream is already closed or another IOException occurs.
 */
public native long skip(long count) throws IOException;
}
