package java.io;

import java.security.Permission;
import java.security.AccessController;
import java.security.PrivilegedAction;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * The class FilePermission is responsible for granting access to files or directories.  The FilePermission
 * is made up of a pathname and a set of actions which are valid for the pathname.<P>
 * The <code>File.separatorChar</code> must be used in all pathnames when constructing a FilePermission.
 * The following descriptions will assume the char is </code>/</code>. A pathname which
 * ends in "/*", implies all the files and directories contained in that directory. If the
 * pathname ends in "/-", it indicates all the files and directories in that directory
 * <b>recursively</b>.
 *
 * @author		OTI
 * @version		initial
 */
public final class FilePermission extends Permission implements Serializable {
	private static final long serialVersionUID = 7930732926638008763L;

	// canonical path of this permission
	private transient String canonPath;

	// list of actions permitted for socket permission in order
	private static final String[] actionList = {"read", "write", "execute", "delete"};

	// "canonicalized" action list
	private String actions;

	// the numeric representation of this action list
	// for implies() to check if one action list is the subset of another.
	transient int mask = -1;

	// global include all permission?
	private transient boolean includeAll = false;
	private transient boolean allDir = false;
	private transient boolean allSubdir = false;

/**
 * Constructs a new FilePermission with the path and actions specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		path		the path to apply the actions to.
 * @param		actions		the actions for the <code>path<code>. May be any
 *							combination of read, write, execute, or delete.
 */
public FilePermission(String path, String actions) {
	super(path);
	init(path, actions);
}

private void init(final String path, String actions) {
	if (actions != null && actions != "") {
		if (path != null) {
			if (path.equals("<<ALL FILES>>")) {
				includeAll = true;
			} else {
				canonPath = (String)AccessController.doPrivileged(new PrivilegedAction() {
					public Object run() {
						try {
							return new File(path).getCanonicalPath();
						} catch (IOException e) {
							return path;
						}
					}});
				if (path.equals("*") || path.endsWith(File.separator + "*")) allDir = true;
				if (path.equals("-") || path.endsWith(File.separator + "-")) allSubdir = true;
			}
			this.actions = toCanonicalActionString(actions);
		// K006e = path is null
		} else throw new NullPointerException(com.ibm.oti.util.Msg.getString("K006e"));
	// K006d = actions invalid
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K006d"));
}

/**
 * Answer the string representing this permissions actions. It must
 * be of the form "read,write,execute,delete", all lower case and
 * in the correct order if there is more than one action.
 *
 * @param		action		the action name
 * @return		the string representing this permission's actions
 */
private String toCanonicalActionString(String action) {
	actions = action.trim().toLowerCase();

	// get the numerical representation of the action list
	mask = getMask(actions);

	// convert the mask to a canonical action list.
	int len = actionList.length;
	// the test mask - shift the 1 to the leftmost position of the actionList
	int highestBitMask = 1 << (len - 1);

	// if a bit of mask is set, append the corresponding action to result
	StringBuffer result = new StringBuffer();
	boolean addedItem = false;
	for (int i=0; i<len; i++) {
		if ((highestBitMask & mask) != 0) {
			if (addedItem) result.append(",");
			result.append(actionList[i]);
			addedItem = true;
		}
		highestBitMask = highestBitMask>>1;
	}
	return result.toString();
}

/**
 * Answers the numerical representation of the argument.
 *
 * @param 		actions		the action names
 * @return		the action mask
 */
private int getMask(String actions) {
	int actionInt = 0, head = 0, tail = 0;
	do {
		tail = actions.indexOf(",", head);
		String action = tail > 0 ? actions.substring(head, tail).trim() : actions.substring(head).trim();
		if (action.equals("read"))
			actionInt |= 8;
		else if (action.equals("write"))
			actionInt |= 4;
		else if (action.equals("execute"))
			actionInt |= 2;
		else if (action.equals("delete"))
			actionInt |= 1;
		// K006f = invalid permission: {0}
		else throw new java.lang.IllegalArgumentException(com.ibm.oti.util.Msg.getString("K006f", action));
		head = tail + 1;
	} while (tail > 0);
	return actionInt;
}

/**
 * Answers the actions associated with the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the actions associated with the receiver.
 */
public String getActions() {
	return actions;
}

/**
 * Check to see if this permission is equal to another.  The
 * two are equal if <code>obj</code> is a FilePermission,
 * they have the same path, and they have the same actions.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		obj			the object to check equality with.
 * @return		<code>true</code> if the two are equal, <code>false</code> otherwise.
 */
public boolean equals(Object obj) {
	if (obj instanceof FilePermission) {
		FilePermission fp = (FilePermission) obj;
		if (fp.actions != actions)
			if (fp.actions == null || !fp.actions.equals(actions)) return false;

		/* Matching actions and both are <<ALL FILES>> ? */
		if (fp.includeAll || includeAll)
			return fp.includeAll == includeAll;
		return fp.canonPath.equals(canonPath);
	}
	return false;
}

/**
 * Indicates whether the argument permission is implied
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		p 			java.security.Permission the permission to check.
 * @return		<code>true</code> if the argument permission is implied by the receiver, and <code>false</code> if it is not.
 */
public boolean implies(Permission p) {
	int match = impliesMask(p);
	return match != 0 && match == ((FilePermission)p).mask;
}

/**
 * Answers an int describing what masks are implied by a
 * specific permission.
 *
 * @param 		p		the permission
 * @return		the mask applied to the given permission
 */
int impliesMask(Permission p) {
	if (!(p instanceof FilePermission)) return 0;
	FilePermission fp = (FilePermission) p;
	int matchedMask = mask & fp.mask;
	// Can't match any bits?
	if (matchedMask == 0) return 0;

	// Is this permission <<ALL FILES>>
	if (includeAll) return matchedMask;

	// We can't imply all files
	if (fp.includeAll) return 0;

	// Scan the length of p checking all match possibilities
	// \- implies everything except \
	int thisLength = canonPath.length();
	if (allSubdir && thisLength == 2 && !fp.canonPath.equals(File.separator))
		return matchedMask;
	// need /- to imply /-
	if (fp.allSubdir && !allSubdir) return 0;
	// need /- or /* to imply /*
	if (fp.allDir && !allSubdir && !allDir) return 0;

	boolean includeDir = false;
	int pLength = fp.canonPath.length();
	// do not compare the * or -
	if (allDir || allSubdir) thisLength--;
	if (fp.allDir || fp.allSubdir) pLength--;
	for (int i = 0; i < pLength; i++) {
		char pChar = fp.canonPath.charAt(i);
		// Is p longer than this permissions canonLength?
		if (i >= thisLength) {
			if (i == thisLength) {
				// Is this permission include all? (must have matched up until this point).
				if (allSubdir) return matchedMask;
				// Is this permission include a dir?  Continue the check afterwards.
				if (allDir) includeDir = true;
			}
			//If not includeDir then is has to be a mismatch.
			if (!includeDir) return 0;
			/**
			 * If we have * for this and find a separator it is invalid.
			 * IE: this is '/a/*' and p is '/a/b/c' we should fail
			 * on the separator after the b. Except for root, canonical paths
			 * do not end in a separator.
			 */
			if (pChar == File.separatorChar) return 0;
		} else {
			// Are the characters matched?
			if (canonPath.charAt(i) != pChar) return 0;
		}
	}
	// Must have matched upto this point or it's a valid file in an include all directory
	if (pLength == thisLength) {
		if (allSubdir) {
			// /- implies /- or /*
			return fp.allSubdir || fp.allDir ? matchedMask : 0;
		} else {
			return allDir == fp.allDir ? matchedMask : 0;
		}
	}
	return includeDir ? matchedMask : 0;
}

/**
 * Answers a new PermissionCollection in which to place FilePermission Objects.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		A new PermissionCollection suitable for storing FilePermission objects.
 */
public java.security.PermissionCollection newPermissionCollection() {
	return new FilePermissionCollection();
}

/**
 * Answers an int representing the hash code value for this FilePermission.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int			the hash code value for this FilePermission.
 */
public int hashCode() {
	return (canonPath == null ? getName().hashCode() : canonPath.hashCode()) + mask;
}

private void writeObject(ObjectOutputStream stream) throws IOException {
	stream.defaultWriteObject();
}

private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
	stream.defaultReadObject();
	init(getName(), actions);
}
}
