package java.io;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 * FilteredOutputStream is a class which takes an output stream and <em>filters</em> the
 * output in some way.  The filtered view may be a buffered output or one which compresses
 * data before actually writing the bytes.  FilterOutputStreams are meant for byte streams.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			FilterInputStream
 */
public class FilterOutputStream extends OutputStream {

	/**
	 * The target OutputStream for this filter.
	 */
	protected OutputStream out;

/**
 * Constructs a new FilterOutputStream on the OutputStream <code>out</code>.  All
 * writes are now filtered through this stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		out			the target OutputStream to filter writes on.
 */
public FilterOutputStream(OutputStream out) {
	this.out = out;
}

/**
 * Close this FilterOutputStream. This implementation closes the target stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 		IOException	If an error occurs attempting to close this stream.
 */
public void close() throws IOException {
	try {
		flush();
	} catch (IOException e) {};
	/* Make sure we clean up this stream if exception fires */
	out.close();
}

/**
 * Flush this FilterOutputStream to ensure all pending data is sent out to the target
 * OutputStream.  This implementation flushes the target OutputStream.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 		IOException	If an error occurs attempting to flush this FilterOutputStream.
 */
public void flush() throws IOException {
	out.flush();
}

/**
 * Writes the entire contents of the byte array <code>buffer</code>
 * to this FilterOutputStream.  This implementation writes the <code>buffer</code>
 * to the target stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the buffer to be written
 *
 * @throws 		IOException	If an error occurs attempting to write to this FilterOutputStream.
 */
public void write(byte buffer[]) throws IOException {
	write(buffer, 0, buffer.length);
}

/**
 * Writes <code>count</code> <code>bytes</code> from the byte array
 * <code>buffer</code> starting at <code>offset</code> to this
 * FilterOutputStream.  This implementation writes the <code>buffer</code>
 * to the target OutputStream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the buffer to be written
 * @param		offset		offset in buffer to get bytes
 * @param		count		number of bytes in buffer to write
 *
 * @throws 		IOException					If an error occurs attempting to write to this FilterOutputStream.
 * @throws		IndexOutOfBoundsException 	If offset or count are outside of bounds.
 */
public void write(byte buffer[], int offset, int count) throws IOException {
	// avoid int overflow, check null buffer
	if (offset <= buffer.length && 0 <= offset && 0 <= count && count <= buffer.length - offset) {
		for (int i=0; i < count; i++)
			// Call write() instead of out.write() since subclasses could override the write() method.
			write(buffer[offset + i]);
	// K002f = Arguments out of bounds
	} else throw new ArrayIndexOutOfBoundsException(com.ibm.oti.util.Msg.getString("K002f"));
}

/**
 * Writes the specified byte <code>oneByte</code> to this FilterOutputStream.  Only
 * the low order byte of <code>oneByte</code> is written.  This implementation writes
 * the byte to the target OutputStream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		oneByte		the byte to be written
 *
 * @throws 		IOException	If an error occurs attempting to write to this FilterOutputStream.
 */
public void write(int oneByte) throws IOException {
	out.write(oneByte);
}
}
