package java.io;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 * PipedWriter is a class which places information on a communications pipe.
 * When two threads want to pass data back and forth, one creates a piped writer
 * and the other creates a piped reader.
 *
 * @author		OTI
 * @version		initial
 *
 * @see 		PipedReader
 */
public class PipedWriter extends Writer {
	/**
	 * The destination PipedReader
	 */
	private PipedReader dest;
	private boolean closed = false;

/**
 * Constructs a new unconnected PipedWriter. The resulting Stream must be
 * connected to a PipedReader before data may be written to it.
 *
 * @author		OTI
 * @version		initial
 */
public PipedWriter() {
	super();
}

/**
 * Constructs a new PipedWriter connected to the PipedReader <code>dest</code>.
 * Any data written to this Writer can be read from the <code>dest</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		dest		the PipedReader to connect to.
 *
 * @throws 	java.io.IOException	if <code>dest</code> is already connected.
 */
public PipedWriter(PipedReader dest) throws IOException {
	super(dest);
	connect(dest);
}

/**
 * Close this PipedWriter.  Any data buffered in the corresponding PipedReader
 * can be read, then -1 will be returned to the reader.  If this Writer is not
 * connected, this method does nothing.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 	java.io.IOException	If an error occurs attempting to close this PipedWriter.
 */
public void close() throws IOException {
	synchronized (lock) {
		/* Is the pipe connected? */
		if (dest != null) {
			dest.done();
			dest = null;
		}
		closed = true;
	}
}

/**
 * Connects this PipedWriter to a PipedReader.  Any data written to this
 * Writer becomes readable in the Reader.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		dest		the destination PipedReader.
 *
 * @throws 	java.io.IOException	If this Writer or the dest is already connected.
 */
public void connect(PipedReader dest) throws IOException {
	synchronized (lock) {
		if (this.dest == null) {
			if (!closed) {
				dest.establishConnection(this);
				this.dest = dest;
			// K0078 = Pipe is closed
			} else throw new IOException(com.ibm.oti.util.Msg.getString("K0078"));
		// K0079 = Already connected
		} else throw new IOException(com.ibm.oti.util.Msg.getString("K0079"));
	}
}

/**
 * Notifies the readers on the PipedReader that characters can be read. This
 * method does nothing if this Writer is not connected.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 	java.io.IOException	If an IO error occurs during the flush.
 */
public void flush() throws IOException {
	if (dest != null) dest.flush();
}

/**
 * Writes <code>count</code> <code>chars</code> from the char array
 * <code>buffer</code> starting at offset <code>index</code> to this
 * PipedWriter.  The written data can now be read from the destination
 * PipedReader. Separate threads should be used for the reader of the
 * PipedReader and the PipedWriter. There may be undesirable results
 * if more than one Thread interacts a input or output pipe.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the buffer to be written
 * @param		offset		offset in buffer to get chars
 * @param		count		number of chars in buffer to write
 *
 * @throws 	java.io.IOException					If the receiving thread was terminated without closing the pipe.
 *													This case is not currently handled correctly.
 * @throws 	java.io.InterruptedIOException		If the pipe is full and the current thread is interrupted waiting
 *													for space to write data.
 *													This case is not currently handled correctly.
 * @throws 	java.lang.NullPointerException		If the receiver has not been connected yet.
 * @throws 	java.lang.IllegalArgumentException	If any of the arguments are out of bounds.
 */
public void write(char buffer[], int offset, int count) throws IOException {
	// avoid int overflow
	if (0 <= offset && offset <= buffer.length && 0 <= count && count <= buffer.length - offset) {
		synchronized(lock) {
			if (!closed) {
				if (dest != null)
					dest.receive(buffer, offset, count);
				// K007b = Pipe Not Connected
				else throw new IOException(com.ibm.oti.util.Msg.getString("K007b"));
			// K0078 = Pipe is closed
			} else throw new IOException(com.ibm.oti.util.Msg.getString("K0078"));
		}
	} else throw new ArrayIndexOutOfBoundsException();
}

/**
 * Writes the character <code>c</code> to this PipedWriter.  The written data can now be read
 * from the destination PipedReader. Separate threads should be used for the reader of the
 * PipedReader and the PipedWriter. There may be undesirable results
 * if more than one Thread interacts a input or output pipe.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		c			the character to be written
 *
 * @throws 	java.io.IOException					If the receiving thread was terminated without closing the pipe.
 *													This case is not currently handled correctly.
 * @throws 	java.io.InterruptedIOException		If the pipe is full and the current thread is interrupted waiting
 *													for space to write data.
 *													This case is not currently handled correctly.
 * @throws 	java.lang.NullPointerException		If the receiver has not been connected yet.
 */
public void write(int c) throws IOException {
	synchronized(lock) {
			if (!closed) {
				if (dest != null)
					dest.receive((char)c);
				// K007b = Pipe Not Connected
				else throw new IOException(com.ibm.oti.util.Msg.getString("K007b"));
			// K0078 = Pipe is closed
			} else throw new IOException(com.ibm.oti.util.Msg.getString("K0078"));
	}
}
}
