package java.io;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * PushbackInputStream is a filter class which allows bytes read to be
 * pushed back into the stream so that they can be reread.  Parsers may
 * find this useful.  There is a progammable limit to the number of bytes
 * which may be pushed back.  If the buffer of pushed back bytes is empty,
 * bytes are read from the source input stream.
 *
 * @author		OTI
 * @version		initial
 */
public class PushbackInputStream extends FilterInputStream {
	/**
	 * The <code>byte</code> array containing the bytes to read.
	 */
	protected byte[] buf;
	/**
	 * The current position within the byte array <code>buf</code>. A value
	 * equal to buf.length indicates no bytes available.  A value of
	 * 0 indicates the buffer is full.
	 */
	protected int pos;

/**
 * Constructs a new PushbackInputStream on the InputStream <code>in</code>.  The size
 * of the pushback buffer is set to the default, or 1 byte.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		in			the InputStream to allow pushback operations on.
 *
 */
public PushbackInputStream(InputStream in) {
	super(in);
	buf = new byte[1];
	pos = 1;
}

/**
 * Constructs a new PushbackInputStream on the InputStream <code>in</code>.  The size
 * of the pushback buffer is set to <code>size</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		in			the InputStream to allow pushback operations on.
 * @param		size		the size of the pushback buffer (<code>size>=0</code>).
 */
public PushbackInputStream(InputStream in, int size) {
	super(in);
	if (size > 0) {
		buf = new byte[size];
		pos = size;
	// K0058 = size must be > 0
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0058"));
}

/**
 * Answers a int representing then number of bytes that are available
 * before this PushbackInputStream will block.  This method returns the
 * number of bytes available in the pushback buffer plus those available
 * in the target stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		int			the number of bytes available before blocking.
 *
 * @throws 	java.io.IOException	If an error occurs in this stream.
 */
public int available() throws IOException {
	if (buf != null) return buf.length - pos + in.available();
	throw new IOException();
}

/**
 * Close this PushbackInputStream. This implementation closes the target stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @throws 		IOException		If an error occurs attempting to close this stream.
 */
public void close() throws IOException {
	if (in != null) {
		in.close();
		in = null;
		buf = null;
	}
}

/**
 * Answers a boolean indicating whether or not this PushbackInputStream supports mark() and
 * reset().  This implementation always answers false since PushbackInputStreams do not support
 * mark/reset.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		boolean		indicates whether or not mark() and reset() are supported.
 */
public boolean markSupported() {
	return false;
}

/**
 * Reads a single byte from this PushbackInputStream and returns the result as
 * an int.  The low-order byte is returned or -1 of the end of stream was
 * encountered.  If the pushback buffer does not contain any available bytes
 * then a byte from the target input stream is returned.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		int			The byte read or -1 if end of stream.
 *
 * @throws 		IOException 	If an IOException occurs.
 */
public int read() throws IOException {
	if (buf != null) {
		// Is there a pushback byte available?
		if (pos < buf.length) return (buf[pos++] & 0xFF);
		// Assume read() in the InputStream will return low-order byte or -1 if end of stream.
		return in.read();
	} else throw new IOException();
}

/**
 * Reads at most <code>length</code> bytes from this PushbackInputStream and stores them in byte
 * array <code>buffer</code> starting at <code>offset</code>. Answer the number of bytes
 * actually read or -1 if no bytes were read and end of stream was encountered.  This implementation
 * reads bytes from the pushback buffer first, then the target stream if more bytes are required to
 * satisfy <code>count</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the byte array in which to store the read bytes.
 * @param		offset		the offset in <code>buffer</code> to store the read bytes.
 * @param		length		the maximum number of bytes to store in <code>buffer</code>.
 * @return 		the number of bytes actually read or -1 if end of stream.
 *
 * @throws 		IOException 	If an IOException occurs.
 */
public int read(byte[] buffer, int offset, int length) throws IOException {
	if (buf != null && buffer != null) {
		// avoid int overflow
		if (0 <= offset && offset <= buffer.length && 0 <= length && length <= buffer.length - offset) {
			int copiedBytes = 0, copyLength = 0, newOffset = offset;
			// Are there pushback bytes available?
			if (pos < buf.length) {
				copyLength = (buf.length - pos >= length) ? length : buf.length - pos;
				System.arraycopy(buf, pos, buffer, newOffset, copyLength);
				newOffset += copyLength;
				copiedBytes += copyLength;
				// Use up the bytes in the local buffer
				pos += copyLength;
			}
			// Have we copied enough?
			if (copyLength == length) return length;
			int inCopied = in.read(buffer, newOffset, length - copiedBytes);
			if (inCopied > 0) return inCopied + copiedBytes;
			if (copiedBytes == 0) return inCopied;
			return copiedBytes;
		} else throw new ArrayIndexOutOfBoundsException();
	} else if (buf == null) throw new IOException();
	throw new NullPointerException();
}

/**
 * Skips <code>count</code> number of bytes in this PushbackInputStream.  Subsequent
 * <code>read()</code>'s will not return these bytes unless <code>reset()</code>
 * is used. This implementation skips <code>count</code> number of bytes in the
 * buffer and/or the target stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		count		the number of bytes to skip.
 * @return		the number of bytes actually skipped.
 *
 * @throws 		IOException 	If the stream is already closed or another IOException occurs.
 */
public long skip(long count) throws IOException {
	if (in != null) {
		if (count <= 0) return 0;
		int numSkipped = 0;
		if (pos < buf.length) {
			numSkipped += (count < buf.length - pos) ? count : buf.length - pos;
			pos += numSkipped;
		}
		if (numSkipped < count)
			numSkipped += in.skip(count - numSkipped);
		return numSkipped;
	// K0059 = Stream is closed
	} else throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
}

/**
 * Push back all the bytes in <code>buffer</code>. The bytes are pushed so that
 * they would be read back buffer[0], buffer[1], etc.  If the push back buffer cannot
 * handle the entire contents of <code>buffer</code>, an IOException will be thrown.
 * Some of the buffer may already be in the buffer after the exception is thrown.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the byte array containing bytes to push back into the stream.
 *
 * @throws 		IOException 	If the pushback buffer becomes, or is, full.
 */
public void unread(byte[] buffer) throws IOException {
	unread(buffer, 0, buffer.length);
}

/**
 * Push back <code>length</code> number of bytes in <code>buffer</code> starting at <code>offset</code>.
 * The bytes are pushed so that they would be read back buffer[offset], buffer[offset+1], etc.
 * If the push back buffer cannot handle the bytes copied from <code>buffer</code>, an
 * IOException will be thrown. Some of the bytes may already be in the buffer after the exception
 * is thrown.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer		the byte array containing bytes to push back into the stream.
 * @param		offset		the location to start taking bytes to push back.
 * @param		length		the number of bytes to push back.
 *
 * @throws 		IOException 	If the pushback buffer becomes, or is, full.
 */
public void unread(byte[] buffer, int offset, int length) throws IOException {
	if (length > pos)
		// K007e = Pushback buffer full
		throw new IOException(com.ibm.oti.util.Msg.getString("K007e"));
	// avoid int overflow
	if (0 <= offset && offset <= buffer.length && 0 <= length && length <= buffer.length - offset) {
		for (int i = offset + length - 1; i >= offset; i--)
			unread(buffer[i]);
	} else throw new ArrayIndexOutOfBoundsException();
}

/**
 * Push back one <code>byte</code>.  Takes the byte <code>oneByte</code> and puts in
 * in the local buffer of bytes to read back before accessing the target input stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		oneByte		the byte to push back into the stream.
 *
 * @throws 		IOException 	If the pushback buffer is already full.
 */
public void unread(int oneByte) throws IOException {
	if (buf != null) {
		if (pos != 0)
			buf[--pos] = (byte)oneByte;
		// K007e = Pushback buffer full
		else throw new IOException(com.ibm.oti.util.Msg.getString("K007e"));
	} else throw new IOException();
}
}
