package java.lang;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.net.URL;

/**
 * An instance of class Package contains information about
 * a Java package. This includes implementation and
 * specification versions. Typically this information is
 * retrieved from the manifest.
 * <p>
 * Packages are managed by class loaders. All classes loaded
 * by the same loader from the same package share a Package
 * instance.
 *
 * @version		initial
 *
 * @see			java.lang.ClassLoader
 */
public class Package {
	private String name;
	private String specificationTitle;
	private String specificationVendor;
	private String specificationVersion;
	private String implementationTitle;
	private String implementationVendor;
	private String implementationVersion;
	private URL sealBase;

/**
 * Create a new package with the specified information.
 *
 * @param		name					the name of this package
 * @param		specificationTitle		the title of the specification
 * @param		specificationVendor		the provider of the specification
 * @param		specificationVersion	the version of the specification
 * @param		implementationTitle		the title of the implementation
 * @param		implementationVendor	the provider of the implementation
 * @param		implementationVersion	the version of the implementation
 * @param		sealBase				the URL that the package is sealed with respect to, or null is unsealed
 *
 */
Package(
		String name,
		String specificationTitle, String specificationVersion, String specificationVendor,
		String implementationTitle, String implementationVersion, String implementationVendor,
		URL sealBase)
{
	this.name = name;
	this.specificationTitle = specificationTitle;
	this.specificationVersion = specificationVersion;
	this.specificationVendor = specificationVendor;
	this.implementationTitle = implementationTitle;
	this.implementationVersion = implementationVersion;
	this.implementationVendor = implementationVendor;
	this.sealBase = sealBase;
}
/**
 * Return the title of the implementation of this
 * package, or null if this is unknown. The format
 * of this string is unspecified.
 *
 * @return 		The implementation title, or null
 */
public String getImplementationTitle() {
	return implementationTitle;
}
/**
 * Return the name of the vendor or organization that
 * provided this implementation of the package, or null
 * if this is unknown. The format of this string is
 * unspecified.
 *
 * @return 		The implementation vendor name, or null
 */
public String getImplementationVendor() {
	return implementationVendor;
}
/**
 * Return the version of the implementation of this
 * package, or null if this is unknown. The format
 * of this string is unspecified.
 *
 * @return 		The implementation version, or null
 */
public String getImplementationVersion() {
	return implementationVersion;
}
/**
 * Return the name of this package in the standard
 * dot notation; for example: "java.lang".
 *
 * @return 		The name of this package
 */
public String getName() {
	return name;
}
/**
 * Attempt to locate the requested package in the caller's
 * class loader. If no package information can be located,
 * null is returned.
 *
 * @param		packageName		The name of the package to find
 * @return		The package requested, or null
 *
 * @see			ClassLoader#getPackage
 */
public static Package getPackage(String packageName) {
	return ClassLoader.getStackClassLoader(1).getPackage(packageName);
}
/**
 * Return all the packages known to the caller's class loader.
 *
 * @return		All the packages known to the caller's classloader
 *
 * @see			ClassLoader#getPackages
 */
public static Package[] getPackages() {
	return ClassLoader.getStackClassLoader(1).getPackages();
}
/**
 * Return the title of the specification this
 * package implements, or null if this is unknown.
 *
 * @return 		The specification title, or null
 */
public String getSpecificationTitle() {
	return specificationTitle;
}
/**
 * Return the name of the vendor or organization that
 * owns and maintains the specification this package
 * implements, or null if this is unknown.
 *
 * @return 		The specification vendor name, or null
 */
public String getSpecificationVendor() {
	return specificationVendor;
}
/**
 * Return the version of the specification this
 * package implements, or null if this is unknown.
 * The version string is a sequence of non-negative
 * integers separated by dots; for example: "1.2.3".
 *
 * @return 		The specification version string, or null
 */
public String getSpecificationVersion() {
	return specificationVersion;
}
/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>equals</code> must answer the same value for this
 * method.
 *
 * @return		the receiver's hash
 */
public int hashCode() {
	return name.hashCode();
}
/**
 * Return true if this package's specification version
 * is compatible with the specified version string.
 * Version strings are compared by comparing each dot
 * separated part of the version as an integer.
 *
 * @param		version		The  version string to compare against
 * @return		true if the package versions are compatible, false otherwise
 *
 * @throws		NumberFormatException		if the package's version string or the one provided is not in the correct format
 */
public boolean isCompatibleWith(String version) throws NumberFormatException {
	if (specificationVersion == null || version == null ||
		specificationVersion.length() == 0 || version.length() == 0)
			throw new NumberFormatException();
	if (specificationVersion.equals(version)) return true;

	// Compare each part
	int length1 = specificationVersion.length();
	int length2 = version.length();
	int start1, start2, end1, end2, part1, part2;
	start1 = start2 = 0;
	while (true) {
		if ((end1 = specificationVersion.indexOf('.', start1)) == -1)
			end1 = length1;
		if ((end2 = version.indexOf('.', start2)) == -1)
			end2 = length2;

		// parse the parts and compare
		part1 = Integer.parseInt(specificationVersion.substring(start1, end1));
		part2 = Integer.parseInt(version.substring(start2, end2));
		if (part1 < part2) return false;
		if (part1 > part2) return true;

		// advance past the '.'
		start1 = end1 + 1;
		start2 = end2 + 1;

		// check the end-of-string cases
		// End of the supplied version, must be >=
		// i.e: 1.1.6 and 1.1 -> true
		if (start2 >= length2) return true;

		// End of the spec version, supplied version continues.
		// We have to have already exceeded the supplied version;
		// equality no long sufficies.
		// i.e. 1.2 and 1.1.6 -> true
		//   1.1 and 1.1.6 -> false
		//   1.1.0 and 1.1 -> true
		if (start1 >= length1) {
			while (start2 < length2) {
				if ((end2 = version.indexOf('.', start2)) == -1)
					end2 = length2;
				part2 = Integer.parseInt(version.substring(start2, end2));
				if (part2 > 0) return false;
				start2 = end2 + 1;
			}
			return true;
		}
	}
}
/**
 * Return true if this package is sealed,
 * false otherwise.
 *
 * @return 		true if this package is sealed, false otherwise
 */
public boolean isSealed() {
	return sealBase != null;
}

/**
 * Returns true if this is a sealed package loaded from the jar at location
 * <code>URL</code>.
 *
 * @param url
 *            the <code>URL</code> to test
 * @return <code>true</code> if this is a sealed package and the supplied
 *         {@link URL} specifies the location of the jar file from where
 *         this package was loaded. Otherwise <code>false</code>.
 */
public boolean isSealed(URL url) {
	return sealBase != null && sealBase.sameFile(url);
}
/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return		a printable representation for the receiver.
 */
public String toString() {
	StringBuffer buffer = new StringBuffer();
	buffer.append("package ");
	buffer.append(name);
	if (specificationTitle != null && specificationTitle.length() > 0) {
		buffer.append(", ");
		buffer.append(specificationTitle);
	}
	if (specificationVersion != null && specificationVersion.length() > 0) {
		buffer.append(", version ");
		buffer.append(specificationVersion);
	}
	return buffer.toString();
}

}
