package java.lang;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.File;

/**
 * The Runtime interface.
 */

public class Runtime {
	// single instance of Runtime class.
	private static final Runtime runtime = new Runtime();
/**
 * Prevent instantiation with private constructor
 */
private Runtime() {
}

/**
 * Execute progAray[0] in a separate platform process
 * The new process inherits the environment of the caller.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		progArray		the array containing the program to execute as
 *								well as any arguments to the program.
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String[] progArray) throws java.io.IOException{
	return exec(progArray, null);
}

/**
 * Execute progArray[0] in a separate platform process
 * The new process uses the environment provided in envp
 *
 * @author		OTI
 * @version		initial
 *
 * @param		progArray	the array containing the program to execute a
 *							well as any arguments to the program.
 *
 * @param		envp	    the array containing the environment to start
 *                          the new process in.
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String[] progArray, String[] envp) throws java.io.IOException{
	return exec(progArray, envp, null);
}

/**
 * Execute progArray[0] in a separate platform process
 * The new process uses the environment provided in envp
 *
 * @author		OTI
 * @version		initial
 *
 * @param		progArray	the array containing the program to execute a
 *							well as any arguments to the program.
 *
 * @param		envp	    the array containing the environment to start
 *                          the new process in.
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String[] progArray, String[] envp, File directory) throws java.io.IOException{
	SecurityManager smngr = System.getSecurityManager();
	if (smngr != null)
		smngr.checkExec(progArray[0]);
	if (envp == null) envp = new String[0];
	return com.ibm.oti.lang.SystemProcess.create(progArray,envp, directory);
}

/**
 * Execute program in a separate platform process
 * The new process inherits the environment of the caller.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		prog		the name of the program to execute
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String prog) throws java.io.IOException{
	return exec(prog, null);
}

/**
 * Execute prog in a separate platform process
 * The new process uses the environment provided in envp
 *
 * @author		OTI
 * @version		initial
 *
 * @param		prog     	the name of the program to execute
 *
 * @param		envp	    the array containing the environment to start
 *                          the new process in.
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String prog, String[] envp) throws java.io.IOException{
	return exec(prog, envp, null);
}

/**
 * Execute prog in a separate platform process
 * The new process uses the environment provided in envp
 *
 * @author		OTI
 * @version		initial
 *
 * @param		prog     	the name of the program to execute
 *
 * @param		envp	    the array containing the environment to start
 *                          the new process in.
 * @param		directory	the initial directory for the subprocess, or null
 *					to use the directory of the current process
 *
 * @exception	java.io.IOException
 *							if the program cannot be executed
 *
 * @exception	SecurityException
 *							if the current SecurityManager disallows
 *							program execution
 *
 * @see			SecurityManager#checkExec
 */
public Process exec(String prog, String[] envp, File directory) throws java.io.IOException{
	//use a regular StringTokenizer to break the command_line into
	//small peaces. By convention the first argument is the name of the
	//command.

	//System.out.println("cmd-->"+prog);

	//the first token is the file_to_load; the rest is the arguments
	int argsLenghPlusOne;
	int i=0;
	java.util.StringTokenizer slicer = new java.util.StringTokenizer(prog);
	String[] command = new String[argsLenghPlusOne=slicer.countTokens()];
	while (i<argsLenghPlusOne)
		command[i++] = slicer.nextToken();
	return exec(command, envp, directory);
}

/**
 * Causes the virtual machine to stop running, and the
 * program to exit. If runFinalizersOnExit(true) has been
 * invoked, then all finalizers will be run first.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		code		the return code.
 *
 * @exception	SecurityException
 *							if the running thread is not allowed
 *							to cause the vm to exit.
 *
 * @see			SecurityManager#checkExit
 */
public void exit (int code) {
	SecurityManager smngr = System.getSecurityManager();
	if (smngr != null)
		smngr.checkExit(code);
	exitImpl(code);
}

private native void exitImpl(int code);

/**
 * Answers the amount of free memory resources which
 * are available to the running program.
 *
 * @author		OTI
 * @version		initial
 */
public native long freeMemory();

/**
 * Indicates to the virtual machine that it would be a
 * good time to collect available memory. Note that, this
 * is a hint only.
 *
 * @author		OTI
 * @version		initial
 */
public native void gc();
/**
 * Return the single Runtime instance
 *
 * @author		OTI
 * @version		initial
 */
public static Runtime getRuntime(){
	return runtime;
}

/**
 * Loads and links the library specified by the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		pathName		the absolute (ie: platform dependent)
 *								path to the library to load
 *
 * @exception	UnsatisfiedLinkError
 *							if the library could not be loaded
 * @exception	SecurityException
 *							if the library was not allowed to be loaded
 */
public synchronized void load(String pathName) {
	SecurityManager smngr = System.getSecurityManager();
	if (smngr != null)
		smngr.checkLink(pathName);
	ClassLoader.loadLibraryWithPath(pathName, ClassLoader.callerClassLoader(), null);
}

/**
 * Loads and links the library specified by the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		libName		the name of the library to load
 *
 * @exception	UnsatisfiedLinkError
 *							if the library could not be loaded
 * @exception	SecurityException
 *							if the library was not allowed to be loaded
 */
public void loadLibrary(String libName) {
	ClassLoader.loadLibraryWithClassLoader(libName, ClassLoader.callerClassLoader());
}

/**
 * Provides a hint to the virtual machine that it would
 * be useful to attempt to perform any outstanding
 * object finalizations.
 *
 * @author		OTI
 * @version		initial
 */
public native void runFinalization();

/**
 * Answers the total amount of memory resources which
 * is available to (or in use by) the running program.
 *
 * @author		OTI
 * @version		initial
 */
public native long totalMemory();

public void traceInstructions(boolean enable) {}
public void traceMethodCalls(boolean enable) {}

/**
 * Registers a new virtual-machine shutdown hook.
 *
 * @param		hook	the hook (a Thread) to register
 *
 * @author		OTI
 * @version		initial
 */
public void addShutdownHook(Thread hook) {
	com.ibm.oti.vm.VM.addShutdownHook(hook);
}

/**
 * De-registers a previously-registered virtual-machine shutdown hook.
 *
 * @param		hook	the hook (a Thread) to de-register
 * @return		true if the hook could be de-registered
 *
 * @author		OTI
 * @version		initial
 */
public boolean removeShutdownHook(Thread hook) {
	return com.ibm.oti.vm.VM.removeShutdownHook(hook);
}

/**
 * Causes the virtual machine to stop running, and the
 * program to exit. Finalizers will not be run first.
 * Shutdown hooks will not be run.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		code		the return code.
 *
 * @exception	SecurityException
 *							if the running thread is not allowed
 *							to cause the vm to exit.
 *
 * @see			SecurityManager#checkExit
 */
public void halt (int code) {
	SecurityManager smngr = System.getSecurityManager();
	if (smngr != null)
		smngr.checkExit(code);
	haltImpl(code);
}

private native void haltImpl(int code);

/**
 * Return the number of processors, always at least one.
 */
public int availableProcessors() {
	return availableProcessorsImpl();
}

private native int availableProcessorsImpl();

/**
 * Return the maximum memory that will be used by the virtual machine,
 * or Long.MAX_VALUE.
 */
public long maxMemory() {
	return maxMemoryImpl();
}

private native long maxMemoryImpl();

}
