package java.lang;

import java.io.FileDescriptor;

import java.security.AccessController;
import java.security.AccessControlContext;
import java.security.Permission;
import java.security.Security;
import java.security.SecurityPermission;
import java.io.File;
import java.io.FilePermission;
import java.net.SocketPermission;
import java.util.PropertyPermission;
import java.util.StringTokenizer;

import com.ibm.oti.util.PriviAction;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * SecurityManager is the abstract superclass of the classes
 * which can provide security verification for a running
 * program.
 *
 * @author		OTI
 * @version		initial
 */
public class SecurityManager {
	static private volatile String[] securePackageList = null;
	static private volatile String packageAccess;

	/**
	 * Optimization: Don't create a new PropertyPermission each time.
	 */
	private PropertyPermission permissionToReadWriteAllProperties =
		new PropertyPermission("*", "read,write");

/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 */
public SecurityManager() {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(
			RuntimePermission.permissionToCreateSecurityManager);
	Class type = Security.class; // initialize Security properties
}

/**
 * Checks whether the running program is allowed to
 * accept socket connections.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		host	the address of the host which is attempting to connect
 * @param		port	the port number to check
 */
public void checkAccept(String host, int port) {
	if (host == null) throw new NullPointerException();
	checkPermission(new SocketPermission(host + ':' + port, "accept"));
}

/**
 * Checks whether the running program is allowed to
 * modify the thread.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		thread		the thread we are attempting to modify
 */
public void checkAccess(Thread thread) {
	// Only worry about system threads.
	ThreadGroup group = thread.getThreadGroup();
	if (group != null && group.parent == null)
		checkPermission(RuntimePermission.permissionToModifyThread);
}

/**
 * Checks whether the running program is allowed to
 * modify the thread group.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		group		the thread group we are attempting to modify
 */
public void checkAccess(ThreadGroup group) {
	// Only worry about system threads.
	if (group == null) throw new NullPointerException();
	if (group.parent == null)
		checkPermission(RuntimePermission.permissionToModifyThreadGroup);
}

/**
 * Checks whether the running program is allowed to
 * establish socket connections. A -1 port indicates
 * the caller is trying to resolve the hostname.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		host String
 *					the address of the host to connect to.
 * @param		port int
 *					the port number to check, or -1 for resolve.
 */
public void checkConnect(String host, int port) {
	if (host == null) throw new NullPointerException();
	if (port > 0)
		checkPermission(new SocketPermission(host + ':' + port, "connect"));
	else
		checkPermission(new SocketPermission(host, "resolve"));
}

/**
 * Checks whether the given security context is allowed to
 * establish socket connections. A -1 port indicates the
 * caller is trying to resolve the hostname.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		host String
 *					the address of the host to connect to.
 * @param		port int
 *					the port number to check, or -1 for resolve.
 * @param		context Object
 *					the security context to use for the check.
 */
public void checkConnect(String host, int port, Object context) {
	if (port > 0)
		checkPermission(
			new SocketPermission(host + ':' + port, "connect"),
			context);
	else
		checkPermission(
			new SocketPermission(host, "resolve"),
			context);
}

/**
 * Checks whether the running program is allowed to
 * create a class loader.
 *
 * @author		OTI
 * @version		initial
 */
public void checkCreateClassLoader() {
	checkPermission(RuntimePermission.permissionToCreateClassLoader);
}

/**
 * Checks whether the running program is allowed to
 * delete the file named by the argument, which should
 * be passed in canonical form.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file		the name of the file to check
 */
public void checkDelete(String file) {
	checkPermission(new FilePermission(file, "delete"));
}

/**
 * Checks whether the running program is allowed to
 * execute the specified platform specific command.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		cmd		the command line
 */
public void checkExec(String cmd) {
	checkPermission(new FilePermission(new File(cmd).isAbsolute() ? cmd : "<<ALL FILES>>", "execute"));
}

/**
 * Checks whether the running program is allowed to
 * terminate itself.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		status		the status to return from the exit.
 */
public void checkExit(int status) {
	checkPermission(RuntimePermission.permissionToExitVM);
}

/**
 * Checks whether the running program is allowed to
 * load the specifed native library.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		libName		the name of the library to load
 */
public void checkLink(String libName) {
	if (libName == null) throw new NullPointerException();
	checkPermission(new RuntimePermission("loadLibrary." + libName));
}

/**
 * Checks whether the running program is allowed to
 * listen on the specified port.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		port int
 *					the port number to check
 */
public void checkListen(int port) {
	if (port == 0)
		checkPermission(new java.net.SocketPermission("localhost:1024-", "listen"));
	else
		checkPermission(new java.net.SocketPermission("localhost:" + port, "listen"));
}

/**
 * Checks whether the running program is allowed to
 * access members. The default is to allow access to
 * public members (i.e. java.lang.reflect.PUBLIC) and
 * to classes loaded by the same loader as the original
 * caller (i.e. the method that called the reflect API).
 *
 * Due to the nature of the check, overriding
 * implementations cannot call super.checkMemberAccess()
 * since the stack would no longer be of the expected
 * shape.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		cls			?
 * @param 		type		Either java.lang.reflect.Member.PUBLIC or DECLARED
 */
public void checkMemberAccess(Class cls, int type) {
	if (cls == null) throw new NullPointerException();
	if(type == java.lang.reflect.Member.PUBLIC) return;
	//
	//	Need to compare the classloaders.
	//	Stack shape is
	//		<user code>						<- want this class
	//		Class.getDeclared*();
	//		Class.checkMemberAccess();
	//		SecurityManager.checkMemberAccess();	<- current frame
	//
	// Use getClassLoaderImpl() since getClassLoader()
	// returns null for the bootstrap class loader.
	if(ClassLoader.getStackClassLoader(3) == cls.getClassLoaderImpl()) return;

	// Forward off to the permission mechanism.
	checkPermission(new RuntimePermission("accessDeclaredMembers"));
}

/**
 * Checks whether the running program is allowed to
 * join, leave or send to a multicast address.
 *
 * @author		OTI
 * @version		initial
 */
public void checkMulticast(java.net.InetAddress maddr) {
	checkPermission(new SocketPermission(maddr.getHostAddress(), "accept,connect"));
}

/**
 * Checks whether the running program is allowed to
 * join, leave or send to a multicast address.
 *
 * @author		OTI
 * @version		initial
 * @deprecated  use SecurityManager#checkMulticast(java.net.InetAddress)
 */
public void checkMulticast(java.net.InetAddress maddr, byte ttl) {
	checkPermission(new SocketPermission(maddr.getHostAddress(), "accept,connect"));
}

/**
 * Checks whether the running program is allowed to
 * access the specified package.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		packageName	the name of the package to be accessed.
 */
public void checkPackageAccess(String packageName) {
	if (packageName == null) throw new NullPointerException();
	String securePackages = getSecurityProperty("package.access");
	String[] packageList;
	if (securePackages != packageAccess) {
		if (securePackages != null) {
			StringTokenizer tokenizer = new StringTokenizer(securePackages, ", ");
			int i = 0;
			packageList = new String[tokenizer.countTokens()];
			while (tokenizer.hasMoreTokens()) {
				packageList[i++] = tokenizer.nextToken();
			}
			packageAccess = securePackages;
		} else {
			packageList = new String[0];
		}
		securePackageList = packageList;
	} else {
		packageList = securePackageList;
	}
	for (int i=0; i<packageList.length; i++) {
		if (packageName.startsWith(packageList[i])) {
			checkPermission(new RuntimePermission("accessClassInPackage." + packageName));
			return;
		}
	}
}

/**
 * Checks whether the running program is allowed to
 * define new classes in the specified package.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		packageName	the name of the package to add a class to.
 */
public void checkPackageDefinition(String packageName) {
	if (packageName == null) throw new NullPointerException();
	String securePackages = getSecurityProperty("package.definition");
	if (securePackages != null) {
		StringTokenizer tokenizer = new StringTokenizer(securePackages, ", ");
		while (tokenizer.hasMoreTokens()) {
			if (packageName.startsWith(tokenizer.nextToken())) {
				checkPermission(new RuntimePermission("defineClassInPackage." + packageName));
				return;
			}
		}
	}
}

private String getSecurityProperty(final String property) {
    return (String) AccessController.doPrivileged(
            PriviAction.getSecurityProperty(property));
}

/**
 * Checks whether the running program is allowed to
 * access the system properties.
 *
 * @author		OTI
 * @version		initial
 */
public void checkPropertiesAccess() {
	checkPermission(permissionToReadWriteAllProperties);
}

/**
 * Checks whether the running program is allowed to
 * access a particular system property.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key			the name of the property to be accessed.
 */
public void checkPropertyAccess(String key) {
	checkPermission(new PropertyPermission(key, "read"));
}

/**
 * Checks whether the running program is allowed to
 * read from the file whose descriptor is the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		fd		the file descriptor of the file to check
 */
public void checkRead(FileDescriptor fd) {
	if (fd == null) throw new NullPointerException();
	checkPermission(RuntimePermission.permissionToReadFileDescriptor);
}

/**
 * Checks whether the running program is allowed to
 * read from the file named by the argument, which should
 * be passed in canonical form.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file String
 *					the name of the file or directory to check.
 */
public void checkRead(String file) {
	checkPermission(new FilePermission(file, "read"));
}

/**
 * Checks whether the given security context is allowed to
 * read from the file named by the argument, which should
 * be passed in canonical form.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file String
 *					the name of the file or directory to check.
 * @param		context Object
 *					the security context to use for the check.
 */
public void checkRead(String file, Object context) {
	checkPermission(new FilePermission(file, "read"), context);
}

/**
 * Checks whether the running program is allowed to perform
 * the security operation named by the target.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		target String
 *					the name of the operation to perform.
 */
public void checkSecurityAccess(String target) {
	checkPermission(new SecurityPermission(target));
}

/**
 * Checks whether the running program is allowed to
 * set the net object factories.
 *
 * @author		OTI
 * @version		initial
 */
public void checkSetFactory() {
	checkPermission(RuntimePermission.permissionToSetFactory);
}

/**
 * Checks whether the running program is allowed to
 * create a top level window.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		window		The non-null window for which to check access
 */
public boolean checkTopLevelWindow (Object window) {
	if (window == null) throw new NullPointerException();
	try {
		Class awtPermission = Class.forName("java.awt.AWTPermission");
		Class[] args = new Class[] { java.lang.String.class };
		java.lang.reflect.Constructor constructor = awtPermission.getConstructor(args);
		Object[] constructorArgs = new Object[] {"showWindowWithoutWarningBanner"};
		java.security.Permission perm = (java.security.Permission) constructor.newInstance(constructorArgs);
		checkPermission(perm);
	} catch (Exception e) {
		return false;
	}
	return true;
}

/**
 * Checks whether the running program is allowed to
 * access the system clipboard.
 *
 * @author		OTI
 * @version		initial
 */
public void checkSystemClipboardAccess () {
	try {
		Class awtPermission = Class.forName("java.awt.AWTPermission");
		Class[] args = new Class[] { java.lang.String.class };
		java.lang.reflect.Constructor constructor = awtPermission.getConstructor(args);
		Object[] constructorArgs = new Object[] {"accessClipboard"};
		java.security.Permission perm = (java.security.Permission) constructor.newInstance(constructorArgs);
		checkPermission(perm);
		return;
	} catch (ClassNotFoundException e) {
	} catch (NoSuchMethodException e) {
	} catch (InstantiationException e) {
	} catch (IllegalAccessException e) {
	} catch (java.lang.reflect.InvocationTargetException e) {
	}
	throw new SecurityException();
}

/**
 * Checks whether the running program is allowed to
 * access the AWT Event queue. Since we don't support AWT,
 * the answer is no.
 *
 * @author		OTI
 * @version		initial
 */
public void checkAwtEventQueueAccess () {
	try {
		Class awtPermission = Class.forName("java.awt.AWTPermission");
		Class[] ar = new Class[] { java.lang.String.class };
		java.lang.reflect.Constructor constructor = awtPermission.getConstructor(ar);
		Object[] constructorArgs = new Object[] {"accessEventQueue"};
		java.security.Permission perm = (java.security.Permission) constructor.newInstance(constructorArgs);
		checkPermission(perm);
		return;
	} catch (ClassNotFoundException e) {
	} catch (NoSuchMethodException e) {
	} catch (InstantiationException e) {
	} catch (IllegalAccessException e) {
	} catch (java.lang.reflect.InvocationTargetException e) {
	}
	throw new SecurityException();
}

/**
 * Checks whether the running program is allowed to
 * start a new print job.
 *
 * @author		OTI
 * @version		initial
 */
public void checkPrintJobAccess() {
	checkPermission(RuntimePermission.permissionToQueuePrintJob);
}

/**
 * Checks whether the running program is allowed to
 * read from the file whose descriptor is the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		fd		the file descriptor of the file to check
 */
public void checkWrite(FileDescriptor fd) {
	if (fd==null) throw new NullPointerException();
	checkPermission(RuntimePermission.permissionToWriteFileDescriptor);
}

/**
 * Checks whether the running program is allowed to
 * write to the file named by the argument, which should
 * be passed in canonical form.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file		the name of the file to check
 */
public void checkWrite(String file) {
	checkPermission(new FilePermission(file, "write"));
}

/**
 * Answers an array containing one entry for each method
 * in the stack. Each entry is the java.lang.Class which
 * represents the class in which the method is defined.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Class[]
 *					all of the classes in the stack.
 */
protected Class[] getClassContext() {
	return Class.getStackClasses(-1, false);
}

/**
 * Answers the thread group which should be used to instantiate
 * new threads. By default, this is the same as the thread group
 * of the thread running this method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		ThreadGroup
 *					The thread group to create new threads in.
 */
public ThreadGroup getThreadGroup () {
	return Thread.currentThread().getThreadGroup();
}

/**
 * Answers an object which encapsulates the security state
 * of the current point in the execution. In our case, this
 * is an AccessControlContext.
 *
 * @author		OTI
 * @version		initial
 */
public Object getSecurityContext () {
	return AccessController.getContext();
}

/**
 * Checks whether the running program is allowed to
 * access the resource being guarded by the given
 * Permission argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		permission		the permission to check
 */
public void checkPermission(Permission permission) {
	AccessController.checkPermission(permission);
}

/**
 * Checks whether the running program is allowed to
 * access the resource being guarded by the given
 * Permission argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		permission		the permission to check
 */
public void checkPermission(Permission permission, Object context) {
	// Must be an AccessControlContext. If we don't check
	// this, then applications could pass in an arbitrary
	// object which circumvents the security check.
	if (context instanceof AccessControlContext)
		((AccessControlContext) context).checkPermission(permission);
	else
		throw new SecurityException();
}

}
