package java.lang;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.*;
import java.util.Properties;
import java.util.PropertyPermission;
import java.security.*;

/**
 * Class System provides a standard place for programs
 * to find system related information. All System API
 * is static.
 *
 * @author		OTI
 * @version		initial
 */
public final class System {

	// The standard input, output, and error streams.
	// Typically, these are connected to the shell which
	// ran the Java program.
	/**
	 * Default input stream
	 */
	public static final InputStream in;
	/**
	 * Default output stream
	 */
	public static final PrintStream out;
	/**
	 * Default error output stream
	 */
	public static final PrintStream err;

	//Get a ref to the Runtime instance for faster lookup
	private static final Runtime RUNTIME = Runtime.getRuntime();
	/**
	 * The System Properties table.
	 */
	private static Properties systemProperties;

	/**
	 * The System default SecurityManager.
	 */
	private static SecurityManager security;

	// Initialize all the slots in System on first use.
	static {

		// Fill in the properties from the VM information.
		ensureProperties();
		// Set up standard in, out, and err.
		err = new String.ConsolePrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.err)));
		out = new String.ConsolePrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.out)));
		in = new BufferedInputStream(new FileInputStream(FileDescriptor.in));
	}

static void completeInitialization() {
	String extraLibs = (String)systemProperties.get("com.ibm.util.extralibs.properties");
	if (extraLibs != null) {
		int lastIndex = 0;
		while (lastIndex < extraLibs.length()) {
			int index = extraLibs.indexOf(':', lastIndex);
			if (index == -1) index = extraLibs.length();
			String libName = extraLibs.substring(lastIndex, index).trim();
			lastIndex = index + 1;
			if (libName.length() == 0) continue;
			try {
				String extensionName = "com.ibm.oti." + libName + ".SystemPropertyExtension";
				Class extensionCls = Class.forName(extensionName);
				Object extension = extensionCls.newInstance();
				try {
					if (extension instanceof com.ibm.oti.util.SystemPropertiesHook) {
						systemProperties = (Properties)((com.ibm.oti.util.SystemPropertiesHook)extension).extendSystemProperties(systemProperties);
					} else {
					((com.ibm.oti.util.SystemProperties)extension).extendSystemProperties(systemProperties);
					}
				} catch (Throwable t) {
					t.printStackTrace();
				}
			} catch (Throwable t) {
			}
		}
	}
	// Must install a SecurityManager before user code runs
	installSecurityManager();
}

private static void installSecurityManager() {
	if (System.getSecurityManager() == null) {
		String clName = System.getProperty("java.security.manager");
		if (clName != null) {
			SecurityManager sm = null;
			if (clName.length() == 0) {
				sm = new SecurityManager();
			} else {
				try {
					Class cl = Class.forName(clName);
					sm = (SecurityManager)cl.newInstance();
				} catch (Exception e) {
					// K00e3 = Could not create specified security manager: {0}
					throw new InternalError(com.ibm.oti.util.Msg.getString("K00e3", clName));
				}
			}
			System.setSecurityManager(sm);
		}
	}
}

/**
 * Sets the value of the static slot "in" in the receiver
 * to the passed in argument.
 *
 * @param		newIn 		the new value for in.
 */
public static void setIn(InputStream newIn) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(
			RuntimePermission.permissionToSetIO);
	setFieldImpl("in", newIn);
}

/**
 * Sets the value of the static slot "out" in the receiver
 * to the passed in argument.
 *
 * @param		newOut 		the new value for out.
 */
public static void setOut(java.io.PrintStream newOut) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(
			RuntimePermission.permissionToSetIO);
	setFieldImpl("out", newOut);
}

/**
 * Sets the value of the static slot "err" in the receiver
 * to the passed in argument.
 *
 * @param		newErr  	the new value for err.
 */
public static void setErr(java.io.PrintStream newErr) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(
			RuntimePermission.permissionToSetIO);
	setFieldImpl("err", newErr);
}

/**
 * Prevents this class from being instantiated.
 */
private System() {
}

/**
 * Copies the contents of <code>array1</code> starting at offset <code>start1</code>
 * into <code>array2</code> starting at offset <code>start2</code> for
 * <code>length</code> elements.
 *
 * @param		array1 		the array to copy out of
 * @param		start1 		the starting index in array1
 * @param		array2 		the array to copy into
 * @param		start2 		the starting index in array2
 * @param		length 		the number of elements in the array to copy
 */
public static native void arraycopy(Object array1, int start1, Object array2, int start2, int length);

/**
 * Private version of the arraycopy method used by the jit for reference arraycopies
 */
private static void arraycopy(Object[] A1, int offset1, Object[] A2, int offset2, int length) {
	if (A1 == null || A2 == null) throw new NullPointerException();
	if (offset1 >= 0 && offset2 >= 0 && length >= 0 &&
		length <= A1.length - offset1 && length <= A2.length - offset2)
	{
		// Check if this is a forward or backwards arraycopy
		if (A1 != A2 || offset1 > offset2 || offset1+length <= offset2) {
			for (int i = 0; i < length; ++i) {
				A2[offset2+i] = A1[offset1+i];
			}
		} else {
			for (int i = length-1; i >= 0; --i) {
				A2[offset2+i] = A1[offset1+i];
			}
		}
	} else throw new ArrayIndexOutOfBoundsException();
}

/**
 * Answers the current time expressed as milliseconds since
 * the time 00:00:00 UTC on January 1, 1970.
 *
 * @return		the time in milliseconds.
 */
public static native long currentTimeMillis();

private static final int InitLocale = 0;
private static final int PlatformEncoding = 1;
private static final int FileEncoding = 2;
private static final int OSEncoding = 3;

/**
 * If systemProperties is unset, then create a new one based on the values
 * provided by the virtual machine.
 */
private static void ensureProperties() {
	systemProperties = new Properties();

	String platformEncoding = null;
	String fileEncoding, osEncoding = null;
	String definedFileEncoding = getEncoding(FileEncoding);
	String definedOSEncoding = getEncoding(OSEncoding);
	if (definedFileEncoding != null) {
		fileEncoding = definedFileEncoding;
		// if file.encoding is defined, and os.encoding is not, use the detected
		// platform encoding for os.encoding
		if (definedOSEncoding == null) {
			platformEncoding = getEncoding(PlatformEncoding);
			osEncoding = platformEncoding;
		} else {
			getEncoding(InitLocale);
		}
	} else {
		platformEncoding = getEncoding(PlatformEncoding);
		fileEncoding = platformEncoding;
	}
	// if os.encoding is not defined, file.encoding will be used
	if (osEncoding == null)
		osEncoding = definedOSEncoding;
	if (osEncoding != null)
		systemProperties.put("os.encoding", osEncoding);

	systemProperties.put("file.encoding", fileEncoding);

	systemProperties.put("microedition.commports",getCommPortList());

	systemProperties.put("java.specification.vendor", "Sun Microsystems Inc.");
	systemProperties.put("java.specification.name", "J2ME Foundation Specification");
	systemProperties.put("java.specification.version", "1.1");
	systemProperties.put("java.version", "J2ME Foundation Specification v1.1");

	systemProperties.put("com.ibm.oti.configuration", "foun11");
	systemProperties.put("com.ibm.oti.configuration.dir", "jclFoundation11" );

	String[] list = getPropertyList();
	for (int i=0; i<list.length; i+=2) {
		String key = list[i];
		if (key == null) break;
		systemProperties.put(key, list[i+1]);
	}

	String consoleEncoding = (String)systemProperties.get("console.encoding");
	if (consoleEncoding == null) {
		if (platformEncoding == null)
			platformEncoding = getEncoding(PlatformEncoding);
		consoleEncoding = platformEncoding;
		systemProperties.put("console.encoding", consoleEncoding);
	}

	systemProperties.put("com.ibm.oti.jcl.build", "20061020_1321");
}
/**
 * Causes the virtual machine to stop running, and the
 * program to exit. If runFinalizersOnExit(true) has been
 * invoked, then all finalizers will be run first.
 *
 * @param		code		the return code.
 *
 * @throws		SecurityException 	if the running thread is not allowed to cause the vm to exit.
 *
 * @see			SecurityManager#checkExit
 */
public static void exit(int code) {
	RUNTIME.exit(code);
}

/**
 * Indicate to the virtual machine that it would be a
 * good time to collect available memory. Note that, this
 * is a hint only.
 */
public static void gc() {
	RUNTIME.gc();
}

/**
 * Answers the system properties. Note that this is
 * not a copy, so that changes made to the returned
 * Properties object will be reflected in subsequent
 * calls to getProperty and getProperties.
 * <p>
 * Security managers should restrict access to this
 * API if possible.
 *
 * @return		the system properties
 */
public static Properties getProperties() {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPropertiesAccess();
	return systemProperties;
}

/**
 * Answers the system properties without any security
 * checks. This is used for access from within java.lang.
 *
 * @return		the system properties
 */
static Properties internalGetProperties() {
	return systemProperties;
}
/**
 * Answers the value of a particular system property.
 * Answers null if no such property exists,
 * <p>
 * The properties currently provided by the virtual
 * machine are:
 * <pre>
 *     java.vendor.url
 *     java.class.path
 *     user.home
 *     java.class.version
 *     os.version
 *     java.vendor
 *     user.dir
 *     user.timezone
 *     path.separator
 *     os.name
 *     os.arch
 *     line.separator
 *     file.separator
 *     user.name
 *     java.version
 *     java.home
 * </pre>
 *
 * @param		prop		the system property to look up
 * @return		the value of the specified system property, or null if the property doesn't exist
 */
public static String getProperty(String prop) {
	return getProperty(prop, null);
}

/**
 * Answers the value of a particular system property.
 * If no such property is found, answers the defaultValue.
 *
 * @param		prop			the system property to look up
 * @param		defaultValue	return value if system property is not found
 * @return		the value of the specified system property, or defaultValue if the property doesn't exist
 */
public static String getProperty(String prop, String defaultValue) {
	if (prop.length() == 0) throw new IllegalArgumentException();
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPropertyAccess(prop);
	return systemProperties.getProperty(prop, defaultValue);
}

/**
 * Sets the value of a particular system property.
 *
 * @param		prop		the system property to change
 * @param		value		the value to associate with prop
 * @return		the old value of the property, or null
 */
public static String setProperty(String prop, String value) {
	if (prop.length() == 0) throw new IllegalArgumentException();
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(
			new PropertyPermission(prop, "write"));
	return (String)systemProperties.setProperty(prop, value);
}

/**
 * Answers an array of Strings containing key..value pairs
 * (in consecutive array elements) which represent the
 * starting values for the system properties as provided
 * by the virtual machine.
 *
 * @return		the default values for the system properties.
 */
private static native String [] getPropertyList();

/**
 * Return a string listing the available comm ports.
 */
private static native String getCommPortList();

/**
 * Return the requested encoding.
 * 		0 - initialize locale
 * 		1 - detected platform encoding
 * 		2 - command line defined file.encoding
 * 		3 - command line defined os.encoding
 */
private static native String getEncoding(int type);

/**
 * Answers the active security manager.
 *
 * @return		the system security manager object.
 */
public static SecurityManager getSecurityManager() {
	return security;
}

/**
 * Answers an integer hash code for the parameter.
 * The hash code returned is the same one that would
 * be returned by java.lang.Object.hashCode(), whether
 * or not the object's class has overridden hashCode().
 * The hash code for null is 0.
 *
 * @param		anObject	the object
 * @return		the hash code for the object
 *
 * @see			java.lang.Object#hashCode
 */
public static native int identityHashCode(Object anObject);

/**
 * Loads the specified file as a dynamic library.
 *
 * @param 		pathName	the path of the file to be loaded
 */
public static void load(String pathName) {
	SecurityManager smngr = System.getSecurityManager();
	if (smngr != null)
		smngr.checkLink(pathName);
	ClassLoader.loadLibraryWithPath(pathName, ClassLoader.callerClassLoader(), null);
}

/**
 * Loads and links the library specified by the argument.
 *
 * @param		libName		the name of the library to load
 *
 * @throws		UnsatisfiedLinkError	if the library could not be loaded
 * @throws		SecurityException 		if the library was not allowed to be loaded
 */
public static void loadLibrary(String libName) {
	ClassLoader.loadLibraryWithClassLoader(libName, ClassLoader.callerClassLoader());
}

/**
 * Provides a hint to the virtual machine that it would
 * be useful to attempt to perform any outstanding
 * object finalizations.
 */
public static void runFinalization() {
	RUNTIME.runFinalization();
}

/**
 * Answers the system properties. Note that the object
 * which is passed in not copied, so that subsequent
 * changes made to the object will be reflected in
 * calls to getProperty and getProperties.
 * <p>
 * Security managers should restrict access to this
 * API if possible.
 *
 * @param		p			the property to set
 */
public static void setProperties(Properties p) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPropertiesAccess();
	if (p == null) ensureProperties();
	else systemProperties = p;
}

/**
 * Sets the active security manager. Note that once
 * the security manager has been set, it can not be
 * changed. Attempts to do so will cause a security
 * exception.
 *
 * @param		s			the new security manager
 *
 * @throws		SecurityException 	if the security manager has already been set.
 */
public static void setSecurityManager(final SecurityManager s) {
	final SecurityManager currentSecurity = security;
	if (s != null) {
		try {
			// Preload classes used for checkPackageAccess(),
			// otherwise we could go recursive
			s.checkPackageAccess("java.lang");
			// must preload the default Policy
			Policy policy = (Policy)AccessController.doPrivileged(new com.ibm.oti.util.PriviAction());
			ProtectionDomain oldDomain = null;
			if (currentSecurity != null) {
				oldDomain = currentSecurity.getClass().getPDImpl();
			}
			ProtectionDomain newDomain = s.getClass().getPDImpl();
			if (newDomain != null && newDomain != oldDomain) {
				// must preload the dynamic permissions from the policy before installing
				policy.getPermissions(newDomain);
			}
		} catch (Exception e) {}
	}
	if (currentSecurity != null) {
		currentSecurity.checkPermission(
			RuntimePermission.permissionToSetSecurityManager);
	}
	security = s;
}

/**
 * Answers the platform specific file name format for the shared
 * library named by the argument.
 *
 * @param		userLibName 	the name of the library to look up.
 * @return		the platform specific filename for the library
 */
public static native String mapLibraryName(String userLibName);

/**
 * Sets the value of the named static field in the receiver
 * to the passed in argument.
 *
 * @param		fieldName 	the name of the field to set, one of in, out, or err
 * @param		stream 		the new value of the field
 */
private static native void setFieldImpl(String fieldName, Object stream);

}
