package java.lang.reflect;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * This class models a field. Information about the field
 * can be accessed, and the field's value can be accessed dynamically.
 *
 * @author		OTI
 * @version		initial
 */
public final class Field extends AccessibleObject implements Member
{
	private Class declaringClass;
	private Class type;
	private String name;
	private long vm1;

/**
 * Prevent this class from being instantiated.
 */
private Field() {}

/**
 * Compares the specified object to this Field and answer if they
 * are equal. The object must be an instance of Field with the same
 * defining class and name.
 *
 * @param		object		the object to compare
 * @return		true if the specified object is equal to this Field, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals(Object object)
{
	Field field;

	if(this == object) return true;
	if(!(object instanceof Field)) return false;
	field = (Field)object;
	if(!getName().equals(field.getName())) return false;
	if(getDeclaringClass() != field.getDeclaringClass()) return false;
	if(getType() != field.getType()) return false;
	return true;
}

/**
 * Return the value of the field in the specified object.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * The value of the field is returned. If the type of this field is a base type, the field
 * value is automatically wrapped.
 *
 * @param		object		the object to access
 * @return		the field value, possibly wrapped
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public Object get(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getImpl(object);
}

private boolean requiresProtectedCheck() {
	return !isAccessible() && ((getModifiers() & (Modifier.STATIC | Modifier.PROTECTED)) == Modifier.PROTECTED);
}

private native Object getImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

private boolean samePackage(Class accessorClass) {
	if(declaringClass.equals(accessorClass)) {
		return true;
	}
	String fieldDeclaringClassPkg = getPackageName(declaringClass.getName());
	String callerClassPkg = getPackageName(accessorClass.getName());
	return fieldDeclaringClassPkg.equals(callerClassPkg);
}

private void doProtectedCheck(Class accessorClass, Object object) throws IllegalAccessException {
	if(!samePackage(accessorClass)) {
		if(!accessorClass.isInstance(object)) {
			if(!declaringClass.isInstance(object)) {
				throw new IllegalArgumentException();
			}
			throw new IllegalAccessException();
		}
	}
}

private static String getPackageName(String name) {
	int index = name.lastIndexOf('.');
	if (index >= 0) return name.substring(0, index);
	return "";
}

/**
 * Return the value of the field in the specified object as a boolean.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public boolean getBoolean(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getBooleanImpl(object);
}

private native boolean getBooleanImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the value of the field in the specified object as a byte.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public byte getByte(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getByteImpl(object);
}

private native byte getByteImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the value of the field in the specified object as a char.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public char getChar(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getCharImpl(object);
}

private native char getCharImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the java.lang.Class associated with the class that defined this field.
 *
 * @return		the declaring class
 */
public Class getDeclaringClass()
{
	return declaringClass;
}

/**
 * Return the value of the field in the specified object as a double.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public double getDouble(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getDoubleImpl(object);
}

private native double getDoubleImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the value of the field in the specified object as a float.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public float getFloat(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getFloatImpl(object);
}

private native float getFloatImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the value of the field in the specified object as an int.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public int getInt(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getIntImpl(object);
}

private native int getIntImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the value of the field in the specified object as a long.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 *
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public long getLong(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getLongImpl(object);
}

private native long getLongImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

/**
 * Return the modifiers for the modelled field.
 * The Modifier class should be used to decode the result.
 *
 * @return		the modifiers
 *
 * @see			java.lang.reflect.Modifier
 */
public native int getModifiers();

/**
 * Return the name of the modelled field.
 *
 * @return		the name
 */
public String getName()
{
	if(name != null) return name;
	return getNameImpl();
}

private native String getNameImpl();

/**
 * Return the value of the field in the specified object as a short.
 * This reproduces the effect of <code>object.fieldName</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * @param		object		the object to access
 * @return		the field value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public short getShort(Object object) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	return getShortImpl(object);
}

private native short getShortImpl(Object object) throws IllegalAccessException, IllegalArgumentException;

native String getSignature();

/**
 * Return the java.lang.Class associated with the type of this field.
 *
 * @return		the type
 */
public Class getType()
{
	if(type != null) return type;
	return getTypeImpl();
}

private native Class getTypeImpl();

/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 * <p>
 * The hash code for a Field is the hash code of the field's name.
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode()
{
	return getName().hashCode();
}

/**
 * Set the value of the field in the specified object to the boolean value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the field type is a base type, the value is automatically unwrapped. If the unwrap fails,
 * an IllegalArgumentException is thrown. If the value cannot be converted to the field type via
 * a widening conversion, an IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void set(Object object, Object value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setImpl(object, value);
}

private native void setImpl(Object object, Object value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the boolean value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setBoolean(Object object, boolean value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setBooleanImpl(object, value);
}

private native void setBooleanImpl(Object object, boolean value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the byte value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setByte(Object object, byte value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setByteImpl(object, value);
}

private native void setByteImpl(Object object, byte value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the char value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setChar(Object object, char value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setCharImpl(object, value);
}

private native void setCharImpl(Object object, char value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the double value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setDouble(Object object, double value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setDoubleImpl(object, value);
}

private native void setDoubleImpl(Object object, double value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the float value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setFloat(Object object, float value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setFloatImpl(object, value);
}

private native void setFloatImpl(Object object, float value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the int value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setInt(Object object, int value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setIntImpl(object, value);
}

private native void setIntImpl(Object object, int value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the long value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setLong(Object object, long value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setLongImpl(object, value);
}

private native void setLongImpl(Object object, long value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Set the value of the field in the specified object to the short value.
 * This reproduces the effect of <code>object.fieldName = value</code>
 * <p>
 * If the modelled field is static, the object argument is ignored.
 * Otherwise, if the object is null, a NullPointerException is thrown.
 * If the object is not an instance of the declaring class of the method, an
 * IllegalArgumentException is thrown.
 * <p>
 * If this Field object is enforcing access control (see AccessibleObject) and the modelled
 * field is not accessible from the current context, an IllegalAccessException is thrown.
 * <p>
 * If the value cannot be converted to the field type via a widening conversion, an
 * IllegalArgumentException is thrown.
 *
 * @param		object		the object to access
 * @param		value		the new value
 *
 * @throws		NullPointerException 		if the object is null and the field is non-static
 * @throws		IllegalArgumentException 	if the object is not compatible with the declaring class
 * @throws		IllegalAccessException 		if modelled field is not accessible
 */
public void setShort(Object object, short value) throws IllegalAccessException, IllegalArgumentException {
	if(requiresProtectedCheck()) {
		doProtectedCheck(getStackClass(-1), object);
	}
	setShortImpl(object, value);
}

private native void setShortImpl(Object object, short value) throws IllegalAccessException, IllegalArgumentException;

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 * <p>
 * The format of the string is:
 * <ul>
 * <li>modifiers (if any)
 * <li>return type
 * <li>declaring class name
 * <li>'.'
 * <li>field name
 * </ul>
 * <p>
 * For example:
 *		<code>public static java.io.InputStream java.lang.System.in</code>
 *
 * @return		a printable representation for the receiver
 */
public String toString()
{
	StringBuffer buf;
	Class current;
	int arity = 0;

	buf = new StringBuffer();
	String modifier = Modifier.toString(getModifiers());
	if (modifier.length() > 0) {
		buf.append(modifier);
		buf.append(" ");
	}

	current = getType();
	while(current.isArray())
	{
		current = current.getComponentType();
		arity++;
	}
	buf.append(current.getName());
	for(;arity > 0; arity--) buf.append("[]");

	buf.append(" ");
	buf.append(getDeclaringClass().getName());
	buf.append(".");
	buf.append(getName());
	return buf.toString();
}

}
