package java.lang.reflect;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2001, 2006  All Rights Reserved
 */

import java.util.*;
import java.lang.ref.WeakReference;
import com.ibm.oti.lang.reflect.ProxyClassFile;

/**
 * This class provides methods to creating dynamic proxy classes and instances.
 *
 * @version		initial
 * @see		java.lang.reflect.InvocationHandler
 */
public class Proxy implements java.io.Serializable {
private static final long serialVersionUID = -2222568056686623797L;

private static Map loaderCache = new WeakHashMap(); // maps class loaders to created classes by interface names
private static Map proxyCache = new WeakHashMap(); // to find previously created types
private static int NextClassNameIndex = 0;

protected InvocationHandler h;

private Proxy() {}

protected Proxy(InvocationHandler h) {
	this.h = h;
}

/**
 * Return the dynamically built class for the given interfaces, build a new
 * one when necessary. The order of the interfaces is important.
 *
 * The interfaces must be visible from the supplied class loader; no
 * duplicates are permitted. All non-public interfaces must be defined in
 * the same package.
 *
 * @version initial
 *
 * @param loader
 *            the class loader that will define the proxy class.
 * @param interfaces
 *            an array of <code>Class</code> objects, each one identifying
 *            an interface that the new proxy must implement.
 * @return a proxy class that implements all of the interfaces referred to
 *         in the contents of <code>interfaces</code>.
 * @exception IllegalArgumentException
 *                if any of the restrictions on the parameters are violated.
 * @exception NullPointerException
 *                if the interfaces or any of its elements are null.
 */
public static Class getProxyClass(ClassLoader loader, Class[] interfaces) throws IllegalArgumentException {
	// check that interfaces are a valid array of visible interfaces
	if (interfaces == null) throw new NullPointerException();
	String commonPackageName = null;
	for (int i = 0, length = interfaces.length; i < length; i++) {
		Class next = interfaces[i];
		if (next == null) throw new NullPointerException();
		String name = next.getName();
		if (!next.isInterface())
			// K00ed = {0} is not an interface
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00ed", name));
		if (loader != next.getClassLoader()) {
			try {
					if (next != Class.forName(name, false, loader))
						// K00ee = {0} is not visible from class loader
						throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00ee", name));
			} catch (ClassNotFoundException ex) {
				// K00ee = {0} is not visible from class loader
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00ee", name));
			}
		}
		for (int j = i+1; j < length; j++)
			if (next == interfaces[j])
				// K00ef = {0} appears more than once
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00ef", name));
		if (!Modifier.isPublic(next.getModifiers())) {
			int last = name.lastIndexOf('.');
			String p = last == -1 ? "" : name.substring(0, last);
			if (commonPackageName == null)
				commonPackageName = p;
			else if (!commonPackageName.equals(p))
				// K00f0 = non-public interfaces must be in the same package
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00f0"));
		}
	}

	// search cache for matching proxy class using the class loader
	synchronized (loaderCache) {
		Map interfaceCache = (Map) loaderCache.get(loader);
		if (interfaceCache == null)
			loaderCache.put(loader, (interfaceCache = new HashMap()));

		String interfaceKey = "";
		if (interfaces.length == 1) {
			interfaceKey = interfaces[0].getName();
		} else {
			StringBuffer names = new StringBuffer();
			for (int i = 0, length = interfaces.length; i < length; i++) {
				names.append(interfaces[i].getName());
				names.append(' ');
			}
			interfaceKey = names.toString();
		}

		Class newClass;
		WeakReference ref = (WeakReference)interfaceCache.get(interfaceKey);
		if (ref == null) {
			String nextClassName = "$Proxy" + NextClassNameIndex++;
			if (commonPackageName != null)
				nextClassName = commonPackageName + "." + nextClassName;
			byte[] classFileBytes = ProxyClassFile.generateBytes(nextClassName, interfaces);
			if (loader == null) loader = ClassLoader.getSystemClassLoader();
			newClass = defineClassImpl(loader, nextClassName.replace('.', '/'), classFileBytes);
			// Need a weak reference to the class so it can
			// be unloaded if the class loader is discarded
			interfaceCache.put(interfaceKey, new WeakReference(newClass));
			synchronized(proxyCache) {
				proxyCache.put(newClass, ""); // the value is unused
			}
		} else {
			newClass = (Class)ref.get();
		}
		return newClass;
	}
}

/**
 * Return an instance of the dynamically build class for the given
 * interfaces that forwards methods to the specified invocation handler.
 *
 * The interfaces must be visible from the supplied class loader; no
 * duplicates are permitted. All non-public interfaces must be defined in
 * the same package.
 *
 * @version initial
 *
 * @param loader
 *            the class loader that will define the proxy class.
 * @param interfaces
 *            the list of interfaces to implement.
 * @param h
 *            the invocation handler for the forwarded methods.
 * @return a a new proxy object that delegates to the handler <code>h</code>.
 * @exception IllegalArgumentException
 * @exception NullPointerException
 *                if the interfaces or any of its elements are null.
 */
public static Object newProxyInstance(ClassLoader loader, Class[] interfaces, InvocationHandler h) throws IllegalArgumentException {
	if (h != null) {
		try {
			return (Object) getProxyClass(loader, interfaces).
				getConstructor(new Class[] {InvocationHandler.class}).
				newInstance(new Object[] {h});
		} catch (NoSuchMethodException ex) {
			throw new InternalError(ex.toString());
		} catch (IllegalAccessException ex) {
			throw new InternalError(ex.toString());
		} catch (InstantiationException ex) {
			throw new InternalError(ex.toString());
		} catch (InvocationTargetException ex) {
			throw new InternalError(ex.getTargetException().toString());
		}
	} else throw new NullPointerException();
}

/**
 * Return whether the supplied class is a dynamically generated proxy class.
 *
 * @version		initial
 *
 * @param		cl	the class.
 * @return		true if the class is a proxy class and false otherwise.
 * @exception		NullPointerException	if the class is null.
 */
public static boolean isProxyClass(Class cl) {
	if (cl != null) {
		synchronized(proxyCache) {
			return proxyCache.containsKey(cl);
		}
	}
	throw new NullPointerException();
}

/**
 * Return the proxy instance's invocation handler.
 *
 * @version initial
 *
 * @param proxy
 *            the proxy instance.
 * @return the proxy's invocation handler object.
 * @exception IllegalArgumentException
 *                if the supplied <code>proxy</code> is not a proxy
 *                object.
 */
public static InvocationHandler getInvocationHandler(Object proxy) throws IllegalArgumentException {
	if (isProxyClass(proxy.getClass()))
		return ((Proxy) proxy).h;

	// K00f1 = not a proxy instance
	throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K00f1"));
}

private static native Class defineClassImpl(ClassLoader classLoader, String className, byte[] classFileBytes);

}
