package java.math;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * BigDecimal objects represent an arbitrary precisioned decimal
 * Number. They contain values that cannot be changed.
 * Thus, most operations on the BigDecimal object yield new
 * instances of BigDecimal.
 * <p>
 * BigDecimal is respresented by an unscaled BigInteger value and an integer
 * representing the scale of the object.  The scale of the BigDecimal
 * is the number of digits after the decimal point.  Eg. 1.234 would
 * have a scale of 3 and an unscaled value of 1234.  Therefore,
 * decimal representation of a BigDecimal is BigIntegerValue/10^scale.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			java.math.BigInteger
 */

public class BigDecimal extends java.lang.Number implements java.lang.Comparable{
 	private static final long serialVersionUID = 6108874887143696463L;

/**
  * Rounding mode Constants
  */
	public final static int ROUND_UP = 0;
	public final static int ROUND_DOWN = 1;
	public final static int ROUND_CEILING = 2;
	public final static int ROUND_FLOOR = 3;
	public final static int ROUND_HALF_UP = 4;
	public final static int ROUND_HALF_DOWN = 5;
	public final static int ROUND_HALF_EVEN = 6;
	public final static int ROUND_UNNECESSARY = 7;

/**
  * BigInteger constants
  */
	private final static BigInteger TEN = new BigInteger("10");
	private final static BigInteger TWO = BigInteger.valueOf(2);
	private final static BigDecimal ONE = new BigDecimal(BigInteger.ONE,0);
	private final static BigDecimal ZERO = new BigDecimal(BigInteger.ZERO,0);
/**
  * number of digits after the decimal.  Zero included
  */
	private int scale;

/**
  * unscaled representation of the number
  */
	private BigInteger intVal;

/**
  * number of times U and K have been set or fetched.
  * A number is represented by ((sum of) U * 10^K).
  */
	private transient int setCount = 0;
	private transient int getCount = 0;

/**
  * an array storing all the Us
  */
	private transient int[] uArray =  new int[64];

/**
  * value of the first K
  */
	private transient int firstK = 0;

/**
 * Constructs a BigDecimal with unscaled value initialized as
 * bval and scale as 0.
 *
 * @author		OTI
 * @version		initial
 */
public BigDecimal(BigInteger bval){
	intVal = bval;
	scale = 0;
}

/**
 * Constructs a BigDecimal with unscaled value initialized as
 * bval and scale as scale from the argument.
 *
 * @author		OTI
 * @version		initial
 */
public BigDecimal(BigInteger bval, int sc){
	if (sc < 0) throw new NumberFormatException();
	intVal = bval;
	scale = sc;
}

/**
 * Constructs a BigDecimal with a double value
 * as an arugment.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	NumberFormatException
 *					If the is Infinity, Negative Infinity or NaN.
 */
public BigDecimal(double bval){
	final int p = 1023 + 52; // the power offset (precision)
	final long signMask = 0x8000000000000000L; // the mask to get the sign of the number
	final long eMask    = 0x7FF0000000000000L; // the mask to get the power bits
	final long fMask    = 0x000FFFFFFFFFFFFFL; // the mask to get the significand bits

	long inputNumberBits = Double.doubleToLongBits (bval);
	long sign = (inputNumberBits & signMask) >> 63;	// the value of the sign--0 is positive, 1 is negative
	long e    = (long)((inputNumberBits & eMask) >> 52);	// the value of the 'exponent bigs' of the val
	long f    = (long)(inputNumberBits & fMask);		// the value of the 'mantissa bits' of the val
	int pow = 0;
	boolean zero = false;
	String signString;
	setCount = 0;
 	getCount = 0;
	if (sign == 0)
		signString = "";
	else
		signString = "-";

	if (e == 2047)
		//number is infinity or not a number
		throw new NumberFormatException();

	if (e == 0) {
		if (f == 0){
			this.intVal = BigInteger.ZERO;
			this.scale = 0;
			zero = true;
		}else
			pow = (int)(e - p + 1);	// a denormalized number
	} else {
		// 0 < e < 2047
		// a "normalized" number
		f = f | 0x0010000000000000L;
		pow = (int)(e - p);
	}
	if (!zero){
		String finalVal = signString + freeFormat (pow, f);
		this.intVal = new BigInteger(finalVal.toString());
	}
}

/**
 * Constructs a BigDecimal from the strong which can only contan digits of 0-9, a
 * decimal point and a negative sign.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	NumberFormatException
 *					If the argument contained characters other than digits.
 */
public BigDecimal(String bval){
	if (bval.length() > 0 && bval.charAt(0) == '+')
		bval = bval.substring(1, bval.length());
	int expIndex = bval.indexOf('e', 0);
	if (expIndex == -1) expIndex = bval.indexOf('E', 0);
	String expString = null;
	if (expIndex >= 0) {
		expString = new String(bval).substring(expIndex + 1, bval.length());
		bval = bval.substring(0, expIndex);
	}
	int period = bval.indexOf('.', 0);
	if (period == -1) {
		intVal = new BigInteger(bval);
		scale = 0;
	} else {
		int max = bval.length();
		String sub1 = bval.substring(0, period);
		String sub2 = bval.substring(period + 1, max);
		intVal = new BigInteger(sub1 + sub2);
		scale = max - period - 1;
	}
	if (expString != null) {
		if (expString.length() > 0 && expString.charAt(0) == '+') {
			expString = expString.substring(1, expString.length());
			if (expString.indexOf('-') >= 0)
				throw new NumberFormatException();
		}

		Integer expVal = new Integer(expString);
		scale -= expVal.intValue();
		for (; scale < 0; scale++) // same as multiply by 10
			intVal = intVal.shiftLeft(1).add(intVal.shiftLeft(3));
	}
}

/**
 * Answers the absolute value of this BigDecimal.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		absolute value of the receiver.
 */
public BigDecimal abs(){
	return new BigDecimal(unscaledValue().abs(),scale());
}

/**
 * Answers the sum of the receiver and argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal 	The sum of adding two BigDecimal.
 */
public BigDecimal add(BigDecimal bval){
	int bvalScale = bval.scale();
	BigInteger bvalintValue = bval.unscaledValue();
	if (scale == bvalScale)
		return new BigDecimal(intVal.add(bvalintValue),scale);

	int maxScale = (scale > bvalScale) ? scale: bvalScale;
	BigInteger decRes;
	if (maxScale == scale){
		BigInteger scaleNum = TEN.pow(maxScale - bvalScale);
		decRes = intVal.add(bvalintValue.multiply(scaleNum));
	}else{
		BigInteger scaleNum = TEN.pow(maxScale - scale);
		decRes = bvalintValue.add(intVal.multiply(scaleNum));
	}
	return new BigDecimal(decRes,maxScale);
}

/**
 * Compares an receiver to the argument Object.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int			0 - equal; 1 - this > val; -1 - this < val
 * @exception	ClassCastException
 *							if the argument is not of type BigDecimal
 */
public int compareTo(Object o){
	if (!(o instanceof BigDecimal))
		throw new ClassCastException();
	return compareTo((BigDecimal)o);
}

/**
 * Compares the receiver BigDecimal and argument BigDecimal
 * e.x 1.00 & 1.0 will return 0 in compareTo.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int 		0 - equal; 1 - this > val; -1 - this < val.
 */
public int compareTo(BigDecimal bval){
	BigDecimal result = this.subtract(bval);
	return result.signum();
}

/**
 * Generates and formats digits of a nonExponential value.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String		String representation of the formated number.
 */
private String freeFormat (int pow, long f) {
	char[] formattedDecimal = new char[309];
	int charPos = 0;
	int U;
	int k;
	int oldScale = 0;
	digitGenerator (pow, f);
	boolean allFraction = false;

	if (getCount < setCount)
		U = uArray[getCount];
	else
		U = -1;
	k = firstK - getCount;
	getCount++;

	if (k < 0){
		allFraction = true;
		oldScale = (-k) - 1;
	}
	do {
		while (k >= 0){
			if (U != -1){
				formattedDecimal[charPos] = Character.forDigit(U,10);
				charPos++;
			} else
				if (k >= -1) {
					formattedDecimal[charPos] = '0';
					charPos++;
				}
			if (getCount < setCount)
				U = uArray[getCount];
			else
				U = -1; // nothing significant to print
			k = firstK - getCount;
			getCount++;
		}

		if (U != -1) {
			formattedDecimal[charPos] = Character.forDigit(U,10);
			charPos++;
			oldScale++;
		}
		if (getCount < setCount)
			U = uArray[getCount];
		else
			U = -1; // nothing significant to print
		k = firstK - getCount;
		getCount++;

	} while (U != -1 || k >= -1);
	if (oldScale != 0 && !allFraction)
		oldScale = (-k) - 1;

	scale = oldScale;
	return new String(formattedDecimal,0,charPos);
}

/**
 * Generate the floating-point with precision based on the
 * Fixed-Precision Positive Floating-point algorithm.
 *
 * @author		OTI
 * @version		initial
 */
private void digitGenerator (int pow, long f) {
	boolean low, high;
	int U, k;
	BigInteger M_minus, M_plus, R, S;
	BigInteger cache, cache2; // used to store values that come up multiple times to increase efficiency

	R = BigInteger.valueOf(f).shiftLeft (Math.max(pow, 0));
	S = BigInteger.ZERO.setBit(Math.max(0, -pow));
	M_minus = BigInteger.ZERO.setBit(Math.max (pow, 0));
	M_plus = M_minus.abs(); // just a copy since M_minus > 0;

	k = 0;

	while ((cache = R.multiplyByTen()).compareTo(S) < 0) {
		k--;
		R = cache;
		M_minus = M_minus.multiplyByTen();
		M_plus =M_plus.multiplyByTen();
	}

	cache = R.shiftLeft(1).add(M_plus);
	while (cache.compareTo(S.shiftLeft(1)) >= 0) {
		S = S.multiplyByTen();
		k++;
	}

	firstK = k - 1;
	cache = S.shiftLeft(1);
	while (true) {
		k--;
		BigInteger a = R.multiplyByTen();
		BigInteger remainder;
		int digit = 0;
		for (int i = 3; i >= 0; i--) {
			remainder = a.subtract(S.shiftLeft(i));
			if (remainder.compareTo(BigInteger.ZERO) >= 0) {
				a = remainder;
				digit += (1 << i);
			}
		}
		U = digit;
		R = a;

		M_minus = M_minus.multiplyByTen();
		M_plus = M_plus.multiplyByTen();
		cache2 = R.shiftLeft(1);
		low = cache2.compareTo(M_minus) < 0;
		high = cache2.compareTo(cache.subtract(M_plus)) > 0;

		if (low || high)
			break;

		uArray [setCount] = U;
		setCount++;
	}
	if (low && !high){
		uArray [setCount] = U;
		setCount++;
	}else
		if (high && !low){
			uArray [setCount] = U+1;
			setCount++;
		}else
			if (cache2.compareTo(S) <= 0){
				uArray [setCount] = U;
				setCount++;
			}else{
				uArray [setCount] = U + 1;
				setCount++;
			}
}

/**
 * Answers the result of (this / val).
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		result of this/val.
 */
public BigDecimal divide(BigDecimal bval, int roundingMode){
	return divide(bval,scale,roundingMode);
}

/**
 * Answers the result of (this / val) and whose scale is specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal				result of this/val.
 * @exception 	ArithmeticException		division by zero.
 * @exception 	IllegalArgumentException
 *					roundingMode is not valid.
 */
public BigDecimal divide(BigDecimal bval, int bscale, int roundingMode){
	if (bscale < 0)
		throw new ArithmeticException();
	if (roundingMode < ROUND_UP || roundingMode > ROUND_UNNECESSARY)
		throw new IllegalArgumentException();

	BigInteger divider, result[];
	int bvalScale = bval.scale();
	int sign = intVal.signum() * bval.intVal.signum();

	if (bvalScale < scale){
		BigDecimal newVal = bval.movePointRight(scale);
		divider = newVal.abs().intVal;
		result = intVal.abs().divideAndRemainder(divider);
	} else {
		divider = bval.abs().intVal;
		if (bvalScale > scale){
			BigDecimal newThis = movePointRight(bvalScale);
			result = newThis.abs().intVal.divideAndRemainder(divider);
		} else
			result = intVal.abs().divideAndRemainder(divider);
	}
	int i = bscale;
	BigInteger finalRem = result[1];
	StringBuffer finalVal = new StringBuffer();
	if (sign < 0) finalVal.append('-');
	finalVal.append(result[0].toString());
	// Loop an extra time to get the first rounding digit
	while (i >= 0 && !result[1].equals(BigInteger.ZERO)) {
		BigInteger divisor = result[1].multiplyByTen();
		result = divisor.divideAndRemainder(divider);
		if (i > 0) finalVal.append(result[0].toString());
		if (i == 1) finalRem = result[1];
		i--;
	}
	boolean round = i < 0;
	if (round && roundingMode == ROUND_UNNECESSARY)
		throw new ArithmeticException();
	while (i-- > 0) finalVal.append('0');
	BigInteger quotient = new BigInteger(finalVal.toString());
	if (round) {
		quotient = roundValues(quotient, finalRem, result, sign, roundingMode);
	}
	return new BigDecimal(quotient, bscale);
}

/**
 * Converts this BigDecimal to a double.  If magnitude of the BigDecimal
 * value is larger than what can be represented by a double, either
 * Infinity or -Infinity is returned.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		double		the value of the receiver.
 */
public double doubleValue(){
	try{
		Double db = new Double(this.toString());
		return db.doubleValue();
	}catch(NumberFormatException e){
		if (this.signum() == -1)
			return Double.NEGATIVE_INFINITY;
		else
			return Double.POSITIVE_INFINITY;
	}
}
/*
 * Compares the argument to the receiver, and answers true
 * if they represent the <em>same</em> object using a class
 * specific comparison. The implementation in Object answers
 * true only if the argument is the exact same object as the
 * receiver (==).
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o Object
 *					the object to compare with this object.
 * @return		boolean
 *					<code>true</code>
 *						if the object is the same as this object
 *					<code>false</code>
 *						if it is different from this object.
 * @see			hashCode
 */
public boolean equals(Object obj) {
	if (!(obj instanceof BigDecimal))
		return false;
	return ((unscaledValue().compareTo(((BigDecimal)obj).unscaledValue()) == 0) && scale() == ((BigDecimal)obj).scale());
}

/**
 * Converts this BigDecimal to a float.If magnitude of the BigDecimal
 * value is larger than what can be represented by a float, either
 * Infinity or -Infinity is returned.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		float		the value of the receiver.
 */
public float floatValue(){
	try{
		Float fl = new Float(this.toString());
		return fl.floatValue();
	}catch(NumberFormatException e){
		if (this.signum() == -1)
			return Float.NEGATIVE_INFINITY;
		else
			return Float.POSITIVE_INFINITY;
	}

}

 /**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>.equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int
 *					the receiver's hash.
 *
 * @see			#equals(Object)
 */
public int hashCode() {
	return (intVal.hashCode() ^ scale);

}

/**
 * Converts this BigDecimal to an int.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the value of the receiver.
 */
public int intValue(){
	BigInteger intVal = this.toBigInteger();
	return intVal.intValue();
}

/**
 * Converts this BigDecimal to a long.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		long	long representation of the receiver.
 */
public long longValue(){
	BigInteger longValue = this.toBigInteger();
	return longValue.longValue();
}

/**
 * Answers the max value between the receiver and this BigDecimal.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		max BigDecimal.
 */
public BigDecimal max(BigDecimal bval){
	return this.compareTo(bval) > 0 ? this : bval;
}

/**
 * Answers the min value between the receiver and argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		max BigDecimal.
 */
public BigDecimal min(BigDecimal bval){
	return this.compareTo(bval) >= 0 ? bval: this;
}

/**
 * Moves the decimal point of this BigDecimal n places to the left.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					new BigDecimal with decimal moved n places to the left.
 */
public BigDecimal movePointLeft(int n){
	if (n < 0)
		return this.movePointRight(-n);
	return new BigDecimal(unscaledValue(),scale+n);
}

/**
 * Moves the decimal point of this BigDecimal n places to the right.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					new BigDecimal with decimal moved n places to the right.
 */
public BigDecimal movePointRight(int n){
	if (n < 0)
		return this.movePointLeft(-n);
	if (scale-n < 0){
		StringBuffer stBuf = new StringBuffer().append(intVal);
		for (int i = 0; i<n-scale;i++)
			stBuf.append("0");
		return new BigDecimal(new BigInteger(stBuf.toString()),0);
	}
	else
		return new BigDecimal(intVal,scale-n);
}

/**
 * Answers the multiplication result of the receiver and argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					result of multiplying two bigDecimals.
 */
public BigDecimal multiply(BigDecimal bval){
	BigInteger res = intVal.multiply(bval.unscaledValue());
	int resScale = scale() + bval.scale();
	return new BigDecimal(res,resScale);
}

/**
 * Negates this BigDecimal value.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					new BigDecimal with value negated.
 */
public BigDecimal negate(){
	return new BigDecimal(intVal.negate(),scale);
}

/**
 * Returns the scale of this BigDecimal.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int 	scale value.
 */
public int scale(){
	return scale;
}

/**
 * Sets the scale of this BigDecimal.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					a BigDecimal with the same value, but specified scale.
 */
public BigDecimal setScale(int newScale){
	return setScale(newScale,ROUND_UNNECESSARY);
}

/**
 * Sets the scale of this BigDecimal.  The unscaled value
 * is determined by the rounding Mode
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal
 *					a BigDecimal with the same value, but specified cale.
 * @exception 	ArithmeticException
 *					rounding mode must be specified if lose of precision due to setting scale.
 * @exception	IllegalArgumentException
 *					invalid rounding mode
 */
public BigDecimal setScale(int newScale, int roundingMode){
	if (newScale < 0) throw new ArithmeticException();
	if (roundingMode < ROUND_UP || roundingMode > ROUND_UNNECESSARY)
		throw new IllegalArgumentException();

	if (scale < newScale){
		StringBuffer stBuf = new StringBuffer("1");
		for (int i = 0; i<newScale-scale;i++) stBuf.append('0');
		BigInteger multiplier = new BigInteger(stBuf.toString());
		BigInteger newValue = intVal.multiply(multiplier);
		return new BigDecimal(newValue, newScale);
	}else if (scale > newScale){
		//split up the unscaled value into two parts
		BigInteger divisor = TEN.pow(scale-newScale);
		BigInteger[] result = intVal.divideAndRemainder(divisor);
		if (!result[1].equals(BigInteger.ZERO)) {
			if (roundingMode == ROUND_UNNECESSARY) {
				// K004f = rounding necessary
				throw new ArithmeticException(com.ibm.oti.util.Msg.getString("K004f"));
			}
			BigInteger[] furRes = null;
			if (roundingMode >= ROUND_HALF_UP) {
				divisor = TEN.pow(scale-newScale-1);
				furRes = result[1].divideAndRemainder(divisor);
				furRes[0] = furRes[0].abs();
			}
			result[0] = roundValues(result[0], result[1], furRes, intVal.signum(), roundingMode);
		}
		return new BigDecimal(result[0],newScale);
	}else
		return this;
}
/**
 * This method takes the number to round, the remainder, an array containing
 * the positive first digit of the remainder and the remaining remainder, and
 * the sign of the result. The sign of the number to round cannot be used since
 * it may be zero. Round the number according to the rounding rules specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger
 *					Returns BigInteger value that has been rounded correctly
 * @exception	IllegalArgumentException
 *					not valid rounding mode
 */

private static BigInteger roundValues(BigInteger num, BigInteger rem, BigInteger[] first, int sign, int mode){
	int digit;
	switch (mode){
	 	case(BigDecimal.ROUND_CEILING) : {
	 			if (sign >= 0 && !rem.equals(BigInteger.ZERO))
	 				return num.add(BigInteger.ONE);
	 			else
	 				return num;
	 	}
		case(BigDecimal.ROUND_DOWN)	: return num;
		case(BigDecimal.ROUND_FLOOR) : {
				if (sign < 0 && !rem.equals(BigInteger.ZERO))
					return num.add(BigInteger.ONE.negate());
				else
				 	return num;
		}
		case(BigDecimal.ROUND_HALF_DOWN) : {
				digit = first[0].intValue();
				if (digit > 5 || (digit == 5 &&
					!(first[1].equals(BigInteger.ZERO))))
				{
					if (sign >= 0)
						return num.add(BigInteger.ONE);
					else
						return num.add(BigInteger.ONE.negate());
				}else
				  	return num;
		}
		case(BigDecimal.ROUND_HALF_EVEN) : {
				digit = first[0].intValue();
				if (digit > 5 || (digit == 5 &&
					!(first[1].equals(BigInteger.ZERO))))
				{
					if (sign >= 0)
						return num.add(BigInteger.ONE);
					else
						return num.add(BigInteger.ONE.negate());
				} else if (digit == 5){
					BigInteger remainder = num.mod(BigInteger.ONE.add(BigInteger.ONE));
					if (remainder.equals(BigInteger.ZERO))
						return num;
					else{
						if (sign >= 0)
							return num.add(BigInteger.ONE);
						else
							return num.add(BigInteger.ONE.negate());
					}
				} else
					return num;
		}
		case(BigDecimal.ROUND_HALF_UP) : {
				if (first[0].intValue() >= 5){
					if (sign >= 0)
						return num.add(BigInteger.ONE);
					else
						return num.add(BigInteger.ONE.negate());
				}else
				  	return num;
		}
		case(BigDecimal.ROUND_UP) : {
				if (!rem.equals(BigInteger.ZERO)){
				   	if (sign >= 0)
						return num.add(BigInteger.ONE);
					else
						return num.add(BigInteger.ONE.negate());
				}else
				   	return num;
		}
		// K0050 = wrong rounding mode
		default : throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0050"));
	}
}
/**
 * Answers the signum function of this instance.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		-1, 0, or 1 if the receiver is negative, zero, or positive.
 */
public int signum(){
	return intVal.signum();
}

/**
 * Answers the subtract result of the receiver and argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal 	The result of adding two BigDecimal.
 */
public BigDecimal subtract(BigDecimal bval){
	int bvalScale = bval.scale();
	BigInteger bvalintValue = bval.unscaledValue();
	BigInteger decRes;
	if (scale == bvalScale){
		decRes = intVal.subtract(bvalintValue);
		if (decRes.equals(BigInteger.ZERO))
			bvalScale = 0;
		return new BigDecimal(decRes,bvalScale);
	}
	int maxScale = (scale > bvalScale) ? scale: bvalScale;
	if (maxScale == scale){
		BigInteger scaleNum = TEN.pow(maxScale - bvalScale);
		decRes = intVal.subtract(bvalintValue.multiply(scaleNum));
	}else{
		BigInteger scaleNum = TEN.pow(maxScale - scale);
		decRes = intVal.multiply(scaleNum).subtract(bvalintValue);
	}
	if (decRes.equals(BigInteger.ZERO))
		maxScale = 0;
	return new BigDecimal(decRes,maxScale);
}

/**
 * Converts this to a BigInteger.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	BigDecimal equivalent of bigInteger.
 */
public BigInteger toBigInteger(){
	BigInteger divisor = TEN.pow(scale);
	return (intVal.divide(divisor));
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					a printable representation for the receiver.
 */
public String toString() {
	String unscaledChar = intVal.toString();
	int max = unscaledChar.length();
	boolean negative = unscaledChar.charAt(0) == '-';
	StringBuffer stBuf = new StringBuffer();
	if (scale >= (negative ? max - 1 : max)){
		if (negative){
			stBuf = new StringBuffer().append("-0.");
			String temp = new String(unscaledChar).substring(1,max);
			unscaledChar = temp;
			max = unscaledChar.length();
		}else
			stBuf = new StringBuffer().append("0.");
		for (int i = 0; i<scale-max; i++)
			stBuf = stBuf.append("0");
		stBuf = stBuf.append(unscaledChar);
		return stBuf.toString();
	} else {
		String sub1 = new String(unscaledChar).substring(0,max-scale);
		if (scale == 0)
			return sub1;
		String sub2 = new String(unscaledChar).substring(max-scale(),max);
		return (sub1 + "." + sub2);
	}

}

/**
 * Returns an unscaled value of this BigDecimal.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger		The unscaled value.
 */
public BigInteger unscaledValue(){
	return intVal;
}

/**
 * Translate long value into a BigDecimal with scale of zero.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		BigDecimal equivalence of a long value.
 */
public static BigDecimal valueOf(long bval){
	return new BigDecimal(BigInteger.valueOf(bval),0);
}

/**
 * Translate long unscaled value into a BigDecimal specified by the scale.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigDecimal		BigDecimal equalvalence of a long value.
 * @exception 	NumberFormatException
 *					the scale value is < 0;
 */
public static BigDecimal valueOf(long bval, int scale){
	if (scale < 0) {
		// K0051 = scale value < than zero
		throw new NumberFormatException(com.ibm.oti.util.Msg.getString("K0051"));
	}
	BigInteger bvalBigInteger = BigInteger.valueOf(bval);
	if (bvalBigInteger == BigInteger.ZERO)
		return BigDecimal.ZERO;
	if (bvalBigInteger == BigInteger.ONE)
		return BigDecimal.ONE;
	return new BigDecimal(bvalBigInteger, scale);
}

}
