package java.net;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class models a socket for sending & receiving
 * datagram packets.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			DatagramPacket
 */

import java.io.IOException;
import java.net.SocketOptions;
import java.security.*;
import com.ibm.oti.util.PriviAction;

/**
 * This class represents sockets to be used in datagram oriented
 * (non-streaming) protocols.
 *
 * @author		OTI
 * @version		initial
 */
public class DatagramSocket {

	DatagramSocketImpl impl;
	InetAddress address;
	int port = -1;
	static DatagramSocketImplFactory factory;
	boolean isBound = false;
	private  boolean isConnected = false;
	private  boolean isClosed = false;
	Object lock = new Object();

/**
 * Constructs a datagram socket, bound to any available port
 * on the localhost.
 *
 * @throws 		SocketException if a problem occurs creating or binding the socket
 */
public DatagramSocket() throws SocketException {
	this(0);
}

/**
 * Answers a datagram socket, bound to the nominated port
 * on the localhost.
 *
 * @param		aPort	the port to bind on the localhost
 *
 * @throws 		SocketException if a problem occurs creating or binding the socket
 */
public DatagramSocket(int aPort) throws SocketException {
	super();
	checkListen(aPort);
	createSocket(aPort, InetAddress.ANY);
}

/**
 * Constructs a datagram socket, bound to the nominated localhost/port.
 *
 * @param		aPort	the port on the localhost to bind
 * @param		addr	the address on the multihomed localhost to bind
 *
 * @throws 		SocketException if a problem occurs creating or binding the socket
 */
public DatagramSocket(int aPort, InetAddress addr) throws SocketException {
	super();
	checkListen(aPort);
	createSocket(aPort, null == addr ? InetAddress.ANY : addr);
}

/**
 * Sent prior to attempting to bind the socket, check that the port
 * is within the valid port range and verify with the security manager
 * that the port may be bound by the current context.
 *
 * @param		aPort	the port on the localhost that is to be bound
 */
void checkListen(int aPort) {
	if (aPort < 0 || aPort > 65535)
		// K0325 = Port out of range: {0}
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0325", aPort));
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkListen(aPort);
}

/**
 * Close the socket.
 */
// In the documentation jdk1.1.7a/guide/net/miscNet.html, this method is noted as not
// being synchronized.
public void close() {
	isClosed = true;
	impl.close();
}

/**
 * Connect the datagram socket to a remote host and port.  The host and port
 * are validated, thereafter the only validation on send() and receive() is
 * that the packet address/port matches the connected target.
 *
 * @param		anAddress	the target address
 * @param		aPort	the target port
 */
public void connect(InetAddress anAddress, int aPort) {
	if (anAddress == null || aPort < 0 || aPort > 65535)
		// K0032 = Address null or destination port out of range
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0032"));

	synchronized(lock){

		if (isClosed()) return;
		try {
			checkClosedAndBind(true);
		} catch (SocketException e) {}

		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			if (anAddress.isMulticastAddress())
				security.checkMulticast(anAddress);
			else
				security.checkConnect(anAddress.getHostName(), aPort);
		}

		try {;
			impl.connect(anAddress, aPort);
		}
		catch (SocketException e) {
			// not connected at the native level just do what we did before
		}
		address = anAddress;
		port = aPort;
		isConnected = true;
	}
}

/**
 * 'Disconnect' the datagram socket from a remote host and port.
 * This method may be called on an unconnected socket.
 */
public void disconnect() {
	if (isClosed() || !isConnected()) return;
	impl.disconnect();
	address = null;
	port = -1;
	isConnected = false;
 }

synchronized void createSocket(int aPort, InetAddress addr) throws SocketException {
	impl = factory != null ? factory.createDatagramSocketImpl() : createSocketImpl();
	impl.create();
	try {
		impl.bind(aPort, addr);
		isBound = true;
	} catch (SocketException e) {
		close();
		throw e;
	}
}

/**
 * Answer a concrete instance of DatagramSocketImpl, either as declared
 * in the system properties or the default, PlainDatagramSocketImpl.  The
 * latter does not support security checks.
 *
 * @return		DatagramSocketImpl	the concrete instance
 *
 * @exception	SocketException	if an error occurs during the instantiation of
 *								a type declared in the system properties
 */
DatagramSocketImpl createSocketImpl() throws SocketException {
	Object impl = null;
	String prefix;
	prefix = (String)AccessController.doPrivileged(new PriviAction("impl.prefix", "Plain"));
	try {
		Class aClass = Class.forName("java.net." + prefix + "DatagramSocketImpl");
		impl = aClass.newInstance();
	} catch (Exception e) {
 		// K0033 = Unknown socket type
 		throw new SocketException(com.ibm.oti.util.Msg.getString("K0033"));
	}
	return (DatagramSocketImpl)impl;
}

/**
 * Answer the remote address to which the socket is connected.
 * Answer null if the socket is not connected.
 *
 * @return		the remote address to which the socket is connected
 */
public InetAddress getInetAddress() {
	return address;
}

/**
 * Answer the local address to which the socket is bound.
 * This address may be any of the addresses on a multihomed machine
 * if the security policy is not satisfied.
 *
 * @return		the local address to which the socket is bound
 */
public InetAddress getLocalAddress() {
	if (isClosed()) {
		return null;
	}
	if (!isBound()) {
		return InetAddress.ANY;
	}
	InetAddress anAddr = impl.getLocalAddress();
	try {
		SecurityManager security = System.getSecurityManager();
		if (security != null)
			security.checkConnect(anAddr.getHostName(), -1);
	} catch (SecurityException e) {
		return InetAddress.ANY;
	}
	return anAddr;
}

/**
 * Answer the local port to which the socket is bound.
 *
 * @return	int		local port to which the socket is bound
 */
public int getLocalPort() {
	if (isClosed()) {
		return -1;
	}
	if (!isBound()) {
		return 0;
	}
	return impl.getLocalPort();
}

/**
 * Answer the remote port to which the socket is connected.
 * Answer -1 if the socket is not connected.
 *
 * @return	int		remote port to which the socket is connected
 */
public int getPort() {
	return port;
}

boolean isMulticastSocket() {
	return false;
}

/**
 * Answer the socket receive buffer size (SO_RCVBUF).
 *
 * @return	int		socket receive buffer size
 *
 * @exception SocketException when an error occurs
 */
public synchronized int getReceiveBufferSize() throws SocketException {
	checkClosedAndBind(false);
	return ((Integer)impl.getOption(SocketOptions.SO_RCVBUF)).intValue();
}

/**
 * Answer the socket send buffer size (SO_SNDBUF).
 *
 * @return	int		socket send buffer size
 *
 * @exception SocketException when an error occurs
 */
public synchronized int getSendBufferSize() throws SocketException {
	checkClosedAndBind(false);
	return ((Integer)impl.getOption(SocketOptions.SO_SNDBUF)).intValue();
}

/**
 * Answer the socket receive timeout (SO_RCVTIMEOUT), in milliseconds.
 * Zero implies the timeout is disabled.
 *
 * @return	int		socket receive timeout
 *
 * @exception SocketException when an error occurs
 */
public synchronized int getSoTimeout() throws SocketException {
	checkClosedAndBind(false);
	return ((Integer)impl.getOption(SocketOptions.SO_TIMEOUT)).intValue();
}

/**
 * Receive on this socket into the packet argument.  This method blocks
 * until a packet is received or, if a timeout has been defined, the timeout
 * period expires.  If this is a connected socket, the packet host/port are
 * compared to the connection host/port otherwise the security manager if present
 * is queried whether the packet host/port is acceptable.
 * Any packets from unacceptable origins will be silently discarded.
 * The packet fields are set according to the data received.
 * If the received data is longer than the packet buffer, it is truncated.
 *
 * @param		pack		the DatagramPacket to receive data into
 *
 * @exception	java.io.IOException		If a receive error occurs.
 */
public synchronized void receive(DatagramPacket pack) throws IOException {
	checkClosedAndBind(true);

	boolean secure = true;

	InetAddress senderAddr = null;

	int senderPort = 0;
	DatagramPacket tempPack = new DatagramPacket(new byte[1], 1);
	boolean copy = false;

	SecurityManager security = System.getSecurityManager();
	if (address != null || security != null) { // The socket is connected
		// Check pack before peeking
		if (pack == null) throw new NullPointerException();
		secure = false;
		while (!secure) {
			copy = false;
			try {
				senderPort = impl.peekData(tempPack);
				senderAddr = tempPack.getAddress();
			} catch (SocketException e) {
				if (e.getMessage().equals("The socket does not support the operation")) {
					tempPack = new DatagramPacket(new byte[pack.length], pack.getLength());
					impl.receive(tempPack);
					senderAddr = tempPack.getAddress();
					senderPort = tempPack.getPort();
					copy = true;
				} else throw e;
			}
			if (address == null) {
				try {
					security.checkAccept(senderAddr.getHostName(), senderPort);
					if (!copy) secure = true;
					break;
				} catch (SecurityException e) {
					if (!copy) {
						if (tempPack == null)
							tempPack = new DatagramPacket(new byte[pack.length], pack.length);
						impl.receive(tempPack);
					}
				}
			} else
			if (port == senderPort && address.equals(senderAddr)) {
				if (!copy) secure = true;
				break;
			} else if (!copy) {
				if (tempPack == null)
					tempPack = new DatagramPacket(new byte[pack.length], pack.length);
				impl.receive(tempPack);
			}
		}
	}
	if (copy) {
		System.arraycopy(tempPack.getData(), 0, pack.getData(), pack.getOffset(), tempPack.getLength());
		pack.setLength(tempPack.getLength());
		pack.setAddress(tempPack.getAddress());
		pack.setPort(tempPack.getPort());
	}
	if (secure){
		impl.receive(pack);
	}
}

/**
 * Send the packet on this socket.  The packet must satisfy the
 * security policy before it may be sent.
 *
 * @param		pack		the DatagramPacket to send
 *
 * @exception	java.io.IOException		If a send error occurs.
 */
public void send(DatagramPacket pack) throws IOException {
	checkClosedAndBind(true);

	InetAddress packAddr = pack.getAddress();
	if (address != null) { // The socket is connected
		if (packAddr != null) {
			if (!address.equals(packAddr) || port != pack.getPort())
				// K0034 = Packet address mismatch with connected address
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0034"));
		} else {
			pack.setAddress(address);
			pack.setPort(port);
		}
	} else {
		// not connected so the target address cannot be null
		if (packAddr == null){
			// K0331 = address is null
			throw new NullPointerException(com.ibm.oti.util.Msg.getString("K0331"));
		}
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			if (packAddr.isMulticastAddress())
				security.checkMulticast(packAddr);
			else
				security.checkConnect(packAddr.getHostName(), pack.getPort());
		}
	}
	impl.send(pack);
}

/**
 * Set the socket send buffer size.
 *
 * @param		size	the buffer size, in bytes
 *
 * @exception	java.net.SocketException If an error occurs while setting
 * the size or the size is invalid.
 */
public synchronized void setSendBufferSize(int size) throws SocketException {
	if (size >= 1) {
		checkClosedAndBind(false);
		impl.setOption(SocketOptions.SO_SNDBUF, new Integer(size));
	// K0035 = Zero or negative buffer size
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0035"));
}

/**
 * Set the socket receive buffer size.
 *
 * @param		size	the buffer size, in bytes
 *
 * @exception	java.net.SocketException If an error occurs while setting
 * the size or the size is invalid.
 */
public synchronized void setReceiveBufferSize(int size) throws SocketException {
	if (size >= 1) {
		checkClosedAndBind(false);
		impl.setOption(SocketOptions.SO_RCVBUF, new Integer(size));
	// K0035 = Zero or negative buffer size
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0035"));
}

/**
 * Set the SO_RCVTIMEOUT to <code>timeout</code>, in milliseconds.
 * The receive timeout defines the period a socket will block waiting
 * to receive data, before throwing an InterruptedIOException.
 *
 * @param		timeout		the timeout period, in milliseconds
 *
 * @exception	java.net.SocketException If an error occurs while setting
 * the timeout or the period is invalid.
 */
public synchronized void setSoTimeout(int timeout) throws SocketException {
	if (timeout >= 0) {
		checkClosedAndBind(false);
		impl.setOption(SocketOptions.SO_TIMEOUT, new Integer(timeout));
	// K0036 = Invalid negative timeout
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0036"));
}

/**
 * Sets the 'client' socket implementation factory for the application.
 * This may only be executed once during the application's lifetime.
 *
 * @param		fac			the socket factory to set
 * @exception	IOException	thrown if the factory has already been set
 */
public static synchronized void setDatagramSocketImplFactory(DatagramSocketImplFactory fac) throws IOException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSetFactory();
	if (factory == null)
		factory = fac;
	// K0044 = The factory has already been set
	else throw new SocketException(com.ibm.oti.util.Msg.getString("K0044"));
}

/**
 * Constructs a DatagramSocket using the specified DatagramSocketImpl. The
 * DatagramSocket is not bound.
 *
 * @param socketImpl the DatagramSocketImpl to use
 */
protected DatagramSocket(DatagramSocketImpl socketImpl) {
	impl = socketImpl;
}

/**
 * Constructs a DatagramSocket bound to the host/port specified by the SocketAddress,
 * or an unbound DatagramSocket if the SocketAddress is null.
 *
 * @param 		localAddr	the local machine address and port to bind to
 *
 * @throws 		IllegalArgumentException 	if the SocketAddress is not supported
 * @throws 		SocketException 			if a problem occurs creating or binding the socket
 */
public DatagramSocket(SocketAddress localAddr) throws SocketException {
	if (localAddr != null) {
		if (!(localAddr instanceof InetSocketAddress))
			// K0316 = SocketAddress {0} not supported
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", localAddr.getClass())); //$NON-NLS-1$
		checkListen(((InetSocketAddress)localAddr).getPort());
	}
	impl = factory != null ? factory.createDatagramSocketImpl() : createSocketImpl();
	impl.create();
	if (localAddr != null) {
		try {
			bind(localAddr);
		} catch (SocketException e) {
			close();
			throw e;
		}
	}
}

void checkClosedAndBind(boolean bind) throws SocketException {
	if (isClosed()) {
		// K003d = Socket is closed
		throw new SocketException(com.ibm.oti.util.Msg.getString("K003d"));
	}
	if (bind && !isBound()) {
		checkListen(0);
		impl.bind(0, InetAddress.ANY);
		isBound = true;
	}
}

/**
 * Bind the DatagramSocket to the nominated local host/port.
 *
 * @param		localAddr	the local machine address and port to bind on
 *
 * @throws 		IllegalArgumentException 	if the SocketAddress is not supported
 * @throws 		SocketException 			if the socket is already bound, or a problem occurs during the bind
 */
public void bind(SocketAddress localAddr) throws SocketException {
	checkClosedAndBind(false);
	int port = 0;
	InetAddress addr = InetAddress.ANY;
	if (localAddr != null) {
		if (!(localAddr instanceof InetSocketAddress))
			// K0316 = SocketAddress {0} not supported
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", localAddr.getClass())); //$NON-NLS-1$
		InetSocketAddress inetAddr = (InetSocketAddress)localAddr;
		addr = inetAddr.getAddress();
		if (addr == null)
			// K0317 = Host is unresolved: {0}
			throw new SocketException(com.ibm.oti.util.Msg.getString("K0317", inetAddr.getHostName())); //$NON-NLS-1$
		port = inetAddr.getPort();
		checkListen(port);
	}
	impl.bind(port, addr);
	isBound = true;
}

/**
 * Connect the datagram socket to a remote host and port.  The host and port
 * are validated, thereafter the only validation on send() and receive() is
 * that the packet address/port matches the connected target.
 *
 * @param		remoteAddr	the target address and port
 *
 * @exception	SocketException if a problem occurs during the connect
 */
public void connect(SocketAddress remoteAddr) throws SocketException {
	if (remoteAddr == null){
		// K0318 = SocketAddress is null
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0318"));
	}

	if (!(remoteAddr instanceof InetSocketAddress)) {
		// K0316 = SocketAddress {0} not supported
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", remoteAddr.getClass())); //$NON-NLS-1$
	}

	InetSocketAddress inetAddr = (InetSocketAddress)remoteAddr;
	if (inetAddr.getAddress() == null) {
		// K0317 = Host is unresolved: {0}
		throw new SocketException(com.ibm.oti.util.Msg.getString("K0317", inetAddr.getHostName())); //$NON-NLS-1$
	}

    synchronized(lock){
		// make sure the socket is open
		checkClosedAndBind(true);

		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			if (inetAddr.getAddress().isMulticastAddress()){
				security.checkMulticast(inetAddr.getAddress());
			}
			else {
				security.checkConnect(inetAddr.getAddress().getHostName(), inetAddr.getPort());
			}
		}

		// now try to do the connection at the native level.  To be compatible for the case
		// when the address is inaddr_any we just
		// eat the exception an act as if we are connected at the java level
		try
		{
			impl.connect(inetAddr.getAddress(), inetAddr.getPort());
		}
		catch (Exception e) {
			// not connected at the native level just do what we did before
		}

		// if we get here then we connected ok
		address = inetAddr.getAddress();
		port = inetAddr.getPort();
		isConnected = true;
	}
}

/**
 * Return if the socket is bound to a local address
 * and port.
 *
 * @return		<code>true</code> if the socket is bound to a local address, <code>false</code> otherwise.
 */
public boolean isBound() {
	return isBound;
}

/**
 * Return if the socket is connected.
 *
 * @return		<code>true</code> if the socket is connected, <code>false</code> otherwise.
 */
public boolean isConnected() {
	return isConnected;
}

/**
 * Answer the remote SocketAddress for this socket, or null
 * if the socket is not connected.
 *
 * @return		the remote socket address
 */
public SocketAddress getRemoteSocketAddress() {
	if (!isConnected())
		return null;
	return new InetSocketAddress(getInetAddress(), getPort());
}

/**
 * Answer the local SocketAddress for this socket, or null
 * if the socket is not bound.
 * <p>
 * This is useful on multihomed hosts.
 *
 * @return		the local socket address
 */
public SocketAddress getLocalSocketAddress() {
	if (!isBound())
		return null;
	return new InetSocketAddress(getLocalAddress(), getLocalPort());
}

/**
 * Set the SO_REUSEADDR socket option.
 *
 * @param		reuse	the socket SO_REUSEADDR option setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setReuseAddress(boolean reuse) throws SocketException {
	checkClosedAndBind(false);
	impl.setOption(SocketOptions.SO_REUSEADDR, reuse ? Boolean.TRUE : Boolean.FALSE);
}

/**
 * Get the state of the SO_REUSEADDR socket option.
 *
 * @return		<code>true</code> if the SO_REUSEADDR is enabled, <code>false</code> otherwise.
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public boolean getReuseAddress() throws SocketException {
	checkClosedAndBind(false);
	return ((Boolean)impl.getOption(SocketOptions.SO_REUSEADDR)).booleanValue();
}

/**
 * Set the SO_BROADCAST socket option.
 *
 * @param		broadcast	the socket SO_BROADCAST option setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setBroadcast(boolean broadcast) throws SocketException {
	checkClosedAndBind(false);
	impl.setOption(SocketOptions.SO_BROADCAST, broadcast ? Boolean.TRUE : Boolean.FALSE);
}

/**
 * Get the state of the SO_BROADCAST socket option.
 *
 * @return		<code>true</code> if the SO_BROADCAST is enabled, <code>false</code> otherwise.
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public boolean getBroadcast() throws SocketException {
	checkClosedAndBind(false);
	return ((Boolean)impl.getOption(SocketOptions.SO_BROADCAST)).booleanValue();
}

/**
 * Set the IP_TOS socket option.
 *
 * @param		value	the socket IP_TOS setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setTrafficClass(int value) throws SocketException {
	checkClosedAndBind(false);
	if (value < 0 || value > 255)
		throw new IllegalArgumentException();
	impl.setOption(SocketOptions.IP_TOS, new Integer(value));
}

/**
 * Get the IP_TOS socket option.
 *
 * @return		the IP_TOS socket option value
 *
 * @throws 		SocketException		if the option is invalid
 */
public int getTrafficClass() throws SocketException {
	checkClosedAndBind(false);
	return ((Number)impl.getOption(SocketOptions.IP_TOS)).intValue();
}

/**
 * Return if the socket is closed.
 *
 * @return		<code>true</code> if the socket is closed, <code>false</code> otherwise.
 */
public boolean isClosed() {
	return isClosed;
}
}
