package java.net;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.IOException;

/**
 * ServerSocket create connections between 'host' and 'client' machines.
 * The ServerSocket listens on a well known port and upon a connection request,
 * instantiates a 'host' sockets, which carries on future communication with the
 * requesting 'client' socket, so that the server socket may continue listening
 * for connection requests.  They are passive objects, having no execution thread
 * of their own to listen on.
 */
public class ServerSocket {

	SocketImpl impl;
	static SocketImplFactory factory;
	private volatile boolean isCreated = false;
	private boolean isBound = false;
	private boolean isClosed = false;

/**
 * Construct a ServerSocket, which is not bound to any port.
 * The default number of pending connections may be backlogged.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			Socket
 */
public ServerSocket() throws IOException {
	impl = factory != null ? factory.createSocketImpl() : new PlainServerSocketImpl();
}

/**
 * Construct a ServerSocket, bound to the nominated port on
 * the default localhost.  The default number of pending
 * connections may be backlogged.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aport		the port number to listen for connection requests on
 * @see			Socket
 */
public ServerSocket(int aport) throws IOException {
	this(aport, defaultBacklog(), InetAddress.ANY);
}

/**
 * Construct a ServerSocket, bound to the nominated port on
 * the default localhost.  The number of pending connections
 * that may be backlogged is a specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aport		the port number to listen for connection requests on
 * @param		backlog		the number of pending connection requests, before
 *							requests are rejected
 * @see			Socket
 */
public ServerSocket(int aport, int backlog) throws IOException {
	this(aport, backlog, InetAddress.ANY);
}

/**
 * Construct a ServerSocket, bound to the nominated local host/port.
 * The number of pending connections that may be backlogged is a specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aport		the port number to listen for connection requests on
 * @param		localAddr	the local machine address to bind on
 * @param		backlog		the number of pending connection requests, before
 *							requests are rejected
 * @see			Socket
 */
public ServerSocket(int aport, int backlog, InetAddress localAddr) throws IOException {
	super();
	checkListen(aport);
	impl = factory != null ? factory.createSocketImpl() : new PlainServerSocketImpl();
	InetAddress addr = localAddr == null ? InetAddress.ANY : localAddr;

	synchronized (this) {
		impl.create(true);
		isCreated = true;
		try {
			impl.bind(addr, aport);
			isBound = true;
			impl.listen(backlog > 0 ? backlog : defaultBacklog());
		} catch (IOException e) {
			close();
			throw e;
		}
	}
}

/**
 * Retrieve the first connection request and answer the 'host' socket
 * that will conduct further communications with the requesting
 * 'client' socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Socket	the 'host' socket
 * @exception	IOException	if an error occurs while instantiating the 'host' socket
 */
public Socket accept() throws IOException {
	checkClosedAndCreate(false);
	if (!isBound())
		// K031f = Socket is not bound
		throw new SocketException(com.ibm.oti.util.Msg.getString("K031f"));

	//If a SOCKS proxy is being used, accept does strange things.
	//Instead of returning a new Socket and allowing this ServerSocket
	//to be used for another accept, it actually uses the current ServerSocket
	//as the accepted Socket. So, closing the returned socket will close the
	//ServerSocket as well. The ServerSocket cannot be used for a second accept.
	if (PlainSocketImpl.usingSocks()) {
		if (impl instanceof PlainSocketImpl) {
			try {
				SecurityManager security = System.getSecurityManager();
				if (security != null) {
					security.checkAccept(
						getInetAddress().getHostName(),
						getLocalPort()
					);
				}
			} catch (SecurityException e) {
				close();
				throw e;
			}

			((PlainSocketImpl) impl).socksAccept();
			return new Socket(impl);
		} else {
			// K0041 = Socket implementation does not support SOCKS.
			throw new IOException(com.ibm.oti.util.Msg.getString("K0041")); //$NON-NLS-1$
		}
	}

	Socket aSocket = new Socket();
	try {
		synchronized (this) {
			implAccept(aSocket);
		};
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkAccept(
				aSocket.getInetAddress().getHostName(),
				aSocket.getPort()
			);
		}
	} catch (SecurityException e) {
		aSocket.close();
		throw e;
	} catch (IOException e) {
		aSocket.close();
		throw e;
	}
	return aSocket;
}

/**
 * Check whether the server may listen for connection requests on <code>aport</code>.
 * Throw an exception if the port is outside the valid range or
 * does not satisfy the security policy.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aPort	the candidate port to listen on
 */
void checkListen(int aPort) {
	if (aPort < 0 || aPort > 65535)
		// K0325 = Port out of range: {0}
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0325", aPort)); //$NON-NLS-1$
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkListen(aPort);
}

/**
 * Close this server socket.  Any attempt to connect to this socket
 * thereafter will fail.
 *
 * @author		OTI
 * @version		initial
 */
public void close() throws IOException {
	isClosed = true;
	impl.close();
}

/**
 * Answer the default number of pending connections on
 * a server socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the default number of pending connection requests
 */
static int defaultBacklog() {
	return 50;
}

/**
 * Answer the local IP address for this server socket.
 * Return null if the socket is not bound.
 * This is useful on multihomed hosts.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		InetAddress	the local address
 */
public InetAddress getInetAddress() {
	if (!isBound()) return null;
	return impl.getInetAddress();
}

/**
 * Answer the local port for this server socket.
 * Return -1 if the socket is not bound.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the local port the server is listening on
 */
public int getLocalPort() {
	if (!isBound()) return -1;
	return impl.getLocalPort();
}

/**
 * Answer the time-out period of this server socket.  This is the time
 * the server will wait listening for connections, before exiting.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the listening timeout
 * @exception	SocketException	thrown if option cannot be retrieved
 */
public synchronized int getSoTimeout() throws IOException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_TIMEOUT)).intValue();
}

/**
 * Invoke the server socket implementation to accept a connection
 * on the newly created <code>aSocket</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aSocket	the concrete socketImpl to accept the connection request on
 * @exception	IOException	thrown if connection cannot be accepted
 */
protected final void implAccept(Socket aSocket) throws IOException {
	impl.accept(aSocket.impl);
	aSocket.accepted();
}

/**
 * Set the server socket implementation factory.  This method may only be invoked
 * with sufficient security and only once during the application lifetime.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		aFactory	the streaming socket factory to be used for further
 *							socket instantiations
 * @exception	IOException	thrown if the factory is already set
 */
public static synchronized void setSocketFactory(SocketImplFactory aFactory) throws IOException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSetFactory();
	if (factory == null)
		factory = aFactory;
	// K0042 = Socket implementation factory already set
	else throw new SocketException(com.ibm.oti.util.Msg.getString("K0042")); //$NON-NLS-1$
}

/**
 * Set the listen time-out period for this server socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		timeout		the time to wait for a connection request
 * @exception	SocketException	thrown if an error occurs during setting the option
 */
public synchronized void setSoTimeout(int timeout) throws SocketException {
	checkClosedAndCreate(true);
	if (timeout >= 0)
		impl.setOption(SocketOptions.SO_TIMEOUT, new Integer(timeout));
	// K0036 = Invalid negative timeout
	else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0036")); //$NON-NLS-1$
}

/**
 * Answers a string containing a concise, human-readable
 * description of the server socket.  The <code>port</code> field
 * is reported a zero, as there is no connection formed to the server.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String		the description
 */
public String toString() {
	StringBuffer result = new StringBuffer(64);
	result.append("ServerSocket["); //$NON-NLS-1$
	if (!isBound())
		return result.append("unbound]").toString(); //$NON-NLS-1$
	return result.append("addr="). //$NON-NLS-1$
		append(getInetAddress()).
		append(",port=0,localport="). //$NON-NLS-1$
		append(getLocalPort()).
		append("]"). //$NON-NLS-1$
		toString();
}

/**
 * Bind the ServerSocket to the nominated local host/port.
 * The default number of pending connections may be backlogged.
 *
 * @author		IBM
 * @version		initial
 *
 * @param		localAddr	the local machine address and port to bind on
 *
 * @exception IllegalArgumentException if the SocketAddress
 * 		is not supported
 * @exception IOException if the socket is already bound, or a problem occurs
 * 		during the bind
 */
public void bind(SocketAddress localAddr) throws IOException {
	bind(localAddr, defaultBacklog());
}

/**
 * Bind the ServerSocket to the nominated local host/port.
 * The number of pending connections that may be backlogged is a specified.
 *
 * @author		IBM
 * @version		initial
 *
 * @param		localAddr	the local machine address and port to bind on
 * @param		backlog		the number of pending connection requests, before
 *							requests are rejected
 *
 * @exception IllegalArgumentException if the SocketAddress
 * 		is not supported
 * @exception IOException if the socket is already bound, or a problem occurs
 * 		during the bind
 */
public void bind(SocketAddress localAddr, int backlog) throws IOException {
	checkClosedAndCreate(true);
	if (isBound())
		// K0315 = Socket is already bound
		throw new BindException(com.ibm.oti.util.Msg.getString("K0315")); //$NON-NLS-1$
	int port = 0;
	InetAddress addr = InetAddress.ANY;
	if (localAddr != null) {
		if (!(localAddr instanceof InetSocketAddress))
			// K0316 = SocketAddress {0} not supported
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", localAddr.getClass())); //$NON-NLS-1$
		InetSocketAddress inetAddr = (InetSocketAddress)localAddr;
		if ((addr = inetAddr.getAddress()) == null)
			// K0317 = Host is unresolved: {0}
			throw new SocketException(com.ibm.oti.util.Msg.getString("K0317", inetAddr.getHostName())); //$NON-NLS-1$
		port = inetAddr.getPort();
	}
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkListen(port);

	synchronized (this) {
		try {
			impl.bind(addr, port);
			isBound = true;
			impl.listen(backlog > 0 ? backlog : defaultBacklog());
		} catch (IOException e) {
			close();
			throw e;
		}
	}
}

/**
 * Answer the local SocketAddress for this server socket, or null
 * if the socket is not bound.
 * This is useful on multihomed hosts.
 */
public SocketAddress getLocalSocketAddress() {
	if (!isBound())
		return null;
	return new InetSocketAddress(getInetAddress(), getLocalPort());
}

/**
 * Return if the server socket is bound to a local address
 * and port.
 */
public boolean isBound() {
	return isBound;
}

/**
 * Return if the server socket is closed.
 */
public boolean isClosed() {
	return isClosed;
}

/**
 * Check if the socket is closed, and throw an exception.
 */
private void checkClosedAndCreate(boolean create) throws SocketException {
	if (isClosed())
		// K003d = Socket is closed
		throw new SocketException(com.ibm.oti.util.Msg.getString("K003d"));

	if (!create || isCreated) return;

	synchronized (this) {
		if (isCreated) return;
		try {
			impl.create(true);
		} catch (SocketException e) {
			throw e;
		} catch (IOException e) {
			throw new SocketException(e.toString());
		}
		isCreated = true;
	}
}

/**
 * Set the SO_REUSEADDR socket option.
 *
 * @param		reuse	the socket SO_REUSEADDR option setting
 */
public void setReuseAddress(boolean reuse) throws SocketException {
	checkClosedAndCreate(true);
	impl.setOption(SocketOptions.SO_REUSEADDR, reuse ? Boolean.TRUE : Boolean.FALSE);
}

/**
 * Get the state of the SO_REUSEADDR socket option.
 */
public boolean getReuseAddress() throws SocketException {
	checkClosedAndCreate(true);
	return ((Boolean)impl.getOption(SocketOptions.SO_REUSEADDR)).booleanValue();
}

/**
 * Set the socket receive buffer size.
 *
 * @author		IBM
 * @version		initial
 *
 * @param		size	the buffer size, in bytes
 *
 * @exception	java.net.SocketException If an error occurs while setting
 * the size or the size is invalid.
 */
public void setReceiveBufferSize(int size) throws SocketException {
	checkClosedAndCreate(true);
	if (size >= 1)
		impl.setOption(SocketOptions.SO_RCVBUF, new Integer(size));
	// K0035 = Zero or negative buffer size
	else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0035")); //$NON-NLS-1$
}

/**
 * Answer the socket receive buffer size (SO_RCVBUF).
 *
 * @author		IBM
 * @version		initial
 *
 * @return	int		socket receive buffer size
 */
public int getReceiveBufferSize() throws SocketException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_RCVBUF)).intValue();
}
}
