package java.net;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.FileDescriptor;

import java.security.AccessController;
import com.ibm.oti.util.PriviAction;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * This class represents sockets to be used in connection-oriented
 * (streaming) protocols.
 *
 * @author		OTI
 * @version		initial
 */
public class Socket {

	SocketImpl impl;
	static SocketImplFactory factory;
	private volatile boolean isCreated = false;
	private boolean isBound = false;
	private boolean isConnected = false;
	private boolean isClosed = false;
	private boolean isInputShutdown = false;
	private boolean isOutputShutdown = false;
	private Object connectLock = new Object();

	static final int MULTICAST_IF = 1;
	static final int MULTICAST_TTL = 2;
	static final int TCP_NODELAY = 4;
	static final int FLAG_SHUTDOWN = 8;
	static final int FLAG_BIND = 16;

	// Fill in the JNI id caches

	private static native void oneTimeInitialization( boolean jcl_supports_ipv6 );
	static {
			oneTimeInitialization( true );
	}

/**
 * Construct a connection-oriented Socket.  The Socket is created in the
 * <code>factory</code> if declared, or otherwise of the default type.
 *
 * @see			SocketImplFactory
 */
public Socket() {
	impl = factory != null ? factory.createSocketImpl() : new PlainSocketImpl2();
}

/**
 * Construct a stream socket connected to the nominated destination host/port.
 * By default, the socket binds it to any available port on the default localhost.
 *
 * @param		dstName		the destination host to connect to
 * @param		dstPort		the port on the destination host to connect to
 *
 * @throws		UnknownHostException	if the host cannot be resolved
 * @throws		IOException				if an error occurs while instantiating the socket
 */
public Socket(String dstName, int dstPort) throws UnknownHostException, IOException {
	this();
	InetAddress dstAddress = InetAddress.getByName(dstName);
	checkDestination(dstAddress, dstPort);
	startupSocket(dstAddress, dstPort, null, 0, true);
}

/**
 * Construct a stream socket connected to the nominated destination host/port.
 * The socket is bound it to the nominated localAddress/port.
 *
 * @param		dstName			the destination host to connect to
 * @param		dstPort			the port on the destination host to connect to
 * @param		localAddress	the local host address to bind to
 * @param		localPort		the local port to bind to
 *
 * @throws		UnknownHostException	if the host cannot be resolved
 * @throws		IOException				if an error occurs while instantiating the socket
 */
public Socket(String dstName, int dstPort, InetAddress localAddress, int localPort) throws IOException {
	this();
	InetAddress dstAddress = InetAddress.getByName(dstName);
	checkDestination(dstAddress, dstPort);
	startupSocket(dstAddress, dstPort, localAddress, localPort, true);
}

/**
 * Construct a stream socket connected to the nominated destination host address/port.
 * By default, the socket binds it to any available port on the default localhost.
 *
 * @param		dstAddress	the destination host address to connect to
 * @param		dstPort		the port on the destination host to connect to
 *
 * @throws		IOException		if an error occurs while instantiating the socket
 */
public Socket(InetAddress dstAddress, int dstPort) throws IOException {
	this();
	checkDestination(dstAddress, dstPort);
	startupSocket(dstAddress, dstPort, null, 0, true);
}

/**
 * Construct a stream socket connected to the nominated destination
 * host address/port. The socket is bound it to the nominated localAddress/port.
 *
 * @param		dstAddress		the destination host address to connect to
 * @param		dstPort			the port on the destination host to connect to
 * @param		localAddress	the local host address to bind to
 * @param		localPort		the local port to bind to
 *
 * @throws		IOException		if an error occurs while instantiating the socket
 */
public Socket(InetAddress dstAddress, int dstPort, InetAddress localAddress, int localPort) throws IOException {
	this();
	checkDestination(dstAddress, dstPort);
	startupSocket(dstAddress, dstPort, localAddress, localPort, true);
}

/**
 * Creates an unconnected socket, wrapping the <code>socketImpl</code>
 * argument.
 *
 * @param		anImpl		the socket to wrap
 *
 * @throws		SocketException 	if an error occurs assigning the implementation
 */
protected Socket(SocketImpl anImpl) throws SocketException {
	impl = anImpl;
}

/**
 * Check the connection destination satisfies the security policy and
 * is in the valid port range.
 *
 * @param		destAddr	the destination host address
 * @param		dstPort		the port on the destination host
 */
void checkDestination(InetAddress destAddr, int dstPort) {
	if (dstPort < 0 || dstPort > 65535)
		// K0032 = Address null or destination port out of range
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0032"));
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkConnect(destAddr.getHostName(), dstPort);
}

/**
 * Close the socket.  It is not valid to use the socket thereafter.
 *
 * @throws		IOException		if an error occurs during the close
 */
public synchronized void close() throws IOException {
	isClosed = true;
	impl.close();
}

/**
 * Answer the remote address to which the socket is connected.
 *
 * @return		the remote address to which the socket is connected
 */
public InetAddress getInetAddress() {
	if (!isConnected())
		return null;
	return impl.getInetAddress();
}

/**
 * Answer the socket input stream, to read byte data off the socket.
 * Note, multiple input streams may be created on a single socket.
 *
 * @return		a byte oriented read stream for this socket
 *
 * @throws		IOException		if an error occurs creating the stream
 *
 * @see			java.net.SocketInputStream
 */
public InputStream getInputStream() throws IOException {
	checkClosedAndCreate(false);
	if (isInputShutdown())
		// K0321 = Socket input is shutdown
		throw new SocketException(com.ibm.oti.util.Msg.getString("K0321"));
	return impl.getInputStream();
}

/**
 * Answer the SO_KEEPALIVE option for this socket.
 *
 * @return		the socket SO_KEEPALIVE option setting
 *
 * @throws		SocketException		if an error occurs on the option access
 */
public boolean getKeepAlive() throws SocketException {
	checkClosedAndCreate(true);
	return ((Boolean)impl.getOption(SocketOptions.SO_KEEPALIVE)).booleanValue();
}

/**
 * Answer the local address to which the socket is bound.
 *
 * @return		the local address to which the socket is bound
 */
public InetAddress getLocalAddress() {
	if (!isBound())
		return InetAddress.ANY;
	return Socket.getSocketLocalAddressImpl(impl.fd, InetAddress.preferIPv6Addresses());
}

/**
 * Answer the local port to which the socket is bound.
 *
 * @return		the local port to which the socket is bound
 */
public int getLocalPort() {
	if (!isBound())
		return -1;
	return impl.getLocalPort();
}

/**
 * Answer the socket output stream, for writing byte data on the socket.
 * Note, multiplie output streams may be created on a single socket.
 *
 * @return		OutputStream	a byte oriented write stream for this socket
 *
 * @throws		IOException		if an error occurs creating the stream
 *
 * @see			java.net.SocketOutputStream
 */
public OutputStream getOutputStream() throws IOException {
	checkClosedAndCreate(false);
	return impl.getOutputStream();
}

/**
 * Answer the remote port to which the socket is connected
 *
 * @return		the remote port to which the socket is connected.
 */
public int getPort() {
	if (!isConnected())
		return 0;
	return impl.getPort();
}

/**
 * Answer the linger-on-close timeout for this socket.
 * The SO_LINGER option, a value of -1 indicates it is disabled.
 *
 * @return		the linger-on-close timeout for this socket
 *
 * @throws		SocketException		if an error occurs on the option access
 */
public int getSoLinger() throws SocketException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_LINGER)).intValue();
}

/**
 * Answer the socket receive buffer size (SO_RCVBUF).
 *
 * @return		socket receive buffer size
 *
 * @throws 		SocketException		if an error occurs on the option access
 */
public synchronized int getReceiveBufferSize() throws SocketException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_RCVBUF)).intValue();
}

/**
 * Answer the socket send buffer size (SO_SNDBUF).
 *
 * @return		socket send buffer size
 *
 * @throws 		SocketException		if an error occurs on the option access
 */
public synchronized int getSendBufferSize() throws SocketException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_SNDBUF)).intValue();
}

/**
 * Answer the socket read timeout.
 * The SO_TIMEOUT option, a value of 0 indicates it is disabled and
 * a read operation will block indefinitely waiting for data.
 *
 * @return		the socket read timeout
 *
 * @throws 		SocketException		if an error occurs on the option access
 */
public synchronized int getSoTimeout() throws SocketException {
	checkClosedAndCreate(true);
	return ((Integer)impl.getOption(SocketOptions.SO_TIMEOUT)).intValue();
}

/**
 * Answer true if the socket is using Nagle's algorithm.
 * The TCP_NODELAY option setting.
 *
 * @return		the socket TCP_NODELAY option setting
 *
 * @throws 		SocketException		if an error occurs on the option access
 */
public boolean getTcpNoDelay() throws SocketException {
	checkClosedAndCreate(true);
	return ((Boolean)impl.getOption(SocketOptions.TCP_NODELAY)).booleanValue();
}

/**
 * Query the IP stack for the local address to which this socket is bound.
 *
 * @param		aFD			the socket descriptor
 * @param 		preferIPv6Addresses address preference for nodes that support both IPv4 and IPv6
 * @return		the local address to which the socket is bound
 */
static native InetAddress getSocketLocalAddressImpl(FileDescriptor aFD, boolean preferIPv6Addresses);

/**
 * Query the IP stack for the local port to which this socket is bound.
 *
 * @param		aFD			the socket descriptor
 * @param 		preferIPv6Addresses address preference for nodes that support both IPv4 and IPv6
 * @return		int			the local port to which the socket is bound
 */
static native int getSocketLocalPortImpl(FileDescriptor aFD, boolean preferIPv6Addresses);

/**
 * Query the IP stack for the nominated socket option.
 *
 * @param		aFD			the socket descriptor
 * @param		opt			the socket option type
 * @return		the nominated socket option value
 *
 * @throws		SocketException		if the option is invalid
 */
static native Object getSocketOptionImpl(FileDescriptor aFD, int opt) throws SocketException;

/**
 * Set the nominated socket option in the IP stack.
 *
 * @param		aFD			the socket descriptor
 * @param		opt			the option selector
 * @param		optVal		the nominated option value
 *
 * @throws		SocketException		if the option is invalid or cannot be set
 */
static native void setSocketOptionImpl(FileDescriptor aFD, int opt, Object optVal) throws SocketException;

/**
 * Set the SO_KEEPALIVE option for this socket.
 *
 * @param		value		the socket SO_KEEPALIVE option setting
 *
 * @throws		SocketException		if an error occurs setting the option
 */
public void setKeepAlive(boolean value) throws SocketException {
	checkClosedAndCreate(true);
	impl.setOption(SocketOptions.SO_KEEPALIVE, value ? Boolean.TRUE : Boolean.FALSE);
}

static native int getSocketFlags();

/**
 * Close the socket in the IP stack.
 *
 * @param		aFD			the socket descriptor
 */
static native void socketCloseImpl(FileDescriptor aFD);

/**
 * Sets the 'client' socket implementation factory for the application.
 * This may only be executed once during the application's lifetime.
 *
 * @param		fac			the socket factory to set
 *
 * @throws		IOException		if the factory has already been set
 */
public static synchronized void setSocketImplFactory(SocketImplFactory fac) throws IOException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSetFactory();
	if (factory == null)
		factory = fac;
	// K0044 = The factory has already been set
	else throw new SocketException(com.ibm.oti.util.Msg.getString("K0044"));
}

/**
 * Set the socket send buffer size.
 *
 * @param		size		the buffer size, in bytes
 *
 * @throws		SocketException		if an error occurs while setting the size or the size is invalid.
 */
public synchronized void setSendBufferSize(int size) throws SocketException {
	checkClosedAndCreate(true);
	if (size >= 1)
		impl.setOption(SocketOptions.SO_SNDBUF, new Integer(size));
	// K0035 = Zero or negative buffer size
	else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0035"));
}

/**
 * Set the socket receive buffer size.
 *
 * @param		size	the buffer size, in bytes
 *
 * @throws		SocketException 	tf an error occurs while setting the size or the size is invalid.
 */
public synchronized void setReceiveBufferSize(int size) throws SocketException {
	checkClosedAndCreate(true);
	if (size >= 1)
		impl.setOption(SocketOptions.SO_RCVBUF, new Integer(size));
	// K0035 = Zero or negative buffer size
	else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0035"));
}

/**
 * Set the SO_LINGER option, with the specified time, in seconds.
 * The SO_LINGER option is silently limited to 65535 seconds.
 *
 * @param		on			if linger is enabled
 * @param		timeout		the linger timeout value, in seconds
 *
 * @throws		SocketException		if an error occurs setting the option
 */
public void setSoLinger(boolean on, int timeout) throws SocketException {
checkClosedAndCreate(true);
	if (!on || 0 <= timeout) {
		int val = on ? (65535 < timeout ? 65535 : timeout) : -1;
		impl.setOption(SocketOptions.SO_LINGER, new Integer(val));
	// K0045 = Attempted to set a negative SoLinger
	} else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0045"));
}

/**
 * Set the read timeout on this socket.
 * The SO_TIMEOUT option, is specified in milliseconds.
 * The read operation will block indefinitely for a zero value.
 *
 * @param		timeout		the read timeout value
 *
 * @throws		SocketException		if an error occurs setting the option
 */
public synchronized void setSoTimeout(int timeout) throws SocketException {
	checkClosedAndCreate(true);
	if (timeout >= 0)
		impl.setOption(SocketOptions.SO_TIMEOUT, new Integer(timeout));
	// K0036 = Invalid negative timeout
	else throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0036"));
}

/**
 * Set whether the socket is to use Nagle's algorithm.
 * The TCP_NODELAY option setting.
 *
 * @param		on		the socket TCP_NODELAY option setting
 *
 * @throws		SocketException		if an error occurs setting the option
 */
public void setTcpNoDelay(boolean on) throws SocketException {
	checkClosedAndCreate(true);
	impl.setOption(SocketOptions.TCP_NODELAY, new Boolean(on));
}

/**
 * Creates a stream socket, binds it to the nominated local address/port,
 * then connects it to the nominated destination address/port.
 *
 * @param		dstAddress		the destination host address
 * @param		dstPort			the port on the destination host
 * @param		localAddress	the address on the local machine to bind
 * @param		localPort		the port on the local machine to bind
 *
 * @throws		IOException 	thrown if a error occurs during the bind or connect operations
 */
void startupSocket(InetAddress dstAddress, int dstPort, InetAddress localAddress, int localPort, boolean streaming) throws IOException {
	if (localPort < 0 || localPort > 65535)
		// K0046 = Local port declared out of range
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0046"));

	InetAddress addr = localAddress == null ? InetAddress.ANY : localAddress;
	synchronized (this) {
		impl.create(streaming);
		isCreated = true;
		try {
			if (!streaming || !PlainSocketImpl.usingSocks())
				impl.bind(addr, localPort);
			isBound = true;
			impl.connect(dstAddress, dstPort);
			isConnected = true;
		} catch (IOException e) {
			impl.close();
			throw e;
		}
	}
}

/**
 * Answers a string containing a concise, human-readable
 * description of the socket.
 *
 * @return		the description
 */
public String toString() {
	if (!isConnected())
		return "Socket[unconnected]"; //$NON-NLS-1$
	return impl.toString();
}

/**
 * Shutdown the input portion of the socket.
 *
 * @throws 		IOException			if an error occurs while closing the socket input
 * @throws		SocketException		if the socket is closed
 */
public void shutdownInput() throws IOException {
	checkClosedAndCreate(false);
	impl.shutdownInput();
	isInputShutdown = true;
}

/**
 * Shutdown the output portion of the socket.
 *
 * @throws 		IOException			if an error occurs while closing the socket output
 * @throws		SocketException		if the socket is closed
 */
public void shutdownOutput() throws IOException {
	checkClosedAndCreate(false);
	impl.shutdownOutput();
	isOutputShutdown = true;
}

/**
 * Check if the socket is closed, and throw an exception. Otherwise
 * create the underlying SocketImpl.
 *
 * @throws		SocketException		if the socket is closed
 */
private void checkClosedAndCreate(boolean create) throws SocketException {
	if (isClosed())
		// K003d = Socket is closed
		throw new SocketException(com.ibm.oti.util.Msg.getString("K003d"));
	if (!create) {
		if (!isConnected())
			// K0074 = Not connected
			throw new SocketException(com.ibm.oti.util.Msg.getString("K0074"));
		// a connected socket must be created
	}
	if (isCreated) return;
	synchronized (this) {
		if (isCreated) return;
		try {
			impl.create(true);
		} catch (SocketException e) {
			throw (SocketException)e;
		} catch (IOException e) {
			throw new SocketException(e.toString());
		}
		isCreated = true;
	}
}

/**
 * Answer the local SocketAddress for this socket, or null
 * if the socket is not bound.
 * <p>
 * This is useful on multihomed hosts.
 *
 * @return		the local socket address
 */
public SocketAddress getLocalSocketAddress() {
	if (!isBound())
		return null;
	return new InetSocketAddress(getLocalAddress(), getLocalPort());
}

/**
 * Answer the remote SocketAddress for this socket, or null
 * if the socket is not connected.
 *
 * @return		the remote socket address
 */
public SocketAddress getRemoteSocketAddress() {
	if (!isConnected())
		return null;
	return new InetSocketAddress(getInetAddress(), getPort());
}

/**
 * Return if the socket is bound to a local address
 * and port.
 *
 * @return		<code>true</code> if the socket is bound to a local address, <code>false</code> otherwise.
 */
public boolean isBound() {
	return isBound;
}

/**
 * Return if the socket is connected.
 *
 * @return		<code>true</code> if the socket is connected, <code>false</code> otherwise.
 */
public boolean isConnected() {
	return isConnected;
}

/**
 * Return if the socket is closed.
 *
 * @return		<code>true</code> if the socket is closed, <code>false</code> otherwise.
 */
public boolean isClosed() {
	return isClosed;
}

/**
 * Bind the Socket to the nominated local host/port.
 *
 * @param		localAddr	the local machine address and port to bind on
 *
 * @throws 		IllegalArgumentException 	if the SocketAddress is not supported
 * @throws 		IOException 				if the socket is already bound, or a problem occurs during the bind
 */
public void bind(SocketAddress localAddr) throws IOException {
	checkClosedAndCreate(true);
	if (isBound())
		// K0315 = Socket is already bound
		throw new BindException(com.ibm.oti.util.Msg.getString("K0315")); //$NON-NLS-1$

	int port = 0;
	InetAddress addr = InetAddress.ANY;
	if (localAddr != null) {
		if (!(localAddr instanceof InetSocketAddress))
			// K0316 = SocketAddress {0} not supported
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", localAddr.getClass())); //$NON-NLS-1$
		InetSocketAddress inetAddr = (InetSocketAddress)localAddr;
		if ((addr = inetAddr.getAddress()) == null)
			// K0317 = Host is unresolved: {0}
			throw new SocketException(com.ibm.oti.util.Msg.getString("K0317", inetAddr.getHostName())); //$NON-NLS-1$
		port = inetAddr.getPort();
	}

	synchronized (this) {
		try {
			if (!PlainSocketImpl.usingSocks())
				impl.bind(addr, port);
			isBound = true;
		} catch (IOException e) {
			impl.close();
			throw e;
		}
	}
}

/**
 * Connect the Socket to the host/port specified by the SocketAddress.
 *
 * @param 		remoteAddr		the remote machine address and port to connect to
 *
 * @throws 		IllegalArgumentException 	if the SocketAddress is not supported
 * @throws		IOException 				if the socket is already connected, or a problem occurs during the connect
 */
public void connect(SocketAddress remoteAddr) throws IOException {
	connect(remoteAddr, 0);
}

/**
 * Connect the Socket to the host/port specified by the SocketAddress with
 * a specified timeout.
 *
 * @param 		remoteAddr	the remote machine address and port to connect to
 * @param		timeout		the millisecond timeout value, the connect will block indefinitely for a zero value.
 *
 * @throws 		IllegalArgumentException 	if the timeout is negative, or the SocketAddress is not supported
 * @throws 		IOException 				if the socket is already connected, or a problem occurs during the connect
 */
public void connect(SocketAddress remoteAddr, int timeout) throws IOException {
	checkClosedAndCreate(true);
	if (timeout < 0)
		// K0036 = Invalid negative timeout
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0036"));
	if (isConnected())
		// K0079 = Already connected
		throw new SocketException(com.ibm.oti.util.Msg.getString("K0079")); //$NON-NLS-1$
	if (remoteAddr == null)
		// K0318 = SocketAddress is null
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0318"));

	if (!(remoteAddr instanceof InetSocketAddress))
		// K0316 = SocketAddress {0} not supported
		throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0316", remoteAddr.getClass())); //$NON-NLS-1$
	InetSocketAddress inetAddr = (InetSocketAddress)remoteAddr;
	InetAddress addr;
	if ((addr = inetAddr.getAddress()) == null)
		// K0317 = Host is unresolved: {0}
		throw new SocketException(com.ibm.oti.util.Msg.getString("K0317", inetAddr.getHostName())); //$NON-NLS-1$
	int port = inetAddr.getPort();

	checkDestination(addr, port);
	synchronized (connectLock) {
		try {
			if (!isBound()) {
// socket allready created at this point by earlier call or checkClosedAndCreate
// this caused us to lose socket options on create
//				impl.create(true);
				if (!PlainSocketImpl.usingSocks())
			    impl.bind(InetAddress.ANY, 0);
				isBound = true;
			}
			impl.connect(remoteAddr, timeout);
			isConnected = true;
		} catch (IOException e) {
			impl.close();
			throw e;
		}
	}
}

/**
 * Return if {@link #shutdownInput} has been called.
 *
 * @return		<code>true</code> if <code>shutdownInput</code> has been called, <code>false</code> otherwise.
 */
public boolean isInputShutdown() {
	return isInputShutdown;
}

/**
 * Return if {@link #shutdownOutput} has been called.
 *
 * @return		<code>true</code> if <code>shutdownOutput</code> has been called, <code>false</code> otherwise.
 */
public boolean isOutputShutdown() {
	return isOutputShutdown;
}

/**
 * Set the SO_REUSEADDR socket option.
 *
 * @param		reuse	the socket SO_REUSEADDR option setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setReuseAddress(boolean reuse) throws SocketException {
	checkClosedAndCreate(true);
	impl.setOption(SocketOptions.SO_REUSEADDR, reuse ? Boolean.TRUE : Boolean.FALSE);
}

/**
 * Get the state of the SO_REUSEADDR socket option.
 *
 * @return		<code>true</code> if the SO_REUSEADDR is enabled, <code>false</code> otherwise.
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public boolean getReuseAddress() throws SocketException {
	checkClosedAndCreate(true);
	return ((Boolean)impl.getOption(SocketOptions.SO_REUSEADDR)).booleanValue();
}

/**
 * Set the SO_OOBINLINE socket option. When this option is enabled,
 * out of band data is recieved in the normal data stream.
 *
 * @param		oobinline	the socket SO_OOBINLINE option setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setOOBInline(boolean oobinline) throws SocketException {
	checkClosedAndCreate(true);
	impl.setOption(SocketOptions.SO_OOBINLINE, oobinline ? Boolean.TRUE : Boolean.FALSE);
}

/**
 * Get the state of the SO_OOBINLINE socket option.
 *
 * @return		<code>true</code> if the SO_OOBINLINE is enabled, <code>false</code> otherwise.
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public boolean getOOBInline() throws SocketException {
	checkClosedAndCreate(true);
	return ((Boolean)impl.getOption(SocketOptions.SO_OOBINLINE)).booleanValue();
}

/**
 * Set the IP_TOS socket option.
 *
 * @param		value	the socket IP_TOS setting
 *
 * @throws		SocketException		if the socket is closed or the option is invalid.
 */
public void setTrafficClass(int value) throws SocketException {
	checkClosedAndCreate(true);
	if (value < 0 || value > 255)
		throw new IllegalArgumentException();
	impl.setOption(SocketOptions.IP_TOS, new Integer(value));
}

/**
 * Get the IP_TOS socket option.
 *
 * @return		the IP_TOS socket option value
 *
 * @throws 		SocketException		if the option is invalid
 */
public int getTrafficClass() throws SocketException {
	checkClosedAndCreate(true);
	return ((Number)impl.getOption(SocketOptions.IP_TOS)).intValue();
}

/**
 * Send the single byte of urgent data on the socket.
 *
 * @param value the byte of urgent data
 *
 * @exception IOException when an error occurs sending urgent data
 */
public void sendUrgentData(int value) throws IOException {
	if (!impl.supportsUrgentData()) {
		// K0333 = Urgent data not supported
		throw new SocketException(com.ibm.oti.util.Msg.getString("K0333"));
	}
	impl.sendUrgentData(value);
}

/**
 * Set the appropriate flags for a Socket created by ServerSocket.accept().
 *
 * @see ServerSocket#implAccept
 */
void accepted() {
	isCreated = isBound = isConnected = true;
}

static boolean preferIPv4Stack() {
	 String result = (String)AccessController.doPrivileged(new PriviAction("java.net.preferIPv4Stack"));
		if ("true".equals(result))
			return true;
		return false;
}

}
