package java.net;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 * The SocketInputStream supports the streamed reading of bytes
 * from a socket.  Multiple streams may be opened on a socket, so
 * care should be taken to manage opened streams and coordinate
 * read operations between threads.
 */

import java.io.*;

class SocketInputStream extends InputStream {

	SocketImpl socket;
/**
 * Constructs a SocketInputStream for the <code>socket</code>.
 * Read operations are forwarded to the <code>socket</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		socket		the socket to be read
 * @see			Socket
 */

public SocketInputStream(SocketImpl socket) {

	super();
	this.socket = socket;
}
/**
 * Answer the number of bytes that may be read without blocking.
 * Zero indicates a read operation would block.
 * This call itself does not block, but may throw an IOException.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the number of bytes that may be read without blocking
 * @exception	IOException		thrown if an error occurs during the test
 */

public int available() throws IOException {

	return socket.available();
}
/**
 * Close the stream and the underlying socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	IOException		thrown if an error occurs during the close
 */

public void close() throws IOException {

	socket.close();
	super.close();
}
/**
 * Read a single byte from the socket, answering the value as an <code>int</code>.
 * This call may block indefinitely, depending upon whether data is available
 * and whether the read timeout option has been set on the socket.
 * A value of -1 indicates 'end-of-file'.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int				the value read
 * @exception	IOException		thrown if an error occurs during the read
 */

public int read() throws IOException {

	byte[] buffer = new byte[1];
	int result = socket.read(buffer, 0, 1);
	return (-1 == result)? result : buffer[0] & 0xFF;
}
/**
 * Read a buffer.length number of bytes from the socket, into the <code>buffer</code>.
 * This call may block indefinitely, depending upon whether data is available
 * and whether the read timeout option has been set on the socket.
 * The number of bytes actually read is returned; a value of -1 indicates
 * 'end-of-file'.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer			the buffer to read into
 * @return		int				the number of bytes actually read
 * @exception	IOException		thrown if an error occurs during the read
 */

public int read(byte[] buffer) throws IOException {

	return read(buffer, 0, buffer.length);
}
/**
 * Read a <code>count</code> number of bytes from the socket, into the
 * <code>buffer</code> at an <code>offset</code>.
 * This call may block indefinitely, depending upon whether data is available
 * and whether the read timeout option has been set on the socket.
 * The number of bytes actually read is returned; a value of -1 indicates
 * 'end-of-file'.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer			the buffer to read into
 * @param		offset			the offset into the buffer to start filling
 * @param		count			the maximum number of bytes to read
 * @return		int				the number of bytes actually read
 * @exception	IOException, ArrayIndexOutOfBoundsException
 *								thrown if the argument bounds are incorrect or
 * 								an error occurs during the read
 */

public int read(byte[] buffer, int offset, int count) throws IOException {

	// K0047 = buffer is null
	if (null == buffer) throw new IOException(com.ibm.oti.util.Msg.getString("K0047"));
	if (0 == count) return 0;

	if (0 > offset || offset >= buffer.length)
		// K002e = Offset out of bounds
		throw new ArrayIndexOutOfBoundsException(com.ibm.oti.util.Msg.getString("K002e"));
	if (0 > count  || offset + count > buffer.length)
		// K002f = Arguments out of bounds
		throw new ArrayIndexOutOfBoundsException(com.ibm.oti.util.Msg.getString("K002f"));

	return socket.read(buffer, offset, count);
}

/**
 * Skips <code>n</code> number of bytes in this InputStream.  Subsequent
 * <code>read()</code>'s will not return these bytes unless <code>reset()</code>
 * is used.  This method may perform multiple reads to read <code>n</code>
 * bytes.  This implementation reads <code>n</code> bytes into a temporary
 * buffer.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		n			the number of bytes to skip.
 * @return		the number of bytes actually skipped.
 *
 * @exception 	java.io.IOException If the stream is already closed or another IOException occurs.
 */

public long skip(long n) throws IOException {

	return (0 == n)? 0: super.skip(n);
}
}
