package java.net;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 * The SocketOutputStream supports the streamed writing of bytes
 * to the socket.  Multiple streams may be opened on a socket, so
 * care should be taken to manage opened streams and coordinate
 * write operations between threads.
 */

import java.io.*;

class SocketOutputStream extends OutputStream {

	SocketImpl socket;
/**
 * Constructs a SocketOutputStream for the <code>socket</code>.
 * Write operations are forwarded to the <code>socket</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		socket		the socket to be written
 * @see			Socket
 */

public SocketOutputStream(SocketImpl socket) {

	super();
	this.socket = socket;
}
/**
 * Close the stream and the underlying socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	IOException		thrown if an error occurs during the close
 */

public void close() throws IOException {

	socket.close();
	super.close();
}
/**
 * Write the <code>buffer</code> to the socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer			the buffer to write
 * @exception	IOException		thrown if an error occurs during the write
 */

public void write(byte[] buffer) throws IOException {

	socket.write(buffer, 0, buffer.length);
}
/**
 * Write the <code>count</code> number of bytes from the <code>buffer</code>
 * to the socket, starting at <code>offset</code>.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer			the buffer to write
 * @param		offset			the offset in buffer to start writing
 * @param		count			the number of bytes to write
 * @exception	IOException, IndexOutOfBoundsException
 *								thrown if an error occurs during the write
 */
public void write(byte[] buffer, int offset, int count) throws IOException {
	// avoid int overflow
	if (buffer != null) {
		if (0 <= offset && offset <= buffer.length && 0 <= count && count <= buffer.length - offset) {
			socket.write(buffer, offset, count);
		// K002f = Arguments out of bounds
		} else throw new ArrayIndexOutOfBoundsException(com.ibm.oti.util.Msg.getString("K002f"));
	// K0047 = buffer is null
	} else throw new NullPointerException(com.ibm.oti.util.Msg.getString("K0047"));
}
/**
 * Write a single byte, the lowest-order byte from an <code>int</code>
 * to the socket.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		oneByte			the value to write
 * @exception	IOException		thrown if an error occurs during the write
 */

public void write(int oneByte) throws IOException {

	byte[] buffer = new byte[1];
	buffer[0] = (byte)(oneByte & 0xFF);

	socket.write(buffer, 0, 1);
}
}
