package java.net;

import java.net.InetAddress;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import com.ibm.oti.util.Inet6Util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * SocketPermissions represent permission to access resources via sockets.
 * The name of the permission should be either the (possibly wildcarded
 * (eg. *.company.com)) DNS style name of the of the host for which access is
 * being requested, or its IP address in standard nn.nn.nn.nn ("dot")
 * notation. The action list can be made up of any of the following:
 * <dl>
 * <dt>connect</dt>
 *		<dd>requests permission to connect to the host</dd>
 * <dt>listen</dt>
 *		<dd>requests permission to listen for connections from the host</dd>
 * <dt>accept</dt>
 *		<dd>requests permission to accept connections from the host</dd>
 * <dt>resolve</dt>
 *		<dd>requests permission to resolve the host name</dd>
 * </dl>
 * Note that "resolve" is implied when any (or none) of the others are present.
 * <p>
 * Access to a particular port can be requested by appending a colon
 * and a single digit to the name (eg. "*.company.com:7000"). A range of
 * port numbers can also be specified, by appending a pattern of the
 * form <low>-<high> where <low> and <high> are valid port numbers.
 * If either <low> or <high> is omitted it is equivalent to entering
 * the lowest or highest possible value respectively.
 * For example:
 * <pre>
 *     SocketPermission("www.company.com:7000-", "connect", "accept")
 * </pre>
 * represents permission to connect to and accept connections from
 * www.company.com on ports in the range 7000 to 65535.
 *
 * @author		OTI
 * @version		initial
 *
 */
public final class SocketPermission extends java.security.Permission
	implements java.io.Serializable
{
	private static final long serialVersionUID = -7204263841984476862L;

	//Bit masks for each of the possible actions
	static final int SP_CONNECT = 1;
	static final int SP_LISTEN = 2;
	static final int SP_ACCEPT = 4;
	static final int SP_RESOLVE = 8;
	// list of actions permitted for socket permission in order, indexed by mask value
	private static final String[] actionNames = {"","connect", "listen","", "accept","","","", "resolve"};

	//If a wildcard is present store the information
	private transient boolean isPartialWild = false;
	private transient boolean isWild = false;

	// The highest port number
	private static final int HIGHEST_PORT = 65535;
	// The lowest port number
	private static final int LOWEST_PORT = 0;

	transient String hostName; //Host name as returned by InetAddress
	transient String ipString; // IP address as returned by InetAddress
	transient boolean resolved = false; // IP address has been resolved

	// the port range;
	transient int portMin = LOWEST_PORT;
	transient int portMax = HIGHEST_PORT;

	private String actions; //List of all actions allowed by this permission
	transient int actionsMask = SP_RESOLVE;

/**
 * Constructs an instance of this class.
 * The host name can be a DNS name, an individual hostname, an ip address
 * or the empty string which implies localhost. The port or port range is optional.
 * <p>
 * The action list is a comma-seperated list which can consist of "connect", "listen",
 * "accept", and "resolve". They are case-insensitive and can be put together in
 * any order. "resolve" is always implied.
 *
 * @author		OTI
 * @version		initial
 *
 * @param host java.lang.String		the host name
 * @param action java.lang.String	the action string
 */
public SocketPermission(String host, String action) {
	super(host.equals("") ? "localhost" : host);
	hostName = getHostString(host);
	if (action == null)
		throw new NullPointerException();
	if (action.equals(""))
		throw new IllegalArgumentException();
	else
		setActions(action);
	actions = toCanonicalActionString(action);
	parsePort(host); //Use host since we are only checking for port presence
}

/**
 * Compares the argument to the receiver, and answers true
 * if they represent the equal objects using a class
 * specific comparison.
 * <p>
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o		the object to compare with this object
 * @return		<code>true</code>
 *					if the object is the same as this object
 *				<code>false</code>
 *					if it is different from this object
 * @see			#hashCode
 */

 public boolean equals(Object o) {
	if (this == o) return true;
	if (this.getClass() != o.getClass()) return false;
	SocketPermission sp = (SocketPermission) o;
	if (!hostName.equals(sp.hostName)) {
		if (getIPString() == null || !ipString.equals(sp.getIPString()))
			return false;
	}
	if (this.portMin != sp.portMin) return false;
	if (this.portMax != sp.portMax) return false;
	return this.actionsMask == sp.actionsMask;
}

/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>.equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int
 *					the receiver's hash.
 *
 * @see			#equals
 */
public int hashCode() {
		return hostName.hashCode() ^ actionsMask ^ portMin ^ portMax;
}

/**
 * Answers the canonical action list of this SocketPermission in the order:
 * connect, listen, accept, resolve.
 *
 * @author		OTI
 * @version		initial
 *
 * @return java.lang.String		the canonical action list
 */
public String getActions() {
	return actions;
}

/**
 * Stores the actions for this permission as a bit field
 *
 * @param actions java.lang.String		the action list
 */
private void setActions(String actions) throws IllegalArgumentException {
	if (actions.equals("")) return;
	boolean parsing = true;
	String action;
	StringBuffer sb = new StringBuffer();
	int pos = 0, length = actions.length();
	while(parsing) {
		char c;
		sb.setLength(0);
		while (pos < length && (c = actions.charAt(pos++)) != ',')
			sb.append(c);
		if (pos == length)
			parsing = false;
		action = sb.toString().trim().toLowerCase();
		if (action.equals(actionNames[SP_CONNECT]))
				actionsMask |= SP_CONNECT;
		else if (action.equals(actionNames[SP_LISTEN]))
				actionsMask |= SP_LISTEN;
		else if (action.equals(actionNames[SP_ACCEPT]))
				actionsMask |= SP_ACCEPT;
		else if (action.equals(actionNames[SP_RESOLVE])){}
		else
			// K0048 = Invalid action specified: {0}
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0048", action));
	}
}

/**
 * Check the permission to see if the actions requested by the argument permission
 * are permissable. All argument permission actions, host and port must be implied
 * by this permission in order to return true. This permission may imply additional
 * actions etc. not present in the argument permission.
 *
 * @author		OTI
 * @version		initial
 *
 * @return boolean		true if this permission implies <code>p</code>, and false otherwise
 * @param p java.security.Permission	the other socket permission
 */
public boolean implies(java.security.Permission p) {
	SocketPermission sp;
	try {
		sp = (SocketPermission)p;
	} catch (ClassCastException e) { return false; }

	// tests if the action list of p is the subset of the one of the receiver
	if ((actionsMask & sp.actionsMask) != sp.actionsMask) return false;

	// only check the port range if the action string of the current object
	// is not "resolve"
	if (!p.getActions().equals("resolve"))
		if ((sp.portMin < this.portMin) || (sp.portMax > this.portMax)) return false;

	//Verify the host is valid
	return checkHost(sp);
}

/**
 * Answers a PermissionCollection for storing SocketPermission objects.
 *
 * @author		OTI
 * @version		initial
 *
 * @return java.security.PermissionCollection	a permission collection
 */
public java.security.PermissionCollection newPermissionCollection() {
	return new SocketPermissionCollection();
}

/**
 * Parses the port string into the lower and higher bound of the port range.
 *
 * @author		OTI
 * @version		initial
 *
 */

private void parsePort(String hostString) throws IllegalArgumentException {
	int negidx = -1;
	int len = -1;
	int lastIdx = hostString.lastIndexOf(':');
	int idx = hostString.indexOf(':');
	int endOfIPv6Addr = hostString.lastIndexOf(']');
	if ((endOfIPv6Addr == -1) && (idx != lastIdx)) {
		// there are no square braces, but there are more than one ':'
		// which implies an IPv6 address with no port, or an illegal argument
		// check for valid IPv6 address
		if(Inet6Util.isValidIP6Address(hostString)) {
			return;
		} else {
			//throw an invalid argument exception
			// K004a = Invalid port number specified
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K004a"));
		}
	}
	// if there is a colon and it occurs after the ']' then there is a port to be parsed
	if ((lastIdx > -1) && (lastIdx > endOfIPv6Addr)){
		try{
			len = hostString.length();
			// look for a '-' after the colon
			negidx = hostString.indexOf('-',lastIdx);
			if (negidx == lastIdx + 1) {
				portMax = Integer.parseInt(hostString.substring(lastIdx + 2, len));
			} else {
				if (negidx != -1 && (negidx != len - 1)){ //A port range was provided
					portMin = Integer.parseInt(hostString.substring(lastIdx + 1, negidx));
					portMax = Integer.parseInt(hostString.substring(negidx + 1, len));
				}
				else{
					if (negidx == -1){
						portMin = Integer.parseInt(hostString.substring(lastIdx + 1, len));
						portMax = portMin;
					}
					else{
						portMin = Integer.parseInt(hostString.substring(lastIdx + 1, negidx));
					}
				}
			}
			if (portMax < portMin)
				// K0049 = MinPort is greater than MaxPort
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0049"));

			} catch (NumberFormatException e){
				// K004a = Invalid port number specified
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K004a"));
			}
	}
}

/**
 * Creates a canonical action list.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 	action java.lang.String
 *
 * @return 	java.lang.String
 */

private String toCanonicalActionString(String action) {
	if (action == null || action.equals("") || actionsMask == SP_RESOLVE)
		return actionNames[SP_RESOLVE]; //If none specified return the implied action resolve
	StringBuffer sb = new StringBuffer();
	if ((actionsMask & SP_CONNECT) == SP_CONNECT){
		sb.append(',');
		sb.append(actionNames[SP_CONNECT]);
		}
	if ((actionsMask & SP_LISTEN) == SP_LISTEN){
		sb.append(',');
		sb.append(actionNames[SP_LISTEN]);
		}
	if ((actionsMask & SP_ACCEPT) == SP_ACCEPT){
		sb.append(',');
		sb.append(actionNames[SP_ACCEPT]);
		}
	sb.append(',');
	sb.append(actionNames[SP_RESOLVE]);//Resolve is always implied
	return actions = sb.substring(1, sb.length()); //Don't copy the first ','.
}

private String getIPString() {
	if (!resolved) {
		try {
			ipString = InetAddress.getHostNameInternal(hostName);
		} catch (UnknownHostException e) {}
		resolved = true;
	}
	return ipString;
}

private String getHostString(String host) throws IllegalArgumentException {
	int idx = -1;
	idx = host.indexOf(':');
	if (isPartialWild = (host.length() > 0 && host.charAt(0) == '*')) {
		resolved = true;
		if (isWild = (host.length() == 1)) {
				return host;
		} else {
			if (idx > -1) {
				host = host.substring(0, idx);
			}
			return host.toLowerCase();
		}
	}

	int lastIdx = host.lastIndexOf(':');
	if ((idx > -1) && (idx == lastIdx)) {
		host = host.substring(0, idx);
	}
	else {
		// likely host is or contains an IPv6 address
		if (lastIdx != -1) {
			if(Inet6Util.isValidIP6Address(host)) {
				return host.toLowerCase();
			} else if (Inet6Util.isValidIP6Address(host.substring(0, lastIdx))) {
				host = host.substring(0, lastIdx);
			} else {
				// K004a = Invalid port number specified
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K004a"));
			}
		}
	}
	return host.toLowerCase();
}

/**
  * Determines whether or not this permission could refer to the same host as sp
  *
  */
boolean checkHost (SocketPermission sp) {
	if (isPartialWild)
		 if (isWild)
			return true; //Match on any host
		else {
			int length = hostName.length() - 1;
			return sp.hostName.regionMatches(
				sp.hostName.length() - length, hostName, 1, length);
		}
	// The ipString may not be the same, some hosts resolve to
	// multiple ips (on unix try www.ibm.com)
	return (getIPString() != null && ipString.equals(sp.getIPString())) ||
		hostName.equals(sp.hostName);
}

private void writeObject(ObjectOutputStream stream) throws IOException {
	stream.defaultWriteObject();
}

private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
	stream.defaultReadObject();
	//Initialize locals
	isPartialWild = false;
	isWild = false;
	portMin = LOWEST_PORT;
	portMax = HIGHEST_PORT;
	actionsMask = SP_RESOLVE;
	hostName = getHostString(getName());
	parsePort(getName());
	setActions(actions);
}

}
