
package java.security;

import java.security.spec.*;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides the functionality of an algorithm
 * parameter generator.
 *
 * @author		OTI
 * @version		initial
 */

public class AlgorithmParameterGenerator {

	/**
	 * Key prefix for algorithm name lookup.
	 */
	private static final String KEY_PREFIX = "AlgorithmParameterGenerator.";

	/**
	 * Name of the signature algorithm represented by the receiver.
	 */
 	private String algorithmName;

 	/**
 	 * Provider of the key algorithm represented by the receiver.
 	 */
	private Provider provider;

	/**
	 * The actual provider-supplied implementation.
	 */
	private AlgorithmParameterGeneratorSpi algorithmParameterGeneratorSpi;

/**
 * Constructs a new instance of this class with the given
 * algorithm parameter generator implementation, provider
 * and name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		paramGenSpi AlgorithmParameterGeneratorSpi
 *					The parameter generator implementation.
 * @param		provider Provider
 *					The provider of the parameter generator.
 * @param		algName String
 *					The name of the algorithm.
 */
protected AlgorithmParameterGenerator (
	AlgorithmParameterGeneratorSpi paramGenSpi,
	Provider provider,
	String algName)
{
	super();
	setProvider (provider);
	setAlgorithm(algName);
	algorithmParameterGeneratorSpi = paramGenSpi;
}

/**
 * Internal code that creates a new AlgorithmParameterGenerator for
 * a given AlgorithmParameterGeneratorSpi, algorithm class, and name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider Provider
 *					The provider that is supplying the implementation for the service.
 * @param		algorithmClass Class
 *					The class that implements the algorithm.
 * @param		algName String
 *					The name of the algorithm.
 * @return		AlgorithmParameterGenerator
 *					which matches the arguments.
 */
private static AlgorithmParameterGenerator createAlgorithmParameterGenerator (
	Provider provider,
	Class algorithmClass,
	String algName)
		throws NoSuchAlgorithmException
{
	try {
		AlgorithmParameterGeneratorSpi providedAlgorithmParameterGenerator =
			(AlgorithmParameterGeneratorSpi) algorithmClass.newInstance();
		return new AlgorithmParameterGenerator (
			providedAlgorithmParameterGenerator,
			provider,
			algName);
	} catch (IllegalAccessException e) {}
	  catch (InstantiationException e) {}
	throw new NoSuchAlgorithmException(algName);
}

/**
 * Answers the newly generated parameters.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		AlgorithmParameters
 *					the newly generated parameters.
 */
public final AlgorithmParameters generateParameters () {
	return algorithmParameterGeneratorSpi.engineGenerateParameters();
}

/**
 * Answers the standard Java Security name for the
 * algorithm being used by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the name of the algorithm
 */
public final String getAlgorithm() {
	return algorithmName;
}

/**
 * Answers a new AlgorithmParameterGenerator which is capable of
 * running the algorithm described by the argument.
 * <p>
 * The result will be an instance of AlgorithmParameterGenerator
 * which implements that algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					Name of the algorithm desired
 * @return		AlgorithmParameterGenerator
 *					a concrete implementation for the desired algorithm.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found
 */
public static AlgorithmParameterGenerator getInstance (String algorithmName)
 	throws NoSuchAlgorithmException
{
	// *** IMPORTANT ***
	// This method is declared to return AlgorithmParameterGenerator, while
	// providers are supposed to subclass AlgorithmParameterGeneratorSpi.
	// Thus, we cannot simply return an instance of a subclass of
	// AlgorithmParameterGeneratorSpi. This is why we have the instance
	// variable pointing to the actual implementation.

	if (algorithmName == null)
		throw new java.lang.IllegalArgumentException();

	return toAlgorithmParameterGeneratorImplementation(algorithmName);
}

/**
 * Answers a new AlgorithmParameterGenerator which is capable of
 * running the algorithm described by the argument, using the
 * specified provider.
 * <p>
 * The result will be an instance of AlgorithmParameterGenerator
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					Name of the algorithm desired
 * @param		providerName java.lang.String
 *					Name of the provider, which must implement the algorithm
 * @return		AlgorithmParameterGenerator
 *					a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found
 * @exception	NoSuchProviderException
 *					If the provider cannot be found
 */
public static AlgorithmParameterGenerator getInstance (
	String algorithmName,
	String providerName)
			throws NoSuchAlgorithmException, NoSuchProviderException
{
	// *** IMPORTANT ***
	// This method is declared to return AlgorithmParameterGenerator, while
	// providers are supposed to subclass AlgorithmParameterGeneratorSpi.
	// Thus, we cannot simply return an instance of a subclass of
	// AlgorithmParameterGeneratorSpi. This is why we have the instance
	// variable pointing to the actual implementation.

	if (providerName == null || algorithmName == null)
		throw new java.lang.IllegalArgumentException();

	Provider provider = Security.getProvider (providerName);
	if (provider == null)
		throw new NoSuchProviderException (providerName);

	return toAlgorithmParameterGeneratorImplementation(algorithmName, provider);
}

/**
 * Answers a new AlgorithmParameterGenerator which is capable of
 * running the algorithm described by the argument, using the
 * specified provider.
 * <p>
 * The result will be an instance of AlgorithmParameterGenerator
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithm java.lang.String
 *					Name of the algorithm desired
 * @param		provider  java.security.Provider
 *					The provider, which must implement the algorithm
 * @return		AlgorithmParameterGenerator
 *					a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found
 */
public static AlgorithmParameterGenerator getInstance( String algorithm, Provider provider)
 throws NoSuchAlgorithmException
{
	if( (algorithm == null) || (provider == null) ){
		throw new IllegalArgumentException();
	}

	return toAlgorithmParameterGeneratorImplementation(algorithm, provider);

}

/**
 * Answers the Provider of the algorithm generator represented
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public final Provider getProvider () {
	return provider;
}

/**
 * Initializes the receiver with the specified parameter size and
 * a default source of randomness.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		size int
 *					Size of the parameters
 */
public final void init (int size) {
	algorithmParameterGeneratorSpi.engineInit(size, new SecureRandom());
}

/**
 * Initializes the receiver with the specified parameter size and
 * source of randomness.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		size int
 *					Size of the parameters
 * @param		random SecureRandom
 *					New secure random to use
 */
public final void init (int size, SecureRandom random) {
	algorithmParameterGeneratorSpi.engineInit(size, random);
}

/**
 * Initializes the receiver with the specified parameters and
 * a default source of randomness.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		genParamSpec AlgorithmParameterSpec
 *					parameters for this algorithm
 */
public final void init (AlgorithmParameterSpec genParamSpec)
	throws InvalidAlgorithmParameterException
{
	algorithmParameterGeneratorSpi.engineInit(genParamSpec, new SecureRandom());
}

/**
 * Initializes the receiver with the specified parameters and
 * source of randomness.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		genParamSpec AlgorithmParameterSpec
 *					parameters for this algorithm
 * @param		random SecureRandom
 *					New secure random to use
 */
 public final void init (
 	AlgorithmParameterSpec genParamSpec,
	SecureRandom random)
		throws InvalidAlgorithmParameterException
{
	algorithmParameterGeneratorSpi.engineInit(genParamSpec, random);
}

/**
 * Set the algorithm being used by the receiver to
 * the argument which should be a standard Java
 * Security algorithm name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					the name of the algorithm
 */
void setAlgorithm (String algorithmName) {
	this.algorithmName = algorithmName;
}

/**
 * Set the provider being used by the receiver to
 * the argument, which should be an instance of a
 * subclass of Provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider Provider
 *					the Provider for the receiver
 */
void setProvider (Provider provider) {
	this.provider = provider;
}

/**
 * Answers an AlgorithmParameterGenerator for the named algorithm.
 * The result may be supplied by any provider which implements
 * the algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					the name of the algorithm to search for.
 * @return		AlgorithmParameterGenerator
 *					for the named algorithm.
 *.
 * @throws		NoSuchAlgorithmException
 *					If no provider implements the algorithm
 */
private static AlgorithmParameterGenerator toAlgorithmParameterGeneratorImplementation (
	String algorithmName)
		throws NoSuchAlgorithmException
{
	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toAlgorithmParameterGeneratorImplementation (algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			// Just skip to next provider
		}
	}
	// None of the available providers implement the algorithm.
	throw new NoSuchAlgorithmException(algorithmName);
}

/**
 * Answers a AlgorithmParameterGenerator for the named algorithm, as
 * supplied by the given provider. If the provider does not implement
 * the algorithm, throw a NoSuchAlgorithmException.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					the name of the algorithm to search for.
 * @param		provider java.security.Provider
 *					the provider which should implement the algorithm.
 *
 * @return		AlgorithmParameterGenerator
 *					for the algorithm name supplied by the given provider.
 *
 * @throws		NoSuchAlgorithmException
 *					If the provider does not implement the algorithm
 */
private static AlgorithmParameterGenerator toAlgorithmParameterGeneratorImplementation (
	String algorithmName,
	Provider provider)
		throws NoSuchAlgorithmException
{
	// First try to find the class corresponding to the algorithm name
	String algorithmParameterGeneratorClassName;
	try {
		algorithmParameterGeneratorClassName = provider.lookupProperty(KEY_PREFIX, algorithmName);
	} catch (ClassCastException e) {
		throw new NoSuchAlgorithmException(algorithmName);
	}

	// If not found, exception
	if (algorithmParameterGeneratorClassName == null)
		throw new NoSuchAlgorithmException(algorithmName);

	// Now try to instantiate the digest.
	try {
		Class algorithmParameterGeneratorClass = Class.forName (
			algorithmParameterGeneratorClassName,
			true,
			provider.getClass().getClassLoader());
		return createAlgorithmParameterGenerator (
			provider,
			algorithmParameterGeneratorClass,
			algorithmName);
	} catch (ClassNotFoundException ex) {
			throw new NoSuchAlgorithmException(algorithmName);
	}
}

}
