
package java.security;

import java.io.*;
import java.security.spec.*;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides functionality for dealing with the
 * parameters to the various security algorithms.
 *
 * @author		OTI
 * @version		initial
 */

public class AlgorithmParameters {

	/**
	 * Key prefix for algorithm name lookup
	 */
	private static final String KEY_PREFIX = "AlgorithmParameters.";

	/**
	 * Name of the signature algorithm represented by the receiver
	 */
 	private String algorithmName;

 	/**
 	 * Provider of the key algorithm represented by the receiver
 	 */
	private Provider provider;

	/**
	 * The actual provider-supplied implementation
	 */
	private AlgorithmParametersSpi algorithmParametersSpi;

	private boolean initialized = false;

/**
 * Constructs a new instance of this class capable of holding the
 * parameter for the specified algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		paramSpi AlgorithmParametersSpi
 *					The actual provider-supplied implementation.
 * @param		provider Provider
 *					The provider of the algorithm parameters.
 * @param		algName String
 *					Algorithm name.
 */
protected AlgorithmParameters (
	AlgorithmParametersSpi paramSpi,
	Provider provider,
	String algName)
{
	super();
	setProvider (provider);
	setAlgorithm(algName);
	algorithmParametersSpi = paramSpi;
}

/**
 * Internal code to creat an AlgorithmParameters for a given
 * AlgorithmParametersSpi as supplied by a particular provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider Provider
 *					The provider that is supplying the implementation
 *					 for the service.
 * @param		algorithmClass Class
 *					The class that implements the algorithm.
 * @param		algName String
 *					The name of the algorithm.
 * @return		AlgorithmParameters
 *					for a given AlgorithmParametersSpi from the provider.
 */
private static AlgorithmParameters createAlgorithmParameters (
	Provider provider,
	Class algorithmClass,
	String algName)
		throws NoSuchAlgorithmException
{
	try {
		AlgorithmParametersSpi providedDigest =
			(AlgorithmParametersSpi) algorithmClass.newInstance();
		AlgorithmParameters result =
			new AlgorithmParameters(providedDigest, provider, algName);
		return result;
	} catch (IllegalAccessException e) {}
	  catch (InstantiationException e) {}
	throw new NoSuchAlgorithmException(algName);
}

/**
 * Answers the standard Java security name for the
 * algorithm for which the receiver can hold the
 * arguments.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the name of the algorithm.
 */
public final String getAlgorithm () {
	return algorithmName;
}

/**
 * Answers the parameters in their primary encoding format.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]
 *					the parameters in their primary encoding format.
 *
 * @exception	IOException
 *					if encoding errors are detected.
 */
public final byte[] getEncoded () throws IOException{
	if (initialized)
		return algorithmParametersSpi.engineGetEncoded();
	throw new IOException();
}

/**
 * Answers the parameters in the specified format. If the
 * specified format is null, return them in the primary format.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		format String
 *					The desired encoding format name.
 * @return		byte[]
 *					the parameters in the specified encoding format.
 *
 * @exception	IOException
 *					if encoding errors are detected
 */
public final byte[] getEncoded (String format) throws IOException {
	// return in "primary encoding format" - which should be returned
	// by getEncoded()
	if( format == null ){
		return getEncoded();
	}

	if (initialized)
		return algorithmParametersSpi.engineGetEncoded(format);
	throw new IOException();
}

/**
 * Answers a new AlgorithmParameters for the algorithm described
 * by the argument. Throw NoSuchAlgorithmException if no provider
 * provides the named algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					Name of the desired algorithm.
 * @return		AlgorithmParameters
 *					a concrete implementation of a holder for
 *					the parameters for the desired algorithm.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found
 */
public static AlgorithmParameters getInstance (String algorithmName)
	throws NoSuchAlgorithmException
{
	// *** IMPORTANT ***
	// This method is declared to return AlgorithmParameters, while
	// providers are supposed to subclass AlgorithmParametersSpi. So,
	// we cannot simply return an instance of a subclass of
	// AlgorithmParametersSpi. This is why we have the instance
	// variable pointing to the actual implementation

	if (algorithmName != null)
		return toAlgorithmParametersImplementation(algorithmName);
	throw new IllegalArgumentException();
}

/**
 * Answers a new AlgorithmParameters which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of AlgorithmParameters
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					Name of the desired algorithm.
 * @param		providerName String
 *					Name of the provider which has to implement the algorithm.
 * @return		AlgorithmParameters
 *					a concrete implementation of a holder for
 *					the parameters for the desired algorithm.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found.
 * @exception	NoSuchProviderException
 *					If the provider cannot be found.
 */
public static AlgorithmParameters getInstance (
	String algorithmName,
	String providerName)
		throws	NoSuchAlgorithmException, NoSuchProviderException
{

	// *** IMPORTANT ***
	// This method is declared to return AlgorithmParameters, while
	// providers are supposed to subclass AlgorithmParametersSpi. So,
	// we cannot simply return an instance of a subclass of
	// AlgorithmParametersSpi. This is why we have the instance
	// variable pointing to the actual implementation

	if (providerName == null || algorithmName == null)
		throw new java.lang.IllegalArgumentException();
	Provider provider = Security.getProvider(providerName);
	if (provider != null)
		return toAlgorithmParametersImplementation(algorithmName, provider);

	throw new NoSuchProviderException(providerName);
}

/**
 * Answers a new AlgorithmParameters which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of AlgorithmParameters
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithm String
 *					Name of the desired algorithm.
 * @param		provider Provider
 *					Provider which has to implement the algorithm.
 * @return		AlgorithmParameters
 *					a concrete implementation of a holder for
 *					the parameters for the desired algorithm.
 *
 * @exception	NoSuchAlgorithmException
 *					If the algorithm cannot be found.
 */
public static AlgorithmParameters getInstance( String algorithm, Provider provider)
 throws NoSuchAlgorithmException
{
	if (algorithm == null || provider == null){
		throw new java.lang.IllegalArgumentException();
	}

	return toAlgorithmParametersImplementation( algorithm, provider );
}

/**
 * Answers the parameter specification for the receiver, instantiated
 * from the provided class.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		paramSpec Class
 *					Class to use when instantiating the result.
 *
 * @return		AlgorithmParameterSpec
 *					the parameter specification for the receiver.
 * @exception	InvalidParameterSpecException
 *					if the requested specification is not appropriate
 *					for the receiver.
 */
public final AlgorithmParameterSpec getParameterSpec (Class paramSpec)
	throws InvalidParameterSpecException
{
	if (initialized)
		return algorithmParametersSpi.engineGetParameterSpec(paramSpec);
	throw new InvalidParameterSpecException();
}

/**
 * Answers the Provider of the algorithm parameters represented
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider.
 */
public final Provider getProvider () {
	return provider;
}

/**
 * Initializes the receiver with the given parameters which
 * are encoded in the primary format.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		params byte[]
 *					Parameters encoded in the primary format for the receiver.
 * @exception	IOException
 *					if decoding errors are detected.
 */
public final void init (byte[] params) throws IOException {
	if (!initialized) {
		algorithmParametersSpi.engineInit(params);
		initialized = true;
	} else
		throw new IOException();
}

/**
 * Initializes the receiver with the parameters encoded in the
 * given named format.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		params byte[]
 *					Parameters encoded in the named format.
 * @param		format String
 *					name of the format in which the parameters are encoded.
 * @exception	IOException
 *					if decoding errors are detected.
 */
public final void init (byte[] params, String format) throws IOException {
	if (!initialized) {
		algorithmParametersSpi.engineInit(params, format);
		initialized = true;
	} else
		throw new IOException();
}

/**
 * Initializes the receiver with the given parameter specification.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		paramSpec AlgorithmParameterSpec
 *					The parameter specification to be used.
 * @exception	InvalidParameterSpecException
 *					if the provided specification is not appropriate
 *					for the receiver.
 */
public final void init (AlgorithmParameterSpec paramSpec)
	throws InvalidParameterSpecException
{
	if (!initialized) {
		algorithmParametersSpi.engineInit(paramSpec);
		initialized = true;
	} else {
		throw new InvalidParameterSpecException();
	}
}

/**
 * Sets the algorithm being used by the receiver to the argument
 * which should be the name of a standard Java security algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					the name of the algorithm.
 */
void setAlgorithm (String algorithmName) {
	this.algorithmName = algorithmName;
}

/**
 * Sets the provider being used by the receiver to the argument
 * which should be an instance of a subclass of Provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider
 *				 	the Provider for the receiver
 */
void setProvider (Provider provider) {
	this.provider = provider;
}

/**
 * Answers a new AlgorithmParameters which is capable of
 * holding the arguments for the named algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					Name of the algorithm which the result will
 *					be used for.
 * @return		AlgorithmParameters
 *					for the named algorithm.
 *
 * @exception	NoSuchAlgorithmException
 *					If there are no providers of the algorithm.
 */
private static AlgorithmParameters toAlgorithmParametersImplementation(
	String algorithmName)
		throws NoSuchAlgorithmException
{
	// *** NOTE *** - This method is declared to return AlgorithmParameters, while
	// providers are supposed to subclass AlgorithmParametersSpi. So, we
	// cannot simply return an instance of a subclass of AlgorithmParametersSpi.
	// This is why we have the instance variable pointing to the actual implementation

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toAlgorithmParametersImplementation (algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new NoSuchAlgorithmException(algorithmName);
}

/**
 * Answers an AlgorithmParameters capable of holding the arguments
 * for the named algorithm, if one can be supplied by the given
 * provider. Throw NoSuchAlgorithmException if one can not be
 * provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					Name of the algorithm which the result will
 *					be used for.
 * @param		provider java.security.Provider
 *					the provider desired for the algorithm.
 * @return		AlgorithmParameters
 *					for the named algorithm as supplied by
 *					the given provider.
 *
 * @throws		NoSuchAlgorithmException
 *					If the provider does not implement the algorithm
 */
private static AlgorithmParameters toAlgorithmParametersImplementation (
	String algorithmName,
	Provider provider)
		throws NoSuchAlgorithmException
{
	// First try to find the class corresponding to the algorithm name
	String algorithmParametersClassName;
	try {
		algorithmParametersClassName = provider.getProperty(KEY_PREFIX + algorithmName);
	} catch (ClassCastException e) {
		throw new NoSuchAlgorithmException(algorithmName);
	}

	// If not found, exception
	if (algorithmParametersClassName == null)
		throw new NoSuchAlgorithmException(algorithmName);

	// Now try to instantiate the object
	try {
		Class algorithmParametersClass = Class.forName (
			algorithmParametersClassName,
			true,
			provider.getClass().getClassLoader());
		return createAlgorithmParameters (
			provider,
			algorithmParametersClass,
			algorithmName);
	} catch (ClassNotFoundException ex) {
		throw new NoSuchAlgorithmException(algorithmName);
	}
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a printable representation for the receiver.
 */
public final String toString () {
	if (initialized)
		return algorithmParametersSpi.engineToString();

	return null;
}

}
