
package java.security;

import java.io.*;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 * This class implements a stream that computes a message digest hash as the
 * bytes are read from it.
 *
 * @author		OTI
 * @version		initial
 */

public class DigestInputStream extends java.io.FilterInputStream {

	/**
	 * 	The digest to use when computing the hash.
	 */
	protected MessageDigest digest;

	/**
	 * True if the digest should be computed for the next chunck
	 * of bytes read.
	 */
	private boolean on;

/**
 * Constructs a new DigestInputStream on an existing stream
 * with the given MessageDigest.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		in java.io.InputStream
 *					source of the bytes to digest.
 * @param		digest java.security.MessageDigest
 *					digest to use when computing the hash.
 *
 * @see			#on
 * @see			MessageDigest
 */
public DigestInputStream (java.io.InputStream in, MessageDigest digest) {
	super(in);
	setMessageDigest (digest);
	on (true);
}

/**
 * Answers the MessageDigest which the receiver uses
 * when computing the hash.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		MessageDigest
 *					the digest the receiver uses when
 *					computing the hash.
 *
 */
public MessageDigest getMessageDigest () {
	return digest;
}

/**
 * Enables or disables the digest function (default is on).
 *
 * @author		OTI
 * @version		initial
 *
 * @param		on boolean
 *					true if the digest should be computed,
 *					and false otherwise.
 *
 * @see	MessageDigest
 */
public void on (boolean on) {
	this.on = on;
}

/**
 * Reads the next byte and answers it as an int. Updates the digest
 * for the byte if this fuction is enabled.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		int
 *					the byte which was read or -1 at end of stream.
 *
 * @exception 	java.io.IOException
 *					If reading the source stream causes an IOException.
 */
public int read () throws IOException {
	int result = super.read();
	if (on && result >= 0)
		digest.engineUpdate ((byte) result);
	return result;
}

/**
 * Reads at most <code>count</code> bytes from the Stream and stores them in the
 * byte array <code>buffer</code> starting at <code>offset</code>. Answer the
 * number of bytes actually read or -1. Updates the digest for the bytes being
 * read if this fuction is enabled.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer byte[]
 *					the byte array in which to store the read bytes.
 * @param		offset int
 *					the offset in <code>buffer</code> to store the read bytes.
 * @param		count int
 *					the maximum number of bytes to store in <code>buffer</code>.
 * @return 		int
 *					the number of bytes actually read or -1 if end of stream.
 *
 * @exception 	java.io.IOException
 *					If reading the source stream causes an IOException.
 */

public int read (byte[] buffer, int offset, int count) throws IOException {
	int read = super.read(buffer, offset, count);
	if (on && read > 0)
		digest.engineUpdate (buffer, offset, read);
	return read;
}

/**
 * Sets the MessageDigest which the receiver will use
 * when computing the hash.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		digest MessageDigest
 *					the digest to use when computing the hash.
 *
 * @see			MessageDigest
 * @see			#on
 */
public void setMessageDigest (MessageDigest digest) {
	this.digest = digest;
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					a printable representation for the receiver.
 */
public String toString () {
	StringBuffer answer = new StringBuffer("[Digest Input Stream]");
	if (digest != null) {
		answer.append(" ");
		answer.append(digest.toString());
	}
	if (on)
		answer.append(", <on>");
	else
		answer.append(", <off>");
	return answer.toString();
}

}
