
package java.security;

import java.util.Arrays;
import java.util.Vector;
import java.io.Serializable;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2001, 2006  All Rights Reserved
 */

/**
 * Identities represent entities that have a public key to authenticate them.
 * It has a set of certificates linking this identity's public key and principals.
 *
 * @author		OTI
 * @version		initial
 *
 * @deprecated 	Use java.security.KeyStore, java.security.cert package, and java.security.Principal
 */
public abstract class Identity implements Principal, Serializable {
	private static final long serialVersionUID = 3609922007826600659L;

	String name;
	PublicKey publicKey;
	Vector certificates;
	IdentityScope scope;
	String info;

/**
 * Constructs a new instance of this class.
 *
 */
protected Identity() {
	certificates = new Vector();
}
/**
 * Constructs a new instance of this class with the name provided.
 *
 * @param		name		the name of this identity.
 */
public Identity(String name) {
	this.name = name;
	certificates = new Vector();
}
/**
 * Constructs a new instance of this class with the name and scope provided.
 *
 * @param		name		the name of this identity.
 * @param		scope		the scope of this identity.
 *
 * @throws		KeyManagementException	if an identity with this name already exists in this scope
 */
public Identity(String name, IdentityScope scope) throws KeyManagementException {
	this(name);
	this.scope = scope;
	if (scope != null) {
		scope.addIdentity(this);
	}
}
/**
 * Returns the scope of this identity.
 *
 * @return		the scope of this identity.
 */
public final IdentityScope getScope() {
	return scope;
}
/**
 * Returns the public key of this identity.
 *
 * @return		the public key of this public key.
 */
public PublicKey getPublicKey() {
	return publicKey;
}
/**
 * Sets the public key of this identity.
 *
 * @param		publicKey		the public key to assign to this identity.
 *
 * @throws		KeyManagementException	an identity in the scope has the same public key
 */
public void setPublicKey(PublicKey publicKey) throws KeyManagementException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("setIdentityPublicKey");
	if (scope!=null && scope.getIdentity(publicKey) != null)
		// K0194 = An identity in the scope has the same public key
		throw new KeyManagementException(com.ibm.oti.util.Msg.getString("K0194"));
	this.publicKey = publicKey;
	this.certificates = new Vector();
}
/**
 * Returns the name of this identity.
 *
 * @return		the name of this identity.
 */
public final String getName() {
	return name;
}
/**
 * Returns the general information about this identity.
 *
 * @return		the general information about this identity.
 */
public String getInfo() {
	return info;
}
/**
 * Sets the general information about this identity.
 *
 * @param		info		the information about this identity.
 */
public void setInfo(String info) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("setIdentityInfo");
	this.info = info;
}
/**
 * Returns the set of certificates of this identity.
 *
 * @return		the certificates of this identity.
 */
public Certificate[] certificates() {
	return (Certificate[])certificates.toArray(new Certificate[0]);
}
/**
 * Adds the given certificate to the set of certificates.
 *
 * @param		cert		the certificate to add.
 *
 * @throws		KeyManagementException	if the public key does not match the one on the certificate
 */
public void addCertificate(Certificate cert) throws KeyManagementException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("addIdentityCertificate");
	PublicKey certPublicKey = cert.getPublicKey();
	if (publicKey!=null && !Arrays.equals(publicKey.getEncoded(), certPublicKey.getEncoded()))
		// K0195 = The existing public key and the one contained in the certificate do not match.
		throw new KeyManagementException(com.ibm.oti.util.Msg.getString("K0195"));
	this.publicKey = certPublicKey;
	certificates.add(cert);
}
/**
 * Removes the given certificate from the set of certificates.
 *
 * @param		cert		the certificate to remove.
 *
 * @throws		KeyManagementException	if the certificate is missing
 */
public void removeCertificate(Certificate cert) throws KeyManagementException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("removeIdentityCertificate");
	if (!certificates.contains(cert))
		// K0196 = Certificate is missing
		throw new KeyManagementException(com.ibm.oti.util.Msg.getString("K0196"));
	certificates.remove(cert);
}
/**
 * Returns true if this identity and the given object are equal.
 *
 * @param		identity	the object to compare to.
 * @return		true if this identity and the given object are equal.
 */
public final boolean equals(Object identity) {
	// first check if this and identity refer to the same object
	if (this == identity)
		return true;
	if (!(identity instanceof Identity))
		return false;
	Identity converted = (Identity) identity;
	boolean sameName = getName() == null ? converted.getName() == null : getName().equals(converted.getName());
	if (sameName) {
		boolean sameScope = getScope() == null ? converted.getScope() == null : getScope().equals(converted.getScope());
		if (sameScope) return true;
	}
	return identityEquals(converted);
}
/**
 * Returns true if this identity and the given identity are equal.
 *
 * @return		true if this identity and the given identity are equal.
 */
protected boolean identityEquals(Identity identity) {
	if (identity == null)
		return false;
	boolean sameName = getName() == null ? identity.getName() == null : getName().equals(identity.getName());
	if (!sameName) return false;
	boolean sameKey = getPublicKey() == null ? identity.getPublicKey() == null : getPublicKey().equals(identity.getPublicKey());
	return sameKey;
}
/**
 * Returns a String representation of this identity.
 *
 * @return	String	the String representation of this identity.
 */
public String toString() {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("printIdentity");
	return "Name : " + this.getName() + "\nPublic key : " + this.publicKey;
}
/**
 * Returns a String representation of this identity, detailed if detail is true.
 *
 * @param		detail		if true a detailed string representation is returned.
 * @return		the string representation of this identity.
 */
public String toString(boolean detail) {
	String output = toString();
	if (detail) {
		StringBuffer result = new StringBuffer();
		result.append ("\nGeneral information : ");
		result.append (getInfo());
		result.append ("\nCertificates :\n");
		Certificate[] certs = certificates();
		for (int i=0;i<certs.length;i++)
			result.append(certs[i].toString(false));
		output += result.toString();
	}
	return output;
}
/**
 * Returns a hashcode of this identity.
 *
 * @return		the hashcode of this identity.
 */
public int hashCode() {
	int hashCode = 0;
	if (name != null)
		hashCode = name.hashCode();
	if (publicKey != null)
		hashCode ^= publicKey.hashCode();
	if (scope != null)
		hashCode ^= scope.hashCode();
	return hashCode;
}
}
