
package java.security;

import java.security.spec.*;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * Instances of this class are used to convert back and forth
 * between keys and key specifications.
 *
 * @author		OTI
 * @version		initial
 */

public class KeyFactory {

	/**
	 * Key prefix for algorithm name lookup.
	 */
	private static final String KEY_PREFIX = "KeyFactory.";

	/**
	 * Name of the signature algorithm represented by the receiver.
	 */
 	private String algorithmName;

 	/**
 	 * Provider of the key algorithm represented by the receiver.
 	 */
 	private Provider provider;

 	/**
 	 * The actual provider-supplied implementation.
 	 */
	private KeyFactorySpi keyFactorySpi;

/**
 * Constructs a new instance of this class with the given
 * implementation, provider and named algorithm.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		keyFactorySpi KeyFactorySpi
 *					The actual key factory implementation.
 * @param		provider Provider
 *					The provider of the key factory.
 * @param		algName String
 *					Algorithm name.
 */
protected
KeyFactory (KeyFactorySpi keyFactorySpi, Provider provider, String algName) {
	super();
	setProvider (provider);
	setAlgorithm(algName);
	this.keyFactorySpi = keyFactorySpi;
}

/**
 * Creates a KeyFactory for a given KeyFactorySpi from a provider
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider	The provider that is supplying the implementation for the service
 * @param		algorithmClass	The class that implements the algorithm
 * @param		algName			The name of the algorithm
 *
 * @return		a KeyFactory for a given KeyFactorySpi from a provider
 */
private static KeyFactory createKeyFactory(Provider provider, Class algorithmClass, String algName) throws NoSuchAlgorithmException {
	try {
		KeyFactorySpi providedKeyFactory = (KeyFactorySpi) algorithmClass.newInstance();
		return new KeyFactory (providedKeyFactory, provider, algName);
	} catch (IllegalAccessException e) {}
	  catch (InstantiationException e) {}

	throw new NoSuchAlgorithmException(algName);
}
/**
 * Generates a private key for the given key specification.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		keySpec KeySpec	the key specification to be used for key generation.
 *
 * @exception	InvalidKeySpecException 	if the provided key specification is not appropriate
 */
 public final PrivateKey generatePrivate(KeySpec keySpec) throws InvalidKeySpecException {
	return keyFactorySpi.engineGeneratePrivate(keySpec);
}
/**
 * Answers the standard Java Security name for the
 * algorithm being used by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the name of the algorithm
 */
public final String getAlgorithm() {
	return algorithmName;
}

/**
 * Answers a new KeyFactory which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of KeyFactory
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @param		providerName java.lang.String	Name of the provider which has to implement the algorithm
 * @return		KeyFactory	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 * @exception	NoSuchProviderException		If the provider cannot be found
 */
 public static KeyFactory getInstance(String algorithmName,
									 String providerName)
							  throws NoSuchAlgorithmException,
									 NoSuchProviderException {
	// *** NOTE *** - This method is declared to return KeyFactory, while
	// providers are supposed to subclass KeyFactorySpi. So, we
	// cannot simply return an instance of a subclass of KeyFactorySpi.
	// This is why we have the instance variable pointing to the actual implementation

	if (providerName==null) throw new java.lang.IllegalArgumentException();
	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	if( providerName.equals("") ){
		throw new IllegalArgumentException();
	}

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toKeyFactoryImplementation(algorithmName, provider);
 }

 /**
  * Answers a new KeyFactory which is capable of
  * running the algorithm described by the argument.
  * The result will be an instance of KeyFactory
  * which implements that algorithm.
  *
  *
  * @author		OTI
  * @version		initial
  *
  * @param		algorithm java.lang.String	Name of the algorithm desired
  * @param		provider Provider Provider which has to implement the algorithm
  * @return		KeyFactory	a concrete implementation for the algorithm desired.
  *
  * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
  */
public static KeyFactory getInstance(String algorithm, Provider provider)
 throws NoSuchAlgorithmException
{
 	if( (algorithm == null ) || (provider == null) ){
 		throw new IllegalArgumentException();
 	}

 	return toKeyFactoryImplementation(algorithm, provider);
}

/**
 * Answers a key specification for a given key.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key 		Key		the key for which the specification is to be computed.
 * @param		keySpec 	Class	the class of the result specification
 *
 * @return		a key specification for a given key. The class of the result object will be <code>keySpec</code>.
 * @exception	InvalidKeySpecException 	if the provided key specification is not appropriate
 */
public final KeySpec getKeySpec(Key key, Class keySpec) throws InvalidKeySpecException {
	return keyFactorySpi.engineGetKeySpec(key,keySpec);
}
/**
 * Returns the Provider of the key factory represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public final Provider getProvider() {
	return provider;
}

/**
 * Set the algorithm being used by the receiver to
 * the argument which should be a standard Java
 * Security algorithm name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					the name of the algorithm
 */
void setAlgorithm(String algorithmName) {
	this.algorithmName = algorithmName;
}
/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}

/**
 * Answers a KeyFactory for the algorithm name supplied by any provider.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					the name of the algorithm to search for
 *
 * @return		The KeyFactory for the algorithm name supplied by any provider
 *.
 * @throws NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static KeyFactory toKeyFactoryImplementation (String algorithmName) throws NoSuchAlgorithmException{

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toKeyFactoryImplementation (algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new NoSuchAlgorithmException(algorithmName);
}

/**
 * Answers a KeyFactory for the algorithm name supplied by the given provider.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName	java.lang.String
 *					the name of the algorithm to search for
 * @param		provider		java.security.Provider
 *					the provider desired for the algorithm.
 *
 * @return		The KeyFactory for the algorithm name supplied by the given provider.
 *
 * @throws NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static KeyFactory toKeyFactoryImplementation (
		String algorithmName,Provider provider) throws NoSuchAlgorithmException {

	// First try to find the class corresponding to the algorithm name
	String keyFactoryClassName;
	try {
		keyFactoryClassName = provider.lookupProperty(KEY_PREFIX, algorithmName);
	} catch (ClassCastException e) {
		throw new NoSuchAlgorithmException(algorithmName);
	}

	// If not found, exception
	if (keyFactoryClassName == null)
		throw new NoSuchAlgorithmException(algorithmName);

	// Now try to instantiate the digest.
	try {
		Class keyFactClass = Class.forName(keyFactoryClassName, true, provider.getClass().getClassLoader());
		return createKeyFactory (provider, keyFactClass, algorithmName);
	} catch (ClassNotFoundException ex) {
			throw new NoSuchAlgorithmException(algorithmName);
	}
}

/**
 * Translates a generic key (any provider, trusted or not) to an instance of
 * this key factory.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key 		Key		the key for which the new key is to be computed.
 *
 * @return		a key for the given key.
 * @exception	InvalidKeyException if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
public final Key translateKey(Key key) throws InvalidKeyException {
	return keyFactorySpi.engineTranslateKey(key);
}

/**
 * Generates a public key for the given key specification.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		keySpec KeySpec	the key specification to be used for key generation.
 *
 * @exception	InvalidKeySpecException 	if the provided key specification is not appropriate
 */
public final PublicKey generatePublic(KeySpec keySpec) throws InvalidKeySpecException {
	return keyFactorySpi.engineGeneratePublic(keySpec);
}

/**
 * Answers a new KeyFactory which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of KeyFactory
 * which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @return		KeyFactory	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 */
 public static KeyFactory getInstance(String algorithmName) throws NoSuchAlgorithmException {
	// *** NOTE *** - This method is declared to return KeyFactory, while
	// providers are supposed to subclass KeyFactorySpi. So, we
	// cannot simply return an instance of a subclass of KeyFactorySpi.
	// This is why we have the instance variable pointing to the actual implementation

	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	return toKeyFactoryImplementation(algorithmName);
 }

}
