
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides the functionality of a KeyPairGenerator algorithm.
 * This class is is abstract and extends from KeyPairGeneratorSpi to be compatible with
 * JDK, which does it for historical reasons (check the JDK's Javadoc files).
 *
 * @author		OTI
 * @version		initial
 */

 import java.security.spec.*;

 public abstract class KeyPairGenerator extends KeyPairGeneratorSpi {
 	private static final String KEY_PREFIX = "KeyPairGenerator.";	// Key prefix for algorithm name lookup

 	private String algorithmName;	// Name of the KeyPairGenerator algorithm represented by the receiver
	private Provider provider;		// Provider of the KeyPairGenerator algorithm represented by the receiver
	// *** WARNING *** - The provider is not passed as parameter to the constructor. It means
	// it will be uninitialized until an actual binding to a concrete class by a provider
	// is performed, in getInstance.

	static private class Wrapper extends KeyPairGenerator {
		KeyPairGeneratorSpi	keyPairGeneratorProvider;

		Wrapper (KeyPairGeneratorSpi keyPairGeneratorProvider, String algorithmName) {
			super(algorithmName);	// Just because there is no empty constructor in the superclass
			this.keyPairGeneratorProvider = keyPairGeneratorProvider;
		};

		public KeyPair generateKeyPair() {
			return keyPairGeneratorProvider.generateKeyPair();
		};

		public void initialize(int keySize, SecureRandom secureRandom) {
			keyPairGeneratorProvider.initialize(keySize, secureRandom);
		};

		 public void initialize(AlgorithmParameterSpec params, SecureRandom secureRandom) throws InvalidAlgorithmParameterException {
			keyPairGeneratorProvider.initialize (params, secureRandom);
		};
	}
/**
 * Constructs a new instance of this class with the algorithm provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	algName	String	Algorithm name
 */
 protected KeyPairGenerator(String algName) {
	super();
	setAlgorithm (algName);
}
/**
 * Answers a newly generated key pair.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		a newly generated key pair.
 *
 */
 public final KeyPair genKeyPair() {
	 return generateKeyPair();
 }
/**
 * Answers the standard Java Security name for the
 * algorithm being used by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the name of the algorithm
 */
public String getAlgorithm() {
	return algorithmName;
}
/**
 * Answers a new KeyPairGenerator which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * KeyPairGenerator which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @return		KeyPairGenerator	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static KeyPairGenerator getInstance(String algorithmName) throws NoSuchAlgorithmException {
	// *** NOTE *** - This method is declared to return KeyPairGenerator, while
	// providers are supposed to subclass KeyPairGeneratorSpi. So, we
	// cannot simply return an instance of a subclass of KeyPairGeneratorSpi.
	// This is why we have the inner type Wrapper

	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	return toKeyPairGeneratorImplementation(algorithmName);
}
/**
 * Answers a new KeyPairGenerator  which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * KeyPairGenerator  which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @param		providerName java.lang.String	Name of the provider which has to implement the algorithm
 * @return		KeyPairGenerator 	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 * @exception	NoSuchProviderException		If the provider cannot be found
 */
public static KeyPairGenerator  getInstance(String algorithmName,
									String providerName)
							 throws NoSuchAlgorithmException,
									NoSuchProviderException {

	// *** NOTE *** - This method is declared to return KeyPairGenerator , while
	// providers are supposed to subclass KeyPairGenerator Spi. So, we
	// cannot simply return an instance of a subclass of KeyPairGenerator.
	// This is why we have the inner type Wrapper

	if (providerName==null) throw new java.lang.IllegalArgumentException();
	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	if( providerName.equals("") ){
		throw new IllegalArgumentException();
	}

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toKeyPairGeneratorImplementation(algorithmName, provider);
}

/**
 * Answers a new KeyPairGenerator  which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * KeyPairGenerator  which implements that algorithm.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithm java.lang.String	Name of the algorithm desired
 * @param		provider Provider Provider which has to implement the algorithm
 * @return		KeyPairGenerator 	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static KeyPairGenerator getInstance(String algorithm, Provider provider)
 throws NoSuchAlgorithmException
{
 	if( (algorithm == null ) || (provider == null) ){
 		throw new IllegalArgumentException();
 	}

 	return toKeyPairGeneratorImplementation(algorithm, provider);
}

/**
 * Returns the Provider of the key pair generation represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public final Provider getProvider() {
	return provider;
}
/**
 * Initialize the receiver to use a default secure random generator,
 * and generate keys with the given size.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @param	keySize	size of keys, in bits
 *
 */
public void initialize(int keySize) {
	initialize (keySize, new SecureRandom());
}
/**
 * Initialize the receiver to use a given secure random generator,
 * and generate keys keys with the given size.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @param	keySize	size of keys, in bits
 * @param	secureRandom	SecureRandom			New secure random to use
 *
 */
 public void initialize(int keySize, SecureRandom secureRandom) {
	throw new UnsupportedOperationException ();	// Subclasses must override
 }
/**
 * Initialize the receiver to use a default secure random generator,
 * and generate keys from the provided set of parameters.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @param	params	AlgorithmParameterSpec	parameters for key generation
 *
 */
public void initialize(AlgorithmParameterSpec params) throws InvalidAlgorithmParameterException {
	initialize (params, new SecureRandom());
}
/**
 * Initialize the receiver to use a given secure random generator,
 * and generate keys from the provided set of parameters.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @param	params	AlgorithmParameterSpec	parameters for key generation
 * @param	secureRandom	SecureRandom			New secure random to use
 *
 */
 public void initialize(AlgorithmParameterSpec params, SecureRandom secureRandom) throws InvalidAlgorithmParameterException {
	super.initialize (params, secureRandom);	// See JDK's Javadoc for class KeyPairGenerator
 }
/**
 * Set the algorithm being used by the receiver to
 * the argument which should be a standard Java
 * Security algorithm name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName String
 *					the name of the algorithm
 */
void setAlgorithm(String algorithmName) {
	this.algorithmName = algorithmName;
}
/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}
/**
 * Answers a KeyPairGenerator  for the algorithm name supplied by any provider.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName java.lang.String
 *					the name of the algorithm to search for
 *
 * @return		The KeyPairGenerator for the algorithm name supplied by any provider
 *.
 * @throws NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static KeyPairGenerator toKeyPairGeneratorImplementation (String algorithmName) throws NoSuchAlgorithmException{

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toKeyPairGeneratorImplementation (algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new NoSuchAlgorithmException(algorithmName);
}
/**
 * Answers a KeyPairGenerator for the algorithm name supplied by the given provider.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		algorithmName	java.lang.String
 *					the name of the algorithm to search for
 * @param		provider		java.security.Provider
 *					the provider desired for the algorithm.
 *
 * @return		The KeyPairGenerator for the algorithm name supplied by the given provider.
 *
 * @throws NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static KeyPairGenerator toKeyPairGeneratorImplementation (
		String algorithmName,
		Provider provider) throws NoSuchAlgorithmException {

	// First try to find the class corresponding to the algorithm name
	String keyPairGeneratorClassName;
	try {
		keyPairGeneratorClassName = provider.lookupProperty(KEY_PREFIX, algorithmName);
	} catch (ClassCastException e) {
		throw new NoSuchAlgorithmException(algorithmName);
	}

	// If not found, exception
	if (keyPairGeneratorClassName == null)
		throw new NoSuchAlgorithmException(algorithmName);

	// Now try to instantiate the digest.
	try {
		Class keyPairGeneratorClass = Class.forName(keyPairGeneratorClassName, true, provider.getClass().getClassLoader());
		KeyPairGeneratorSpi providedKeyPairGenerator = (KeyPairGeneratorSpi) keyPairGeneratorClass.newInstance();
		KeyPairGenerator generator;
		if (providedKeyPairGenerator instanceof KeyPairGenerator)
			generator = (KeyPairGenerator) providedKeyPairGenerator;
		else
			generator = new Wrapper(providedKeyPairGenerator, algorithmName);
		generator.setProvider(provider);
		return generator;
	} catch (ClassNotFoundException ex) {
	} catch (IllegalAccessException e) {
	} catch (InstantiationException e) {
	} catch (ClassCastException e) {
	}
	throw new NoSuchAlgorithmException(algorithmName);
}

/**
 * Answers a newly generated key pair.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a newly generated key pair.
 *
 */
public KeyPair generateKeyPair() {
	return null;
}
}
