
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides the functionality of a keystore (a simple
 * "database" of keys and certificates).
 *
 * @author		OTI
 * @version		initial
 */

 import java.util.*;
 import java.io.*;
 import java.security.cert.*;
 import com.ibm.oti.util.PriviAction;

 public class KeyStore {
	private static final String KEY_PREFIX = "KeyStore.";	// Key prefix for service lookup //$NON-NLS-1$

 	private String type;				// Type of keystore represented by the receiver
	private Provider provider;			// Provider of the keystore represented by the receiver
	private KeyStoreSpi keyStoreSpi;	// The actual provider-supplied implementation
	private boolean loaded = false;

/**
 * Constructs a new instance of this class with the given keystore type.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	keyStoreSpi		KeyStoreSpi		The actual keystore implementation
 * @param	provider		Provider		The provider of the key factory
 * @param	type			String			The keystore type
 */
protected KeyStore(KeyStoreSpi keyStoreSpi, Provider provider, String type) {
	super();
	setProvider (provider);
	setType(type);
	this.keyStoreSpi = keyStoreSpi;
}
/**
 * Answers the alias names for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	Enumeration	All alias names for the receiver
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final Enumeration aliases()  throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineAliases();
}
/**
 * Checks if the given alias exists in the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	true if the alias exists, false otherwise
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final boolean containsAlias(String alias) throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineContainsAlias(alias);
}
/**
 * Creates a KeyStore for a given KeyStoreSpi from a provider
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider		Provider	The provider that is supplying the implementation for the service
 * @param		keyStoreClass	Class		The class that implements the type wanted
 * @param		type			String		The type desired
 *
 * @return		a KeyStore for a given KeyStoreSpi from a provider
 */
private static KeyStore createKeyStore(Provider provider, Class keyStoreClass, String type) throws KeyStoreException {
	try {
		KeyStoreSpi providedKeyStore = (KeyStoreSpi) keyStoreClass.newInstance();
		return new KeyStore (providedKeyStore, provider, type);
	} catch (IllegalAccessException e) {
		// Ignore exception}
	} catch (InstantiationException e) {
		// Ignore exception
	}
	throw new KeyStoreException(type);
}
/**
 * Deletes as given alias entry from the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be deleted.
 *
 * @exception	KeyStoreException 	If the alias cannot be removed
 */
 public final void deleteEntry(String alias) throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	keyStoreSpi.engineDeleteEntry(alias);
 }

/**
 * Answers a certificate associated with the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Certificate	a certificate associated with the given alias.
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final java.security.cert.Certificate getCertificate(String alias) throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineGetCertificate(alias);
}

/**
 * Answers the alias for a given Certificate.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		cert	Certificate		The Certificate to be searched for.
 *
 * @return	String	The alias for the given Certificate, null if the alias does not exist or if the
 *					Certificate does not exist in the receiver.
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final String getCertificateAlias(java.security.cert.Certificate cert)  throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineGetCertificateAlias(cert);
}
/**
 * Answers the Certificate chain associated with teh given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Certificate[]	a chain of certificates associated with the given alias.
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final java.security.cert.Certificate[] getCertificateChain(String alias) throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineGetCertificateChain(alias);
}

/**
 * Answers the creation date for the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Date	The date when the entry for this alias was created,
 *					null if the alias does not exist
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 */
public final Date getCreationDate(String alias)   throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineGetCreationDate(alias);
}
/**
 * Answers a new KeyStore of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of keystore desired
 * @return		KeyStore	a concrete implementation for the KeyStore type desired.
 *
 * @exception	KeyStoreException 	If the type cannot be found
 */
public static KeyStore getInstance(String type) throws KeyStoreException  {
	// *** NOTE *** - This method is declared to return KeyStoreException , while
	// providers are supposed to subclass KeyStoreException Spi. So, we
	// cannot simply return an instance of a subclass of KeyStoreException Spi.
	// This is why we have the instance variable pointing to the actual implementation

	if (type==null) throw new java.lang.IllegalArgumentException();

	return toKeyStoreImplementation(type);
 }
/**
 * Answers a new KeyStore of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of keystore desired
 * @param		providerName java.lang.String	Name of the provider which has to implement the algorithm
 * @return		KeyStore	a concrete implementation for the KeyStore type desired.
 *
 * @exception	KeyStoreException 	If the type cannot be found
 * @exception	NoSuchProviderException		If the provider cannot be found
 */
public static KeyStore getInstance(String type, String providerName)
											throws KeyStoreException, NoSuchProviderException {
	// *** NOTE *** - This method is declared to return KeyStoreException , while
	// providers are supposed to subclass KeyStoreException Spi. So, we
	// cannot simply return an instance of a subclass of KeyStoreException Spi.
	// This is why we have the instance variable pointing to the actual implementation

	if (providerName==null) throw new java.lang.IllegalArgumentException();
	if (type==null) throw new java.lang.IllegalArgumentException();

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toKeyStoreImplementation(type,provider);
 }

/**
 * Answers a new KeyStore of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of keystore desired
 * @param		provider Provider Provider which has to implement the algorithm
 * @return		KeyStore	a concrete implementation for the KeyStore type desired.
 *
 * @exception	KeyStoreException 	If the type cannot be found
 */
public static KeyStore getInstance(String type,Provider provider)
 throws KeyStoreException
{
 	if( (type == null ) || (provider == null) ){
 		throw new IllegalArgumentException();
 	}

 	return toKeyStoreImplementation(type, provider);
}

/**
 * Answers the key associated with the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 * @param		password	char[]		Password to use when fetching the key for the alias
 *
 * @return	Key		key associated with the given alias,
 *					null if the alias does not exist
 *
 * @exception	KeyStoreException		 	If the keystore has not been initialized
 * @exception	NoSuchAlgorithmException 	If the algorithm to recover the key cannot be found
 * @exception	UnrecoverableKeyException 	If the key can't be recovered
 */
public final Key getKey(String alias,	char[] password)
                 throws KeyStoreException,
                        NoSuchAlgorithmException,
                        UnrecoverableKeyException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineGetKey(alias, password);
}
/**
 * Returns the Provider of the keystore represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public final Provider getProvider() {
	return provider;
}
/**
 * Returns the keystore type
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String	type of keystore being used
 */
public final String getType() {
	return type;
}
/**
 * Answers whether the alias entry represents a trusted certificate entry
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 *
 * @return	boolean		true if the alias entry represents a trusted certificate entry,
 *						false otherwise
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 *
 */
public final boolean isCertificateEntry(String alias) throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineIsCertificateEntry(alias);
}

/**
 * Answers whether the alias entry represents a key entry
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 *
 * @return	boolean		true if the alias entry represents a key entry,
 *						false otherwise
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 *
 */
public final boolean isKeyEntry(String alias) throws KeyStoreException {
	if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineIsKeyEntry(alias);
}
/**
 * Loads the keystore from the given input stream, using the password
 * provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	stream		InputStream		stream from where to read the keystore
 *
 * @exception	IOException					An IO problem was found when reading the keystore
 * @exception	NoSuchAlgorithmException	The algorithm to check the integrity of the keystore cannot be found
 * @exception	CertificateException		If a certificate in the keystore could not be loaded.
 *
 */
public final void load(InputStream stream, char[] password)
					 throws IOException,
							NoSuchAlgorithmException,
							CertificateException {

	keyStoreSpi.engineLoad(stream, password);
	loaded = true;
}
/**
 * Assigns a certificate to an alias
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias	String			an alias for the Certificate
 * @param	cert	Certificate		The Certificate for which an alias is being registered
 *
 * @exception	KeyStoreException	if the alias already exists for an untrusted Certificate
 *
 */
public final void setCertificateEntry(String alias, java.security.cert.Certificate cert)
										throws KeyStoreException {

    if (!loaded) throw new KeyStoreException();
	keyStoreSpi.engineSetCertificateEntry(alias, cert);
}

/**
 * Assigns the given key to an alias. If the key is a PrivateKey,
 * a Certificate chain for the corresponding public key must be provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias	String			the alias name
 * @param	key		byte[]			key in protected format, to be associated with the alias
 * @param	chain	Certificate[]	chain of certificates for the corresponding public key
 *
 * @exception	KeyStoreException	if the operation fails
 *
 */
public final void setKeyEntry(String alias,
						byte[] key,
						java.security.cert.Certificate[] chain)
							throws KeyStoreException {

    if (!loaded) throw new KeyStoreException();

	keyStoreSpi.engineSetKeyEntry(alias, key, chain);
}

/**
 * Assigns the given key to an alias. If the key is a PrivateKey,
 * a Certificate chain for the corresponding public key must be provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias		String			the alias name
 * @param	key			Key				key to be associated with the alias
 * @param	password	char[]			the passwrod to protect the key
 * @param	chain		Certificate[]	chain of certificates for the corresponding public key
 *
 * @exception	KeyStoreException		if the operation fails
 *
 */
public final void setKeyEntry(String alias,
					    Key key,
						char[] password,
						java.security.cert.Certificate[] chain)
								throws KeyStoreException
{
    if (!loaded) throw new KeyStoreException();

	keyStoreSpi.engineSetKeyEntry(alias, key, password, chain);
}

/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}
/**
 * Set the keystore type being used by the receiver to
 * the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type String
 *					the type of the keystore
 */
void setType(String type) {
	this.type = type;
}
/**
 * Retrieves the number of entries in the keystore
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the number of entries in the keystore
 *
 * @exception	KeyStoreException	If the Keystore has not been initialized (loaded)
 *
 */
public final int size() throws KeyStoreException {
    if (!loaded) throw new KeyStoreException();
	return keyStoreSpi.engineSize();
}
/**
 * Stores the keystore into the given input stream, using the password
 * provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	stream		OutputStream		stream from where to store the keystore
 *
 * @exception	KeyStoreException		 	If the keystore has not been initialized
 * @exception	IOException		An IO problem was found when writing the keystore
 * @exception	NoSuchAlgorithmException	The data integrity algorithm for the keystore cannot be found
 * @exception	CertificateException		If a certificate in the keystore could not be stored.
 *
 */
public final void store(OutputStream stream, char[] password)
                 throws KeyStoreException,
                        IOException,
                        NoSuchAlgorithmException,
                        CertificateException {
    if (!loaded) throw new KeyStoreException();
	keyStoreSpi.engineStore (stream,password);
}
/**
 * Answers a new KeyStore of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of keystore desired
 * @return		KeyStore	a concrete implementation for the KeyStore type desired.
 *
 * @exception	KeyStoreException 	If the type cannot be found
 */
private static KeyStore toKeyStoreImplementation (String type) throws KeyStoreException{

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = providers[i];
		try {
			return toKeyStoreImplementation (type, provider);
		} catch (KeyStoreException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new KeyStoreException(type);
}
/**
 * Answers a KeyStore for the type name
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type java.lang.String
 *					the type of KeyStore desired
 *
 * @param		provider		java.security.Provider
 *					the provider desired for the type.
 *
 * @return		The KeyStore for the type name
 *
 * @throws KeyStoreException 	If the requested type is not available
 */
private static KeyStore toKeyStoreImplementation (String type, Provider provider) throws KeyStoreException{

	// First try to find the class corresponding to the algorithm name
	String keyStoreClassName;
	try {
		keyStoreClassName = provider.lookupProperty(KEY_PREFIX, type);
	} catch (ClassCastException e) {
		throw new KeyStoreException(type);
	}

	// If not found, exception
	if (keyStoreClassName == null)
		throw new KeyStoreException(type);

	// Now try to instantiate the digest.
	try {
		Class keyStoreClass = Class.forName(keyStoreClassName, true, provider.getClass().getClassLoader());
		return createKeyStore (provider, keyStoreClass, type);
	} catch (ClassNotFoundException ex) {
			throw new KeyStoreException(type);
	}
}

/**
 * Answers the default keystore implementation type
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @return		The default KeyStore implementation type
 */
public static final String getDefaultType () {
	String result = (String)AccessController.doPrivileged(
	        PriviAction.getSecurityProperty("keystore.type")); //$NON-NLS-1$
	if (result == null) return "jks"; //$NON-NLS-1$
	return result;
}

}
