
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2002  All Rights Reserved
 */

/**
 * This class is a Service Provider Interface (therefore the Spi suffix) for
 * keystore mechanisms.
 *
 *
 * @author		OTI
 * @version		initial
 */

 import java.util.*;
 import java.io.*;
 import java.security.cert.*;

public abstract class KeyStoreSpi {
/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 */
public KeyStoreSpi() {
	super();
}
/**
 * Answers the alias names for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	Enumeration	All alias names for the receiver
 */
public abstract Enumeration engineAliases();
/**
 * Checks if the given alias exists in the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	true if the alias exists, false otherwise
 */
public abstract boolean engineContainsAlias(String alias);
/**
 * Deletes a given alias entry from the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be deleted.
 *
 * @exception	KeyStoreException 	If the alias cannot be removed
 */
 public abstract void engineDeleteEntry(String alias) throws KeyStoreException;
/**
 * Answers a certificate associated with the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Certificate	a certificate associated with the given alias.
 */
public abstract java.security.cert.Certificate engineGetCertificate(String alias);
/**
 * Answers the alias for a given Certificate.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		cert	Certificate		The Certificate to be searched for.
 *
 * @return	String	The alias for the given Certificate, null if the alias does not exist or if the
 *					Certificate does not exist in the receiver.
 */
public abstract String engineGetCertificateAlias(java.security.cert.Certificate cert);
/**
 * Answers the Certificate chain associated with the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Certificate[]	a chain of certificates associated with the given alias.
 */
public abstract java.security.cert.Certificate[] engineGetCertificateChain(String alias);
/**
 * Answers the creation date for the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias	String	The alias to be searched for.
 *
 * @return	Date	The date when the entry for this alias was created,
 *					null if the alias does not exist
 */
public abstract Date engineGetCreationDate(String alias);
/**
 * Answers the key associated with the given alias.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 * @param		password	char[]		Password to use when fetching the key for the alias
 *
 * @return	Key		key associated with the given alias,
 *					null if the alias does not exist
 *
 * @exception	NoSuchAlgorithmException 	If the algorithm to recover the key cannot be found
 * @exception	UnrecoverableKeyException 	If teh key can't be recovered
 */
public abstract Key engineGetKey(String alias,
								 char[] password)
						  throws NoSuchAlgorithmException,
								 UnrecoverableKeyException;
/**
 * Answers whether the alias entry represents a trusted certificate entry
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 *
 * @return	boolean		true if the alias entry represents a trusted certificate entry,
 *						false otherwise
 *
 */
public abstract boolean engineIsCertificateEntry(String alias);
/**
 * Answers whether the alias entry represents a key entry
 *
 * @author		OTI
 * @version		initial
 *
 * @param		alias		String		The alias to be searched for.
 *
 * @return	boolean		true if the alias entry represents a key entry,
 *						false otherwise
 *
 */
public abstract boolean engineIsKeyEntry(String alias);
/**
 * Loads the keystore from the given input stream, using the password
 * provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	stream		InputStream		stream from where to read the keystore
 *
 * @exception	IOException					An IO problem was found when reading the keystore
 * @exception	NoSuchAlgorithmException	The algorithm to check the integrity of the keystore cannot be found
 * @exception	CertificateException		If a certificate in the keystore could not be loaded.
 *
 */
public abstract void engineLoad(InputStream stream,
								char[] password)
						 throws IOException,
								NoSuchAlgorithmException,
								CertificateException;
/**
 * Assigns a certificate to an alias
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias	String			an alias for the Certificate
 * @param	cert	Certificate		The Certificate for which an alias is being registered
 *
 * @exception	KeyStoreException	if the alias already exists for an untrusted Certificate
 *
 */
public abstract void engineSetCertificateEntry(String alias,
											   java.security.cert.Certificate cert)
										throws KeyStoreException;
/**
 * Assigns the given key to an alias. If the key is a PrivateKey,
 * a Certificate chain for the corresponding public key must be provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias	String			the alias name
 * @param	key		byte[]			key in protected format, to be associated with the alias
 * @param	chain	Certificate[]	chain of certificates for the corresponding public key
 *
 * @exception	KeyStoreException	if the operation fails
 *
 */
public abstract void engineSetKeyEntry(String alias,
									   byte[] key,
									   java.security.cert.Certificate[] chain)
								throws KeyStoreException;
/**
 * Assigns the given key to an alias. If the key is a PrivateKey,
 * a Certificate chain for the corresponding public key must be provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	alias		String			the alias name
 * @param	key			Key				key to be associated with the alias
 * @param	password	char[]			the passwrod to protect the key
 * @param	chain		Certificate[]	chain of certificates for the corresponding public key
 *
 * @exception	KeyStoreException		if the operation fails
 *
 */
public abstract void engineSetKeyEntry(String alias,
									   Key key,
									   char[] password,
									   java.security.cert.Certificate[] chain)
								throws KeyStoreException;
/**
 * Retrieves the number of entries in the keystore
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the number of entries in the keystore
 *
 */
public abstract int engineSize();
/**
 * Stores the keystore into the given input stream, using the password
 * provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	stream		OutputStream		stream from where to store the keystore
 *
 * @exception	IOException		An IO problem was found when writing the keystore
 * @exception	NoSuchAlgorithmException	The data integrity algorithm for the keystore cannot be found
 * @exception	CertificateException		If a certificate in the keystore could not be stored.
 *
 */
public abstract void engineStore(OutputStream stream,
								 char[] password)
						  throws IOException,
								 NoSuchAlgorithmException,
								 CertificateException;

}
