
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides the functionality of a digest algorithm, like MD5 or SHA.
 * This class is is abstract and extends from MessageDigestSpi to be compatible with
 * JDK, which does it for historical reasons (check the JDK's Javadoc files).
 *
 * @author		OTI
 * @version		initial
 *
 */
public abstract class MessageDigest extends MessageDigestSpi {

 	private static final String KEY_PREFIX = "MessageDigest.";	// Key prefix for algorithm name lookup //$NON-NLS-1$

	private String algorithmName;	// Name of the digest algorithm represented by the receiver
	private Provider provider;		// Provider of the digest algorithm represented by the receiver
	// *** WARNING *** - The provider is not passed as parameter to the constructor. It means
	// it will be uninitialized until an actual binding to a concrete class by a provider
	// is performed, in getInstance.

	// This one is tricky. Providers are supposed to subclass MessageDigestSpi, but
	// method getInstance in MessageDigest returns a MessageDigest. Therefore, there is no
	// way we can return the user-provider class directly (incompatible types). We need
	// to return a MessageDigest (or subclass) instance, but somehow have a reference
	// to the provider digest instance. This is what the Wrapper class below implements.
	static private class Wrapper extends MessageDigest {
		MessageDigestSpi	providerDigest;

		Wrapper (MessageDigestSpi providerDigest, String algorithmName) {
			super(algorithmName);	// Just because there is no empty constructor in the superclass
			this.providerDigest = providerDigest;
		}

		public Object clone() throws CloneNotSupportedException {
			Wrapper clone = new Wrapper((MessageDigestSpi)providerDigest.clone(), getAlgorithm());
			clone.setProvider(getProvider());
			return clone;
		}

		protected byte[] engineDigest() {
			return providerDigest.engineDigest();
		}

		protected void engineReset() {
			providerDigest.engineReset();
		}

		protected void engineUpdate(byte bytesToHash[], int offset, int count) {
			providerDigest.engineUpdate(bytesToHash, offset, count);
		}

		protected void engineUpdate(byte byteToHash) {
			providerDigest.engineUpdate(byteToHash);
		}

		protected int engineGetDigestLength() {
			return providerDigest.engineGetDigestLength();
		}
	}

/**
 * Create a new MessageDigest with its algorithm
 * set to the argument.
 *
 *
 * @param		algorithmName java.lang.String
 *					the algorithm that the receiver will represent
 */
protected MessageDigest(String algorithmName) {
	setAlgorithm(algorithmName);
}

/**
 * Answers a new instance of the same class as the receiver,
 * whose slots have been filled in with the values in the
 * slots of the receiver.
 * <p>
 * Classes which wish to support cloning must specify that
 * they implement the Cloneable interface, since the native
 * implementation checks for this.
 *
 * @return		a shallow copy of this object
 * @exception	CloneNotSupportedException
 *						if the component does not implement the interface Cloneable
 */
public Object clone() throws CloneNotSupportedException {
	return super.clone();
		// The only reason we have this method here is to be compatible with JDK
		// in terms of having the same public methods in the same public classes.
		// JDK has clone() in this class, and so do we.
}

/**
 * Computes and answers the final hash value that
 * the receiver represents. After the digest is computed
 * the receiver is reset.
 *
 * @return		the hash the receiver computed
 *
 * @see			#reset
 */
public
byte[] digest() {
	byte[] answer = engineDigest();
	reset();
	return answer;
}

/**
 * Includes the bytes of the argument in the hash value
 * computed by the receiver, and then computes the final
 * digest value.
 *
 * @param		bytesToHash byte[]
 *					the source array
 * @return		the hash the receiver computed
 *
 * @see			#update(byte)
 * @see			#update(byte[])
 * @see			#update(byte[], int, int)
 * @see			#digest()
 * @see			#digest(byte[])
 * @see			#digest(byte[], int, int)
 */
public byte[] digest(byte bytesToHash[]) {
	update(bytesToHash);
	return digest();
}

/**
 * Computes the digest and stores it into the buffer passed as parameter.
 *
 * @param		buffer byte[]
 *					the array into which to store the digest
 * @param		offset
 *					the starting offset into the array
 * @param		length
 *					the number of bytes available to store the digest
 *
 * @exception	DigestException		If an error occurs
 *
 * @return		the number of bytes copied
 * @see			#digest()
 * @see			#digest(byte[])
 * @see			#digest(byte[], int, int)
 */
public int digest(byte[] buffer, int offset, int length) throws DigestException {
	return engineDigest(buffer, offset, length);
}

/**
 * Answers the standard Java Security name for the
 * algorithm being used by the receiver.
 *
 * @return		String
 *					the name of the algorithm
 */
public
final String getAlgorithm() {
	return algorithmName;
}
/**
 * Return the engine digest length in bytes. Default is 0.
 *
 * @return		int		the engine digest length in bytes
 *
 */
public
final int getDigestLength() {
	return engineGetDigestLength();
}

/**
 * Answers a new MessageDigest which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * MessageDigest which implements that algorithm.
 *
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @return		MessageDigest	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static MessageDigest getInstance(String algorithmName) throws NoSuchAlgorithmException {
	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = providers[i];
		try {
			return toMessageDigestImplementation(algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			/*Ignore - The next provider might have the algorithn*/
		}
	}

	// Scanned all, found nothing
	throw new NoSuchAlgorithmException(algorithmName);
}

/**
 * Answers a new MessageDigest which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * MessageDigest which implements that algorithm.
 *
 *
 * @param		algorithmName java.lang.String	Name of the algorithm desired
 * @param		providerName java.lang.String	Name of the provider which has to implement the algorithm
 * @return		MessageDigest	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 * @exception	NoSuchProviderException		If the provider cannot be found
 */
public static MessageDigest getInstance(String algorithmName, String providerName) throws NoSuchAlgorithmException, NoSuchProviderException {
	if (providerName == null || providerName == "") throw new IllegalArgumentException(); //$NON-NLS-1$

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toMessageDigestImplementation(algorithmName, provider);
}

/**
 * Answers a new MessageDigest which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * MessageDigest which implements that algorithm.
 *
 *
 * @param		algorithm java.lang.String	Name of the algorithm desired
 * @param		provider Provider Provider which has to implement the algorithm
 * @return		MessageDigest	a concrete implementation for the algorithm desired.
 *
 * @exception	NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static MessageDigest getInstance(String algorithmName, Provider provider) throws NoSuchAlgorithmException {
	return toMessageDigestImplementation(algorithmName, provider);
}

/**
 * Returns the Provider of the digest represented by the receiver.
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public
final Provider getProvider() {
	return provider;
}
/**
 * Does a simply byte-per-byte compare of the two digests.
 *
 * @param		digesta		One of the digests to compare
 * @param		digestb		The digest to compare to
 *
 * @return		<code>true</code>
 *						if the two hashes are equal
 *				<code>false</code>
 *						if the two hashes are not equal
 */
public static boolean isEqual(byte[] digesta, byte[] digestb) {
	return java.util.Arrays.equals (digesta, digestb);
}
/**
 * Puts the receiver back in an initial state, such that
 * it is ready to compute a new hash.
 *
 * @see			java.security.MessageDigest.Wrapper#engineReset()
 */
public void reset() {
	engineReset();
}
/**
 * Set the algorithm being used by the receiver to
 * the argument which should be a standard Java
 * Security algorithm name.
 *
 * @param		algorithmName String
 *					the name of the algorithm
 */
void setAlgorithm(String algorithmName) {
	this.algorithmName = algorithmName;
}
/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}
/**
 * Answers a MessageDigest for the algorithm name supplied by the given provider.
 *
 *
 * @param		algorithmName	java.lang.String
 *					the name of the algorithm to search for
 * @param		provider		java.security.Provider
 *					the provider desired for the algorithm.
 *
 * @return		The message digest for the algorithm name supplied by the given provider.
 *
 */
private static MessageDigest toMessageDigestImplementation (String algorithmName, Provider provider) throws NoSuchAlgorithmException {
	if (provider == null) throw new IllegalArgumentException();
	if (algorithmName == null) throw new IllegalArgumentException();

	// First try to find the class corresponding to the algorithm name
	String digestClassName;
	try {
		digestClassName = provider.lookupProperty(KEY_PREFIX, algorithmName);
		if (digestClassName == null) throw new NoSuchAlgorithmException(algorithmName);

		Class digestClass = Class.forName(digestClassName, true, provider.getClass().getClassLoader());
		MessageDigestSpi providedDigest = (MessageDigestSpi) digestClass.newInstance();
		MessageDigest digest;
		if (providedDigest instanceof MessageDigest)
			digest = (MessageDigest) providedDigest;
		else
			digest = new Wrapper(providedDigest, algorithmName);
		digest.setProvider(provider);
		return digest;
	} catch (ClassNotFoundException ex) { /*Ignore*/
	} catch (IllegalAccessException e) { /*Ignore*/
	} catch (InstantiationException e) { /*Ignore*/
	} catch (ClassCastException e) { /*Ignore*/
	}
	throw new NoSuchAlgorithmException(algorithmName);
}
/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return		a printable representation for the receiver.
 */
public String toString () {
	// K03a3 = {0} message digest from {1}
	return com.ibm.oti.util.Msg.getString("K03a3", new String[] {getAlgorithm(), getProvider().getName()}); //$NON-NLS-1$
}

/**
 * Includes the bytes of the argument in the hash value
 * computed by the receiver.
 *
 * @param		bytesToHash byte[]
 *					the source array
 */
public void update(byte bytesToHash[]) {
	engineUpdate(bytesToHash, 0, bytesToHash.length);
}

/**
 * Includes a range of bytes from the first argument
 * in the hash value computed by the receiver.
 *
 * @param		bytesToHash byte[]
 *					the source array
 * @param		offset
 *					the starting offset into the array
 * @param		count
 *					the number of bytes to include in the hash
 */
public void update(byte bytesToHash[], int offset, int count) {
	engineUpdate(bytesToHash, offset, count);
}

/**
 * Includes the argument in the hash value computed
 * by the receiver.
 *
 * @param		byteToHash byte
 *					the byte to feed to the hash algorithm
 *
 * @see			#reset()
 */
public void update (byte byteToHash) {
	engineUpdate(byteToHash);
}

}
