
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class is a Service Provider Interface (therefore the Spi suffix) for digest algorithms
 * to be supplied by providers. Examples of digest algorithms are MD5 and SHA.
 *
 * A digest is a secure hash function for a stream of bytes, like a fingerprint for the
 * stream of bytes.
 *
 * @author		OTI
 * @version		initial
 */
public abstract class MessageDigestSpi {

/**
 * Constructs a new instance of this class
 *
 * @author		OTI
 * @version		initial
 */
public MessageDigestSpi() {
	super();
}

/**
 * Answers a new instance of the same class as the receiver,
 * whose slots have been filled in with the values in the
 * slots of the receiver.
 * <p>
 * Classes which wish to support cloning must specify that
 * they implement the Cloneable interface, since the native
 * implementation checks for this.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a shallow copy of this object
 * @exception	CloneNotSupportedException
 *						if the component does not implement the interface Cloneable
 */
public Object clone() throws CloneNotSupportedException {
	return super.clone();
		// The only reason we have this method here is to be compatible with JDK
		// in terms of having the same public methods in the same public classes.
		// JDK has clone() in this class, and so do we.
}

/**
 * Actually does the work of computing the final
 * hash value that the receiver represents, and answers
 * the resulting value. Note that the caller must invoke
 * <code>reset</code> before using the receiver further.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the hash the receiver computed
 *
 * @see			MessageDigest#reset
 */
protected abstract byte[] engineDigest();

/**
 * Computes the digest and copies it into the buffer provided.
 *
 * This method should be abstract, but it is not in JDK. So, we
 * aren't either, for compatibility reasons.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer	where to store the digest into
 * @param		offset	where in the buffer to store the digest
 * @param		length	how many bytes are available for storing the hash
 *
 * @throws		DigestException	if the <code>length</code> is not big enough for the hash
 */
protected int engineDigest(byte[] buffer, int offset,int length) throws DigestException {
	// K0047 = buffer is null
	if (buffer == null) throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0047")); //$NON-NLS-1$
	// K03a2 = buffer is too small
	if (buffer.length < offset + length) throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K03a2")); //$NON-NLS-1$

	// K03a2 = buffer is too small
	byte[] digest = engineDigest();
	if (digest.length > buffer.length - offset) throw new DigestException(com.ibm.oti.util.Msg.getString("K03a2")); //$NON-NLS-1$
	if (length < digest.length) throw new DigestException("Digest length exceeds passed value");

	int toCopy = length;
	if (digest.length < toCopy)
		toCopy = digest.length;
	System.arraycopy (digest, 0, buffer, offset, toCopy);
	engineReset();
	return toCopy;
}

/**
 * Return the engine digest length in bytes. Default is 0.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the engine digest length in bytes
 *
 */
protected int engineGetDigestLength() {
	return 0;
}

/**
 * Puts the receiver back in an initial state, such that
 * it is ready to compute a new hash.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			MessageDigest#reset()
 */
protected abstract void engineReset();
/**
 * Includes a range of bytes from the first argument
 * in the hash value computed by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		bytesToHash byte[]
 *					the source array
 * @param		offset
 *					the starting offset into the array
 * @param		count
 *					the number of bytes to include in the hash
 */
protected abstract void engineUpdate(byte bytesToHash[], int offset, int count) ;
/**
 * Includes the argument in the hash value computed
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		byteToHash byte
 *					the byte to feed to the hash algorithm
 *
 * @see			MessageDigest#reset()
 */
protected abstract void engineUpdate(byte byteToHash);
}
