
package java.security;

import java.io.Serializable;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * Abstract superclass of all classes which represent
 * permission to access system resources.
 *
 * @author		OTI
 * @version		initial
 */
public abstract class Permission implements Guard, Serializable {
	private static final long serialVersionUID =  -5636570222231596674L;

	/**
	 * The name of the permission.
	 */
	private String name;

/**
 * Constructs a new instance of this class with its
 * name set to the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		permissionName String
 *					the name of the permission.
 */
public Permission(String permissionName) {
	name = permissionName;
}

/**
 * Compares the argument to the receiver, and answers true
 * if they represent the <em>same</em> object using a class
 * specific comparison. The implementation in Object answers
 * true only if the argument is the exact same object as the
 * receiver (==).
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o Object
 *					the object to compare with this object.
 * @return		boolean
 *					<code>true</code>
 *						if the object is the same as this object
 *					<code>false</code>
 *						if it is different from this object.
 * @see			#hashCode
 */
public abstract boolean equals (Object o);

/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>.equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int
 *					the receiver's hash.
 *
 * @see			#equals
 */
public abstract int hashCode();

/**
 * Checks that the receiver is granted in the current
 * access control context (Guard interface). Note that
 * the argument is not currently used.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		object Object
 *					ignored.
 *
 * @exception	java.lang.SecurityException
 *					If access is not granted
 */
public void checkGuard(Object object) throws SecurityException {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(this);
}

/**
 * Answers the actions associated with the receiver.
 * Subclasses should return their actions in canonical
 * form. If no actions are associated with the receiver,
 * the empty string should be returned.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the receiver's actions.
 */
public abstract String getActions();

/**
 * Answers the name of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the receiver's name.
 */
public final String getName() {
	return name;
}

/**
 * Indicates whether the argument permission is implied
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean
 *					<code>true</code> if the argument permission
 *					is implied by the receiver, and
 *					<code>false</code> if it is not.
 * @param		permission Permission
 *					the permission to check.
 */
public abstract boolean implies(Permission permission);

/**
 * Answers a new PermissionCollection for holding permissions
 * of this class. Answer null if any permission collection can
 * be used.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		PermissionCollection or null
 *					a suitable permission collection for
 *					instances of the class of the receiver.
 */
public PermissionCollection newPermissionCollection() {
	return null;
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					a printable representation for the receiver.
 */
public String toString() {
	String actions = getActions();
	return
		"(" + this.getClass().getName() +
		" " + this.getName() +
		" " + actions +
		")";
}

}
