
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * Abstract superclass of classes which represent the
 * system security policy.
 *
 * @author		OTI
 * @version		initial
 */
public abstract class Policy {

	/**
	 * The currently in place security policy.
	 */
	static Policy policy;

/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 */
public Policy() {
	super();
}

/**
 * Answers the current system security policy. If no policy has been
 * instantiated then this is done using the security property
 * <EM>policy.provider</EM>
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Policy
 *					the current system security policy.
 */
public static Policy getPolicy() {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkPermission(SecurityPermission.permissionToGetPolicy);
	return getPolicyImpl();
}

static Policy getPolicyImpl() {
	if (policy == null) {
		String policyName = Security.getProperty("policy.provider");
		try {
			if (policyName != null)
				policy = (Policy)Class.forName(policyName).newInstance();
		} catch (Exception e) {}
		if (policy == null) {
			// if instantiating the policy.provider failed for any reason,
			// assume use of the default policy provider
			policy = new com.ibm.oti.util.DefaultPolicy();
		}
	}
	return policy;
}

/**
 * Sets the system-wide policy object if it is permitted by the
 * security manager.
 *
 * @author		OTI
 * @version		initial
 *
 * @param 		p Policy
 *					the policy object that needs to be set.
 */
public static void setPolicy (Policy p) {
	SecurityManager security = System.getSecurityManager ();
	if (security != null)
		security.checkPermission (SecurityPermission.permissionToSetPolicy);
	policy = p;
}

/**
 * Answers a PermissionCollection describing what permissions
 * are available to the given CodeSource based on the current
 * security policy.
 * <p>
 * Note that this method is <em>not</em> called for classes which
 * are in the system domain (i.e. system classes). System classes
 * are <em>always</em> given full permissions (i.e. AllPermission).
 * This can not be changed by installing a new Policy.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		cs CodeSource
 *					the code source to compute the permissions for.
 * @return		PermissionCollection
 *					the permissions the code source should have.
 */
public abstract PermissionCollection getPermissions(CodeSource cs) ;

/**
 * Reloads the policy configuration, depending on how the type of source location
 * for the policy information.
 *
 * @author		OTI
 * @version		initial
 *
 */
public abstract void refresh();

/**
 * Answers a PermissionCollection describing what permissions
 * are available to the given ProtectionDomain (more specifically,
 * its CodeSource) based on the current security policy.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		domain ProtectionDomain
 *					the protection domain to compute the permissions for.
 * @return		PermissionCollection
 *					the permissions the code source should have.
 */
public PermissionCollection getPermissions(ProtectionDomain domain){
	return getPermissions( domain.getCodeSource() );
}

/**
 * Answers whether the Permission is implied by the PermissionCollection
 * of the Protection Domain
 * @param domain ProtectionDomain for which Permission to be checked
 * @param permission Permission for which authorization is to be verified
 * @return boolean Permission implied by ProtectionDomain
 */
public boolean implies(ProtectionDomain domain, Permission permission){
	return getPermissions( domain ).implies( permission );
}

}
