
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.IOException;
import java.io.InputStream;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Enumeration;

/**
 * Common superclass for providers of Java Security API implementations.
 * Such implementations may provide a number of services like Message digest,
 * key generation, certificate support, secure random number generation, etc.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			MessageDigest
 * @see			Signature
 * @see			KeyPairGenerator
 * @see			KeyFactory
 * @see			KeyStore
 * @see			java.security.cert.CertificateFactory
 * @see			AlgorithmParameters
 * @see			AlgorithmParameterGenerator
 * @see			SecureRandom
 */
public abstract class Provider extends java.util.Properties {

	private static final long serialVersionUID = -4298000515446427739L;	// To be serialization-compatible with JDK

	// These fields have to be declared in this order to be compatible with
	// JDK's serialization, so that Providers can be cross-loaded across VMs.
	private String name;		// Name of the provider.
	private String info;		// Generic description about what is being provided.
	private double version;		// Version number for the services being provided.

/**
 * Constructs a new Provider with the given name, version and info.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @param	name		name for this provider
 * @param	version		version number for the services being provided
 * @param	info		generic description of the services being provided
 */
protected Provider(String name, double version, String info) {
	super();
	this.name = name;
	this.version = version;
	this.info = info;
}

/**
 * Clears this provider so that it no longer contains the properties used to
 * look up facilities implemented by the provider.
 * First, if there is a security manager, its checkSecurityAccess method is
 * called with the string "clearProviderProperties."+name (where name is
 * the provider name) to see if it's ok to clear this provider. If the default
 * implementation of checkSecurityAccess is used (that is, that method is
 * not overriden), then this results in a call to the security manager's
 * checkPermission method with a
 * SecurityPermission("clearProviderProperties."+name) permission.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			java.util.Hashtable#isEmpty
 * @see			java.util.Hashtable#size
 */
public synchronized void clear() {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("clearProviderProperties." + name);
	super.clear();
}

/**
 * Returns an unmodifiable Set view of the property entries contained in this Provider.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		an unmodifiable Set of the property entries
 */
public Set entrySet() {
	return Collections.unmodifiableSet(super.entrySet());
}

/**
 * Returns the generic information about the services being provided.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		String	generic description of the services being provided
 */
public String getInfo() {
	return info;
}

/**
 * Returns the name of this provider.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		String	name of the provider
 */
public String getName() {
	return name;
}
/**
 * Returns the version number for teh services being provided
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		double	version number for the services being provided
 */
public double getVersion() {
	return version;
}

/**
 * Returns an unmodifiable Set view of the property keys contained in this Provider.
 *
 * @author		OTI
 * @version		initial
 *
 *
 * @return		an unmodifiable Set of the property keys
 */
public Set keySet() {
	return Collections.unmodifiableSet(super.keySet());
}

/**
 * Loads properties from the specified InputStream. The properties are of
 * the form <code>key=value</code>, one property per line.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		in	the input stream
 */
public synchronized void load(InputStream in) throws IOException {
	super.load(in);
}

/**
 * Sets the key property to have the specified value.
 * First, if there is a security manager, its checkSecurityAccess method is
 * called with the string "putProviderProperty."+name, where name is the
 * provider name, to see if it's ok to set this provider's property values.
 * If the default implementation of checkSecurityAccess is used (that is,
 * that method is not overriden), then this results in a call to the security
 * manager's checkPermission method with a
 * SecurityPermission("putProviderProperty."+name) permission.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key	the property key to add
 * @param		value	the property value to add
 * @return		the old value associated with the specified key, null if the key did
 *				not exist
 *
 * @see			java.util.Hashtable#elements()
 * @see			java.util.Hashtable#get(Object)
 * @see			java.util.Hashtable#keys()
 * @see			java.lang.Object#equals
 */
public synchronized Object put(Object key, Object value) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("putProviderProperty." + name);
	return super.put(key, value);
}

/**
 * Copies all of the mappings from the specified Map to this provider.
 * These mappings will replace any properties that this provider had
 * for any of the keys currently in the specified Map.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		map the Map to copy mappings from
 */
public synchronized void putAll(Map map) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("putProviderProperty." + name);

	Iterator it = map.entrySet().iterator();
	while (it.hasNext()) {
		Map.Entry entry = (Map.Entry)it.next();
		super.put(entry.getKey(), entry.getValue());
	}
}

/**
 * Removes the key property (and its corresponding value).
 * First, if there is a security manager, its checkSecurityAccess method
 * is called with the string ""removeProviderProperty."+name, where
 * name is the provider name, to see if it's ok to remove this provider's
 * properties. If the default implementation of checkSecurityAccess is used
 * (that is, that method is not overriden), then this results in a call to the
 * security manager's checkPermission method with a
 * SecurityPermission("removeProviderProperty."+name) permission.

 *
 * @author		OTI
 * @version		initial
 *
 * @param		key	the key to remove
 * @return		the value associated with the specified key, null if the specified key
 *				did not exist
 *
 * @see			java.util.Hashtable#get(Object)
 * @see			java.util.Hashtable#put(Object, Object)
 */
public synchronized Object remove(Object key) {
	SecurityManager security = System.getSecurityManager();
	if (security != null)
		security.checkSecurityAccess("removeProviderProperty." + name);
	return super.remove(key);
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a printable representation for the receiver.
 */
public String toString() {
	return name + " version " + version;
}

/**
 * Returns an unmodifiable Collection of the property values contained in this Provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a unmodifiable Collection of the values
 */
public Collection values() {
	return Collections.unmodifiableCollection(super.values());
}

/* (non-Javadoc)
 * @see java.util.Properties#getProperty(java.lang.String, java.lang.String)
 */
public String getProperty(String name, String defaultValue) {
    String result = getProperty( name );

    if( result == null ){
        return defaultValue;
    }

    return defaultValue;
}

/* (non-Javadoc)
 * @see java.util.Properties#getProperty(java.lang.String)
 */
public String getProperty(String name) {
    int index = name.indexOf('.');
    if( index == -1 ){
        return null;
    }

    String prefix = name.substring(0, index + 1 );
    String algorithm = name.substring(index + 1, name.length());

    return lookupProperty( prefix, algorithm );
}

String lookupProperty(String property) {

    //Use the Properties.getProperty to determine if property is present
    //in its original form
	String result = super.getProperty(property);
	if (result != null) return result;

    //Look for the property in a case-insensitive fashion
    String upper = property.toUpperCase();
	Enumeration keyEnum = keys();
	while (keyEnum.hasMoreElements()) {
		String key = (String)keyEnum.nextElement();
		if (key.toUpperCase().equals(upper))
			return getProperty(key);
	}

    //Property was not found
	return null;
}

String lookupProperty(String prefix, String name) {
	String property = prefix + name;
	String result = lookupProperty(property);
	if (result != null) return result;
	result = lookupProperty("Alg.Alias." + property);
	if (result != null)
		return lookupProperty(prefix + result);
	return null;
}

}
