
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.util.*;
import java.security.spec.*;
import java.security.cert.*;

/**
 * This class provides the functionality of a signature algorithm, like DSA or RSA.
 * This class is abstract and extends from SignatureSpi to be compatible with
 * JDK, which does it for historical reasons (check the JDK's Javadoc files).
 *
 * @author		OTI
 * @version		initial
 */
public abstract class Signature extends SignatureSpi {

	protected static final int UNINITIALIZED = 0;	// State = UNINITIALIZED
	protected static final int SIGN = 2;			// State = SIGN
	protected static final int VERIFY = 3;			// State = VERIFY
	protected int state;			// Current state of the receiver - check possible states above

	// Used in toString()
	private static final String[] STATE_DESCRIPTION = {"UNINITIALIZED", "SIGN" , "VERIFY"};

 	private static final String KEY_PREFIX = "Signature.";	// Key prefix for algorithm name lookup

 	private String algorithmName;	// Name of the signature algorithm represented by the receiver
	private Provider provider;		// Provider of the signature algorithm represented by the receiver
	// *** WARNING *** - The provider is not passed as parameter to the constructor. It means
	// it will be uninitialized until an actual binding to a concrete class by a provider
	// is performed, in getInstance.

	// This one is tricky. Providers are supposed to subclass SignatureSpi, but
	// method getInstance in Signature returns a Signature. Therefore, there is no
	// way we can return the user-provider class directly (incompatible types). We need
	// to return a Signature (or subclass) instance, but somehow have a reference
	// to the provider signature instance. This is what the Wrapper class below implements.
	static private class Wrapper extends Signature {
		SignatureSpi	signatureProvider;

		Wrapper (SignatureSpi signatureProvider, String algorithmName) {
			super(algorithmName);	// Just because there is no empty constructor in the superclass
			this.signatureProvider = signatureProvider;
		};

		/**
		 * @see 		java.lang.Object#clone()
		 */
		public Object clone() throws CloneNotSupportedException {
			Wrapper clone = new Wrapper((SignatureSpi)signatureProvider.clone(), getAlgorithm());
			clone.setProvider(getProvider());
			return clone;
		};

		protected void engineInitSign(PrivateKey privateKey) throws InvalidKeyException {
			signatureProvider.engineInitSign (privateKey);
		};

		protected void engineInitVerify(PublicKey publicKey) throws InvalidKeyException {
			signatureProvider.engineInitVerify (publicKey);
		};

		protected byte[] engineSign() throws SignatureException {
			return signatureProvider.engineSign ();
		};

		protected void engineUpdate(byte[] buffer, int offset, int length) throws SignatureException{
			signatureProvider.engineUpdate (buffer,offset,length);
		};

		protected void engineUpdate(byte b) throws SignatureException {
			signatureProvider.engineUpdate (b);
		};

		protected boolean engineVerify(byte[] sigBytes) throws SignatureException {
			return signatureProvider.engineVerify (sigBytes);
		};

		/**
		 * @see 		java.lang.Object#toString()
		 */
		public String toString () {
			if (signatureProvider != null)
				return signatureProvider.toString ();
			else
				return super.toString ();
		}

	}

/**
 * Constructs a new instance of this class with the algorithm provided.
 *
 * @param		algName		Algorithm name
 */
protected Signature(String algName) {
	super();
	setAlgorithm (algName);
	state = UNINITIALIZED;
}
/**
 * Answers a new instance of the same class as the receiver,
 * whose slots have been filled in with the values in the
 * slots of the receiver.
 * <p>
 * Classes which wish to support cloning must specify that
 * they implement the Cloneable interface, since the native
 * implementation checks for this.
 *
 * @return		a shallow copy of this object
 *
 * @throws		CloneNotSupportedException 	if the component does not implement the interface Cloneable
 */
public Object clone() throws CloneNotSupportedException {
	return super.clone();
		// The only reason we have this method here is to be compatible with JDK
		// in terms of having the same public methods in teh same public classes.
		// JDK has clone() in this class, and so do we.
}
/**
 * Answers the standard Java Security name for the
 * algorithm being used by the receiver.
 *
 * @return		the name of the algorithm
 */
public final String getAlgorithm() {
	return algorithmName;
}
/**
 * Answers a new Signature which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * Signature which implements that algorithm.
 *
 * @param		algorithmName	Name of the algorithm desired
 * @return		a concrete implementation for the algorithm desired.
 *
 * @throws		NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static Signature getInstance(String algorithmName) throws NoSuchAlgorithmException {
	// *** NOTE *** - This method is declared to return Signature, while
	// providers are supposed to subclass SignatureSpi. So, we
	// cannot simply return an instance of a subclass of SignatureSpi.
	// This is why we have the inner type Wrapper

	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	return toSignatureImplementation(algorithmName);
}
/**
 * Answers a new Signature which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * Signature which implements that algorithm.
 *
 * @param		algorithmName	Name of the algorithm desired
 * @param		providerName	Name of the provider which has to implement the algorithm
 * @return		a concrete implementation for the algorithm desired.
 *
 * @throws		NoSuchAlgorithmException	If the algorithm cannot be found
 * @throws		NoSuchProviderException		If the provider cannot be found
 */
public static Signature getInstance(String algorithmName,
									String providerName)
							 throws NoSuchAlgorithmException,
									NoSuchProviderException {

	// *** NOTE *** - This method is declared to return Signature, while
	// providers are supposed to subclass SignatureSpi. So, we
	// cannot simply return an instance of a subclass of SignatureSpi.
	// This is why we have the inner type Wrapper

	if (providerName==null) throw new java.lang.IllegalArgumentException();
	if (algorithmName==null) throw new java.lang.IllegalArgumentException();

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toSignatureImplementation(algorithmName, provider);
}

/**
 * Answers a new Signature which is capable of
 * running the algorithm described by the argument.
 * The result will be an instance of a subclass of
 * Signature which implements that algorithm.
 *
 * @param		algorithm 	Name of the algorithm desired
 * @param		provider 	Provider which has to implement the algorithm
 * @return		a concrete implementation for the algorithm desired.
 *
 * @throws		NoSuchAlgorithmException	If the algorithm cannot be found
 */
public static Signature getInstance(String algorithm,Provider provider)
 throws NoSuchAlgorithmException
{
	if( (algorithm == null) || (provider == null) ){
		throw new IllegalArgumentException();
	}

	return toSignatureImplementation(algorithm, provider);

}

/**
 * Returns the Provider of the signature represented by the receiver.
 *
 * @return		an instance of a subclass of java.security.Provider
 */
public final Provider getProvider() {
	return provider;
}
/**
 * Initialize the receiver with the specified private key, to be used for signing purposes.
 *
 * @param		privateKey 		the key to be used when signing
 *
 * @throws		InvalidKeyException 	if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
public final void initSign(PrivateKey privateKey) throws InvalidKeyException {
	state = SIGN;
	engineInitSign (privateKey);
}
/**
 * Initialize the receiver with the specified private key, to be used for signing purposes.
 *
 * @param		privateKey		the key to be used when signing
 * @param		random 			the random number generator to use
 *
 * @throws		InvalidKeyException 	if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
public final void initSign(PrivateKey privateKey, SecureRandom random) throws InvalidKeyException {
	state = SIGN;
	engineInitSign (privateKey, random);
}

/**
 * Initialize the receiver with the specified public key, to be used for verification purposes.
 *
 * @param		publicKey		the key to be used when verifying
 *
 * @throws		InvalidKeyException 	if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
public final void initVerify(PublicKey publicKey) throws InvalidKeyException {
	state = VERIFY;
	engineInitVerify(publicKey);
}

/**
 * Initialize the receiver with the specified certificate, using public key from certificate,
 * to be used for verification purposes.
 *
 * @param		certificate		the certificate to be used when verifying
 *
 * @throws		InvalidKeyException 	if the specified key is not valid, cannot be used for digital signature purposes
 * 										(improperly encoded, parameters missing, etc)
 */
public final void initVerify(java.security.cert.Certificate certificate) throws InvalidKeyException {
	state = VERIFY;
	if (certificate instanceof X509Certificate) {
		X509Certificate x509 =  (X509Certificate) certificate;
		Set critical = x509.getCriticalExtensionOIDs();
		if (critical != null && !critical.isEmpty()) {
			// keyUsageOID "2.5.29.15"
			if (critical.contains("2.5.29.15")) {
         		boolean [] keyUsage = x509.getKeyUsage();
				if (!keyUsage[0]) {
         			// K00e4 = Key usage is critical and cannot be used for digital signature purposes.
         			throw new InvalidKeyException(com.ibm.oti.util.Msg.getString("K00e4"));
				}
			}
 		}
	}
	engineInitVerify(certificate.getPublicKey());
}

/**
 * Set the algorithm being used by the receiver to
 * the argument which should be a standard Java
 * Security algorithm name.
 *
 * @param		algorithmName	the name of the algorithm
 */
void setAlgorithm(String algorithmName) {
	this.algorithmName = algorithmName;
}

/**
 * Initializes the receiver with the given parameters.
 *
 * @param		params		the parameters for this algorithm
 *
 * @throws		InvalidAlgorithmParameterException 		if the specified parameters are not valid for the
 * 														receiver or if they have been set already
 *
 */
 public final void setParameter(AlgorithmParameterSpec params) throws InvalidAlgorithmParameterException {
	 engineSetParameter (params);
 }

 /**
  * Returns the algorithm parameters in use with this Signature
  *
  * @return 	algorithm parameters in use with this Signature
 */
public final AlgorithmParameters getParameters(){
 	return engineGetParameters();
 }

/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}
/**
 * Answers the signature bytes of the data updated so far.
 *
 * @return		the signature bytes of the data updated so far.
 *
 * @throws		SignatureException		if the engine is not initialized properly
 *
 * @see			SignatureSpi#engineUpdate(byte)
 * @see			SignatureSpi#engineUpdate(byte[], int, int)
 *
 */
 public final byte[] sign() throws SignatureException {
	if (state != SIGN) throw new SignatureException();
	return engineSign();
 }
/**
 * Finishes the signing and stores the result into the buffer provided.
 *
 * @param		buffer		the buffer into which the result should be stored.
 * @param		offset		offset in the buffer where to store the result
 * @param		length		how many bytes allowed for storing
 * @return		how many bytes were copied into the buffer
 *
 * @throws		SignatureException		if the receiver is not initialized properly
 */
public final int sign(byte[] buffer, int offset, int length) throws SignatureException {
	if (state != SIGN) throw new SignatureException();
	return engineSign (buffer, offset, length);
}
/**
 * Answers a Signature for the algorithm name supplied by any provider.
 *
 * @param		algorithmName	the name of the algorithm to search for
 * @return		the Signature for the algorithm name supplied by any provider
 *.
 * @throws 		NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static Signature toSignatureImplementation (String algorithmName) throws NoSuchAlgorithmException{

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toSignatureImplementation (algorithmName, provider);
		} catch (NoSuchAlgorithmException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new NoSuchAlgorithmException(algorithmName);
}
/**
 * Answers a Signature for the algorithm name supplied by the given provider.
 *
 * @param		algorithmName	the name of the algorithm to search for
 * @param		provider		the provider desired for the algorithm.
 * @return		the Signature for the algorithm name supplied by the given provider.
 *
 * @throws 		NoSuchAlgorithmException	If the provider does not implement such algorithm
 */
private static Signature toSignatureImplementation (
		String algorithmName,
		Provider provider) throws NoSuchAlgorithmException {

	// First try to find the class corresponding to the algorithm name
	String signatureClassName;
	try {
		signatureClassName = (String) provider.lookupProperty(KEY_PREFIX, algorithmName);
	} catch (ClassCastException e) {
		throw new NoSuchAlgorithmException(algorithmName);
	}

	// If not found, exception
	if (signatureClassName == null)
		throw new NoSuchAlgorithmException(algorithmName);

	// Now try to instantiate the digest.
	try {
		Class signatureClass = Class.forName(signatureClassName, true, provider.getClass().getClassLoader());
		SignatureSpi providedSignature = (SignatureSpi) signatureClass.newInstance();
		Signature signature;
		if (providedSignature instanceof Signature)
			signature = (Signature) providedSignature;
		else

			signature = new Wrapper(providedSignature, algorithmName);
		signature.setProvider(provider);
		return signature;
	} catch (ClassNotFoundException ex) {
	} catch (IllegalAccessException e) {
	} catch (InstantiationException e) {
	} catch (ClassCastException e) {
	}
	throw new NoSuchAlgorithmException(algorithmName);
}
/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return		a printable representation for the receiver.
 */
public String toString () {
	StringBuffer result = new StringBuffer (128);
	result.append (getAlgorithm());
	result.append (" Signature (");
	result.append (STATE_DESCRIPTION[state]);
	result.append (')');
	return result.toString();
}
/**
 * Update the bytes signed so far with the extra bytes provided.
 *
 * @param		data		the extra bytes to be signed.
 *
 * @throws		SignatureException  	if the receiver is not initialized properly
 */
public final void update(byte[] data) throws SignatureException {
	if (state == UNINITIALIZED) throw new SignatureException();
	engineUpdate (data, 0, data.length);
}
/**
 * Update the bytes signed so far with the extra bytes provided.
 *
 * @param		buffer		the extra bytes to be signed.
 * @param		offset		offset in the buffer where to start getting bytes for signing
 * @param		length		how many bytes to get for signing
 *
 * @throws		SignatureException  	if the receiver is not initialized properly
 */
public final void update(byte[] buffer, int offset, int length) throws SignatureException {
	if (state == UNINITIALIZED) throw new SignatureException();
	engineUpdate (buffer, offset, length);
}
/**
 * Update the bytes signed so far with the extra byte provided.
 *
 * @param		b			the extra byte to be signed.
 *
 * @throws		SignatureException  	if the receiver is not initialized properly
 */
public final void update(byte b) throws SignatureException {
	if (state == UNINITIALIZED) throw new SignatureException();
	engineUpdate (b);
}
/**
 * Verifies the passed signature.
 *
 * @param		signature	the signature bytes to verify
 * @return		<code>true</code> if the signature was verified successfully, <code>false</code> otherwise.
 *
 * @throws		SignatureException  	if the receiver is not initialized properly or if the key is not encoded properly.
 */
public final boolean verify(byte[] signature) throws SignatureException {
	if (state != VERIFY) throw new SignatureException();
	return engineVerify(signature);
}

/**
 * Verifies the passed signature.
 *
 * @param		signature	array containing the signature bytes to verify
 * @param		offset		offset at which the signature begins
 * @param		length		length of the signature
 * @return		<code>true</code> if the signature was verified successfully, <code>false</code> otherwise.
 *
 * @throws		SignatureException  	if the receiver is not initialized properly or if the key is not encoded properly.
 */
public final boolean verify(byte[] signature, int offset, int length) throws SignatureException {
	if (state != VERIFY) throw new SignatureException();
	return engineVerify(signature, offset, length);
}

}
