
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class is a Service Provider Interface (therefore the Spi suffix) for signature algorithms
 * to be supplied by providers. Examples of signature algorithms are SHA1 with DSA, SHA1 with RSA,
 * MD5 with RSA, etc.
 *
 *
 * @author		OTI
 * @version		initial
 */

 import java.security.spec.*;

public abstract class SignatureSpi {

	protected SecureRandom appRandom;	// Application-defined secure random number generation
/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 */
public SignatureSpi() {
	super();
}

/**
 * Answers a new instance of the same class as the receiver,
 * whose slots have been filled in with the values in the
 * slots of the receiver.
 * <p>
 * Classes which wish to support cloning must specify that
 * they implement the Cloneable interface, since the native
 * implementation checks for this.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a complete copy of this object
 * @exception	CloneNotSupportedException
 *						if the component does not implement the interface Cloneable
 */
public Object clone() throws CloneNotSupportedException {
	SignatureSpi result = (SignatureSpi) super.clone();
	try {
		// SecureRandom does not have clone(),
		result.appRandom = (SecureRandom) this.appRandom.getClass().newInstance();
	}
	catch (IllegalAccessException iae) {
		throw new CloneNotSupportedException();
	}
	catch (InstantiationException ie) {
		throw new CloneNotSupportedException();
	}
	return result;
}

/**
 * Initialize the receiver with the specified private key, to be used for signing purposes.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		privateKey PrivateKey	the key to be used when signing
 *
 * @exception	InvalidKeyException if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
protected abstract void engineInitSign(PrivateKey privateKey) throws InvalidKeyException;
/**
 * Initialize the receiver with the specified private key, to be used for signing purposes.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		privateKey	PrivateKey		the key to be used when signing
 * @param		random 		SecureRandom	the random number generator to use
 *
 * @exception	InvalidKeyException if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
protected void engineInitSign(PrivateKey privateKey, SecureRandom random) throws InvalidKeyException {
	appRandom = random;
	engineInitSign (privateKey);
}

/**
 * Initialize the receiver with the specified public key, to be used for verification purposes.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		publicKey	PublicKey		the key to be used when verifying
 *
 * @exception	InvalidKeyException if the specified key is not valid (improperly encoded, parameters missing, etc)
 */
protected abstract void engineInitVerify(PublicKey publicKey) throws InvalidKeyException;

/**
 * Initializes the receiver with the given parameters.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		params	AlgorithmParameterSpec		the parameters for this algorithm
 *
 * @exception	InvalidAlgorithmParameterException
 *					if the specified parameters are not valid for the receiver or if they have
 *					been set already
 *
 */
protected void engineSetParameter(AlgorithmParameterSpec params) throws InvalidAlgorithmParameterException {
	throw new UnsupportedOperationException(); // As defined in the Javadoc
}

/**
 * Returns the parameters in use for the Signature operation
 *
 * @return AlgorithmParameters parameters in use for Signature operation
 *
 * @throws UnsupportedOperationException if the provider does not implement
 */
protected AlgorithmParameters engineGetParameters(){
	throw new UnsupportedOperationException();
}

/**
 * Answers the signature bytes of the data updated so far.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	the signature bytes of the data updated so far.
 *
 * @exception	SignatureException	if the engine is not initialized properly
 *
 * @see		#engineUpdate(byte)
 * @see		#engineUpdate(byte[], int, int)
 */
protected abstract byte[] engineSign() throws SignatureException;

/**
 * Finishes the signing and stores the result into the buffer provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer	byte[]		the buffer into which the result should be stored.
 * @param		offset	int			offset in the buffer where to store the result
 * @param		length	int			how many bytes allowed for storing
 *
 * @return		how many bytes were copied into the buffer
 *
 * @exception	SignatureException  if the receiver is not initialized properly
 */
protected int engineSign(byte[] buffer,int offset, int length) throws SignatureException {
	byte[] result = engineSign();
	if (length < result.length)
		throw new SignatureException();	// No partial digests
	System.arraycopy (result, 0 , buffer, offset, result.length);
	return result.length;
}

/**
 * Update the bytes signed so far with the extra bytes provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer	byte[]		the extra bytes to be signed.
 * @param		offset	int			offset in the buffer where to start getting bytes for signing
 * @param		length	int			how many bytes to get for signing
 *
 * @exception	SignatureException  if the receiver is not initialized properly
 */
protected abstract void engineUpdate(byte[] buffer, int offset, int length) throws SignatureException;

/**
 * Update the bytes signed so far with the extra byte provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		b	byte		the extra byte to be signed.
 *
 * @exception	SignatureException  if the receiver is not initialized properly
 */
protected abstract void engineUpdate(byte b) throws SignatureException;
/**
 * Verifies the passed signature.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		sigBytes	byte[]		the signature bytes to verify
 *
 * @return		true if the signature was verified successfully, false otherwise.
 *
 * @exception	SignatureException  if the receiver is not initialized
 *					properly or if the key is not encoded properly.
 */
protected abstract boolean engineVerify(byte[] sigBytes) throws SignatureException;

/**
 * Verifies the passed signature.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		sigBytes	byte[]		array containing the signature bytes to verify
 * @param		offset		int			offset at which signature begins
 * @param		length		int			length of the signature
 *
 * @return		true if the signature was verified successfully, false otherwise.
 *
 * @exception	SignatureException  if the receiver is not initialized
 *					properly or if the key is not encoded properly.
 */
protected boolean engineVerify(byte[] sigBytes, int offset, int length)
throws SignatureException
{
	byte[] data = new byte[ length ];
	System.arraycopy(sigBytes, offset, data, 0, length );
	return engineVerify( data );
}

}
