
package java.security;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * This class provides the functionality of a Serializable object whose bytes (in serialized form)
 * can ve signed and verified using a signature algorithm from any JCA Provider installed.
 *
 * @author		OTI
 * @version		initial
 */

import java.io.*;

public final class SignedObject implements Serializable {
	private static final long serialVersionUID = 720502720485447167L;

	private byte[] content;		// Serialized form of the target object
	private byte[] signature;	// Signature bytes for the target object
	private String thealgorithm;	// Signature algorithm for the target object

/**
 * Constructs a new instance of this class with the object to be signed provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	object	Serializable	Object whose bytes are to be signed
 * @param	key		PrivateKey		Key to be used to sign the object
 * @param	signatureAlgorithm	Signature	Signature code to use to sign the bytes
 *
 * @exception IOException if errors are detected when serializing the object
 * @exception InvalidKeyException if the key provided is not valid
 * @exception SignatureException if errors are detected when signing the serilized bytes
 */
public SignedObject(Serializable object, PrivateKey key, Signature signatureAlgorithm) throws IOException, InvalidKeyException, SignatureException {
	super();
	thealgorithm = signatureAlgorithm.getAlgorithm();
	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	ObjectOutputStream oos = new ObjectOutputStream(baos);
	oos.writeObject (object);
	oos.close();
	content = baos.toByteArray();
	signatureAlgorithm.initSign (key);
	signatureAlgorithm.update (content);
	signature = signatureAlgorithm.sign();
}

/**
 * Answers the standard Java Security name for the
 * signature algorithm being used by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the name of the algorithm
 */
public String getAlgorithm() {
	return thealgorithm;
}

/**
 * Answers the actual signature bytes for the object held by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]
 *					The actual bytes representing the signature for the object held by the receiver
 */
public byte[] getSignature() {
	return (byte[])signature.clone();
}

/**
 * Verifies the integrity of the object held by the rceiver by verifying its signature
 * according to the key provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean	true if the signature verification succeeds, false otherwise.
 */
public boolean verify(PublicKey key, Signature signatureAlgorithm) throws InvalidKeyException, SignatureException {
	signatureAlgorithm.initVerify(key);
	signatureAlgorithm.update(content);
	return signatureAlgorithm.verify (getSignature());
}

/**
 * Answers the object whose bytes were signed
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Object
 *					the actual target object held by the receiver.
 */
public Object getObject() throws IOException, ClassNotFoundException {
	ByteArrayInputStream bais = new ByteArrayInputStream(content);
	ObjectInputStream ois = new ObjectInputStream(bais);
	return ois.readObject ();
}

}
