
package java.security.cert;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.util.Arrays;
import java.security.*;
import java.io.Serializable;
import java.io.ByteArrayInputStream;
import java.io.InvalidObjectException;
import java.io.ObjectStreamException;
import java.io.NotSerializableException;

/**
 * Abstract class to represent identity certificates. It represents
 * a way to verify the binding of a Principal and its public key.
 * Examples are X.509, PGP, and SDSI.
 *
 *
 * @author		OTI
 * @version		initial
 */
public abstract class Certificate implements Serializable {
	private static final long serialVersionUID = -6751606818319535583L;

	private String type;	// Type of certificate

/**
 * Constructs a new instance of this class with its type.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	type	String	Type of certificate
 */
protected Certificate(String type) {
	this.type = type;
}

/**
 * Compares the argument to the receiver, and answers true
 * if they represent the <em>same</em> object using a class
 * specific comparison. The implementation in Object answers
 * true only if the argument is the exact same object as the
 * receiver (==).
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o		the object to compare with this object
 * @return		<code>true</code>
 *					if the object is the same as this object
 *				<code>false</code>
 *					if it is different from this object
 * @see			#hashCode
 */
public boolean equals (Object o) {
	if (o == null) return false;

	Certificate c = null;
	try {
		c = (Certificate) o;
	} catch (ClassCastException e) {return false;};

	try {
		return Arrays.equals(this.getEncoded(), c.getEncoded());
	} catch (CertificateEncodingException cee) {
		return false;
	}
}
/**
 * Answers the encoded representation for this certificate.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the encoded representation for this certificate.
 */
public abstract byte[] getEncoded() throws CertificateEncodingException;
/**
 * Answers the public key corresponding to this certificate.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the public key corresponding to this certificate.
 */
public abstract PublicKey getPublicKey();
/**
 * Answers the certificate type represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the certificate type represented by the receiver.
 */
public final String getType() {
	return type;
}
/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	try {
		java.util.zip.CRC32 hash = new java.util.zip.CRC32();
		hash.update(getEncoded());
		return (int)hash.getValue();
	} catch (CertificateEncodingException cee) {
		return super.hashCode();
	}
}
/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a printable representation for the receiver.
 */
public abstract String toString ();

/**
 * Verifies that this certificate was signed with the given public key.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	key	PublicKey	public key for which verification should be performed.
 *
 * @exception CertificateException		if encoding errors are detected
 * @exception NoSuchAlgorithmException	if an unsupported algorithm is detected
 * @exception InvalidKeyException		if an invalid key is detected
 * @exception NoSuchProviderException	if there is no default provider
 * @exception SignatureException		if signature errors are detected
 */
public abstract void verify(PublicKey key)
					 throws CertificateException,
							NoSuchAlgorithmException,
							InvalidKeyException,
							NoSuchProviderException,
							SignatureException;
/**
 * Verifies that this certificate was signed with the given public key.
 * Uses the signature algorithm given by the provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	key	PublicKey		public key for which verification should be performed.
 * @param	sigProvider	String	the name of teh signature provider.
 *
 * @exception CertificateException		if encoding errors are detected
 * @exception NoSuchAlgorithmException	if an unsupported algorithm is detected
 * @exception InvalidKeyException		if an invalid key is detected
 * @exception NoSuchProviderException	if there is no default provider
 * @exception SignatureException		if signature errors are detected
 */
public abstract void verify(PublicKey key,
							String sigProvider)
					 throws CertificateException,
							NoSuchAlgorithmException,
							InvalidKeyException,
							NoSuchProviderException,
							SignatureException;

/**
 * Representation of a Certificate used for serialization.
 *
 * @author		OTI
 * @version		initial
 */
protected static class CertificateRep implements Serializable {
	static final long serialVersionUID = -8563758940495660020L;
	String type;
	byte[] data;

/**
 * Create this representation of a certificate.
 *
 * @param	type	the type of the certificate
 * @param	encoded	the encoded representation of the certificate
 *
 * @author		OTI
 * @version		initial
 */
protected CertificateRep(String type, byte[] encoded) {
	this.type = type;
	data = encoded;
}

/**
 * Return the certificate represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 */
protected Object readResolve() throws ObjectStreamException {
	try {
		CertificateFactory factory = CertificateFactory.getInstance(type);
		Certificate certificate = factory.generateCertificate(new ByteArrayInputStream(data));
		return certificate;
	} catch (CertificateException e) {
		throw new InvalidObjectException(e.toString());
	}
}
}

/**
 * Return a representation of the receiver for serialization.
 *
 * @author		OTI
 * @version		initial
 */
protected Object writeReplace() throws ObjectStreamException {
	byte[] encoded = null;
	try {
		encoded = getEncoded();
	} catch (CertificateEncodingException e) {
		throw new NotSerializableException(e.toString());
	}
	return new CertificateRep(type, encoded);
}
}
