
package java.security.cert;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class provides the functionality of a certificate factory algorithm.
 *
 * @author		OTI
 * @version		initial
 */
 import java.security.*;
 import java.io.*;
import java.util.*;

public class CertificateFactory {
	private static final String KEY_PREFIX = "CertificateFactory.";	// Key prefix for algorithm name lookup

 	private String type		;	// Certificate type being used
	private Provider provider;		// Provider of the key algorithm represented by the receiver
	private CertificateFactorySpi certificateFactorySpi;	// The actual provider-supplied implementation

/**
 * Constructs a new instance of this class with the algorithm provided.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	certFacSpi		CertificateFactorySpi	The actual certificate factory implementation
 * @param	provider		Provider		The provider of the certificate factory
 * @param	type			String			Certificate type
 */
protected CertificateFactory(CertificateFactorySpi certFacSpi, Provider provider, String type) {
	super();
	setProvider (provider);
	setType (type);
	this.certificateFactorySpi = certFacSpi;
}
/**
 * Creates a CertificateFactory for a given CertificateFactorySpi from a provider
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider	Provider	The provider that is supplying the implementation for the service
 * @param		certClass	Class		The class that implements the type wanted
 * @param		type		String		The type desired
 *
 * @return		a CertificateFactory for a given CertificateFactorySpi from a provider
 *
 * @throws 		CertificateException	if parsing problems are detected
 *
 */
private static CertificateFactory createCertificateFactory(Provider provider, Class certClass, String type) throws CertificateException {
	try {
		CertificateFactorySpi providedCertificateFactory = (CertificateFactorySpi) certClass.newInstance();
		return new CertificateFactory (providedCertificateFactory, provider, type);
	} catch (IllegalAccessException e) {}
	  catch (InstantiationException e) {}

	throw new CertificateException(type);
}

/**
 * Generates and initializes a Certificate from data from the provided
 * input stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		inStream InputStream	Stream from where data is read to create the Certificate
 *
 * @return		Certificate 	an initialized Certificate
 * @exception	CertificateException 	if parsing problems are detected
 */
 public final Certificate generateCertificate(InputStream inStream) throws CertificateException{
	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	return certificateFactorySpi.engineGenerateCertificate(inStream);
}
/**
 * Generates and initializes a collection of Certificates from data from the provided
 * input stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		inStream InputStream	Stream from where data is read to create the Certificates
 *
 * @return		Collection	 	an initialized collection of Certificates
 * @exception	CertificateException 	if parsing problems are detected
 */
 public final Collection generateCertificates(InputStream inStream) throws CertificateException {
	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	return certificateFactorySpi.engineGenerateCertificates(inStream);
}
/**
 * Generates and initializes a Certificate Revocation List from data from the provided
 * input stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		inStream InputStream	Stream from where data is read to create the CRL
 *
 * @return		CRL 	an initialized Certificate Revocation List
 * @exception	CRLException 	if parsing problems are detected
 */
 public final CRL generateCRL(InputStream inStream) throws CRLException {

	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CRLException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	return certificateFactorySpi.engineGenerateCRL(inStream);
}

 /**
  * Generates a <code>CertPath</code> from data from the provided <code>InputStream</code>. The
  * default encoding is assumed.
  *
  * @param inStream  InputStream with PKCS7 or PkiPath encoded data
  *
  * @return CertPath	 a CertPath initialized from the provided data
  *
  * @throws CertificateException  if parsing problems are detected
  */
public final CertPath generateCertPath(InputStream inStream) throws CertificateException {
	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}

	return certificateFactorySpi.engineGenerateCertPath( inStream );
}

/**
 * Generates a <code>CertPath</code> from data from the provided <code>InputStream</code>. The
 * encoding is that specified by the encoding parameter.
 *
 * @param inStream  InputStream containing certificate path data in specified encoding
 * @param encoding  encoding of the data in the input stream
 *
 * @return CertPath  a CertPath initialized from the provided data
 *
 * @throws CertificateException  if parsing problems are detected
 * @throws UnsupportedOperationException  if the provider does not implement this method
 */
public final CertPath generateCertPath(InputStream inStream, String encoding) throws CertificateException{
	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	if( encoding == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", encoding ));
	}

	inStream = new BufferedInputStream(inStream);
	return certificateFactorySpi.engineGenerateCertPath( inStream, encoding );
}

/**
 * Generates a <code>CertPath</code> from the provided List of Certificates.  The encoding is
 * the default encoding.
 *
 * @param 	certificates  List	containing certificates in a format supported by the CertificateFactory
 *
 * @return 	CertPath  			a CertPath initialized from the provided data
 *
 * @throws CertificateException  if parsing problems are detected
 * @throws UnsupportedOperationException  if the provider does not implement this method
 */
public final CertPath generateCertPath(List certificates) throws CertificateException {
	if( certificates == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CertificateException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	return certificateFactorySpi.engineGenerateCertPath( certificates );
}

/**
 * Returns an Iterator over the supported CertPath encodings (as Strings). The first
 * element is the default encoding.
 * @return Iterator  Iterator over supported CertPath encodings (as Strings)
 */
public final Iterator getCertPathEncodings(){
	return certificateFactorySpi.engineGetCertPathEncodings();
}

/**
 * Generates and initializes a collection of Certificate Revocation List from data from the provided
 * input stream.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		inStream InputStream	Stream from where data is read to create the CRLs
 *
 * @return	Collection		an initialized collection of Certificate Revocation List
 * @exception	CRLException 	if parsing problems are detected
 */
 public final Collection generateCRLs(InputStream inStream) throws CRLException{
	if( inStream == null ){
		// K00a5 = Invalid parameter - {0}
		throw new CRLException(com.ibm.oti.util.Msg.getString( "K00a5", "null" ));
	}
	return certificateFactorySpi.engineGenerateCRLs(inStream);
}
/**
 * Answers a new CertificateFactory of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of certificate desired
 * @return		CertificateFactory	a concrete implementation for the certificate type desired.
 *
 * @exception	CertificateException	If the type cannot be found
 */
public static final CertificateFactory getInstance(String type) throws CertificateException {
	// *** NOTE *** - This method is declared to return CertificateFactory, while
	// providers are supposed to subclass CertificateFactorySpi. So, we
	// cannot simply return an instance of a subclass of CertificateFactorySpi.
	// This is why we have the instance variable pointing to the actual implementation

	// K0337 = null type not allowed
	if (type==null) throw new CertificateException(com.ibm.oti.util.Msg.getString("K0337"));

	return toCertificateFactoryImplementation(type);
 }
/**
 * Answers a new CertificateFactory of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of certificatem desired
 * @param		providerName java.lang.String	Name of the provider which has to implement the algorithm
 * @return		CertificateFactory	a concrete implementation for the certificate type desired.
 *
 * @exception	CertificateException	If the type cannot be found
 * @exception	NoSuchProviderException		If the provider cannot be found
 */
public static final CertificateFactory getInstance(String type,
												   String providerName)
											throws CertificateException,
												   NoSuchProviderException {
	// *** NOTE *** - This method is declared to return CertificateFactory, while
	// providers are supposed to subclass CertificateFactorySpi. So, we
	// cannot simply return an instance of a subclass of CertificateFactorySpi.
	// This is why we have the instance variable pointing to the actual implementation

	if (providerName==null) throw new java.lang.IllegalArgumentException();

	// K0337 = null type not allowed
	if (type==null) throw new CertificateException(com.ibm.oti.util.Msg.getString("K0337"));

	Provider provider = Security.getProvider (providerName);
	if (provider == null) throw new NoSuchProviderException (providerName);

	return toCertificateFactoryImplementation(type, provider);

 }

/**
 * Answers a new CertificateFactory of the given type.
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type	java.lang.String	Type of certificatem desired
 * @param		provider java.security.Provider Provider which has to implement the algorithm
 * @return		CertificateFactory	a concrete implementation for the certificate type desired.
 *
 * @exception	CertificateException	If the type cannot be found
 */
public static final CertificateFactory getInstance(String type, Provider provider)
 throws CertificateException
{
	// K0337 = null type not allowed
	if (type==null) throw new CertificateException(com.ibm.oti.util.Msg.getString("K0337"));

	if(provider == null ){
		throw new IllegalArgumentException();
	}

	return toCertificateFactoryImplementation(type, provider);
}

/**
 * Returns the Provider of the certificate factory represented by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Provider
 *					an instance of a subclass of java.security.Provider
 */
public final Provider getProvider() {
	return provider;
}
/**
 * Returns the Certificate type
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String	type of certificate being used
 */
public final String getType() {
	return type;
}
/**
 * Set the provider being used by the receiver to
 * the argument which should be an instance of a subclass of Provider
 *
 * @author		OTI
 * @version		initial
 *
 * @param		provider 	the Provider for the receiver
 */
void setProvider(Provider provider) {
	this.provider = provider;
}
/**
 * Set the certificate type being used by the receiver to
 * the argument.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type String
 *					the type of the certificate
 */
void setType(String type) {
	this.type = type;
}
/**
 * Answers a CertificateFactory for the type name
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type java.lang.String
 *					the type of certificate desired
 *
 * @return		The CertificateFactory for the type name
 *.
 * @throws CertificateException 	If the requested type is not available
 */
private static CertificateFactory toCertificateFactoryImplementation (String type) throws CertificateException{

	Provider[] providers = Security.getProviders();
	for (int i = 0; i < providers.length; i++) {
		Provider provider = (Provider) providers[i];
		try {
			return toCertificateFactoryImplementation (type, provider);
		} catch (CertificateException e) {
			// Just skip to next provider
		}
	}

	// Scanned all, found nothing
	throw new CertificateException(type);
}
/**
 * Answers a CertificateFactory for the type name
 *
 *
 * @author		OTI
 * @version		initial
 *
 * @param		type java.lang.String
 *					the type of certificate desired
 *
 * @param		provider		java.security.Provider
 *					the provider desired for the type.
 *
 * @return		The CertificateFactory for the type name
 *.
 * @throws CertificateException 	If the requested type is not available
 */
private static CertificateFactory toCertificateFactoryImplementation (String type, Provider provider) throws CertificateException{

	// First try to find the class corresponding to the algorithm name
	String certificateFactoryClassName;
	try {
		certificateFactoryClassName = lookupProperty(provider, KEY_PREFIX, type);
	} catch (ClassCastException e) {
		throw new CertificateException(type);
	}

	// If not found, exception
	if (certificateFactoryClassName == null)
		throw new CertificateException(type);

	// Now try to instantiate the certificate factory.
	try {
		Class certFactClass = Class.forName(certificateFactoryClassName, true, provider.getClass().getClassLoader());
		return createCertificateFactory (provider, certFactClass, type);
	} catch (ClassNotFoundException ex) {
			throw new CertificateException(type);
	}
}

private static String lookupProperty(Provider provider, String property) {
	String result = provider.getProperty(property);
	if (result != null) return result;
	String upper = property.toUpperCase();
	Enumeration keyEnum = provider.keys();
	while (keyEnum.hasMoreElements()) {
		String key = (String)keyEnum.nextElement();
		if (key.toUpperCase().equals(upper))
			return provider.getProperty(key);
	}
	return null;
}

private static String lookupProperty(Provider provider, String prefix, String name) {
	String property = prefix + name;
	String result = lookupProperty(provider, property);
	if (result != null) return result;
	result = lookupProperty(provider, "Alg.Alias." + property);
	if (result != null)
		return provider.getProperty(prefix + result);
	return null;
}

}
