
package java.security.cert;
/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 *	This class represents X.509 Certificate Revocation Lists (CRLs). They are used
 * to indicate that a given Certificate has expired already.
 *
 * @author		OTI
 * @version		initial
 *
 * @see CRL
 */

import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Principal;
import java.security.PublicKey;
import java.security.SignatureException;
import java.util.Arrays;
import java.util.Date;
import java.util.Set;
import javax.security.auth.x500.X500Principal;

public abstract class X509CRL extends CRL implements X509Extension {

/**
 * Constructs a new instance of this class
 *
 * @author		OTI
 * @version		initial
 *
 */
protected X509CRL() {
	super ("X.509");
}

/**
 * Return the CRL, DER-encoded.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]		the parameters in their primary encoding format.
 * @exception	CRLException	if encoding errors are detected
 */
public abstract byte[] getEncoded() throws CRLException;

/**
 * Return the issuer distinguished name for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Principal		The entity that signed the receiver, the issuer
 */
public abstract Principal getIssuerDN();

/**
 * Return the CRL 'nextUpdate' field
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Date	the CRL 'nextUpdate' field
 */
public abstract Date getNextUpdate();

/**
 * Return the revoked CRL given its serial number.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		serialNumber	BigInteger	The serial number of the desired Certificate
 * @return		X509CRLEntry	the revoked CRL given its serial number.
 */
public abstract X509CRLEntry getRevokedCertificate(BigInteger serialNumber);

/**
 * Return all the revoked Certificates in the receiver, as a Set of X509CRLEntry.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Set		all the revoked Certificates in the receiver, as a Set of X509CRLEntry.
 */
public abstract Set getRevokedCertificates();

/**
 * Return the signature algorithm name for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String	the signature algorithm name for the receiver.
 */
public abstract String getSigAlgName();

/**
 * Return the signature algorithm ID (OID) for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String	the signature algorithm ID (OID) for the receiver.
 */
public abstract String getSigAlgOID();

/**
 * Return the signature algorithm parameters for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]	the signature algorithm parameters for the receiver.
 */
public abstract byte[] getSigAlgParams();

/**
 * Return the actual signature for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]	the actual signature for the receiver.
 */
public abstract byte[] getSignature();

/**
 * Return the field 'tbsCertList' for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	CRLException	if encoding errors are detected
 * @return		byte[]	the field 'tbsCertList' for the receiver.
 */
public abstract byte[] getTBSCertList() throws CRLException;

/**
 * Return the field 'thisUpdate' for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Date	the field 'thisUpdate' for the receiver.
 */
public abstract Date getThisUpdate();

/**
 * Return the version number for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Date	the version number for the receiver.
 */
public abstract int getVersion();

/**
 * Verifies if the CRL was signed with the PrivateKey corresponding to the
 * PublicKey passed.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key		PublicKey		PublicKey to be used for signature verification
 * @exception	CRLException				if encoding errors are detected
 * @exception	NoSuchAlgorithmException	if unsupported algorithms are encountered
 * @exception	InvalidKeyException			if the provided key is not valid
 * @exception	NoSuchProviderException		if a provider for the signature algorithm can not be found
 * @exception	SignatureException			if signature errors are found
 */
public abstract void verify(PublicKey key)
                     throws CRLException,
                            NoSuchAlgorithmException,
                            InvalidKeyException,
                            NoSuchProviderException,
                            SignatureException;

/**
 * Verifies if the CRL was signed with the PrivateKey corresponding to the
 * PublicKey passed, from the given signature provider.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key			PublicKey	PublicKey to be used for signature verification
 * @param		sigProvider	String		name of the signature provider to use
 * @exception	CRLException				if encoding errors are detected
 * @exception	NoSuchAlgorithmException	if unsupported algorithms are encountered
 * @exception	InvalidKeyException			if the provided key is not valid
 * @exception	NoSuchProviderException		if a provider for the signature algorithm can not be found
 * @exception	SignatureException			if signature errors are found
 */
public abstract void verify(PublicKey key,
                            String sigProvider)
                     throws CRLException,
                            NoSuchAlgorithmException,
                            InvalidKeyException,
                            NoSuchProviderException,
                            SignatureException;

/**
 * Compares the argument to the receiver, and answers true
 * if they represent entries that are equal.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o Object
 *					the object to compare with this object.
 * @return		boolean
 *					<code>true</code>
 *						if the object is euqal to the receiver
 *					<code>false</code>
 *						if it is different from teh receiver
 * @see			#hashCode
 */
public boolean equals (Object o) {
	if (o == null) return false;
	if ( !(o instanceof X509CRL)) return false;
	X509CRL param = (X509CRL) o;
	try {
		return Arrays.equals (this.getEncoded(), param.getEncoded());
	} catch (CRLException e) {
		return false;
	}
}

/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>.equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int
 *					the receiver's hash.
 *
 * @see			#equals
 */
public int hashCode() {
	// The spec says the hashCode has to be computed from the encoded representation.
	// This sort of goes against the fact that a hash function should be fast. Otherwise,
	// what is the benefit of hashed collections ?
	// Unfortunately, here we are... converting to the encoded representation and then
	// doing a math with the bytes to make sure the same byte[] content will produce the same hash
	try {
		byte[] encoded = getEncoded();
		int hash = 1;
		for (int i = 0; i < encoded.length; i++) {
			hash *= encoded[i] & 0xff;
		}
		return hash;
	} catch (CRLException e) {
		return 0;
	}
}

/**
 * Returns the issuer of the CRL as X500Principal
 * @return X500Principal issuer
 */
public X500Principal getIssuerX500Principal(){
	return new X500Principal( getIssuerDN().getName() );
}

}
