
package java.security.cert;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

/**
 *	Abstract class for entries in a CRL, where the Certificates are of type X.509
 *
 * @author		OTI
 * @version		initial
 *
 * @see X509CRL
 */

import java.util.*;
import java.math.*;

public abstract class X509CRLEntry extends Object implements X509Extension {

/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 *
 */
public X509CRLEntry() {
}

/**
 * Compares the argument to the receiver, and answers true
 * if they represent entries that are equal.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o Object
 *					the object to compare with this object.
 * @return		boolean
 *					<code>true</code>
 *						if the object is euqal to the receiver
 *					<code>false</code>
 *						if it is different from teh receiver
 * @see			#hashCode
 */
public boolean equals (Object o) {
	if (o == null) return false;
	if ( !(o instanceof X509CRLEntry)) return false;
	X509CRLEntry param = (X509CRLEntry) o;
	try {
		return Arrays.equals (this.getEncoded(), param.getEncoded());
	} catch (CRLException e) {
		return false;
	}
}

/**
 * Return the ASN.1 DER encoded representation for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	CRLException	if encoding problems are detected.
 * @return		byte[]		the ASN.1 DER encoded representation for the receiver
 *
 */
public abstract byte[] getEncoded() throws CRLException;

/**
 * Return the serial number for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger		the serial number for the receiver
 *
 */
public abstract BigInteger getSerialNumber();

/**
 * Return the revocation date for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean		the revocation date for the receiver.
 *
 */
public abstract Date getRevocationDate();

/**
 * Return true if the receiver has extensions, false otherwise.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean		true if the receiver has extensions, false otherwise.
 *
 */
public abstract boolean hasExtensions();

/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>.equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int
 *					the receiver's hash.
 *
 * @see			#equals
 */
public int hashCode() {
	// The spec says the hashCode has to be computed from the encoded representation.
	// This sort of goes against the fact that a hash function should be fast. Otherwise,
	// what is the benefit of hashed collections ?
	// Unfortunately, here we are... converting to the encoded representation and then
	// doing a math with the bytes to make sure the same byte[] content will produce the same hash
	try {
		byte[] encoded = getEncoded();
		int hash = 1;
		for (int i = 0; i < encoded.length; i++) {
			hash *= encoded[i] & 0xff;
		}
		return hash;
	} catch (CRLException e) {
		return 0;
	}
}

/**
 * Answers a string containing a concise, human-readable
 * description of the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					a printable representation for the receiver.
 */
public abstract String toString ();

}
