
package java.security.cert;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 *	This class represents X.509 Certificates.
 *
 * @author		OTI
 * @version		initial
 *
 * @see Certificate
 */

import java.util.*;
import java.math.*;
import java.security.*;
import javax.security.auth.x500.X500Principal;

public abstract class X509Certificate extends Certificate implements X509Extension {
	private static final long serialVersionUID = -6439739127092946520L;

/**
 * Constructs a new instance of this class
 *
 * @author		OTI
 * @version		initial
 *
 */
protected X509Certificate() {
	super ("X.509");
}

/**
 * Checks if the certificate represented by the receiver is still valid.
 * Throws an exception if the certificate is not valid.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	CertificateExpiredException		The certificate is no longer valid, it has expired
 * @exception	CertificateNotYetValidException	The certificate is not valid yet
 */
public abstract void checkValidity() throws CertificateExpiredException, CertificateNotYetValidException;

/**
 * Checks if the certificate represented by the receiver is valid for the
 * date provided. Throws an exception if the certificate is not valid.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		date	Date	date to use when testing the validity of the certificate
 * @exception	CertificateExpiredException		The certificate is no longer valid, it has expired
 * @exception	CertificateNotYetValidException	The certificate is not valid yet
 */
public abstract void checkValidity(Date date) throws CertificateExpiredException, CertificateNotYetValidException;

/**
 * Return the path length constraint for the BasicConstraints for the receiver.
 * The value comes from the BasicConstraints extension (OID = 2.5.29.19)
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the path length constraint for the receiver if BasicConstraints is defined.
 * 						Return -1 otherwise.
 *
 */
public abstract int getBasicConstraints();

/**
 * Return the issuer distinguished name for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Principal	the issuer distinguished name for the receiver
 */
public abstract Principal getIssuerDN();

/**
 * Return the issuer X500Principal for the receiver
 *
 * @return X500Principal  the issuer principal for the reciever
 */
public X500Principal getIssuerX500Principal(){
	return new X500Principal( getIssuerDN().getName() );
}

/**
 * Return the issuer unique ID for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean[]	the issuer unique ID for the receiver. null if not present.
 */
public abstract boolean[] getIssuerUniqueID();

/**
 * Return a bit array representation for the key usage for the receiver. The
 * value comes from the KeyUsage extension (OID = 2.5.29.15)
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean[]	 a bit array representation for the key usage for the receiver.
 */
public abstract boolean[] getKeyUsage();

/**
 * Return the Date until when the certificate is valid
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Date		the Date until when the certificate is valid
 */
public abstract Date getNotAfter();

/**
 * Return the Date until when the certificate starts to be valid
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Date		the Date until when the certificate starts to be valid
 */
public abstract Date getNotBefore();

/**
 * Return the serial number for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	the serial number for the receiver
 */
public abstract BigInteger getSerialNumber();

/**
 * Return the certificate signature algorithm name for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String		the certificate signature algorithm name for the receiver.
 */
public abstract String getSigAlgName();

/**
 * Return the certificate signature OID for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String		the certificate signature OID for the receiver.
 */
public abstract String getSigAlgOID();

/**
 * Return the signature algorithm parameters for the receiver, DER-encoded.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]		the signature algorithm parameters for the receiver, DER-encoded. null if not present.
 */
public abstract byte[] getSigAlgParams();

/**
 * Return the actual signature value for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte[]		the actual signature value for the receiver
 */
public abstract byte[] getSignature();

/**
 * Return the subject distinguished name for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Principal	the subject distinguished name for the receiver
 */
public abstract Principal getSubjectDN();

/**
 * Return the subject X500Principal for the receiver
 *
 * @return X500Principal  the subject principal for the reciever
 */
public X500Principal getSubjectX500Principal(){
	return new X500Principal( getSubjectDN().getName() );
}

/**
 * Return the subject unique ID for the receiver
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean[]	the subject unique ID for the receiver
 */
public abstract boolean[] getSubjectUniqueID();

/**
 * Return the tbsCertificate for the receiver, DER-encoded
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean[]	the subject unique ID for the receiver
 */
public abstract byte[] getTBSCertificate() throws CertificateEncodingException;

/**
 * Return the version number for this certificate (1 for v1, 2 for v2, 3 for v3)
 *
 * @author		OTI
 * @version		initial
 *
 * @return		int		the version number for this certificate (1 for v1, 2 for v2, 3 for v3)
 */
public abstract int getVersion();

/**
 * Returns the ExtendedKeyUsage extension (ASN1 OID 2.5.29.37) as a List of OIDs of supported
 * key usage(s)
 * @return List extended key usage OIDs
 * @throws CertificateParsingException if an error occurs reading the extension
 * @throws UnsupportedOperationException if not implemented by X509Certificate implementation
 */
public List getExtendedKeyUsage() throws CertificateParsingException {
	throw new UnsupportedOperationException();
}

/**
 * Returns a Collection containing the SubjectAltName extension for the Certificate if present.
 * null is returned if the extension is not present.
 *
 * The following is the definition of the ASN.1 SubjectAltName field of X509Certificates:
 *     SubjectAltName ::= GeneralNames
 *
 *     GeneralNames ::= SEQUENCE SIZE (1..MAX) OF GeneralName
 *
 *     GeneralName ::= CHOICE {
 *         otherName                       [0]     OtherName,
 *         rfc822Name                      [1]     IA5String,
 *         dNSName                         [2]     IA5String,
 *         x400Address                     [3]     ORAddress,
 *         directoryName                   [4]     Name,
 *         ediPartyName                    [5]     EDIPartyName,
 *         uniformResourceIdentifier       [6]     IA5String,
 *         iPAddress                       [7]     OCTET STRING,
 *         registeredID                    [8]     OBJECT IDENTIFIER}
 *
 * The elements in the Collection are Lists where the first element is an Integer defining the
 * type of the GeneralName (i.e., 0=OtherName, 1=rfc822Name, ... ) and the second is the
 * GeneralName.
 *
 * rfc822Name, dNSName and uniformResourceIdentifiers are returned as Strings.
 * directoryName elements are returned as Strings.
 * iPAddress for IPv4 addresses is returned as a String in dotted-quad notation.  IPv6
 * 	addresses are returned in the form "a1:a2:...:a8" where a1-a8 are the eight 16-bit pieces
 * 	of the address.
 * Elements of type OBJECT IDENTIFIER are returned in their String representation as a series
 * 	of positive integers seperated by periods.
 * Other types are returned as byte arrays.
 *
 * @return Collection Collection of Lists of General Names encoded as described above.
 * @throws CertificateParsingException
 */
public Collection getSubjectAlternativeNames() throws CertificateParsingException{
	throw new UnsupportedOperationException();
}

/**
 * Returns a Collection containing the IssuerAltName extension for the Certificate if present.
 * null is returned if the extension is not present.
 *
 * The following is the definition of the ASN.1 IssuerAltName field of X509Certificates:
 *     IssuerAltName ::= GeneralNames
 *
 * The definition of GeneralNames and the manner in which its values are output
 * is describe in getSubjectAlternativeNames.
 *
 * @return Collection Collection of Lists of General Names encoded as described above.
 * @throws CertificateParsingException
 */
public Collection getIssuerAlternativeNames() throws CertificateParsingException{
	throw new UnsupportedOperationException();
}

}
