package java.security.spec;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This interface represents RSA-specific private keys represented using
 * the Chinese Remainder Theorem.
 *
 * @author		OTI
 * @version		initial
 *
 * @see 		Key
 * @see 		Signature
 */

import java.math.*;

public class RSAPrivateCrtKeySpec extends java.security.spec.RSAPrivateKeySpec {

	private BigInteger pubExpE;
	private BigInteger primeP;
	private BigInteger primeQ;
	private BigInteger dPCRT;
	private BigInteger dQCRT;
	private BigInteger qInvCRT;

/**
 * Create a RSAPrivateCRTKeySpec
 *
 * @param modN RSA modulus (n)
 * @param pubExpE RSA public exponent (e)
 * @param privExpD RSA private exponent (d)
 * @param primeP RSA prime modulus factor (p)
 * @param primeQ RSA prime modulus factor (q)
 * @param dPCRT RSA CRT exponent for p (dP)
 * @param dQCRT RSA CRT exponent for q (dQ)
 * @param qInvCRT RSA CRT co-efficient (qInv)
 */
public RSAPrivateCrtKeySpec(
        BigInteger modN,
		BigInteger pubExpE,
		BigInteger privExpD,
		BigInteger primeP,
		BigInteger primeQ,
		BigInteger dPCRT,
		BigInteger dQCRT,
		BigInteger qInvCRT)
{
	super (modN, privExpD);

    this.pubExpE = pubExpE;
	this.primeP = primeP;

	this.primeQ = primeQ;
	this.dPCRT = dPCRT;
	this.dQCRT = dQCRT;
	this.qInvCRT = qInvCRT;
}

/**
 * Return the RSA public exponent for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	public exponent for the receiver.
 *
 */
public BigInteger getPublicExponent() {
	return pubExpE;
}

/**
 * Return the RSA CRT coefficient for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	CRT coefficient for the receiver.
 *
 */
public BigInteger getCrtCoefficient() {
	return qInvCRT;
}

/**
 * Return the RSA prime P for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	prime P for the receiver.
 *
 */
public BigInteger getPrimeP() {
	return primeP;
}

/**
 * Return the RSA prime Q for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	prime Q for the receiver.
 *
 */
public BigInteger getPrimeQ() {
	return primeQ;
}

/**
 * Return the RSA prime exponent P for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	prime exponent P for the receiver.
 *
 */
public BigInteger getPrimeExponentP() {
	return dPCRT;
}

/**
 * Return the RSA prime exponent Q for the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		BigInteger	prime exponent Q for the receiver.
 *
 */
public BigInteger getPrimeExponentQ() {
	return dQCRT;
}

}
