package java.text;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.util.Arrays;
import java.util.Vector;

/**
 * ChoiceFormat is used to assocate strings with ranges of double
 * values. The strings and ranges are either specified using arrays
 * or with a pattern which is parsed to determine the Strings and
 * ranges.
 *
 * @author		OTI
 * @version		initial
 */

public class ChoiceFormat extends NumberFormat {
	private static final long serialVersionUID = 1795184449645032964L;

	private double[] choiceLimits;
	private String[] choiceFormats;

/**
 * Constructs a new ChoiceFormat with the specified ranges and associated strings.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		limits	an array of double, the ranges are greater or equal to the value in lower index
 *				up to less than the value in the next higher index. The bounds of the lowest and highest
 *				indexes are negative and positive infinity.
 * @param		formats	the strings associated with the ranges. The lower bound of the associated range is
 *				at the same index as the string.
 */
public ChoiceFormat(double[] limits, String[] formats) {
	setChoices(limits, formats);
}
/**
 * Constructs a new ChoiceFormat with the strings and ranges parsed from the
 * specified pattern.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		template	the pattern of strings and ranges
 *
 * @exception	IllegalArgumentException then an error occurs parsing the pattern
 */
public ChoiceFormat(String template) {
	applyPattern(template);
}
/**
 * Parses the pattern to determine new strings and ranges for this ChoiceFormat.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		template	the pattern of strings and ranges
 */
public void applyPattern(String template) {
	boolean first = true;
	double[] limits = new double[5];
	Vector formats = new Vector();
	int length = template.length(), limitCount = 0, index = 0;
	StringBuffer buffer = new StringBuffer();
	NumberFormat format = NumberFormat.getInstance();
	ParsePosition position = new ParsePosition(0);
	try {
		while (true) {
			index = skipWhitespace(template, index);
			if (index >= length) {
				if (limitCount == limits.length) choiceLimits = limits;
				else {
					choiceLimits = new double[limitCount];
					System.arraycopy(limits, 0, choiceLimits, 0, limitCount);
				}
				choiceFormats = new String[formats.size()];
				for (int i=0; i<formats.size(); i++)
					choiceFormats[i] = (String)formats.elementAt(i);
				return;
			}

			position.setIndex(index);
			Number value = format.parse(template, position);
			index = skipWhitespace(template, position.getIndex());
			if (position.getErrorIndex() != -1 || index >= length)
				throw new IllegalArgumentException();
			char ch = template.charAt(index++);
			if (limitCount == limits.length) {
				double[] newLimits = new double[limitCount * 2];
				System.arraycopy(limits, 0, newLimits, 0, limitCount);
				limits = newLimits;
			}
			double next;
			switch(ch) {
				case '#':
				case '\u2264':
					next = value.doubleValue();
					break;
				case '<':
					if (first) throw new IllegalArgumentException();
					next = nextDouble(value.doubleValue());
					break;
				default:
					throw new IllegalArgumentException();
			}
			first = false;
			if (limitCount > 0 && next <= limits[limitCount-1])
				throw new IllegalArgumentException();
			buffer.setLength(0);
			position.setIndex(index);
			upTo(template, position, buffer, '|');
			index = position.getIndex();
			limits[limitCount++] = next;
			formats.addElement(buffer.toString());
		}
	} catch (IllegalArgumentException e) {
		choiceFormats = null;
		choiceLimits = null;
	}
}
/**
 * Answers a new instance of ChoiceFormat with the same
 * ranges and strings as this ChoiceFormat.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a shallow copy of this ChoiceFormat
 *
 * @see			java.lang.Cloneable
 */
public Object clone() {
	ChoiceFormat clone = (ChoiceFormat)super.clone();
	clone.choiceLimits = (double[])choiceLimits.clone();
	clone.choiceFormats = (String[])choiceFormats.clone();
	return clone;
}
/**
 * Compares the specified object to this ChoiceFormat and answer if they
 * are equal. The object must be an instance of ChoiceFormat and have the
 * same limits and formats.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to compare with this object
 * @return		true if the specified object is equal to this ChoiceFormat, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals(Object object) {
	if (this == object) return true;
	if (!(object instanceof ChoiceFormat)) return false;
	ChoiceFormat choice = (ChoiceFormat)object;
	return Arrays.equals(choiceLimits, choice.choiceLimits) &&
		Arrays.equals(choiceFormats, choice.choiceFormats);
}
/**
 * Appends to the specified StringBuffer the string associated with the range
 * in which the specified double value fits.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value	the double to format
 * @param		buffer	the StringBuffer
 * @param		field	a FieldPosition which is ignored
 * @return		the StringBuffer parameter <code>buffer</code>
 */
public StringBuffer format(double value, StringBuffer buffer, FieldPosition field) {
	if (Double.isNaN(value) || (choiceLimits.length > 1 && value < choiceLimits[1]))
		return buffer.append(choiceFormats[0]);
	for (int i=2; i<choiceLimits.length; i++)
		if (value >= choiceLimits[i-1] && value < choiceLimits[i])
			return buffer.append(choiceFormats[i-1]);
	return buffer.append(choiceFormats[choiceFormats.length-1]);
}
/**
 * Appends to the specified StringBuffer the string associated with the range
 * in which the specified long value fits.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value	the long to format
 * @param		buffer	the StringBuffer
 * @param		field	a FieldPosition which is ignored
 * @return		the StringBuffer parameter <code>buffer</code>
 */
public StringBuffer format(long value, StringBuffer buffer, FieldPosition field) {
	return format((double)value, buffer, field);
}
/**
 * Answers the Strings associated with the ranges of this ChoiceFormat.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an array of String
 */
public Object[] getFormats() {
	return choiceFormats;
}
/**
 * Answers the ranges of this ChoiceFormat.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an array of double, the ranges are greater or equal to the value in lower index
 *				up to less than the value in the next higher index. The bounds of the lowest and highest
 *				indexes are negative and positive infinity.
 */
public double[] getLimits() {
	return choiceLimits;
}
/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	int hashCode = 0;
	for (int i=0; i<choiceLimits.length; i++) {
		long v = Double.doubleToLongBits(choiceLimits[i]);
		hashCode += (int)(v ^ (v >>> 32)) + choiceFormats[i].hashCode();
	}
	return hashCode;
}
/**
 * Answers the double value which is closest to the specified double but
 * larger.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value	a double value
 * @return		the next larger double value
 */
public static final double nextDouble(double value) {
	if (value == Double.POSITIVE_INFINITY) return value;
	long bits;
	// Handle -0.0
	if (value == 0) bits = 0;
	else bits = Double.doubleToLongBits(value);
	return Double.longBitsToDouble(value < 0 ? bits - 1 : bits + 1);
}
/**
 * Answers the double value which is closest to the specified double but
 * either larger or smaller as specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value	a double value
 * @param		increment	true to get a larger value, false to get a smaller value
 * @return		the next larger or smaller double value
 */
public static double nextDouble(double value, boolean increment) {
	return increment ? nextDouble(value) : previousDouble(value);
}
/**
 * Parse a Double from the specified String starting at the index specified by
 * the ParsePosition. The String is compared to the strings of this ChoiceFormat
 * and if a match occurs, the answer is the lower bound of the corresponding range.
 * If the string is successfully parsed, the index of the ParsePosition is updated
 * to the index following the parsed text.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string	the String to parse
 * @param		position	the ParsePosition, updated on return with the index following
 *							the parsed text, or on error the index is unchanged and the
 *							error index is set to the index where the error occurred
 * @return		a Double resulting from the parse, or Double.NaN if there is an error
 */
public Number parse(String string, ParsePosition position) {
	int offset = position.getIndex();
	for (int i=0; i<choiceFormats.length; i++) {
		if (string.startsWith(choiceFormats[i], offset)) {
			position.setIndex(offset + choiceFormats[i].length());
			return new Double(choiceLimits[i]);
		}
	}
	position.setErrorIndex(offset);
	return new Double(Double.NaN);
}
/**
 * Answers the double value which is closest to the specified double but
 * smaller.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value	a double value
 * @return		the next smaller double value
 */
public static final double previousDouble(double value) {
	if (value == Double.NEGATIVE_INFINITY) return value;
	long bits;
	// Handle 0.0
	if (value == 0) bits = 0x8000000000000000L;
	else bits = Double.doubleToLongBits(value);
	return Double.longBitsToDouble(value <= 0 ? bits + 1 : bits - 1);
}
/**
 * Sets the ranges and associated strings of this ChoiceFormat.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		limits	an array of double, the ranges are greater or equal to the value in lower index
 *				up to less than the value in the next higher index. The bounds of the lowest and highest
 *				indexes are negative and positive infinity.
 * @param		formats	the strings associated with the ranges. The lower bound of the range is at the
 *				same index as the string.
 */
public void setChoices(double[] limits, String[] formats) {
	if (limits.length != formats.length) throw new IllegalArgumentException();
	choiceLimits = limits;
	choiceFormats = formats;
}
private int skipWhitespace(String string, int index) {
	int length = string.length();
	while (index < length && Character.isWhitespace(string.charAt(index))) index++;
	return index;
}
/**
 * Answers the pattern of this ChoiceFormat which specified the
 * ranges and their associated strings.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	the pattern
 */
public String toPattern() {
	if (choiceLimits == null) {
		return "";
	}
	StringBuffer buffer = new StringBuffer();
	for (int i=0; i<choiceLimits.length; i++) {
		String previous = String.valueOf(previousDouble(choiceLimits[i]));
		String limit = String.valueOf(choiceLimits[i]);
		if (previous.length() < limit.length()) {
			buffer.append(previous);
			buffer.append('<');
		} else {
			buffer.append(limit);
			buffer.append('#');
		}
		boolean quote;
		if (quote = (choiceFormats[i].indexOf('|') != -1))
			buffer.append('\'');
		buffer.append(choiceFormats[i]);
		if (quote) buffer.append('\'');
		buffer.append('|');
	}
	if (buffer.length() > 0) {
		buffer.setLength(buffer.length() - 1);
	}
	return buffer.toString();
}
}
