package java.text;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.util.Comparator;
import java.util.Locale;
import java.util.Vector;
import com.ibm.oti.util.ExtendedResourceBundle;
import java.security.AccessController;
import com.ibm.oti.util.PriviAction;

/**
 * Collator is an abstract class which is the root of classes which provide
 * Locale specific String comparison to determine their ordering with respect
 * to each other.
 *
 * @author		OTI
 * @version		initial
 */

public abstract class Collator implements Comparator, Cloneable {
	static final int EQUAL = 0;
	static final int GREATER = 1;
	static final int LESS = -1;

	/**
	 * Constant used to specify the decomposition rule.
	 */
	public static final int NO_DECOMPOSITION = 0;
	/**
	 * Constant used to specify the decomposition rule.
	 */
	public static final int CANONICAL_DECOMPOSITION = 1;
	/**
	 * Constant used to specify the decomposition rule.
	 */
	public static final int FULL_DECOMPOSITION = 2;

	/**
	 * Constant used to specify the collation strength.
	 */
	public static final int PRIMARY = 0;
	/**
	 * Constant used to specify the collation strength.
	 */
	public static final int SECONDARY = 1;
	/**
	 * Constant used to specify the collation strength.
	 */
	public static final int TERTIARY = 2;
	/**
	 * Constant used to specify the collation strength.
	 */
	public static final int IDENTICAL = 3;

	private static int CACHE_SIZE;
	private static Vector cache = new Vector();

	private int strength = TERTIARY, decomposition = CANONICAL_DECOMPOSITION;

	static {
		// CACHE_SIZE includes key and value, so needs to be double
		String cacheSize = (String)AccessController.doPrivileged(new PriviAction("collator.cache"));
		if (cacheSize != null) {
			try {
				CACHE_SIZE = Integer.parseInt(cacheSize);
			} catch (NumberFormatException e) {
				CACHE_SIZE = 6;
			}
		} else CACHE_SIZE = 6;
	}

/**
 * Constructs a new instance of this Collator.
 *
 * @author		OTI
 * @version		initial
 */
protected Collator() {
}
/**
 * Answers a new Collator with the same decomposition rule and strength
 * value as this Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a shallow copy of this Collator
 *
 * @see			java.lang.Cloneable
 */
public Object clone() {
	try {
		return super.clone();
	} catch (CloneNotSupportedException e) {return null;}
}
/**
 * Compares the two objects to determine their relative ordering.
 * The objects must be Strings.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object1	the first String to compare
 * @param		object2	the second String to compare
 * @return		an int < 0 if object1 is less than object2,
 *				0 if they are equal, and > 0 if object1 is greater than object2
 *
 * @exception	ClassCastException when the objects are not Strings
 */
public int compare(Object object1, Object object2) {
	return compare((String)object1, (String)object2);
}
/**
 * Compares the two Strings to determine their relative ordering.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string1	the first String to compare
 * @param		string2	the second String to compare
 * @return		an int < 0 if string1 is less than string2,
 *				0 if they are equal, and > 0 if string1 is greater than string2
 */
public abstract int compare(String string1, String string2);
/**
 * Compares the specified object to this Collator and answer if they
 * are equal. The object must be an instance of Collator and have the
 * same strength and decomposition values.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to compare with this object
 * @return		true if the specified object is equal to this Collator, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals(Object object) {
	if (!(object instanceof Collator)) return false;
	Collator collator = (Collator)object;
	return strength == collator.strength && decomposition == collator.decomposition;
}
/**
 * Compares the two Strings using the collation rules to determine
 * if they are equal.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string1	the first String to compare
 * @param		string2	the second String to compare
 * @return		true if the strings are equal using the collation rules, false otherwise
 */
public boolean equals(String string1, String string2) {
	return compare(string1, string2) == 0;
}

/**
 * Gets the list of installed Locales which support Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an array of Locale
 */
public static Locale[] getAvailableLocales() {
	return Locale.getAvailableLocales();
}

/**
 * Answers a CollationKey for the specified String for this Collator
 * with the current decomposition rule and stength value.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a CollationKey
 */
public abstract CollationKey getCollationKey(String string);
/**
 * Answers the decomposition rule for this Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the decomposition rule, either NO_DECOMPOSITION, CANONICAL_DECOMPOSITION
 *				or FULL_DECOMPOSITION
 */
public int getDecomposition() {
	return decomposition;
}
/**
 * Answers a Collator instance which is appropriate for the default Locale.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a Collator
 */
public static Collator getInstance() {
	return getInstance(Locale.getDefault());
}
/**
 * Answers a Collator instance which is appropriate for the specified Locale.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		locale	the Locale
 * @return		a Collator
 */
public static Collator getInstance(Locale locale) {
	String key = locale.toString();
	for (int i=cache.size()-1; i>=0; i-=2) {
		if (cache.elementAt(i).equals(key))
			return (Collator)((Collator)cache.elementAt(i-1)).clone();
	}
	ExtendedResourceBundle bundle = (ExtendedResourceBundle)Format.getBundle(locale);
	String rules = (String)bundle.getObject(com.ibm.oti.locale.Locale.COLLATION);
	try {
		int decomp = CANONICAL_DECOMPOSITION;
		try {
			Object decompProp = bundle.getObject(com.ibm.oti.locale.Locale.COLLATIONDECOMP);
			decomp = ((Integer)decompProp).intValue();
		} catch (java.util.MissingResourceException e) {}
		Collator collator = new RuleBasedCollator(CollationRules.DEFAULTRULES + rules,
				decomp);
		collator.setDecomposition(NO_DECOMPOSITION);
		if (CACHE_SIZE > 0) {
			if (cache.size() >= CACHE_SIZE) {
				cache.removeElementAt(0);
				cache.removeElementAt(0);
			}
			cache.addElement(collator.clone());
			cache.addElement(key);
		}
		return collator;
	} catch (ParseException e) {throw new IllegalArgumentException(e.getMessage());}
}
/**
 * Answers the strength value for this Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the strength value, either PRIMARY, SECONDARY, TERTIARY, or IDENTICAL
 */
public int getStrength() {
	return strength;
}
/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals(Object)
 * @see			#equals(String, String)
 */
public abstract int hashCode();
/**
 * Sets the decomposition rule for this Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value the decomposition rule, either NO_DECOMPOSITION, CANONICAL_DECOMPOSITION
 *				or FULL_DECOMPOSITION
 *
 * @exception	IllegalArgumentException when the decomposition rule is not valid
 */
public void setDecomposition(int value) {
	switch(value) {
		case NO_DECOMPOSITION:
		case CANONICAL_DECOMPOSITION:
		case FULL_DECOMPOSITION:
			break;
		default:
			throw new IllegalArgumentException();
	}
	decomposition = value;
}
/**
 * Sets the strength value for this Collator.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		value the strength value, either PRIMARY, SECONDARY, TERTIARY, or IDENTICAL
 *
 * @exception	IllegalArgumentException when the strength value is not valid
 */
public void setStrength(int value) {
	switch(value) {
		case PRIMARY:
		case SECONDARY:
		case TERTIARY:
		case IDENTICAL:
			break;
		default:
			throw new IllegalArgumentException();
	}
	strength = value;
}
}
