package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * AbstractList is an abstract implementation of the List interface, optimized for a
 * backing store which supports random access. This implementation does not support
 * adding or replacing. A subclass must implement the abstract methods get() and size().
 *
 * @author		OTI
 * @version		initial
 */

public abstract class AbstractList extends AbstractCollection implements List {
	protected transient int modCount = 0;

private class SimpleListIterator implements Iterator {
	int pos = -1, expectedModCount;
	int lastPosition = -1;

SimpleListIterator() {
	expectedModCount = modCount;
}
public boolean hasNext() {
	return pos + 1 < size();
}
public Object next() {
	if (expectedModCount == modCount) {
		try {
			Object result = get(pos + 1);
			lastPosition = ++pos;
			return result;
		} catch (IndexOutOfBoundsException e) {
			throw new NoSuchElementException();
		}
	} else throw new ConcurrentModificationException();
}
public void remove() {
	if (expectedModCount == modCount) {
		try {
			AbstractList.this.remove(lastPosition);
		} catch (IndexOutOfBoundsException e) {
			throw new IllegalStateException();
		}
		if (modCount != expectedModCount)
			expectedModCount++;
		if (pos == lastPosition) pos--;
		lastPosition = -1;
	} else throw new ConcurrentModificationException();
}
}

private final class FullListIterator extends SimpleListIterator implements ListIterator {
FullListIterator(int start) {
	if (0 <= start && start <= size()) pos = start - 1;
	else throw new IndexOutOfBoundsException();
}
public void add(Object object) {
	if (expectedModCount == modCount) {
		try {
			AbstractList.this.add(pos + 1, object);
		} catch (IndexOutOfBoundsException e) {
			throw new NoSuchElementException();
		}
		pos++;
		lastPosition = -1;
		if (modCount != expectedModCount)
			expectedModCount++;
	} else throw new ConcurrentModificationException();
}
public boolean hasPrevious() {
	return pos >= 0;
}
public int nextIndex() {
	return pos + 1;
}
public Object previous() {
	if (expectedModCount == modCount) {
		try {
			Object result = get(pos);
			lastPosition = pos;
			pos--;
			return result;
		} catch (IndexOutOfBoundsException e) {
			throw new NoSuchElementException();
		}
	} else throw new ConcurrentModificationException();
}
public int previousIndex() {
	return pos;
}
public void set(Object object) {
	if (expectedModCount == modCount) {
		try {
			AbstractList.this.set(lastPosition, object);
		} catch (IndexOutOfBoundsException e) {
			throw new IllegalStateException();
		}
	} else throw new ConcurrentModificationException();
}
}

private static final class SubAbstractListRandomAccess extends SubAbstractList implements RandomAccess {
	SubAbstractListRandomAccess(AbstractList list, int start, int end) {
		super(list, start, end);
	}
}

private static class SubAbstractList extends AbstractList {
	private final AbstractList fullList;
	private int offset, size;

private static final class SubAbstractListIterator implements ListIterator {
	private final SubAbstractList subList;
	private final ListIterator iterator;
	private int start, end;

SubAbstractListIterator(ListIterator it, SubAbstractList list, int offset, int length) {
	iterator = it;
	subList = list;
	start = offset;
	end = start + length;
}
public void add(Object object) {
	iterator.add(object);
	subList.sizeChanged(true);
	end++;
}
public boolean hasNext() {
	return iterator.nextIndex() < end;
}
public boolean hasPrevious() {
	return iterator.previousIndex() >= start;
}
public Object next() {
	if (iterator.nextIndex() < end) return iterator.next();
	throw new NoSuchElementException();
}
public int nextIndex() {
	return iterator.nextIndex() - start;
}
public Object previous() {
	if (iterator.previousIndex() >= start) return iterator.previous();
	throw new NoSuchElementException();
}
public int previousIndex() {
	int previous = iterator.previousIndex();
	if (previous >= start) return previous - start;
	return -1;
}
public void remove() {
	iterator.remove();
	subList.sizeChanged(false);
	end--;
}
public void set(Object object) {
	iterator.set(object);
}
}

SubAbstractList(AbstractList list, int start, int end) {
	fullList = list;
	modCount = fullList.modCount;
	offset = start;
	size = end - start;
}
public void add(int location, Object object) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size) {
			fullList.add(location+offset, object);
			size++;
			modCount = fullList.modCount;
		} else throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
public boolean addAll(int location, Collection collection) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size) {
			boolean result = fullList.addAll(location + offset, collection);
			if (result) {
				size += collection.size();
				modCount = fullList.modCount;
			}
			return result;
		} else throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
public boolean addAll(Collection collection) {
	if (modCount == fullList.modCount) {
		boolean result = fullList.addAll(size, collection);
		if (result) {
			size += collection.size();
			modCount = fullList.modCount;
		}
		return result;
	} else throw new ConcurrentModificationException();
}
public Object get(int location) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size)
			return fullList.get(location+offset);
		throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
public Iterator iterator() {
	return listIterator(0);
}
public ListIterator listIterator(int location) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size)
			return new SubAbstractListIterator(fullList.listIterator(location + offset), this, offset, size);
		throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
public Object remove(int location) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size) {
			Object result = fullList.remove(location + offset);
			size--;
			modCount = fullList.modCount;
			return result;
		} else throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
protected void removeRange(int start, int end) {
	if (start != end) {
		if (modCount == fullList.modCount) {
			fullList.removeRange(start + offset, end + offset);
			size -= end - start;
			modCount = fullList.modCount;
		} else throw new ConcurrentModificationException();
	}
}
public Object set(int location, Object object) {
	if (modCount == fullList.modCount) {
		if (0 <= location && location <= size)
			return fullList.set(location + offset, object);
		throw new IndexOutOfBoundsException();
	} else throw new ConcurrentModificationException();
}
public int size() {
	return size;
}
void sizeChanged(boolean increment) {
	if (increment) size++;
	else size--;
	modCount = fullList.modCount;
}
}

/**
 * Constructs a new instance of this AbstractList.
 *
 * @author		OTI
 * @version		initial
 */
protected AbstractList() {
}
/**
 * Inserts the specified object into this List at the specified
 * location. The object is inserted before any previous element at
 * the specified location. If the location is equal to the size of
 * this List, the object is added at the end.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location the index at which to insert
 * @param		object	the object to add
 *
 * @exception	UnsupportedOperationException when adding to this List
 *				is not supported
 * @exception	ClassCastException when the class of the object is inappropriate
 *				for this List
 * @exception	IllegalArgumentException when the object cannot be added
 *				to this List
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 */
public void add(int location, Object object) {
	throw new UnsupportedOperationException();
}
/**
 * Adds the specified object at the end of this List.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to add
 * @return		true
 *
 * @exception	UnsupportedOperationException when adding to this List
 *				is not supported
 * @exception	ClassCastException when the class of the object is inappropriate
 *				for this List
 * @exception	IllegalArgumentException when the object cannot be added
 *				to this List
 */
public boolean add(Object object) {
	add(size(), object);
	return true;
}
/**
 * Inserts the objects in the specified Collection at the specified location
 * in this List. The objects are added in the order they are returned
 * from the Collection iterator.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location the index at which to insert
 * @param		collection	the Collection of objects
 * @return		true if this List is modified, false otherwise
 *
 * @exception	UnsupportedOperationException when adding to this List
 *				is not supported
 * @exception	ClassCastException when the class of an object is inappropriate
 *				for this List
 * @exception	IllegalArgumentException when an object cannot be added
 *				to this List
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 */
public boolean addAll(int location, Collection collection) {
	Iterator it = collection.iterator();
	while (it.hasNext())
		add(location++, it.next());
	return !collection.isEmpty();
}
/**
 * Removes all elements from this List, leaving it empty.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	UnsupportedOperationException when removing from this List
 *				is not supported
 *
 * @see			List#isEmpty
 * @see			List#size
 */
public void clear() {
	removeRange(0, size());
}
/**
 * Compares the specified object to this List and answer if they
 * are equal. The object must be a List which contains the same
 * objects in the same order.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to compare with this object
 * @return		true if the specified object is equal to this List, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals(Object object) {
	if (this == object) return true;
	if (object instanceof List) {
		List list = (List)object;
		if (list.size() != size()) return false;

		Iterator it1 = iterator(), it2 = list.iterator();
		while (it1.hasNext()) {
			Object e1 = it1.next(), e2 = it2.next();
			if (!(e1 == null ? e2 == null : e1.equals(e2))) return false;
		}
		return true;
	}
	return false;
}
/**
 * Answers the element at the specified location in this List.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location	the index of the element to return
 * @return		the element at the specified index
 *
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 */
public abstract Object get(int location);
/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	int result = 1;
	Iterator it = iterator();
	while (it.hasNext()) {
		Object object = it.next();
		result = (31 * result) + (object == null ? 0 : object.hashCode());
	}
	return result;
}
/**
 * Searches this List for the specified object and returns the index
 * of the first occurrence.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to search for
 * @return		the index of the first occurrence of the object
 */
public int indexOf(Object object) {
	ListIterator it = listIterator();
	if (object != null) {
		while (it.hasNext())
			if (object.equals(it.next())) return it.previousIndex();
	} else {
		while (it.hasNext())
			if (it.next() == null) return it.previousIndex();
	}
	return -1;
}
/**
 * Answers an Iterator on the elements of this List. The elements are
 * iterated in the same order that they occur in the List.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an Iterator on the elements of this List
 *
 * @see			Iterator
 */
public Iterator iterator() {
	return new SimpleListIterator();
}
/**
 * Searches this List for the specified object and returns the index
 * of the last occurrence.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to search for
 * @return		the index of the last occurrence of the object
 */
public int lastIndexOf(Object object) {
	ListIterator it = listIterator(size());
	if (object != null) {
		while (it.hasPrevious())
			if (object.equals(it.previous())) return it.nextIndex();
	} else {
		while (it.hasPrevious())
			if (it.previous() == null) return it.nextIndex();
	}
	return -1;
}
/**
 * Answers a ListIterator on the elements of this List. The elements are
 * iterated in the same order that they occur in the List.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a ListIterator on the elements of this List
 *
 * @see			ListIterator
 */
public ListIterator listIterator() {
	return listIterator(0);
}
/**
 * Answers a ListIterator on the elements of this List. The elements are
 * iterated in the same order that they occur in the List. The iteration
 * starts at the specified location.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location the index at which to start the iteration
 * @return		a ListIterator on the elements of this List
 *
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 *
 * @see			ListIterator
 */
public ListIterator listIterator(int location) {
	return new FullListIterator(location);
}
/**
 * Removes the object at the specified location from this
 * List.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location	the index of the object to remove
 * @return		the removed object
 *
 * @exception	UnsupportedOperationException when removing from this List
 *				is not supported
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 */
public Object remove(int location) {
	throw new UnsupportedOperationException();
}
/**
 * Removes the objects in the specified range from the start to the, but
 * not including, end index.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		start	the index at which to start removing
 * @param		end		the index one past the end of the range to remove
 *
 * @exception	UnsupportedOperationException when removing from this List
 *				is not supported
 * @exception	IndexOutOfBoundsException when <code>start < 0
 */
protected void removeRange(int start, int end) {
	Iterator it = listIterator(start);
	for (int i=start; i < end; i++) {
		it.next();
		it.remove();
	}
}
/**
 * Replaces the element at the specified location in this List with
 * the specified object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		location	the index at which to put the specified object
 * @param		object	the object to add
 * @return		the previous element at the index
 *
 * @exception	UnsupportedOperationException when replacing elements in this List
 *				is not supported
 * @exception	ClassCastException when the class of an object is inappropriate
 *				for this List
 * @exception	IllegalArgumentException when an object cannot be added
 *				to this List
 * @exception	IndexOutOfBoundsException when <code>location < 0 || >= size()</code>
 */
public Object set(int location, Object object) {
	throw new UnsupportedOperationException();
}
/**
 * Answers a List of the specified portion of this List from the start
 * index to one less than the end index. The returned List is backed by
 * this list so changes to one are reflected by the other.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		start	the index at which to start the sublist
 * @param		end		the index one past the end of the sublist
 * @return		a List of a portion of this List
 *
 * @exception	IndexOutOfBoundsException when <code>start < 0 or <code>end > size()</code>
 * @exception	IllegalArgumentException when <code>start > end</code>
 */
public List subList(int start, int end) {
	if (0 <= start && end <= size()) {
		if (start <= end){
			if (this instanceof RandomAccess)
				return new SubAbstractListRandomAccess(this, start, end);
			else
				return new SubAbstractList(this, start, end);
		} else throw new IllegalArgumentException();
	} else throw new IndexOutOfBoundsException();
}
}
