package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2004, 2006  All Rights Reserved
 */

import java.io.Serializable;
import com.ibm.oti.util.ExtendedResourceBundle;

/**
 * This class represents a currency as identified in the ISO 4217 currency codes.
 */
public final class Currency implements Serializable {

	private static final long serialVersionUID = -158308464356906721L;

	private static Hashtable codesToCurrencies = new Hashtable();

	private String currencyCode;

	private static String currencyVars = "EURO, HK, PREEURO"; //$NON-NLS-1$

	private transient int defaultFractionDigits;

	/**
	 * @param currencyCode
	 */
	private Currency(String currencyCode) {
		this.currencyCode = currencyCode;
	}

	/***
	 * Answers the currency instance for this curency code.
	 * <p>
	 * @author		IBM
	 * @version		initial
	 *
	 * @param  		currencyCode java.lang.String
	 * @return 		currency java.util.Currency
	 *
	 * @throws 		java.lang.IllegalArgumentException
	 * 		 		if the currency code is not a supported ISO 4217 currency code
	 */
	public static Currency getInstance(String currencyCode) {
		Currency currency = (Currency) codesToCurrencies.get(currencyCode);

		if (currency == null) {
			ResourceBundle bundle = Locale.getBundle("ISO4CurrenciesToDigits", Locale.getDefault()); //$NON-NLS-1$
			currency = new Currency(currencyCode);

			String defaultFractionDigits = null;
			try {
				defaultFractionDigits = bundle.getString(currencyCode);
			} catch (MissingResourceException e) {
				// K0322 = Not a supported ISO 4217 Currency Code: {0}
				throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0322", currencyCode)); //$NON-NLS-1$
			}
			currency.defaultFractionDigits = Integer.parseInt(defaultFractionDigits);
			codesToCurrencies.put(currencyCode, currency);
		}

		return currency;
	}

	/***
	 * Answers the currency instance for this locale.
	 *
	 * @author		IBM
 	 * @version		initial
	 *
	 * @param  		locale	 java.util.Locale
	 * @return 		currency java.util.Currency
	 *
	 * @throws 		java.lang.IllegalArgumentException
	 * 		 		if the locale's country is not a supported ISO 3166 Country
	 */
	public static Currency getInstance(Locale locale) {
		String country = locale.getCountry();
		String variant = locale.getVariant();
		if (!variant.equals("") && currencyVars.indexOf(variant) > -1) //$NON-NLS-1$
			country = country + "_" + variant; //$NON-NLS-1$

		ResourceBundle bundle = Locale.getBundle("ISO4Currencies", Locale.getDefault()); //$NON-NLS-1$
		String currencyCode = null;
		try {
			currencyCode = bundle.getString(country);
		} catch (MissingResourceException e) {
			// K0323 = Not a supported ISO 3166 Country locale: {0}
			throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString("K0323", locale.toString())); //$NON-NLS-1$
		}

		if (currencyCode.equals("None")) //$NON-NLS-1$
			return null;

		return getInstance(currencyCode);
	}

	public String getCurrencyCode() {
		return currencyCode;
	}

	public String getSymbol() {
		return getSymbol(Locale.getDefault());
	}

	/***
	 * Return the symbol for this currency in the given locale.
	 * <p>
	 *
	 * If the locale doesn't have any countries
	 * (e.g. <code>Locale.JAPANESE, new Locale("en","")</code>),
	 * currencyCode is returned.
	 * <p>
	 * First the locale bundle is checked, if the locale has the same currency,
	 * the CurrencySymbol in this locale bundle is returned.
	 * <p>
	 * Then a currency bundle for this locale is searched.
	 * <p>
	 * If a currency bundle for this locale does not exist, or
	 * there is no symbol for this currency in this bundle,
	 * than <code>currencyCode</code> is returned.
	 * <p>
	 *
	 * @author		IBM
 	 * @version		initial
 	 *
	 * @param  		locale java.lang.String 		locale
	 * @return 		symbol java.lang.String 		the representation of this Currency's symbol in this locale
	 */
	public String getSymbol(Locale locale) {
		if (locale.getCountry().equals("")) //$NON-NLS-1$
			return currencyCode;

		// check in the Locale bundle first, if the local has the same currency
		ExtendedResourceBundle localeBundle = (ExtendedResourceBundle) Locale.getBundle("Locale", locale); //$NON-NLS-1$
		if (((String)localeBundle.getObject(com.ibm.oti.locale.Locale.INTL_CURRENCY_SYMBOL)).equals(currencyCode))
			return (String)localeBundle.getObject(com.ibm.oti.locale.Locale.CURRENCY_SYMBOL);

		//search for a Currency bundle
		ResourceBundle bundle = null;
		try {
			bundle = Locale.getBundle("Currency", locale); //$NON-NLS-1$
		} catch (MissingResourceException e) {
			return currencyCode;
		}

		//is the bundle found for a different country?	(for instance the default locale's currency bundle)
		if (!bundle.getLocale().getCountry().equals(locale.getCountry()))
			return currencyCode;

		// check if the currency bundle for this locale
		// has an entry for this currency
		String result = (String) bundle.handleGetObject(currencyCode);
		if (result != null)
			return result;
		else
			return currencyCode;
	}

	public int getDefaultFractionDigits() {
		return defaultFractionDigits;
	}

	public String toString() {
		return currencyCode;
	}

	private Object readResolve() {
		return getInstance(currencyCode);
	}
}
