package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * Date represents a specific moment in time, to the millisecond.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			System#currentTimeMillis
 * @see			Calendar
 * @see			GregorianCalendar
 * @see			SimpleTimeZone
 * @see			TimeZone
 */

public class Date implements Serializable, Cloneable, Comparable {
	private static final long serialVersionUID = 7523967970034938905L;

	transient private long milliseconds;

/**
 * Initializes this Date instance to the current date and time.
 *
 * @author		OTI
 * @version		initial
 */
public Date () {
	this (System.currentTimeMillis ());
}

/**
 * Initializes this Date instance using the specified millisecond value. The
 * value is the number of milliseconds since Jan. 1, 1970 GMT.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		milliseconds	the number of milliseconds since Jan. 1, 1970 GMT
 */
public Date (long milliseconds) {
	this.setTime (milliseconds);
}

/**
 * Answers if this Date is after the specified Date.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		date	a Date instance to compare
 * @return		true if this Date is after the specified Date, false otherwise
 */
public boolean after (Date date) {
	return milliseconds > date.milliseconds;
}

/**
 * Answers if this Date is before the specified Date.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		date	a Date instance to compare
 * @return		true if this Date is before the specified Date, false otherwise
 */
public boolean before (Date date) {
	return milliseconds < date.milliseconds;
}

/**
 * Answers a new Date with the same millisecond value as this Date.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a shallow copy of this Date
 *
 * @see			java.lang.Cloneable
 */
public Object clone() {
	try {
		return super.clone();
	} catch (CloneNotSupportedException e) {return null;}
}

/**
 * Compare the receiver to the specified Object to determine
 * the relative ordering.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	an Object
 * @return		an int < 0 if this Date is less than the specified Date,
 *				0 if they are equal, and > 0 if this Date is greater
 *
 * @exception	ClassCastException when object is not a Date
 */
public int compareTo(Object object) {
	return compareTo((Date)object);
}

/**
 * Compare the receiver to the specified Date to determine
 * the relative ordering.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		date	a Date
 * @return		an int < 0 if this Date is less than the specified Date,
 *				0 if they are equal, and > 0 if this Date is greater
 */
public int compareTo(Date date) {
	if (milliseconds < date.milliseconds) return -1;
	if (milliseconds == date.milliseconds) return 0;
	return 1;
}

/**
 * Compares the specified object to this Date and answer if they
 * are equal. The object must be an instance of Date and have the
 * same millisecond value.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to compare with this object
 * @return		true if the specified object is equal to this Date, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals (Object object) {
	return (object == this) || (object instanceof Date) &&
		(milliseconds == ((Date) object).milliseconds);
}

/**
 * Answers this Date as a millisecond value. The value is the number of
 * milliseconds since Jan. 1, 1970 GMT.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the number of milliseconds since Jan. 1, 1970 GMT.
 */
public long getTime () {
	return milliseconds;
}

/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	return (int)(milliseconds >>> 32) ^ (int) milliseconds;
}

/**
 * Sets this Date to the specified millisecond value. The value is the number of
 * milliseconds since Jan. 1, 1970 GMT.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		milliseconds	the number of milliseconds since Jan. 1, 1970 GMT.
 */
public void setTime (long milliseconds) {
	this.milliseconds = milliseconds;
}

/**
 * Answers the string representation of this Date in the
 * format: Tue Jun 22 13:07:00 GMT 1999
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the string representation of this Date
 */
public String toString () {
	return new SimpleDateFormat("E MMM dd HH:mm:ss z yyyy", Locale.US).format(this);
}

private void writeObject(ObjectOutputStream stream) throws IOException {
	stream.defaultWriteObject();
	stream.writeLong(getTime());
}
private void readObject(ObjectInputStream stream)
throws IOException, ClassNotFoundException {
	stream.defaultReadObject();
	setTime(stream.readLong());
}
}
