package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * HashSet is an implementation of Set. All optional operations are supported,
 * adding and removing. The elements can be any objects.
 *
 * @author		OTI
 * @version		initial
 */
public class HashSet extends AbstractSet implements Set, Cloneable, Serializable {
	private static final long serialVersionUID = -5024744406713321676L;
	transient HashMap backingMap;

/**
 * Contructs a new empty instance of HashSet.
 *
 * @author		OTI
 * @version		initial
 */
public HashSet() {
	this(new HashMap());
}
/**
 * Constructs a new instance of HashSet with the specified capacity.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		capacity	the initial capacity of this HashSet
 */
public HashSet(int capacity) {
	this(new HashMap(capacity));
}
/**
 * Constructs a new instance of HashSet with the specified capacity and load factor.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		capacity	the initial capacity
 * @param		loadFactor	the initial load factor
 */
public HashSet(int capacity, float loadFactor) {
	this(new HashMap(capacity, loadFactor));
}
/**
 * Constructs a new instance of HashSet containing the unique elements in the
 * specified collection.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		collection	the collection of elements to add
 */
public HashSet(Collection collection) {
	this(new HashMap(collection.size() < 6 ? 11 : collection.size() * 2));
	Iterator it = collection.iterator();
	while (it.hasNext()) add(it.next());
}

HashSet(HashMap backingMap) {
	this.backingMap = backingMap;
}

/**
 * Adds the specified object to this HashSet.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to add
 * @return		true when this HashSet did not already contain the object, false otherwise
 */
public boolean add(Object object) {
	return backingMap.put(object, this) == null;
}
/**
 * Removes all elements from this HashSet, leaving it empty.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			#isEmpty
 * @see			#size
 */
public void clear() {
	backingMap.clear();
}
/**
 * Answers a new HashSet with the same elements and size as this HashSet.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a shallow copy of this HashSet
 *
 * @see			java.lang.Cloneable
 */
public Object clone() {
	try {
		HashSet clone = (HashSet)super.clone();
		clone.backingMap = (HashMap)backingMap.clone();
		return clone;
	} catch (CloneNotSupportedException e) {return null;}
}
/**
 * Searches this HashSet for the specified object.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to search for
 * @return		true if <code>object</code> is an element of this HashSet, false otherwise
 */
public boolean contains(Object object) {
	return backingMap.containsKey(object);
}
/**
 * Answers if this HashSet has no elements, a size of zero.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		true if this HashSet has no elements, false otherwise
 *
 * @see			#size
 */
public boolean isEmpty() {
	return backingMap.isEmpty();
}
/**
 * Answers an Iterator on the elements of this HashSet.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an Iterator on the elements of this HashSet
 *
 * @see			Iterator
 */
public Iterator iterator() {
	return backingMap.keySet().iterator();
}
/**
 * Removes an occurrence of the specified object from this
 * HashSet.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		object	the object to remove
 * @return		true if this HashSet is modified, false otherwise
 */
public boolean remove(Object object) {
	return backingMap.remove(object) != null;
}
/**
 * Answers the number of elements in this HashSet.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the number of elements in this HashSet
 */
public int size() {
	return backingMap.size();
}
private void writeObject(ObjectOutputStream stream) throws IOException {
	stream.defaultWriteObject();
	stream.writeInt(backingMap.elementData.length);
	stream.writeFloat(backingMap.loadFactor);
	stream.writeInt(backingMap.elementCount);
	for (int i=backingMap.elementData.length; --i >= 0;) {
		HashMap.Entry entry = backingMap.elementData[i];
		while (entry != null) {
			stream.writeObject(entry.key);
			entry = entry.next;
		}
	}
}
private void readObject(ObjectInputStream stream)
throws IOException, ClassNotFoundException {
	stream.defaultReadObject();
	int length = stream.readInt();
	float loadFactor = stream.readFloat();
	backingMap = createBackingMap(length, loadFactor);
	int elementCount = stream.readInt();
	for (int i=elementCount; --i >= 0;) {
		Object key = stream.readObject();
		backingMap.put(key, this);
	}
}

HashMap createBackingMap(int capacity, float loadFactor) {
	return new HashMap(capacity, loadFactor);
}
}
