package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * Observable is used to notify a group of Observer objects when a change
 * occurs.
 *
 * @author		OTI
 * @version		initial
 */

public class Observable {
	final Vector observers = new Vector();
	boolean changed = false;
/**
 * Constructs a new Observable object.
 *
 * @author		OTI
 * @version		initial
 */
public Observable() {
}
/**
 * Adds the specified Observer to the list of observers.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	observer the Observer to add
 */
public synchronized void addObserver(Observer observer) {
	if (observer == null) {
		throw new NullPointerException();
	}
	if (!observers.contains(observer))
		observers.addElement(observer);
}
/**
 * Clears the changed flag for this Observable.
 *
 * @author		OTI
 * @version		initial
 */
protected synchronized void clearChanged() {
	changed = false;
}
/**
 * Answers the number of Observers in the list of observers.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	the number of observers
 */
public synchronized int countObservers() {
	return observers.size();
}
/**
 * Removes the specified Observer from the list of observers.
 *
 * @author		OTI
 * @version		initial
 *
 * @param	observer the Observer to remove
 */
public synchronized void deleteObserver(Observer observer) {
	observers.removeElement(observer);
}
/**
 * Removes all Observers from the list of observers.
 *
 * @author		OTI
 * @version		initial
 */
public synchronized void deleteObservers() {
	observers.setSize(0);
}
/**
 * Answers the changed flag for this Observable.
 *
 * @author		OTI
 * @version		initial
 *
 * @return	true when the changed flag for this Observable is set, false otherwise
 */
public synchronized boolean hasChanged() {
	return changed;
}
/**
 * If the changed flag is set, calls the <code>update()</code> method for
 * every Observer in the list of observers using null as the argument. Clears
 * the changed flag.
 *
 * @author		OTI
 * @version		initial
 */
public void notifyObservers() {
	notifyObservers(null);
}
/**
 * If the changed flag is set, calls the <code>update()</code> method for
 * every Observer in the list of observers using the specified argument.
 * Clears the changed flag.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		data	the argument passed to update()
 */
public void notifyObservers(Object data) {
	if (changed) {
		// Must clone the vector in case deleteObserver is called
		Vector clone = (Vector)observers.clone();
		int size = clone.size();
		for (int i = 0; i < size; i++)
			((Observer)clone.elementAt(i)).update(this, data);
		clearChanged();
	}
}
/**
 * Sets the changed flag for this Observable.
 *
 * @author		OTI
 * @version		initial
 */
protected synchronized void setChanged() {
	changed = true;
}
}
