package java.util;

import java.security.Permission;
import java.security.BasicPermission;
import java.security.PermissionCollection;
import java.io.IOException;
import java.io.ObjectStreamField;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * PropertyPermission objects represent permission to
 * access system properties.
 *
 * @author		OTI
 * @version		initial
 */
public final class PropertyPermission extends BasicPermission {
	private static final long serialVersionUID = 885438825399942851L;
	transient private boolean read, write;

/**
 * Constructs a new instance of this class.
 *
 * @author		OTI
 * @version		initial
 * @param		name java.lang.String
 *					the (possibly wildcarded) name of the property.
 * @param		actions java.lang.String
 *					the actions which are applicable to it.
 */
public PropertyPermission(String name, String actions) {
	super(name);
	decodeActions(actions);
}

private void decodeActions(String actions) {
	StringTokenizer tokenizer =
		new StringTokenizer(actions.toLowerCase(), " \t\n\r,");
	while (tokenizer.hasMoreTokens()) {
		String token = tokenizer.nextToken();
		if (token.equals("read"))
			read = true;
		else if (token.equals("write"))
			write = true;
		else
			throw new IllegalArgumentException();
	}
	if (!read && !write) throw new IllegalArgumentException();
}

/**
 * Compares the argument to the receiver, and answers true
 * if they represent the <em>same</em> object using a class
 * specific comparison. In this case, the receiver must be
 * for the same property as the argument, and must have the
 * same actions.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		o		the object to compare with this object
 * @return		<code>true</code>
 *					if the object is the same as this object
 *				<code>false</code>
 *					if it is different from this object
 * @see			#hashCode
 */
public boolean equals(Object o) {
	if (super.equals(o)) {
		PropertyPermission pp = (PropertyPermission) o;
		return read == pp.read && write == pp.write;
	}
	return false;
}

/**
 * Answers the actions associated with the receiver.
 * The result will be either "read", "write", or
 * "read,write".
 *
 * @author		OTI
 * @version		initial
 *
 * @return		String
 *					the actions associated with the receiver.
 */
public String getActions() {
	return read ? (write ? "read,write" : "read") : "write";
}

/**
 * Answers an integer hash code for the receiver. Any two
 * objects which answer <code>true</code> when passed to
 * <code>equals</code> must answer the same value for this
 * method.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	return super.hashCode();
}

/**
 * Indicates whether the argument permission is implied
 * by the receiver.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		boolean
 *					<code>true</code> if the argument permission
 *					is implied by the receiver,
 *					and <code>false</code> if it is not.
 * @param		permission java.security.Permission
 *					the permission to check
 */
public boolean implies(Permission permission) {
	if (super.implies(permission)) {
		PropertyPermission pp = (PropertyPermission) permission;
		return (read || !pp.read) && (write || !pp.write);
	}
	return false;
}

/**
 * Answers a new PermissionCollection for holding permissions
 * of this class. Answer null if any permission collection can
 * be used.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a new PermissionCollection or null
 *
 * @see			java.security.BasicPermissionCollection
 */
public PermissionCollection newPermissionCollection() {
	return new PropertyPermissionCollection();
}

private static final ObjectStreamField[] serialPersistentFields = {
	new ObjectStreamField("actions", String.class)
};
private void writeObject(ObjectOutputStream stream) throws IOException {
	ObjectOutputStream.PutField fields = stream.putFields();
	fields.put("actions", getActions());
	stream.writeFields();
}
private void readObject(ObjectInputStream stream)
throws IOException, ClassNotFoundException {
	ObjectInputStream.GetField fields = stream.readFields();
	String actions = (String)fields.get("actions", "");
	decodeActions(actions);
}
}
