
package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

/**
 * String tokenizer is used to break a string apart into tokens.
 *
 * If returnDelimiters is false, successive calls to nextToken() return
 * maximal blocks of characters that do not contain a delimiter.
 *
 * If returnDelimiters is true, delimiters are considered to be tokens, and
 * successive calls to nextToken() return either a one character delimiter,
 * or a maximal block of text between delimiters.
 *
 * @author		OTI
 * @version		initial
 *
 */

public class StringTokenizer implements Enumeration {
	private final String string;
	private String delimiters;
	private final boolean returnDelimiters;
	private int position;

/**
 * Constructs a new StringTokenizer for string using whitespace as the
 * delimiter, returnDelimiters is false.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string		the string to be tokenized
 */
public StringTokenizer(String string) {
	this(string, " \t\n\r\f", false);
}

/**
 * Constructs a new StringTokenizer for string using the specified
 * delimiters, returnDelimiters is false.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string		the string to be tokenized
 * @param		delimiters	the delimiters to use
 */
public StringTokenizer(String string, String delimiters) {
	this(string, delimiters, false);
}
/**
 * Constructs a new StringTokenizer for string using the specified delimiters and
 * returning delimiters as tokens when specified.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		string		the string to be tokenized
 * @param		delimiters	the delimiters to use
 * @param		returnDelimiters		true to return each delimiter as a token
 */
public StringTokenizer(String string, String delimiters, boolean returnDelimiters) {
	if (string != null) {
		this.string = string;
		this.delimiters = delimiters;
		this.returnDelimiters = returnDelimiters;
		this.position = 0;
	} else throw new NullPointerException();
}
/**
 * Returns the number of unprocessed tokens remaining in the string.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		number of tokens that can be retreived before an exception will result
 *
 */
public int countTokens() {
	int count = 0;
	boolean inToken = false;
	for (int i = position, length = string.length(); i < length; i++) {
		if (delimiters.indexOf(string.charAt(i), 0) >= 0) {
			if (returnDelimiters) count++;
			if (inToken) {
				count++;
				inToken = false;
			}
		} else {
			inToken = true;
		}
	}
	if (inToken) count++;
	return count;
}

/**
 * Returns true if unprocessed tokens remain.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		true if unprocessed tokens remain
 */
public boolean hasMoreElements() {
	return hasMoreTokens();
}

/**
 * Returns true if unprocessed tokens remain.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		true if unprocessed tokens remain
 */
public boolean hasMoreTokens() {
	int length = string.length();
	if (position < length) {
		if (returnDelimiters) return true; // there is at least one character and even if it is a delimiter it is a token

	 	// otherwise find a character which is not a delimiter
  		for (int i = position; i < length; i++)
 			if (delimiters.indexOf(string.charAt(i), 0) == -1) return true;
	}
	return false;
}

/**
 * Returns the next token in the string as an Object.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		next token in the string as an Object
 * @exception	NoSuchElementException
 *					if no tokens remain
 */
public Object nextElement() {
	return nextToken();
}

/**
 * Returns the next token in the string as a String.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		next token in the string as a String
 * @exception	NoSuchElementException
 *					if no tokens remain
 */
public String nextToken() {
	int i = position;
	int length = string.length();

	if (i < length) {
		if (returnDelimiters) {
			if (delimiters.indexOf(string.charAt(position), 0) >= 0)
				return String.valueOf(string.charAt(position++));
			for (position++; position < length; position++)
				if (delimiters.indexOf(string.charAt(position), 0) >= 0)
					return string.substring(i, position);
			return string.substring(i);
		}

		while (i < length && delimiters.indexOf(string.charAt(i), 0) >= 0) i++;
		position = i;
		if (i < length) {
			for (position++; position < length; position++)
				if (delimiters.indexOf(string.charAt(position), 0) >= 0)
					return string.substring(i, position);
			return string.substring(i);
		}
	}
	throw new NoSuchElementException();
}

/**
 * Returns the next token in the string as a String. The delimiters used
 * are changed to the specified delimiters.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		delimiters	the new delimiters to use
 * @return		next token in the string as a String
 * @exception	NoSuchElementException
 *					if no tokens remain
 */
public String nextToken(String delimiters) {
	this.delimiters = delimiters;
	return nextToken();
}
}
