package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2006  All Rights Reserved
 */

class TimeZoneTable extends TimeZone {
	private int rawOffset;
	private SimpleTimeZone[] timezones;
	private int[][] dateOffsets;
	private long[] longOffsets;

/**
 * Creates a TimeZoneTable with an ordered array of SimpleTimeZones.
 * The TimeZones must all have the same ID. The dateOffsets and longOffsets
 * are arrays of int[] and long[] which specify the localized date and
 * gmt date of the time to change between SimpleTimeZones.
 *
 * @param timezones a non-null array of SimpleTimeZone, ordered by start year
 * @param dateOffsets an int[] containing the localized date to change to the next SimpleTimeZone
 * @param longOffsets a long[] containing the gmt date to change to the next SimpleTimezone
 */
public TimeZoneTable(SimpleTimeZone[] timezones, int[][] dateOffsets, long[] longOffsets) {
	super();
	this.timezones = timezones;
	this.dateOffsets = dateOffsets;
	this.longOffsets = longOffsets;
	setID(timezones[0].getID());
	rawOffset = timezones[timezones.length-1].getRawOffset();
}

/**
 * Gets the offset from GMT of this TimeZone for the specified date and time. The
 * offset includes daylight savings time if the specified date and time are within
 * the daylight savings time period.
 *
 * @param		era	the GregorianCalendar era, either GregorianCalendar.BC or
 *					GregorianCalendar.AD
 * @param		year	the year
 * @param		month	the Calendar month
 * @param		day	the day of the month
 * @param		dayOfWeek	the Calendar day of the week
 * @param		time	the time of day in milliseconds
 * @return		the offset from GMT in milliseconds
 */
public int getOffset(int era, int year, int month, int day, int dayOfWeek, int time) {
	if (era == GregorianCalendar.AD) {
		for (int i=dateOffsets.length-1; i>=0; i--) {
			if (year < dateOffsets[i][0]) continue;
			if (year == dateOffsets[i][0]) {
				if (month < dateOffsets[i][1]) continue;
				if (month == dateOffsets[i][1]) {
					if (day < dateOffsets[i][2]) continue;
					if (day == dateOffsets[i][2]) {
						if (time < dateOffsets[i][3]) continue;
					}
				}
			}
			return timezones[i+1].getOffset(era, year, month, day, dayOfWeek, time);
		}
	}
	return timezones[0].getOffset(era, year, month, day, dayOfWeek, time);
}

/**
 * Compares the specified object to this TimeZone and answer if they
 * are equal. The object must be an instance of TimeZoneTable and have the
 * same properties.
 *
 * @param		object	the object to compare with this object
 * @return		true if the specified object is equal to this TimeZone, false otherwise
 *
 * @see			#hashCode
 */
public boolean equals(Object object) {
	if (!(object instanceof TimeZoneTable)) return false;
	TimeZoneTable table = (TimeZoneTable)object;
	if (getRawOffset() != table.getRawOffset() || !(getID().equals(table.getID())) ||
			timezones.length != table.timezones.length)
		return false;
	for (int i=0; i<timezones.length; i++) {
		if (!timezones[i].equals(table.timezones[i])) return false;
	}
	for (int i=0; i<dateOffsets.length; i++) {
		for (int j=0; j<dateOffsets[0].length; j++) {
			if (dateOffsets[i][j] != table.dateOffsets[i][j]) return false;
		}
		if (longOffsets[i] != table.longOffsets[i]) return false;
	}
	return true;
}

/**
 * Answers an integer hash code for the receiver. Objects which are
 * equal answer the same value for this method.
 *
 * @return		the receiver's hash
 *
 * @see			#equals
 */
public int hashCode() {
	int hashcode = 0;
	for (int i=0; i<timezones.length; i++) {
		hashcode += timezones[i].hashCode();
	}
	return hashcode;
}

/**
 * Answers if the specified TimeZone has the same raw offset and
 * daylight savings time rules as this TimeZone.
 *
 * @param		zone	a TimeZone
 * @return		true when the TimeZones have the same raw offset and
 *				daylight savings time rules, false otherwise
 */
public boolean hasSameRules(TimeZone zone) {
	if (!(zone instanceof TimeZoneTable)) return false;
	TimeZoneTable table = (TimeZoneTable)zone;
	if (getRawOffset() != table.getRawOffset() || timezones.length != table.timezones.length)
		return false;
	for (int i=0; i<timezones.length; i++) {
		if (!timezones[i].hasSameRules(table.timezones[i])) return false;
	}
	for (int i=0; i<dateOffsets.length; i++) {
		for (int j=0; j<dateOffsets[0].length; j++) {
			if (dateOffsets[i][j] != table.dateOffsets[i][j]) return false;
		}
		if (longOffsets[i] != table.longOffsets[i]) return false;
	}
	return true;
}

/**
 * Answers if this TimeZone has a daylight savings time period.
 *
 * @return		true if this time zone has a daylight savings time period, false otherwise
 */
public boolean useDaylightTime() {
	return timezones[timezones.length-1].useDaylightTime();
}

/**
 * Gets the offset for standard time from GMT for this TimeZone.
 *
 * @return		the offset from GMT of standard time in milliseconds
 */
public int getRawOffset() {
	return rawOffset;
}

/**
 * Sets the offset for standard time from GMT for this TimeZone.
 *
 * @param		offset	the offset from GMT of standard time in milliseconds
 */
public void setRawOffset (int offset) {
	rawOffset = offset;
	for (int i=0; i<timezones.length; i++) {
		timezones[i].setRawOffset(offset);
	}
}

/**
 * Answers if the specified Date is in the daylight savings time period
 * for this TimeZone.
 *
 * @param		time	a Date
 * @return		true when the Date is in the daylight savings time period, false otherwise
 */
public boolean inDaylightTime(Date time) {
	// check for null pointer
	long millis = time.getTime();
	for (int i=longOffsets.length-1; i>=0; i--) {
		if (millis < longOffsets[i]) continue;
		return timezones[i+1].inDaylightTime(time);
	}
	return timezones[0].inDaylightTime(time);
}

/**
 * Gets the offset from GMT of this TimeZone for the specified date. The
 * offset includes daylight savings time if the specified date is within
 * the daylight savings time period.
 *
 * @param		time 	the date in milliseconds since January 1, 1970 00:00:00 GMT
 * @return		the offset from GMT in milliseconds
 */
public int getOffset(long time) {
	for (int i=longOffsets.length-1; i>=0; i--) {
		if (time < longOffsets[i]) continue;
		return timezones[i+1].getOffset(time);
	}
	return timezones[0].getOffset(time);
}

/**
 * Answers the string representation of this TimeZoneTable.
 *
 * @return		the string representation of this TimeZoneTable
 */
public String toString() {
	StringBuffer buffer = new StringBuffer(256);
	buffer.append(getClass().getName());
	buffer.append('[');
	for (int i=0; i<timezones.length; i++) {
		if (i != 0) {
			buffer.append("\n");
			buffer.append(longOffsets[i-1]);
			buffer.append(" : ");
		}
		buffer.append(timezones[i]);
	}
	buffer.append(']');
	return buffer.toString();
}

/**
 * Gets the daylight savings offset in milliseconds for this TimeZone.
 *
 * @return		the daylight savings offset in milliseconds
 */
public int getDSTSavings() {
	for (int i=longOffsets.length-1; i>=0; i--) {
		if (timezones[i].useDaylightTime())
			return timezones[i].getDSTSavings();
	}
	return 0;
}

/**
 * Answers a new SimpleTimeZone with the same ID, rawOffset and daylight savings
 * time rules as this SimpleTimeZone.
 *
 * @return		a shallow copy of this SimpleTimeZone
 *
 * @see			java.lang.Cloneable
 */
public Object clone() {
	TimeZoneTable zone = (TimeZoneTable) super.clone();
	zone.timezones = new SimpleTimeZone[timezones.length];
	for (int i=0; i<timezones.length; i++) {
		zone.timezones[i] = (SimpleTimeZone)timezones[i].clone();
	}
	return zone;
}

}
