package java.util;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.lang.ref.WeakReference;
import java.lang.ref.ReferenceQueue;

/**
 * WeakHashMap is an implementation of Map with keys which
 * are WeakReferences. The key/value mapping is removed when
 * the key is no longer referenced. All optional operations
 * are supported, adding and removing. Keys and values can
 * be any objects.
 *
 * @author		OTI
 * @version		initial
 */

public class WeakHashMap extends AbstractMap implements Map {
	private final ReferenceQueue referenceQueue;
	int elementCount;
	Entry[] elementData;
	private final int loadFactor;
	private int threshold;

	transient int modCount = 0;

	private static final int DEFAULT_SIZE = 16;

	private static final class Entry extends WeakReference implements Map.Entry {
		int hash;
		boolean isNull;
		Object value;
		Entry next;
		interface Type {
			Object get(Map.Entry entry);
		}
		Entry(Object key, Object object, ReferenceQueue queue) {
			super(key, queue);
			isNull = key == null;
			hash = isNull ? 0 : key.hashCode();
			value = object;
		}

		public Object getKey() {return super.get();}
		public Object getValue() {return value;}
		public Object setValue(Object object) {
			Object result = value;
			value = object;
			return result;
		}
		public boolean equals(Object other) {
			if (!(other instanceof Map.Entry))
				return false;
			Map.Entry entry = (Map.Entry) other;
			Object key = super.get();
			return (key == null ? key == entry.getKey() : key.equals(entry.getKey())) &&
				(value == null ? value == entry.getValue() : value.equals(entry.getValue()));
		}
		public int hashCode() {
			return hash + (value == null ? 0 : value.hashCode());
		}
		public String toString() {
			return super.get() + "=" + value;
		}
	}

class HashIterator implements Iterator {
	private int position = 0, expectedModCount;
	private Entry currentEntry, nextEntry;
	private Object nextKey;
	final Entry.Type type;

HashIterator(Entry.Type type) {
	this.type = type;
	expectedModCount = modCount;
}

public boolean hasNext() {
	if (nextEntry != null) return true;
	while (true) {
		if (nextEntry == null) {
			while (position < elementData.length)
				if ((nextEntry = elementData[position++]) != null) break;
			if (nextEntry == null) return false;
		}
		// ensure key of next entry is not gc'ed
		nextKey = nextEntry.get();
		if (nextKey != null || nextEntry.isNull) {
			return true;
		}
		nextEntry = nextEntry.next;
	}
}
public Object next() {
	if (expectedModCount == modCount) {
		if (hasNext()) {
			currentEntry = nextEntry;
			nextEntry = currentEntry.next;
			Object result = type.get(currentEntry);
			// free the key
			nextKey = null;
			return result;
		} else throw new NoSuchElementException();
	} else throw new ConcurrentModificationException();
}
public void remove() {
	if (expectedModCount == modCount) {
		if (currentEntry != null) {
			removeEntry(currentEntry);
			currentEntry = null;
			expectedModCount++;
			// cannot poll() as that would change the expectedModCount
		} else throw new IllegalStateException();
	} else throw new ConcurrentModificationException();
}
}

/**
 * Contructs a new empty instance of WeakHashMap.
 *
 * @author		OTI
 * @version		initial
 */
public WeakHashMap() {
	this(DEFAULT_SIZE);
}

/**
 * Constructs a new instance of WeakHashMap with the specified capacity.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		capacity	the initial capacity of this WeakHashMap
 *
 * @exception	IllegalArgumentException when the capacity is less than zero
 */
public WeakHashMap(int capacity) {
	if (capacity >= 0) {
		elementCount = 0;
		elementData = new Entry[capacity == 0 ? 1 : capacity];
		loadFactor = 7500;	// Default load factor of 0.75
		computeMaxSize();
		referenceQueue = new ReferenceQueue();
	} else throw new IllegalArgumentException();
}

/**
 * Constructs a new instance of WeakHashMap with the specified capacity and
 * load factor.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		capacity	the initial capacity
 * @param		loadFactor	the initial load factor
 *
 * @exception	IllegalArgumentException when the capacity is less than zero
 *				or the load factor is less or equal to zero
 */
public WeakHashMap(int capacity, float loadFactor) {
	if (capacity >= 0 && loadFactor > 0) {
		elementCount = 0;
		elementData = new Entry[capacity == 0 ? 1 : capacity];
		this.loadFactor = (int)(loadFactor * 10000);
		computeMaxSize();
		referenceQueue = new ReferenceQueue();
	} else throw new IllegalArgumentException();
}

/**
 * Constructs a new instance of HashMap containing the mappings from the
 * specified Map.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		map	the mappings to add
 */
public WeakHashMap(Map map) {
	this(map.size() < 6 ? 11 : map.size() * 2);
	putAll(map);
}

/**
 * Removes all mappings from this WeakHashMap, leaving it empty.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			#isEmpty
 * @see			#size
 */
public void clear() {
	if (elementCount > 0) {
		elementCount = 0;
		Arrays.fill(elementData, null);
		modCount++;
		while (referenceQueue.poll() != null);
	}
}

private void computeMaxSize() {
	threshold = (int)((long)elementData.length * loadFactor / 10000);
}

/**
 * Searches this WeakHashMap for the specified key.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key	the object to search for
 * @return		true if <code>key</code> is a key of this WeakHashMap, false otherwise
 */
public boolean containsKey(Object key) {
	return getEntry(key) != null;
}

/**
 * Answers a Set of the mappings contained in this WeakHashMap. Each
 * element in the set is a Map.Entry. The set is backed by this
 * WeakHashMap so changes to one are relected by the other. The set does
 * not support adding.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		a Set of the mappings
 */
public Set entrySet() {
	poll();
	return new AbstractSet() {
		public int size() {return WeakHashMap.this.size();}
		public void clear() {WeakHashMap.this.clear();}
		public boolean remove(Object object) {
			if (contains(object)) {
				WeakHashMap.this.remove(((Map.Entry)object).getKey());
				return true;
			}
			return false;
		}
		public boolean contains(Object object) {
			if (object instanceof Map.Entry) {
				Entry entry = getEntry(((Map.Entry)object).getKey());
				if (entry != null) {
					Object key = entry.get();
					if (key != null || entry.isNull) {
						return object.equals(entry);
					}
				}
			}
			return false;
		}
		public Iterator iterator() {
			return new HashIterator(new Entry.Type() {
				public Object get(Map.Entry entry) {return entry;}});
		}
	};
}

/**
 * Answers a Set of the keys contained in this WeakHashMap. The set is backed
 * by this WeakHashMap so changes to one are reflected by the other. The set does
 * not support adding.
 *
 * @author		IBM
 * @version		initial
 *
 * @return		a Set of the keys
 */
public Set keySet() {
	poll();
	if (keySet == null) {
		keySet = new AbstractSet() {
			public boolean contains(Object object) {return containsKey(object);}
			public int size() {return WeakHashMap.this.size();}
			public void clear() {WeakHashMap.this.clear();}
			public boolean remove(Object key) {
				if (containsKey(key)) {
					WeakHashMap.this.remove(key);
					return true;
				}
				return false;
			}
			public Iterator iterator() {
				return new HashIterator(new Entry.Type() {
					public Object get(Map.Entry entry) {return entry.getKey();}});
			}
		};
	}
	return keySet;
}
/**
 * Answers a Collection of the values contained in this WeakHashMap. The
 * collection is backed by this WeakHashMap so changes to one are reflected
 * by the other. The collection does not support adding.
 *
 * @author		IBM
 * @version		initial
 *
 * @return		a Collection of the values
 */
public Collection values() {
	poll();
	if (valuesCollection == null) {
		valuesCollection = new AbstractCollection() {
			public int size() {return WeakHashMap.this.size();}
			public void clear() {WeakHashMap.this.clear();}
			public boolean contains(Object object) {return containsValue(object);}
			public Iterator iterator() {
				return new HashIterator(new Entry.Type() {
					public Object get(Map.Entry entry) {return entry.getValue();}});
			}
		};
	}
	return valuesCollection;
}

/**
 * Answers the value of the mapping with the specified key.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key	the key
 * @return		the value of the mapping with the specified key
 */
public Object get(Object key) {
	poll();
	if (key != null) {
		int index = (key.hashCode() & 0x7FFFFFFF) % elementData.length;
		Entry entry = elementData[index];
		while (entry != null) {
			if (key.equals(entry.get())) return entry.value;
			entry = entry.next;
		}
		return null;
	}
	Entry entry = elementData[0];
	while (entry != null) {
		if (entry.isNull) return entry.value;
		entry = entry.next;
	}
	return null;
}

Entry getEntry(Object key) {
	poll();
	if (key != null) {
		int index = (key.hashCode() & 0x7FFFFFFF) % elementData.length;
		Entry entry = elementData[index];
		while (entry != null) {
			if (key.equals(entry.get())) return entry;
			entry = entry.next;
		}
		return null;
	}
	Entry entry = elementData[0];
	while (entry != null) {
		if (entry.isNull) return entry;
		entry = entry.next;
	}
	return null;
}

/**
 * Searches this WeakHashMap for the specified value,
 * and returns true, if at least one entry has this object as its value.
 *
 * @author		IBM
 * @version		initial
 *
 * @param		value	the object to search for
 * @return		true if <code>value</code> is a value in this WeakHashMap, false otherwise
 */
public boolean containsValue(Object value) {
	poll();
	if (value != null) {
		for (int i=elementData.length; --i >= 0;) {
			Entry entry = elementData[i];
			while (entry != null) {
				Object key = entry.get();
				if ((key != null || entry.isNull) && value.equals(entry.value))
					return true;
				entry = entry.next;
			}
		}
	} else {
		for (int i=elementData.length; --i >= 0;) {
			Entry entry = elementData[i];
			while (entry != null) {
				Object key = entry.get();
				if ((key != null || entry.isNull) && entry.value == null)
					return true;
				entry = entry.next;
			}
		}
	}
	return false;
}

/**
 * Answers if this WeakHashMap has no elements, a size of zero.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		true if this HashMap has no elements, false otherwise
 *
 * @see			#size
 */
public boolean isEmpty() {
	return size() == 0;
}

void poll() {
	Entry toRemove;
	while ((toRemove = (Entry)referenceQueue.poll()) != null) {
		removeEntry(toRemove);
	}
}

void removeEntry(Entry toRemove) {
	Entry entry, last = null;
	int index = (toRemove.hash & 0x7FFFFFFF) % elementData.length;
	entry = elementData[index];
	// Ignore queued entries which cannot be found, the user could
	// have removed them before they were queued, i.e. using clear()
	while (entry != null) {
		if (toRemove == entry) {
			modCount++;
			if (last == null) elementData[index] = entry.next;
			else last.next = entry.next;
			elementCount--;
			break;
		}
		last = entry;
		entry = entry.next;
	}
}

/**
 * Maps the specified key to the specified value.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key the key
 * @param		value the value
 * @return		the value of any previous mapping with the specified key or
 *				null if there was no mapping
 */
public Object put(Object key, Object value) {
	poll();
	int index = 0;
	Entry entry;
	if (key != null) {
		index = (key.hashCode() & 0x7FFFFFFF) % elementData.length;
		entry = elementData[index];
		while (entry != null && !key.equals(entry.get()))
			entry = entry.next;
	} else {
		entry = elementData[0];
		while (entry != null && !entry.isNull)
			entry = entry.next;
	}
	if (entry == null) {
		modCount++;
		if (++elementCount > threshold) {
			rehash();
			index = key == null ? 0 : (key.hashCode() & 0x7FFFFFFF) % elementData.length;
		}
		entry = new Entry(key, value, referenceQueue);
		entry.next = elementData[index];
		elementData[index] = entry;
		return null;
	}
	Object result = entry.value;
	entry.value = value;
	return result;
}

private void rehash() {
	int length = elementData.length<<1;
	if (length == 0) length = 1;
	Entry[] newData = new Entry[length];
	for (int i=0; i<elementData.length; i++) {
		Entry entry = elementData[i];
		while (entry != null) {
			int index = entry.isNull ? 0 : (entry.hash & 0x7FFFFFFF) % length;
			Entry next = entry.next;
			entry.next = newData[index];
			newData[index] = entry;
			entry = next;
		}
	}
	elementData = newData;
	computeMaxSize();
}

/**
 * Removes a mapping with the specified key from this
 * WeakHashMap.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		key	the key of the mapping to remove
 * @return		the value of the removed mapping or null if key is not a key
 *				in this WeakHashMap
 */
public Object remove(Object key) {
	poll();
	int index = 0;
	Entry entry, last = null;
	if (key != null) {
		index = (key.hashCode() & 0x7FFFFFFF) % elementData.length;
		entry = elementData[index];
		while (entry != null && !key.equals(entry.get())) {
			last = entry;
			entry = entry.next;
		}
	} else {
		entry = elementData[0];
		while (entry != null && !entry.isNull) {
			last = entry;
			entry = entry.next;
		}
	}
	if (entry != null) {
		modCount++;
		if (last == null) elementData[index] = entry.next;
		else last.next = entry.next;
		elementCount--;
		return entry.value;
	}
	return null;
}

/**
 * Answers the number of mappings in this WeakHashMap.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the number of mappings in this WeakHashMap
 */
public int size() {
	poll();
	return elementCount;
}
}
