package java.util.jar;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.*;
import java.util.*;
import java.security.AccessController;
import com.ibm.oti.util.PriviAction;

class InitManifest {
	private static final boolean useNative = com.ibm.oti.vm.VM.useNatives();
	private byte[] inbuf = new byte[1024];
	private int inbufCount = 0, inbufPos = 0;
	private byte[] buffer = new byte[5];
	private char[] charbuf = new char[0];
	private ByteArrayOutputStream out = new ByteArrayOutputStream(256);
	private String encoding;
	private com.ibm.oti.io.CharacterConverter converter;

	private Map attributeNames = new HashMap();

	InitManifest(InputStream is, Attributes main, Map entries, Map chunks, String verString) throws IOException {
		encoding = (String)AccessController.doPrivileged(new PriviAction("manifest.read.encoding"));
		if ("".equals(encoding)) {
			encoding = null;
		}

		Attributes current = main;
		ArrayList list = new ArrayList();
		readLines(is, list);
		Iterator it = list.iterator();
		while (it.hasNext())
			addAttribute((String)it.next(), current);

		// Check for version attribute
		if (verString != null && main.getValue(verString) == null)
			// K0009 = Missing version string: {0}
			throw new IOException(com.ibm.oti.util.Msg.getString("K0009", verString));

		list.clear();
		byte[] chunk = null;
		while (chunks == null ? readLines(is,list) : (chunk = nextChunk(is, list)) != null) {
			it = list.iterator();
			String line = (String)it.next();
			if (line.length() < 7 || !line.substring(0, 5).toLowerCase().equals("name:"))
				// K000a = Entry is not named
				throw new IOException(com.ibm.oti.util.Msg.getString("K000a"));
			String name = line.substring(6, line.length()); //Name: length required space char
			current = new Attributes(12);
			if (chunks != null) chunks.put(name, chunk);
			entries.put(name, current);
			while (it.hasNext())
				addAttribute((String)it.next(), current);
			list.clear();
		}

	}

	private void addLine(int length, List lines) throws IOException {
		if (encoding != null) {
			lines.add(new String(buffer, 0, length, encoding));
		} else {
			if (converter == null) {
				try {
					if (useNative) {
						lines.add(com.ibm.oti.util.Util.convertFromUTF8(buffer, 0, length));
					} else {
						if (charbuf.length < length) charbuf = new char[length];
						lines.add(com.ibm.oti.util.Util.convertUTF8WithBuf(buffer, charbuf, 0, length));
					}
				} catch (UTFDataFormatException e) {
					converter = new com.ibm.oti.io.CharacterConverter();
				}
			}
			if (converter != null) {
				if (charbuf.length < length) charbuf = new char[length];
				// If invalid UTF8, convert bytes to chars setting the upper bytes to zeros
				converter.convert(buffer, 0, charbuf, 0, length);
				lines.add(new String(charbuf, 0, length));
			}
		}
	}

	private byte[] nextChunk(InputStream in, List lines) throws IOException {
		if (inbufCount == -1) return null;
		byte next;
		int pos = 0;
		boolean blankline = false, lastCr = false;
		out.reset();
		while (true) {
			if (inbufPos == inbufCount) {
				if ((inbufCount = in.read(inbuf)) == -1) {
					if (out.size() == 0) return null;
					if (blankline) addLine(pos, lines);
					return out.toByteArray();
				}
				inbufPos = 0;
			}
			next = inbuf[inbufPos++];
			if (lastCr) {
				if (next != '\n') {
					inbufPos--;
					next = '\r';
				} else {
					if (out.size() == 0) continue;
					out.write('\r');
				}
				lastCr = false;
			} else if (next == '\r') {
				lastCr = true;
				continue;
			}
			if (blankline) {
				if (next == ' ') {
					out.write(next);
					blankline = false;
					continue;
				}
				addLine(pos, lines);
				if (next == '\n') {
					out.write(next);
					return out.toByteArray();
				}
				pos = 0;
			} else if (next == '\n') {
				if (out.size() == 0) continue;
				out.write(next);
				blankline = true;
				continue;
			}
			blankline = false;
			out.write(next);
			if (pos == buffer.length) {
			    byte[] newBuf = new byte[buffer.length * 2];
			    System.arraycopy(buffer, 0, newBuf, 0, buffer.length);
			    buffer = newBuf;
			}
			buffer[pos++] = next;
		}
	}

	private boolean readLines(InputStream in, List lines) throws IOException {
		if (inbufCount == -1) return false;
		byte next;
		int pos = 0;
		boolean blankline = false, lastCr = false;
		while (true) {
			if (inbufPos == inbufCount) {
				if ((inbufCount = in.read(inbuf)) == -1) {
					if (blankline) addLine(pos, lines);
					return lines.size() != 0;
				}
				inbufPos = 0;
			}
			next = inbuf[inbufPos++];
			if (lastCr) {
				if (next != '\n') {
					inbufPos--;
					next = '\r';
				}
				lastCr = false;
			} else if (next == '\r') {
				lastCr = true;
				continue;
			}
			if (blankline) {
				if (next == ' ') {
					blankline = false;
					continue;
				}
				addLine(pos, lines);
				if (next == '\n') return true;
				pos = 0;
			} else if (next == '\n') {
				if (pos == 0 && lines.size() == 0) continue;
				blankline = true;
				continue;
			}
			blankline = false;
			if (pos == buffer.length) {
			    byte[] newBuf = new byte[buffer.length * 2];
			    System.arraycopy(buffer, 0, newBuf, 0, buffer.length);
			    buffer = newBuf;
			}
			buffer[pos++] = next;
		}
	}

	/* Get the next attribute and add it */
	private void addAttribute(String line, Attributes current) throws IOException {
		String header;
		int hdrIdx = line.indexOf(':');
		if (hdrIdx < 1)
			// K000b = Invalid attribute {0}
			throw new IOException(com.ibm.oti.util.Msg.getString("K000b", line));
		header = line.substring(0, hdrIdx);
		Attributes.Name name = (Attributes.Name)attributeNames.get(header);
		if (name == null) {
			try {
				name = new Attributes.Name(header);
			} catch (IllegalArgumentException e) {
				throw new IOException(e.toString());
			}
			attributeNames.put(header, name);
		}
		if (hdrIdx + 1 >= line.length() || line.charAt(hdrIdx + 1) != ' ')
			// K000b = Invalid attribute {0}
			throw new IOException(com.ibm.oti.util.Msg.getString("K000b", line));
		current.put(name, line.substring(hdrIdx + 2, line.length())); //+2 due to required SPACE char
	}
}
