package java.util.jar;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.util.zip.*;
import java.io.*;
import java.security.*;

/**
 * JarFile is used to read jar entries and their associated data from jar files.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			JarInputStream
 * @see			JarEntry
 */
public class JarFile extends java.util.zip.ZipFile {
	public static final String MANIFEST_NAME = "META-INF/MANIFEST.MF";
	static final String META_DIR = "META-INF/";

	private Manifest manifest;
	private ZipEntry manifestEntry;
	JarVerifier verifier;

	static final class JarFileInputStream extends FilterInputStream {
		private long count;
		private ZipEntry zipEntry;
		private JarVerifier verifier;
		private JarVerifier.VerifierEntry entry;
		private MessageDigest digest;

		JarFileInputStream(InputStream is, ZipEntry ze, JarVerifier ver){
			super(is);
			if (ver != null) {
				zipEntry = ze;
				verifier = ver;
				count = zipEntry.getSize();
				entry = verifier.initEntry(ze.getName());
				if (entry != null) digest = entry.digest;
			}
		}
		public int read() throws IOException {
			int r = super.read();
			if (entry != null) {
				if (r != -1) {
					digest.update((byte)r);
					count--;
				}
				if (r == -1 || count <= 0) {
					JarVerifier.VerifierEntry temp = entry;
					entry = null;
					verifier.verifySignatures(temp, zipEntry);
				}
			}
			return r;
		}
		public int read(byte[] buf, int off, int nbytes) throws IOException {
			int r = super.read(buf, off, nbytes);
			if (entry != null) {
				if (r != -1) {
					int size = r;
					if (count < size) size = (int)count;
					digest.update(buf, off, size);
					count -= r;
				}
				if (r == -1 || count <= 0) {
					JarVerifier.VerifierEntry temp = entry;
					entry = null;
					verifier.verifySignatures(temp, zipEntry);
				}
			}
			return r;
		}
		public long skip(long nbytes) throws IOException {
			long count = 0, rem = 0;
			byte[] buf = new byte[4096];
			while (count < nbytes) {
				int x = read(buf, 0, (rem = nbytes - count) > buf.length ? buf.length : (int)rem);
				if (x == -1) return count;
				count += x;
			}
			return count;
		}
	}

/**
 * Create a new JarFile using the contents of file.
 *
 * @author		OTI
 * @version		initial
 *
 * @param file 	java.io.File
 * @exception 	java.io.IOException If the file cannot be read.
 */
public JarFile(java.io.File file) throws java.io.IOException {
	this(file, true);
}

/**
 * Create a new JarFile using the contents of file.
 *
 * @author		OTI
 * @version		initial
 *
 * @param file 	java.io.File
 * @param verify	verify a signed jar file
 * @exception 	java.io.IOException If the file cannot be read.
 */
public JarFile(java.io.File file, boolean verify) throws java.io.IOException {
	super(file);
	if (verify)
		verifier = new JarVerifier(file.getPath());
	readMetaEntries();
}

/**
 * Create a new JarFile using the contents of file.
 *
 * @author		OTI
 * @version		initial
 *
 * @param file 	java.io.File
 * @param verify	verify a signed jar file
 * @param mode	the mode to use, either OPEN_READ or OPEN_READ | OPEN_DELETE
 * @exception 	java.io.IOException If the file cannot be read.
 */
public JarFile(java.io.File file, boolean verify, int mode) throws java.io.IOException {
	super(file, mode);
	if (verify)
		verifier = new JarVerifier(file.getPath());
	readMetaEntries();
}

/**
 * Create a new JarFile from the contents of the file specified by filename.
 *
 * @author		OTI
 * @version		initial
 *
 * @param filename 	java.lang.String
 * @exception 	 	java.io.IOException If fileName cannot be opened for reading.
 */
public JarFile(String filename) throws java.io.IOException {
	this(filename, true);

}

/**
 * Create a new JarFile from the contents of the file specified by filename.
 *
 * @author		OTI
 * @version		initial
 *
 * @param filename 	java.lang.String
 * @param verify		verify a signed jar file
 * @exception 	 	java.io.IOException If fileName cannot be opened for reading.
 */
public JarFile(String filename, boolean verify) throws java.io.IOException {
	super(filename);
	if (verify)
		verifier = new JarVerifier(filename);
	readMetaEntries();
}

/**
 * Return an enumeration containing the JarEntrys contained in
 * this JarFile.
 *
 * @author		OTI
 * @version		initial
 *
 * @return 		java.util.Enumeration
 * @exception 	java.lang.IllegalStateException If this JarFile has been closed.
 */
public java.util.Enumeration entries() {
	class JarFileEnumerator implements java.util.Enumeration{
		java.util.Enumeration ze;
		JarFile jf;
		JarFileEnumerator(java.util.Enumeration zenum, JarFile jf) {
			ze = zenum;
			this.jf = jf;
			}
		public boolean hasMoreElements(){
			return ze.hasMoreElements();
			}
		public Object nextElement(){
			JarEntry je = new JarEntry((ZipEntry)ze.nextElement());
			je.parentJar = jf;
			if (verifier != null)
				je.certificates = verifier.getCertificates(je.getName());
			return je;
			}
		}
	return new JarFileEnumerator(super.entries(), this);
}
/**
 * Return the JarEntry specified by name or null if no such entry exists.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		name	the name of the entry in the jar file
 * @return 		java.util.jar.JarEntry
 */
public JarEntry getJarEntry(String name){
	return (JarEntry)getEntry(name);
}
/**
 * Returns the Manifest object associated with this JarFile or null
 * if no manifest entry exists.
 *
 * @author		OTI
 * @version		initial
 *
 * @return java.util.jar.Manifest
 */
public Manifest getManifest() throws IOException {
	if (manifest != null) return manifest;
	if (manifestEntry != null) {
		ByteArrayInputStream is = (ByteArrayInputStream)super.getInputStream(manifestEntry);
		if (verifier != null) {
			byte[] buf = new byte[is.available()];
			is.mark(buf.length);
			is.read(buf, 0, buf.length);
			is.reset();
			verifier.addMetaEntry(manifestEntry.getName(), buf);
		}
		try {
			manifest = new Manifest(is, verifier != null);
		} finally {
			is.close();
		}
		manifestEntry = null;
	}
	return manifest;
}

private void readMetaEntries() throws IOException {
	ZipEntry []metaEntries = this.getMetaEntriesImpl(null);
	int dirLength = META_DIR.length();

	boolean signed = false;
	if(null != metaEntries) {
		for(int i=0;i<metaEntries.length;i++) {
			ZipEntry entry = metaEntries[i];
			String entryName = entry.getName();
			if (manifestEntry == null && manifest == null &&
				entryName.regionMatches(true, dirLength, MANIFEST_NAME, dirLength, MANIFEST_NAME.length() - dirLength))
			{
				manifestEntry = entry;
				if (verifier == null) break;
			} else if (verifier != null && entryName.length() > dirLength &&
				(entryName.regionMatches(true, entryName.length() - 3, ".SF", 0, 3) ||
				entryName.regionMatches(true, entryName.length() - 4, ".DSA", 0, 4) ||
				entryName.regionMatches(true, entryName.length() - 4, ".RSA", 0, 4)))
			{
				signed = true;
				InputStream is = super.getInputStream(entry);
				byte[] buf = new byte[is.available()];
				is.read(buf, 0, buf.length);
				is.close();
				verifier.addMetaEntry(entryName, buf);
			}
		}
	}
	if (!signed) verifier = null;
}

/**
 * Return an InputStream for reading the decompressed contents
 * of ze.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		ze	the ZipEntry to read from
 * @return 		java.io.InputStream
 * @exception 	java.io.IOException If an error occured while creating the InputStream.
 */
public InputStream getInputStream(ZipEntry ze) throws IOException {
	if (manifestEntry != null) getManifest();
	if (verifier != null) {
		verifier.setManifest(getManifest());
		if (verifier.readCertificates()) {
			verifier.removeMetaEntries();
			if (manifest != null) manifest.removeChunks();
			if (!verifier.isSignedJar()) verifier = null;
		}
	}
	InputStream in = super.getInputStream(ze);
	if (in == null) return null;
	return new JarFileInputStream(in, ze, ze.getSize() >= 0 ? verifier : null);
}

/**
 * Return the JarEntry specified by name or null if no such entry exists
 *
 * @author		OTI
 * @version		initial
 *
 * @param		name	the name of the entry in the jar file
 * @return 		java.util.jar.JarEntry
 */
public ZipEntry getEntry (String name) {
	ZipEntry ze = super.getEntry(name);
	if (ze == null)
		return ze;
	JarEntry je = new JarEntry(ze);
	je.parentJar = this;
	if (verifier != null)
		je.certificates = verifier.getCertificates(je.getName());
	return je;
}

private native ZipEntry []getMetaEntriesImpl (byte[] buf);

}

