package java.util.zip;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

import java.io.InputStream;
import java.io.IOException;

/**
 * The CheckedInputStream class is used to maintain
 * a running Checksum of all data read from a stream.
 *
 * @author		OTI
 * @version		initial
 *
 */
public class CheckedInputStream extends java.io.FilterInputStream {

	private Checksum check;

/**
 * Constructs a new CheckedInputStream on InputStream is.
 * The Checksum will be calculated using the algorithm
 * implemented by csum.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		is		InputStream to calculate checksum from
 * @param		csum	Type of Checksum to calculate
 */
public CheckedInputStream(InputStream is, Checksum csum){
	super(is);
	check =  csum;
}
/**
 * Reads a byte of data from the underlying stream and
 * recomputes the Checksum with the byte data.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		-1 if end of stream, a single byte value otherwise
 */
public int read () throws IOException{
	int x = in.read();
	if (x != -1)
		check.update(x);
	return x;
}
/**
 * Reads up to nbytes of data from the underlying stream, storing
 * it in buf, starting at offset off. The Checksum is updated
 * with the bytes read.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Number of bytes read, -1 if end of stream
 */
public int read (byte[] buf, int off, int nbytes) throws IOException{
	int x = in.read(buf, off, nbytes);
	if (x != -1)
		check.update(buf, off, x);
	return x;
}

/**
 * Returns the Checksum calculated on the stream thus far.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		A java.util.zip.Checksum
 */
public Checksum getChecksum () {
	return check;
}
/**
 * Skip upto nbytes of data on the underlying stream.
 * Any skipped bytes are added to the running Checksum value.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		nbytes long  Number of bytes to skip
 * @return		Number of bytes skipped
 */
public long skip(long nbytes ) throws IOException{
	if (nbytes < 1) return 0;
	long skipped = 0;
	byte[] b = new byte[2048];
	int x, v;
	while (skipped != nbytes){
		x = in.read(b, 0, (v = (int)(nbytes - skipped)) > b.length ? b.length : v);
		if (x == -1)
			return skipped;
		else{
			check.update(b, 0, x);
			skipped += x;
			}
		}
	return skipped;
}

}
