package java.util.zip;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2004  All Rights Reserved
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.EOFException;

/**
 * The GZIPInputStream class is used to read data stored in the GZIP
 * format.
 *
 * @author		OTI
 * @version		initial
 *
 */
public class GZIPInputStream extends java.util.zip.InflaterInputStream {
	protected CRC32 crc = new CRC32();
	protected boolean eos = false;
	public final static int GZIP_MAGIC = 0x8b1f;
	private static final int FHCRC = 2;
	private static final int FEXTRA = 4;
	private static final int FNAME = 8;
	private static final int FCOMMENT = 16;

/**
 * Construct a GZIPInputStream to read from GZIP data from the underlying stream
 *
 * @author		OTI
 * @version		initial
 *
 * @param		is  InputStream to read data from
 */
public GZIPInputStream(InputStream is) throws IOException {
	this(is, BUF_SIZE);
}
/**
 * Construct a GZIPInputStream to read from GZIP data from the underlying stream.
 * Set the internal buffer size to size
 *
 * @author		OTI
 * @version		initial
 *
 * @param		is  	InputStream to read data from
 * @param		size	Internal read buffer size
 */
public GZIPInputStream(InputStream is, int size) throws IOException {
	super(is, new Inflater(true), size);
	byte[] header = new byte[10];
	readFully(header, 0, header.length);
	if (getShort(header, 0) != GZIP_MAGIC)
		// K0020 = Unknown format
		throw new IOException(com.ibm.oti.util.Msg.getString("K0020"));
	int flags = header[3];
	boolean hcrc = (flags & FHCRC) != 0;
	if (hcrc) crc.update(header, 0, header.length);
	if ((flags & FEXTRA) != 0) {
		readFully(header, 0, 2);
		if (hcrc) crc.update(header, 0, 2);
		int length = getShort(header, 0);
		while (length > 0) {
			int max = length > buf.length ? buf.length : length;
			int result = in.read(buf, 0, max);
			if (result == -1) throw new EOFException();
			if (hcrc) crc.update(buf, 0, result);
			length -= result;
		}
	}
	if ((flags & FNAME) != 0) readZeroTerminated(hcrc);
	if ((flags & FCOMMENT) != 0) readZeroTerminated(hcrc);
	if (hcrc) {
		readFully(header, 0, 2);
		int crc16 = getShort(header, 0);
		if ((crc.getValue() & 0xffff) != crc16)
			// K0077 = Crc mismatch
			throw new IOException(com.ibm.oti.util.Msg.getString("K0077"));
		crc.reset();
	}
}
private long getLong(byte[] buf, int off){
	long l = 0;
	l |= (buf[off] & 0xFF);
	l |= (buf[off+1] & 0xFF) << 8;
	l |= (buf[off+2] & 0xFF) << 16;
	l |= ((long)(buf[off+3] & 0xFF)) << 24;
	return l;
}
private int getShort(byte[] buf, int off){
	return (buf[off] & 0xFF) | ((buf[off+1] & 0xFF) << 8);
}
/**
 * Reads and decompresses GZIP data from the underlying stream
 * into buf.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer	Buffer to receive data
 * @param		off		Offset in buffer to store data
 * @param		nbytes	Number of bytes to read
 */
public int read (byte[] buffer, int off, int nbytes) throws IOException{
	// avoid int overflow, check null buffer
	if (off <= buffer.length && nbytes >= 0 && off >= 0 && buffer.length - off >= nbytes) {
		int val = super.read(buffer, off, nbytes);
		if (val != -1)
			crc.update(buffer, off, val);
		else if (!eos) {
			eos = true;
			// Get non-compressed bytes read by fill
			int size = 0;
			byte[] b = new byte[8];
			int inB = inf.getTotalIn();
			if (inB > 0) {
				int diff = inB % buf.length;
				if (diff != 0 || len != buf.length) {
					size = len - diff;
					if (size > b.length) size = b.length;
					System.arraycopy(buf, diff, b, 0, size);
				}
			}
			readFully(b, size, b.length - size);
			if (getLong(b, 0) != crc.getValue())
				// K0077 = Crc mismatch
				throw new IOException(com.ibm.oti.util.Msg.getString("K0077"));
			if ((int)getLong(b, 4) != inf.getTotalOut())
				// K00ae = Size mismatch
				throw new IOException(com.ibm.oti.util.Msg.getString("K00ae"));
		}
		return val;
	} else throw new ArrayIndexOutOfBoundsException();
}
/**
 * Closes this stream and any underlying streams.
 *
 * @author		OTI
 * @version		initial
 *
 */
public void close () throws IOException{
	eos = true;
	super.close();
}

private void readFully(byte[] buf, int offset, int length) throws IOException {
	int result;
	while (length > 0) {
		result = in.read(buf, offset, length);
		if (result == -1) throw new EOFException();
		offset += result;
		length -= result;
	}
}

private void readZeroTerminated(boolean hcrc) throws IOException {
	int result;
	while ((result = in.read()) > 0) {
		if (hcrc) crc.update(result);
	}
	if (result == -1) throw new EOFException();
	// Add the zero
	if (hcrc) crc.update(result);
}
}
