package java.util.zip;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.EOFException;

/**
 * InflaterOuputStream read data which has been compressed using the DEFLATE
 * compression method.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			Inflater
 * @see			DeflaterOutputStream
 */
public class InflaterInputStream extends java.io.FilterInputStream {

	protected Inflater inf;
	protected byte[] buf;
	protected int len;
	boolean closed = false;
	boolean eof = false;
	static final int BUF_SIZE = 512;
/**
 * Constructs a new InflaterOutputStream on is
 *
 * @author		OTI
 * @version		initial
 *
 * &param		is		The InputStream to read data from
 */
public InflaterInputStream(InputStream is){
	this(is, new Inflater(), BUF_SIZE);
}
/**
 * Constructs a new InflaterOutputStream on is, using
 * the Inflater provided in inf.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		is		The InputStream to read data from
 * @param		inf		The Inflater to use for decompression
 */
public InflaterInputStream(InputStream is, Inflater inf){
	this(is, inf, BUF_SIZE);
}
/**
 * Constructs a new InflaterOutputStream on is, using
 * the Inflater provided in inf. The dsize of the
 * inflation buffer is determined by bsize.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		is		The InputStream to read data from
 * @param		inf		The Inflater to use for decompression
 * @param		bsize	sie of the inflation buffer
 */
public InflaterInputStream(InputStream is, Inflater inf, int bsize){
	super(is);
	if (is == null || inf == null)
		throw new NullPointerException();
	if (bsize <= 0)
		throw new IllegalArgumentException();
	this.inf = inf;
	buf = new byte[bsize];
}
/**
 * Reads a single byte of decompressed data.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		byte read
 * @exception	IOException		If an error occurs reading
 */
public int read () throws IOException {
	byte[] b = new byte[1];
	if (read(b, 0, 1) == -1) return -1;
	return b[0] & 0xff;
}
/**
 * Reads up to nbytes of decompressed data and stores it
 * in buf starting at off.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		buffer			Buffer to store into
 * @param		off				offset in buffer to store at
 * @param		nbytes			number of bytes to store
 * @return		Number of uncompressed bytes read
 * @exception	IOException		If an error occurs reading
 */
public int read (byte[] buffer, int off, int nbytes) throws IOException {
    /* [MSG "K0059", "Stream is closed"] */
    if (closed)
        throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));

    if (null == buffer)
        throw new NullPointerException();

    if (off < 0 || nbytes < 0 || off + nbytes > buffer.length)
        throw new IndexOutOfBoundsException();

    if (nbytes == 0)
        return 0;

    if (inf.finished()) {
        eof = true;
        return -1;
    }

    // avoid int overflow, check null buffer
    if (off <= buffer.length && nbytes >= 0 && off >= 0
            && buffer.length - off >= nbytes) {
        if (nbytes == 0)
            return 0;
        do {
            if (inf.needsInput())
                fill();
            int result;
            try {
                result = inf.inflate(buffer, off, nbytes);
            } catch (DataFormatException e) {
                if (len == -1)
                    throw new EOFException();
                throw new IOException(e.getMessage());
            }
            if (result > 0)
                return result;
            else if (inf.finished()) {
                eof = true;
                return -1;
            } else if (inf.needsDictionary())
                return -1;
            else if (len == -1)
                throw new EOFException();
            // If result == 0, fill() and try again
        } while (true);
    }
    throw new ArrayIndexOutOfBoundsException();
}

protected void fill () throws IOException {
	// K0059 = Stream is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
	if ((len = in.read(buf)) > 0)
		inf.setInput(buf, 0, len);
}
/**
 * Skips up to nbytes of uncompressed data.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		nbytes			Number of bytes to skip
 * @return		Number of uncompressed bytes skipped
 * @exception	IOException		If an error occurs skipping
 */
public long skip (long nbytes) throws IOException {
	if (nbytes >= 0) {
		long count = 0, rem = 0;
		while (count < nbytes) {
			int x = read(buf, 0, (rem = nbytes - count) > buf.length ? buf.length : (int)rem);
			if (x == -1) {
				eof = true;
				return count;
			}
			count += x;
		}
		return count;
	} else throw new IllegalArgumentException();
}
/**
 * Returns 0 if this stream has been closed,
 * 1 otherwise.
 *
 * @author		OTI
 * @version		initial
 *
 * @exception	IOException		If an error occurs
 */
public int available() throws IOException {
	// K0059 = Stream is closed
	if (closed) throw new IOException(com.ibm.oti.util.Msg.getString("K0059"));
	if (eof)
		return 0;
	return 1;
}
/**
 * Closes the stream
 *
 * @author		OTI
 * @version		initial
 *
 *  @exception	IOException 	If an error occurs closing the stream
 */
public void close() throws IOException{
	if (!closed) {
		inf.end();
		closed = true;
		eof = true;
		super.close();
	}
}

}
