package java.util.zip;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2006  All Rights Reserved
 */

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * ZipFile is used to read zip entries and their associated data from zip files.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			ZipInputStream
 * @see			ZipEntry
 */

public class ZipFile implements ZipConstants {
	private String fileName;

	long descriptor = -1;
	private int size = -1;
	private int mode;
	private Object lock = new Object();

	static { ntvinit(); }

	/**
	 * Open zip file for read.
	 */
	public static final int OPEN_READ = 1;
	/**
	 * Delete zip file when closed.
	 */
	public static final int OPEN_DELETE = 4;

/**
 * Constructs a new ZipFile opened on the specified File.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file	the File
 */
public ZipFile(File file) throws ZipException, IOException {
	this(file.getPath());
}

/**
 * Constructs a new ZipFile opened on the specified File using the
 * specified mode.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		file	the File
 * @param		mode the mode to use, either OPEN_READ or OPEN_READ | OPEN_DELETE
 */
public ZipFile(File file, int mode) throws IOException {
	if (mode == OPEN_READ || mode == (OPEN_READ | OPEN_DELETE)) {
		fileName = file.getPath();
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(fileName);
			if ((mode & OPEN_DELETE) != 0) {
				security.checkDelete(fileName);
			}
		}
		this.mode = mode;
		openZip(fileName);
	} else
		throw new IllegalArgumentException();
}

/**
 * Constructs a new ZipFile opened on the specified file path name.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		filename	the file path name
 */
public ZipFile(String filename) throws IOException {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
		security.checkRead(filename);
	}
	fileName = filename;
	openZip(fileName);
}

private void openZip(String filename) throws IOException {
	int result;
	synchronized(lock) {
		result = openZipImpl(com.ibm.oti.util.Util.getBytes(fileName));
	}
	if (result != 0) {
		switch (result) {
			// K01c3 = Unable to open: {0}
			case 1: throw new ZipException(com.ibm.oti.util.Msg.getString("K01c3", fileName)); //$NON-NLS-1$
			// K01c4 = Invalid zip file: {0}
			case 2: throw new ZipException(com.ibm.oti.util.Msg.getString("K01c4", fileName)); //$NON-NLS-1$
			default:
				throw new OutOfMemoryError();
		}
	}
}

protected void finalize() throws IOException {
	close();
}

/**
 * Closes this ZipFile.
 *
 * @author		OTI
 * @version		initial
 */
public void close() throws IOException {
	if (fileName != null) {
		// Only close initialized instances
		synchronized(lock) {
			closeZipImpl();
		}
		if ((mode & OPEN_DELETE) != 0) {
		    AccessController.doPrivileged( new PrivilegedAction(){
		        public Object run(){
		        	new File(fileName).delete();
		            return null;
		        }
		    });
		}
	}
}
/**
 * Answers all of the zip entries contained in this ZipFile.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		an Enumeration of the zip entries
 */
public Enumeration entries() {
	synchronized(lock) {
		return new ZFEnum();
	}
}
/**
 * Gets the zip entry with the specified name from this ZipFile.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		entryName	the name of the entry in the zip file
 * @return		a ZipEntry or null if the entryname does not exist in the zip file
 */
public ZipEntry getEntry(String entryName) {
	if (entryName != null) {
		ZipEntry entry;
		synchronized(lock) {
			if(descriptor == -1) {
				// the descriptor is set to -1 by native code to indicate the zip was closed
				// K00b7 = File is closed
				throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K00b7")); //$NON-NLS-1$
			}
			entry = getEntryImpl(descriptor, entryName);
		}
		return entry;
	} else throw new NullPointerException();
}
/**
 * Answers an input stream on the data of the specified ZipEntry.
 *
 * @author		OTI
 * @version		initial
 *
 * @param		entry	the ZipEntry
 * @return		an input stream on the ZipEntry data
 */
public InputStream getInputStream(ZipEntry entry) throws IOException {
	byte[] buf;
	synchronized(lock) {
		if(descriptor == -1) {
			// the descriptor is set to -1 by native code to indicate the zip was closed
			// K00b7 = File is closed
			throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K00b7")); //$NON-NLS-1$
		}
		buf = inflateEntryImpl2(descriptor, entry.getName());
	}
	if (buf == null) return null;
	return new com.ibm.oti.util.AccessibleByteArrayInputStream(buf);
}
/**
 * Gets the file name of this ZipFile.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		the file name of this ZipFile
 */
public String getName() {
	return fileName;
}

private native int openZipImpl(byte[] fileName);
private native void closeZipImpl() throws IOException;
private native ZipEntry getEntryImpl(long descriptor, String entryName);
private native byte[] inflateEntryImpl2(long descriptor, String entryName) throws ZipException;

/**
 * Returns the number of ZipEntries in this ZipFile.
 *
 * @author		OTI
 * @version		initial
 *
 * @return		Number of entries in this file
 */
public int size() {
	if (size != -1) return size;

	size = 0;
	Enumeration e = entries();
	while (e.hasMoreElements()) {
		size++;
		e.nextElement();
	}
	return size;

}
private static native void ntvinit();

class ZFEnum implements java.util.Enumeration{
	private long nextEntryPointer;
	private ZipEntry current;

	ZFEnum() {
		nextEntryPointer = resetZip(descriptor);
		current = getNextEntry(descriptor, nextEntryPointer);
	}
	private native long resetZip(long descriptor);
	private native ZipEntry getNextEntry(long descriptor, long nextEntryPointer);

	public boolean hasMoreElements(){
		synchronized(lock) {
			if(descriptor == -1) {
				// the descriptor is set to -1 by native code to indicate the zip was closed
				// K00b7 = File is closed
				throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K00b7")); //$NON-NLS-1$
			}
		}
		return current != null;
	}
	public Object nextElement() {
		if (current == null) throw new NoSuchElementException();
		ZipEntry ze = current;
		synchronized(lock) {
			if(descriptor == -1) {
				// the descriptor is set to -1 by native code to indicate the zip was closed
				// K00b7 = File is closed
				throw new IllegalStateException(com.ibm.oti.util.Msg.getString("K00b7")); //$NON-NLS-1$
			}
			current = getNextEntry(descriptor, nextEntryPointer);
		}
		return ze;
	}
}
}
