package javax.security.auth.x500;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1997, 2005  All Rights Reserved
 */

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.security.Principal;

import com.ibm.oti.util.ASN1Exception;

/**
 * This class represents X500's distinguished names.
 */
public final class X500Principal implements Principal, Serializable {

	/* [PR 117156] - TCK errors in serialization */
	/**
	 * serialVersionUID comes from Foundation 1.1 specification
     */
	private static final long serialVersionUID = -500463348111345721l ;
	/**
	 *  RFC 1779 String format of X.500 Distinguished Names.
	 */
	public static final String RFC1779 = "RFC1779"; //$NON-NLS-1$
	/**
	 * RFC 2253 String format of X.500 Distinguished Names.
	 */
	public static final String RFC2253 = "RFC2253"; //$NON-NLS-1$
	/**
	 * Canonical String format of X.500 Distinguished Names
	 */
	public static final String CANONICAL = "CANONICAL"; //$NON-NLS-1$

	/* [PR 117156] - Throw exceptions for constructors with invalid arguments */
	private transient com.ibm.oti.security.provider.X500Principal principal = null;

	/**
	 * Creates an X500 Principal from data (ASN1 DER encoding)
	 *
	 * @param 		name 		distinguished name in ASN1 DER encoding
	 *
	 * @throws 		IllegalArgumentException 	If an encoding error is present in data
	 */
	public X500Principal( byte[] name ){
		try	{
			principal = new com.ibm.oti.security.provider.X500Principal( name );
		} catch ( ASN1Exception e )	{
			throw new IllegalArgumentException( e.getMessage() );
		}
	}

	/**
	 * Creates an X500 Principal from data (ASN1 DER encoding) in the InputStream
	 * The read position in the stream is set to the first byte after the
	 * X500Principal that is read
	 *
	 * @param 		is 			InputStream with current position at start of X500Principal
	 *
	 * @throws 		NullPointerException 		If is null
	 * @throws 		IllegalArgumentException 	If an encoding error is present in data
	 */
	public X500Principal( InputStream is ){

		/* [PR 117156] - Throw exceptions for constructors with invalid arguments */
		if( is == null ){
			throw new NullPointerException();
		}

		try	{
			principal = new com.ibm.oti.security.provider.X500Principal( is );
		} catch ( ASN1Exception e )	{
			throw new IllegalArgumentException( e.getMessage() );
		}

	}

	/**
	 * Creates an X500 Principal from a String representation of the distinguished name
	 *
	 * @param 		name 		<code>String</code> representation of distinguished name
	 *
	 * @throws 		NullPointerException 	If name null
	 */
	public X500Principal(String name){
		if( name == null ){
			throw new NullPointerException();
		}

		principal = new com.ibm.oti.security.provider.X500Principal( name );

	}

	/**
	 * Returns the distinguished name as a String as defined by RFC 2253
	 *
	 * @return		the distinguished name
	 */
	public String getName(){
		return getName( RFC2253 );
	}

	/**
	 * Returns the distinguished name as a String as defined by the provided
	 * encoding
	 *
	 * @param 		format 		encoding format
	 * @return 		distinguished name in specified format
	 *
	 * @throws 		IllegalArgumentException 	If specified format invalid
	 */
	public String getName( String format ){
		return principal.getName( mapAPIConstantToInternal( format ) );
	}

	/**
	 * Returns the distinguished name encoded in ASN.1 DER format
	 *
	 * @return 		name encoded in ASN.1 DER format
	 */
	public byte[] getEncoded(){
		return principal.getEncoded();
	}

	/**
	 * Returns a String representation of this X500Principal
	 *
	 * @return 		representation of X500Principal
	 */
	public String toString(){
		return principal.toString();
	}

	/**
	 * Compares for equality:
	 * Returns true if o is an X500Principal and its RFC_2459 representation is equivalent
	 * to that of the receiver.
	 *
	 * @param 		o			Object to test for equality
	 * @return 		<code>true</code> if objects are equal, <code>false</code> otherwise.
	 */
	public boolean equals( Object o ){

		if( (o instanceof X500Principal ) == false ){
			return false;
		}

		return this.getName( CANONICAL ).equals( ((X500Principal) o).getName( CANONICAL ) );

	}

	/**
	 * Returns a hash code for this X500Principal
	 * Hash code is equal to getName( RFC_2459 ).hashCode();
	 *
	 * @return 		hash code of X500Principal
	 */
	public int hashCode(){
		return this.getName( CANONICAL ).hashCode();
	}

	private int mapAPIConstantToInternal( String representation ){

		if( representation.equals( RFC1779 ) ){
			return com.ibm.oti.security.provider.X500Principal.OUTPUT_RFC1779;
		}
		if( representation.equals( RFC2253 ) ){
			return com.ibm.oti.security.provider.X500Principal.OUTPUT_RFC2253;
		}
		if( representation.equals( CANONICAL ) ){
			return com.ibm.oti.security.provider.X500Principal.OUTPUT_CANONICAL;
		}

		throw new IllegalArgumentException();

	}

	/* [PR 117156] - Throw exceptions for constructors with invalid arguments */
	private void writeObject(ObjectOutputStream stream) throws IOException {
		stream.writeObject(getEncoded());
	}

	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		byte[] encoded = (byte[]) stream.readObject();
		principal = new com.ibm.oti.security.provider.X500Principal(encoded);
	}

}
