package org.placelab.spotter;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;

import org.ubicollab.lm.sensor.wifi.model.*;

public class WiFiSpotter {
	private boolean spotterLibraryLoaded=false, spotterInitialized=false;
	
	/** Initialize the spotter
     * 
     * @param wifiInterface The wireless interface to use.  If null or the empty string "", use
     * the first found wireless interface.
     * 
     * @return true on successful init
     */
    public native void spotter_init(String wifiInterface) throws Exception;
    
    /** Shutdown the spotter.
     */
    public native void spotter_shutdown() throws Exception;
    
    /** Poll the spotter.
     * 
     * @return an array of BSSID strings
     */
    public native String[] spotter_poll() throws Exception;
    
    public WiFiSpotter(){
    	
    }
    
    public void open() throws Exception {
        if (spotterInitialized) return;
        
        try {
            if (!spotterLibraryLoaded) {
                spotterLoadLibrary();
                spotterLibraryLoaded = true;
            }
            if (!spotterInitialized) {
                //String wifiDevice = PlacelabProperties.get("placelab.wifi_device");
            	String wifiDevice = "";
            	
                if (!wifiDevice.equalsIgnoreCase(""))
                    System.out.println("Using explicitly requested WiFi interface \""+wifiDevice+"\".");

                // initialize the spotter
                spotter_init(wifiDevice);
                
                
                spotterInitialized = true;
                
                getMeasurementImpl();
            }
        } catch (UnsatisfiedLinkError e) {
            e.printStackTrace();
            throw new Exception("MSG: " + e.getMessage() + 
                  "\n*** WiFi spotter library unavailable, running without spotter");
        } catch (IOException e) {
            throw new Exception("MSG: " + e.getMessage() + 
            "*** WiFi spotter library unavailable, running without spotter\n" +
                    "*** Error loading spotter library, running without spotter");
        }
    }
    
    public void close() {
        if (spotterInitialized) {
            try {
                spotter_shutdown();
            } catch (Exception se) {
                System.err.println(se); 
            } finally {
                spotterInitialized = false;
            }
        }
    }
    
    public synchronized WifiReading[] getMeasurementImpl() throws Exception {
        if (!spotterInitialized || !spotterLibraryLoaded) {
            throw new Exception("Spotter not initialized");
        }

        
        // call the native method
        String sarr[] = spotter_poll();

        // putting these in a hashtable will eliminate dupes
        Hashtable readings = new Hashtable();
        for (int i = 0; sarr != null && i <= (sarr.length - 5); i += 5) {
            try {
                sawAP(readings, StringUtil.canonicalizeBSSID(sarr[i]),
                        sarr[i + 1], Integer.parseInt(sarr[i + 2]), !"0"
                                .equals(sarr[i + 3]), !"0".equals(sarr[i + 4]));
            } catch (Exception ex) {
                throw new Exception("WiFiSpotter: Malformed reply from native spotter:\n"+ex.getMessage());
            }
        }
        
        WifiReading[] wifiReadings = null;        
        
        if( readings.size() > 0 ){   
        	
        	wifiReadings = new WifiReading[readings.size()];
        	int i = 0;
        	
        	for(Enumeration it = readings.elements(); it.hasMoreElements();){
        		WifiReading apReading = (WifiReading) it.nextElement();
        		wifiReadings[i++] = apReading;
        	}
        	              
        	return wifiReadings;
        } else {
        	return null;
        }            
    }
    
    private void sawAP(Hashtable readings, String bssid, String ssid, int rssi,
            boolean wep, boolean inf) {
        /*
         * Sometimes cards return funky RSSI values; until we figure out what is
         * going on, we do the hack below (yatin) if (rssi <= -100) rssi = -65;
         * if (rssi >= -35) rssi = -65;
         */

        if (rssi < 0xffff0001)
            /*
             * One particular chipset (acx100 / I-Mate PDA 2k) likes to return a
             * RSSI padded with 2 useless bytes. The two bytes are exactly the
             * same for every RSSI value, so we just shift them out.
             *  
             */
            rssi = rssi >> 16;
        else if (rssi < -149)
            /*
             * our code expects RSSI values to be between
             * -Types.NETSTUMBLER_RSSI_ADJUSTMENT and zero check for it
             */
            rssi = -149;
        else if (rssi > 0)
            rssi = 0;

        try {
            //WiFiReading reading = new WiFiReading(bssid, ssid, rssi, wep, inf);
        	//System.out.println(bssid + " " + ssid + " " + rssi + " " + wep + " " + inf);
        	
        	WifiReading apReading = new WifiReading(ssid, bssid, rssi);        	
            
        	//System.out.println(apReading.getBSSID());
        	
        	readings.put(bssid, apReading);
        	
        } catch (IllegalArgumentException iae) {
            System.err.println("Ignoring invalid WiFi reading: " + iae);
        }
    }
    
    private void spotterLoadLibrary() throws IOException {
        String likelyLibName = System.mapLibraryName("spotter");
        // Get the extension
        int idx = likelyLibName.lastIndexOf('.');
        String ext;
        String pre;
        if (idx != -1) {
            ext = likelyLibName.substring(idx, likelyLibName.length());
            pre = likelyLibName.substring(0, idx);
        } else {
            pre = "spotter";
            ext = "";
        }

        if (likelyLibName.endsWith("dll")) {
            // then its a windows of some form
            String osName = System.getProperty("os.name");
            if (osName.equalsIgnoreCase("Windows CE")) {
                likelyLibName = "spotter_ce.dll";
            }
        }
        Enumeration e = this.getClass().getClassLoader().getResources(likelyLibName);
        boolean found = false;
        File tempFile = File.createTempFile(pre, ext);
        
        while (e.hasMoreElements()) {
            URL jarURL = (URL) e.nextElement();
            InputStream is = jarURL.openStream();
            // now squirt this lib into tempFile
            FileOutputStream fo = new FileOutputStream(tempFile);
            byte[] buffer = new byte[1024];
            int bytesRead;
            while ((bytesRead = is.read(buffer)) != -1) {
                fo.write(buffer, 0, bytesRead);
            }
            fo.close();
            String path = tempFile.getAbsolutePath();
            try {
                System.load(path);
                found = true;
                break;
            } catch (UnsatisfiedLinkError ule) {
                tempFile.delete();
                tempFile = File.createTempFile(pre, ext);
            }
        }
        if (!found) {        	        
            tempFile.delete();            
            System.load("\\TEMP\\spotter.dll");
            //System.loadLibrary("spotter");
        } else {
            tempFile.deleteOnExit();
        }
    }
}
