package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

/**
 * This class represents events which model an action occurring
 * on a <code>Component</code>.  The actual definition for what
 * that action is is defined by the component.  In the case of a
 * <code>Button</code> it would represent the pressing of the button.
 * <p>
 * This event is propagated to all <code>ActionListener</code> that
 * are registered on the <code>Component</code> using the
 * <code>addActionListener</code> method, action events occurring
 * on that component will be delivered to the action listener via
 * the sending of the <code>actionPerformed</code> method.
 * <p>
 * @author		IBM
 * @version		initial
 *
 */

public class ActionEvent extends java.awt.AWTEvent {

	static final long serialVersionUID = -7671078796273832149L;

	// The action command string of the source of the receiver
	String actionCommand;

	// The keyboard and mouse modifiers state
	int modifiers;

	// Action event ID constants
	public static final int ACTION_FIRST = 1001;
	public static final int ACTION_LAST = 1001;
	public static final int ACTION_PERFORMED = 1001;

	// Action event key modifiers
	// Alt
	public static final int ALT_MASK = 8;
	// Ctrl
	public static final int CTRL_MASK = 2;
	// Meta
	public static final int META_MASK = 4;
	// Shift
	public static final int SHIFT_MASK = 1;

	long when = 0;

	/**
	 * Constructs a new <code>ActionEvent</code> using the
	 * specified id and commandString.  The receiver will be
	 * an event that occured in the specified source object.
	 *
	 * @author		IBM
	 * @version 	initial
	 *
	 * @param	source			the source object of the receiver
	 * @param	id				the id of the receiver
	 * @param	commandString	the command string for the receiver
	 *
	 */
	public ActionEvent(Object source, int id, String commandString) {
		this(source, id, commandString, 0);
	}

	/**
	 * Constructs a new <code>ActionEvent</code> using the
	 * specified id, command string and modifiers.  The receiver will be
	 * an event that occured in the specified source object.
	 *
	 * @author		IBM
	 * @version 	initial
	 *
	 * @param	source			the source object of the receiver
	 * @param	id				the id of the receiver
	 * @param	commandString	the command string for the receiver
	  * @param	modifiers		the modifiers for the receiver
	 *
	 */
	public ActionEvent(Object source, int id, String commandString, int modifiers) {
		super(source, id);

		this.actionCommand = commandString;
		this.modifiers = modifiers;
	}

	/**
	 * Answer the string representing the action command for
	 * the receiver.
	 * <p>
	 * Depending on the type of action a component may want to
	 * specify a different command string for different types of
	 * the same action.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the action command
	 */
	public String getActionCommand() {
		return actionCommand;
	}

	/**
	 * Answer the modifiers for the receiver.
	 * This will be a combination (integer addition) of one or
	 * more of the following:
	 * 	- ALT_MASK
	 * 	- CTRL_MASK
	 * 	- META_MASK
	 * 	- SHIFT_MASK
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	an int	the modifiers
	 */
	public int getModifiers() {
		return modifiers;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case ACTION_PERFORMED:
				buffer.append("ACTION_PERFORMED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown type"); //$NON-NLS-1$
		}
		buffer.append(",cmd=");	 //$NON-NLS-1$
		buffer.append(getActionCommand());
		String modifiersText = KeyEvent.getKeyModifiersText(getModifiers());
		if (modifiersText.length() > 0) {
			buffer.append(",modifiers="); //$NON-NLS-1$
			buffer.append(modifiersText);
		}
		return buffer.toString();
	}

	public ActionEvent(Object source, int id, String command, long when, int modifiers) {
	    this(source, id, command, modifiers);
	    this.when = when;
	}

	public long getWhen() {
		return when;
	}
}
