package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Adjustable;

/**
 * This class represents events which model an adjustment occurring
 * on a <code>Component</code>.  These events occur on objects which
 * are implementors of <code>Adjustable</code>.
 * <p>
 * This event is propagated to all <code>AdjustmentListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addAdjustmentListener</code> method, adjustment events occurring
 * on that component will be delivered to the adjustment listener via
 * the sending of the <code>adjustmentValueChanged</code> method.
 * <p>
 * @author		IBM
 * @version		initial
 *
 */
public class AdjustmentEvent extends java.awt.AWTEvent {

	static final long serialVersionUID = 5700290645205279921L;

	// The adjustable that generated the receiver
	Adjustable adjustable;

	// The type of adjustment
	int adjustmentType;

	// The new value of the adjustable that generated the receiver
	int value;

	// Adjustment event ID constants
	public static final int ADJUSTMENT_FIRST = 601;
	public static final int ADJUSTMENT_LAST = 601;
	public static final int ADJUSTMENT_VALUE_CHANGED = 601;

	// Adjustment types
	public static final int BLOCK_DECREMENT = 3;
	public static final int BLOCK_INCREMENT = 4;
	public static final int TRACK = 5;
	public static final int UNIT_DECREMENT = 2;
	public static final int UNIT_INCREMENT = 1;

	/**
	 * Constructs a new <code>AdjustmentEvent</code> using the
	 * specified id, type and value.  The receiver will be
	 * an event that occured in the specified source object.
	 *
	 * @author		IBM
	 * @version 	initial
	 *
	 * @param	source	the source object of the receiver
	 * @param	id		the id of the receiver
	 * @param	type	the type for the receiver
	 * @param	value	the value for the receiver
	 *
	 */
	public AdjustmentEvent(Adjustable source, int id, int type, int value) {
		super(source, id);

		this.adjustable = source;
		this.adjustmentType = type;
		this.value = value;
	}

	/**
	 * Answer the source object for the receiver.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Adjustable	the source of the
	 * 									adjustment event
	 */
	public Adjustable getAdjustable() {
		return adjustable;
	}

	/**
	 * Answer the type of adjustment that the receiver represents.
	 * This is one of the following:
	 * 	<code>BLOCK_DECREMENT</code>
	 * 	<code>BLOCK_INCREMENT</code>
	 * 	<code>TRACK</code>
	 * 	<code>UNIT_DECREMENT</code>
	 * 	<code>UNIT_INCREMENT</code>
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	an int	the type of adjustment
	 */
	public int getAdjustmentType() {
		return adjustmentType;
	}

	/**
	 * Answer the value of the receiver.
	 * That is the value of the source object when the receiver was
	 * created.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	an int	the value of the receivers source
	 */
	public int getValue() {
		return value;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author		IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case ADJUSTMENT_VALUE_CHANGED:
				buffer.append("ADJUSTMENT_VALUE_CHANGED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",adjType="); //$NON-NLS-1$
		switch (getAdjustmentType()) {
			case UNIT_INCREMENT:
				buffer.append("UNIT_INCREMENT"); //$NON-NLS-1$
				break;
			case UNIT_DECREMENT:
				buffer.append("UNIT_DECREMENT"); //$NON-NLS-1$
				break;
			case BLOCK_DECREMENT:
				buffer.append("BLOCK_DECREMENT"); //$NON-NLS-1$
				break;
			case BLOCK_INCREMENT:
				buffer.append("BLOCK_INCREMENT"); //$NON-NLS-1$
				break;
			case TRACK:
				buffer.append("TRACK"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",value="); //$NON-NLS-1$
		buffer.append(getValue());
		return buffer.toString();
	}
}
