package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Component;
import java.awt.Rectangle;

/**
 * This class represents events which model a component event occurring
 * on a <code>Component</code>.
 * <p>
 * This event is propagated to all <code>ComponentListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addComponentListener</code> method, component events occurring
 * on that component will be delivered to the component listener via
 * the sending of one of
 * the following methods:
 * 	- <code>componentHidden</code>
 * 	- <code>componentMoved</code>
 * 	- <code>componentResized</code>
 * 	- <code>componentShown</code>
 * <p>
 * Component events will be handled by AWT regardless of the existence of
 * listeners, they are used to notify listeners of what is happening.
 *
 * @author      IBM
 * @version		initial
 *
 */
public class ComponentEvent extends java.awt.AWTEvent {

	static final long serialVersionUID = 8101406823902992965L;

	// Component event ID constants
	public static final int COMPONENT_FIRST = 100;
	public static final int COMPONENT_HIDDEN = 103;
	public static final int COMPONENT_LAST = 103;
	public static final int COMPONENT_MOVED = 100;
	public static final int COMPONENT_RESIZED = 101;
	public static final int COMPONENT_SHOWN = 102;

	/**
	 * Constructs a new <code>ComponentEvent</code> using the
	 * specified id.  The receiver will be an event that occured
	 * in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source	the source object of the receiver
	 * @param	id		the id of the receiver
	 *
	 */
	public ComponentEvent(Component source, int id) {
		super(source, id);
	}

	/**
	 * Answer the source object for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Component	the source of the
	 * 									component event
	 *
	 * @see	<code>getSource()</code>
	 */
	public Component getComponent() {
		return (Component)getSource();
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		int id = getID();
		switch (id) {
			case COMPONENT_MOVED:
				buffer.append("COMPONENT_MOVED"); //$NON-NLS-1$
				appendBounds(buffer);
				break;
			case COMPONENT_RESIZED:
				buffer.append("COMPONENT_RESIZED"); //$NON-NLS-1$
				appendBounds(buffer);
				break;
			case COMPONENT_SHOWN:
				buffer.append("COMPONENT_SHOWN"); //$NON-NLS-1$
				break;
			case COMPONENT_HIDDEN:
				buffer.append("COMPONENT_HIDDEN"); //$NON-NLS-1$
				break;
		}
		return buffer.toString();
	}

	private void appendBounds(StringBuffer buffer) {
		Rectangle bounds = getComponent().getBounds();
		buffer.append(" ("); //$NON-NLS-1$
		buffer.append(bounds.x);
		buffer.append(',');
		buffer.append(bounds.y);
		buffer.append(' ');
		buffer.append(bounds.width);
		buffer.append('x');
		buffer.append(bounds.height);
		buffer.append(')');
	}
}
