package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Component;
import java.awt.Container;

/**
 * This class represents events which model a change in the
 * child tree of a container occurring on a <code>Container</code>.
 * <p>
 * This event is propogated to all <code>ContainerListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addContainerListener</code> method, container events occurring
 * on that component will be delivered to the container listener via
 * the sending of one of the following methods:
 * 	- <code>componentAdded</code>
 * 	- <code>componentRemoved</code>
 * <p>
 * Container events will be handled by AWT regardless of the existence of
 * listeners, they are used to notify listeners of what is happening.
 *
 * @author      IBM
 * @version		initial
 *
 */
public class ContainerEvent extends ComponentEvent {

	static final long serialVersionUID = -4114942250539772041L;

	// The component that is being added or remove from the source
	Component child;

	// Container event ID constants
	public static final int COMPONENT_ADDED = 300;
	public static final int CONTAINER_FIRST = 300;
	public static final int CONTAINER_LAST = 301;
	public static final int COMPONENT_REMOVED = 301;

	/**
	 * Constructs a new <code>ContainerEvent</code> using the
	 * specified id, type and value.  The receiver will be
	 * an event that occured in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source	the source object of the receiver
	 * @param	id		the id of the receiver
	 * @param	child	the child being added or removed from the receiver
	 *
	 */
	public ContainerEvent(Component source, int id, Component child) {
		super(source, id);

		this.child = child;
	}

	/**
	 * Answer the child being added or removed in the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return		a java.lang.Component	the child being added or removed
	 */
	public Component getChild() {
		return child;
	}

	/**
	 * Answer the source for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Container	the source for the receiver
	 *
	 * @see	<code>getSource()</code>
	 */
	public Container getContainer() {
		return (Container)getSource();
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case COMPONENT_ADDED:
				buffer.append("COMPONENT_ADDED"); //$NON-NLS-1$
				break;
			case COMPONENT_REMOVED:
				buffer.append("COMPONENT_REMOVED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",child="); //$NON-NLS-1$
		buffer.append(child.getName());
		return buffer.toString();
	}
}
