package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Component;

/**
 * This class represents events which model a keyboard focus change
 * on a <code>Component</code>.
 * This event is propagated to all <code>FocusListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addFocusListener</code> method, focus events occurring
 * on that component will be delivered to the focus listener via
 * the sending of one of the following methods:
 * 	- <code>focusGained</code>
 * 	- <code>focusLost</code>
 *
 * @author      IBM
 * @version		initial
 *
 */
public class FocusEvent extends ComponentEvent {

	static final long serialVersionUID = 523753786457416396L;

	// Determines whether the source has lost or gain the focus temporarily
	boolean temporary;

	// Focus event ID constants
	public static final int FOCUS_FIRST = 1004;
	public static final int FOCUS_GAINED = 1004;
	public static final int FOCUS_LAST = 1005;
	public static final int FOCUS_LOST = 1005;

	Component opposite;

	/**
	 * Constructs a new <code>FocusEvent</code> using the
	 * specified id.  The receiver will be an event that occured
	 * in the specified source object.
	 * In this case it is assumed that the focus change is not
	 * temporary and will remain with the source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 *
	 */
	public FocusEvent(Component source, int id) {
		this(source, id, false);
	}

	/**
	 * Constructs a new <code>FocusEvent</code> using the
	 * specified id and temporary flag.  The receiver will be
	 * an event that occured in the specified source object.
	 * Temporary indicates that the focus may not remain with the
	 * source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 * @param	temporary	the temporary flag for the receiver
	 *
	 */
	public FocusEvent(Component source, int id, boolean temporary) {
		super(source, id);

		this.temporary = temporary;
	}

	/**
	 * Answer the whether the receiver represents a temporary focus change.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a boolean	<code>true</code> if the focus state
	 * 						of the receiver is temporary
	 *						<code>false</code> otherwise
	 */
	public boolean isTemporary() {
		return temporary;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case FOCUS_GAINED:
				buffer.append("FOCUS_GAINED"); //$NON-NLS-1$
				break;
			case FOCUS_LOST:
				buffer.append("FOCUS_LOST"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(isTemporary() ? ",temporary" : ",permanent"); //$NON-NLS-1$ //$NON-NLS-2$
		buffer.append(",opposite="); //$NON-NLS-1$
		buffer.append(opposite);
		return buffer.toString();
	}

	// TODO PPro 1.1 - change callers of old ctor to set opposite for 1.1
	public FocusEvent(Component source, int id, boolean temporary, Component opposite) {
		this(source, id, temporary);
		this.opposite = opposite;
	}

	/**
	 * Answer the other component involved in the change of focus.
	 *
	 * @author      IBM
	 * @version		initial
	 */
	public Component getOppositeComponent() {
		return opposite;
	}
}
