package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.ItemSelectable;

/**
 * This class is represent events which model
 * the selection and de-selection of an item.  Source objects which
 * want to create these events will implement the
 * <code>ItemSelectable</code> interface.
 * <p>
 * When an <code>ItemListener</code> is added to a
 * <code>Component</code> using the <code>addItemListener</code>
 * method, item events occurring on that component will be
 * delivered to the item listener via the sending of the
 * <code>itemStateChanged</code> method.

 * @author      IBM
 * @version		initial
 *
 */

public class ItemEvent extends java.awt.AWTEvent {

	static final long serialVersionUID = -608708132447206933L;

	// The item whose state is being changed
	Object item;

	// The state change that occured in the source
	int stateChange;

	public static final int DESELECTED = 2;
	public static final int SELECTED = 1;

	// Input event modifiers
	public static final int ITEM_FIRST = 701;
	public static final int ITEM_LAST = 701;
	public static final int ITEM_STATE_CHANGED = 701;

	/**
	 * Constructs a new <code>ItemEvent</code> using the
	 * specified id.  The receiver will be an event that occured
	 * in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 * @param	item		the item whose state has changed
	 * @param	stateChange	the state change that occured in the source
	 *
	 */
	public ItemEvent(ItemSelectable source, int id, Object item, int stateChange) {
		super(source, id);

		this.item = item;
		this.stateChange = stateChange;
	}

	/**
	 * Answer the item whose state
	 * is being changed.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.Object	the item
	 */
	public Object getItem() {
		return item;
	}

	/**
	 * Answer the source object for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.ItemSelectable	the source
	 *
	 * @see	<code>getSource()</code>
	 */
	public ItemSelectable getItemSelectable() {
		return (ItemSelectable)getSource();
	}

	/**
	 * Answer the state change that occured in the source.
	 *
	 * This will be one of the following:
	 *	<code>DESELECTED</code>
	 *	<code>SELECTED</code>
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	an int	the state change
	 */
	public int getStateChange() {
		return stateChange;
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.lang.String	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case ITEM_STATE_CHANGED:
				buffer.append("ITEM_STATE_CHANGED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",item="); //$NON-NLS-1$
		buffer.append(getItem());
		buffer.append(",stateChange="); //$NON-NLS-1$
		switch (getStateChange()) {
			case SELECTED:
				buffer.append("SELECTED"); //$NON-NLS-1$
				break;
			case DESELECTED:
				buffer.append("DESELECTED"); //$NON-NLS-1$
				break;
			default:
				buffer.append("unknown"); //$NON-NLS-1$
		}
		return buffer.toString();
	}
}
