package java.awt.event;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2005  All Rights Reserved
 */

import java.awt.Window;

/**
 * This class represents events which model a window action
 * occurring on a <code>Window</code>.
 * <p>
 * This event is propagated to all <code>WindowListeners</code> that
 * are registered on the <code>Component</code> using the
 * <code>addWindowListener</code> method, window events occurring
 * on that component will be delivered to the window listener via
 * the sending of one of the following methods:
 * 	- <code>windowActivated</code>
 * 	- <code>windowClosed</code>
 * 	- <code>windowClosing</code>
 * 	- <code>windowDeactivated</code>
 * 	- <code>windowDeiconified</code>
 * 	- <code>windowIconified</code>
 * 	- <code>windowOpened</code>
 *
 * @author      IBM
 * @version		initial
 *
 */
public class WindowEvent extends ComponentEvent {

	static final long serialVersionUID = -1567959133147912127L;

	// Window event ID constants
	public static final int WINDOW_ACTIVATED = 205;
	public static final int WINDOW_CLOSED = 202;
	public static final int WINDOW_CLOSING = 201;
	public static final int WINDOW_DEACTIVATED = 206;
	public static final int WINDOW_DEICONIFIED = 204;
	public static final int WINDOW_FIRST = 200;
	public static final int WINDOW_ICONIFIED = 203;
	public static final int WINDOW_LAST = 208;
	public static final int WINDOW_OPENED = 200;
	public static final int WINDOW_GAINED_FOCUS = 207;
	public static final int WINDOW_LOST_FOCUS = 208;

	Window opposite;

	/**
	 * Constructs a new <code>WindowEvent</code> using the
	 * specified id.  The receiver will be an event that occured
	 * in the specified source object.
	 *
	 * @author      IBM
	 * @version 	initial
	 *
	 * @param	source		the source object of the receiver
	 * @param	id			the id of the receiver
	 *
	 */
	public WindowEvent(java.awt.Window source, int id) {
		super(source, id);
	}

	/**
	 * Answer the source object for the receiver.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a java.awt.Window	the source object for the receiver
	 *
	 * @see	<code>getSource()</code>
	 */
	public Window getWindow() {
		return (Window) getSource();
	}

	/**
	 * Answer a string with the relevant information describing the state
	 * of the receiver.  This will not be <code>null</code>.
	 *
	 * @author      IBM
	 * @version		initial
	 *
	 * @return	a <code>String</code> 	the param string.
	 */
	public String paramString() {
		StringBuffer buffer = new StringBuffer(super.paramString());
		switch (getID()) {
			case WINDOW_OPENED :
				buffer.append("WINDOW_OPENED"); //$NON-NLS-1$
				break;
			case WINDOW_CLOSING :
				buffer.append("WINDOW_CLOSING"); //$NON-NLS-1$
				break;
			case WINDOW_CLOSED :
				buffer.append("WINDOW_CLOSED"); //$NON-NLS-1$
				break;
			case WINDOW_ICONIFIED :
				buffer.append("WINDOW_ICONIFIED"); //$NON-NLS-1$
				break;
			case WINDOW_DEICONIFIED :
				buffer.append("WINDOW_DEICONIFIED"); //$NON-NLS-1$
				break;
			case WINDOW_ACTIVATED :
				buffer.append("WINDOW_ACTIVATED"); //$NON-NLS-1$
				break;
			case WINDOW_DEACTIVATED :
				buffer.append("WINDOW_DEACTIVATED"); //$NON-NLS-1$
				break;
			case WINDOW_GAINED_FOCUS :
				buffer.append("WINDOW_GAINED_FOCUS"); //$NON-NLS-1$
				break;
			case WINDOW_LOST_FOCUS :
				buffer.append("WINDOW_LOST_FOCUS"); //$NON-NLS-1$
				break;
			default :
				buffer.append("unknown"); //$NON-NLS-1$
		}
		buffer.append(",opposite="); //$NON-NLS-1$
		buffer.append(opposite);
		return buffer.toString();
	}

	public WindowEvent(Window source, int id, Window opposite) {
		this(source, id);
		this.opposite = opposite;
	}

	public Window getOppositeWindow() {
		return opposite;
	}
}
